#include "AStarSearchStrategy.hpp"
#include "OutsideCostEstimatorBuilder.hpp"
#include <boost/program_options.hpp>
#include "Path.hpp"
#include "LazyAgenda.hpp"
#include "BestFirstComparator.hpp"

namespace po = boost::program_options;

AStarSearchStrategy::AStarSearchStrategy(const string &configFile) {

  string outsideCostConfigFile;

  // Compile the config file
  po::options_description config("Configuration");
  config.add_options()
	  ("OutsideCostEstimatorConfigFile", po::value<string>(&outsideCostConfigFile), "The config file for the outside cost estimator")
	  ;
  po::variables_map vm;
  iFile ifs(configFile.c_str());
  store(parse_config_file(ifs, config), vm);
  notify(vm);

  outsideCostConfigFile = Path::maybe_compose(configFile, outsideCostConfigFile);
  m_pOutsideCost = OutsideCostEstimatorBuilder::Instance()->build(outsideCostConfigFile);
  if(!m_pOutsideCost)
	THROW("Error: OutsideCostEstimator not built!");

  // The A* algorithm requires best first parsing
  m_pAgenda = new LazyAgenda<BestFirstComparator>();
}

AStarSearchStrategy::~AStarSearchStrategy() {
  if(m_pOutsideCost)
    delete m_pOutsideCost;
}

void AStarSearchStrategy::setGrammar(Grammar *g) {
  assert(g != NULL);
  NaiveWMTG *nwg = dynamic_cast<NaiveWMTG*>(g);
  if(!nwg) {
	cerr << typeid(g).name() << endl;
    THROW("grammar is not a NaiveWMTG");
  }
  m_pOutsideCost->precompute(nwg);
}

void AStarSearchStrategy::initialize(SentenceTuple &sent) {
  AgendaSearchStrategy::initialize(sent);
  m_sent = sent;
  m_pOutsideCost->initialize(sent);
}

void AStarSearchStrategy::addItem(Item *item) {
  // Compute the outside cost before adding to the agenda
  SCORE score = (*m_pOutsideCost)(m_sent, item);
  cout << *item << " score: " << score << endl;
  item->setOutsideScore(score);
  AgendaSearchStrategy::addItem(item);
}

void AStarSearchStrategy::clear() {
  m_pOutsideCost->reset();
  AgendaSearchStrategy::clear();
}

