#include <string>
#include <fstream>
#include "GPDefs.hpp"
#include "Exceptions.hpp"

#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

#include "AbsoluteCountPruningStrategy.hpp"
#include "FactoredItemIndex.hpp"
#include "NaiveItemIndex.hpp"

/*
 * Constructor.
 * @param configFile The config file.
 */
AbsoluteCountPruningStrategy::AbsoluteCountPruningStrategy (const string& configFile) {

  unsigned int arity ;
  string itemIndexType;

  // compile the config file.
  BPO::options_description config("Configuration");
  config.add_options()
    ("ItemIndexType", BPO::value<string>(&itemIndexType), "The type of the item index.")
    ("AbsoluteCount", BPO::value<unsigned int>(&m_itemCount), "The count of items left after pruning.")
    ("Dimensionality", BPO::value<unsigned int>(&arity), "How many components for factored index?")
    ;
  BPO::variables_map vm;
  cerr << "AbsolutePruningStrategy: Compiling config file " << configFile << endl;
  iFile ifs(configFile.c_str());
  store(parse_config_file(ifs, config), vm);
  notify(vm);
  
  if(debug(1)) {
    cerr << "Option values of \"AbsoluteCountPruningStrategy\"" << endl << endl;
    cerr << OptionsPrinter(config, vm) << endl;
  }
  
  if(itemIndexType == ""){
    THROW("Error: The type of item index is not specified for "
	<<"the absolute count pruning strategy!"<<endl);
  }

  string strItemIndexType = tolower(itemIndexType);

  if(strItemIndexType == "factored"){
    // default arity = 2
    m_pItemIndex = new FactoredItemIndex(arity);

  } else if(strItemIndexType == "naive"){
    m_pItemIndex = new NaiveItemIndex();

  } else if(strItemIndexType == "none") {
    m_pItemIndex = NULL;
    if (m_itemCount >= 0) {
      THROW("Error: your threshold is positive, but your item "
	    << "index type is \'none\'." << endl);
    }
  } else {
    THROW("Error: unknown item index type!"<<endl);
  }

}

bool AbsoluteCountPruningStrategy::prunable(const Item* item) const {
  // If the relative pruning threshold is less than 0, then
  // there is no pruning at all.
  if(m_itemCount <= 0){ return false; }

  arity_t d;

  const DSV* span = item->getDSV();

  // the number of times that the factored item index says "prune".
  size_t numBad = 0;
  // the number of factored indexes (should be 1 for terminal/preterminal scan/load items).
  size_t numFactors = min(span->size(), m_pItemIndex->numFactors());

  // retrieves items from the chart according to the span.
  for(d = 0; d < numFactors; ++d){
    vector<Item*> edges;
    // retrieve method also returns the best item f-value
    // in the same index entry, but we don't use it here.
    m_pItemIndex->retrieve(*span, d, edges);
    SCORE itemScore = item->getScore();
    if (debug(3))
      cerr << ": item score = " << itemScore << " ; count = " << edges.size() << endl;
    if(prunable(itemScore, edges)){
      numBad ++;
      if (debug(3)){
	cerr << "Item pruned:" << endl;
	item->dump(cerr);
	cerr << endl;
      }
      return true;
    }
  }
  // If any of the indexes say that this item
  // should be pruned, then we prune it.
  return false;
}

/* This method will simply the vector counting the number of items with a
  better score than the input score. It that number exceeds the absolute
  count threshold, return true. */
bool AbsoluteCountPruningStrategy::prunable(SCORE score, vector<Item*>& edges) const {
  // if the original absolute threshold is less than
  // 0, then there will be no pruning at all.
  if ( m_itemCount <= 0 )
    return false;

  // if the number of items covering a span is lower
  // than the threshold do not prune the item
  if (m_itemCount > edges.size())
    return false;

  return true;
}

/** Inserts a chart index into the item index table.*/
void AbsoluteCountPruningStrategy::memoize(const Item* item) {
  if(m_pItemIndex){
    m_pItemIndex->insert(item);
  }
}

/* Resets the thresholds to the initial values, clears the item index.*/
void AbsoluteCountPruningStrategy::reset() {
  if(m_pItemIndex){
    m_pItemIndex->clear();
  }
}
