#include "BigramGrammar.hpp"
#include "DebugMacros.hpp"
#include "HedgeWithExtrema.hpp"

BigramGrammar::BigramGrammar(const string& configFile){

  string modelFileStem;
  BPO::options_description opt("BigramGrammar options");
  opt.add_options()
    ("model-filestem", BPO::value<string>(&modelFileStem), "")
    ;
  BPO::variables_map vm;
  cerr << "Compiling config file " << configFile << endl;
  iFile ifs(configFile.c_str());
  store(parse_config_file(ifs, opt), vm);
  notify(vm);

  m_bigrams.loadModel(modelFileStem);
}

void BigramGrammar::getPossibleConsForTLink(const TLink&, const DSV&, vector<Item* >& ret) const {
  // vector of substitute items
  vector<Item* > subs;
  // assumes that Cons have been proposed by another grammar higher up the chain of responsibility (pattern)
  for(vector<Item*>::const_iterator it = ret.begin(); it != ret.end(); ++it){
    Hedge* pH = dynamic_cast<Hedge*>(*it);
    arity_t outputDimension = pH->arity() - 1;
    const LexNT* lnt = static_cast<const LexNT*>(pH->getNTLink(outputDimension));
    const Terminal& theTerm = lnt->lex();
    HedgeWithExtrema* hwe = new HedgeWithExtrema(pH, theTerm, theTerm);
    SCORE mixinScore = mixInScoreForTerminal(theTerm);
    // assuming simple addition of costs
    hwe->setInsideScore(pH->getInsideScore() + mixinScore);
    hwe->setModelScore(pH->getModelScore() + mixinScore);
    subs.push_back(hwe);
    delete *it;
  }

  // swap in enhanced items
  ret = subs;
};

SCORE BigramGrammar::mixInScoreForTerminal(const Terminal& theTerm) const {

  DEBUG(5, "Output terminal=" << theTerm << endl);
  bool found = false;
  const string* word = GlobalVocab::s_terminalSet->getWord(theTerm, found);
  assert(found);
  SCORE prob = -(m_bigrams.getLogUniProb(*word));
  return prob;
}

void BigramGrammar::getPossibleConsHeaded(const Item* heir, const Item* depend, 
					  const Pair<bool, PAV>& armnt, vector<HRVIpair>& ret) {

  const PAV& thePAV = armnt.second;
  arity_t outputDimension = thePAV.arity() - 1;
  if (thePAV.hasGap(outputDimension))
    THROW("BigramGrammar not yet equipped to handle discontinuities");

  // assume heir comes first
  const HedgeWithExtrema* left = dynamic_cast<const HedgeWithExtrema*>(heir);
  const HedgeWithExtrema* right = dynamic_cast<const HedgeWithExtrema*>(depend);
  if (! left || ! right) THROW("antecedents have no extrema");

  // swap left and right if depend comes first
  bool heirRoleIsOne = armnt.first;
  const PrecArray& outputPA = thePAV.getPA(outputDimension);
  if((outputPA[0] == 2 && heirRoleIsOne) || (outputPA[0] == 1 && ! heirRoleIsOne)) {
    swap(left, right);
  }

  Terminal leftTerm, rightTerm;
  if (left->getSize(outputDimension) > 0)
    leftTerm = left->getLeftEx();
  else
    leftTerm = right->getLeftEx();

  if (right->getSize(outputDimension) > 0)
    rightTerm = right->getRightEx();
  else
    rightTerm = left->getRightEx();

  bool mixingIn;
  if (left->getSize(outputDimension) > 0 && right->getSize(outputDimension) > 0)
    mixingIn = true;

  // vector of substitute items
  vector<HRVIpair> subs;
  // assumes that consequents have been proposed by another grammar
  // higher up the chain of responsibility (pattern)
  vector<HRVIpair >::const_iterator retIt;
  for(retIt = ret.begin(); retIt != ret.end(); ++retIt){
    Hedge* pConseq = dynamic_cast<Hedge*>(retIt->second);
    HedgeWithExtrema* hwe = new HedgeWithExtrema(pConseq, leftTerm, rightTerm);
    if (mixingIn) {
      SCORE mixinScore = mixInScoreForNonTerminal(leftTerm, rightTerm);
      // assuming simple addition of costs
      hwe->setInsideScore(pConseq->getInsideScore() + mixinScore);
      hwe->setModelScore(pConseq->getModelScore() + mixinScore);
    }
    subs.push_back(HRVIpair(retIt->first, hwe));
    delete retIt->second;
  }

  // swap in enhanced items
  ret = subs;
};

SCORE BigramGrammar::mixInScoreForNonTerminal(const Terminal& leftTerm, const Terminal& rightTerm) const {
  bool found = false;
  const string* leftWord = GlobalVocab::s_terminalSet->getWord(leftTerm, found);
  assert(found);
  const string* rightWord = GlobalVocab::s_terminalSet->getWord(rightTerm, found);
  assert(found);
  DEBUG(5, "extrema = " << leftWord << " and " << rightWord << endl);
  SCORE pRightGivenLeft = -(m_bigrams.getLogProb(*rightWord, *leftWord));
  SCORE pRightUni = -(m_bigrams.getLogUniProb(*rightWord));
  return (pRightGivenLeft - pRightUni);
}
