#include "BottomUpLogic.hpp"
#include "MoreMath.hpp"
#include "Hedge.hpp"
#include "NodeRefHedge.hpp"
#include "HashChart.hpp"
#include "LinearPrecedenceComparator.hpp"
#include <queue>
#include "HeadMTree.hpp"
#include "HeadCompose.hpp"
#include "Scan.hpp"
#include "Exceptions.hpp"
#include "GPDiscChart.hpp"

#include "TerminationTestBuilder.hpp"

#include "IndexedWMTG.hpp"

#include "DebugMacros.hpp"
#include "Path.hpp"

#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

/** Constructor.
 */
BottomUpLogic::
BottomUpLogic(const string& configFile,
	      BSP<BPO::options_description> optionsDescription)
  : _super(configFile, addOptionDescriptionsTo(optionsDescription)), m_pWMTG(NULL) {

  BPO::variables_map vm;
  iFile ifs(configFile.c_str());
  BPO::store(parse_config_file(ifs, *optionsDescription), vm);
  BPO::notify(vm);

  string optTerminationTestConfigFile;
  if(vm.count("TerminationTestConfigFile")){
    optTerminationTestConfigFile = vm["TerminationTestConfigFile"].as<string>();
  }

  string TerminationTestConfigFile = 
    Path::maybe_compose(configFile, optTerminationTestConfigFile);
  LexNT lnt(2,2);
  Nonterminal* l1 = new LexNT(2,2);
  Nonterminal* l2 = new LexNT(2,2);
  NTLink defaultLabel(l1, l2);
  DSV defaultSpan(m_arity);
  m_pGoal = TerminationTestBuilder::Instance()->build(TerminationTestConfigFile, 
						  defaultLabel, defaultSpan);
}

//! Destructor
BottomUpLogic::~BottomUpLogic() { }

//! set the grammar -- override baseclass's method
void BottomUpLogic::setGrammar(Grammar* gr) {
  
  // If it's not null, then check to make sure that it's a wmtg,
  // then build the chart and parsing goal.
  if (WMTG* wmtg = dynamic_cast<WMTG*>(gr)) {
      
    m_pGrammar = wmtg;
    m_pWMTG = wmtg;
  } else
    THROW("BottomUpLogic requires a grammar that implements the WMTG interface!" << endl);

  // Build the chart
  //  if (!m_spMainChart) {
  //#ifdef DISC_CHART
  //  if ( m_pWMTG->fanout() > this->arity()) {
  //    int spansPerDim = (m_pWMTG->fanout() / this->arity());
  //    if(debug(1))
  //      cout<<"Building chart with max spans per dimension: "<<spansPerDim<<endl;
  //    m_spMainChart.reset(new HedgeDChart<Hedge>( this->inputArity(), spansPerDim));
  //  }
  //  else
  //#endif
  //    m_spMainChart.reset(new HedgeChart(arity()));
  //}

  // construct the default goal. This has to be done here because the
  // grammar isn't initialized in the logic's constructor.  A better
  // approach might be to make the rootlabel() method static, so that
  // it can be called in the constructor.
  if (!m_pGoal)
    THROW("No parsing goal in BottomUpLogic" << endl);
  m_pGoal->redefine(m_pWMTG->rootLabel());

}

/** Backup the working chart and then clear its content. */
void BottomUpLogic::saveCharts() {
  m_savedCharts.push_back(m_spMainChart);
  m_spMainChart.reset(m_spMainChart->memorylessClone());
}

// modified to compute compatible arrangements for trigger by
// consulting grammar, then pass these compatible arrangements on to
// the chart.
//! store the item in the chart
/// param triggerItem currently passed in by reference to pointer, so that
/// a new item can be swapped for the original item and the pointer
/// updated to point to the new item.  That way, new inferences will
/// always have antecedents that are in the chart.
MemoResult BottomUpLogic::memoize(Item* & triggerItem) {

  MemoResult memoResult;

  if(Hedge* he = dynamic_cast<Hedge*>(triggerItem)){
    static BSP<set<Pair<bool,PAV> > > arnmts(new set<Pair<bool,PAV> >);
    arnmts->clear();
    // if we are going to index this item, we need to obtain
    // the compatible Arrangements.
    m_pWMTG->getArrangementFor(arnmts, triggerItem);

    // insert the item into the chart
    m_spMainChart->insert(he, memoResult, *arnmts);

    // the pruner does its memoization here too if necessary
    if(m_pPruner && memoResult == INSERT){
      // update he, cuz triggerItem was swapped
      he = dynamic_cast<Hedge*>(triggerItem);
      if(m_pPruner->debug(3))
	cerr << "Pruner considering memoizing " << *he;
      // don't memoize Scan consequents
      if (dynamic_cast<Compose*>(he -> getInference().get())) {
	// Note that pruner need not create a copy, cuz at this point
	// triggerItem is a pointer to an item in the chart.
	m_pPruner->memoize(triggerItem);
	if(m_pPruner->debug(3))
	  cerr << "memoized it!" << endl;
      } else
	if(m_pPruner->debug(3))
	  cerr << endl;
    }
  }
  else {
    THROW("Problem in BottomUpLogic::memoize. Item is not a Hedge." << endl);
  }

  //NB: We create two copies of each Item. One in the grammar,
  //and another when we insert it in the chart, because there's
  //no other way to put something in a hash bucket. The following code
  //deletes the first copy whenever a new copy is created.
  return memoResult;
}

void BottomUpLogic::genScans(arity_t inDim, vector<Item*>& itemVec) {
  unsigned short sz = m_sent[inDim].size();
  for(unsigned short i = 0; i < sz; ++i){
    scanOneWord(inDim, i, itemVec);
    Item::incrementBatchID();
  }
}
 
void BottomUpLogic::scanOneWord(arity_t inDim, size_t position, vector<Item*>& itemVec) {
  // generate the TLink
  TLink tlink(m_arity);
  // initialize by setting all components to UNDEF/UNDEF
  for(arity_t d1 = 0; d1 < m_arity; ++d1)
    tlink[d1] = UNDEF_T;
  // then reset the component that matters
  tlink[inDim] = m_sent[inDim][position];

  // generate the DSV
  DSV aDSV(m_arity);
  aDSV[inDim] = DSpan(1);
  aDSV[inDim][0].first = position;
  aDSV[inDim][0].second = position+1;

  if(debug(3)){
    cerr << "Scanned TLink: " << tlink << "with DSV " << aDSV << endl;
  }

  genScansHelper(tlink, aDSV, itemVec);
}

// generate the Scans that have this TLink & DSpan as antecedent
void BottomUpLogic::genScansHelper(TLink& tlk, DSV& span, vector<Item*>& itemVec) {

  // first create the items
  vector<Item*> tags;
  m_pWMTG->getUnaryCons(tlk, span, tags);

  // Generate the inferences
  vector<Item*>::const_iterator iter;
  for(iter = tags.begin(); iter!= tags.end(); iter++){
    // Even though the sc variable isn't used anywhere, the call
    // to the Scan constructor sets the pointer from the Hedge
    // back to the Scan inference. The current code will segfault
    // otherwise.
    Scan* sc = new Scan(static_cast<Hedge*>(*iter), tlk);
    sc = sc; // duck compiler warning
    itemVec.push_back(*iter);
  }
}

void BottomUpLogic::genNewInfs(const Item* trigger, vector<Item*>& allNewItems) {
  allNewItems.clear();

  // if the trigger is NULL, then we generate scans
  if(!trigger){
    if(debug(3))
      dout() << "BUL "<< m_uniqueID << " generating Scan inferences" << endl;

    vector<Item*> someNewItems;
    vector<Item*>::iterator ii;
    for(arity_t inDim = 0; inDim < m_sent.arity(); ++inDim) {
      someNewItems.clear();
      genScans(inDim, someNewItems);
      //       // If this logic is being used as the supervising logic for
      //       // learning, then we hide the DSVs on the components that we
      //       // want to hide from the learning logic.
      //       if (inDim > m_inputArity) {
      // 	DSV emptyDSV;
      // 	Item* pI;
      // 	for(ii = someNewItems.begin(); ii != someNewItems.end(); ++ii) {
      // 	  pI = *ii;
      // 	  Vector<DSV>* theDSVV = new Vector<DSV>(1);
      // 	  DSV hiddenDSV = *(pI->getDSV());
      // 	  (*theDSVV)[0] = hiddenDSV;
      // 	  pI->setHiddenDSVs(theDSVV);
      // 	  pI->setDSV(emptyDSV);
      // 	}
      //       }
      allNewItems.insert(allNewItems.end(), someNewItems.begin(), someNewItems.end());
    }

    if(debug(3))
      dout() << "Generated " << allNewItems.size() << " Scans" << endl;

  } else {

    const Hedge* hedge = static_cast<const Hedge*>(trigger);

    if(debug(3)){
      dout() << "BUL "<< m_uniqueID << " generating Compose inferences for trigger "
	     << *hedge << endl;
    }

    genComposes(hedge, allNewItems);

    if(debug(3)){
      dout()<<allNewItems.size()<<" inferences generated by BottomUpLogic." << endl;
    }
  }
}

//! Generates Composes inferences.
void BottomUpLogic::
genComposes(const Hedge* trigger, vector<Item*>& itemVec) {
  // Get, from the grammar, the arrangements that are compatible with
  // the trigger.
  static BSP<set<Pair<bool,PAV> > > arnmts(new set<Pair<bool,PAV> >);
  arnmts->clear();
  m_pWMTG->getArrangementFor(arnmts, trigger);

  // for each arrangement
  DEBUG(5, "<LOOP>" << endl);
  set< Pair<bool,PAV> >::iterator it_arnmt;
  for(it_arnmt = arnmts->begin(); it_arnmt != arnmts->end(); ++it_arnmt) {

    if(debug(5))
      dout() << "BUL::genComposes for arrangement: " << *it_arnmt << " ";

    // find the matching items from the chart for this trigger and PAV
    vector<Item*> matchingItems;
    matchItem(trigger, *it_arnmt, matchingItems);
    m_numMatches+=matchingItems.size();

    // for each matching item
    vector<Item*>::const_iterator it;
    static google::sparse_hash_set<Item*> seenItems;
    for(it = matchingItems.begin(); it != matchingItems.end(); ++it){
      if(seenItems.find(*it) == seenItems.end()){
	m_numUniqueMatches++;
	processMatchingItem(trigger, *it, *it_arnmt, itemVec);
	seenItems.insert(*it);
      }
    }
    seenItems.clear();
  }
  DEBUG(5, "</LOOP>" << endl);
  Item::incrementBatchID();
}

//! process trigger and one matching item: generate an inference for them
void BottomUpLogic::processMatchingItem(const Hedge* trigger, Item* matchingItem,
					const Pair<bool,PAV>& arrangement, vector<Item*>& itemVec) {
  static Log<SCORE> logfun(LOGBASE);
  if (debug(5)) {
    Hedge* matcher = static_cast<Hedge*>(matchingItem);
    cerr << "FOUND matchItem: " << matcher->getNTLink()
	 << " with span=" << *(matcher->getDSV()) << endl;
  }

  vector<Inference*> vecInf;
  m_pWMTG->getInferences(trigger, matchingItem, arrangement, vecInf);
  for(unsigned int i = 0; i < vecInf.size(); i++){
    // compute outside score of consequent
    Hedge* conseq = static_cast<Hedge*>(vecInf[i]->getConseq());
     
    //SCORE oscore = outsideCostEstimator()->operator()(snt(), conseq);
    //if(oscore < POSINF && conseq->getScore() < POSINF){
    if(conseq->getScore() < POSINF){
      //conseq->setOutsideScore(oscore);
      itemVec.push_back(conseq);
    }
    else{
      delete conseq;
      // deleting item should automatically delete BSP<Inference>
      //      delete vecInf[i];
    }
  }
}

/*! Exports a parse forest of subtrees, each subtree being associated
 * with its inside score.
 *
 * The union of the yields of the subtrees will cover the
 * entire sentence tuple, and there is no crossing brackets
 * among the yields of the subtrees.
 *
 * This method is usually called if there is no complete
 * parse obtained.
 */
inline WTreeVec BottomUpLogic::exportParses() {
  WTreeVec ret;
  Vector<const Item*> items;
  Vector<const Item*> itemsV;
  Vector<const Item*> items1Dtarg;
  
  exportMaxCover(items);

  //items come back sorted by span size, so we re-sort by position
  Vector<const Item*>::iterator sit;

  // \todo: shouldn't need a separate vector for this

  //seperate out those that are 1D on target side.  They can be added later.
  // this makes sorting easier.
  for(sit = items.begin(); sit != items.end(); sit++){
    if((*sit)->getDSV()->getFanoutV()[0] != 0){
      itemsV.push_back(*sit);
    }
    else{
      items1Dtarg.push_back(*sit);
    }
  }

  LinearPrecedenceComparator lpc;
  
  //sort main vector and 1D target vector.
  sort(itemsV.begin(), itemsV.end(), lpc);
  sort(items1Dtarg.begin(), items1Dtarg.end(), lpc);
  
  Vector<const Item*>::iterator it;
  Vector<const Item*>::iterator itr;
  Vector<const Item*>::iterator D2it = items1Dtarg.begin();
  queue<short> heirs;
  for(itr = itemsV.begin(); itr != itemsV.end() || D2it !=  items1Dtarg.end(); itr++){
    if(D2it != items1Dtarg.end() && //if target1D is not at the end AND 
       (itr == itemsV.end() ||  // either there are no more in the main set or
	(!(*itr)->getDSpan(1).empty() && //the main span1 is not empty 
	 //and is to the left in the second dimension.
	((*itr)->getDSpan(1)[0].first > (*D2it)->getDSpan(1)[0].second)))){
      //use the 1D target iterator instead.
      it= D2it;
      //adjust the iterators accordingly
      D2it++;
      --itr;
    }else{
      //else its normal so use the main iterator.
      it = itr;
    }
    
    SCORE s = (*it)->getScore();
    HeadMTree* tr = new HeadMTree;
    for(arity_t i = 0; i < arity(); i++)
      heirs.push(i);

    exportParseHelper((const Item*)*it, *tr, tr->begin(), true, heirs);
    ret.push_back(pair<HeadMTree*, SCORE>(tr, s));
  }
  return ret;
}

void BottomUpLogic::exportParseHelper(const Item* item,
				      HeadMTree& tr,
				      HeadMTree::iterator tree_it,
				      bool isHeir,
				      queue<short> heirQ ) {
  HeadMTree::iterator curr_it = tree_it;

  if(item) {
    HeadMTreeNode data(item->arity());
    const Hedge* ahedge =
      dynamic_cast<const Hedge*>(item);

    if(ahedge) {
      // TODO: we want to get rid of Inference::getPAV()
      BGET_PAV(data)=
	ahedge->getInference()->getPAV(); 
      // PAV::undef(ahedge->arity());

      BGET_NT_LABEL(data)=ahedge->getNTLink();
      curr_it = tr.append_child(tree_it, data);

      if(isHeir){
	if(heirQ.size() != 0)
	  tr.setHeirQueue(curr_it,heirQ);
	else
	  tr.setHeir(curr_it);
      }
    }
  }

  BSP<Inference> inf = item->getInference();
  
  if(inf) {
    if( dynamic_cast<Compose*>(inf.get())) {
      vector<const Item*> items = inf->getAnt();

      for(short i = 0; i < (short)items.size(); i++){
	bool foundHeir = false;

	//This queue holds which dimensions curr_it is an heir for.
	queue<short> heirDims;
	for(unsigned int j = 0; j < arity(); j++){
	  if(dynamic_cast<HeadCompose*>(inf.get())){
	    if(((HeadCompose*)inf.get())->heirIndices().getRole(j) == i+1){
	      heirDims.push(j);
	      foundHeir=true;
	    }
	  }
	}
	exportParseHelper((Item*)items[i], tr, curr_it, foundHeir, heirDims);
      }
    } else if (dynamic_cast<Scan*>(inf.get())) {
      //WE HAVE A SCAN. ADD TERMINAL CHILD.
      HeadMTreeNode data2(item->arity());
      Pair<TLink, DSV> pair(((Scan*)inf.get())->getTLink(),*(item->getDSV()));
      BGET_EITHER_LABEL(data2)= pair;
      tr.append_child(curr_it, data2);
    } else
      THROW("Unknown inference type");
  }
}

/** Adds this class' option description to the option description list of the hierarchy */
BSP<BPO::options_description> BottomUpLogic::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  //optionsDescription->add_options()
  // ("SomeBottomUpLogicOption", BPO::value<string>()->default_value("my default"), "test option")
  // ;
  optionsDescription->add_options()
    ("TerminationTestConfigFile", BPO::value<string>(), "config file for the Parsing Goals");
  return optionsDescription;
}

