#ifndef _BottomUpLogic_H_
#define _BottomUpLogic_H_

#include <iostream>
#include <set>
#include "Logic.hpp"
#include "Compose.hpp"
#include "WMTG.hpp"
#include <boost/program_options.hpp>
#include "Chart.hpp"

//uncomment DISC_CHART constant to include code for discontinuous charts
//switch to Discontinuous chart by setting these values in config files
//1. Increase FanOut from 2 to 4, or higher if supported (grammar).
// A GPDiscChart is created when FANOUT EXCEEDS ARITY, otherwise
// grammar does not allow gaps and a HashChart is used.
// See BottomUpLogic::setGrammar method.
//2. update PAV file to one that includes PAV's with gaps (grammar)
//3. You may want to turn up debug level, to display chart query results
//The standard HashChart can be used when constant is included,
// just reset above values in config files.
#define DISC_CHART 1

using namespace std;

//! multiparsing logic involving Scan and Compose inferences;
//! uses a WMTG in GCNF
class BottomUpLogic : public Logic {
   typedef Logic _super;

 private:

  //! adds the additional options that the configFile contains
  BSP<BPO::options_description>
  addOptionDescriptionsTo(BSP<BPO::options_description>
			  priorOptions);

 public:

  /// Constructor.
  BottomUpLogic
  (const string& configFile,
   BSP<BPO::options_description> priorOptions =
   BSP<BPO::options_description>(new BPO::options_description())
   );

  //! Destructor.
  virtual ~BottomUpLogic();

  /** Backup the working chart and then clear its content. */
  virtual void saveCharts();

  //! set the grammar -- override baseclass's method
  virtual void setGrammar(Grammar* gr);

  /*! Exports a parse forest of subtrees, each subtree being associated
   * with its inside score.
   *
   * The union of the yields of the subtrees will cover the
   * entire sentence tuple, and there is no crossing brackets
   * among the yields of the subtrees.
   *
   * This method is usually called if there is no complete
   * parse obtained.
   */
  virtual WTreeVec exportParses();

protected: // methods

  //! store the item in the chart
  virtual MemoResult memoize(Item*& triggerItem);

  //! Generates new inferences triggered by the input item.

  virtual void genNewInfs(const Item* triggerItem, vector<Item*>& itemList);

  //! Find matched items for the input item. This version just casts
  //! the reference to a pointer, so that it can be dynamic_casted.
  inline virtual void matchItem(const Item * item,
                                const Pair<bool,PAV>& arrg,
				vector<Item*> & matchedItems) {
    m_spMainChart->matchItem(*item, arrg, matchedItems);
  }

  //! Generate Scan inferences
  virtual void genScans(arity_t inDim, vector<Item*>& itemVec);
  //! generate one Scan inference
  void scanOneWord(arity_t inDim, size_t position, vector<Item*>& itemVec);

  // generate the Scans that have this TLink & DSpan as antecedent
  void genScansHelper(TLink& tlk, DSV& span, vector<Item*>& itemVec);

  //! Generates Compose inferences triggered by the input item.
  virtual void genComposes(const Hedge* item, vector<Item*>& itemVec);

  //! helper function for genComposes
  virtual void processMatchingItem(const Hedge* trigger, Item* matchingItem,
			   const Pair<bool,PAV>& arrangement,
			   vector<Item*>& itemVec);


private: // methods

  //! A helper function to export the parse tree rooted by an item.
  //! \param rootItem : the item leading the parse tree being exported.
  //! \param mTree : the finally exported tree.
  //! \param it : the sync tree iterator to which kids will be added.
  //! \param isHeir : whether the rootItem is heir.
  //! \param heir : queue marking which dimensions this item is an heir in.
  void exportParseHelper(const Item *rootItem,
                         HeadMTree &mTree,
			 HeadMTree::iterator it,
			 bool isHeir,
			 queue<short> heir);

protected: // data

  //! The WMTG. This reference is in addition to the generic grammar
  //reference in the base Logic. This way, the Logic can require a
  //WMTG only when necessary, and use the other pointer at other times.
  WMTG* m_pWMTG;

  // Number of dimensions in the input
  virtual arity_t inputArity() const { return arity(); }

};

#endif
