#include "BottomUpTranslationLogic.hpp"
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

//! constructor
BottomUpTranslationLogic::
BottomUpTranslationLogic(const string& configFile,
			 BSP<BPO::options_description> optionsDescription)
  : _super(configFile, addOptionDescriptionsTo(optionsDescription)) {

  // override parent's default input arity to be one less than the logic's arity
  if (m_inputArity == m_arity)
    --m_inputArity;
}

//! Destructor
BottomUpTranslationLogic::~BottomUpTranslationLogic() { }

bool BottomUpTranslationLogic::genLoads(arity_t outDim, vector<Item*>& itemVec) {

  // dummy, to be used repeatedly
  DSV aDSV(m_arity);

  // create a m_arity-dimensional UNDEF_T terminal
  TLink tlink(m_arity);
  for(arity_t d1 = 0; d1 < m_arity; ++d1)
    tlink[d1] = UNDEF_T;

  //load all terminals on this output component of the grammar
  // work directly on the grammar's terminal set
  StringVocabulary* pTerms = m_pWMTG->terminalSet();
  StringVocabulary::const_iterator svci;
  for(svci = pTerms->begin(); svci != pTerms->end(); ++svci) {
    tlink[outDim] = static_cast<unsigned int>(svci->second);

    // create the items for which this tlink is the antecedent
    genScansHelper(tlink, aDSV, itemVec);
  }

  return true;
}

void BottomUpTranslationLogic::
genNewInfs(const Item* trigger, vector<Item*>& itemVec) {
  itemVec.clear();

  // if the trigger is NULL, then we generate scans and loads
  if(!trigger){
    if(debug(3)){
      dout() << "Generating Load and Scan Inferences from BUTL " << m_uniqueID << endl;
    }

    // the -1 is necessary cuz the sentence component indexes start at 0
    for(arity_t outDim = arity() - 1; outDim > m_inputArity - 1; --outDim) {
      bool succeeded = genLoads(outDim, itemVec);
      if (not succeeded) {
	itemVec.clear();
	return;
      }
    }
    short scanCount = itemVec.size();
    if(debug(3))
      dout() << "Generated " << scanCount << " Loads" << endl;

    for(arity_t inDim = 0; inDim < m_inputArity; ++inDim)
      genScans(inDim, itemVec);
    if(debug(3))
      dout() << "Generated " << itemVec.size() - scanCount << " Scans" << endl;

  } else {
    const Hedge* hedge = static_cast<const Hedge*>(trigger);

    if(debug(3)){
      dout() << "Generating Compose Inferences from BUTL " << m_uniqueID << endl;
    }
    genComposes(hedge, itemVec);
    if(debug(3)){
      dout()<<itemVec.size()<<" inferences generated by BUTL " << m_uniqueID << endl;
    }
  }
}

/** Adds this class' option description
    to the option description list of the hierarchy */
BSP<BPO::options_description>
BottomUpTranslationLogic::addOptionDescriptionsTo(BSP<BPO::options_description>
						  optionsDescription){
  //optionsDescription->add_options() ("BottomUpTranslationLogicTestOption",
  //BPO::value<string>()->default_value("my default"), "test option") ;
  return optionsDescription;
}

