
#include "CSPseudoMTG.hpp"
#include "CSHeadCompose.hpp"

CSPseudoMTG::CSPseudoMTG(arity_t arity, const string& configFile,
			 BSP<BPO::options_description> priorOptions)
  : _grammar(arity),
    _super(arity, configFile, addOptionDescriptionsTo(priorOptions)),
    m_featureDimPriority(0) {

  BPO::variables_map vm;
  cerr << "CSPseudoMTG: Compiling config file " << configFile << endl;
  iFile ifs(configFile.c_str());
  store(parse_config_file(ifs, *priorOptions), vm);
  notify(vm);

  if(this->debug(1)) {
    cerr << "Option values of \"CSPseudoMTG\"" << endl << endl;
    cerr << OptionsPrinter(*priorOptions, vm) << endl;
  }

  if(vm.count("featureDimPriority"))
    m_featureDimPriority = vm["featureDimPriority"].as<unsigned int>();

};

CSPseudoMTG::~CSPseudoMTG() {};

ostream& CSPseudoMTG::put(ostream& o) const {
  o << "CSPseudoMTG" << endl;
  return o;
};

/// given two items, gets the conqequent and returns an inference
/// unlike the superclass, uses context-aware inferences
void CSPseudoMTG::getInferences(const Item* pTrigger, const Item* pMatcher,
				const Pair<bool, PAV>& arment, vector<Inference*>& vInf) {
  // start with ordinary composition
  _super::getInferences(pTrigger, pMatcher, arment, vInf);
  // we're gonna replace all of the _super's inferences with more detailed ones
  vector<Inference*> vInfRevised;

  while (! vInf.empty()) {
    Inference* pInf = vInf.back();
    vInf.pop_back();
    // note that Scans are generated outside of getInferences()
    HeadCompose* pHC = dynamic_cast<HeadCompose*>(pInf);
    if (not pHC)
      THROW("Only HeadComposes allowed by CSPseudoMTG for now");
    Hedge* pConseq = dynamic_cast<Hedge*>(pHC->getConseq());
    if (not pConseq)
      THROW("Couldn't get Hedge consequent from HeadCompose");
    
    // now generate context-aware features pertaining to the consequent
    static WVector<GlobalVocab::FeatureType> featVec;
    featVec.clear();
    getFeats(pConseq, featVec);

    // create context-aware inference
    vector<const Item*> ants(pHC->getAnt());
    HeirRoleVector hrv(pHC->heirIndices());
    // the HC's arnmt might be different from the one passed to getInfs due to PAV normalization
    const Pair<bool, PAV>& newArnmt = pHC->getArnmt();
    // this should automatically delete the previous BSP<Inference> in the conseq
    Inference* pCSHC = new CSHeadCompose(ants[0], ants[1], pConseq, newArnmt, hrv, featVec);
    SCORE infScore = getInfScore(pCSHC);
    pConseq->setModelScore(infScore);
    vInfRevised.push_back(pCSHC);
  }

  // swap in the revisions
  vInf = vInfRevised;
};

void CSPseudoMTG::getFeats(Hedge* pConseq, WVector<GlobalVocab::FeatureType>& featVec) {
  if (m_featureDimPriority == 0) {
    m_monoGrammars[0]->getFeats(pConseq, featVec);
    m_monoGrammars[1]->getFeats(pConseq, featVec);
  } else if (m_featureDimPriority == 1) {
    m_monoGrammars[1]->getFeats(pConseq, featVec);
    m_monoGrammars[0]->getFeats(pConseq, featVec);
  } else
    THROW("features can be extracted from only 2 components for now");
};

SCORE CSPseudoMTG::getInfScore(Inference* pInf) const {
  const CSHeadCompose* pCSHC = dynamic_cast<const CSHeadCompose*>(pInf);
  if (not pCSHC){
    return _super::getInfScore(pInf);
  }
  static WVector<GlobalVocab::FeatureType> featVec = pCSHC->getFeatVec();
  // put the model hook here
  return 0;
}

/** Adds this class' option description to the option description list of the hierarchy */
BSP<BPO::options_description> 
CSPseudoMTG::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  optionsDescription->add_options()
    ("featureDimPriority", BPO::value<unsigned int>(), 
     "from which dimension should features be extracted first?")
    ;
  return optionsDescription;
}
