#include <string>
#include <stdio.h>
#include <stdlib.h>

#include "Path.hpp"
#include "CSTreeConstrainedMTG.hpp"
#include "FeatureGenerator.hpp"
#include "Terminal.hpp"
#include "SynCat.hpp"
#include "LexNT.hpp"
#include "NodeRefHedge.hpp"
#include "HeadCompose.hpp"

/// constructor
CSTreeConstrainedMTG::CSTreeConstrainedMTG(arity_t arity, const string& configFile,
					   BSP<BPO::options_description> priorOptions)
  : _grammar(arity), 
    _super(arity, configFile, addOptionDescriptionsTo(priorOptions)),
    m_pFeatGen(NULL) {

  string optFeatGenConfigFile;

  BPO::variables_map vm;
  cerr << "CSTreeConstrainedMTG: Compiling config file " << configFile << endl;
  iFile ifs(configFile.c_str());
  store(parse_config_file(ifs, *priorOptions), vm);
  notify(vm);

  if(vm.count("FeatGenConfig")) optFeatGenConfigFile = vm["FeatGenConfig"].as<string>();

  if(this->debug(1)) {
    cerr << "Option values of \"CSTreeConstrainedMTG\"" << endl << endl;
    cerr << OptionsPrinter(*priorOptions, vm) << endl;
  }

  if(optFeatGenConfigFile != "") {
    string FGConfigFile = Path::maybe_compose(configFile, optFeatGenConfigFile);
    m_pFeatGen = new FeatureGenerator<PositionTree>(arity, FGConfigFile);
  }
};

CSTreeConstrainedMTG::~CSTreeConstrainedMTG() {};

/**
 * Updates the content of the grammar.
 * @param tuple The sentence tuple.
 * @return If there is an empty parse read, then return false.
 */
bool CSTreeConstrainedMTG::prepareFor(SentenceTuple& tuple) {
  bool ret = _super::prepareFor(tuple) 
    && m_pFeatGen->prepareFor(tuple)
    && m_pFeatGen->prepareFor(m_pRestrictingTree, m_constrainedDim);

  return ret;
};

void CSTreeConstrainedMTG::
getFeats(Hedge* pConseq, WVector<GlobalVocab::FeatureType>& featVec) {

  const BSP<Inference> pInf = pConseq->getInference();
  if (debug(3)) {
    cerr << "CSTCMTG::getFeats on " << *pConseq << endl;
    cerr << " HComp:= ";
    pInf->dump(cerr);
  }
  const vector<const Item*> antVec = pInf->getAnt();
  const NodeRefHedge* pAnt1 = dynamic_cast<const NodeRefHedge*>(antVec[0]);
  const NodeRefHedge* pAnt2 = dynamic_cast<const NodeRefHedge*>(antVec[1]);
  const NodeRefHedge* nrhConseq = dynamic_cast<const NodeRefHedge*>(pConseq);
  if (! nrhConseq || ! pAnt1 || ! pAnt2)
    THROW("item is not a NodeRefHedge");
  const PositionTree::iterator conseqNode = nrhConseq->getNode()[m_constrainedDim];
  const PositionTree::iterator antNode1 = pAnt1->getNode()[m_constrainedDim];
  const PositionTree::iterator antNode2 = pAnt2->getNode()[m_constrainedDim];

  if (m_isHeaded) {
    const BSP<HeadCompose> spHC = static_pointer_cast<HeadCompose>(pInf);
    if (! spHC)
      THROW("Not a HeadCompose");
    HeirRoleVector hrv = spHC->heirIndices();
    role_t theRole = hrv.getRole(m_constrainedDim);

    const DSV& ant1DSV = *(pAnt1->getDSV());
    const DSV& ant2DSV = *(pAnt2->getDSV());

    static ScoreSet<GlobalVocab::FeatureType> featCount;
    featCount.clear();
    // the consequent label
    SynCat sc = pConseq->getNTLink().getNT(m_constrainedDim)->getSynCat();
    const string* pLabel = GlobalVocab::s_synCatSet->getWord(sc);
    string featName =  string("dim") + char('0' + m_constrainedDim) + " label " + *pLabel;
    m_pFeatGen->addOneFeature(featCount, featName, "conseq ");
    // production features, including the PAV
    const PAV& thePAV = pInf->getPAV();
    const NTLink& theLHS = nrhConseq->getNTLink();
    NTLink rhsHead, rhsDep;
    if (theRole == 1) {
      rhsHead = pAnt1->getNTLink();
      rhsDep = pAnt2->getNTLink();
    } else if (theRole == 2) {
      rhsHead = pAnt2->getNTLink();
      rhsDep = pAnt1->getNTLink();
    } else      
      THROW("illegal heir role " << theRole << " on dim " << (int)m_constrainedDim);
      
    m_pFeatGen->getProdFeats(featCount, theLHS, rhsHead, rhsDep, thePAV, "conseq ");

    arity_t otherDim = 1 - m_constrainedDim;
    // for word->null case, use only the 1D item
    // first check for the case where both antecedents are 1D source, i.e. PAV [ 1 2 | ]
    if (pAnt1->getSize(otherDim) == 0 && pAnt2->getSize(otherDim) == 0) {
      // use the non-heir item
      if (theRole == 2)
	m_pFeatGen->getFeats(featCount, antNode1, m_constrainedDim, ant1DSV[m_constrainedDim], "");
      else
	m_pFeatGen->getFeats(featCount, antNode2, m_constrainedDim, ant2DSV[m_constrainedDim], "");
    } else if (pAnt1->getSize(otherDim) == 0)
      m_pFeatGen->getFeats(featCount, antNode1, m_constrainedDim, ant1DSV[m_constrainedDim], "");
    else if (pAnt2->getSize(otherDim) == 0)
      m_pFeatGen->getFeats(featCount, antNode2, m_constrainedDim, ant2DSV[m_constrainedDim], "");
    else if (theRole == 1) {
      m_pFeatGen->getFeats(featCount, antNode1, m_constrainedDim, ant1DSV[m_constrainedDim], "heir ");
      m_pFeatGen->getFeats(featCount, antNode2, m_constrainedDim, ant2DSV[m_constrainedDim], "dep ");
    } else if (theRole == 2) {
      m_pFeatGen->getFeats(featCount, antNode2, m_constrainedDim, ant2DSV[m_constrainedDim], "heir ");
      m_pFeatGen->getFeats(featCount, antNode1, m_constrainedDim, ant1DSV[m_constrainedDim], "dep ");
    } else
      THROW("illegal heir role " << theRole << " on dim " << (int)m_constrainedDim);
    if (debug(3))
      cerr << "FeatCount " << featCount << endl;
    featVec.push_back(featCount);
  } else
    THROW("you probably don't want an unheaded MTG here");
}

/** Adds this class' option description to the option description list of the hierarchy */
BSP<BPO::options_description> 
CSTreeConstrainedMTG::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  optionsDescription->add_options()
    ("FeatGenConfig", BPO::value<string>(), "config file for the feature generator")
    ;
  return optionsDescription;
}
