#ifndef _Chart_H_
#define _Chart_H_

#include "GPDefs.hpp"
#include "Item.hpp"
#include "Dbug.hpp"
#include "HeadMTree.hpp"
#include "Logic.hpp"

class Chart : public Dbug {
 public:

  //! Constructor(s)
  Chart(const arity_t& dim, const arity_t& inputArity) 
    : m_arity(dim), m_inputArity(inputArity), m_uniqueID(++sm_chartCount) 
  {}

   //! Destructor(s).
  virtual ~Chart(){}

  //! Clone a Chart of the same type without copying the content.
  virtual Chart* memorylessClone() const = 0;

  //! Returns the language dimensionality.
  arity_t arity() const { return m_arity;}


  //! Finds the matched items for the input Hedge.
  //! \param trigger : the Hedge (trigger) to be matched.
  //! \param arrg : the arrangement to be satisfied by trigger and matches.
  virtual void matchItem(const Item &trigger,
                         const Pair<bool,PAV>& arrg,
		         vector<Item*> &matchedItems) = 0;

  //! Inserts an item, returns the pointer to the inserted item in chart.
  //! \param item : the item to be inserted.
  //! \param insertRes : the result of the insertion.
  //! - INSERT: the item inserted by this method is new;
  //! - UPDATE: there is an item with the same signature already
  //! in chart, and the inside score of the inserted
  //! item is larger than the old one;
  //! - NOEFFECT: there is an item with the same signature
  //! already in chart, and has larger inside score;
  //! - IGNORE: this item is ignored. I.e. because of having
  //! minimum infinity score.
  //! - ERROR : unrecognized item type.
  //! \param arrgs : the arrangements compatible with this item, according to the grammar.
  virtual Item* insert(Item *item,
		       MemoResult &insertRes,
		       const set< Pair<bool,PAV> >& arrgs) = 0;

  /// insertion into the closed list only, with no indexing (and so no matching)
  virtual void insertNoIndex(Item* pItem) {
    THROW("Not implemented");
  }

  //! Get the score of an item. Including both the inside and outside
  //! scores.
  virtual pair<SCORE, SCORE> getIOScores(const Item* item) const = 0;

  //! Check whether a given Item is already on the chart.
  virtual bool derived(const Item* item) const = 0;

  /// export all items in the chart
  virtual void exportAll(Vector<const Item*>& VpI) const = 0;

  /*! \brief Exports a forest of subtrees, each subtree being
   * associated with its inside score.
   *
   * The union of the yields of the subtrees will cover the
   * entire sentence tuple, and there is no crossing brackets
   * among the yields of the subtrees.
   *
   * This method is usually called if there is no complete
   * parse obtained.
   */
  virtual WTreeVec exportParses()
  {THROW("Chart::exporParses() not implemented");}

  //! Export the set of items that maximally cover the
  //! entire span, and the spans of items are not overlapped.
  virtual void exportMaxCover(vector<Item*>& items)
  {THROW("Chart::exportMaxCover() not implemented");}

  //! Clears the content.
  virtual void clear() = 0;

  //! Prints the statistics of this class.
  virtual void printStats(ostream& out) const = 0;

   //! Returns the total number of items stored inside.
  virtual size_t size() const = 0;

protected: // methods
  //! Updates the information of the first argument by the second.
  MemoResult updateScoreAndInference(Item* updated,
				     const Item* updating);

protected: // data

  //! Language dimensionality.
  arity_t m_arity;

  //! number of dimensions for which we shall manage absolute positions
  arity_t m_inputArity;

  /// for diagnostics
  static size_t sm_chartCount; // how many have we constructed so far?
  size_t m_uniqueID;  // for each instance

};

#endif
