#ifndef _CompactAgenda_H_
#define _CompactAgenda_H_

#include <queue>
#include <set>
#include "FHeap.hpp"
#include "Agenda.hpp"
#include "Hedge.hpp"
#include "Exceptions.hpp"

class Item;

using namespace std;

//using namespace boost;

//! A CompactAgenda class never stores duplicate (in the sense
//! of derived item) items. It is only suitable to
//! BestFirstParser.
template<class Comp>
class CompactAgenda :
#ifdef BOOST_MEM_POOL
#include "BoostMemPool.hpp"
  public boost::fibonacci_heap_with_alloc<Item *, Comp>
#else
		public fibonacci_heap<Item *, Comp>
#endif
	      , public Agenda {
#ifdef BOOST_MEM_POOL
  typedef boost::fibonacci_heap_with_alloc<Item *, Comp> _super;
#else
  typedef fibonacci_heap<Item *, Comp> _super;
#endif
 public:

  //! The pointer type used by boost::fibonacci_heap_with_alloc to the value stored.
  typedef typename _super::pointer pointer;

  //! Const iterator.
  typedef typename _super :: const_iterator const_iterator;

  //! begin.
  const_iterator begin() const {
    return _super::begin();
  }

  //! end.
  const_iterator end() const {
    return _super :: end();
  }

  //! Constructor.
  CompactAgenda() : _super() {}

  //! Constructor.
  CompactAgenda(const Comp& comp) : _super(comp)
  {}

  //! Destructor.
  ~CompactAgenda() {
    while(!empty()) {
      Item* inf = _super::top();
      _super :: pop();
      delete inf;
    }
  }

  bool empty() const {
    return _super :: empty();
  }

  //! Clears the content of the Agenda.
  void clear(){
    m_hedges.clear();
    // m_hooks.clear();
    // m_engagedHedges.clear();
  }

  //! Overloads the push interface.
  void push(Item* const& val);

  //! Overloads the pop interface.
  void pop();

  Item *const & top() const{
    return _super :: top();
  }

  size_t size() const{
    return _super::size();
  }

  //! Dumps the content for diagnostics.
  void dump(ostream& out) const {
    out<<"BEGIN of AGENDA:\n";
    const_iterator i;
    for(i = begin(); i != end(); i++){
      //c[i]->getConseq()->dump(out); out<<endl;
      (*i)->dump(cout); cout<<endl;
    }
    out<<"END of AGENDA\n\n";
  }

  //! Returns true if the input item has been in this agenda.
  bool exists(const Item&) const;

 protected:

#ifdef BOOST_MEM_POOL
  typedef hash_map<Hedge, pair<pointer,bool>,
		    hash<Hedge>,
		    equal_to<Hedge>,
		    boost::decorated_fast_pool_allocator<Hedge> > hedge_table_type;

  template<class ItemT>
  void rememberItem(const ItemT& item,
		    hash_map<ItemT, pair<pointer, bool>,
		    hash<ItemT>,
		    equal_to<ItemT>,
		    boost::decorated_fast_pool_allocator<ItemT> >& table);
#else

  typedef hash_map<Hedge, pair<pointer,bool> > hedge_table_type;

  //! Helper function to modify the item hash tables
  //! when an item is pushed into agenda.
  template<class ItemT>
  void rememberItem(const ItemT& item,
		    hash_map<ItemT, pair<pointer, bool> >& table);
#endif

  //! Memoize the hedge consequents of items on agenda.
  hedge_table_type m_hedges;

};

template<class Comp>
void CompactAgenda<Comp>::push(Item* const& val) {
  const Item* derivedItem = val;
  if(const Hedge* hedge = dynamic_cast<const Hedge*>(derivedItem)){
    rememberItem(*hedge, m_hedges);
  } else {
    THROW("ERROR: unrecognized item pushed into CompactAgenda!"<<endl);
  }
}

template< class Comp>
void CompactAgenda<Comp>::pop() {
  const Item* inf = top();
  const Item* derivedItem = inf;

  if(const Hedge* hedge =
     dynamic_cast<const Hedge*>(derivedItem)){
    m_hedges.find(*hedge)->second.second = false;
  }
  _super :: pop();
}

template<class Comp> template<class ItemT>
void CompactAgenda<Comp>::
#ifdef BOOST_MEM_POOL
rememberItem(const ItemT& item,
	     hash_map<ItemT, pair<pointer, bool>,
	     hash<ItemT>,
	     equal_to<ItemT>,
	     boost::decorated_fast_pool_allocator<ItemT> >& table)
#else
  rememberItem(const ItemT& item,
	       hash_map<ItemT, pair<pointer, bool> >& table)
#endif 
{
  Comp compare;

#ifdef BOOST_MEM_POOL
  typename hash_map<ItemT, pair<pointer, bool>,
    hash<ItemT>,
    equal_to<ItemT>,
    boost::decorated_fast_pool_allocator<ItemT> >
    ::iterator iter;
#else
  typename hash_map<ItemT, pair<pointer, bool> >
    ::iterator iter;
#endif
  iter = table.find(item);
  if(iter != table.end()){
    if(iter->second.second){ // duplicate
      if(compare(iter->second.first->data(), (Item*) &item)) {//gt
	Item* tmp = iter->second.first->data();
	increase(iter->second.first, (Item*)&item);
	delete tmp;
      } 
    } else {
      iter->second.second = true;
      iter->second.first = _super :: push((Item*)&item);
    }
  } else {
    table[item].second = true;
    table[item].first = _super :: push((Item*)&item);
  }
}

template<class Comp>
bool CompactAgenda<Comp>::exists(const Item& item) const {
  if(const Hedge* hedge = dynamic_cast<const Hedge*>(&item)){
    //typename hash_map<Hedge, pair<pointer, bool> >::const_iterator it;
    typename hedge_table_type :: const_iterator it;
    it = m_hedges.find(*hedge);
    if(it != m_hedges.end() && it->second.second){
      return true;
    }
  }

  return false;
}

#endif
