#ifndef _COMPOSITEWMTGA_H_
#define _COMPOSITEWMTGA_H_

#include "GrammarMix.hpp"
#include "Hedge.hpp"
#include "WMTG.hpp"
#include "HasW2WDict.hpp"
#include "ScoreDefs.hpp"
#include "Exceptions.hpp"
#include <boost/program_options.hpp>

using namespace std;

/// A composite of 2 WMTGs.  The first one is responsible for
/// generating inference consequents.  The second one can generate
/// additional consequents, rescore the existing ones, or modify the
/// list of consequents in other ways.  E.g., if the 2nd WMTG is just
/// a BigramGrammar (i.e. a target LM), then it will just rescore the
/// existing consequents.
class CompositeWMTGA : public GrammarMix, public WMTG, public HasW2WDict {

  typedef WMTG _super;
  typedef GrammarMix _mixbase;
  typedef Grammar _diamondTop;

  BSP<BPO::options_description>
  addOptionDescriptionsTo(BSP<BPO::options_description> priorOptions);

public:

  CompositeWMTGA(arity_t dim, MixtureType mType = UNIFORM);

  CompositeWMTGA(arity_t dim, const string& configFile,
		 BSP<BPO::options_description> priorOptions =
		 BSP<BPO::options_description>(new BPO::options_description())
		 );

  //! destructor
  virtual ~CompositeWMTGA() {}

  //! add a new grammar to the mixture
  //! all grammars must implement the WMTG interface
  /// N.B.: This method cannot/willnot be used during construction,
  /// because at that time, the _mixbase does not have access to
  /// methods in its subclasses.
  virtual size_t add(Grammar* grammar) {
    WMTG* pWMTG = dynamic_cast<WMTG*>(grammar);
    if(!pWMTG)
      THROW("grammar must be a WMTG");
    return _mixbase::add(grammar);
  }

  virtual void clear();

  //! cost associated with having this item as a root
  virtual SCORE rootCost(const Item* item)const{
    return m_pPrimaryGrammar->rootCost(item) + m_pSecondaryGrammar->rootCost(item);
  }

  virtual void getArrangementFor(BSP<set<Pair<bool,PAV> > >& arrangement, const Item* trigger) const {
    m_pPrimaryGrammar->getArrangementFor(arrangement, trigger);
  }

  virtual void getPossibleConsHeaded(const Item* heir, const Item* depend,
				     const Pair<bool,PAV>& pav, vector<HRVIpair >& parents);

  virtual void getPossibleConsForTLink ( const TLink&, const DSV&, vector<Item*>& ) const;

  virtual SCORE getInfScore(Inference* pInf) const {
    // for now, delegate to secondary grammar only
    return m_pPrimaryGrammar->getInfScore(pInf);
  }

  /// Compile w2w model from the productions, assuming GCNF.
  /// Currently, this ignores links involving a null in the 0th
  /// component.
  virtual BSP<W2WDictionary> getW2WDict() const;

private: // methods

  // this method must be called after construction, because when the
  // _mixbase is being initialized, it does not yet have access to the
  // _self::add method
  void init();

protected: // data

  WMTG* m_pPrimaryGrammar;
  WMTG* m_pSecondaryGrammar;

};

#endif // _WMTGGRAMMARMIX_H_
