#ifndef _WMTGGRAMMARMIX2_H_
#define _WMTGGRAMMARMIX2_H_

#include "GrammarMix.hpp"
#include "Hedge.hpp"
#include "LearnableWMTG.hpp"
#include "HasW2WDict.hpp"
#include "ScoreDefs.hpp"
#include "Exceptions.hpp"
#include <boost/program_options.hpp>

using namespace std;

/// A composite of 2 WMTGs.  The first one is responsible for
/// generating inference consequents.  The second one does all the 
// scoring and everything else. This is used for a tree transducing 
// translator for example.
class CompositeWMTGB : public GrammarMix,  public HasW2WDict, public LearnableWMTG {

  typedef LearnableWMTG _super;
  typedef GrammarMix _mixbase;
  typedef Grammar _diamondTop;

  BSP<BPO::options_description>
  addOptionDescriptionsTo(BSP<BPO::options_description> priorOptions);

public:

  CompositeWMTGB(arity_t dim, MixtureType mType = UNIFORM);

  CompositeWMTGB(arity_t dim, const string& configFile,
		 BSP<BPO::options_description> priorOptions =
		 BSP<BPO::options_description>(new BPO::options_description())
		 );

  //! destructor
  virtual ~CompositeWMTGB() {}

  //! add a new grammar to the mixture
  //! all grammars must implement the WMTG interface
  /// N.B.: This method cannot/willnot be used during construction,
  /// because at that time, the _mixbase does not have access to
  /// methods in its subclasses.
  virtual size_t add(Grammar* grammar) {
    WMTG* pWMTG = dynamic_cast<WMTG*>(grammar);
    if(!pWMTG)
      THROW("grammar must be a WMTG");
    return _mixbase::add(grammar);
  }

  virtual void clear();
  
  //! e.g., some grammars use this to rebuild indexes
  virtual void estimate(){
    m_pSecondaryGrammar->estimate();
  }

  //! cost associated with having this item as a root
  virtual SCORE rootCost(const Item* item)const{
    return m_pSecondaryGrammar->rootCost(item);
  }

  virtual void getArrangementFor(BSP<set<Pair<bool,PAV> > >& arrangement, const Item* trigger) const {
    m_pPrimaryGrammar->getArrangementFor(arrangement, trigger);
  }

  virtual void getPossibleConsHeaded(const Item* heir, const Item* depend,
				     const Pair<bool,PAV>& pav, vector<HRVIpair >& parents);

  virtual void getPossibleConsForTLink ( const TLink&, const DSV&, vector<Item*>& ) const;


  virtual SCORE getInfScore(Inference* pInf) const {
    // for now, delegate to secondary grammar only
    return m_pSecondaryGrammar->getInfScore(pInf);
  }

  // test whether a feature exists in the model
  virtual bool knows(Production* pProd) const{
    return m_pSecondaryGrammar->knows(pProd);
  }

  //! Decompose a multitree into productions and increment them in.
  virtual void increment(HeadMTree& h, SCORE s = UNITYSCORE){
   m_pSecondaryGrammar->increment(h,s);
  }
  
  //! Increment a production, propagating to subfeatures
  virtual void increment(Production* p, SCORE s = UNITYSCORE)  {
    m_pSecondaryGrammar->increment(p,s);
  }
  

  virtual void increment(const Grammar* g){
    THROW("CompositeWMTGB::increment not implemented");
    //    m_pSecondaryGrammar->LearnableGrammar::increment(g);
  }

  virtual SCORE mass(){
    return m_pSecondaryGrammar->mass();
  }
  //! create an inference from the given tree node and add it to the set
  virtual void treeNode2infs(Vector<Inference*>& v, const HeadMTree::iterator& h, const DSV& conseqDSV, 
			     const DSV& antDSV1 = DSV(), const DSV& antDSV2 = DSV()){
    m_pSecondaryGrammar->treeNode2infs(v,h,conseqDSV,antDSV1,antDSV2);
    
  }
  
  //! update the score of a production, NOT propagating to subfeatures
  virtual void updateFeatureScore(Production* p, SCORE s){
    m_pSecondaryGrammar->updateFeatureScore(p,s);
    
  }
  
  /// call-back method to simulate virtual template parameter for UpdateRule::trees2gradient()
  /// computes the margin and return its total cost
  virtual SCORE trees2gradient(const vector<HeadMTree::iterator>& supervisorMargin, 
			       const vector<HeadMTree::iterator>& learnerMargin,
			       const vector<HeadMTree::iterator>& intersectionMargin,
			       Grammar* pCurrModel,
			       const UpdateRule* uRule) {
    return m_pSecondaryGrammar->trees2gradient(supervisorMargin, learnerMargin, intersectionMargin, pCurrModel, uRule);
  }

  bool prepareFor(SentenceTuple& tuple) {
    m_pPrimaryGrammar->prepareFor(tuple); 
    m_pSecondaryGrammar->prepareFor(tuple);
    return true;
  }

  /// Compile w2w model from the productions, assuming GCNF.
  /// Currently, this ignores links involving a null in the 0th
  /// component.
  virtual BSP<W2WDictionary> getW2WDict() const;

  istream& get(istream& is){
    return m_pSecondaryGrammar->get(is);
  }
  
  ostream& put(ostream& os)const{
    return m_pSecondaryGrammar->put(os);
  }

  WMTG* getPrimaryGrammar(){
    return  m_pPrimaryGrammar;
  }
  
  LearnableWMTG* getSecondaryGrammar(){
    return  m_pSecondaryGrammar;
  }

private: // methods

  // this method must be called after construction, because when the
  // _mixbase is being initialized, it does not yet have access to the
  // _self::add method
  void init();

protected: // data

  WMTG* m_pPrimaryGrammar;
  LearnableWMTG* m_pSecondaryGrammar;

};

#endif // _WMTGGRAMMARMIX_H_
