#include "DSV.hpp"
#include "NTLink.hpp"
#include "Exceptions.hpp"
#include "Dbug.hpp"
#include "DebugMacros.hpp"
#include "MoreMath.hpp"

//! Empty constructor.
DSV::DSV() {};

//! Constructor.
DSV::DSV(const arity_t& dim) : m_DSV(dim) { }

//! Constructor. Create DSV for lex link
DSV::DSV(const Location& loc ): m_DSV(loc.size()) {
  for(unsigned int i = 0; i < loc.size(); i++) {
    if(loc[i] != MAX_UINT)
      (*this)[i] = DSpan(Span(loc[i], loc[i]+1));
  }
}

//! Construct from a vector of lengths, assuming no gaps
DSV::DSV(vector<short>& sizes) : m_DSV(sizes.size()) {
  for(arity_t i = 0; i < arity(); i++)
    m_DSV[i] = DSpan(Span(0, sizes[i]));
}

//! Copy constructor from a vector.size() (== D) dim Yield.
DSV::DSV(vector<Span>& yieldSpan): m_DSV(yieldSpan.size()) {
  for(arity_t i = 0; i < arity(); i++){
    if(yieldSpan[i].first < yieldSpan[i].second )
      m_DSV[i] = DSpan(yieldSpan[i]);
  }
}

//! Copy constructor from a NTLink.
//! The lexical head is the position of the word in
//! the sentence. When the lex head in one dimension
//! is NULLT, the DSV for that dimension is empty.
DSV::DSV(const NTLink& link) : m_DSV(link.arity()) {
  // THIS IS ONLY USED WITH POSITION TREES.
  // WHICH ARE ALWAYS LEXICALIZED.

  if(!(dynamic_cast<const LexNT*>(link.getNT(0)))){
    THROW("DSV(PNTL) needs LexNT not implemented");
  }

  for(arity_t d = 0; d < arity(); d++)
    if((static_cast<const LexNT*>(link.getNT(d)))->lex() != NULLT){
      Span pe;
      pe.first = (static_cast<const LexNT*>(link.getNT(d)))->lex();
      pe.second = pe.first + 1;
      m_DSV[d] = DSpan(pe);
    }
}

//! Construct from a vector of DSpans.
// Added for testing purposes, may not be necessary
DSV::DSV(Vector<DSpan> dspans) {
  for (unsigned int i = 0; i < dspans.size(); i++)
    m_DSV.push_back(dspans[i]);
}

//! Destructor.
DSV::~DSV() { }

//! Equality test.
bool DSV::operator==(const DSV& other) const {
  if(other.arity() != arity())
    return false;
  for(arity_t i = 0; i < arity(); i++) {
    if(m_DSV[i] != (other.m_DSV)[i]) return false;
  }
  return true;
}

//! Less than test.
bool DSV::operator<(const DSV& other) const {
  return cmp(other) < 0;
}

//! Greater than test.
bool DSV::operator>(const DSV& other) const {
  return !operator<(other);
}

//! Assignment
DSV& DSV::operator=(const DSV& other) {
  // assert(this->arity() == other.arity());
  m_DSV.resize(other.size());
  for(arity_t i = 0; i < arity(); i++){
    m_DSV[i] = other[i];
  }
  return *this;
}

//! For hashing.
size_t DSV::hash() const {
  size_t ret = 0;
  arity_t a = m_DSV.size();
  for(arity_t d = 0; d < a; d++) {
    ret *= 5231;
    ret += m_DSV[d].hash();
  }
  return ret;
}

//! Join operation. If the operation fails(there is overlapped
//! region), then return empty DSV.
DSV& DSV::operator+=(const DSV& other) {
  arity_t aThis = arity();
  arity_t aOther = other.arity();
  arity_t maxAr = max( aThis, aOther);
  for(arity_t i = 0; i < maxAr; i++)
    if (i >= aOther)
      continue; // keep current DSpan
    else if (i >= aThis)
      m_DSV.push_back(other[i]); // substitute other DSpan
    else {
      DSpan aSpan = m_DSV[i] + other[i];
      if(aSpan.size() == 0 &&
	 (m_DSV[i].size() != 0 || other[i].size() != 0)) {
	clear();
	break;
      }
      m_DSV[i] = aSpan;
    }
  return *this;
}

DSV DSV::operator+(const DSV& other) const {
  arity_t aThis = arity();
  arity_t aOther = other.arity();
  arity_t maxAr = max( aThis, aOther);
  DSV ret(maxAr);
  for(arity_t i = 0; i < maxAr; i++)
    if (i >= aOther)
      ret[i] = (*this)[i]; // keep current DSpan
    else if (i >= aThis)
      ret[i] = other[i]; // substitute other DSpan
    else {
      DSpan aSpan = m_DSV[i] + other[i];
      if(aSpan.size() == 0 &&
	 (m_DSV[i].size() != 0 || other[i].size() != 0)) {
	ret.clear();
	break;
      }
      ret[i] = aSpan;
    }
  return ret;
}

//! Remove operation.
DSV& DSV::operator-=(const DSV& other) {
  assert(arity() && arity() == other.arity());
  for(arity_t i =0; i < arity(); i++) {
    DSpan boundList = m_DSV[i] - other[i];
    if(boundList.size() == 0 && !(m_DSV[i] == other[i])){
      clear(); break;
    }
    m_DSV[i] = boundList;
  }
  return *this;
}

DSV DSV::operator-(const DSV& other) const {
  assert(arity() && arity() == other.arity());

  DSV dspan(arity());

  for(arity_t i =0; i < arity(); i++) {
    DSpan boundList = m_DSV[i] - other[i];
    if(boundList.size() == 0 && !(m_DSV[i] == other[i])){
      dspan.clear(); break;
    }
    dspan[i] = boundList;
  }
  return dspan;
}

//! Access the data of a dimension.
const DSpan & DSV::operator[](arity_t d) const {
  return m_DSV[d];
}

//! Non-const version; both are required. Returns the d-th dimension 1D
//! span.
DSpan& DSV::operator[](arity_t d) {
  return m_DSV[d];
}

//! Returns the dimension of this DSV.
size_t DSV::size() const { 
  return m_DSV.size();
}

size_t DSV::getSize() const {
  size_t size = 1;
  bool allzero = true;

  for(arity_t i = 0; i < arity(); i++){
    unsigned int size1D = 0;
    Vector<Span> :: const_iterator iter;
    for(iter = (*this)[i].begin(); iter != (*this)[i].end(); iter++){
      size1D += iter->second - iter->first;
    }
    if(size1D) size *= size1D;
    if(size1D) allzero = false;
  }

  if(allzero) return 0;
  else return size;
}

// checks whether the d-span is empty in the given dimension
bool DSV::empty(const arity_t& d) const {
  if(d >= arity())
    return true;

  return m_DSV[d].empty();
}

//! Returns the length of the span on d'th dimension.
size_t DSV::length(const arity_t& d) const {
  assert(d < arity());

  return (m_DSV[d]).length();
}

//! sum of covered lengths
size_t DSV::getLen() const {
  size_t ret = 0;
  for(arity_t d = 0; d < arity(); ++d){
    ret += length(d);
  }

  return ret;
}

//! Returns the fanout of the DSV.
//! The fanout of a DSV is the maximum number of continuous
//! subspans over all dimensions.
size_t DSV::fanout() const {
  size_t cd = 0;

  for(unsigned int i = 0; i < size(); i++)
    cd += m_DSV[i].size();

  return cd;
}

//! Returns a fanout vector of this DSV.
Vector<unsigned int> DSV::getFanoutV() const {
  Vector<unsigned int> ret(arity());
  for(unsigned int d = 0; d < size(); d++)
    ret[d] = operator[](d).size();
  return ret;
}

//! Returns a fanout vector of this DSV.
vector<unsigned int> DSV::getLengthV() const {
  Vector<unsigned int> ret(arity());
  for(unsigned int d = 0; d < size(); d++)
    ret[d] = length(d);
  return ret;
}

//! Returns the maximum gap over all dimensions.
//! The size of a gap in a DSV is the distance
//! between its left boundary and right boundary.
//! If it is null span in some dimension, the gap
//! for that dimension is 0.
size_t DSV::maxGapSize() const {
  size_t gapSize = 0;

  for(arity_t i = 0; i < arity(); i++){
    if((*this)[i].size() > 1) {
      size_t cur_gap = (*this)[i].gapSize();
      gapSize = (cur_gap > gapSize) ? cur_gap : gapSize;
    }
  }
  return gapSize;
}

//! Judge whether it is empty or empty.
bool DSV::empty() const {
  assert(arity());
  for(arity_t i = 0; i < arity(); i++)
    if(m_DSV[i].size() > 0) return false;
  return true;
}

//! Judges whether the DSV has gap (in at least one dim).
bool DSV::hasGap() const {
  for(arity_t i = 0; i < arity(); i++){
    if((*this)[i].hasGap()) return true;
  }
  return false;
}

//! Judge whether this DSV contains the other.
//! Note that an empty span is always contained by another span,
//! even if it is also an empty span.
bool DSV::contains(const DSV& other) const {

  assert(arity() && other.arity());

  arity_t ar = arity();
  arity_t otherar = other.arity();
  if (ar < otherar)
    return false;
  if (ar > otherar)
    return true;
  for (arity_t i = 0; i < arity() ; i++) {
    if (!((*this)[i].contains(other[i])))
      return false; }
  return true;
}

//! Judge whether this DSV contains the other on the given dim
bool DSV::contains(const arity_t dim, const DSV& other) const {
  assert(arity() >= dim);

  // empty span is always contained in anything
  if (other.arity() < dim)
    return true;

  if ((*this)[dim].contains(other[dim])) return true;

  return false;
}

//! Judge whether the given dimension of this span is contained in the other 1D span
bool DSV::containedIn(const arity_t d, const DSpan& other) const {
  // empty span is always contained in anything
  if (empty(d))
    return true;

  if (((*this)[d]).containedIn(other)) return true;

  return false;
}

//! Judge whether two DSVs are overlapped.
//! Note that an empty DSV is never overlapped with
//! another span, even if it is also an empty span.
bool DSV::overlapped(const DSV& other) const {
  arity_t maxDim = min( arity(), other.arity());
  for(arity_t d = 0; d < maxDim; d++){
    if ((*this)[d].overlapped(other[d])) return true;
  }
  return false;
}

//! Judges overlap between DSV and DSpan.
bool DSV::overlapped(const DSpan& other) const {
  assert(arity());

  for(arity_t d = 0; d < arity(); d++){
    if ((*this)[d].overlapped(other)) return true;
  }
  return false;
}

// //! Decides the symmetric relationship between two DSVs.
// DSV::SpanRelationType DSV::relation(const DSV& other) const{
//   if(this->contains(other) || other.contains(*this)){
//     return CONTAINING;
//   } else if(this->overlapped(other)){
//     return OVERLAPPED;
//   } else {
//     return GAP;
//   }
// }

//! Clears the data inside, but remains the dimension.
void DSV::clear() {
  arity_t dim = arity();
  *this = DSV(dim);
}

//! inserts empty components at the front
void DSV::prependBlanks(const arity_t howmany) {
  m_DSV.insert(m_DSV.begin(), howmany, DSpan()); 
};


//! Orders the DSVs
int DSV::cmp(const DSV& other) const {
  int ret;
  int thisSize = arity();
  int otherSize = other.arity();
  if((ret = MM::cmp(thisSize,otherSize)) !=0) return ret;
  for(arity_t i = 0; i < arity(); i++){
    if((ret = (*this)[i].cmp((*this)[i], other[i])) !=0) return ret ;
  }
  return 0;
}

//! Dimension size
size_t DSV::sizeOfDim(const arity_t d)const{
  return m_DSV[d].size();
}

//! Returns the gaps in the dimension d of this span.
DSV DSV::getGaps(const arity_t d) const {
  assert(d < arity());

  DSV DSV_1d(1);

  DSV_1d.m_DSV[0] = (*this)[d].getGaps();

  return DSV_1d;
}

//! Returns the gaps in the dimension d of this span including
//! the gaps at the start and end of the sentence. Note that if
//! the beginning and end of sentence are given as before or after
//! a gap, respectively, the gap is still included in getGaps.
DSV DSV::getGaps(const arity_t d, const int indfirst,
		 const int indlast) const {
  assert(d < arity());

  DSV DSV_1d(1);

  DSV_1d.m_DSV[0] = (*this)[d].getGaps(indfirst, indlast);

  return DSV_1d;
}

//! In dim d, "other" is at the direction of "this", in other words,
//! the direction is from "this" to "other".
DSV::Direction DSV::to(const DSV& other, const arity_t d) const {
  assert(d < arity());

  DSpan thisSpan = (*this)[d];
  DSpan otherSpan = other[d];

  return (Direction(thisSpan.to(otherSpan)));
}

//! Input. The dimensionality of the span will be automatically
//! detected according to the input format.
istream& operator>>(istream& in, DSV& span) {
  char ch;

  // set dim to be zero.
  span.reset();
  unsigned int dim = 0;
  in >> ws;
  in >> ch; if(in.eof()) return in;
  in >> ws;
  if(ch != '[') {
    THROW(" '[' expected in dspan, got " << ch << endl);
  }

  vector<DSpan> vec(1);

  // enlarges the dimensionality.
  span.setDimity(span.arity() + 1);

  while(1){
    in >> vec[dim];
    if(in.peek() == '|'){
      in >> ch;
      in >> ws;
      vec.resize(vec.size() + 1);
      dim++;
    }
    else if(in.peek() == ']') {
      in >> ch;
      in >> ws;
      break;
    }
  }
  Vector<DSpan> ret(vec.size());
  for(unsigned int i = 0; i < vec.size(); i++)
    ret[i] = vec[i];

  span.m_DSV = ret;
  return in;
}

//! output.
ostream& operator<<(ostream& out, const DSV& span) {

  out<<"[";
  for(arity_t i = 0; i < span.arity(); i++){
    if(i) out<<"|";
    out<<span[i];
  }
  out<<"] ";
  return out;
}

void DSV::print(ostream& out) {
  out<<"< ";
  for(arity_t i = 0; i < arity(); i++){
    Vector<Span> :: const_iterator it;
    for(it = (*this)[i].begin(); it != (*this)[i].end(); it++){
      if(it - (*this)[i].begin() != 0) out<<"_";
      out<<(Span)(*it);
    }
    out<<" , ";
  }
  out<<" >";
}
