#ifndef _DSV_H
#define _DSV_H

#include <vector>
#include <set>
#include <assert.h>
#include <boost/serialization/serialization.hpp>
#include "IOUtil.hpp"
#include "Vector.hpp"
#include "Pair.hpp"
#include "COMPILER-DEPENDENT_HEADERS.hpp"
#include "MoreAlgorithms.hpp"
#include "GPDefs.hpp"
#include "DSpan.hpp"

using namespace std;

class NTLink;

//! what the heck is this? (undocumented)
typedef vector<unsigned short> Location;

class DSV {
public:

  //! Other DSV is at the "Direction" of "this".
  typedef DSpan::Direction Direction;

  //! 1-D continuous boundary
  typedef DSpan::Span Span;

  //! Empty constructor
  DSV();

  //! Constructor.
  DSV(const arity_t& dim) ;

  //! Constructor. Create DSV for lex link
  DSV(const Location& loc );

  //! Construct from a vector of lengths, assuming no gaps
  DSV(vector<short>& sizes) ;

  //! Copy constructor from a vector.size() (== D) dim Yield.
  DSV(vector<Span>& yieldSpan) ;

  /// Copy constructor from a NTLink.
  /// The lexical head is the position of the word in
  /// the sentence. When the lex head in one dimension
  /// is NULLT, the DSV for that dimension is empty.
  /// THIS IS USED ONLY WITH POSITION TREES, WHERE THE NTLINKS ARE
  /// ALWAYS LEXICALIZED with positions.
  DSV(const NTLink& link) ;

  //! Construct from a vector of DSpans.
  // Added for testing purposes, may not be necessary
  DSV(Vector<DSpan> dspans);

  //! Destructor.
  virtual ~DSV() ;

  //! Equality test.
  bool operator==(const DSV& other) const;

  //! Less than test.
  bool operator<(const DSV& other) const;

  //! Greater than test.
  bool operator>(const DSV& other) const;

  //! Assignment
  DSV& operator=(const DSV& other);

  //! For hashing.
  virtual size_t hash() const;

  //! Join operation. If the operation fails(there is overlapped
  //! region), then return empty DSV.
  DSV& operator+=(const DSV& other);

  DSV operator+(const DSV& other) const;

  //! Remove operation.
  DSV& operator-=(const DSV& other);

  DSV operator-(const DSV& other) const;

  //! Access the data of a dimension.
  const DSpan & operator[](arity_t d) const;
  DSpan& operator[](arity_t d);

  //! Returns the dimension of this DSV.
  size_t size() const;

  size_t getSize() const;

  // checks whether the d-span is empty in the given dimension
  bool empty(const arity_t& d) const;

  //! Returns the length of the span on d'th dimension.
  size_t length(const arity_t& d) const;

  //! sum of covered lengths
  size_t getLen() const;

  //! Returns the fanout of the DSV.
  //! The fanout of a DSV is the sum of continuous
  //! subspans over all dimensions.
  size_t fanout() const;

  //! Returns a fanout vector of this DSV.
  Vector<unsigned int> getFanoutV() const;

  //! Returns a length vector of this DSV.
  vector<unsigned int> getLengthV() const;

  //! Returns the maximum gap over all dimensions.
  //! The size of a gap in a DSV is the distance
  //! between its left boundary and right boundary.
  //! If it is null span in some dimension, the gap
  //! for that dimension is 0.
  size_t maxGapSize() const;

  //! Judge whether it is empty or empty.
  bool empty() const;

  //! Judges whether the DSV has gap (in at least one dim).
  bool hasGap() const;

  //! Judge whether this DSV contains the other.
  //! Note that an empty span is always contained by another span,
  //! even if it is also an empty span.
  bool contains(const DSV& other) const;

  bool contains(const arity_t d, const DSV& other) const;

  //! Judge whether the given dimension of this span is contained in the other 1D span
  bool containedIn(const arity_t d, const DSpan& other) const;

  //! Judge whether two DSVs are overlapped.
  //! Note that an empty DSV is never overlapped with
  //! another span, even if it is also an empty span.
  bool overlapped(const DSV& other) const;

  //! Judge whether a DSpan overlaps this DSV
  bool overlapped(const DSpan& other) const;

//   //! Decides the symmetric relationship between two DSVs.
//   SpanRelationType relation(const DSV& other) const;

  //! Clears the data inside, but retains the dimension.
  void clear();

  //! Clears the content and sets the dimensionality to be zero.
  void reset() { m_DSV.clear();}

  //! Returns the arity.
  arity_t arity() const { return (arity_t)(m_DSV.size());}

  //! Sets the arity.
  void setDimity(const arity_t dim) { m_DSV.resize(dim); }

  //! inserts empty components at the front
  void prependBlanks(const arity_t howmany);

  //! Orders the DSV.
  int cmp(const DSV& other) const;

  size_t sizeOfDim(const arity_t d)const;

  /*
   * Some 1-D operations on some certain dimension.
   */

  //! Returns the gaps in the dimension d of this span.
  DSV getGaps(const arity_t d) const;

  //! Returns the gaps in the dimension d of this span including
  //! the gaps at the start and end of the sentence. Note that if
  //! the beginning and end of sentence are given as before or after
  //! a gap, respectively, the gap is still included in getGaps.
  DSV getGaps(const arity_t d, const int indfirst,
	      const int indlast) const;

  //! In dim d, "other" is at the direction of "this", in other words,
  //! the direction is from "this" to "other".
  Direction to(const DSV& other, const arity_t d) const;

  //! Input. The dimensionality of the span will be automatically
  //! detected according to the input format.
  friend istream& operator>>(istream& in, DSV& span);

  //! output.
  friend ostream& operator<<(ostream& out, const DSV& span);

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) { ar & m_DSV; };

  void print(ostream& out);

  void reverse(){ m_DSV.reverse(); }

private:

  Vector<DSpan> m_DSV;

};

namespace HashNS {
  template<>
  class hash<DSV > {
  public:
    size_t operator()(const DSV& dspan) const{
      return dspan.hash();
    }
  };

  template<>
  struct hash<bool> {
    size_t operator() (const bool b) const {
      if(b) return 1;
      else return 0;
    }
  };
}

#endif
