#ifndef _DSpan_H
#define _DSpan_H

#include "Vector.hpp"
#include "Pair.hpp"
#include <cassert>
#include <boost/serialization/serialization.hpp>

using namespace std;

class DSpan {

public:

  //! 1-D boundary
  typedef unsigned short Position;
  /// a segment, usually of the input
  typedef Pair<Position, Position> Span;

  //! Other DSpan is at the "Direction" of "this".
  enum Direction { LEFT, RIGHT, EQUAL, OVERLAP, UNDEF};

  //! Empty constructor.
  DSpan() ;

  //! Construct a DSpan from a vector of Spans.
  DSpan(Vector< Span > pairs) ;

  //! Constructor. Create DSpan from Span
  DSpan(Span bpair) ;

  //! Destructor.
  virtual ~DSpan() ;

  //! Equality test.
  bool operator==(const DSpan& other) const ;

  //! Inequality test.
  bool operator!=(const DSpan& other) const ;

  //! Less than test.
  bool operator<(const DSpan& other) const ;

  //! Greater than test.
  bool operator>(const DSpan& other) const ;

  //! join (concatenate)
  DSpan operator+(const DSpan& other) const ;

  //! remove other from this
  DSpan operator-(const DSpan& other) const ;

  //! For hashing.
  size_t hash() const ;

  //! Assignment
  DSpan& operator=(const DSpan& other) ;

  //! Judge whether it is empty
  bool empty() const ;

  //! output.
  friend ostream& operator<<(ostream& out, const DSpan& span) ;

  //! input.
  friend istream& operator>>(istream& in, DSpan& span);

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) { ar & m_DSpan; };

  //! Judges whether the DSpan has gap.
  bool hasGap() const ;

  //! Returns the gaps in the dspan.
  DSpan getGaps() const ;

  //! Returns the gaps in the dspan including the
  //! gaps at the start and end of the sentence
  DSpan getGaps(const int indfirst, const int indlast) const ;

  //! Judge whether this span contains another span.
  bool contains(const DSpan& other) const ;

  //! Judge whether this span is contained in the other 1D span
  bool containedIn(const DSpan& other) const ;

  /** Judge whether two DSpans are overlapped. Note that an empty DSpan is
      never overlapped with another span, even if it is also an empty
      span. */
  bool overlapped(const DSpan& other) const ;

  //! "Other" is at the direction of "this", in other word,
  //! the direction is from "this" to "other".
  //! Compoares leftmost/lowest boundary of first Span of each DSpan
  Direction to(const DSpan& otherSpan) const ;

  //! Clears the data inside
  void clear() ;

  // ack!  this is wrong!
  //   //! Private binary function to judge whether two
  //   //! boundaries are overlapped.
  //   template<class _Tp>
  //   struct overlap : public binary_function<_Tp, _Tp, bool> {
  //   public:
  //     bool operator()(const _Tp& bd1, const _Tp& bd2) const{
  // 	assert(bd1.second >= bd1.first && bd2.second >= bd2.first);
  //         if((bd1.second <= bd2.first) ||(bd1.first >= bd2.second)) {
  // 	    return false;
  // 	} else {
  // 	    return true;
  // 	}
  //     }
  //   };

  //! Public binary function to judge whether the first boundary
  //! contains the other.
  template<class _Tp>
  struct span_contain: public binary_function<_Tp, _Tp, bool> {
  public:
    bool operator()(const _Tp& bd1, const _Tp& bd2) const{
      assert(bd1.second >= bd1.first && bd2.second >= bd2.first);
      if(bd1.second >= bd2.second && bd1.first <= bd2.first) {
	return true;
      } else {
	return false;
      }
    }
  };

  //! Calculate the size of a dspan (number of distinct pieces)
  unsigned int size() const ;

  //! Calculate the length of a dspan (total number of words)
  size_t length() const ;

  // for simple access from other classes, vector properties on the
  // m_DSpan copied to DSpan
  Span back() const ;

  void resize(unsigned int n) ;

  Span front() const ;

  void push_back(Span bp) ;

  Vector<Span> :: const_iterator begin() const ;

  Vector<Span> :: const_iterator end() const ;

  int cmp(const DSpan& one, const DSpan &other) const ;

  //! Access the data of a dimension.
  Span& operator[](int d) ;
  const Span & operator[](unsigned int d) const ;

  size_t gapSize() const ;

  //given this vector, split the DSpan into many Dspans, where new borders 
  // are inserted at each integer in the vector.
  Vector<DSpan> split(vector<int> vec);

private: // data
  Vector<Span> m_DSpan;

private: // methods
  int cmp(const size_t& x, const size_t& y) const{
    int z = x - y;
    return (z < 0 ? -1 : (z > 0 ? 1 : 0));
  }

  int cmp(const Vector<Span>& x, const Vector<Span>& y) const{
    if(x.size() == 0 && y.size() != 0) return 1;
    else if(x.size() != 0 && y.size() == 0) return -1;
    else if(x.size() != 0 && y.size() != 0){
      size_t x_i = 0, y_i= 0;
      int ret;
      while(x_i != x.size() && y_i != y.size()){
	if((ret = cmp(x[x_i], y[y_i]))!=0) return ret;
	x_i++; y_i++;
      }
      if(x_i != x.size()) return 1;
      else if(y_i != y.size()) return -1;
      else return 0;
    }
    else return 0;
  }

  int cmp(const Span& x, const Span& y) const{
    int ret;
    if((ret = cmp(x.first, y.first)) != 0) return ret;
    else if((ret = cmp(x.second, y.second)) != 0) return ret;
    else return 0;
  }

};

namespace HashNS {
  template<>
  struct hash< DSpan > {
    size_t operator()(const DSpan& e) const{
      return e.hash();
    }

  };
};

#endif
