#include "DimSplitEstimator.hpp"

#include "OptionsPrinter.hpp"
#include "Path.hpp"
#include "GrammarBuilder.hpp"
#include "Chart.hpp"
#include "HeadProduction.hpp"

DimSplitEstimator::DimSplitEstimator(const string& configFile) 
    : _super(configFile) {

  string parserFile;
  string grammarFile;

  // compile the config file.
  BPO::options_description config("Configuration");
  config.add_options()
      ("ParserConfig", BPO::value<string>(&parserFile),
       "The config file for the monolingual parsers.")
      ("GrammarConfig", BPO::value<string>(&grammarFile),
       "The config file for the grammar used by the monolingual parsers.")
      ;
  BPO::variables_map vm;
  cerr << "DimSplitEstimator: Compiling config file "
       << configFile << endl;
  iFile ifs(configFile.c_str());
  store(parse_config_file(ifs, config), vm);
  notify(vm);

  if(debug(1)) {
    cerr << "Option values of \"DimSplitEstimator\"" << endl << endl;
    cerr << OptionsPrinter(config, vm) << endl;
  }

  if(parserFile == "") {
    THROW("parser config file not given");
  }
  if(grammarFile == "") {
    THROW("grammar config file not given");
  }
  m_parserConfigFile = Path::maybe_compose(configFile, parserFile);
  m_grammarConfigFile = Path::maybe_compose(configFile, grammarFile);
}

DimSplitEstimator::~DimSplitEstimator() {
  for(vector<IndexedWMTG*>::iterator itr=m_grammars.begin(); 
      itr!=m_grammars.end(); itr++) {
    // Not deleting this might cause a memory leak, but I DON'T CARE!!!
    // Somewhere the productions are being deleted twice and I can't figure
    // out where.
    /*
    if(*itr)
      delete *itr;
    */
  }

  for(vector<Parser*>::iterator itr=m_parsers.begin(); 
      itr!=m_parsers.end(); itr++) {
    if(*itr)
      delete *itr;
  }

  m_grammars.resize(0);
  m_parsers.resize(0);
}

void DimSplitEstimator::precompute(NaiveWMTG *grammar) {
  m_grammars.resize(0);
  for(size_t i=0; i<grammar->arity(); i++) {
    m_grammars.push_back(new HeadPMTGA(1, m_grammarConfigFile));
    if(!m_grammars[i]) {
      THROW("Could not build Grammar")
    }
    assert(m_grammars[i]->isHeaded());
    /*
    StringVocabulary* newNtSet(1, grammar->ntSet());
    cout << "set size: " << newNtSet->size();
    //StringVocabulary* newNtSet(grammar->ntSet);
    //m_grammars[i]->ntSet() = newNtSet;
    StringVocabulary* newTermSet(1, grammar->terminalSet());
    //StringVocabulary* newTermSet(grammar->terminalSet());
    //m_grammars[i]->terminalSet() = newTermSet;
    */
  }

  // Decompose each production in the grammar into a 
  // set of single dimensional productions
  for(NaiveWMTG::iterator itr=grammar->begin(); itr!=grammar->end(); itr++) {
    // We don't expressly need HeadProductions but this eliminates the 
    // need to figure out what kind of production we're making
    HeadProduction *prd = dynamic_cast<HeadProduction*>(itr->first);
    SCORE cost = itr->second;
    if(prd == NULL) {
      THROW("DimSplitEstimator only works with HeadProductions");
    }

    vector<HeadProduction*> prods(grammar->arity());
    NTLink lhs = prd->lhs();

    Vector<ProdRHSElem> rhs = prd->rhsLinks();
    cout << *prd << endl;
    PAV pav = prd->rhsPerm();
    for(arity_t d=0; d<prd->arity(); d++) {
      NTLink newLhs(1, *(lhs.getNT(d)->clone()));

      // We need the PAV here so we can normalize the rhs
      PAV newPav(prd->rhsPerm().getPA(d));

      // Don't create a completely empty rhs, otherwise wakiness may ensue
      if(newPav.getPA(0).size() == 0)
        continue;

      Vector<ProdRHSElem> newRhs;
      for(Vector<ProdRHSElem>::iterator rhsItr=rhs.begin(); 
          rhsItr!=rhs.end(); rhsItr++) {
        if(Pair<TLink, DSV> *p = boost::get<Pair<TLink, DSV> >(&(*rhsItr))) {
          Vector<Terminal> ts(1, (p->first)[d]);
          TLink newLink(ts);
          //tmpRhs.push_back(newLink);
          newRhs.push_back(newLink);
        } else if(NTLink *p = boost::get<NTLink>(&(*rhsItr))) {
          const LexNT *tmp = dynamic_cast<const LexNT*>((*p).getNT(d));
          if(!tmp)
            THROW("must be a LexNT");
          NTLink newLink(1, *(tmp->clone()));
          //tmpRhs.push_back(newLink);
          newRhs.push_back(newLink);
        } else {
          TLink *p = boost::get<TLink>(&(*rhsItr));
          Vector<Terminal> ts(1, (*p)[d]);
          TLink newLink(ts);
          //tmpRhs.push_back(newLink);
          newRhs.push_back(newLink);
        }
      }

      if(newRhs.size() > 2)
        THROW("new rhs size > 2");

      //newPav = PAV::normalize(newPav);  // make sure to normalize everything
      HeirRoleVector tmpRole = prd->heirRole();
      HeirRoleVector newRole = tmpRole.get1DItem(d);
      if(!newPav.empty()) {
        HeadProduction *newPrd = 
            new HeadProduction(newLhs, newRhs, newPav, newRole);

        // The increment function expects the unloged probabilities
        SCORE unlogCost = exp(-cost);
        cout << *newPrd << endl;
        m_grammars[d]->increment(newPrd, unlogCost);
      }
    }
  }

  // Index all the grammars
  for(arity_t d=0; d<m_grammars.size(); d++) {
    m_grammars[d]->estimate();
  }

  // Create a parser for each dimension
  m_parsers.resize(0);
  for(size_t i=0; i<grammar->arity(); i++) {
    Parser *p = new Parser(m_parserConfigFile, m_grammars[i]);
    if(!p)
      THROW("Parser could not be created");
    //p->setPrintStats(false);
    //p->setGrammar(m_grammars[i]);
    m_parsers.push_back(p);
  }

  m_scores.resize(grammar->arity());
}

void DimSplitEstimator::initialize(SentenceTuple &sent) {
  for(int d=0; d<sent.arity(); d++) {
    m_scores[d] = 0.0;
    m_parsers[d]->clear();
    Vector<Vector<Terminal> > sent1d;
    sent1d.push_back(sent[d]);
    SentenceTuple st(sent1d);
    if(!st.empty()) {
      WTreeVec tree = m_parsers[d]->getParse(st);
      m_scores[d] = 0.0;
      for(WTreeVec::iterator itr=tree.begin(); 
          itr!=tree.end(); itr++) {
        m_scores[d] += itr->second;
      }
      if(tree.size() == 0) {
        cerr << "could not get parse for " << st << endl;
      }
    }
  }
}

SCORE DimSplitEstimator::get1DScore(Parser* parser1D, const Hedge *hedge1D, 
                                    SCORE totalScore) {
  assert(hedge1D->arity() == 1);

  // if the span is empty then it isn't in the tree 
  // and shouldn't be included in the estimate
  if(hedge1D->getDSV()->empty(0))
    return 0.0;

  // Traverse the chart finding the item whose 
  // span exactly matches the given Hedge
  // This is the item in the tree that corresponds 
  // to the given Hedge
  BSP<Chart> chart = parser1D->getLogic()->getChart();
  // The first score is the inside score, the second is the outside score
  SCORE itemScore = (chart->getIOScores(hedge1D)).first;

  if(totalScore - itemScore < 0.0)
    return 0.0;
  else
    return totalScore - itemScore;
}

SCORE DimSplitEstimator::operator()(SentenceTuple &sent, const Item *item) {
  const Hedge *hedge = dynamic_cast<const Hedge*>(item);
  if(!hedge) {
    THROW("item must be a Hedge");
  }
  SCORE cost = 0.0;

  // Combine all the outside costs of the unidimensional parses
  // We want to use the arity of the sentence and not the item
  // since during translation the sentence won't have certain components
  for(int d=0; d<sent.arity(); d++) {
    Item *item1d = item->get1DItem(d);
    Hedge *hedge1d = dynamic_cast<Hedge*>(item1d);
    if(!hedge1d) {
      THROW("could not get a 1d Hedge");
    }
    Parser *parser1d = m_parsers[d];
    cost += get1DScore(parser1d, hedge1d, m_scores[d]);
  }

  return cost;
}

void DimSplitEstimator::reset() {
  // reset all the parsers for the next sentence
  for(vector<Parser*>::iterator itr=m_parsers.begin(); 
      itr!=m_parsers.end(); itr++) {
    Parser *p = (*itr);
    p->clear();
  }
}

