#include "DiscrimLearner.hpp"
#include "OptionsPrinter.hpp"
#include "Path.hpp"
#include "GrammarBuilder.hpp"
#include "LogicBuilder.hpp"
#include "SearchStrategyBuilder.hpp"

/// constructor from a config file, given pre-existing learning grammar
DiscrimLearner::DiscrimLearner(const string& configFile, Grammar* pLG,
			       BSP<BPO::options_description> optionsDescription)
  : _super(configFile, pLG, addOptionDescriptionsTo(optionsDescription)),
  m_pSupervisingGrammar(NULL),
  m_pSupervisingLogic(NULL), 
  m_pSupervisingSearchStrategy(NULL)
{
  string optSGBuilderConfigFile;
  string optSLBuilderConfigFile;
  string optSSSConfigFile;
//   string optLLBuilderConfigFile;
//   string optLSSConfigFile;
  string tbPrefix, modelPrefix;
  string className = "Discriminative Learner";

  // Compile the config file.
  BPO::variables_map vm;
  if(configFile != ""){
    iFile ifs(configFile.c_str());
    store(parse_config_file(ifs, *optionsDescription), vm);
  }
  notify(vm);
  if(debug(1)) {
    cerr << "DiscrimLearner options:" << endl;
    cerr << OptionsPrinter(*optionsDescription, vm) << endl;
  }

  if(vm.count("SupervisingGrammarBuilderConfigFile"))
     optSGBuilderConfigFile = vm["SupervisingGrammarBuilderConfigFile"].as<string>();
  if(vm.count("SupervisingLogicBuilderConfigFile"))
     optSLBuilderConfigFile = vm["SupervisingLogicBuilderConfigFile"].as<string>();
  if(vm.count("SupervisingSearchStrategyConfigFile"))
     optSSSConfigFile = vm["SupervisingSearchStrategyConfigFile"].as<string>();

  if(debug(1)) {
      cerr << "Option values of \"DiscrimLearner\"" << endl << endl;
      cerr << OptionsPrinter(*optionsDescription, vm) << endl;
  }

  // set learning grammar;
  setGrammar(pLG);

  // build the supervising grammar
  string sgBuilderConfigFile = Path::maybe_compose(configFile, 
						   optSGBuilderConfigFile);
  m_pSupervisingGrammar = GrammarBuilder::Instance()->build(sgBuilderConfigFile);
  if(!m_pSupervisingGrammar){
    cerr<<"Error: No Supervising Grammar was built."<<endl;
    exit(EXIT_FAILURE);
  }

  // build the Supervising Logic
  string slBuilderConfigFile = Path::maybe_compose(configFile, optSLBuilderConfigFile);
  m_pSupervisingLogic = LogicBuilder::Instance()->build(slBuilderConfigFile);
  if(!m_pSupervisingLogic){
    cerr<<"Error: No Supervising Logic was built."<<endl;
    exit(EXIT_FAILURE);
  }
  m_pSupervisingLogic->setGrammar(m_pSupervisingGrammar);

  // build the Supervising SearchStrategy
  string sssConfigFile = Path::maybe_compose(configFile, optSSSConfigFile);
  m_pSupervisingSearchStrategy = SearchStrategyBuilder::Instance()->build(sssConfigFile);
  if(!m_pSupervisingSearchStrategy){
    cerr<<"Error: No SupervisingSearchStrategy was built."<<endl;
    exit(EXIT_FAILURE);
  }
}

/// Adds this class's option descriptions to the option description list
/// of the hierarchy
BSP<BPO::options_description>
DiscrimLearner::addOptionDescriptionsTo(BSP<BPO::options_description>
					optionsDescription){
  optionsDescription->add_options()
    ("SupervisingGrammarBuilderConfigFile,g", BPO::value<string>(),
     "The supervising grammarBuilder config file")
    ("SupervisingLogicBuilderConfigFile", BPO::value<string>(), 
     "The configure file of builder for the supervising logic")
    ("SupervisingSearchStrategyConfigFile", BPO::value<string>(), 
     "The configure file of the supervising search strategy")
//     ("LearningLogicBuilderConfigFile", BPO::value<string>(&optLLBuilderConfigFile), 
//      "The configure file of builder for the learning logic")
//     ("LearningSearchStrategyConfigFile", BPO::value<string>(&optLSSConfigFile), 
//        "The configure file of the learning search strategy")
    ;

  return optionsDescription;
}

/// does not delete the learning grammar
DiscrimLearner::~DiscrimLearner() {
  delete m_pSupervisingLogic;
  delete m_pSupervisingSearchStrategy;
  delete m_pSupervisingGrammar;
//   delete m_pLearningLogic;
//   delete m_pLearningSearchStrategy;
}

/// does not clear the learning grammar
void DiscrimLearner::clear() {
  _super::clear();
  m_pSupervisingLogic->clear();
  m_pSupervisingSearchStrategy->clear();
  m_pSupervisingGrammar->clear();
  // The following are cleared in the _super:
//   m_pLearningLogic->clear();
//   m_pLearningSearchStrategy->clear();
}

//! Set the learning grammar for the learner
void DiscrimLearner::setGrammar(Grammar* pGrammar) {
  // another pointer to the base's grammar but with a downcast
  if ((m_pLearningGrammar = dynamic_cast<LearnableWMTG*>(pGrammar)))
    _super::setGrammar(pGrammar);
  else 
    THROW("DiscrimLearner: learnable grammar must be a LearnableWMTG");
};

// void DiscrimLearner::setLearningGrammar(LearnableWMTG* pGrammar) {
//   m_pLearningGrammar = pGrammar;
// }

// LearnableWMTG* DiscrimLearner::getLearningGrammar() const {
//   return m_pLearningGrammar;
// }

void DiscrimLearner::initialize(SentenceTuple& st) {
  _super::initialize(st);
  // clear agenda
  m_pSupervisingSearchStrategy->initialize(st);
  //  m_pLearningSearchStrategy->initialize(st);;

  // clear chart, etc.
  m_pSupervisingLogic->initialize(st);
  //  m_pLearningLogic->initialize(st);
}

/// the weight update method
/// Perform uniform multiclass update, as per Crammer et al.,
/// JMLR'03, p.957.  Subject to the additional constraint that all
/// weights should remain non-negative.
void DiscrimLearner::updateWeights(BSP<Inference> pCorrectPrediction, 
				   list<BSP<Inference> >& errorSet,
				   bool forceCorrectUpdate) {
  Production* pFeatureBag;
  Item* pWrongConseq;
  Item* pCorrectConseq =  pCorrectPrediction->getConseq();
  // "Wrong" inferences might be identical to correct ones, modulo
  // properties that the LG can't see, such as iterators over
  // constraining trees.  So first we delete the ones that are not
  // really wrong in this sense.
  list<BSP<Inference> >::iterator cit;
  for(cit=errorSet.begin(); cit!=errorSet.end(); ++cit) {
    pWrongConseq =  (*cit)->getConseq();
    if (*pWrongConseq == *pCorrectConseq) {
      delete pWrongConseq;
      errorSet.erase(cit);
    }
  }

  if (not errorSet.empty()) {
    // For the rest, update the production cost only, even though the
    // prediction may have been also based on the costs of the
    // antecedents!
    size_t errorSetSize = errorSet.size();
    float updateSize = -1.0 / errorSetSize;
    for(cit=errorSet.begin(); cit != errorSet.end(); ++cit) {
      pWrongConseq = (*cit)->getConseq();
      // predictionScore unnecessary for now
      //  predictionScore =  pCorrectConseq->getModelScore();
      pFeatureBag = m_pLearningGrammar->getFeatureBag(pWrongConseq)->clone();
      m_pLearningGrammar->increment(pFeatureBag, updateSize);
      delete pWrongConseq;
      // do not delete pFeatureBag, because the grammar now has control of it
      //  delete pFeatureBag;
    }
  }

  if (not errorSet.empty()
      or forceCorrectUpdate) {
    // predictionScore unnecessary for now
    // SCORE predictionScore =  pCorrectConseq->getModelScore();
    pFeatureBag = m_pLearningGrammar->getFeatureBag(pCorrectConseq)->clone();
    m_pLearningGrammar->increment(pFeatureBag, 1.0);
    // do not delete pFeatureBag, because the grammar now has control of it
    //  delete pFeatureBag;
  }
}

/// the "main" method
bool DiscrimLearner::learn() {

  TerminationTest::GoalID goalID;
  TerminationTest* pParseGoal = m_pLogic->getGoal();
  // at the beginning, the goal is unachieved
  vector<short> lens = (m_pLogic->snt()).lengths();
  DSV span(lens);
  pParseGoal->reset(span);
  pParseGoal->reset();
  int goalCount = 0;

  vector<Item*> hypothesisList;
  list<BSP<Inference> > errorSet;
  vector<Item*>::const_iterator it;

  // the firstOutDim is the same as the inputArity, cuz the former is
  // 0-indexed, and the latter is 1-indexed.
  arity_t firstOutDim = m_pLogic->inputArity();

  // the item popped off the agenda, possibly empty
  Item* pTrigger = NULL;
  BSP<Inference> pCurrInf;
  vector<Item*> pIVec;
  // foundWayForward is true whenever either logic makes a valid
  // inference.  SLFoundWayForward is true whenever the SL makes a
  // valid inference, which implies that the LL failed to make one.
  bool foundWayForward, SLFoundWayForward;
  bool startedSupervisor = false;
  MemoResult LLMemoResult = IGNORE;
  MemoResult SLMemoResult = IGNORE;

  unsigned int correctHypos = 0;
  unsigned int wrongHypos = 0;
  unsigned int numCorrections = 0;

  while (goalCount == 0) {
    foundWayForward = false;
    SLFoundWayForward = false;

    if (pParseGoal->achieved(pTrigger, m_pLogic, goalID)) {
      if (debug(2))
	cerr << "goal achieved by " << pTrigger;
      ++goalCount;
      break; // quit main loop
    }

    if (debug(3))
      cerr << "LL expanding trigger " << pTrigger;

    LLMemoResult = m_pLogic->expand(pTrigger, hypothesisList);

    switch(LLMemoResult) {
    case IGNORE : {
      delete pTrigger;
      pTrigger = NULL;
      // note: no break -- IGNORE also does enqueue below,
      // in case of initial expansion when there's no trigger
    }
    case INSERT :
    case UPDATE:
    case NOEFFECT:
      // push onto learning open list
      enqueue(m_pSearchStrategy, hypothesisList);
    }

    while ( not m_pSearchStrategy->empty()) {
      pTrigger = m_pSearchStrategy->nextItem();
      pCurrInf = pTrigger->getInference();

      // See if the inference is allowed by the supervising grammar.
      // For full generality, this check should be done by the logic,
      // so that both the ID and LP constraints can be checked.
      // However, for now we assume that the LP constraints are always
      // OK in the input components, so we send it directly to the
      // grammar for checking the ID constraints, bypassing the logic.
      // This design is not great because in fact we also need to
      // check the LP constrains in the output components, which are
      // hidden from the learner, so this really should go through the
      // (supervising) logic.  Anyway, must use the inference cuz the
      // prod doesn't have span info that a PseudoMTG needs to
      // evaluate it.
      if(m_pSupervisingGrammar->allowInf(pCurrInf)) {
	// update trigger, cuz the call to allowInf() might change it
	pTrigger = pCurrInf->getConseq();

	if (debug(3))
	  cerr << "SG allowed it." << endl;

	++correctHypos;
	// Bias the SG to break ties in favor of items proposed by the
	// LG.  Ideally, the SG should be biased further, to prefer
	// acceptable proposals from the LG, even if they cost more
	// than other acceptable proposals.
	//	m_pSupervisingSearchStrategy->addItem(pTrigger);

	foundWayForward = true;
	SLMemoResult = IGNORE; // now it doesn't matter if the SL memoized it
	break; // inner while loop
      } else {

	if (debug(3))
	  cerr << "SG did NOT allow it." << endl;

	++wrongHypos;
	errorSet.push_back(pCurrInf);
	pTrigger = NULL;
      }
    }
    
    // if LG found a way forward
    if (foundWayForward) {
      // make the SL prefer the LL's way
      if (pTrigger->getHiddenDSVs() == NULL) {
	// Some of the LL's hypotheses, such as Scan conseqs, won't have hiddenDSVs.
// 	Item * pTrigger2 = pTrigger->clone();
// 	SLMemoResult = m_pSupervisingLogic->expand(pTrigger2, hypothesisList);
// 	// In this case, the trigger should not be deleted, cuz it'll
// 	// be expanded by the LL.
// 	enqueue(m_pSupervisingSearchStrategy, hypothesisList);
      } else {
	pTrigger->restoreHDSVs(firstOutDim, pIVec);
	while(not pIVec.empty()) {
	  Item * pTrigger2 = pIVec.back(); pIVec.pop_back();
	  SLMemoResult = m_pSupervisingLogic->expand(pTrigger2, hypothesisList);
	  if (SLMemoResult == IGNORE)
	    delete pTrigger2;
	  enqueue(m_pSupervisingSearchStrategy, hypothesisList);
	}
	// post-condition: pIVec is empty
      }
    } else {
      // none of the learningLogic's hypotheses qualified
      if (debug(4))
	cerr << "DL: passing control to SL " << endl;

      // For efficiency, the SL is initialized only if the learner
      // gets lost.
      if (not startedSupervisor) {
	pTrigger = NULL; // just to make sure
	SLMemoResult = m_pSupervisingLogic->expand(pTrigger,hypothesisList);
	enqueue(m_pSupervisingSearchStrategy, hypothesisList);
	startedSupervisor = true;
      }
      while (not foundWayForward 
	     and not m_pSupervisingSearchStrategy->empty()) {
	pTrigger = m_pSupervisingSearchStrategy->nextItem();
	
	if (debug(3))
	  cerr << "SL: expanding new trigger " << *pTrigger;

	// can't rely on LG to expand successfully
	SLMemoResult = m_pSupervisingLogic->expand(pTrigger,hypothesisList);

	switch(SLMemoResult) {
	case IGNORE : {
	  delete pTrigger;
	  pTrigger = NULL;
	  break;
	} 
	case INSERT : {
	  pTrigger = pTrigger->clone();
	  pTrigger->hideDSV(firstOutDim);
	  if ( ! m_pLogic->derived(pTrigger) ) {
	    // let the LG take over again
	    foundWayForward = true;
	    SLFoundWayForward = true;
	    ++numCorrections;
	  }
	}
	case UPDATE:
	case NOEFFECT:
	  break;
	}
	
	// push onto supervisory open list
	enqueue(m_pSupervisingSearchStrategy, hypothesisList);
      }
    }

    if (foundWayForward) {
      // reward pTrigger.inference and/or penalize all e in errorSet
      if (SLFoundWayForward) {
	// force cost decrement for correct features
	updateWeights(pTrigger->getInference(), errorSet, true);
      } else {
	updateWeights(pTrigger->getInference(), errorSet);
      }
      errorSet.clear();
    } else
      // if we're here then both grammars are out of ideas, so fail
      break; // MAIN while

  }
  
  if (debug(1))
    cerr << "goals reached: " << goalCount 
	 << " ; correct hypos: " << correctHypos 
	 << " ; wrong hypos: " << wrongHypos
	 << " ; corrections: " << numCorrections << endl;


  return (goalCount > 0);
}



