#include "GPDefs.hpp"
#include "File.hpp"
#include "Path.hpp"
#include "FeatureGenerator.hpp"
#include "GlobalVocab.hpp"
#include "Exceptions.hpp"
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"
#include "FTreeTransducer.hpp"
#include "PositionTree.hpp"
#include "CSHeadCompose.hpp"

//! Destructor. To clear the memory allocated.
FTreeTransducer::~FTreeTransducer() {
  if (m_pMonoTBStream)
    m_pMonoTBStream->close();
  delete m_pRestrictingTree;
  delete m_pPosFeatGen;
}

FTreeTransducer::FTreeTransducer(arity_t dm, const string& configFile,
				 char** overrideArgV, size_t overrideArgC,
				 BSP<BPO::options_description> optionsDescription)
  : _grammar(dm), 
    _super(dm, configFile, overrideArgV, overrideArgC, addOptionDescriptionsTo(optionsDescription)),
    m_pRestrictingTree(NULL) , m_pPosFeatGen(NULL){
  
  string optMonoTB;
  string optPosFeatGenConfigFile;
  /*
   * Compile the config file.
   */

  BPO::variables_map vm;
  if(overrideArgV){
    if(this->debug(1)){
      cerr<<"FTreeTransducer: Overriding config file arguments..."<<endl;
    }
    store(BPO::command_line_parser(overrideArgC, overrideArgV).
	  options(*optionsDescription).run(), vm);
  }
  if(configFile != ""){
    iFile ifs(configFile.c_str());
    store(parse_config_file(ifs, *optionsDescription), vm);
  }
  notify(vm);

  if(vm.count("TreeStream")) optMonoTB = vm["TreeStream"].as<string>();
  if(vm.count("PosFeatGenConfig")) optPosFeatGenConfigFile = vm["PosFeatGenConfig"].as<string>();

  if(this->debug(1)) {
    cerr << "FTreeTransducer options:" << endl;
    cerr << OptionsPrinter(*optionsDescription, vm) << endl;
  }

  if(optMonoTB == "") {
    THROW("FTreeTransducer needs an inputFile for 1D trees");
  }  
  m_pMonoTBStream = new iFile(optMonoTB.c_str());
  
  if(optPosFeatGenConfigFile != "") {
    string PFGConfigFile = Path::maybe_compose(configFile, optPosFeatGenConfigFile);
    m_pPosFeatGen = new FeatureGenerator<PositionTree>(dm, PFGConfigFile);
  }
}

bool FTreeTransducer::prepareFor(SentenceTuple& tuple) {
  _super::prepareFor(tuple);

  //pass this to the Feature generator 
  m_pPosFeatGen->prepareFor(tuple);

  // get new constraining tree
  PositionTree* parseTree = new PositionTree;
  parseTree->readPositionTree(*(m_pMonoTBStream),1);

  // delete previous parseTree
  if(m_pRestrictingTree)
    delete m_pRestrictingTree;

  m_pRestrictingTree = new PositionTree;
  m_pRestrictingTree->addRoot(*parseTree, *m_pRestrictingTree);

  if((BGET_SPAN(*m_pRestrictingTree->begin()))[0][0].second != tuple[0].size()) {
    THROW("Mismatch in span size of 1D tree and sentence tuples.");
  }

  return true;
};

bool FTreeTransducer::prepareFor(const Vector<HeadMTree*>& treeVec, vector<unsigned int>& startingPosns) {
  return _super::prepareFor(treeVec, startingPosns)
    && m_pPosFeatGen->prepareFor(m_pRestrictingTree);
};

const ScoreSet<GlobalVocab::FeatureType>& FTreeTransducer::
inf2featsHelper(const Hedge* pConH, const Hedge* pAnt1, const Hedge* pAnt2,
		const HeirRoleVector& hrv, const DSV& ant1DSV, const DSV& ant2DSV) const {
  static ScoreSet<GlobalVocab::FeatureType> featCount;
  featCount = _super::inf2featsHelper(pConH, pAnt1, pAnt2, hrv, ant1DSV, ant2DSV);

  //src side has 1D position tree
  const DSpan theDSpan = pConH->getDSpan(0);
  
  //if this node has a span on the source side
  // it might not if the node covers a 1D item on the other dim
  if(theDSpan.size() > 0) {
    static const arity_t constrainedDim = 0;
    static const arity_t otherDim = 1 - constrainedDim;
    DSV dsv(1);
    dsv[constrainedDim] = theDSpan;

    //now run it on PositionTree Node that corresponds to the HeadMTree node
    const PositionTree::iterator treeTop = m_pRestrictingTree->begin();
    PositionTree::iterator ptAntNode1, ptAntNode2;
    if (ant1DSV.empty(constrainedDim))
      ptAntNode1 = treeTop;
    else
      ptAntNode1 = PositionTree::lowestSubsumingNode(treeTop, ant1DSV, constrainedDim);
    if (ant2DSV.empty(constrainedDim))
      ptAntNode2 = treeTop;
    else
      ptAntNode2 = PositionTree::lowestSubsumingNode(treeTop, ant2DSV, constrainedDim);

    role_t theRole = hrv.getRole(constrainedDim);
    // for word->null case, use only the 1D item
    // first check for the case where both antecedents are 1D source, i.e. PAV [ 1 2 | ]
    if (pAnt1->empty(otherDim) && pAnt2->empty(otherDim)) {
      // use the non-heir item
      if (theRole == 2)
	m_pPosFeatGen->getFeats(featCount, ptAntNode1, constrainedDim, ant1DSV[constrainedDim], "");
      else
	m_pPosFeatGen->getFeats(featCount, ptAntNode2, constrainedDim, ant2DSV[constrainedDim], "");
    } else if (pAnt1->empty(otherDim)) // PAV [1 | 2 ] or [ 1 2 | 2 ]
      m_pPosFeatGen->getFeats(featCount, ptAntNode1, constrainedDim, ant1DSV[constrainedDim], "");
    else if (pAnt2->empty(otherDim)) // PAV [ 1 2 | 1 ]
      m_pPosFeatGen->getFeats(featCount, ptAntNode2, constrainedDim, ant2DSV[constrainedDim], "");
    else {	// PAV [1 2 | 1 2] or [1 2 | 2 1]
//       NTLink theLHS = pConH->getNTLink();
//       NTLink rhsHead, rhsDep;
//       PAV thePAV = pConH->getInference()->getPAV();
//       NTLink ntl;
//       TLink t;
      if (theRole == 1) {
// 	ntl = ptAntNode1->getNTLink();
// 	t =  ntl.getTLink();
// 	t[0] =  m_pSentTuple[0][t[0]];
// 	ntl.setTLink(t);
// 	rhsHead =  ntl;

// 	ntl = ptAntNode2->getNTLink();
// 	t =  ntl.getTLink();
// 	t[0] =  m_pSentTuple[0][t[0]];
// 	ntl.setTLink(t);
// 	rhsDep = ntl;

	if(ptAntNode1 != treeTop)
	  m_pPosFeatGen->getFeats(featCount, ptAntNode1, constrainedDim, ant1DSV[constrainedDim], "heir ");
	if(ptAntNode2 != treeTop)
	  m_pPosFeatGen->getFeats(featCount, ptAntNode2, constrainedDim, ant2DSV[constrainedDim], "dep ");
      } 
      else if (theRole == 2) {
// 	ntl = ptAntNode2->getNTLink();
// 	t =  ntl.getTLink();
// 	t[0] =  m_pSentTuple[0][t[0]];
// 	ntl.setTLink(t);
// 	rhsHead =  ntl;
	
// 	ntl = ptAntNode1->getNTLink();
// 	t =  ntl.getTLink();
// 	t[0] =  m_pSentTuple[0][t[0]];
// 	ntl.setTLink(t);
// 	rhsDep =  ntl;

	if(ptAntNode2 != treeTop)
	  m_pPosFeatGen->getFeats(featCount, ptAntNode2, constrainedDim, ant2DSV[constrainedDim], "heir ");
	if(ptAntNode1 != treeTop)
	  m_pPosFeatGen->getFeats(featCount, ptAntNode1, constrainedDim, ant1DSV[constrainedDim], "dep ");
      }
      else
	THROW("illegal heir role " << theRole << " on dim " << constrainedDim);

      // it makes no sense to generate prod features from the position
      // tree, cuz the node structure might not correspond to the
      // binarized multitree
      //      m_pPosFeatGen->getProdFeats(featCount, theLHS, rhsHead, rhsDep, thePAV, "conseq ");
    }
  }

  return featCount;
}

//! Decompose a multitree into inferences in the way specified by the grammar
// assumes conseqNode fanout == 2
const ScoreSet<GlobalVocab::FeatureType>& FTreeTransducer::
treeNode2feats(const HeadMTree::iterator& conseqNode,  
	       const HeadMTree::iterator& hmtAntNode1, const HeadMTree::iterator& hmtAntNode2,
	       const HeirRoleVector& hrv,
	       const DSV& conseqDSV, const DSV& ant1DSV, const DSV& ant2DSV) {
  
  static ScoreSet<GlobalVocab::FeatureType> featCount;
  featCount =   
    _super::treeNode2feats(conseqNode,hmtAntNode1,hmtAntNode2,hrv,conseqDSV,ant1DSV,ant2DSV);

  //src side has 1D position tree
  const DSpan theDSpan = conseqDSV[0];
  
  //if this node has a span on the source side
  // it might not if the node covers a 1D item on the other dim
  if(theDSpan.size() > 0) {
    static const arity_t constrainedDim = 0;
    static const arity_t otherDim = 1 - constrainedDim;
    DSV dsv(1);
    dsv[constrainedDim] = theDSpan;

    //now run it on PositionTree Node that corresponds to the HeadMTree node
    const PositionTree::iterator treeTop = m_pRestrictingTree->begin();
    PositionTree::iterator ptAntNode1, ptAntNode2;
    if (hmtAntNode1->empty(constrainedDim))
      ptAntNode1 = treeTop;
    else
      ptAntNode1 = PositionTree::lowestSubsumingNode(treeTop, ant1DSV, constrainedDim);
    if (hmtAntNode2->empty(constrainedDim))
      ptAntNode2 = treeTop;
    else
      ptAntNode2 = PositionTree::lowestSubsumingNode(treeTop, ant2DSV, constrainedDim);

    HeirRoleVector hrv = HeadMTree::getHRV(conseqNode);
    role_t theRole = hrv.getRole(constrainedDim);
    // for word->null case, use only the 1D item
    // first check for the case where both antecedents are 1D source, i.e. PAV [ 1 2 | ]
    if (hmtAntNode1->empty(otherDim) && hmtAntNode2->empty(otherDim)) {
      if (theRole == 2)
	m_pPosFeatGen->getFeats(featCount, ptAntNode1, constrainedDim, ant1DSV[constrainedDim], "");
      else
	m_pPosFeatGen->getFeats(featCount, ptAntNode2, constrainedDim, ant2DSV[constrainedDim], "");
    } else if (hmtAntNode1->empty(otherDim))
      m_pPosFeatGen->getFeats(featCount, ptAntNode1, constrainedDim, ant1DSV[constrainedDim], "");
    else if (hmtAntNode2->empty(otherDim))
      m_pPosFeatGen->getFeats(featCount, ptAntNode2, constrainedDim, ant2DSV[constrainedDim], "");
    else {
      NTLink theLHS = conseqNode->getNTLink();
      NTLink rhsHead, rhsDep;
      PAV& thePAV = conseqNode->getPAV();
      if (theRole == 1) {
	
	rhsHead = ptAntNode1->getNTLink();
	rhsDep = ptAntNode2->getNTLink();
	if(ptAntNode1 != treeTop)
	  m_pPosFeatGen->getFeats(featCount, ptAntNode1, constrainedDim, ant1DSV[constrainedDim], "heir ");
	if(ptAntNode2 != treeTop)
	  m_pPosFeatGen->getFeats(featCount, ptAntNode2, constrainedDim, ant2DSV[constrainedDim], "dep ");
      } else if (theRole == 2) {
	rhsHead = ptAntNode2->getNTLink();
	rhsDep = ptAntNode1->getNTLink();
	if(ptAntNode2 != treeTop)
	  m_pPosFeatGen->getFeats(featCount, ptAntNode2, constrainedDim, ant2DSV[constrainedDim], "heir ");
	if(ptAntNode1 != treeTop)
	  m_pPosFeatGen->getFeats(featCount, ptAntNode1, constrainedDim, ant1DSV[constrainedDim], "dep ");
      } else
	THROW("illegal heir role " << theRole << " on dim " << constrainedDim);
      
      m_pPosFeatGen->getProdFeats(featCount, theLHS, rhsHead, rhsDep, thePAV, "conseq ");
    }
  }

  return featCount;
};

/** Adds this class' option description to the option description list of the hierarchy */
BSP<BPO::options_description> 
FTreeTransducer::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  optionsDescription->add_options()
    ("TreeStream", BPO::value<string>(), "file containing 1d trees")
    ("PosFeatGenConfig", BPO::value<string>(), "Feature Generator config file")
    ;
  return optionsDescription;
}
