#include "FasterBottomUpTranslationLogic.hpp"

#include "Path.hpp"
#include "Link.hpp"
#include "GlobalVocab.hpp"
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"
#include "Weighted.hpp"
#include "HasW2WDict.hpp"
#include "W2WTM.hpp"
#include "FeatureWMTG.hpp"
#include "CompositeWMTGB.hpp"

//! constructor -- just use the parent class
FasterBottomUpTranslationLogic::FasterBottomUpTranslationLogic
(const string& configFile, BSP<BPO::options_description> optionsDescription) 
  : _super(configFile, addOptionDescriptionsTo(optionsDescription)),
    m_minTranConf(0.0), m_copyUnseen(false), m_maxFertility(10) {

  string optW2WModelFile;

  // Compile the config file.
  BPO::variables_map vm;
  if(configFile != ""){
    iFile ifs(configFile.c_str());
    store(parse_config_file(ifs, *optionsDescription), vm);
  }
  notify(vm);
  if(debug(1)) {
    cerr << "FasterBottomUpTranslationLogic options:" << endl;
    cerr << OptionsPrinter(*optionsDescription, vm) << endl;
  }

  if(vm.count("CopyUnseen")) m_copyUnseen = vm["CopyUnseen"].as<bool>();

  if(vm.count("MinTranConf")) m_minTranConf = vm["MinTranConf"].as<double>();

  if(vm.count("MaxFertility")) m_maxFertility = vm["MaxFertility"].as<unsigned int>();

  // If a w2wTM is passed in, then use it; o.w. it'll be compiled from
  // the grammar in setGrammar().
  if(vm.count("W2WModelFile")) {
    optW2WModelFile = vm["W2WModelFile"].as<string>();

    string w2wModelFile = Path::maybe_compose(configFile, optW2WModelFile);
    iFile modelStream(w2wModelFile);

    ScoreSet<Link<Terminal> > scores;
    //If no vocabs in, integerize the input text.
    if(GlobalVocab::s_noVocabsIn){
      ScoreSet<Link<string> > txtSet;
      modelStream >> txtSet;

      ScoreSet<Link<string> >::iterator i;
      for(i = txtSet.begin(); i != txtSet.end(); i++){
	Terminal t1 = GlobalVocab::s_terminalSet->addWord(i->first.first);
	Terminal t2 = GlobalVocab::s_terminalSet->addWord(i->first.second);
	Link<Terminal> w2w(t1,t2);
	scores.increment(w2w, i->second);
      }
    }
    else{ // else you dont need to integerize
      modelStream >> scores;
    }

    // iterate over the model
    m_spW2WModel.reset( new W2WDictionary);
    ScoreSet<Link<Terminal> >::const_iterator w2wIter;
    for(w2wIter = scores.begin(); w2wIter != scores.end(); ++w2wIter) {
      const Link<Terminal>& tl = w2wIter->first;
      // insert the lexical heads of the LHS into the w2w model
      if (debug(5))
	cerr << "Inserting " << tl << " into w2wTM" << endl;
      m_spW2WModel->insert(tl);
    }

    // add EOSH -> EOSH mapping
    m_spW2WModel->insert(Link<Terminal>(EOSH,EOSH));
  }
}

//! destructor
FasterBottomUpTranslationLogic::~FasterBottomUpTranslationLogic() {}

bool FasterBottomUpTranslationLogic::genLoads(arity_t outDim, vector<Item*>& itemVec) {


  // create a m_arity-dimensional UNDEF_T terminal
  TLink theTLink(m_arity);
  for(arity_t d1 = 0; d1 < m_arity; ++d1){
    theTLink[d1] = UNDEF_T;
  }

  // assuming the input dim is just 0 for now
  Vector<Terminal>& ost = m_sent[0];
  vector<size_t> unseenInputPosns;
  const string* pWord;
  if (debug(2))
    cerr << ost;

  // dummy, to be used repeatedly
  DSV aDSV(m_arity);

  hash_set<Terminal> neededLoad;
  set<Weighted<W2WLink>, less<Weighted<W2WLink> > > currNeededLoad;


  // for each input word
  size_t inLength = ost.size();
  size_t knownWordCount = 0;

  Vector<Terminal> vTrans;
  Vector<Terminal>::iterator vTransIt;
  //some declarations.
  FeatureWMTG* fwmtg;
  set<Weighted<W2WLink>, less<Weighted<W2WLink> > >::const_iterator setci;
  W2WDictionary::const_iterator mmci;
  pair<W2WDictionary::const_iterator, W2WDictionary::const_iterator> transRange;
  Vector<Terminal>::const_iterator ostci;

  // more than one output component currently unsupported
  assert(m_arity - 1 == m_inputArity);


  //check to see what kind of grammar we are dealing with

  fwmtg = dynamic_cast<FeatureWMTG* >(m_pWMTG);
  if(!fwmtg){
    CompositeWMTGB* c = dynamic_cast<CompositeWMTGB* >(m_pWMTG);
    if(c)
      fwmtg = dynamic_cast<FeatureWMTG* >(c->getSecondaryGrammar());
  }
  // set up the w2w model
  BSP<W2WDictionary> spW2WModel; 
  // if we were given one in the config file, then use it
 if(m_spW2WModel) {
    spW2WModel = m_spW2WModel; 
  } else {
    // if available, get a fresh one from the grammar
    const HasW2WDict* hw2wd = dynamic_cast<const HasW2WDict*>(m_pWMTG);
    if (hw2wd) {
      spW2WModel = hw2wd->getW2WDict();
    } //else
      //THROW("No w2w dictionary available to FBUTL!");
  }
  // get access to the w2wTM if it exists
  const W2WTM* pw2wtm = dynamic_cast<const W2WTM*>(m_pWMTG);
  W2WTM::W2WTMType theTM;
  if (pw2wtm)
    theTM = pw2wtm->getW2WTM();

  
  for(size_t posn = 0; posn < inLength; ++posn) {
    Terminal currTerm = ost[posn];
    if (debug(3)) {
      pWord = GlobalVocab::s_terminalSet->getWord(currTerm);
      cerr << "Loading translations of " << currTerm
	   << " ( " << *pWord << " )" << endl;
    };
    currNeededLoad.clear();
    
    if(fwmtg){
      vTrans.clear();
      fwmtg->getPossibleWordTranslations(currTerm, vTrans);
      if (vTrans.begin() == vTrans.end()) {
	if (m_copyUnseen)
	  unseenInputPosns.push_back(posn);
	else if (debug(1)) {
	  pWord = GlobalVocab::s_terminalSet->getWord(currTerm);
	  if (!pWord)
	    cerr << "Warning: No translation found for " << currTerm << endl;
	};
      }
      else{
	++knownWordCount;
	for(vTransIt = vTrans.begin(); vTransIt != vTrans.end(); ++vTransIt) {
	  if (debug(3)) {
	    cerr << *vTransIt << " found!" << endl;
	  }
	  neededLoad.insert(*vTransIt);  
	}
      }
    
    }else{
      
      // find its translation
      transRange = spW2WModel->equal_range(currTerm);
      // if not found
      if (transRange.first == transRange.second) {
	if (m_copyUnseen)
	  unseenInputPosns.push_back(posn);
	else if (debug(1)) {
	  pWord = GlobalVocab::s_terminalSet->getWord(currTerm);
	  if (!pWord)
	    cerr << "Warning: No translation found for " << currTerm << endl;
	};
	
      }else{
	++knownWordCount;
	for(mmci = transRange.first; mmci != transRange.second; ++mmci) {
	  if (pw2wtm) {
	    // add it to the candidate list
	    Terminal tx = mmci->second;
	    W2WLink lk(currTerm, tx);
	    if (debug(5))
	      cerr << "FBUTL " << m_uniqueID << " considering w2wLink " 
		   << currTerm << " , " << tx;
	    ScoreSet<W2WLink>::const_iterator ssci = theTM.find(lk);
	    if (ssci != theTM.end()) {
	      // inserts in order of less comparator
	      currNeededLoad.insert(Weighted<W2WLink>(*ssci));
	      if (debug(5))
		cerr << " with score " << ssci->second;
	    }
	    if (debug(5)) cerr << endl;
	  } else { // just stick in the hashset
	    neededLoad.insert(mmci->second);  
	  }
	}
	if (not currNeededLoad.empty()) {
	  // prune 
	  // iterating through the set will give the translations in order
	  // of increasing cost
	  setci = currNeededLoad.begin();
	  unsigned int fertility = m_maxFertility;
	  while(fertility > 0 && setci != currNeededLoad.end()) {
	    if (debug(5))
	      cerr << "FBUTL " << m_uniqueID << " will load " << setci->first.second << endl;
	    neededLoad.insert(setci->first.second); // 2nd word of the W2WLink
	    ++setci;
	    --fertility;
	  }
	}
      }
    
  
      // estimate translation confidence
      float tranConf = (float) knownWordCount / (float) inLength;
      if (tranConf < m_minTranConf) {
	if (debug(1))
	  cerr << "Translation confidence is only " << tranConf 
	       << " -- aborting." << endl;
	return false;
      }
    }
  }
  short nonNullLoadCount = neededLoad.size();

  if(!fwmtg){
    // load all target words that can vanish on the source side
    transRange = spW2WModel->equal_range(NULLT);
    if (debug(3))
    cerr << "Loading translations of NULL" << endl;
    for(mmci = transRange.first; mmci != transRange.second; ++mmci) {
      // add it to the neededLoad list
      neededLoad.insert(mmci->second);
    }
  }
  else{
    // load all target words that can vanish on the source side
    vTrans.clear();
    fwmtg->getPossibleWordTranslations(NULLT, vTrans);
    if (debug(3))
      cerr << "Loading translations of NULL" << endl;
    for(vTransIt = vTrans.begin(); vTransIt != vTrans.end(); ++vTransIt) {
      if (debug(3)) {
	cerr << *vTransIt << " found!" << endl;
      }
      neededLoad.insert(*vTransIt);  
    }
  }
  

  ///////below here its the same///////
  hash_set<Terminal>::const_iterator hstci;
  for(hstci = neededLoad.begin(); hstci != neededLoad.end(); ++hstci) {
    theTLink[outDim] = *hstci;
    if (debug(3)) {
      pWord = GlobalVocab::s_terminalSet->getWord(*hstci);
      cerr << "Loading " << *hstci << " ( " << *pWord << " )" << endl;
    };

    // create the items for which this tlink is the antecedent
    genScansHelper(theTLink, aDSV, itemVec);
  }

  if (debug(2)) {
    short nullLoadCount = neededLoad.size() - nonNullLoadCount;
    cerr << "Generated " << nullLoadCount + nonNullLoadCount
	 << " Load Antecedents, of which " << nullLoadCount 
	 << " were linked to nothing." << endl;
  }

  if (m_copyUnseen)
    copyUnseenInputs(unseenInputPosns, itemVec);

  return true;
}

void FasterBottomUpTranslationLogic::
copyUnseenInputs(const vector<size_t>& unseenInputPosns, vector<Item*>& ret) {

  // for now, assume input dim is 0 and output dim is 1
  arity_t inDim = 0;
  arity_t outDim = 1;
  // create m_arity-dimensional UNDEF_T terminals
  TLink inTLink(m_arity);
  for(arity_t d1 = 0; d1 < m_arity; ++d1){
    inTLink[d1] = UNDEF_T;
  }
  TLink outTLink(m_arity);
  for(arity_t d1 = 0; d1 < m_arity; ++d1){
    outTLink[d1] = UNDEF_T;
  }

  // prepare the input DSV
  DSV inDSV(m_arity);
  inDSV[inDim] = DSpan(1);

  // dummy output DSV, to be used repeatedly
  DSV outDSV(m_arity);

  vector<Item*> scanCons, loadCons;
  vector<size_t>::const_iterator pci;
  hash_map<TLink,Item*> hmap;
  vector<Item*>::const_iterator iScans, iLoads;

  for(pci = unseenInputPosns.begin(); pci != unseenInputPosns.end(); ++pci) {
    size_t posn = *pci;
    Terminal inTerm = m_sent[inDim][posn];
    // first, send input word to genScansHelper
    inTLink[inDim] = inTerm;
    inDSV[inDim][0].first = posn;
    inDSV[inDim][0].second = posn+1;
    genScansHelper(inTLink, inDSV, scanCons);
    for(iScans = scanCons.begin(); iScans != scanCons.end(); iScans++){
      if(((Hedge*)*iScans)->getNTLink().getNT(0)->getSynCat() != UNKNT)
	return;
    }
    //now, copy the terminal to the output side, and send it to genScansHelper separately
    const string* pWord = GlobalVocab::s_terminalSet->getWord(inTerm);
    Terminal outTerm = GlobalVocab::s_terminalSet->addWord(*pWord);
    outTLink[outDim] = outTerm;
    
    hash_map<TLink,Item*>::iterator hit;
    if((hit = hmap.find(outTLink)) != hmap.end()){
      loadCons.push_back(hit->second);
    }else{
      genScansHelper(outTLink, outDSV, loadCons);
      // insert load conseqs into chart, but don't bother indexing,
      // since they can't match anything anyway
      // a better design will propagate the chart's insertNoIndex method up to logic
      for(iLoads = loadCons.begin(); iLoads != loadCons.end(); ++iLoads) {
	hmap[outTLink] = (*iLoads);
	m_spMainChart->insertNoIndex(*iLoads);
      }
    }
    
    // for each pair in the cross-product of results, send to WMTG::makePretermInference
    // add the conseq of that inference to the return vector
    for(iScans = scanCons.begin(); iScans != scanCons.end(); ++iScans) {
      Item* pItem = *iScans;
      // insert directly into chart to get the update pointer, which
      // remains valid for the inference
      m_spMainChart->insertNoIndex(pItem);
      Hedge* pScanCon = static_cast<Hedge*>(pItem);
      for(iLoads = loadCons.begin(); iLoads != loadCons.end(); ++iLoads) {
	Hedge* pLoadCon = static_cast<Hedge*>(*iLoads);
	Hedge* pH = m_pWMTG->makePretermInference(pScanCon, pLoadCon);
	ret.push_back(pH);
      }
    }

    // clean up
    scanCons.clear();
    loadCons.clear();
  }
}

/** Adds this class' option description
    to the option description list of the hierarchy */
BSP<BPO::options_description>
FasterBottomUpTranslationLogic::addOptionDescriptionsTo
(BSP<BPO::options_description> optionsDescription){

  optionsDescription->add_options()
    ("W2WModelFile", BPO::value<string>(), "w2w model file")
    ("MinTranConf", BPO::value<double>(), "Minimum translation confidence (no output if below this)")
    ("CopyUnseen", BPO::value<bool>(), "Copy unseen input words to the output?")
    ("MaxFertility", BPO::value<unsigned int>()->default_value(10),
     "Max number of Loads per source word")
    ;

  return optionsDescription;
}
