#include "FeatureGenerator.hpp"

#include <string>
#include <stdio.h>
#include <stdlib.h>

#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

#include "NTLink.hpp"
#include "PAV.hpp"
#include "HeadMTreeNode.hpp"
#include "HeadProduction.hpp"
#include "Inference.hpp"
#include "SentenceTuple.hpp"

/** Adds this class' option description to the option description list of the hierarchy */
template<class TreeType> inline
BSP<BPO::options_description> FeatureGenerator<TreeType>::
addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  optionsDescription->add_options()
    ("debug", BPO::value<unsigned int>()->default_value(0), "debug level")
    ("MaxContextWidthLeft", BPO::value<unsigned int>(), "how far out to the left to extract features?")
    ("MaxContextWidthRight", BPO::value<unsigned int>(), "how far out to the right to extract features?")
    ("UseFocusFeats", BPO::value<bool>(), "Use features of the focus word?")
    ("UseProductionFeats", BPO::value<bool>(), "Use production rules as features?")
    ("UseBoundedContextFeats", BPO::value<bool>(), "Use bounded context features?")
    ("UseUnboundedContextFeats", BPO::value<bool>(), "Use unbounded context features?")
    ("UseStructuralFeats", BPO::value<bool>(), "Use structural features?")
    ("UseBUStructureOnly", BPO::value<bool>(), "Use bottom-up structural features only?")
    ("UseTopStructureOnly", BPO::value<bool>(), "Use above structural features only?")
    ("MaxSuffixLength", BPO::value<unsigned int>(), 
     "Max length of suffixes used for suffix features (default 0)")
    ;

  return optionsDescription;
}

template<class TreeType> inline FeatureGenerator<TreeType>::
FeatureGenerator(const arity_t arity, const string& configFile,
		 BSP<BPO::options_description> optionsDescription) 
  : m_pSentTuple(NULL), m_comment(string("")), m_posn2tag(arity),
    m_maxContextWidthLeft(0), m_maxContextWidthRight(0),
    m_useFocusFeats(true), m_useProductionFeats(true), 
    m_useBoundedContextFeats(true), m_useUnboundedContextFeats(true),
    m_useStructuralFeats(true), m_useBUStructureOnly(true), m_useTopStructureOnly(false),
    m_maxSuffixLength(0) {

  addOptionDescriptionsTo(optionsDescription);

  /*
   * Compile the config file.
   */
  BPO::variables_map vm;
  if(configFile != ""){
    iFile ifs(configFile.c_str());
    store(parse_config_file(ifs, *optionsDescription), vm);
  }
  notify(vm);

  for (arity_t dim = 0; dim < arity; ++dim)
    m_posn2tag[dim] = Vector<SynCat>();

  if(vm.count("debug")) debugme(vm["debug"].as<unsigned int>());
  if(vm.count("MaxContextWidthLeft"))
    m_maxContextWidthLeft = vm["MaxContextWidthLeft"].as<unsigned int>();
  if(vm.count("MaxContextWidthRight"))
    m_maxContextWidthRight = vm["MaxContextWidthRight"].as<unsigned int>();
  if(vm.count("UseFocusFeats"))
    m_useFocusFeats = vm["UseFocusFeats"].as<bool>();
  if(vm.count("UseProductionFeats"))
    m_useProductionFeats = vm["UseProductionFeats"].as<bool>();
  if(vm.count("UseBoundedContextFeats"))
    m_useBoundedContextFeats = vm["UseBoundedContextFeats"].as<bool>();
  if(vm.count("UseUnboundedContextFeats"))
    m_useUnboundedContextFeats = vm["UseUnboundedContextFeats"].as<bool>();
  if(vm.count("UseStructuralFeats"))
    m_useStructuralFeats = vm["UseStructuralFeats"].as<bool>();
  if(vm.count("UseBUStructureOnly"))
    m_useBUStructureOnly = vm["UseBUStructureOnly"].as<bool>();
  if(vm.count("UseTopStructureOnly"))
    m_useTopStructureOnly = vm["UseTopStructureOnly"].as<bool>();
  if(vm.count("MaxSuffixLength"))
    m_maxSuffixLength = vm["MaxSuffixLength"].as<unsigned int>();

  if(Dbug::isDebugAll(1)) {
    cerr << "FeatureGenerator options:" << endl;
    cerr << OptionsPrinter(*optionsDescription, vm) << endl;
  }
}

template<class TreeType> inline
FeatureGenerator<TreeType>::~FeatureGenerator() {
  delete m_pSentTuple;
};


template<class TreeType> inline
bool FeatureGenerator<TreeType>::prepareFor(const SentenceTuple& tuple) {
  delete m_pSentTuple; // clean up the previous one
  m_pSentTuple = new SentenceTuple(tuple);
  for(arity_t focusDim = 0; focusDim < m_pSentTuple->size(); ++focusDim) {
    m_posn2tag[focusDim].clear();
    m_posn2tag[focusDim].resize((*m_pSentTuple)[focusDim].size());
  }
  return true;
}

template<class TreeType> inline
bool FeatureGenerator<TreeType>::prepareFor(Vector<const TreeType*>& treeVec) {
  typename Vector<const TreeType*>::const_iterator vtci = treeVec.begin();
  typename Vector<const TreeType*>::const_iterator sentinel = treeVec.end();
  for(; vtci != sentinel; ++vtci) {
    bool ret = prepareFor(*vtci);
    if (not ret)
      return false;
  }
  return true;
}

template<class TreeType> inline
bool FeatureGenerator<TreeType>::prepareFor(const TreeType* pPT, const arity_t focusDim) {
  typename TreeType::iterator iter = pPT->begin();
  typename TreeType::iterator itEnd = pPT->end();
  for(; iter != itEnd; ++iter) {
    /* if it is a leaf */
    if (debug(5))
      cerr << "FeatureGenerator::prepareFor at node " << *iter << endl;
    if( ! iter->empty(focusDim) && TreeType::number_of_children(iter) == 0) {
      short posn = pPT->headPosn(iter, focusDim);
      // now posn will be 0 if the tree is inactive on focusDim
      if (posn > 0) {
	const SynCat sc = pPT->pretermLabel(iter, focusDim);
	m_posn2tag[focusDim][posn - 1] = sc;  // - 1 cuz we want 0-indexed
	if (debug(5)) {
	  cerr << "Recording tag " << sc << " at position " << posn - 1 << " on dim " << (int)focusDim << endl;
	}
      }
    }
  }

  return true;
};

template<class TreeType> inline
void FeatureGenerator<TreeType>::
getFeats(ScoreSet<FeatureType>& featCount, const typename TreeType::iterator& treeIt, 
	 const arity_t focusDim, const DSpan& theDSpan, const string& comment) {

  m_comment = comment;

  if (treeIt == NULL) {
    string featName = string("dim") + char('0' + focusDim) + " NULL treeNode";
    addOneFeature(featCount, featName, m_comment);
    return;
  }

  if (m_useFocusFeats)
    getFocusFeats(featCount, treeIt, focusDim);
  if (m_useBoundedContextFeats)
    getBoundedContextFeats(featCount, focusDim, theDSpan, m_maxContextWidthLeft, m_maxContextWidthRight);
  if (m_useUnboundedContextFeats)
    getUnboundedContextFeats(featCount, treeIt, focusDim);
  if (m_useStructuralFeats)
    getStructuralFeats(featCount, treeIt, focusDim, theDSpan);
};

/// get the lexical heads and SynCats of the item itself
template<class TreeType> inline
void FeatureGenerator<TreeType>::
getFocusFeats(ScoreSet<FeatureType>& featCount, 
	      const typename TreeType::iterator& treeIt, const arity_t focusDim) const {
  SynCat sc;
  string dimNumberString = string("dim") + char('0' + focusDim);
  // terminal
  // feature name will be like dimX.focusWord==Z, where X is the dimension number, 
  // and Z is the word itself
  unsigned headPosn = TreeType::headPosn(treeIt, focusDim);
  Terminal term = (*m_pSentTuple)[focusDim][headPosn - 1]; // -1 cuz it's 0-indexed
  const string* pWord = GlobalVocab::s_terminalSet->getWord(term);
  string wordFeatName = dimNumberString + string(" focusWord") + " = " + *pWord;
  addOneFeature(featCount, wordFeatName);

  // label
  // feature name will be like dimX.focusLabel==Z, where X is the dimension number, 
  // and Z is the label
  if (TreeType::number_of_children(treeIt) == 0) {
    sc = TreeType::pretermLabel(treeIt, focusDim);
  } else {
    sc = BGET_NT_LABEL(*treeIt).getNT(focusDim)->getSynCat();
  }
  const string* pLabel = GlobalVocab::s_synCatSet->getWord(sc);
  string labelFeatName = dimNumberString + string(" focusLabel") + " = " + *pLabel;
  addOneFeature(featCount, labelFeatName);
};

// terminal feature name will be like "word Z at [Left/Right] on dimX", 
// where X is the dimension number and Z is the word itself
// POS tag feature name will be like "label Z at [Left/Right] on dimX", 
// where X is the dimension number and Z is the tag
template<class TreeType> inline
void FeatureGenerator<TreeType>::
getUnboundedContextFeats(ScoreSet<FeatureType>& featCount, 
			 const typename TreeType::iterator& treeIt, const arity_t focusDim) const {
  Terminal term;
  SynCat sc;
  string wordFeatNamePrefix, featName;
  string labelFeatNamePrefix, labelFeatName;
  string dimNumberString = string("dim")  + char('0' + focusDim);

  // first & last word position in current sentence
  short firstPosn = 0; // probably always
  short lastPosn = (*m_pSentTuple)[focusDim].size() - 1; // -1 cuz it's 0-indexed

  short focusPosn = TreeType::headPosn(treeIt, focusDim) - 1; // -1 cuz it's 0-indexed
  if (focusPosn < 0)
    return;
  //  unsigned short focusPosn = nrh->getHeads()[focusDim] - 1; // -1 cuz it's 0-indexed
  //  DSpan& conseqSpan = pConseq->getDSV(focusDim);
  //  unsigned short focusPosn = conseqSpan.front().first; // left boundary of the leftmost span
  for (short absPosn = firstPosn; absPosn <= lastPosn; ++absPosn) {
    // skip focus position
    if (absPosn == focusPosn)
      continue;

    term = (*m_pSentTuple)[focusDim][absPosn];
    sc = m_posn2tag[focusDim][absPosn];

    // word at given posn
    const string* pWord = GlobalVocab::s_terminalSet->getWord(term);
    if (! pWord)
      THROW("Word not found: " << sc);
    Vector<string> wordFeatNamePrefixes;
    getTokenFeatNamePrefixes(wordFeatNamePrefixes, *pWord, dimNumberString);
    // word token features
    Vector<string>::const_iterator vsci = wordFeatNamePrefixes.begin();
    const Vector<string>::const_iterator sentinel = wordFeatNamePrefixes.end();
    for(; vsci != sentinel; ++vsci) {
      if (absPosn < focusPosn) {
	featName = *vsci + " at Left";
      } else {
	featName = *vsci + " at Right";
      }
      addOneFeature(featCount, featName);
    }

    const string* pLabel = GlobalVocab::s_synCatSet->getWord(sc);
    labelFeatNamePrefix = dimNumberString + string(" label ") + *pLabel;
    if (absPosn < focusPosn) {
      labelFeatName = labelFeatNamePrefix + " at Left";
    } else {
      labelFeatName = labelFeatNamePrefix + " at Right";
    }
    addOneFeature(featCount, labelFeatName);
  }
}

// extract feature name prefixes for a particular word token, including 
// * the word itself
// * suffixes
template<class TreeType> inline
void FeatureGenerator<TreeType>::
getTokenFeatNamePrefixes(Vector<string>& featNameVec, const string& inTok,
			 const string& inFeatNamePrefix, const string& inFeatNameSuffix) const {
  unsigned short tokLen = inTok.size();
  // don't take suffixes of length greater than the input token
  size_t currMaxSuffLen = min(tokLen, m_maxSuffixLength);
  string suffix;
  string featNamePrefix = inFeatNamePrefix + string(" word ") + inTok + inFeatNameSuffix;
  featNameVec.push_back(featNamePrefix);

  for(unsigned suffLen = 1; suffLen <= currMaxSuffLen; ++suffLen) {
    suffix = inTok.substr(tokLen - suffLen, suffLen);
    featNamePrefix = inFeatNamePrefix + string(" suffix ") + suffix + inFeatNameSuffix;
    featNameVec.push_back(featNamePrefix);
  }
}

// terminal feature name will be like "word Z at posn [><]=Y on dimX",
// where X is the dimension number, Y is the relative position number
// (including a leading + or -), and Z is the word itself POS tag
// label feature name will be like "label Z at posn [><]=Y on dimX",
// where X is the dimension number, Y is the relative position number
// (including a leading + or -), and Z is the tag
template<class TreeType> inline
void FeatureGenerator<TreeType>::
getBoundedContextFeats(ScoreSet<FeatureType>& featCount, const arity_t focusDim, 
		       const DSpan& theDSpan,
		       const unsigned short maxContextWidthLeft, const unsigned short maxContextWidthRight) const {
  Terminal term;
  SynCat sc;
  string wordFeatNamePrefix, wordFeatName, dimNumberString;
  string labelFeatNamePrefix, labelFeatName;
  string featName, infix;

  static const unsigned short maxPosnStringWidth = 5;
  static char* buffer = (char *) malloc(maxPosnStringWidth);
  short offset, absPosn, leftBracket, rightBracket;
  // first & last word position in current sentence
  short sentenceStartPosn = 0; // probably always
  short SentenceEndPosn = (*m_pSentTuple)[focusDim].size() - 1; // -1 cuz it's 0-indexed
  short spanStartPosn = theDSpan[0].first; // left boundary of first span
  short spanEndPosn = theDSpan[theDSpan.size() - 1].first; // left boundary of last span
  for (short relPosn = 0 - maxContextWidthLeft; relPosn <= maxContextWidthRight; ++relPosn) {
    // skip position zero, cuz the _super takes care of that
    if (relPosn < 0)
      absPosn = spanStartPosn + relPosn;
    else if (relPosn > 0)
      absPosn = spanEndPosn + relPosn;
    else // relPosn == 0
      continue;

    if (absPosn < sentenceStartPosn || absPosn > SentenceEndPosn) {
      term = Terminal(OutOfBoundsT);
      sc = SynCat(OutOfBoundsNT);
    } else {
      term = (*m_pSentTuple)[focusDim][absPosn];
      sc = m_posn2tag[focusDim][absPosn];
    }
    dimNumberString = string("dim")  + char('0' + focusDim);

    // word at given posn
    const string* pWord = GlobalVocab::s_terminalSet->getWord(term);
    if (! pWord)
      THROW("Word not found: " << sc);
    Vector<string> wordFeatNamePrefixes;
    if (absPosn >= sentenceStartPosn && absPosn <= SentenceEndPosn)
      getTokenFeatNamePrefixes(wordFeatNamePrefixes, *pWord, dimNumberString);
    else
      // get the special -oob- token as a feature
      wordFeatNamePrefixes.push_back(dimNumberString + string(" word ") + *pWord);

    // POS tag at given posn
    const string* pLabel = GlobalVocab::s_synCatSet->getWord(sc);
    if (! pLabel)
      THROW("Label not found: " << sc);
    labelFeatNamePrefix = dimNumberString + string(" label ") + *pLabel;

    if (relPosn > 0) {
      infix = string(" at posn <= +");
      leftBracket = relPosn;
      rightBracket = maxContextWidthRight;
    } else {
      infix = string(" at posn >= ");
      leftBracket = 0 - maxContextWidthLeft;
      rightBracket = relPosn;
    }
    // the info appears in every context width from relPosn up to maxContextWidth
    // e.g., if it occurs within an offset of 2, then it also occurs within an offset of 3, etc.
    for(offset = leftBracket; offset <= rightBracket; ++offset) {
      // get string rep'n of current context width in base 10
      snprintf(buffer, maxPosnStringWidth, "%d", offset);

      // word token features
      Vector<string>::const_iterator vsci = wordFeatNamePrefixes.begin();
      const Vector<string>::const_iterator sentinel = wordFeatNamePrefixes.end();
      for(; vsci != sentinel; ++vsci) {
	featName = *vsci + infix + buffer;
	addOneFeature(featCount, featName);
      }

      // label
      labelFeatName = labelFeatNamePrefix + infix + buffer;
      addOneFeature(featCount, labelFeatName);
    }
  }
};

// structural features both above and below the current node
template<class TreeType> inline
void FeatureGenerator<TreeType>::
getStructuralFeats(ScoreSet<FeatureType>& featCount,
		   const typename TreeType::iterator& it, const arity_t focusDim, const DSpan& theDSpan) const {
  //first look at dependents of the node
  if(! m_useTopStructureOnly)
    getStructuralFeatsBelow(featCount, it, focusDim, theDSpan);
  //then get siblings and everything above
  if (! m_useBUStructureOnly)
    getStructuralFeatsAbove(featCount, it, focusDim, theDSpan); 
};

// bottom-up structural features include only the dependents of the current node
template<class TreeType> inline
void FeatureGenerator<TreeType>::
getStructuralFeatsBelow(ScoreSet<FeatureType>& featCount, const typename TreeType::iterator& it, 
			   const arity_t focusDim, const DSpan& theDSpan) const {
  getDependentFeats(featCount, it, focusDim, theDSpan);
};

// structural features include:
// * label of the maximal projection
// * dependents at every X-bar level
// * label/headWord of the parent of the maximal projection (= dependency head)
// * label of the head sibling of the maximal projection
// * whether the head sibling is to the left or to the right
// * label/headWord of all other siblings of the maximal projection
template<class TreeType> inline
void FeatureGenerator<TreeType>::
getStructuralFeatsAbove(ScoreSet<FeatureType>& featCount, const typename TreeType::iterator& treeIt, 
			const arity_t focusDim, const DSpan& theDSpan) const {
  typename TreeType::sibling_iterator si;
  const string* pWord;
  string featName;
  string dimNumberString = string("dim") + char('0' + focusDim);
  // first, find the maximal projection
  typename TreeType::iterator maxProj = TreeType::maximalProjection(treeIt);
  const Nonterminal* pMPNT = BGET_NT_LABEL(*maxProj).getNT(0);
  SynCat maxProjLabel = pMPNT->getSynCat();
  const string* pLabel = GlobalVocab::s_synCatSet->getWord(maxProjLabel);
  featName = dimNumberString + string(" maxProjLabel ") + *pLabel;
  addOneFeature(featCount, featName);

  // next, collect the dependents at every X-bar level
  getDependentFeats(featCount, maxProj, focusDim, theDSpan, treeIt);

  // next, the parent of the maximal projection, if it exists
  typename TreeType::iterator maxProjParent = TreeType::dummyTree.parent(maxProj);
  if (TreeType::dummyTree.is_valid(maxProjParent)) {
    const Nonterminal* pMPParNT  = BGET_NT_LABEL(*maxProjParent).getNT(0);
    SynCat maxProjParentLabel = pMPParNT->getSynCat();
    pLabel = GlobalVocab::s_synCatSet->getWord(maxProjParentLabel);
    featName = dimNumberString + string(" maxProjParentLabel ") + *pLabel;
    addOneFeature(featCount, featName);
    const LexNT* pMParLNT = dynamic_cast<const LexNT*>(pMPParNT);
    if (pMParLNT) {
      Terminal maxProjParentHeadPosn = pMParLNT->lex();
      // - 1 cuz the positions are 1-indexed but the sentences are 0-indexed
      Terminal maxProjParentHead = (*m_pSentTuple)[focusDim][ maxProjParentHeadPosn - 1];
      pWord = GlobalVocab::s_terminalSet->getWord(maxProjParentHead);
      featName = dimNumberString + string(" maxProjParentHead ") + *pWord;
      addOneFeature(featCount, featName);

      // in which direction is the head?
      const Nonterminal* pFocusNT  = BGET_NT_LABEL(*treeIt).getNT(0);
      const LexNT* pFocusLNT = dynamic_cast<const LexNT*>(pFocusNT);
      if (pFocusLNT) {
	Terminal focusHeadPosn = pFocusLNT->lex();
	if ( maxProjParentHeadPosn > focusHeadPosn)
	  featName = dimNumberString + string(" headFollows");
	else 
	  featName = dimNumberString + string(" headPrecedes");
	addOneFeature(featCount, featName);
      }
    }

    // next, all the siblings of the maximal projection
    // iterate over all the children of the parent of the max proj
    for (si = TreeType::dummyTree.begin(maxProjParent);
	 si != TreeType::dummyTree.end(maxProjParent); ++si) {
      if (si == maxProj)
	continue; // skip the max proj itself
      pMPNT = BGET_NT_LABEL(*si).getNT(0);
      SynCat siblingLabel = pMPNT->getSynCat();
      pLabel = GlobalVocab::s_synCatSet->getWord(siblingLabel);
      if (TreeType::dummyTree.isHeir(si, maxProjParent, focusDim)) {
	// we get only the label cuz the headWord would be the same as
	// that of the parent of the maxProj
	featName = dimNumberString + string(" maxProjHeadSiblingLabel ") + *pLabel;
	addOneFeature(featCount, featName);
      } else {
	featName = dimNumberString + string(" maxProjSiblingLabel ") + *pLabel;
	addOneFeature(featCount, featName);
	const LexNT* pMPLNT = dynamic_cast<const LexNT*>(pMPNT);
	if (pMPLNT) {
	  Terminal maxProjSiblingHeadPosn = pMPLNT->lex();
	  // - 1 cuz the positions are 1-indexed but the sentences are 0-indexed
	  Terminal maxProjSiblingHead = (*m_pSentTuple)[focusDim][ maxProjSiblingHeadPosn - 1];
	  pWord = GlobalVocab::s_terminalSet->getWord(maxProjSiblingHead);
	  featName = dimNumberString + string(" maxProjSiblingHead ") + *pWord;
	  addOneFeature(featCount, featName);
	}
      }
    }
  }
};

// collect all of the dependents on the path from inIter down to stopIt on the focusDim
template<class TreeType> inline
void FeatureGenerator<TreeType>::
getDependentFeats(ScoreSet<FeatureType>& featCount, 
		  const typename TreeType::iterator& inIter, 
		  const arity_t focusDim, const DSpan& theDSpan,
		  const typename TreeType::iterator& stopIt) const {
  if (inIter == stopIt)
    return;

  const string* pWord;
  const string* pLabel;
  string featName;
  string dimNumberString = string("dim") + char('0' + focusDim);
  typename TreeType::sibling_iterator si = TreeType::dummyTree.begin(inIter);
  const typename TreeType::sibling_iterator& sentinel = TreeType::dummyTree.end(inIter); 
  for (; si != sentinel; ++si) {
    
    // proceed only if the node is active on focusDim
    if (si->empty(focusDim))
      continue;

    if (! si->hasLabel(focusDim))
      continue;

    const Nonterminal* pNT = BGET_NT_LABEL(*si).getNT(0);
    SynCat depLabel = pNT->getSynCat();
    pLabel = GlobalVocab::s_synCatSet->getWord(depLabel);

    if (TreeType::dummyTree.isHeir(si, inIter, focusDim)) {
      //dont recurse below the stop iterator
      if(si == stopIt)
	continue;
      // we get only the label cuz the headWord would be the same as
      // that of the parent
      featName = dimNumberString + string(" headChildLabel ") + *pLabel;
      addOneFeature(featCount, featName);
      // cast up to more generic iterator type
      typename TreeType::iterator lowerIter(si);
      getDependentFeats(featCount, lowerIter, focusDim, theDSpan); // recurse
    } else {
      unsigned short dependentPosn = TreeType::headPosn(si, focusDim);
      // In the case of a 1D extension on the opposite dim, the span
      // of the node pointed to by inIter can be empty *on the
      // focusDim*.  In that case, there are no dependents to
      // iterate over, even though that node is not a leaf node, and
      // headPosn() will return 0;
      if (dependentPosn == 0)
	return;
      if (dependentPosn > (*m_pSentTuple)[focusDim].size())
	THROW("dependent posn " << dependentPosn);
      // - 1 cuz the positions are 1-indexed but the sentences are 0-indexed
      Terminal dependentTerm = (*m_pSentTuple)[focusDim][ dependentPosn - 1];
      pWord = GlobalVocab::s_terminalSet->getWord(dependentTerm);
      featName = dimNumberString + string(" dependentWord ") + *pWord;
      addOneFeature(featCount, featName);
      // now the label
      featName = dimNumberString + string(" dependentLabel ") + *pLabel;
      addOneFeature(featCount, featName);
    }
  }
};

/// get various types of productions from an inference
template<class TreeType> inline
void FeatureGenerator<TreeType>::
getProdFeats(ScoreSet<FeatureType>& featCount, 
	     const NTLink& theLHS, const NTLink& rhsHead, const NTLink& rhsDep, 
	     const PAV& thePAV, const string& comment) const {

  if (! m_useProductionFeats)
    return;

  // proceed only for 2x2 inferences
  if (thePAV.rank(0) < 2 || thePAV.rank(1) < 2)
    return;
  
  string featName;
  ostringstream PAVStream;
  PAVStream << thePAV;
  featName = "PAV " + PAVStream.str();
  addOneFeature(featCount, featName, comment);

  if (debug(5))
    cerr << "lhs" << theLHS << "; rhs" << rhsHead << " |" << rhsDep << endl;

  ostringstream bilexProdStream, monolexProdStream, delexProdStream;
  bilexProdStream << "lhs" << theLHS << "; rhs" << rhsHead << " |" << rhsDep;
  featName = "bilexProd " + bilexProdStream.str();
  addOneFeature(featCount, featName, comment);

  NTLink delexRHSHead = rhsHead; delexRHSHead.delex();
  NTLink delexRHSDep = rhsDep; delexRHSDep.delex();
  monolexProdStream << "lhs" << theLHS << "; rhs" << delexRHSHead << " |" << delexRHSDep;
  featName = "monolexProd " + monolexProdStream.str();
  addOneFeature(featCount, featName, comment);

  NTLink delexLHS = theLHS; delexLHS.delex();
  delexProdStream << "lhs" << delexLHS << "; rhs" << delexRHSHead << " |" << delexRHSDep;
  featName = "delexProd " + delexProdStream.str();
  addOneFeature(featCount, featName, comment);
};

