#ifndef _FeatureGenerator_H
#define _FeatureGenerator_H

#include "GlobalVocab.hpp"
#include "SentenceTuple.hpp"
#include "ScoreSet.hpp"
#include "Dbug.hpp"

#include "Terminal.hpp"
#include "SynCat.hpp"
#include "LexNT.hpp"
#include "DSpan.hpp"

class NTLink;
class PAV;

template<class TreeType>
class FeatureGenerator : public Dbug {

  BSP<BPO::options_description> 
  addOptionDescriptionsTo(BSP<BPO::options_description> priorOptions);

public:

  typedef GlobalVocab::FeatureType FeatureType;

  //! constructor
  FeatureGenerator<TreeType>(const arity_t dm, const string& configFile,
			     BSP<BPO::options_description> optionsDescription =
			     BSP<BPO::options_description>(new BPO::options_description()));

  ~FeatureGenerator<TreeType>();

  bool prepareFor(const SentenceTuple& tuple);
  /// pre-compute the posn2tag map
  /// The variant with the focusDim parameter allows us to force a
  /// constraining monotree to apply to a particular component.
  bool prepareFor(Vector<const TreeType*>& treeVec);
  //  bool prepareFor(const TreeType* pPT);
  //template<class TreeType>
  bool prepareFor(const TreeType* pPT) {
    size_t tend = m_pSentTuple->size();
    for(size_t dim = 0; dim < tend; ++dim) {
      bool ret = prepareFor(pPT, dim);
      if (not ret)
	return false;
    }
    return true;
  }

  bool prepareFor(const TreeType* pPT, const arity_t focusDim);

  /// can't be const cuz it's sets m_comment
  void getFeats(ScoreSet<FeatureType>& featCount, 
		const typename TreeType::iterator&, const arity_t m_constrainedDim, 
		const DSpan& theSpan, const string& comment = string(""));

  /// add a single feature token into the feature scoreset
  /// just a way to save typing
  inline void addOneFeature( ScoreSet<FeatureType>& featCount, const string& inFeatName, 
			     const string& comment = string ("")) const {
    string featName;
    // if comment supplied, then use it
    if (comment != "")
      featName = comment + inFeatName;
    // otherwise use the pre-existing one
    else
      featName = m_comment + inFeatName;
    VocabularyIndex vi = GlobalVocab::s_featSet->addWord(featName);
    featCount[vi] += UNITYSCORE;
    if (debug(5))
      cerr << "FGen feature: " << featName << endl;
  }

  /// given a particular word token, extract feature name prefixes to be used in feature names computed elsewhere
  /// current features for each token include
  /// * the token itself
  /// * suffixes, up to max length specified in maxSuffLen
  void getTokenFeatNamePrefixes(Vector<string>& featNameVec, const string& inTok,
				const string& inFeatNamePrefix = string(""), 
				const string& inFeatNameSuffix = string("")) const;

  // terminal
  // feature name will be like dimX.focusWord==Z, where X is the dimension number, 
  // and Z is the word itself
  // label
  // feature name will be like dimX.focusLabel==Z, where X is the dimension number, 
  // and Z is the label
  void getFocusFeats(ScoreSet<FeatureType>& featCount, 
		     const typename TreeType::iterator& treeIt, const arity_t focusDim) const;

  /// terminal feature name will be like "word Z at [Left/Right] on dimX", 
  /// where X is the dimension number and Z is the word itself
  /// POS tag feature name will be like "label Z at [Left/Right] on dimX", 
  /// where X is the dimension number and Z is the tag
  void getUnboundedContextFeats(ScoreSet<FeatureType>& featCount, 
				const typename TreeType::iterator&, const arity_t focusDim) const;

  /// terminal feature name will be like "word Z at posn [><]=Y on dimX",
  /// where X is the dimension number, Y is the relative position number
  /// (including a leading + or -), and Z is the word itself POS tag
  /// label feature name will be like "label Z at posn [><]=Y on dimX",
  /// where X is the dimension number, Y is the relative position number
  /// (including a leading + or -), and Z is the tag
  void getBoundedContextFeats(ScoreSet<FeatureType>& featCount, 
			      // const typename TreeType::iterator&, 
			      const arity_t focusDim, const DSpan& theSpan,
			      const unsigned short maxContextWidthLeft, const unsigned short maxContextWidthRight) const;

  /// structural features include:
  /// * label of the maximal projection
  /// * dependents at every x-bar level
  /// * label/headWord of the parent of the maximal projection (= dependency head)
  /// * label of the head sibling of the maximal projection
  /// * whether the head sibling is to the left or the right
  /// * label/headWord of all other siblings of the maximal projection
  void getStructuralFeats(ScoreSet<FeatureType>& featCount, 
			  const typename TreeType::iterator&, const arity_t focusDim, const DSpan& theDSpan) const;

  /// bottom-up structural features include only the dependents of the current node
  void getStructuralFeatsBelow(ScoreSet<FeatureType>& featCount, 
				  const typename TreeType::iterator&, const arity_t focusDim, const DSpan& theDSpan) const;

  /// above structural features include only the sibs and parents, not dependents.
  void getStructuralFeatsAbove(ScoreSet<FeatureType>& featCount, const typename TreeType::iterator&, 
			       const arity_t focusDim, const DSpan& theDSpan) const;

  /// recursive helper function for getStructuralFeats
  /// collect dependents at every X-bar level on the path from topIter down to bottomIter
  void getDependentFeats(ScoreSet<FeatureType>&, 
			 const typename TreeType::iterator& topIter, 
			 const arity_t focusDim, const DSpan& theDSpan,
			 const typename TreeType::iterator& bottomIter = NULL) const;

  /// get various types of productions from an inference
  void getProdFeats(ScoreSet<FeatureType>& featCount, 
		    const NTLink& theLHS, const NTLink& rhsHead, const NTLink& rhsDep, 
		    const PAV& thePAV, const string& comment = string("")) const;

  SentenceTuple* getSntTuple(){
    return m_pSentTuple;
  }


protected: // data

  /// memoize sent tuple to work on
  SentenceTuple* m_pSentTuple;

  string m_comment; // to be prefixed to every feature, empty by default

  /// index word position to POS tag
  /// this should be faster and easier than using the span2leaf index
  /// however, later, we will need to use the span2leaf index for full generality
  /// m_posn2tag[x][y] refers to the y'th tag in the x'th component
  Vector<Vector<SynCat> > m_posn2tag;

  unsigned short m_maxContextWidthLeft;
  unsigned short m_maxContextWidthRight;
  bool m_useFocusFeats;
  bool m_useProductionFeats;
  bool m_useBoundedContextFeats;
  bool m_useUnboundedContextFeats;
  bool m_useStructuralFeats;
  bool m_useBUStructureOnly;
  bool m_useTopStructureOnly;
  unsigned short m_maxSuffixLength;

};

#include "FeatureGenerator.cpp"
#include "ItemFeatureGenerator.cpp"

#endif
