#include "File.hpp"
#include "Path.hpp"
#include "FeatureWMTG.hpp"
#include "FeatureGenerator.hpp"
#include "TLink.hpp"
#include "HeadMTree.hpp"
#include "HedgeTree.hpp"
#include "GlobalVocab.hpp"
#include "Hedge.hpp"
#include "HeadProduction.hpp"
#include "Scan.hpp"
#include "CSHeadCompose.hpp"
#include "Exceptions.hpp"
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

//! Destructor. To clear the memory allocated.
FeatureWMTG::~FeatureWMTG() {}

FeatureWMTG::FeatureWMTG(arity_t dm, const string& configFile,
			 char** overrideArgV, size_t overrideArgC,
			 BSP<BPO::options_description> optionsDescription)
  : _grammar(dm), 
    _super(dm, configFile, addOptionDescriptionsTo(optionsDescription)),
    m_pFeatGen(NULL), m_pItemFeatGen(NULL), m_pMC(NULL) {

  string optFeatGenConfigFile;
  string optItemFeatGenConfigFile;
  string optMulticlassifierConfigFile;
  /*
   * Compile the config file.
   */

  BPO::variables_map vm;
  if(overrideArgV){
    if(this->debug(1)){
      cerr<<"FeatureWMTG: Overriding config file arguments..."<<endl;
    }
    store(BPO::command_line_parser(overrideArgC, overrideArgV).
	  options(*optionsDescription).run(), vm);
  }
  if(configFile != ""){
    iFile ifs(configFile.c_str());
    store(parse_config_file(ifs, *optionsDescription), vm);
  }
  notify(vm);

  if(vm.count("FeatGenConfig")) optFeatGenConfigFile = vm["FeatGenConfig"].as<string>();
  if(vm.count("ItemFeatGenConfig")) optItemFeatGenConfigFile = vm["ItemFeatGenConfig"].as<string>();
  if(vm.count("MulticlassifierConfig")) optMulticlassifierConfigFile = vm["MulticlassifierConfig"].as<string>();

  if(this->debug(1)) {
    cerr << "FeatureWMTG options:" << endl;
    cerr << OptionsPrinter(*optionsDescription, vm) << endl;
  }

  if(optFeatGenConfigFile != "") {
    string FGConfigFile = Path::maybe_compose(configFile, optFeatGenConfigFile);
    m_pFeatGen = new FeatureGenerator<HeadMTree>(dm, FGConfigFile);
  }
  if(optItemFeatGenConfigFile != "") {
    string IFGConfigFile = Path::maybe_compose(configFile, optItemFeatGenConfigFile);
    m_pItemFeatGen = new FeatureGenerator<HedgeTree>(dm, IFGConfigFile);
  }
  if(optMulticlassifierConfigFile != "") {
    string MCConfigFile = Path::maybe_compose(configFile, optMulticlassifierConfigFile);
    m_pMC = new Multiclassifier(MCConfigFile);
  }
}

size_t FeatureWMTG::size() const {
  return m_feats.size();
};

// return the sum of the scores of all features
SCORE FeatureWMTG::mass() {
  return m_feats.sum();
}

//! begin iterator
FeatureWMTG::iterator FeatureWMTG::begin() {
  return m_feats.begin();
}

//! const begin iterator
FeatureWMTG::const_iterator FeatureWMTG::begin() const {
  return m_feats.begin();
}

//! end iterator
FeatureWMTG::iterator FeatureWMTG::end() {
  return m_feats.end();
}

//! const end.iterator
FeatureWMTG::const_iterator FeatureWMTG::end() const {
  return m_feats.end();
}

bool FeatureWMTG::prepareFor(SentenceTuple& tuple) {
  if (! _super::prepareFor(tuple))
    return false;
  if (m_pFeatGen && ! m_pFeatGen->prepareFor(tuple))
    return false;
  if (m_pItemFeatGen && ! m_pItemFeatGen->prepareFor(tuple))
    return false;

  return true;
}

bool FeatureWMTG::prepareFor(HeadMTree* pTree) {
  if (! _super::prepareFor(pTree))
    return false;
  if (m_pFeatGen && ! m_pFeatGen->prepareFor(pTree))
    return false;
  return true;
}

bool FeatureWMTG::prepareFor(const Vector<HeadMTree*>& treeVec, vector<unsigned int>& startingPosns) {
  return LearnableGrammar::prepareFor(treeVec, startingPosns);
}

//! input
istream& FeatureWMTG::get(istream& is){
  _super::get(is);
  is >> m_feats;
  return is;
}

//! output
ostream& FeatureWMTG::put(ostream& os) const {
  _super::put(os);
  os << m_feats;
  return os;
}

//! update the score of a production, NOT propagating to subfeatures
void FeatureWMTG::updateFeatureScore(const GlobalVocab::FeatureType& feat, SCORE sc) {
  m_feats[feat] = sc;
}

//! Increment productions from a grammar with iterators
void FeatureWMTG::increment(const Grammar* pG ) {
  const FeatureWMTG* pFeature = dynamic_cast<const FeatureWMTG*>(pG);
  if (not pFeature)
    THROW("Grammar type mismatch in FeatureWMTG::increment");
  for(const_iterator nwci = pFeature->begin(); nwci != pFeature->end(); ++nwci) {
    pair<const GlobalVocab::FeatureType, SCORE> scoredFeat = *nwci;
    updateFeatureScore(scoredFeat.first, scoredFeat.second);
  }
}

/** Clears the content (only) of the containers. */
void FeatureWMTG::clear() {
  _super::clear();

  // erase container
  m_feats.clear();
}

/// score an inference, assuming it's not a Scan
SCORE FeatureWMTG::getInfScore(Inference* pInf) const {

  if (debug(3)) {
    cerr << "FWMTG::getInfScore working on " << *pInf << endl;
  }

  static const string STRAIGHT_PAV = "straight";
  static const string INVERTED_PAV = "inverted";

  const ScoreSet<GlobalVocab::FeatureType>& featCount = inf2feats(pInf);
  vector<string> strVec;
  const ScoreSet<GlobalVocab::FeatureType>::const_iterator sentinel = featCount.end();
  ScoreSet<GlobalVocab::FeatureType>::const_iterator itFC = featCount.begin();
  for(; itFC != sentinel; ++itFC) {
    GlobalVocab::FeatureType feat = itFC->first;
    const string* pFeatName = GlobalVocab::s_featSet->getWord(feat);
    strVec.push_back(*pFeatName);
  };
  if (debug(3)) {
    cerr << "generated featVec " << strVec;
  }

  HeadCompose* compose = dynamic_cast<HeadCompose*>(pInf);
  if(!compose){
    THROW("Inference not a compose in FeatureWMTG");
  }
  
  
  const arity_t sourceDim = 0;
  const arity_t targetDim = 1;
  const role_t sourceRole = compose->heirIndices().getRole(sourceDim);
  
  const Hedge* pAnt1 = dynamic_cast<const Hedge*>(compose->getAnt()[0]);
  const Hedge* pAnt2 = dynamic_cast<const Hedge*>(compose->getAnt()[1]);

  string source;
  string target;
  if (pAnt1->empty(targetDim) && pAnt2->empty(targetDim)) {
    // use the non-heir item  //PAV [1 2 | ]
    if (sourceRole == 2){
      source =*(GlobalVocab::s_terminalSet->getWord((int)(static_cast<const LexNT*>(pAnt1->getNTLink(sourceDim))->lex())));
      target = "_0_";
    }
    else{
      source = *(GlobalVocab::s_terminalSet->getWord(static_cast<const LexNT*>(pAnt2->getNTLink(sourceDim))->lex()));
      target = "_0_";
    } 
  }
  
  else if (pAnt1->empty(targetDim)){ // PAV [1 | 2 ] or [ 1 2 | 2 ]
    if(pAnt2->empty(sourceDim)){// [1 | 2]
      source = *(GlobalVocab::s_terminalSet->getWord((static_cast<const LexNT*>(pAnt1->getNTLink(sourceDim))->lex())));
      target =*(GlobalVocab::s_terminalSet->getWord((static_cast<const LexNT*>(pAnt2->getNTLink(targetDim))->lex())));
    }
    else{// [ 1 2 | 2 ]
      source = *(GlobalVocab::s_terminalSet->getWord((static_cast<const LexNT*>(pAnt1->getNTLink(sourceDim))->lex())));
      target = "_0_";
    }
  }
  
  else if (pAnt2->empty(targetDim)){ // PAV [ 1 2 | 1 ]
      source = *(GlobalVocab::s_terminalSet->getWord((static_cast<const LexNT*>(pAnt2->getNTLink(sourceDim))->lex())));
      target = "_0_";
  }
  
  else {	// PAV [1 2 | 1 2] or [1 2 | 2 1]
    const PAV thePAV = pInf->getPAV();
    if(thePAV.getPA(1)[0] == 1){
      source = "_0_";
      target =  STRAIGHT_PAV;
    }else{
      source = "_0_";
      target =  INVERTED_PAV;
    }
  }

  SCORE scr = m_pMC->loss_of(source,target, strVec);
  return scr; // for now
}

const ScoreSet<GlobalVocab::FeatureType>& FeatureWMTG::inf2feats(Inference* pInf) const {
  const Item* pConseq = pInf->getConseq();
  const vector<const Item*>& antVec = pInf->getAnt();
  if (antVec.size() != 2)
    THROW("dunno what to do with non-binary infs yet");

  const Hedge* pAnt1 = dynamic_cast<const Hedge*>(antVec[0]);
  const Hedge* pAnt2 = dynamic_cast<const Hedge*>(antVec[1]);
  const Hedge* pConH = dynamic_cast<const Hedge*>(pConseq);
  if (!pConH || !pAnt1 || ! pAnt2)
    THROW("dunno what to do with non-Hedge items");
  const DSV& ant1DSV = *(pAnt1->getDSV());
  const DSV& ant2DSV = *(pAnt2->getDSV());
  const HeadCompose* pHC = dynamic_cast<const HeadCompose*>(pInf);
  const HeirRoleVector& hrv = pHC->heirIndices();

  return inf2featsHelper( pConH, pAnt1, pAnt2, hrv, ant1DSV, ant2DSV);
}

const ScoreSet<GlobalVocab::FeatureType>& FeatureWMTG::
inf2featsHelper(const Hedge* pConH, const Hedge* pAnt1, const Hedge* pAnt2,
		const HeirRoleVector& hrv, const DSV& ant1DSV, const DSV& ant2DSV) const {

  static ScoreSet<GlobalVocab::FeatureType> featCount;
  featCount.clear();

  const arity_t sourceDim = 0;
  const arity_t targetDim = 1;
  const role_t sourceRole = hrv.getRole(sourceDim);
  //  const role_t targetRole = hrv.getRole(targetDim);
  // for word->null case -- use only the 1D item
  // first check for the case where both antecedents are 1D source, i.e. PAV [ 1 2 | ]
  if (pAnt1->empty(targetDim) && pAnt2->empty(targetDim)) {
    // use the non-heir item
    if (sourceRole == 2)
      m_pItemFeatGen->getFeats(featCount, pAnt1, sourceDim, ant1DSV[sourceDim], "");
    else
      m_pItemFeatGen->getFeats(featCount, pAnt2, sourceDim, ant2DSV[sourceDim], "");
  } else if (pAnt1->empty(targetDim)) // PAV [1 | 2 ] or [ 1 2 | 2 ]
    m_pItemFeatGen->getFeats(featCount, pAnt1, sourceDim, ant1DSV[sourceDim], "");
  else if (pAnt2->empty(targetDim)) // PAV [ 1 2 | 1 ]
    m_pItemFeatGen->getFeats(featCount, pAnt2, sourceDim, ant2DSV[sourceDim], "");
  else {	// PAV [1 2 | 1 2] or [1 2 | 2 1]
    // production features, including the PAV
    const PAV thePAV = pConH->getInference()->getPAV();
    const NTLink& theLHS = pConH->getNTLink();
    NTLink rhsHead, rhsDep;
    if (sourceRole == 1) {
      rhsHead = pAnt1->getNTLink();
      rhsDep = pAnt2->getNTLink();
    } else if (sourceRole == 2) {
      rhsHead = pAnt2->getNTLink();
      rhsDep = pAnt1->getNTLink();
    } else
      THROW("illegal heir role " << sourceRole << " on dim " << (int)sourceDim);

    m_pItemFeatGen->getProdFeats(featCount, theLHS, rhsHead, rhsDep, thePAV, "conseq ");
    for(arity_t dim = 0; dim < m_arity; ++dim) {
      if (pAnt1->empty(dim) && pAnt2->empty(dim))
	continue;
      role_t theRole = hrv.getRole(dim);
      if (theRole == 1) {
	if (! pAnt1->empty(dim))
	  m_pItemFeatGen->getFeats(featCount, pAnt1, dim, ant1DSV[dim], "heir ");
	if (! pAnt2->empty(dim))
	  m_pItemFeatGen->getFeats(featCount, pAnt2, dim, ant2DSV[dim], "dep ");
      } else if (theRole == 2) {
	if (! pAnt2->empty(dim))
	  m_pItemFeatGen->getFeats(featCount, pAnt2, dim, ant2DSV[dim], "heir ");
	if (! pAnt1->empty(dim))
	  m_pItemFeatGen->getFeats(featCount, pAnt1, dim, ant1DSV[dim], "dep ");
      } else
	THROW("illegal heir role " << theRole << " on dim " << (int)dim);
      // the consequent label
      string featName;
      SynCat sc;
      sc = pConH->getNTLink().getNT(dim)->getSynCat();
      const string* pLabel = GlobalVocab::s_synCatSet->getWord(sc);
      featName =  string("dim") + char('0' + dim) + " label " + *pLabel;
      m_pItemFeatGen->addOneFeature(featCount, featName, "conseq ");
    }
  }

  return featCount;
};

//! returns possible consequents of a TLink antecedent
void FeatureWMTG::getPossibleConsHeaded(const Item* heir, const Item* dependent,
					const Pair<bool,PAV>& arnmt, vector<HRVIpair > & parents) {
  THROW("Not implemented yet")
  // replace each HeadCompose with a CSHeadCompose
  // replace the consequent modelScore accordingly
};

//! Decompose a multitree into inferences in the way specified by the grammar.
void FeatureWMTG::treeNode2infs(Vector<Inference*>& ret, const HeadMTree::iterator& conseqNode, 
				const DSV& conseqDSV, const DSV& ant1DSV, const DSV& ant2DSV) {
  short fanout = conseqNode.number_of_children();
  if (fanout == 0) {
    // do nothing
  } else if (fanout == 1) {
    Hedge* pConseq = new Hedge(conseqNode, conseqDSV);
    HeadMTree::sibling_iterator sit = HeadMTree::dummyTree.begin(conseqNode);
    const TLink& antTL = BGET_TERM_SPAN_LABEL(*sit).first;
    Inference* pInf = new Scan(pConseq, antTL);
    if (debug(3))
      cerr << "generated Scan " << pInf->getUniqueID() << " from Node: " << *conseqNode 
	   << " Conseq: " << *pConseq << " ANT: " << antTL << endl;
    ret.push_back(pInf);
  } else if (fanout == 2) {
    // nodes of the first and second antecedents
    HeadMTree::sibling_iterator hmtAntNode1 = HeadMTree::dummyTree.begin(conseqNode);
    HeadMTree::sibling_iterator hmtAntNode2 = hmtAntNode1; ++hmtAntNode2;  // the next one over

    Hedge* pConseq = new Hedge(conseqNode, conseqDSV);
    PAV thePAV = BGET_PAV(*conseqNode);
    const Pair<bool, PAV> arnmt(true,thePAV);
    Hedge* pAnt1 = new Hedge(hmtAntNode1, ant1DSV);
    Hedge* pAnt2 = new Hedge(hmtAntNode2, ant2DSV);
    if (debug(3))
      cerr << "FWMTG::treeNode2infs working on conseq " << *pConseq << endl
	   << " ANT1 " << *pAnt1 << endl << " ANT2 " << *pAnt2 << endl
	   <<  " hmtAntNode1 " << *hmtAntNode1 << endl << " hmtAntNode2 " << *hmtAntNode2 << endl;

    if (! m_isHeaded)
      THROW("you probably don't want an unheaded FeatureWMTG");

    HeirRoleVector hrv = HeadMTree::getHRV(conseqNode);

    const ScoreSet<GlobalVocab::FeatureType>& featCount =
      treeNode2feats(conseqNode,hmtAntNode1,hmtAntNode2,hrv,conseqDSV,ant1DSV,ant2DSV);

    CSHeadCompose* pHC = new CSHeadCompose(pAnt1, pAnt2, pConseq, arnmt, hrv);
    pHC->getFeatVec().push_back(featCount);

    if (debug(3))
      cerr << "generated CSHeadCompose " << *pHC << endl;
    ret.push_back(pHC);

  } else 
    THROW("Dunno what to do with fanout of " << fanout);
}

const ScoreSet<GlobalVocab::FeatureType>& FeatureWMTG::
treeNode2feats(const HeadMTree::iterator& conseqNode, 
	       const HeadMTree::iterator& hmtAntNode1, const HeadMTree::iterator& hmtAntNode2,
	       const HeirRoleVector& hrv,
	       const DSV& conseqDSV, const DSV& ant1DSV, const DSV& ant2DSV) {

  static ScoreSet<GlobalVocab::FeatureType> featCount;
  featCount.clear();
  const arity_t sourceDim = 0;
  const arity_t targetDim = 1;
  const role_t sourceRole = hrv.getRole(sourceDim);
  //  const role_t targetRole = hrv.getRole(targetDim);

  // for word->null case -- use only the 1D item
  // first check for the case where both antecedents are 1D source, i.e. PAV [ 1 2 | ]
  if (hmtAntNode1->empty(targetDim) && hmtAntNode2->empty(targetDim)) {
    // use the non-heir item
    if (sourceRole == 2)
      m_pFeatGen->getFeats(featCount, hmtAntNode1, sourceDim, ant1DSV[sourceDim], "");
    else
      m_pFeatGen->getFeats(featCount, hmtAntNode2, sourceDim, ant2DSV[sourceDim], "");
  } else if (hmtAntNode1->empty(targetDim)) // PAV [1 | 2 ] or [ 1 2 | 2 ]
    m_pFeatGen->getFeats(featCount, hmtAntNode1, sourceDim, ant1DSV[sourceDim], "");
  else if (hmtAntNode2->empty(targetDim)) // PAV [ 1 2 | 1 ]
    m_pFeatGen->getFeats(featCount, hmtAntNode2, sourceDim, ant2DSV[sourceDim], "");
  else {	// PAV [1 2 | 1 2] or [1 2 | 2 1]
    // production features, including the PAV
    PAV& thePAV = conseqNode->getPAV();
    NTLink& theLHS = conseqNode->getNTLink();
    NTLink rhsHead, rhsDep;
    if (sourceRole == 1) {
      rhsHead = hmtAntNode1->getNTLink();
      rhsDep = hmtAntNode2->getNTLink();
    } else if (sourceRole == 2) {
      rhsHead = hmtAntNode2->getNTLink();
      rhsDep = hmtAntNode1->getNTLink();
    } else
      THROW("illegal heir role " << sourceRole << " on dim " << (int)sourceDim);

    m_pFeatGen->getProdFeats(featCount, theLHS, rhsHead, rhsDep, thePAV, "conseq ");
    for(arity_t dim = 0; dim < m_arity; ++dim) {
      if (hmtAntNode1->empty(dim) && hmtAntNode2->empty(dim))
	continue;
      role_t theRole = hrv.getRole(dim);
      if (theRole == 1) {
	if (! hmtAntNode1->empty(dim))
	  m_pFeatGen->getFeats(featCount, hmtAntNode1, dim, ant1DSV[dim], "heir ");
	if (! hmtAntNode2->empty(dim))
	  m_pFeatGen->getFeats(featCount, hmtAntNode2, dim, ant2DSV[dim], "dep ");
      } else if (theRole == 2) {
	if (! hmtAntNode2->empty(dim))
	  m_pFeatGen->getFeats(featCount, hmtAntNode2, dim, ant2DSV[dim], "heir ");
	if (! hmtAntNode1->empty(dim))
	  m_pFeatGen->getFeats(featCount, hmtAntNode1, dim, ant1DSV[dim], "dep ");
      } else
	THROW("illegal heir role " << theRole << " on dim " << (int)dim);
      // the consequent label
      string featName;
      SynCat sc;
      sc = conseqNode->getNTLink().getNT(dim)->getSynCat();
      const string* pLabel = GlobalVocab::s_synCatSet->getWord(sc);
      featName =  string("dim") + char('0' + dim) + " label " + *pLabel;
      m_pFeatGen->addOneFeature(featCount, featName, "conseq ");
    }
  }
  return featCount;
};

void FeatureWMTG::getPossibleWordTranslations(const Terminal t, Vector<Terminal>& vec){
  
  //actual word
  const string* pWord = GlobalVocab::s_terminalSet->getWord(t);  
  
  //now send it to ML program and store results in here:
  vector<string>vecRet;

  //get cadidates from BDT
  m_pMC->candidates_of(vecRet, *pWord);

  //iterate through returned vector and pull out strings.
  // then look the trings up in the vocab and insert them.
  vector<string>::iterator it;
  for(it = vecRet.begin(); it != vecRet.end(); it++){
    if(*it != "_0_"){
      Terminal t = GlobalVocab::s_terminalSet->addWord(*it);
      vec.push_back(t);
    }
  }
}
 

/** Adds this class' option description to the option description list of the hierarchy */
BSP<BPO::options_description> 
FeatureWMTG::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  optionsDescription->add_options()
    ("FeatGenConfig", BPO::value<string>(), "config file for the HeadMTree-based feature generator")
    ("ItemFeatGenConfig", BPO::value<string>(), "config file for the item-based feature generator")
    ("MulticlassifierConfig", BPO::value<string>(), "config file for the classifier")
    ;
  return optionsDescription;
}
