#ifndef _FeatureWMTG_H_
#define _FeatureWMTG_H_

#include "Multiclassifier.H"
#include "ScoreSet.hpp"
#include "LearnableWMTG.hpp"

#include <boost/program_options.hpp>
#include <utility>


template<class TreeType>
class FeatureGenerator;

class CSHeadCompose;
class HedgeTree;

//! concrete extension of LearnableWMTG that keeps track of arbitrary features of inferences
class FeatureWMTG : public LearnableWMTG {
  typedef LearnableWMTG _super;
  typedef Grammar _grammar;

  typedef ScoreSet<GlobalVocab::FeatureType> container_type;

  // for serialization only

  BSP<BPO::options_description> addOptionDescriptionsTo(BSP<BPO::options_description> priorOptions);

public:
  FeatureWMTG() {};

  /// constructor from config file
  FeatureWMTG(arity_t dm, const string& configFile,
	    char** overrideArgV = NULL, size_t overrideArgC = 0,
	    BSP<BPO::options_description> priorOptions 
	    = BSP<BPO::options_description>(new BPO::options_description())
	    );

  //! destructor
  virtual ~FeatureWMTG();

  virtual size_t size() const;
  // return the sum of the scores of all features
  virtual SCORE mass();

  //! Iterator to iterate all the MTG productions.
  typedef container_type::iterator iterator;
  typedef container_type::const_iterator const_iterator;

  iterator begin();
  const_iterator begin() const;
  iterator end();
  const_iterator end() const;

  virtual bool prepareFor(SentenceTuple& tuple);
  virtual bool prepareFor(HeadMTree* pTree);     
  virtual bool prepareFor(const Vector<HeadMTree*>& treeVec, vector<unsigned int>& startingPosns);

  bool knows(Production* pProd) const {
    return false;
  };

  //! Increment a production
  virtual void increment(Production* , SCORE = UNITYSCORE) {
    THROW("not implemented");  
  };

  //! Increment features from a grammar
  virtual void increment(const Grammar* );

  //! update the score of a production, NOT propagating to subfeatures
  virtual void updateFeatureScore(Production* pProd, SCORE sc) {
    THROW("not implemented");    
  };

  //! update the score of a production, NOT propagating to subfeatures
  virtual void updateFeatureScore(const GlobalVocab::FeatureType& feat, SCORE sc);

  //! Binarizes all the productions.
  virtual void binarize(vector<StringVocabulary*>&, char* binarizationMethod = "all") {
    THROW("not implemented");
  }

  //! determine whether a certain inference should be allowed
  virtual bool allowInf(const Inference* pInf) {
    return true;
  };

  /// score an inference, assuming it's not a Scan
  virtual SCORE getInfScore(Inference* pInf) const;

  //! returns possible consequents of a TLink antecedent
  virtual void getPossibleConsForTLink(const TLink& t, const DSV& d, vector<Item*>& vItem) const {
    // do nothing
  };

  //! returns possible consequents of a TLink antecedent
  virtual void getPossibleConsHeaded(const Item* heir, const Item* dependent,
                                     const Pair<bool,PAV>& arment, vector<HRVIpair > & parents);

  const ScoreSet<GlobalVocab::FeatureType>& inf2feats(Inference* pInf) const;
  virtual const ScoreSet<GlobalVocab::FeatureType>&
  inf2featsHelper(const Hedge* pConH, const Hedge* pAnt1, const Hedge* pAnt2,
		  const HeirRoleVector& hrv, const DSV& ant1DSV, const DSV& ant2DSV) const;

  //! create an inference from the given tree node and add it to the vector
  void treeNode2infs(Vector<Inference*>&, const HeadMTree::iterator&, const DSV& conseqDSV, 
		     const DSV& antDSV1 = DSV(), const DSV& antDSV2 = DSV());
  //! create an inference from the given tree node and add it to the set
  virtual const ScoreSet<GlobalVocab::FeatureType>&
  treeNode2feats(const HeadMTree::iterator&,
		      const HeadMTree::iterator& hmtAntNode1, const HeadMTree::iterator& hmtAntNode2,
		      const HeirRoleVector& hrv, 
		      const DSV& conseqDSV, const DSV& antDSV1 = DSV(), const DSV& antDSV2 = DSV());

  //! input
  virtual istream& get(istream& is);

  //! output
  virtual ostream& put(ostream& os) const;

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    ar & boost::serialization::base_object<_super>(*this);
    ar & m_feats;
  };

  /** Clears the content (only) of the containers. */
  virtual void clear();

  virtual void getPossibleWordTranslations(const Terminal t, Vector<Terminal>& vec);

protected: // data

  //! HeadMTree-based feature generator, for when we already have a HeadMTree
  FeatureGenerator<HeadMTree>* m_pFeatGen;
  //! Item-based feature generator, for use during inference
  FeatureGenerator<HedgeTree>* m_pItemFeatGen;

  //! the features
  container_type m_feats;

  //! the multiclassifier. 
  Multiclassifier* m_pMC;

};

#endif
