#ifndef _GPDiscChart_C_
#define _GPDiscChart_C_

#include "GPDiscChart.hpp"
#include <queue>

// constructor
template<class ItemT,class ItemInterface> GPDiscChart<ItemT,ItemInterface>::
GPDiscChart(const arity_t& dim, const arity_t& inputArity, int fanout)
  : _super(dim, inputArity), m_discChart( fanout, ((int)dim) ) {
  assert(dim>0 && fanout>0); 
}

// Clone a Chart of the same type as mine without copying the content.
template<class ItemT,class ItemInterface>
GPDiscChart<ItemT,ItemInterface>*
GPDiscChart<ItemT,ItemInterface>:: memorylessClone() const {
  const GPDiscChart< ItemT,ItemInterface>* newChart =
    new GPDiscChart< ItemT,ItemInterface>(m_arity, m_inputArity);
  if(typeid(newChart) == typeid(this)){
    return (GPDiscChart< ItemT>*) newChart;
  } else {
    throw NotImplemented();
    return NULL;
  }
}

// Destructor
template<class ItemT,class ItemInterface> GPDiscChart<ItemT,ItemInterface>::
~GPDiscChart() {}

// Reset the content.
template<class ItemT,class ItemInterface>
void GPDiscChart<ItemT,ItemInterface> :: clear() {
  _super::clear();
  m_discChart.clear();
}

//! Inserts an item, returns the pointer to the inserted item in chart.
//! \param item : the item to be inserted.
//! \param memoRes : the result of the insertion.
//! \param arrgs: the compatible arrangements for this item, according to the grammar.
template<class ItemT,class ItemInterface>
Item* GPDiscChart<ItemT,ItemInterface> ::
insert(Item* pTrigger, MemoResult& memoRes, const set< Pair<bool,PAV> >& arrgs) {

  if (debug(5)) {
    cerr << "GPDiscChart::insert: pTrigger: ";
    pTrigger->dump(cerr);
    cerr << endl;
  }

  Log<SCORE> logfun(LOGBASE);

  //! If the input item has minus infinite score, which means that
  //! this item is underivable, then ignore it.
  if(pTrigger->getScore() == logfun(ZEROSCORE)) {

    memoRes = IGNORE;
    return NULL;
  }

  NTLink label;
  NTLink  temp_lnl;

  // retrieve all items with same coverage
  ItemList* pitem_list = m_discChart.find(pTrigger);

  if(pitem_list == NULL) {
    bool res = m_discChart.insert(pTrigger);   
    _super::insertNoIndex(pTrigger);
    assert(res);
    memoRes = INSERT;
  } else {
    // find equal item 
    Item* existing_item = findItem(pitem_list,pTrigger);

    if (existing_item == NULL) {
      bool res = m_discChart.insert(pTrigger);   // insert if no match found
      assert(res==false);
      memoRes = INSERT;
    } else {
      memoRes = updateScoreAndInference( existing_item /*previous*/,
                                         (Item*)pTrigger /*current */);
      return existing_item;
    }
  }

  //DiscChart insert does not create a copy, so same item is returned here
  return pTrigger;
}

template<class ItemT,class ItemInterface>
void GPDiscChart<ItemT,ItemInterface> ::
printStats(ostream& out) const {
  _super::printStats(out);
  out<<endl<<"Chart Summary:"<<endl;
  RangeIter it = m_discChart.queryAll();

  while (it.has_next()) {
    out<<(const ItemT&)(**it)<<endl;
    ++it;
  }

  out<<"Num of Items: "<<size()<<endl;
}


// added s.w.
// contians  assert as sanity checks,
// many can be removed later
template<class ItemT,class ItemInterface>
void GPDiscChart<ItemT,ItemInterface> ::
matchItem(const Item &trigger,
	  const Pair<bool,PAV>& arrg,
	  vector<Item*> &matchedItems) {
  typedef DSpan::Span Span;

  // Cast information of the input item;
  const DSV* dsv = trigger.getDSV();

  assert( this->arity() <= dsv->arity() );
  //assert( 2 == this->C );
  HRect res( 2 * this->arity() * this->maxSpansPerDim() ); //2CD

  short me    = (arrg.first == 1) ? 1 : 2;  /* 1 or 2 in PAV */
  short other = (arrg.first == 1) ? 2 : 1;  /* 1 or 2 in PAV */

  // convert this trigger/arrangement pair to a HRect
  // for each dimension, add 2 Range's to HRect
  int offset = 2 * this->maxSpansPerDim();

  //for display
  //bool has_gap = 0;

  //get precedence array for this dim
  PAV pav = arrg.second;

  //
  // ******** 'Bookkeepping' arrays **********
  // NOTES:
  // 1. indexes start at zero, not one
  // 2. in comments 'order' refers to span's position in DSpan.
  //    ie, DSpan:[1 4 7 8] consists { 0th span(4-5), 1st span(7-8) }

  // otherIdc[i] =
  //   index of i-th span of 'other' in prec array, or
  //  -1, if i-th span does not exist
  int * otherIdc = new int[this->maxSpansPerDim()];

  // toLeft[i] =
  //  order of 'me' span to the left of other's i-th span
  //  -1, if i-th span does not exist,
  //         or leftmost span is not of me
  //         or no span found to left of other's i-th span
  int * toLeft   = new int[this->maxSpansPerDim()];

  // toRight[i] =
  //  order of 'me' span to the right of other's i-th span
  //  -1, if i-th span does not exist,
  //         or rightmost span is not of me
  //         or no span found to right of other's i-th span
  int * toRight  = new int[this->maxSpansPerDim()];

  // gapLeft[i] =
  //  1,  if gap found to right of other's i-th span
  //  0,  if NO gap found to right of other's i-th span
  // -1,  if ith span does not exist or is leftmost span
  int * gapLeft   = new int[this->maxSpansPerDim()];

  // gapRight[i] =
  //  1,  if gap found to right of other's i-th span
  //  0,  if NO gap found to right of other's i-th span
  // -1,  if ith span does not exist
  int * gapRight  = new int[this->maxSpansPerDim()];

  // EXAMPLE:
  //  MaxSpansPerDim = 2
  //  Prec Array: 0 [ 1 2 0 2 1 ] ~ [other me gap me other]
  //          otherIdc = [0, 4]
  //          toLeft   = [-1, 1]
  //          toRight  = [0, -1]
  //          gapLeft  = [-1, 0]
  //          gapRight = [0, -1]
  // *******************************************

  for(arity_t d = 0; d < this->arity(); d++){
    //declare vars
    DSpan dspan = (*dsv)[d];

    // precedence Indices, ie, indices into precedence array
    // otherIdc[i] == index of others i'th  span in prec array

    for(int i = 0; i < this->maxSpansPerDim(); i++)
      otherIdc[i] = -1;

    for(int i = 0; i < this->maxSpansPerDim(); i++) {
      toRight[i] = -1;
      toLeft[i]  = -1;
      gapRight[i]  = -1;
      gapLeft[i]   = -1;
    }

    //check fanout of dspan
    assert(dspan.size() <= this->maxSpansPerDim());

    //PrecArray pa = (arrg.second)[d];
    PrecArray pa = pav.getPA(d);//= (arrg.second)[d];

    int pa_count_me = 0, pa_count_other = 0, pa_count_gap  = 0;
    int last_type= -1; //me or other, init to neither

    //populate bookkeeping arrays
    for (unsigned short i = 0; i<pa.size(); i++) {
      if(pa[i]==0) {
        pa_count_gap++;
      } else if(pa[i]==me) {
        //check if there is an other span to the left
        if( last_type == other) {
          toRight[pa_count_other-1] = pa_count_me; //may be -1
          gapRight[pa_count_other-1] = (pa[i-1]==0)?1:0;
        }

        last_type = me;
        pa_count_me++;
      } else if(pa[i]==other) {

        otherIdc[pa_count_other] = i;

        if( last_type == me) {
          toLeft[pa_count_other] = pa_count_me - 1; //may be -1
          gapLeft[pa_count_other] = (pa[i-1]==0)?1:0;
        }

        last_type = other;
        pa_count_other++;
      } else assert(0);

      if(i>0) assert( pa[i] != pa[i-1] );
    }
    /*
    //Printouts for debugging ---
    cout<<"otherIdc:\t"<<endl;
    for(int i = 0; i < this->maxSpansPerDim(); i++) cout<<"  i:"<<i<<" "<<otherIdc[i]<<endl;
    cout<<"toLeft:"<<endl;
    for(int i = 0; i < this->maxSpansPerDim(); i++) cout<<"  i:"<<i<<" "<<toLeft[i]<<endl;
    cout<<"toRight:"<<endl;
    for(int i = 0; i < this->maxSpansPerDim(); i++) cout<<"  i:"<<i<<" "<<toRight[i]<<endl;
    cout<<"gapLeft:"<<endl;
    for(int i = 0; i < this->maxSpansPerDim(); i++) cout<<"  i:"<<i<<" "<<gapLeft[i]<<endl;
    cout<<"gapRigth:"<<endl;
    for(int i = 0; i < this->maxSpansPerDim(); i++) cout<<"  i:"<<i<<" "<<gapRight[i]<<endl;
    */

    //for display
    //if(pa_count_gap>0) has_gap=1;

    assert(pa_count_me   <= maxSpansPerDim() );
    assert(pa_count_other<= maxSpansPerDim() );
    assert(pa_count_gap  <= maxSpansPerDim()-1 ); //at most one gap

    //build hyperrectangle
    for(int sp = 0; sp<maxSpansPerDim(); sp++) {
      int precIdx =  otherIdc[sp];
      if(precIdx == -1) {
        res[ d*offset + 2*sp +0] = Range(DEGEN,DEGEN);
        res[ d*offset + 2*sp +1] = Range(DEGEN,DEGEN);
      } else if ( d >= m_inputArity ) {
        res[ d*offset + 2*sp +0] = Range(0,0);
        res[ d*offset + 2*sp +1] = Range(0,0);
      } else if (precIdx == 0) {//1st span in leftmost position
        assert(sp == 0);

        int r1 = UNBOUND,r2 = UNBOUND;
        if(toRight[0]!=-1) {
          Span my_span = dspan[0];
          r2 = my_span.first;
          if(gapRight[0])
	    {  if(r2>0)r2--; }
          else
            r1 = r2;
        }

        res[ d*offset + 0] = Range(0,UNBOUND); //use l1=0 to ensure span not degen
        res[ d*offset + 1] = Range(r1,r2);

      } else {
        int l1 = 0, l2 = UNBOUND; //use l1=0 to ensure span not degen
        int r1 = UNBOUND, r2 = UNBOUND;

        if(toLeft[sp]!=-1) {
          Span my_span = dspan[ toLeft[sp] ];
          l1 = my_span.second;
          if(gapLeft[sp])
            l1++;
          else
            l2=l1;
        }
        if(toRight[sp]!=-1) {
          Span my_span = dspan[ toRight[sp] ];
          r2 = my_span.first;
          if(gapRight[sp])
	    {  if(r2>0)r2--; }
          else
            r1 = r2;
        }

        res[ d*offset + sp*2 + 0] = Range(l1,l2);
        res[ d*offset + sp*2 + 1] = Range(r1,r2);

      }
    }
  }

  //dealloc bookkeepping arrays
  delete otherIdc;
  delete toRight;
  delete toLeft;
  delete gapRight;
  delete gapLeft;

  if(debug(5))
    dout() << "  Query HyperRectangle: " << res << endl;

  //Do the actual query!
  //RangeIter iter = m_discChart.query(res);
  RangeIter iter = this->queryHRect(res);
  int match_count = 0;
  while(iter.has_next()) {
    match_count++;
    matchedItems.push_back((Item*)(*iter));
    if (debug(5)) {
      dout() << "MATCHED ITEM: " << **iter << endl;
    }
    ++iter;
  }
  if(debug(5))
    dout() << "Number of MATCHED ITEMS : " << match_count<< endl;
  //  if(debug(6) && match_count && has_gap)
  //    dout() << " * MATCHED WITH GAP * " << endl;

  return;
}

#endif
