#ifndef _Grammar_H_
#define _Grammar_H_

#include <set>
#include "Dbug.hpp"
#include "UniqueID.hpp"
#include "GPDefs.hpp"
#include "Exceptions.hpp"
#include "StringVocabulary.hpp"
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/is_abstract.hpp>

class Inference;
class SentenceTuple;

using namespace std;

//! Abstract grammar class
class Grammar : public Dbug, public UniqueID {

public:
  //! constructors
  Grammar(arity_t dim);

  // added for serialization
  Grammar() {};

  //! destructor
  virtual ~Grammar();

  //! assignment
  Grammar& operator=(const Grammar& other);

  virtual arity_t arity() const;
  virtual size_t size() const;

  //! Returns the terminal sets
  virtual StringVocabulary* terminalSet();
  virtual const StringVocabulary* terminalSet() const;
  //! Returns the sets of nonterminals used in this grammar.
  virtual StringVocabulary* ntSet();
  virtual const StringVocabulary* ntSet() const;

  // update the grammar based on input, e.g. for caching
  // by default, do nothing
  virtual bool prepareFor(SentenceTuple& tuple) {
    return true;
  };

  // notify the grammar that the end is nigh, in case it wants to save any state
  virtual void clear();

  //! determine whether a certain inference should be allowed
  virtual bool allowInf(BSP<Inference> spInf) { return false;}

  ////////////////////Input/Output////////////////////
  // Even though the Grammar base class has nothing for I/O, these
  // methods are needed for polymorphic I/O of subclasses.

  //! Read.
  virtual istream& get(istream& is) {
    // does nothing by default
    return is;
  }

  //! Write
  virtual ostream& put(ostream& os) const {
    // does nothing by default
    return os;
  }

  //! Adapter of read.
  friend istream& operator >> (istream& in, Grammar& gr)
  { return gr.get(in); }

  //! Adapter of write.
  friend ostream& operator << (ostream& out, const Grammar& gr)
  { return gr.put(out); }

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    ar & boost::serialization::base_object<Dbug>(*this);
    ar & m_arity;
  };

protected:
  //! Grammar dimensionality.
  arity_t m_arity;

};

BOOST_IS_ABSTRACT(Grammar);

#endif
