#ifdef USE_SERIALIZATION
#include "ArchiveMaker.hpp"
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#endif

#include "File.hpp"
#include "WMTG.hpp"
#include "HeadPMTGA.hpp"
#include "HeadPMTGB.hpp"
#include "HeadWMTGB.hpp"
#include "IndexedPMTG.hpp"
#include "TreeConstrainedMTG.hpp"
#include "CSTreeConstrainedMTG.hpp"
#include "PseudoMTG.hpp"
#include "FTreeTransducer.hpp"
#include "CSPseudoMTG.hpp"
#include "FeatureWMTG.hpp"
//#include "FSMWMTG.hpp"
#include "PartialWMTG.hpp"
#include "CompositeWMTGA.hpp"
#include "CompositeWMTGB.hpp"
#include "BigramGrammar.hpp"
#include "Exceptions.hpp"
#include "DebugMacros.hpp"
#include "OptionsPrinter.hpp"
#include <boost/program_options.hpp>
#include "GrammarBuilder.hpp"
#include "Path.hpp"

#ifdef USE_SERIALIZATION
#include <boost/serialization/serialization.hpp>
#endif

GrammarBuilder* GrammarBuilder::_instance = 0;

GrammarBuilder* GrammarBuilder::Instance(){
  if(_instance == 0){
    _instance = new GrammarBuilder;
  }
  return _instance;
}

//! build grammar
Grammar* GrammarBuilder::build(const string& configFile,
                               StringVocabulary* synCatVcb,
                               StringVocabulary* tvcb,
                               char** overrideArgV,
                               size_t overrideArgC) {
  string grammarType;
  string grammarFileName;
  string grammarFileType;
  unsigned int arity = 2;

  // nonterminal vocabulary files.
  string optNTVcbFiles;
  // terminal vocabulary files.
  string optTVcbFiles;

  string optGrammarConfigFile;

  BPO::options_description config("Grammar config file options");
  config.add_options()
    ("nonterminals,n", BPO::value<string>(&optNTVcbFiles), "Control file for input nonterminal vocabs")
    ("terminals,t", BPO::value<string>(&optTVcbFiles), "Control file for input terminal vocabs")
    ("GrammarType", BPO::value<string>(&grammarType), "The grammar type")
    ("GrammarFileName", BPO::value<string>(&grammarFileName), "The grammar file name")
    ("fileType", BPO::value<string>(&grammarFileType)->default_value("txt"), "The grammar file type")
    ("grammar-config", BPO::value<string>(&optGrammarConfigFile), "The grammar config file")
    ("D", BPO::value<unsigned int>(&arity)->default_value(2), "The dimension of the grammar")
    ;

  BPO::variables_map vm;

  if(overrideArgV){
    if(debug(1)){
      cerr<<"GrammarBuilder: Overriding config file arguments..."<<endl;
    }
    store(BPO::command_line_parser(overrideArgC, overrideArgV).
	  options(config).run(), vm);
  }

  if(configFile != ""){
    cerr << "GrammarBuilder: Compiling config file " << configFile << endl;
    iFile ifs(configFile.c_str());
    store(parse_config_file(ifs, config), vm);
  }

  notify(vm);

  if(debug(1)) {
    cerr << OptionsPrinter(config, vm) << endl;
  }

  if (grammarType == "") {
    THROW("No grammar type specified!" << endl);
  }

  if (! arity) {
    THROW("Grammar dimensionlity must be > 0!" << endl);
  }

  vector<string> fileList;
  vector<string>::iterator itFileList;

  // Load NT vocabs, if desired and not already present
  if (synCatVcb == NULL && !GlobalVocab::s_noVocabsIn) {
    StringVocabulary* vocab = new StringVocabulary;
    if(optNTVcbFiles != "") {
      fileList = split(optNTVcbFiles,"|");
      for(itFileList = fileList.begin(); itFileList != fileList.end(); ++itFileList){
	iFile iF(itFileList->c_str());
	iF >> *vocab;
	iF.close();
      }
      GlobalVocab::s_synCatSet = vocab;
    }
  }

  // Load the terminal vocabs, if desired and not already present
  if (tvcb == NULL && !GlobalVocab::s_noVocabsIn) {
    StringVocabulary* vocab = new StringVocabulary;
    if(optTVcbFiles != ""){
      fileList = split(optTVcbFiles, "|");
      for(itFileList = fileList.begin(); itFileList != fileList.end();
	  ++itFileList){
	iFile iF(itFileList->c_str());
	iF >> *vocab;
	iF.close();
      }
      GlobalVocab::s_terminalSet = vocab;
    }
  }

  if(optGrammarConfigFile == ""){
    THROW("Option --grammar-config needed. Try --help" << endl);
  }

  // string grammarConfigFile = Path::maybe_compose(configFile, optGrammarConfigFile);
  string grammarConfigFile = optGrammarConfigFile;

  Grammar* pGrammar = buildOne(grammarType, arity, grammarConfigFile, 
			      overrideArgV, overrideArgC, configFile);

  DEBUG(4, "Constructed " << grammarType << endl);

  if(grammarFileName != "" ){
    if (grammarFileType == "txt") {
      iFile grammarF(grammarFileName);
      grammarF >> * pGrammar;
      grammarF.close();
    }
#ifdef USE_SERIALIZATION
    else {
      LearnableGrammar* lgrammar;
      iFile ifs(grammarFileName.c_str());

      if (grammarFileType == "bar") {
	boost::archive::binary_iarchive ia(ifs);
	ArchiveMaker am;
	return(am.readBinaryArchive(ia, lgrammar)); }
      else if (grammarFileType == "tar") {
        boost::archive::text_iarchive ia(ifs);
	ArchiveMaker am;
	return(am.readTextArchive(ia, lgrammar)); }
      else {
	ifs.close();
	THROW("Error: unknown grammar file type!" << endl);
      }
    }
#endif
  }

  return pGrammar;
}

Grammar* GrammarBuilder::buildOne(const string& grammarType, arity_t arity, 
				  string& grammarConfigFile,
				  char** overrideArgV, size_t overrideArgC, 
				  const string& origConfig) {
  Grammar* pGrammar;

  //if it's not a composite (made up of many subfiles), then add a path.
  if(grammarType != "PseudoMTG" && grammarType != "CSPseudoMTG"){
    grammarConfigFile = Path::maybe_compose(origConfig, grammarConfigFile);
  }

  //check to see type.
  if (grammarType == "PseudoMTG" || grammarType == "CSPseudoMTG") {
    //this is a composite. Split the file name into the config file
    // and the sub-config files.
    vector<string> strvec = split(grammarConfigFile, "|");
    string configtm = Path::maybe_compose(origConfig,strvec[0]);
    PseudoMTG* pPsMTG;
    if (grammarType == "PseudoMTG")
      pPsMTG = new PseudoMTG(arity,configtm);
    else if (grammarType == "CSPseudoMTG")
      pPsMTG = new CSPseudoMTG(arity,configtm);
    else
      pPsMTG = 0; // duck compiler warning
    //get subconfig files, and build appropriate grammars from them.
    for (unsigned int i = 1; i < strvec.size(); i++) {
      vector<string> goalDefine = split(strvec[i], ",");
      if (goalDefine.size() != 2) {
	THROW("The config file is incorrectly formatted."
	      << " It should be of the form: "
	      << "\"PsMTGconfigFile|TreeConstrainedMTG,tcConfigFileName|PartialWMTG,emptyMTGConfigFileName [...] \""
	      << endl);
      }
      Grammar* subGrammar = buildOne(goalDefine[0], arity, goalDefine[1],
				     overrideArgV, overrideArgC, origConfig);
      pPsMTG->add(subGrammar);
    }
    pGrammar = pPsMTG;
  } else if (grammarType == "TreeConstrainedMTG") {
    pGrammar = new TreeConstrainedMTG(arity, grammarConfigFile);
  } else if (grammarType == "CSTreeConstrainedMTG") {
    pGrammar = new CSTreeConstrainedMTG(arity, grammarConfigFile);
  } else if (grammarType == "IndexedWMTG") {
    pGrammar = new IndexedWMTG(arity, grammarConfigFile);
  } else if (grammarType == "IndexedPMTG") {
    pGrammar = new IndexedPMTG(arity, grammarConfigFile);
  } else if (grammarType == "FeatureWMTG") {
    pGrammar = new FeatureWMTG(arity, grammarConfigFile);
  } else if (grammarType == "CompositeWMTGA") {
    pGrammar = new CompositeWMTGA(arity, grammarConfigFile);
  } else if (grammarType == "CompositeWMTGB") {
    pGrammar = new CompositeWMTGB(arity, grammarConfigFile);
  } else if (grammarType == "FTreeTransducer") {
    pGrammar = new FTreeTransducer(arity, grammarConfigFile);
  } else if (grammarType == "BigramGrammar") {
    pGrammar = new BigramGrammar(grammarConfigFile);
  } else if (grammarType == "NaiveWMTG") {
    pGrammar = new NaiveWMTG(arity, grammarConfigFile, overrideArgV, overrideArgC);
  } else if (grammarType == "HeadPMTGA") {
    pGrammar = new HeadPMTGA(arity, grammarConfigFile, overrideArgV, overrideArgC);
  } else if (grammarType == "HeadPMTGB") {
    pGrammar = new HeadPMTGB(arity, grammarConfigFile, overrideArgV, overrideArgC);
  } else if (grammarType == "HeadWMTGB") {
    pGrammar = new HeadWMTGB(arity, grammarConfigFile, overrideArgV, overrideArgC);
}    else if (grammarType == "PartialWMTG") {
    pGrammar = new PartialWMTG(arity,grammarConfigFile);
    //  } else if (grammarType == "FSMWMTG") {
    // pGrammar = new FSMWMTG(arity, grammarConfigFile);
  }else {
    THROW("Error: unknown grammar type!"<<endl);
  };
  return pGrammar;
}
