#include "GrammarIndex.hpp"
#include "TLink.hpp"
#include "CrossProduct.hpp"
#include "Exceptions.hpp"
#include "DebugMacros.hpp"
#include "Hedge.hpp"

arity_t GrammarIndex::arity() const {
  return m_arity;
}

//! Insert a production
void GrammarIndex::insert(Production* pr) {

  DEBUG(5, "GrammarIndex insert " << *pr << ")" << endl);

  arity_t dd;

  if(pr->rank() > 2){
    cerr<<"Sorry, only binarized grammar is supported."<<endl;
    cerr << *pr << endl;
    THROW("rank > 2");
  }

  /*
   * Word/Pos-Tag. Compiled only from terminal productions
   * (In GCNF, rank() == 1 indicates that it is a terminal production).
   */
  if(pr->rank() == 1) {
    // add the 1D word-PoS mappings.
    const NTLink ntl = pr->lhs();
    const TLink l = boost::get<const TLink>(pr->rhsLinks()[0]);
    for(dd = 0; dd < l.arity(); ++dd){
      Nonterminal* nt = ntl.getNT(dd)->clone();
      if (m_word2Tags[dd][l[dd]].find(nt)==m_word2Tags[dd][l[dd]].end()){
	m_word2Tags[dd][l[dd]].insert(nt);
      }
      else{
	delete nt;
      }
    }
  }
  
  /*Binary*/
  if(pr->rank() > 1){
    /*
     * Immediate dominance
     */
    insertImDom(pr);

    // w2w dictionary
    W2WLink lk;
    if (pr->getTLink(lk))
      m_spW2WDict->insert(lk);

#ifdef USE_ARMT_INDEX

    /*
     * Insert the permutation events
     */
    /*
     * Two labels in linear order-> If it is a unary production,
     * the second label is undefined.
     */

    /*the first label must be role 1, and the production should be
     * binary at this point*/
    NTLink lab1 = boost::get<NTLink>(pr->rhsLinks()[0]);
    NTLink lab2 = boost::get<NTLink>(pr->rhsLinks()[1]);

    /* Index entry taking the first label at RHS. */
    m_arrgs[Pair<bool, NTLink>(true, lab1)].insert(pr->rhsPerm());

    if(debug(6))
      cerr << "GrammarIndex inserting " << (pr->rhsPerm())
	   << " at " << true << " " << lab1 << ", " << lab2 << endl;

    /*Index entry taking the second label at RHS*/
    bool role_of_second_label = false;
    /*if the PAV in the production is undefined, then
     * the role of the second label is also true*/
    if(pr->rhsPerm().empty()) {
      role_of_second_label = true;
    }
    m_arrgs[Pair<bool, NTLink>(role_of_second_label, lab2)].insert(pr->rhsPerm());

    if(debug(6))
      cerr << "GrammarIndex: inserting " << (pr->rhsPerm())
	   << " at " << role_of_second_label << " "
	   << lab2 << ", " << lab1 << endl;

#endif
  }
}


//! Returns the possible parent. (Unheaded case)
void GrammarIndex:: getParents(const NTLink& link1,
	                       const NTLink& link2,
			       const PAV& pav,
			       set<NTLink >& parents) const {
  //create some fake heirRoleVectors and use existing index.
  HeirRoleVector hv();
  
  hash_map<Pair<NTLink, NTLink>,
    Set< Pair<PAV, Pair<HeirRoleVector, NTLink> > > > ::const_iterator i;
  
  //bool to say whether we switched the order of heir and dep
  // in the look up.  This is because the imDom is stored
  // lexicographically.
  bool switched = false;
  
  //check for switch or not
  if(!(link1 < link2))
    i = m_imDom.find(Pair<NTLink, NTLink>(link1, link2)); 
  else{
    switched = true;
    i = m_imDom.find(Pair<NTLink, NTLink>(link2, link1)); 
  }

  //perform look up of ordered pair.
 
  Set<Pair< PAV, Pair<HeirRoleVector, NTLink> > >::const_iterator j;
  for(j = i->second.begin(); j != i->second.end(); ++j){
    if(!switched){
      parents.insert(j->second.second);
    }
  }
}

//! Returns the possible parent by query the ID table/
//! This function will replace the function above.
void GrammarIndex:: getParents(const NTLink& link1,
	                       const NTLink& link2,
			       const PAV& pav,
			       set< Pair<HeirRoleVector, NTLink> >& parents) const {

  hash_map<Pair<NTLink, NTLink>,
    Set< Pair<PAV, Pair<HeirRoleVector, NTLink> > > > ::const_iterator i;

  //bool to say whether we switched the order of heir and dep
  // in the look up. This is because the imDom is stored
  // lexicographically.
  bool switched = false;

  //check for switch or not
  Pair<NTLink, NTLink> pe;
  if(!(link1 < link2))
    pe = Pair<NTLink, NTLink>(link1, link2);
  else{
    switched = true;
    pe = Pair<NTLink, NTLink>(link2, link1);
  }

  if(debug(5))
    cerr << "GI:FIND: " << pe << endl;
  //perform look up of ordered pair.
  i = m_imDom.find(pe);

  //if one is found
  if(i != m_imDom.end()){
    if(debug(5))
      cerr << "GI:FOUND: " << *i << endl;
    Set<Pair< PAV, Pair<HeirRoleVector, NTLink> > >::const_iterator j;
    for(j = i->second.begin(); j != i->second.end(); ++j){
      HeirRoleVector hv(j->second.first);
      NTLink lntl;
      if(dynamic_cast<const LexNT*>(link1.getNT(0))){
	TLink l(arity());
	//Tlink l needs to hold the terminals pointed to by the returned
	//HeirRoleVector.
	for(size_t i = 0; i < arity(); i++){
	  if(hv.getRole(i) == 1){
	    if(!switched){
	      l[i] = link1.getTLink()[i];
	    }
	    else{
	      l[i] = link2.getTLink()[i];}
	  }
	  else if(hv.getRole(i) == 2){
	    if(!switched)
	      { l[i] = link2.getTLink()[i];}
	    else
	      {l[i] = link1.getTLink()[i];}
	  }
	  else if(hv.getRole(i) == 0)
	    l[i] = UNDEF_T;
	}
	lntl = NTLink(j->second.second, l);
      }
      else{
	lntl = NTLink(j->second.second);
      }

      if(!switched){
	//insert HeirRoleVector that was returned.
	if(j->first == pav){
	  Pair<HeirRoleVector, NTLink> pe2((*j).second.first,lntl);
	  parents.insert(pe2);
	  if(debug(5))	cerr << "GI:inserting: " << pe2 << endl;
	}
      }
      else{
	if(PAV::swapElement(j->first) == pav){
	  //insert inverted HeirRoleVector, since we switched the inputs.
	  Pair<HeirRoleVector, NTLink> pe2((*j).second.first.invert(),lntl);
	  parents.insert(pe2);
	  if(debug(5))	cerr << "GI:inserting.b: " << pe2 << endl;
	}
      }
    }
  }
}

void GrammarIndex::getParents(const TLink& tl, set<NTLink>& wordTags) const {

  hash_map<Terminal, Set<Nonterminal*, less<Nonterminal*> > > :: const_iterator i;
  arity_t d;

  vector< Set<Nonterminal*, less<Nonterminal*> > > ntsetvec(arity());

  for(d = 0; d < arity(); ++d){
    i = m_word2Tags[d].find(tl[d]);
    if(i != m_word2Tags[d].end()) {
      ntsetvec[d] = i->second;
    }
  }

  NTLink::hyperCrossProduct(ntsetvec, wordTags);
}

#ifdef USE_ARMT_INDEX
void GrammarIndex::getArrangementFor(BSP<set<Pair<bool,PAV> > >& arrgs, const Item* trigger) const {

  const Hedge* pH = dynamic_cast<const Hedge*>(trigger);
  const NTLink& label = pH->getNTLink();

  if(debug(5)){
    cerr << "GrammarIndex::getArrangementFor " << *pH << endl;
  }

  google::sparse_hash_map<Pair<bool, NTLink>, google::sparse_hash_set<PAV> >::const_iterator i;

  /*if the input label is role 1*/
  i = m_arrgs.find(Pair<bool, NTLink>(true, label));
  if(i != m_arrgs.end()){
    google::sparse_hash_set<PAV> ::const_iterator j;
    for(j = i->second.begin(); j != i->second.end(); ++j){
      if(debug(5)) cerr << "found: " << *j << endl;
      if(j->getFanoutV(1) == trigger->getFanoutV()){
	if(debug(5)) cerr << "added: " << *j << endl;
	arrgs->insert(Pair<bool,PAV>(true, *j));
      }
    }
  }

  /*if the input label is role 2*/
  i = m_arrgs.find(Pair<bool, NTLink>(false, label));
  if(i != m_arrgs.end()){
    google::sparse_hash_set<PAV>::const_iterator k;
    for(k = i->second.begin(); k != i->second.end(); ++k){
      if(debug(5)) cerr << "found: " << *k << endl;
      if(k->getFanoutV(2) == trigger->getFanoutV()){
	if(debug(5)) cerr << "added: " << *k << endl;
	arrgs->insert(Pair<bool,PAV>(false, *k));
      }
    }
  }
}
#endif

// insert the Immediate Dominance information in the input
// rule into the index.
void GrammarIndex::insertImDom(const Production* pr) {
  
  const HeadProduction* pHPr = dynamic_cast<const HeadProduction*>(pr);
  if (not pHPr) 
    THROW("Big Oops");

  HeirRoleVector hrv = pHPr->heirRole();
  //  HeirRoleVector hrv = ((HeadProduction*)(pr))->heirRole();
  
  if(pr->rank() == 2){
    //link 1 is left link and link2 is right link in pr->
    NTLink link1 = boost::get<NTLink>(pr->rhsLinks()[0]);
    NTLink link2 = boost::get<NTLink>(pr->rhsLinks()[1]);
    
    //store them in sorted order-> If they are equal, store in both orders.
    if( not (link1 < link2)){
      Pair<NTLink, NTLink> rhsLinks(link1,link2);
      Pair<HeirRoleVector, NTLink> pe2(hrv, pr->lhs());
      if(debug(6)){
	cerr << "GI:insert into imdom: " << pe2 << " at " << rhsLinks << endl;
      }
      Pair<PAV, Pair<HeirRoleVector, NTLink> > pe3(pr->rhsPerm(),pe2);
      m_imDom[rhsLinks].insert(pe3);
    } else {
      Pair<NTLink, NTLink> rhsLinks(link2, link1);
      Pair<HeirRoleVector, NTLink> pe2(hrv.invert(), pr->lhs());
      if(debug(6)){
	cerr << "GI: insert into imdom: (inv)"
	     << pe2 << " at " << rhsLinks << endl;
      }
      Pair<PAV, Pair<HeirRoleVector, NTLink> > pe3(PAV::swapElement(pr->rhsPerm()),pe2);
      m_imDom[rhsLinks].insert(pe3);
    }
  }
}

void GrammarIndex::clear() {
  m_imDom.clear();

  //delete all the nonterminals
  vector<hash_map<Terminal, Set<Nonterminal*, less<Nonterminal*> > > >::iterator i;
  for(i = m_word2Tags.begin(); i != m_word2Tags.end(); i++){
    hash_map<Terminal, Set<Nonterminal*, less<Nonterminal*> > >::iterator j;
    for(j = i->begin(); j != i->end(); j++){
      Set<Nonterminal*, less<Nonterminal*> >::iterator k;
      for(k = j->second.begin(); k != j->second.end(); k++){
	if(*k != NULL){
	  delete *k;
	}
      }
    }
  }
  for(arity_t dd = 0; dd < arity(); ++dd)
    m_word2Tags[dd].clear();
  
#ifdef USE_ARMT_INDEX
  m_arrgs.clear();
#endif
};

