/*
 * The index of a grammar, e.g., the map
 * from the dependent label to its possible
 * parent labels; the map from word to tags.
 */

#ifndef _GrammarIndex_H_
#define _GrammarIndex_H_

#include "Set.hpp"
#include "StringVocabulary.hpp"
#include "HeirRoleVector.hpp"
#include "DSV.hpp"
#include <vector>
#include <set>
#include <map>
#include "MultiMap.hpp"
#include <google/sparse_hash_map>
#include <google/sparse_hash_set>
#include "HeadProduction.hpp"
#include "Vocabulary.hpp"
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/hash_map.hpp>
#include <boost/serialization/hash_set.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/shared_ptr.hpp>
#include "Dbug.hpp"

using namespace std;

class Item;

#define USE_ARMT_INDEX

/**
 * The index of a grammar, e.g., the map
 * from the dependent label to its possible
 * parent labels; the map from word to tags.
 */
class GrammarIndex : public Dbug {
public:

  //! constructor
  GrammarIndex(const arity_t d) :
    m_arity(d), m_word2Tags(d), m_spW2WDict(new W2WDictionary)
  {}

  // added for serialization
  GrammarIndex() {};

  //! destructor
  ~GrammarIndex() {}

  //! Insert a production.
  void insert(Production* );

  //! Returns the possible parent. (unheaded case)
  void getParents(const NTLink& link1,
			  const NTLink& link2,
			  const PAV& pav,
			  set<NTLink >& parents) const;

  //! Returns the possible parent.
  void getParents(const NTLink& heir, const NTLink& dep,
			  const PAV& pav,
			  set<Pair<HeirRoleVector, NTLink> >& parents) const;

  void getParents(const TLink&, set<NTLink>& wordTags) const;

  #ifdef USE_ARMT_INDEX
  void getArrangementFor(BSP<set<Pair<bool,PAV> > >& arrgs, const Item* trigger) const;
  #endif

  arity_t arity() const ;

  // getter for w2w dict
  BSP<W2WDictionary> getW2WDict() {
    return m_spW2WDict;
  }

  void clear();

protected: // methods

  //! Insert the immediate dominance information into the
  //! index.
  void insertImDom(const Production* );

private: // methods

  friend class boost::serialization::access;

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    ar & boost::serialization::base_object<Dbug>(*this);
    ar & m_imDom;
    ar & m_word2Tags;
    #ifdef USE_ARMT_INDEX
    ar & m_arrgs;
    #endif
    ar & m_arity;
  };

protected: // data

  //! Dimension of the Grammar Index.
  arity_t m_arity;

  //! The immediate dominance. Map from the pair of heir and dependent
  //! to the possible parents.
  hash_map<Pair<NTLink , NTLink >,
	   Set<Pair<PAV, Pair<HeirRoleVector,NTLink> > > > m_imDom;

  //! D Word/POS-Tag table.
  vector<hash_map<Terminal, Set<Nonterminal*, less<Nonterminal*> > > > m_word2Tags;

  #ifdef USE_ARMT_INDEX
  /// Map from pair of NTLinks to Arrangements.  Note that the first
  /// NTLink in the pair refers to the heir constituent label.
  google::sparse_hash_map<Pair<bool /*role*/, NTLink /*label*/>, 
			  google::sparse_hash_set<PAV> > m_arrgs;
  #endif

  /// W2W dictionary
  BSP<W2WDictionary> m_spW2WDict;

};

#endif
