#ifndef _HashChart_H_
#define _HashChart_H_

#include "COMPILER-DEPENDENT_HEADERS.hpp"

#include <list>

#include "Hedge.hpp"
#include "ItemSet.hpp"
#include "Corner.hpp"
#include <ext/hash_set>
#include "MoreAlgorithms.hpp"

#ifdef BOOST_MEM_POOL
#include "BoostMemPool.hpp"
#endif

using namespace std;
using namespace HashNS;

/// HashChart contains only Hedges
class HashChart : public ItemSt {
  typedef ItemSt _super;

  typedef BSP<Item> ItemPtr;

  /// Type of the item table.
#ifdef BOOST_MEM_POOL
  typedef hash_set<ItemPtr,
                   HashNS::hash<ItemPtr>,
		   equal_to_ptr<ItemPtr>,
		   boost::decorated_fast_pool_allocator<ItemPtr> >
  item_table_type;
#else
  typedef hash_set<ItemPtr,
                   HashNS::hash<ItemPtr>,
		   equal_to_ptr<ItemPtr> > item_table_type;
#endif

  /// Iterator over the contained hedges.
  typedef item_table_type::iterator iterator;
  typedef item_table_type::const_iterator const_iterator;

  /// items in the same chart cell
  typedef list<ItemPtr> cell_t;

  /// The type of the map from the corner with the direction in which
  /// the matching items will be located to the relevant chart cell.
  typedef hash_map<Corners, cell_t> matchingIndex_t;

public:
  /// Constructors
  HashChart(const arity_t& dim);
  HashChart(const arity_t& dim, const bool so);

  // Clone a HashChart without copying the content.
  HashChart* memorylessClone() const;

  /// Destructor.
  ~HashChart();

  /// Inserts an item, returns the pointer to the inserted item in chart.
  /// \param item : the item to be inserted.
  /// \param memoRes : the result of the insertion.
  /// - INSERT: the item inserted by this method is new;
  /// - UPDATE: there is an item with the same signature already
  /// in chart, and the inside score of the inserted
  /// item is larger than the old one;
  /// - NOEFFECT: there is an item with the same signature
  /// already in chart, and has larger inside score;
  /// - IGNORE: this item is ignored. I.e. because of having
  /// minimum infinity score.
  /// - ERROR : unrecognized item type.
  /// \param arrgs : the arrangements compatible with this item, according to the grammar.
  /// Returns pointer to inserted chart item, if any.  
  /// Original input item always deleted.
  /// That way, the caller can set inferences to have antecedents that
  /// are in the chart.
  Item* insert(Item * item, MemoResult &memoRes, const set< Pair<bool,PAV> >& arrgs);

  /// Finds the matched items for the input Hedge.
  /// \param trigger : the Hedge (trigger) to be matched.
  /// \param arrg : the PAV that governs the matching; the bool is
  /// true if the trigger should have role 1 in the PAV, and false if
  /// the trigger should have role 2
  void matchItem(const Item &trigger, const Pair<bool,PAV>& arrg, vector<Item*>& matchedItems);

  /// export all items in the chart
  void exportAll(Vector<const Item*>& VpI) const;

  /// Clears the content.
  virtual void clear();

  /// Prints the statistics of this class.
  virtual void printStats(ostream& out) const;

protected: // data

  /// index used for finding "matching items"
  matchingIndex_t m_matchingIndex;
};

#endif

