/// \brief Headed Multitree

#ifndef _HeadMTree_H_
#define _HeadMTree_H_

#include <set>
#include "tree.hpp"
#include "HeirTree.hpp"
#include "SpanTree.hpp"
#include "HeadMTreeNode.hpp"
#include "MTree.hpp"
#include "PAVTree.hpp"
#include "NTLink.hpp"
#include "Histogram.hpp"
#include "DSV.hpp"
#include "SentenceTuple.hpp"
#include "boost/tuple/tuple.hpp"
#include "RHSElem.hpp"

class HeadMTree : public HeirTree<MTree> {
public:

  enum HeirLayout {NODELTAS, CROSSWAYS, WHOLEDIM, ONEDELTA};

  virtual ~HeadMTree(){};

  //! Input of a multitree.
  friend istream& operator>>(istream&, HeadMTree&);

  //! Output of a multitree.
  friend ostream& operator<<(ostream&, const HeadMTree&);
  friend ostream& operator<<(ostream&, HeadMTree&);

  //! Prints a subtree.
  virtual void print(ostream& out, const MTree* tr, iterator it) const;

  //! reverse the dimensions.
  void flipTree();
  void flipTree_helper(iterator it);

  /// put the PAVs in canonical form, switching children if necessary
  void normalize();
  void normalizeHelper(iterator it);

  //! extract the TLinks from this tree and return them in a histogram
  void extractTLinks(Histogram<TLink>& w2wHist);
  void extractTLinks_helper(iterator it, Histogram<TLink>& w2wHist);

  static void twoSidedDiff(const WTreeVec& vecGold, 
			   const WTreeVec& vecCandidate, 
			   vector<iterator>& goldNotCandidate,
			   vector<iterator>& candidateNotGold, 
			   vector<iterator>& vecIntersection,
			   bool returnW2Wnodes = true,
			   bool filterNoise = false);
 
  void exportToHash(hash_map< Pair<DSV,Vector<int> > , 
		    iterator >& hmap,
		    hash_map<DSV,Vector<int> >& hmap2,
		    hash_map<Pair<DSV,Vector<int> >, unsigned int> &, 
		    hash_set<Vector<int> > & yields,
		    vector<unsigned int>&);
  
  void exportToHashHelper(iterator& it,
			  hash_map< Pair<DSV,Vector<int> > , iterator >& hmap, 
			  hash_map< DSV,Vector<int> >& hmap2, 
			  hash_map<Pair<DSV,Vector<int> >, unsigned int>& hmapCnts,
			  hash_set<Vector<int> > & yields,
			  Vector<Vector<int> > & yieldVec, 
			  DSV& dsvRet)const;

  //! enumerates different layouts for deltas and tells which type the
  //! two TLinks passed in are.
  HeirLayout getHeirLayout(TLink& antTLink1, TLink& antTLink2, int& posDelta);

  /**
   * Indent printing.
   * @param out The output stream.
   * @param tr The parse tree to be output.
   * @param ntset The global set of nonterminal vocabs
   * @param terminalSet The global set of terminal vocabs
   * @param ptbII If true, the output form is consistent with the PTB II
   * format. Note that the parse tr must be monolingual when
   * ptbII is set true.
   * @param newline whether to break lines
   */
  void indentPrint(ostream& out,
		   StringVocabulary* ntset,
		   StringVocabulary* terminalSet,
		   bool ptbII = false, bool newline =true);

  /**
   * Print a subtree rooted at iterator it. Arrangement 'ind' is the number
   * of spaces for indent printing.
   */
  void indentPrint_helper(ostream&, iterator it,
			  StringVocabulary* ntset,
			  StringVocabulary* terminalSet, int ind,
			  bool ptbII = false, bool newline= true);

  void readHeadMTree(istream& in, iterator it);

  /*****************************************************************
   * The following are some operations that are used
   * to manipulate a parse tree.
   *****************************************************************/

  //! Compares two parses in terms of constituent labels and heirs.
  bool compare(const HeadMTree& tr1, const HeadMTree& tr2,
	       vector<StringVocabulary*> ntset,
	       vector<StringVocabulary*> tset);

  //! Helper.
  bool compare(const HeadMTree& tr1, const iterator i,
	       const HeadMTree& tr2, const iterator j,
	       vector<StringVocabulary*> ntset,
	       vector<StringVocabulary*> tset);

  //! Propagate the POS upward until the root.
  void propogatePOS(iterator it);

  //! Transforms the input lexicalized constituent tree
  //! to a dependency tree. This is done by deleting all
  //! the internal heir nodes and mask all the non-terminals
  //! (excluding the part-of-speech tags).
  //! This function is applicable to multi-dimensional trees.
  void extractDependencyTree(iterator it);

  //! Transforms the input lexicalized constituent tree
  //! to a dependency tree. This is done by deleting all
  //! the internal heir nodes and mask all the non-terminals
  //! (excluding the part-of-speech tags).
  //! This function is applicable to multi-dimensional trees.
  void toDependencyTree( );

  //! Assignment of trees.
  HeadMTree& copyTree(const HeadMTree& source);
  void copyTree_helper(iterator i, const HeadMTree& source, iterator j);

  /** flatten the subtree rooted at the iterator. */
  void flatten(iterator it, HeadMTree& tr);

  void indent(ostream& out, size_t numsp);

  /// return the position of the headword of a node
  /// assumes that spans have already been appended
  /// zero means unknown or nonexistent
  static short headPosn(const iterator& it, arity_t dim);

};


namespace HashNS {
  template<>
  class hash<HeadMTree::pre_order_iterator > {
  public:
    size_t operator()(const HeadMTree::pre_order_iterator& it) const {
      size_t ui = reinterpret_cast<size_t>(addressof(*it));
      return ui;
    }
  };
}

#endif
