#include "HeadPMTGA.hpp"
#include "TLink.hpp"
#include "COMPILER-DEPENDENT_HEADERS.hpp"
#include "HeadMTree.hpp"
#include "HeadProduction.hpp"
#include "LazyCD.hpp"
#include "boost/bind.hpp"
#include "GlobalVocab.hpp"
#include "LexNT.hpp"
#include "MoreAlgorithms.hpp"
#include <functional>
#include <vector>
#include "Exceptions.hpp"
#include "boost/variant.hpp"
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"
#include "NameAddressMap.hpp"
#include <string>
#include "COMPILER-DEPENDENT_HEADERS.hpp"

/*
 * Constructor.
 * @param configFile The config file of the grammar.
 */
HeadPMTGA::
HeadPMTGA(arity_t dm, const string& configFile,
	  char** overrideArgV, size_t overrideArgC,
	  BSP<BPO::options_description> priorOptions
	  )
  : _grammar(dm),
    _indexed(dm, configFile, overrideArgV, overrideArgC, addOptionDescriptionsTo(priorOptions)),
    // m_forceDown(true),
    m_pValDistrib(NULL),
    m_pDepDistrib(NULL),
    m_pHeirDistrib(NULL),
    m_pHeirRoleVectorDistrib(NULL),
    m_depHeirDistribs((size_t)pow(2,(double)m_arity)-2),
    m_permDistribs(m_arity,NULL) {
  for(int i = 0; i < 20; i++){
    m_depDistUsed[i] = false;
  }

  /*
   * All the distributions must use the function pointers dynamically
   * allocated with NameAddressMap in order to be serializable. Thus
   * each new LazyCD must be added to NameAddressMap, and the function
   * pointer determined by lookup in the map thus created.
   */

  m_pValDistrib =
    new LazyCD<ValencyEvent,
	     ValencyEvent::FirstType,
	     ValencyEvent::SecondType,
	     boost::_mfi::cmf0<ValencyEvent::FirstType, ValencyEvent> >
    ((boost::_mfi::cmf0<ValencyEvent::FirstType, ValencyEvent>*)NameAddressMap::m_AddressNameMap.find("boost-cmf0-ValencyEvent")->second);

  m_pHeirRoleVectorDistrib =
    new LazyCD<HeirRoleVectorEvent,
	     HeirRoleVectorEvent::FirstType,
	     HeirRoleVectorEvent::SecondType,
	     boost::_mfi::cmf0<HeirRoleVectorEvent::FirstType, HeirRoleVectorEvent> >
    ((boost::_mfi::cmf0<HeirRoleVectorEvent::FirstType, HeirRoleVectorEvent>*)NameAddressMap::m_AddressNameMap.find("boost-cmf0-HeirRoleVectorEvent")->second);

  m_pHeirDistrib =
    new LazyCD<HeirEvent,
	     HeirEvent::FirstType,
	     HeirEvent::SecondType,
	     boost::_mfi::cmf0<NTLink, HeirEvent> >
    ((boost::_mfi::cmf0<NTLink, HeirEvent>*)NameAddressMap::m_AddressNameMap.find("boost-cmf0-NTLink-HeirEvent")->second);

  m_pDepDistrib =
    new LazyCD<Dependency,
	     Dependency::FirstType,
	     Dependency::SecondType,
	     boost::_mfi::cmf0<NTLink, Dependency> >
    ((boost::_mfi::cmf0<NTLink, Dependency>*)NameAddressMap::m_AddressNameMap.find("boost-cmf0-NTLink-Dependency")->second);

  for(unsigned short d = 0; d < pow(2,(double)arity())-2; d++){
    string fn = "boost-bind-";
    stringstream s;
    s << d;
    fn += s.str();
    m_depHeirDistribs[d] =
      new LazyCD<Dependency,
      Vector<unsigned short>,
      Vector<unsigned short>,
      boost::_bi::bind_t<Vector<unsigned short>,
      boost::_mfi::cmf1<Vector<unsigned short>, Dependency, unsigned short>,
      boost::_bi::list2<boost::arg<1>, boost::_bi::value<unsigned short> > >
      >
      ((boost::_bi::bind_t<Vector<unsigned short>,
	boost::_mfi::cmf1<Vector<unsigned short>, Dependency, unsigned short>,
	boost::_bi::list2<boost::arg<1>, boost::_bi::value<unsigned short> > >*)NameAddressMap::m_AddressNameMap.find(fn)->second);
  }

  /*
   * Distributions of permutation events.
   */
  for(arity_t d = 0; d < arity(); ++d){
    short suffix = d;
    string fn = "boost-cmf0-NTPair-";
    stringstream s;
    s << suffix;
    fn += s.str();

    //set the io dimension. This is needed so we know
    // which vocab to access.
    GlobalVocab::s_ioDIM = d;

    m_permDistribs[d] =
      new LazyCD<PAOrderEvent,
      PAOrderEvent::FirstType,
      PAOrderEvent::SecondType,
      boost::_mfi::cmf0<const PAOrderEvent::NTPair&, ImDom<PAOrderEvent::NTPair, PAV> >
      >((boost::_mfi::cmf0<const PAOrderEvent::NTPair&, ImDom<PAOrderEvent::NTPair, PAV> >*)NameAddressMap::m_AddressNameMap.find(fn)->second);
  }

}

//! Destructor. To clear the memory allocated.
HeadPMTGA::~HeadPMTGA() {
  delete m_pValDistrib;
  delete m_pDepDistrib;
  delete m_pHeirRoleVectorDistrib;
  delete m_pHeirDistrib;
  
  for(arity_t d = 0; d < pow(2,(double)arity())-2; d++)
    delete m_depHeirDistribs[d];
  
  for(arity_t d = 0; d < arity(); ++d)
    delete m_permDistribs[d];
}

//! make it empty
void HeadPMTGA::clear() {
  _super::clear();
  m_pValDistrib->clear();
  m_pDepDistrib->clear();
  m_pHeirRoleVectorDistrib->clear();
  m_pHeirDistrib->clear();
  
  for(arity_t d = 0; d < pow(2,(double)arity())-2; d++)
    m_depHeirDistribs[d]->clear();

  for(arity_t d = 0; d < arity(); ++d)
    m_permDistribs[d]->clear();

  for(int i = 0; i < 20; i++){
    m_depDistUsed[i] = false;
  }
}

//! returns a pointer to the correct distribution given a depend event.
Distribution<Dependency>*
HeadPMTGA::getRelevantDistribution(const Dependency& dep, int& i) const {
  arity_t d = arity();
  if(d == 1){
    i = -1;
    return m_pDepDistrib;
  } else {
    i = dep.getDistributionNumber();
    if(i == -1)
      return m_pDepDistrib;
    else{
      return m_depHeirDistribs[i];
    }
  }
}

void HeadPMTGA::normalize() {
  // do nothing, because the data is stored in LazyCDs
}

vector<HeadPMTGA*> HeadPMTGA::splitGrammar(const string &configFile) {
  vector<HeadPMTGA*> grammars;
  /*
  for(size_t i=0; i<this->arity(); i++) {
    grammars.push_back(new HeadPMTGA(1, configFile));
    if(!grammars[i]) {
      THROW("Could not build Grammar")
    }
    //assert(grammars[i]->isHeaded());
  }

  // We need to decompose each event type into its 1-D components
  typedef LazyCD<ValencyEvent, ValencyEvent::FirstType, ValencyEvent::SecondType, boost::_mfi::cmf0<ValencyEvent::FirstType, ValencyEvent> > ValencyEventCD;
  typedef LazyCD<HeirRoleVectorEvent, HeirRoleVectorEvent::FirstType, HeirRoleVectorEvent::SecondType, boost::_mfi::cmf0<HeirRoleVectorEvent::FirstType, HeirRoleVectorEvent> > HeirRoleEventCD;
  typedef LazyCD<HeirEvent, HeirEvent::FirstType, HeirEvent::SecondType, boost::_mfi::cmf0<NTLink, HeirEvent> > HeirEventCD;
  typedef LazyCD<Dependency, Dependency::FirstType, Dependency::SecondType, boost::_mfi::cmf0<NTLink, Dependency> > DependencyCD;
  typedef LazyCD<PAOrderEvent, PAOrderEvent::FirstType, PAOrderEvent::SecondType, boost::_mfi::cmf0<const PAOrderEvent::NTPair&, ImDom<PAOrderEvent::NTPair, PAV> > > PACV;

  // Set breakpoint here.  Then make sure ntlinks and stuff are 
  // getting added in GrammarIndex class
  ValencyEventCD *vals = dynamic_cast<ValencyEventCD*>(m_pValDistrib);
  for(ValencyEventCD::iterator itr=vals->begin(); itr!=vals->end(); itr++) {
    NTLink link = itr->first.link();
    int valency = itr->first.valency();
    SCORE prob = vals->lookup(itr->first);

    for(arity_t d=0; d<grammars.size(); d++) {
      NTLink newLink(1, *(link[d]->clone()));
      grammars[d]->increment(ValencyEvent(newLink, valency), prob);
    }
  }

  DependencyCD *deps = dynamic_cast<DependencyCD*>(m_pDepDistrib);
  for(DependencyCD::iterator itr=deps->begin(); itr!=deps->end(); itr++) {
    Dependency dep = itr->first;
    SCORE prob = deps->lookup(dep);

    NTLink parent = dep.getParent();
    NTLink depend = dep.getDepend();
    for(arity_t d=0; d<grammars.size(); d++) {
      NTLink newParent(1, *(parent[d]->clone()));
      NTLink newDepend(1, *(depend[d]->clone()));
      Dependency newDep(newParent, newDepend);
      grammars[d]->increment(newDep, prob);
    }
  }

  HeirEventCD *heirs = dynamic_cast<HeirEventCD*>(m_pHeirDistrib);
  for(HeirEventCD::iterator itr=heirs->begin(); itr!=heirs->end(); itr++) {
    SCORE prob = heirs->lookup(itr->first);

    NTLink parent = itr->first.getParent();
    NTLink heir = itr->first.getHeir();
    for(arity_t d=0; d<grammars.size(); d++) {
      NTLink newParent(1, *(parent[d]->clone()));
      NTLink newHeir(1, *(heir[d]->clone()));
      HeirEvent newHeirEvent(newParent, newHeir);
      grammars[d]->increment(newHeirEvent, prob);
    }
  }

  HeirRoleEventCD *heirRV = dynamic_cast<HeirRoleEventCD*>(m_pHeirRoleVectorDistrib);
  for(HeirRoleEventCD::iterator itr=heirRV->begin(); itr!=heirRV->end(); itr++) {
    NTLink link = itr->first.link();
    int valency = itr->first.valency();
    HeirRoleVector hrv = itr->first.heirVector();
    SCORE prob = heirRV->lookup(itr->first);

    for(arity_t d=0; d<grammars.size(); d++) {
      NTLink newLink(1, *(link[d]->clone()));
      HeirRoleVector newHrv = hrv.get1DItem(d);
      ImDom<NTLink, short> first(newLink, valency);
      grammars[d]->increment(HeirRoleVectorEvent(first, newHrv), prob);
    }
  }

  for(arity_t d=0; d<grammars.size(); d++) {
    PACV *perms = dynamic_cast<PACV*>(m_permDistribs[d]);
    for(PACV::iterator itr=perms->begin(); itr!=perms->end(); itr++) {
      SCORE prob = perms->lookup(itr->first);
      Nonterminal * nt1 = itr->first.first.first->clone();
      Nonterminal * nt2 = itr->first.first.second->clone();
      PAOrderEvent newOrd(nt1, nt2, itr->first.second, 1);
      grammars[d]->increment(PAIntEvent(newOrd, 0), prob);
    }
  }

  */
  /*
  for(arity_t d=0; d<grammars.size(); d++) {
    // I might need to copy these later
    //grammars[d]->m_depHeirDistribs.resize(0);
    DependencyCD *heirDistribs = dynamic_cast<DependencyCD*>(this->m_depHeirDistribs[d]);
    grammars[d]->m_depHeirDistribs.push_back(this->m_depHeirDistribs[d]);
    //grammars[d]->m_permDistribs.resize(0);
    grammars[d]->m_permDistribs.push_back(this->m_permDistribs[d]);
  }
  */
  /*

  for(arity_t d=0; d<grammars.size(); d++) {
    // Estimate everything
    grammars[d]->buildGrammarIndex();
    grammars[d]->estimate();
    //cout << "size: " << (grammars[d]->terminalSets()[0])->size() << endl;
  }
  */

  return grammars;
}

// Returns the cost of a model event.
SCORE HeadPMTGA::getFeatureScore(const FeatureTypes& feature) const {
  SCORE score;
  if(const Dependency* dep = boost::get<const Dependency>(&feature) ){
    int i;
    score = getRelevantDistribution(*dep, i)->lookupNegLog(*dep);
    if(debug(5)){
      dout()<<"D"<< i << ": " <<*dep<< " " << score<<endl;
    }
    return score;
  } else if(const HeirRoleVectorEvent* heirVectorEvent 
	    = boost::get<const HeirRoleVectorEvent>(&feature) ){
    score = m_pHeirRoleVectorDistrib->lookupNegLog(*(heirVectorEvent));
    if(debug(5)){
      dout()<<"HVE : "<< *heirVectorEvent << " " << score<<endl;
    }
    return score;
  } else if(const HeirEvent* heirEvent = boost::get<const HeirEvent>(&feature) ){
    score = m_pHeirDistrib->lookupNegLog(*heirEvent);
    if(debug(5)){
      dout()<<"H : "<<*heirEvent << " " <<score<<endl;
    }
    return score;
  } else if(const ValencyEvent* v = boost::get<const ValencyEvent>(&feature)){
    score = m_pValDistrib->lookupNegLog(*v);
    if(debug(5)){

      v->link().put(cerr);
      dout()<<" V : "<<*v<<" "<<score<<endl;
    }
    return score;
  } else if(const PAIntEvent* Ev = boost::get<const PAIntEvent>(&feature)){
    score = m_permDistribs[Ev->second]->lookupNegLog(Ev->first);
    if(debug(5)){
      dout()<<"OP : " << Ev->first << " " << Ev->second <<score<<endl;
    }
    return score;
  }

  // should never get to here
  THROW("Can't get score!" << endl);
  return 0;
}

// Returns the cost of a MTG production.
SCORE HeadPMTGA::getProdScore(Production* rule) const {
  SCORE ret = 0;
  SPVector<FeatureTypes> features = getFeatureBag(rule);
  SPVector<FeatureTypes>::const_iterator it;
  for(it = features.begin(); it != features.end(); ++it){
    ret += getFeatureScore(*(it->get()));
  }
  return ret;
}

// Estimate the probability based on the incremented counts.
void HeadPMTGA::estimate() {
  //first, let the base class do what it needs to do
  IndexedPMTG::estimate();

  GTuring<VocabularyIndex> gt(1);

  cerr<<"Estimating valency events distribution ..."<<endl;
  m_pValDistrib->estimate(gt);

  cerr<<"Estimating dependency events distribution ..."<<endl;
  m_pDepDistrib->estimate(gt);

  cerr<<"Estimating heir vector events distribution ..."<<endl;
  m_pHeirRoleVectorDistrib->estimate(gt);

  cerr<<"Estimating heir events distribution ..."<<endl;
  m_pHeirDistrib->estimate(gt);

  cerr<<"Estimating dep/heir events distribution fr ..."<<endl;

  for(arity_t d = 0; d < pow(2,(double)arity())-2; ++d){
    if(m_depDistUsed[d])
      m_depHeirDistribs[d]->estimate(gt);
  }

  for(arity_t d = 0; d < arity(); ++d){
    cerr<<"Estimating permutation events distribution on "
	<<(int)d<<"th dimension ..."<<endl;
    m_permDistribs[d]->estimate(gt);
  }
}

// increment features. Note that score is not log.
void HeadPMTGA:: increment(const FeatureTypes& feature, SCORE score) {
  if(const Dependency* dep = boost::get<const Dependency>(&feature) ){
    int i;
    if(debug(3))cerr << "HeadPMTGA::increment:Dep " << *dep << endl;
    getRelevantDistribution(*dep,i )->increment(*dep, score);

    if(i >= 0)
      m_depDistUsed[i] = true;
  } else if(const HeirEvent* heirEvent = boost::get<const HeirEvent>(&feature) ){
    if(debug(3)) cerr << "HeadPMTGA::increment:HeirEvent " << *heirEvent << endl;
    m_pHeirDistrib->increment(*heirEvent, score);
    //delete heirEvent;
  } else if(const HeirRoleVectorEvent* heirVectorEvent =
	    boost::get<const HeirRoleVectorEvent>(&feature) ){
    if(debug(3)) cerr << "HeadPMTGA::increment:HeirVE " << *heirVectorEvent << endl;
    m_pHeirRoleVectorDistrib->increment(*(heirVectorEvent), score);
  } else if(const ValencyEvent* v = boost::get<const ValencyEvent>(&feature)) {
    if(debug(3)) cerr << "HeadPMTGA::increment:Valency " << *v << endl;
    m_pValDistrib->increment(*v, score);

    /*
     * Add the new nonterminals
     */
    assert(v->arity());
    arity_t d;
    for(d = 0; d < v->arity(); ++d){
      bool found;
      const string* word = ntSet()->baseVocabulary().getWord(v->link().getNT(d)->nt(), found);
      if(!found){
	cerr<<v->link().getNT(d)->nt()<<" not found!"<<endl;
	cerr<<"Error happens when HeadPMTGA increments a valency event.!"<<endl;
	cerr<<"Possible reason: you are doing alignment, and "
	    <<" the NTVocabularys passed to HeadPMTGA have not appropriate "
	    <<" base class on dimension "<<d<<endl;
      }
      ntSet()->addWord(*word);
    }
  } else if(const PAIntEvent* Ev = boost::get<const PAIntEvent>(&feature)){
    if(debug(3)){
      bool tmp = GlobalVocab::s_noVocabsOut;
      GlobalVocab::s_noVocabsOut = false;
      cerr << "HeadPMTGA::increment:PAOrderEvent " << *Ev << endl;
      GlobalVocab::s_noVocabsOut = tmp;
    }
    PAV pav = Ev->first.child();
    if(pav.fanout() <= fanoutLimit())
      m_permDistribs[Ev->second]->increment(Ev->first, score);
    // delete (&Ev->first);
  } else {
    THROW("Can't increment features." << endl);
  }
}

// Read from istream
istream& HeadPMTGA::get(istream& in) {
  arity_t d;
  vector<SCORE> minprbs;

  _super::get(in);

  /*
   * read valency event distribution.
   */
  cerr<<"Reading valency event distribution ..."<<endl;
  ValencyEvent valE(arity());

  // if(m_estType == MLE){
  m_pValDistrib->read(in, &valE);
  // } else {
  // cerr<<"Unknown estimator type!"<<endl;
  // exit(1);
  // }

  NTLink lnl1, lnl2;
  NTLink ntlnk;

  /*
   * read heir vector event distribution.
   */
  cerr<<"Reading heir vector event distribution ..."<<endl;
  HeirRoleVectorEvent heirVE(arity());
  HeirRoleVector hv = heirVE.heirVector();
  ImDom<NTLink, short> pe = heirVE.getParValPair();

  // if(m_estType == MLE){
  m_pHeirRoleVectorDistrib->read(in, &heirVE);
  // } else {
  // cerr<<"Unknown estimator type!"<<endl;
  // exit(1);
  // }

  /*
   * read heir event distribution.
   */
  cerr<<"Reading heir event distribution ..."<<endl;
  HeirEvent heirE(arity());
  m_pHeirDistrib->read(in, &heirE);

  /*
   * read dependency event distribution.
   */
  cerr<<"Reading dependency distribution ..."<<endl;
  Dependency depE(arity());
  m_pDepDistrib->read(in, &depE);

  cerr<<"Reading dependencyHeir distributions ..."<<endl;
  for(d = 0; d < pow(2,(double)arity())-2; d++){
    unsigned int rhsSize = Dependency::getLHSsizefromNumber(d, arity());
    unsigned int lhsSize = arity()*4 - rhsSize;

    Vector<unsigned short> rhv(rhsSize,0);
    Vector<unsigned short> lhv(lhsSize,0);

    m_depHeirDistribs[d]->read(in, &depE);
  }
  /*
   * read permutation event distribution.
   */
  cerr<<"Reading permutation distributions ..."<<endl;

  for(d = 0; d < arity(); d++){
    //set the dimensionality of the io (for vocab use)
    GlobalVocab::s_ioDIM = d;
    PAOrderEvent ordE;
    m_permDistribs[d]->read(in, &ordE);
  }
  return in;
}

//! Write
ostream& HeadPMTGA:: put(ostream& out) const {
  _super::put(out);

  /*
   * write valency event distribution.
   */
  cerr<<"Writing valency event distribution ..."<<endl;
  // if(m_estType == MLE)
  out<<*((LazyCD<ValencyEvent,
	  ValencyEvent::FirstType,
	  ValencyEvent::SecondType>*)m_pValDistrib);

  /*
   * write heir vector event distribution.
   */
  cerr<<"Writing heir vector event distribution ..."<<endl;
  // if(m_estType == MLE)
  out<<*((LazyCD<HeirRoleVectorEvent,
	  HeirRoleVectorEvent::FirstType,
	  HeirRoleVectorEvent::SecondType>*) m_pHeirRoleVectorDistrib);

  /*
   * write heir event distribution.
   */
  cerr<<"Writing heir event distribution ..."<<endl;
  // if(m_estType == MLE)
  out<<*((LazyCD<HeirEvent,
	  HeirEvent::FirstType,
	  HeirEvent::SecondType>*) m_pHeirDistrib);

  /*
   * write dependency event distribution.
   */
  cerr<<"Writing dependency distribution ..."<<endl;
  // if(m_estType == MLE)
  out<<*((LazyCD<Dependency,
	  Dependency::FirstType,
	  Dependency::SecondType>*) m_pDepDistrib);

  cerr<<"Writing Dependency/Heir distribution ..."<<endl;
  // if(m_estType == MLE)
  for(short d = 0; d < pow(2,(double)arity())-2; d++)
    out<<*((LazyCD<Dependency,
	    Dependency::FirstType,
	    Dependency::SecondType>*) m_depHeirDistribs[d]);

  /*
   * write permutation event distribution.
   */
  cerr<<"Writing permutation distributions ..."<<endl;
  for(arity_t d = 0; d < arity(); d++){
    GlobalVocab::s_ioDIM = d;
    // if(m_estType == MLE)
    out<<*((LazyCD<PAOrderEvent,
	    PAOrderEvent::FirstType,
	    PAOrderEvent::SecondType>*) m_permDistribs[d]);
  }
  return out;
}

SPVector<HeadPMTGA::FeatureTypes> HeadPMTGA ::
getFeatureBag(Production* pProd) const {

  if(!pProd)
    THROW("prodrule is not a HeadProduction!");

  if(pProd->rank() > 2)
    cerr << "rank is " << pProd->rank() << endl;
  assert(pProd->rank() <= 2);

  SPVector<FeatureTypes> features;

  // valency event
  features.push_back(BSP<FeatureTypes>
		     (new FeatureTypes(ValencyEvent(pProd->lhs(), pProd->rank()))));

  // if it's not headed, then there are no other features of interest
  HeadProduction* pHProd = dynamic_cast<HeadProduction*>(pProd);
  if (not pHProd)
    return features;

  //Heir Vector Event
  if(pHProd->rank() >= 1){
    features.push_back(BSP<FeatureTypes>(new FeatureTypes(HeirRoleVectorEvent(ImDom<NTLink, short>(pHProd->lhs(), pHProd->rank()), pHProd->heirRole().normal()))));
  }
  // Heir Event.
  if(pHProd->rank() >= 1) {
    NTLink* nt = pHProd->heirRole().getHeirEvent(pHProd->rhsLinks());
    if(const LexNT* xx = dynamic_cast<const LexNT*>(nt->getNT(0))){
      xx = xx; // duck compiler warning
      TLink tl(arity());
      for(unsigned int i =0; i < arity(); i++) {
	    short s = pHProd->heirRole().getRole(i);
	    if (s > 0) {
	      if(TLink* t = boost::get<TLink>(&(pHProd->rhsLinks()[s-1]))) {
	        tl[i] = (*t)[i];
	      }
	      if(NTLink* ntt = boost::get<NTLink>(&pHProd->rhsLinks()[s-1])) {
	        tl[i] = ntt->getTLink()[i];
	      }
	    }
      }
      pHProd->lhs().setTLink(tl);
    }
    features.push_back(BSP<FeatureTypes>(new FeatureTypes(HeirEvent(pHProd->lhs(), *nt))));
    delete nt;
  }
  // Dep Event
  if(pHProd->rank() > 1){
    for(arity_t i = 0 ; i < pHProd->arity(); i++){
      NTLink* nt = pHProd->heirRole().getDependency( boost::get<NTLink>(pHProd->rhsLinks()[i]), i);
      if(nt != NULL){
	features.push_back(BSP<FeatureTypes>(new FeatureTypes(Dependency(pHProd->lhs(), *nt))));
      }
      delete nt;
    }
  }
  //PAIntEvent
  if(pHProd->rank() == 2){
    for(unsigned int i = 0; i < boost::get<NTLink>(pHProd->rhsLinks().front()).size(); i++){
      Nonterminal* ls = boost::get<NTLink>(pHProd->rhsLinks()[0]).getNT(i)->clone();
      Nonterminal* rs = boost::get<NTLink>(pHProd->rhsLinks()[1]).getNT(i)->clone();
      if(!(ls->isInactive() && rs->isInactive())){
	PAV pav1 = PAV(pHProd->rhsPerm().getPA(i));
	GlobalVocab::s_ioDIM = i;
	features.push_back(BSP<FeatureTypes>
			   (new FeatureTypes(PAIntEvent(PAOrderEvent(ls, rs, pav1, i),i))));
      } else{
	delete ls;
	delete rs;
      }
    }
  }
  return features;
}

BSP<BPO::options_description> HeadPMTGA::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDesc){
  //optionsDesc->add_options()
  // ("HeadPMTGATestOption", BPO::value<string>()->default_value("default value"), "test option")
  // ;
  return optionsDesc;
}
