#include "HeadPMTGB.hpp"
#include "TLink.hpp"
#include "COMPILER-DEPENDENT_HEADERS.hpp"
#include "HeadMTree.hpp"
#include "HeadProduction.hpp"
#include "LazyCD.hpp"
#include "GlobalVocab.hpp"
#include "boost/bind.hpp"
#include "MoreAlgorithms.hpp"
#include <functional>
#include <vector>
#include "Exceptions.hpp"
#include "boost/variant.hpp"
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

/*
 * Constructor.
 * @param configFile The config file of the grammar.
 */
HeadPMTGB::HeadPMTGB(arity_t dm, const string& configFile,
		     char** overrideArgV, size_t overrideArgC,
		     BSP<BPO::options_description> optionsDescription
		     )
  :
  _grammar(dm),
  _indexed(dm,configFile, overrideArgV, overrideArgC, addOptionDescriptionsTo(optionsDescription)),
  // m_forceDown(true),
  m_pValDistrib(NULL),
  m_pDepDistrib(NULL),
  m_pHeirDistrib(NULL),
  m_pHeirRoleVectorDistrib(NULL),
  m_depHeirDistribs((size_t)pow(2,(double)m_arity)-2),
  m_pPermDistrib(NULL) {
  //say that none of the dep/heir distributions have been used.

  for(int i = 0; i < 20; i++){
    m_depDistUsed[i] = false;
  }

  m_pValDistrib =
    new LazyCD<ValencyEvent,
    ValencyEvent::FirstType,
    ValencyEvent::SecondType,
    boost::_mfi::cmf0<ValencyEvent::FirstType, ValencyEvent> >
    (new boost::_mfi::cmf0<ValencyEvent::FirstType, ValencyEvent>(boost::mem_fn(( NTLink (ValencyEvent::*)()const)&ValencyEvent::link)));

  m_pHeirRoleVectorDistrib =
    new LazyCD<HeirRoleVectorEvent,
    HeirRoleVectorEvent::FirstType,
    HeirRoleVectorEvent::SecondType,
    boost::_mfi::cmf0<HeirRoleVectorEvent::FirstType, HeirRoleVectorEvent> >
    (new boost::_mfi::cmf0<HeirRoleVectorEvent::FirstType, HeirRoleVectorEvent>(boost::mem_fn( (HeirRoleVectorEvent::FirstType (HeirRoleVectorEvent::*)()const)& HeirRoleVectorEvent::getParValPair)));

  boost::mem_fn(&HeirEvent::getParent);
  m_pHeirDistrib =
    new LazyCD<HeirEvent,
    HeirEvent::FirstType,
    HeirEvent::SecondType,
    boost::_mfi::cmf0<NTLink, HeirEvent> >
    (new boost::_mfi::cmf0<NTLink, HeirEvent>(boost::mem_fn(&HeirEvent::getParent)));

  m_pDepDistrib =
    new LazyCD<Dependency,
    Dependency::FirstType,
    Dependency::SecondType,
    boost::_mfi::cmf0<NTLink, Dependency> >
    (new boost::_mfi::cmf0<NTLink, Dependency>(boost::mem_fn((NTLink (Dependency::*)() const)&Dependency::getParent)));

  for(unsigned short d = 0; d < pow(2,(double)arity())-2; d++){
    m_depHeirDistribs[d] =
      new LazyCD<Dependency,
      Vector<unsigned short>,
      Vector<unsigned short>,
      boost::_bi::bind_t<Vector<unsigned short>,
      boost::_mfi::cmf1<Vector<unsigned short>, Dependency, unsigned short>,
      boost::_bi::list2<boost::arg<1>, boost::_bi::value<unsigned short> > >
      >
      (new boost::_bi::bind_t<Vector<unsigned short>,
      boost::_mfi::cmf1<Vector<unsigned short>, Dependency, unsigned short>,
      boost::_bi::list2<boost::arg<1>, boost::_bi::value<unsigned short> > > 
       ( boost::bind<Vector<unsigned short> >(boost::mem_fn(&Dependency::getFromDesc), _1, d)));
  }

  /*
   * Distributions of permutation events.
   */
  m_pPermDistrib =
    new LazyCD<PAVOrderEvent,
    PAVOrderEvent::FirstType,
    PAVOrderEvent::SecondType,
    boost::_mfi::cmf0<ImDom<NTLink,
    NTLink>, PAVOrderEvent> >
   ((new boost::_mfi::cmf0<ImDom<NTLink, NTLink>, 
     PAVOrderEvent>(boost::mem_fn((PAVOrderEvent::FirstType ( PAVOrderEvent::*)()const) 
				  &PAVOrderEvent::links))));

}

//! Destructor. To clear the memory allocated.
HeadPMTGB::~HeadPMTGB() {
  delete m_pValDistrib;
  delete m_pDepDistrib;
  delete m_pHeirRoleVectorDistrib;
  delete m_pHeirDistrib;
  delete m_pPermDistrib;
  
  for(arity_t d = 0; d < pow(2,(double)arity())-2; d++)
    delete m_depHeirDistribs[d];
}

//! make it empty
void HeadPMTGB::clear() {
  _super::clear();

  m_pValDistrib->clear();
  m_pDepDistrib->clear();
  m_pHeirRoleVectorDistrib->clear();
  m_pHeirDistrib->clear();
  m_pPermDistrib->clear();
  
  for(arity_t d = 0; d < pow(2,(double)arity())-2; d++)
    m_depHeirDistribs[d]->clear();

  for(int i = 0; i < 20; i++){
    m_depDistUsed[i] = false;
  }
};


//! returns a pointer to the correct distribution given a depend event.
Distribution<Dependency>*
HeadPMTGB::getRelevantDistribution(const Dependency& dep, int& i)const{
  arity_t d = arity();
  if(d == 1){
    i = -1;
    return m_pDepDistrib;
  } else {
    i = dep.getDistributionNumber();
    if(i == -1)
      return m_pDepDistrib;
    else{
      return m_depHeirDistribs[i];
    }
  }
}

void HeadPMTGB::normalize() {
  // do nothing, because the data is stored in LazyCDs
}

// Returns the cost of a model event.
SCORE HeadPMTGB::getFeatureScore(const FeatureTypes& feature) const {
  SCORE score;
  if(const Dependency* dep = boost::get<const Dependency>(&feature) ){
    int i;
    score = getRelevantDistribution(*dep, i)->lookupNegLog(*dep);
    if(debug(5)){
      dout()<<"D"<< i << ": " <<*dep<< " " << score<<endl;
    }
    return score;
  } else if(const HeirRoleVectorEvent* heirVectorEvent =
	    boost::get<const HeirRoleVectorEvent>(&feature) ){
    score = m_pHeirRoleVectorDistrib->lookupNegLog(*(heirVectorEvent));
    if(debug(5)){
      dout()<<"HVE : "<< *heirVectorEvent << " " << score<<endl;
    }

    return score;
  } else if(const HeirEvent* heirEvent = boost::get<const HeirEvent>(&feature) ){
    score = m_pHeirDistrib->lookupNegLog(*heirEvent);
    if(debug(5)){
      dout()<<"H : "<<*heirEvent << " " <<score<<endl;
    }
    return score;
  } else if(const ValencyEvent* v = boost::get<const ValencyEvent>(&feature)){
    score = m_pValDistrib->lookupNegLog(*v);
    if(debug(5)){
      v->link().put(cerr);
      dout()<<" V : "<<*v<<" "<<score<<endl;
    }
    return score;
  } else if(const PAVOrderEvent* ordE = boost::get<const PAVOrderEvent>(&feature)){
    score = m_pPermDistrib->lookupNegLog(*ordE);
    if(debug(5)){
      dout()<<"O : "<<score<<endl;
    }
    return score;
  }

  // should never get to here
  THROW("Can't get score!" << endl);
  return 0;
}

// Returns the cost of a MTG production.
SCORE HeadPMTGB::getProdScore(Production* rule) const {
  SCORE ret = 0;

  SPVector<FeatureTypes> features = getFeatureBag(rule);
  SPVector<FeatureTypes>::const_iterator it;
  for(it = features.begin(); it != features.end(); ++it){
    ret += getFeatureScore(*(it->get()));
  }
  return ret;
}

// Estimate the probability based on the incremented counts.
void HeadPMTGB::estimate() {
  //first, let the base class do what it needs to do
  IndexedPMTG::estimate();

  GTuring<VocabularyIndex> gt(1);

  cerr<<"Estimating valency events distribution ..."<<endl;
  m_pValDistrib->estimate(gt);

  cerr<<"Estimating dependency events distribution ..."<<endl;
  m_pDepDistrib->estimate(gt);

  cerr<<"Estimating heir vector events distribution ..."<<endl;
  m_pHeirRoleVectorDistrib->estimate(gt);

  cerr<<"Estimating heir events distribution ..."<<endl;
  m_pHeirDistrib->estimate(gt);

  cerr<<"Estimating dep/heir events distribution fr ..."<<endl;

  for(arity_t d = 0; d < pow(2,(double)arity())-2; ++d)
    if(m_depDistUsed[d])
      m_depHeirDistribs[d]->estimate(gt);

  cerr<<"Estimating the distributions of permutation events ... "<<endl;
  m_pPermDistrib->estimate(gt);

}

// increment features. Note that score is not log.
void HeadPMTGB:: increment(const FeatureTypes& feature, SCORE score) {
  if(const Dependency* dep = boost::get<const Dependency>(&feature) ){
    int i;
    if(debug(3))cerr << "HeadPMTGB::increment:Dep " << *dep << endl;
    getRelevantDistribution(*dep,i )->increment(*dep, score);
    if(i >= 0)
      m_depDistUsed[i] = true;
  } else if(const HeirEvent* heirEvent = boost::get<const HeirEvent>(&feature) ){
    if(debug(3)) cerr << "HeadPMTGB::increment:HeirEvent " << *heirEvent << endl;
    m_pHeirDistrib->increment(*heirEvent, score);
  } else if(const HeirRoleVectorEvent* heirVectorEvent =
	    boost::get<const HeirRoleVectorEvent>(&feature) ){
    if(debug(3)) cerr << "HeadPMTGB::increment:HeirVE " << *heirVectorEvent << endl;
    m_pHeirRoleVectorDistrib->increment(*(heirVectorEvent), score);
  } else if(const ValencyEvent* v = boost::get<const ValencyEvent>(&feature)){
    if(debug(3)) cerr << "HeadPMTGB::increment:Valency " << *v << endl;
    m_pValDistrib->increment(*v, score);

    /*
     * Add the new nonterminals.
     */
    assert(v->arity());
    arity_t d;
    for(d = 0; d < v->arity(); ++d){
      bool found;
      const string* word = ntSet()->baseVocabulary().getWord(v->link().getNT(d)->nt(), found);
      if(!found){
	cerr<<v->link().getNT(d)->nt()<<" not found!"<<endl;
	cerr<<"Error happens when HeadPMTGB increments a valency event.!"<<endl;
	cerr<<"Possible reason: you are doing alignment, and "
	    <<" the NTVocabularys passed to HeadPMTGB have not appropriate "
	    <<" base class on dimension "<<d<<endl;
      }
      ntSet()->addWord(*word);
    }
  } else
    if(const PAVOrderEvent* ordE = boost::get<const PAVOrderEvent>(&feature)){
      if(debug(3)) cerr << "HeadPMTGB::increment:PAVOrderEvent " << *ordE << endl;
      m_pPermDistrib->increment(*ordE, score);
    }

    else {
      THROW("Cannot increment features." << endl);
    }
}

// Read from istream
istream& HeadPMTGB::get(istream& in) {
  arity_t d;
  vector<SCORE> minprbs;

  _super::get(in);


  /*
   * read valency event distribution.
   */
  cerr<<"Reading valency event distribution ..."<<endl;
  ValencyEvent valE(arity());

  // if(m_estType == MLE){
  m_pValDistrib->read(in, &valE);
  // } else {
  // cerr<<"Unknown estimator type!"<<endl;
  // exit(1);
  // }

  NTLink lnl1, lnl2;
  NTLink ntlnk;

  /*
   * read heir vector event distribution.
   */
  cerr<<"Reading heir vector event distribution ..."<<endl;
  HeirRoleVectorEvent heirVE(arity());
  HeirRoleVector hv = heirVE.heirVector();
  ImDom<NTLink, short> pe = heirVE.getParValPair();

  // if(m_estType == MLE){
  m_pHeirRoleVectorDistrib->read(in, &heirVE);
  // } else {
  // cerr<<"Unknown estimator type!"<<endl;
  // exit(1);
  // }

  /*
   * read heir event distribution.
   */
  cerr<<"Reading heir event distribution ..."<<endl;
  HeirEvent heirE(arity());
  m_pHeirDistrib->read(in, &heirE);

  /*
   * read dependency event distribution.
   */
  cerr<<"Reading dependency distribution ..."<<endl;
  Dependency depE(arity());
  m_pDepDistrib->read(in, &depE);

  cerr<<"Reading dependencyHeir distributions ..."<<endl;
  for(d = 0; d < pow(2,(double)arity())-2; d++){
    unsigned int rhsSize = Dependency::getLHSsizefromNumber(d, arity());
    unsigned int lhsSize = arity()*4 - rhsSize;

    Vector<unsigned short> rhv(rhsSize,0);
    Vector<unsigned short> lhv(lhsSize,0);

    m_depHeirDistribs[d]->read(in, &depE);
  }

  cerr<<"Reading permutation distributions ..."<<endl;
  PAVOrderEvent ordE(arity());
  m_pPermDistrib->read(in, &ordE);

  return in;
}

//! Write
ostream& HeadPMTGB:: put(ostream& out) const {
  arity_t d;

  _super::put(out);

  cerr<<"Writing valency event distribution ..."<<endl;
  out<<*((LazyCD<ValencyEvent,
	  ValencyEvent::FirstType,
	  ValencyEvent::SecondType>*)m_pValDistrib);

  cerr<<"Writing heir vector event distribution ..."<<endl;
  out<<*((LazyCD<HeirRoleVectorEvent,
	  HeirRoleVectorEvent::FirstType,
	  HeirRoleVectorEvent::SecondType>*) m_pHeirRoleVectorDistrib);

  cerr<<"Writing heir event distribution ..."<<endl;
  out<<*((LazyCD<HeirEvent,
	  HeirEvent::FirstType,
	  HeirEvent::SecondType>*) m_pHeirDistrib);

  cerr<<"Writing dependency distribution ..."<<endl;
  out<<*((LazyCD<Dependency,
	  Dependency::FirstType,
	  Dependency::SecondType>*) m_pDepDistrib);

  cerr<<"Writing Dependency/Heir distribution ..."<<endl;
  for(d = 0; d < pow(2,(double)arity())-2; d++)
    out<<*((LazyCD<Dependency,
	    Dependency::FirstType,
	    Dependency::SecondType>*) m_depHeirDistribs[d]);

  cerr<<"Writing permutation distributions ..."<<endl;
  out<<*((LazyCD<PAVOrderEvent,
	  PAVOrderEvent::FirstType,
	  PAVOrderEvent::SecondType>*)m_pPermDistrib);

  return out;
}

// SCORE HeadPMTGB::rootCost(const Item* item)const{
//   NTLink ntl = ((Hedge*)(item))->getNTLink();

//   try{
//   SCORE s = ((LazyCD<HeirEvent,
// 	      HeirEvent::FirstType,
// 	      HeirEvent::SecondType,
// 	      boost::_mfi::cmf0<NTLink, HeirEvent> >*)
// 	     (m_pHeirDistrib))->getDen()->lookup(ntl); 

//   return (-log(s));
//   }catch(KeyNotFound<NTLink>){
//     return 999999;
//   }
// }

/**
 * Decomposes an MTG production into features
 */
SPVector<HeadPMTGB::FeatureTypes> HeadPMTGB::
getFeatureBag(Production* pProd) const {

  assert(pProd->rank() <= 2);

  SPVector<FeatureTypes> features;

  // valency event
  features.push_back(BSP<FeatureTypes>
		     (new FeatureTypes(ValencyEvent(pProd->lhs(), pProd->rank()))));

  HeadProduction* pHProd = dynamic_cast<HeadProduction*>(pProd);

  // if it's not headed, then there are no other features of interest
  if (not pHProd)
    return features;

  // heir role vector and heir event
  if(pHProd->rank() >= 1){
    Vector<HeirRoleVectorEvent> hvs(1, HeirRoleVectorEvent(ImDom<NTLink, short>(pHProd->lhs(), pHProd->rank()), pHProd->heirRole()));
    Vector<HeirRoleVectorEvent>::iterator it;
    for(it = hvs.begin(); it != hvs.end(); it++)
      features.push_back(BSP<FeatureTypes>(new FeatureTypes(HeirRoleVectorEvent(*it))));
    NTLink* nt = pHProd->heirRole().getHeirEvent(pHProd->rhsLinks());
    features.push_back(BSP<FeatureTypes>(new FeatureTypes(HeirEvent(pHProd->lhs(), *nt))));
  }

  // dependency events
  for(size_t i = 0; i < pHProd->rank(); i++)
    if(boost::get<NTLink>(&pHProd->rhsLinks()[i])){
      NTLink* nt = pHProd->heirRole().getDependency(boost::get<NTLink>(pHProd->rhsLinks()[i]),i);
      if(nt != NULL){
	features.push_back(BSP<FeatureTypes>(new FeatureTypes(Dependency(pHProd->lhs(), *nt))));
      }
      delete nt;
    }

  // permutation event, only comes from n-ary productions with n > 1.
  if(pHProd->rank() > 1){
    PAVOrderEvent::FirstType links;
    links.parent() = boost::get<NTLink>(pHProd->rhsLinks()[0]);
    links.child() = boost::get<NTLink>(pHProd->rhsLinks()[1]);
    features.push_back(BSP<FeatureTypes>
		       (new FeatureTypes(PAVOrderEvent(links, pHProd->rhsPerm()))));
  }

  return features;

}

BSP<BPO::options_description> HeadPMTGB::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  optionsDescription->add_options()
    ("HeadPMTGBTestOption", BPO::value<string>()->default_value("default value"), "test option")
    ;
  return optionsDescription;
}
