#include "HeadProduction.hpp"
#include "TLink.hpp"
#include "Exceptions.hpp"
#include "Dbug.hpp"
#include "LexNT.hpp"
#include "HeadCompose.hpp"
#include "boost/variant.hpp"

  //! copy constructor
HeadProduction::HeadProduction(const HeadProduction& other) 
  : _super(other), m_heirRole(other.m_heirRole) 
{};

HeadProduction::HeadProduction(const HeadMTree::iterator& it) 
  : _super(it) {
  m_heirRole = HeadMTree::getHRV(it);
  if(m_rhs.size() == 1){
    TLink* tl = boost::get<TLink>(&m_rhs[0]);
    for(unsigned int i = 0; i < arity(); i++)
      //if its not a null, then let the heir point to it.
      if((*tl)[i] != NULLT)
       m_heirRole.setRole(i,1);
  }
}

//! constructor from inference
HeadProduction::HeadProduction(const Inference* pInf) 
  : _super(pInf) {
  const HeadCompose* pHC = dynamic_cast<const HeadCompose*>(pInf);
  if (pHC)
    m_heirRole = pHC->heirIndices();
};

//! constructor
HeadProduction::
HeadProduction(const NTLink& lhss,
	       const Vector<ProdRHSElem>& rhslinks,
	       const PAV& pav,
	       const HeirRoleVector& heirIndex)
  : _super(lhss, rhslinks, pav) {

  // if links vector is empty, then it is a terminal production.
  if(rhslinks.size() == 0) {
    if(heirIndex.containsHeir()) {
      THROW("In HeadProduction constructor: heir index "
	    <<"should be 0 in all dimensions for terminal prod."<<endl);
    }
  }

  m_heirRole = heirIndex;
}

// Assignment.
HeadProduction&
HeadProduction::operator=(const HeadProduction& other) {
  _super::operator=(other);
  m_heirRole  = other.heirRole();
  return *this;
}

// Equality test.
bool HeadProduction::
operator==(const HeadProduction& other) const {
  return _super::operator==(other) && m_heirRole  == other.heirRole();
}


bool HeadProduction::less(const HeadProduction& other) const {
  return (cmp(other) < 0);
}

int HeadProduction::cmp(const HeadProduction& other) const {
  int ret;

  // super
  if((ret = _super::cmp(other)) != 0) return ret;

  // hrv
  if((ret = m_heirRole.cmp(other.m_heirRole)) != 0) return ret;

  return 0;
}

void HeadProduction::clear() {
  _super::clear();
}

ostream& HeadProduction::put(ostream& out) const {

  if (Dbug::isDebugAll(3))
    cerr << "HProd:= " << m_uniqueID;

  HeadProduction aProd(*this);

  /*output lhs*/
  out << aProd.lhs();
  
  /*output rhs links*/
  for(size_t r = 0; r < aProd.rank(); ++r) {
    out<<"^"<<aProd.rhsLinks()[r];
  }
  
  /*output rhs permutation.*/
  // terminal production doesn't output the permutation.
  if(aProd.rank()) {
    out<<"^"<<aProd.rhsPerm();
  }
  
  /*output the heir role*/
  out<<"^" << aProd.heirRole();

  return out;
}

ostream& operator<<(ostream& out,
	            const HeadProduction& bProd) {
  bProd.put(out);
  return out;
}

istream& operator>>(istream& in, HeadProduction& aProd) {
  aProd.clear();
  in >> ws;
  string buffer;
  getline(in, buffer, '\n');

  vector<string> strs = split(buffer, "^");
  deque<string> strs1;
  copy(strs.begin(), strs.end(), back_inserter<deque<string> >(strs1));

  // every aProd at least has lhs label and heirRole (even if there is no heir).
  if(strs1.size() <= 1) return in;
  // heirRoleVector
  strs = split(strs1.back().c_str(), " ");
  HeirRoleVector hv(strs.size());
  for(unsigned int i = 0; i < strs.size(); i++){
    hv.setRole(i,atoi(strs[i].c_str()));
  }
  aProd.m_heirRole = hv;

  strs1.pop_back();

  // lhs
#ifdef USE_STRSTREAM
  istrstream ist(strs1[0].c_str(), strs1.size());
#else
  istringstream ist(strs1[0]);
#endif
  NTLink el(2);
  ist >> ws;
  ist >> el;
  aProd.lhs() = el;
  strs1.pop_front();

  // pav. If, now, the size of strs1 is still nonzero, then
  // the rank will not be zero.
  if(strs1.size()){
#ifdef USE_STRSTREAM
    istrstream ist1(strs1.back().c_str(), strs1.back().size());
#else
    istringstream ist1(strs1.back());
#endif
    ist >> ws;
    ist1 >> aProd.rhsPerm();
    strs1.pop_back();
  }

  // rhs label.
  while(strs1.size()){
#ifdef USE_STRSTREAM
    istrstream ist1(strs1.front().c_str(), strs1.front().size());
#else
    istringstream ist1(strs1.front());
#endif
    in >> ws;
    ist1 >> ws;
    //if there is a {, then it's an NTLink
    if(strs1.front().c_str()[0] == '{' ||
       ( strs1.front().c_str()[1] == '{' && strs1.front().c_str()[0] == ' ')){
      NTLink elem(2);
      ist1 >> elem;
      aProd.rhsLinks().push_back(elem);
    }
    //else it's a TLink
    else{
      TLink elem;
      ist1 >> elem;
      aProd.rhsLinks().push_back(elem);
    }
    strs1.pop_front();
  }

  return in;
}

istream& HeadProduction::get(istream& in){
  in >> *this;
  return in;
}

size_t HeadProduction::hash() const {
  size_t ret = _super::hash();

  // heir index.
  ret += heirRole().hash();

  return ret;
}

bool HeadProduction::getTLink(W2WLink& lk) const {
  if (rank() != 2)
    return false;

  if(!isW2WProduction())
    return false;
  
  const NTLink* nt1 = boost::get<const NTLink>(addressof(m_rhs[0]));
  const NTLink* nt2 = boost::get<const NTLink>(addressof(m_rhs[1]));

  if( nt1 && nt2) {
    TLink tl1 = nt1->getTLink();
    TLink tl2 = nt2->getTLink();
    // Assuming they are both 2D, even if one of the slots is inactive (null).
    // Also assuming that each is active in at least one slot.
    
    //Preterminal productions (word + word)
    if (tl1.active(0) && not tl2.active(0) &&  not tl1.active(1) && tl2.active(1)    ) {
      lk = W2WLink(tl1[0], tl2[1]);
      return true;
    } else if (tl2.active(0) && not tl1.active(0) &&  not tl2.active(1) && tl1.active(1)  ) {
      lk = W2WLink(tl2[0], tl1[1]);
      return true;
    }//inactive in one dim  (word + \ep)
    else if(not tl1.active(1)&& not tl2.active(1)){
      if(m_heirRole.getRole(0) == 1)
	lk = W2WLink(tl2[0], tl2[1]);
      else
	lk = W2WLink(tl1[0], tl1[1]);
      return true;
    }//inactive in other dim (word + \ep)
    else if(not tl1.active(0)&& not tl2.active(0)){
      if(m_heirRole.getRole(1) == 1)
	lk = W2WLink(tl2[0], tl2[1]);
      else
	lk = W2WLink(tl1[0], tl1[1]);
      return true;
    }//inactive in one spot (word + \ep)
    else if (not tl1.active(0)){
      lk = W2WLink(tl2[0], tl1[1]);
      return true;
    }//inactive in one spot (word + \ep)
    else if (not tl1.active(1)){
      lk = W2WLink(tl1[0], tl1[1]);
      return true;
    }//inactive in one spot (word + \ep)
    else if (not tl2.active(0)) {
      lk = W2WLink(tl1[0], tl2[1]);
      return true;
    }
    else if (not tl2.active(1)) {
      lk = W2WLink(tl2[0], tl2[1]);
      return true;
    }
  } else 
    THROW("Binary production has something other than NTLinks on the RHS!");

  return false;
}


//! Return the data simply as a vector of ints.
Vector<int> HeadProduction::makeVec()const{
  Vector<int> ret;
  
  for(unsigned int i = 0; i != lhs().size(); i++){
    ret.push_back(lhs().getNT(i)->getSynCat());
    ret.push_back(lhs().getTLink()[i]);
  }
  
  ret.push_back(heirRole().getIndex());
  ret.push_back( rhsPerm().getIndex());
  
  for(size_t r = 0; r < this->rank(); ++r) {  
    if(const NTLink* nt = boost::get<NTLink>(&rhsLinks()[r])){
      for(unsigned int i = 0; i < nt->size(); i++){
	ret.push_back((*nt).getNT(i)->getSynCat());
	ret.push_back(nt->getTLink()[i]);
      }
    }
    
    else if(const TLink* t = boost::get<TLink>(&rhsLinks()[r])){
      for(unsigned int i = 0; i < t->size(); i++){
	ret.push_back((*t)[r]);
      }
    }
  }
  return ret;
}
