#include "HeadWMTG.hpp"
#include "StringVocabulary.hpp"
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

/*
 * Constructor.
 * @param configFile The config file of the grammar.
 */
HeadWMTG::
HeadWMTG(arity_t dm, const string& configFile,
     char** overrideArgV, size_t overrideArgC,
     BSP<BPO::options_description> optionsDescription )
  : _grammar(dm),
    _super(dm, configFile, overrideArgV, overrideArgC,
	   addOptionDescriptionsTo(optionsDescription)) {

  BPO::variables_map vm;
  if(overrideArgV)
    store(BPO::command_line_parser(overrideArgC, overrideArgV).options(*optionsDescription).run(), vm);
  if(configFile != ""){
    iFile ifs(configFile.c_str());
    store(parse_config_file(ifs, *optionsDescription), vm);
  }
  notify(vm);
  if(debug(1)) {
    cerr << "HeadWMTG options:" << endl;
    cerr << OptionsPrinter(*optionsDescription, vm) << endl;
  }
}

//! Destructor. To clear the memory allocated.
HeadWMTG::~HeadWMTG() {}

// Returns the cost of a MTG production.
SCORE HeadWMTG::getProdScore(Production* rule) const {
  SCORE sum = 0;

  SPVector<FeatureTypes> features = getFeatureBag(rule);
  SPVector<FeatureTypes>::const_iterator it;
  for(it = features.begin(); it != features.end(); ++it){
    sum += getFeatureScore(*(it->get()));
  }
  size_t featCount = features.size();
  return sum / featCount;
}

/** Increment {\it binary/unary} MTG rules.
    @param score is the increment size
    If the fanout of this rule is greater than the fanout limit, then
    this rule is rejected. The fanout of a production is the
    fanout of the involved PAV */
void HeadWMTG:: increment(Production* pProd, SCORE score) {
  if(debug(3))
    cerr << "HeadWMTG " << m_uniqueID << "::incr by" << score << *pProd << endl;
  // Only valid for productions with rank at most 2.
  assert(pProd->rhsPerm().rank() <= 2);

  // The fanout limit test.
  if(pProd->fanout() > m_fanoutLimit){
    return;
  }
  FanoutVect fanoutv(pProd->rhsPerm().getFanoutV(1));
  FanoutVect fanoutv1(pProd->rhsPerm().getFanoutV(2));

  if( fanoutv.overallFanout() > m_fanoutLimit 
      || fanoutv1.overallFanout() > m_fanoutLimit){
    return;
  }

  // Increment the features into the Grammar.
  SPVector<FeatureTypes> features;
  features = getFeatureBag(pProd);
  SCORE updateSize = score / features.size();
  SPVector<FeatureTypes>::const_iterator it;
  for(it = features.begin(); it != features.end(); ++it){
    updateFeatureScore(it->get(), updateSize);
  }

  // remember the whole production (this might eventually disappear)
  _super::updateFeatureScore(pProd, score);
}

/// break the tree down into a vector of features
/// return the number of features just added
size_t HeadWMTG::tree2features(const HeadMTree& inTree, vector<BSP<FeatureTypes> >& ret) const {
  // do NOT clear ret, in case the caller wants to loop over a bunch of trees
  //  ret.clear();
  size_t sz = 0;
  vector<BSP<Production> > vProds;
  vector<BSP<Production> >::const_iterator vProdIter;
  _super::tree2features(inTree, vProds);
  for(vProdIter = vProds.begin(); vProdIter != vProds.end(); ++vProdIter) {
    const SPVector<FeatureTypes>& vFeats = getFeatureBag(*vProdIter);
    ret.insert(ret.end(), vFeats.begin(), vFeats.end());
    sz += vFeats.size();
  } 
  return sz;
};

/// break a tree node down into a vector of features
/// return the number of features so generated
size_t HeadWMTG::treeNode2features(const HeadMTree::iterator& inNode, 
				   vector<BSP<FeatureTypes> >& ret) const {
  // do NOT clear ret, in case the caller wants to loop over a bunch of trees
  //  ret.clear();
  size_t sz = 0;
  vector<BSP<Production> > vProds;
  vector<BSP<Production> >::const_iterator vProdIter;
  vector<BSP<FeatureTypes> >::const_iterator ftci;
  _super::treeNode2features(inNode, vProds);
  for(vProdIter = vProds.begin(); vProdIter != vProds.end(); ++vProdIter) {
    const SPVector<FeatureTypes>& vFeats = getFeatureBag(*vProdIter);
    if (debug(4))
	cerr << vFeats << endl;
    ret.insert(ret.end(), vFeats.begin(), vFeats.end());
    sz += vFeats.size();
  } 
  return sz;
};

/// call-back method to simulate virtual template parameter for UpdateRule::trees2gradient()
/// computes the margin and return its total cost
SCORE HeadWMTG::trees2gradient(const vector<HeadMTree::iterator>& supervisorMargin, 
			       const vector<HeadMTree::iterator>& learnerMargin,
			       const vector<HeadMTree::iterator>& intersectionMargin,
			       Grammar* pCurrModel,
			       const UpdateRule* uRule) {

  // update productions too
  SCORE ret =_super::trees2gradient(supervisorMargin, learnerMargin, intersectionMargin, pCurrModel, uRule);

  HeadWMTG* pCurrGrammar = dynamic_cast<HeadWMTG*>(pCurrModel);
  if (not pCurrGrammar)
    THROW("HeadWMTG::trees2gradient currModel must be at least a HeadWMTG");
  SPVector<FeatureTypes> supervisorFeats, knownSupervisorFeats, 
    unknownSupervisorFeats, learnerFeats, intersectionFeats;
  size_t supervisorFeatVecSize = 0;
  size_t learnerFeatVecSize = 0;
  size_t intersectionFeatVecSize = 0;
  vector<HeadMTree::iterator>::const_iterator tici;
  SPVector<FeatureTypes>::const_iterator vpci;

  if (debug(3))
    cerr << "learner margin (LearnableWMTG):" << endl;
  // convert each node in the learner's margin to features
  for(tici = learnerMargin.begin(); tici != learnerMargin.end(); ++tici) {
    ++learnerFeatVecSize;
    treeNode2features(*tici, learnerFeats);
  };

  // convert each node in the supervisor's margin to features
  if (debug(3))
    cerr << "supervisor margin (LearnableWMTG):" << endl;
  for(tici = supervisorMargin.begin(); tici != supervisorMargin.end(); ++tici) {
    ++supervisorFeatVecSize; 
    treeNode2features(*tici, supervisorFeats);
  };

  // convert each node in the supervisor's margin to features
  if (debug(3))
    cerr << "intersection margin (LearnableWMTG):" << endl;
  for(tici = intersectionMargin.begin(); tici != intersectionMargin.end(); ++tici) {
    ++intersectionFeatVecSize;
    treeNode2features(*tici, intersectionFeats);
  };

  // find out how many features are previously unseen
  size_t knownFeats = 0;
  for(vpci = supervisorFeats.begin(); vpci != supervisorFeats.end(); ++vpci) {
    if (pCurrGrammar->knows(vpci->get())) {
      ++knownFeats;
      knownSupervisorFeats.push_back(*vpci);
    } else
      unknownSupervisorFeats.push_back(*vpci);
  }

  // get feature update sizes
  SCORE supervisorFeatGradient, learnerFeatGradient, unseenFeatGradient,intersectionFeatGradient;
  uRule->getFeatureGradients(knownFeats, learnerFeatVecSize,intersectionFeatVecSize,
			     supervisorFeatGradient, learnerFeatGradient, intersectionFeatGradient,
			     unseenFeatGradient);

  SCORE posMass = 0.0;
  SCORE negMass = 0.0;
  // update previously unseen supervisor features
  for(vpci = unknownSupervisorFeats.begin(); vpci != unknownSupervisorFeats.end(); ++vpci) {
    updateFeatureScore(vpci->get(), unseenFeatGradient);
    negMass += unseenFeatGradient;
  }

  // update each known supervisor feature
  for(vpci = knownSupervisorFeats.begin(); vpci != knownSupervisorFeats.end(); ++vpci) {
    updateFeatureScore(vpci->get(), supervisorFeatGradient);
    negMass += supervisorFeatGradient;
  }

  // update each intersection feature
  for(vpci = intersectionFeats.begin(); vpci != intersectionFeats.end(); ++vpci) {
    updateFeatureScore(vpci->get(), intersectionFeatGradient);
    negMass += intersectionFeatGradient;
  }

  // update each learner feature
  for(vpci = learnerFeats.begin(); vpci != learnerFeats.end(); ++vpci) {
    updateFeatureScore(vpci->get(), learnerFeatGradient);
    posMass += learnerFeatGradient;
  }

  // each unseen feature counts a whole 1, since we definitely want to
  // keep training if we've just seen stuff that we've never seen
  // before.
  ret += posMass - negMass + unknownSupervisorFeats.size();

  if (debug(2)) {
    cerr << unknownSupervisorFeats.size() << " unknownSuperFeats; ";
    cerr << knownSupervisorFeats.size() << " knownSuperFeats; ";
    cerr << learnerFeats.size() << " learnerFeats" << endl;
//     cerr << "HeadWMTG update mass = " << posMass - negMass
// 	 << " ConvergenceValue = " << ret << endl;
  }

  return ret;
};


BSP<BPO::options_description> HeadWMTG::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDesc){
  //optionsDesc->add_options()
  // ("HeadWMTGTestOption", BPO::value<string>()->default_value("default value"), "test option")
  // ;
  return optionsDesc;
}
