#ifndef _HeadWMTG_H_
#define _HeadWMTG_H_

#include "Vector.hpp"
#include "BoostMacros.hpp"
#include "StringVocabulary.hpp"
#include "ValencyEvent.hpp"
#include "Dependency.hpp"
#include "HeirEvent.hpp"
#include "HeirRoleVectorEvent.hpp"
#include "PAOrderEvent.hpp"
#include "PAVOrderEvent.hpp"
#include "PAIntEvent.hpp"

#include "IndexedPMTG.hpp"

#include <boost/serialization/serialization.hpp>
#include <boost/serialization/is_abstract.hpp>

class HeadWMTG : virtual public IndexedWMTG {

  typedef IndexedWMTG _super;
  typedef Grammar _grammar;

private:

  BSP<BPO::options_description> 
  addOptionDescriptionsTo(BSP<BPO::options_description> priorOptions);

protected:
  // for serialization only
  HeadWMTG() {};

public:

  //this is a hack. It will be removed
  // when we redesign genproc
  typedef boost::variant<ValencyEvent, HeirRoleVectorEvent, HeirEvent,
			 Dependency, PAIntEvent, PAVOrderEvent,NTLink > FeatureTypes;

  //! Destructor.
  virtual ~HeadWMTG();

  /**
   * Constructor.
   * @param configFile The config file of the grammar.
   */
  HeadWMTG(arity_t dm, const string& configFile,
	   char** overrideArgV = NULL, size_t overrideArgC = 0,
	   BSP<BPO::options_description> priorOptions =
	   BSP<BPO::options_description>(new BPO::options_description())
	   );

  //! Increment binary/unary MTG rules.
  //!
  //! If the fanout of this rule is greater than the fanout limit,
  //! then this rule is rejected. The fanout of a production is the
  //! fanout of the involved PAV (we only consider binary/unary
  //! productions here).
  virtual void increment(Production* , SCORE count = UNITYSCORE);
  
  //! Increment features from a grammar
  virtual void increment(const Grammar* pG ) {
    _super::increment(pG);
  };

  /// break the tree down into a vector of features
  /// return the number of features in the vector
  size_t tree2features( const HeadMTree& inTree, vector<BSP<FeatureTypes> >& ret) const;

  /// break a tree node down into a vector of features
  /// return the number of features so generated
  size_t treeNode2features(const HeadMTree::iterator& inNode, 
				     vector<BSP<FeatureTypes> >& ret) const;

  /// call-back method to simulate virtual template parameter for UpdateRule::trees2gradient()
  /// computes the margin and return its total cost
  virtual SCORE trees2gradient(const vector<HeadMTree::iterator>& supervisorMargin, 
			       const vector<HeadMTree::iterator>& learnerMargin,
			       const vector<HeadMTree::iterator>& intersectionMargin,
			       Grammar* pCurrModel,
			       const UpdateRule* uRule);

protected: // methods

  //! Returns the cost of a MTG production.
  virtual SCORE getProdScore(Production*) const;

  // test whether a feature exists in the model
  virtual bool knows(FeatureTypes* pFT) const = 0;

  //! Returns the cost of a model event.
  virtual SCORE getFeatureScore(const FeatureTypes&) const = 0;

  //! non-virtual wrapper for pointers
  void updateFeatureScore(const FeatureTypes* pFT, SCORE sc) {
    increment(*pFT, sc);
  }

  //! increment events. Note that score is not log.
  virtual void increment(const FeatureTypes&, SCORE score) = 0;

  /// Decompose an inference or a production into features
  virtual SPVector<FeatureTypes> getFeatureBag(Production* prod) const = 0;
  SPVector<FeatureTypes> getFeatureBag(BSP<Production> prod) const {
    return getFeatureBag(prod.get());
  };

  friend class boost::serialization::access;

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    ar & boost::serialization::base_object<_super>(*this);
  };
};

BOOST_IS_ABSTRACT(HeadWMTG);
namespace HashNS {
  template<> struct hash<HeadWMTG::FeatureTypes> {
    size_t operator()(const HeadWMTG::FeatureTypes& p1) const {
      static VariantHash hasher;
      return (boost::apply_visitor(hasher, p1));
    };
  };
  template<> struct hash<HeadWMTG::FeatureTypes*> {
    size_t operator()(const HeadWMTG::FeatureTypes* p1) const {
      static VariantHash hasher;
      return (boost::apply_visitor(hasher, *p1));
    };
  };
  template<> struct hash<BSP<HeadWMTG::FeatureTypes> > {
    size_t operator()(const BSP<HeadWMTG::FeatureTypes> p1) const {
      static VariantHash hasher;
      return (boost::apply_visitor(hasher, *p1));
    };
  };
};

namespace std {
  template<>
  struct less<HeadWMTG::FeatureTypes> {
    typedef HeadWMTG::FeatureTypes FeatureTypes;
    bool operator()(FeatureTypes& f1, FeatureTypes& f2) const {
      static VariantLess vtor;
      return (boost::apply_visitor(vtor, f1, f2));
    };
    bool operator()(const FeatureTypes& f1, const FeatureTypes& f2) const {
      static VariantLess vtor;
      return (boost::apply_visitor(vtor, f1, f2));
    };
    bool operator()(const BSP<FeatureTypes> f1, const BSP<FeatureTypes> f2) const {
      static VariantLess vtor;
      return (boost::apply_visitor(vtor, *f1,*f2));
    };
  };
};

#endif
