#include "HeadWMTGB.hpp"
#include "TLink.hpp"
#include "COMPILER-DEPENDENT_HEADERS.hpp"
#include "HeadMTree.hpp"
#include "HeadProduction.hpp"
#include "LazyCD.hpp"
#include "GlobalVocab.hpp"
#include "boost/bind.hpp"
#include "MoreAlgorithms.hpp"
#include <functional>
#include <vector>
#include "Exceptions.hpp"
#include "boost/variant.hpp"
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

/*
 * Constructor.
 * @param configFile The config file of the grammar.
 */
HeadWMTGB::HeadWMTGB(arity_t dm, const string& configFile,
		     char** overrideArgV, size_t overrideArgC,
		     BSP<BPO::options_description> optionsDescription
		     )
  :
  _grammar(dm),
  _indexed(dm,configFile, overrideArgV, overrideArgC, addOptionDescriptionsTo(optionsDescription)),
  // m_forceDown(true),
  m_pValDistrib( new ScoreSet<ValencyEvent>),
  m_pDepDistrib( new ScoreSet<Dependency>),
  m_pHeirDistrib( new ScoreSet<HeirEvent>),
  m_pHeirRoleVectorDistrib(new ScoreSet<HeirRoleVectorEvent>),
  m_pPermDistrib( new ScoreSet<PAVOrderEvent>),
  m_depHeirDistribs((size_t)pow(2,(double)m_arity)-2) {

  for(unsigned short d = 0; d < pow(2,(double)arity())-2; d++){
    m_depHeirDistribs[d] = new ScoreSet<Dependency>;
  }
}

//! Destructor. To clear the memory allocated.
HeadWMTGB::~HeadWMTGB() {
  delete m_pValDistrib;
  delete m_pDepDistrib;
  delete m_pHeirRoleVectorDistrib;
  delete m_pHeirDistrib;
  delete m_pPermDistrib;
  
  for(arity_t d = 0; d < pow(2,(double)arity())-2; d++)
    delete m_depHeirDistribs[d];
}

//! make it empty
void HeadWMTGB::clear() {
  _super::clear();

  m_pValDistrib->clear();
  m_pDepDistrib->clear();
  m_pHeirRoleVectorDistrib->clear();
  m_pHeirDistrib->clear();
  m_pPermDistrib->clear();
  
  for(arity_t d = 0; d < pow(2,(double)arity())-2; d++)
    m_depHeirDistribs[d]->clear();
};


//! returns a pointer to the correct distribution given a depend event.
ScoreSet<Dependency>*
HeadWMTGB::getRelevantScoreSet(const Dependency& dep, int& i)const{
  arity_t d = arity();
  if(d == 1){
    i = -1;
    return m_pDepDistrib;
  } else {
    i = dep.getDistributionNumber();
    if(i == -1)
      return m_pDepDistrib;
    else{
      assert(abs(i) < pow(2,(double)arity())-2);
      return m_depHeirDistribs[i];
    }
  }
}

// test whether a feature exists in the model
bool HeadWMTGB::knows(FeatureTypes* pFT) const {
  if(const Dependency* dep = boost::get<const Dependency>(pFT) ){
    int i;
    ScoreSet<Dependency>* ssdp = getRelevantScoreSet(*dep, i);
    return (ssdp->contains(*dep));
  } else if(const HeirRoleVectorEvent* heirVectorEvent =
	    boost::get<const HeirRoleVectorEvent>(pFT) ){
    return (m_pHeirRoleVectorDistrib->contains(*heirVectorEvent));
  } else if(const HeirEvent* heirEvent = boost::get<const HeirEvent>(pFT) ){
    return (m_pHeirDistrib->contains(*heirEvent));
  } else if(const ValencyEvent* v = boost::get<const ValencyEvent>(pFT)){
    return (m_pValDistrib->contains(*v));
  } else if(const PAVOrderEvent* ordE = boost::get<const PAVOrderEvent>(pFT)){
    return (m_pPermDistrib->contains(*ordE));
  } else {
    // should never get to here
    THROW("Unknown feature type!" << endl);
  }

  return false;
};

// Returns the cost of a model event.
SCORE HeadWMTGB::getFeatureScore(const FeatureTypes& feature) const {
  SCORE score;

  if(const Dependency* dep = boost::get<const Dependency>(&feature) ){
    int i;
    ScoreSet<Dependency>* pSDep = getRelevantScoreSet(*dep, i);
    score = pSDep->lookup(*dep);
    if(debug(5)){
      dout()<<"D " << i << ": " << *dep;
    }
  } else if(const HeirRoleVectorEvent* heirVectorEvent =
	    boost::get<const HeirRoleVectorEvent>(&feature) ){
    if(debug(5)){
      dout()<<"HVE: "<< *heirVectorEvent;
    }
    score = m_pHeirRoleVectorDistrib->lookup(*(heirVectorEvent));
  } else if(const HeirEvent* heirEvent = boost::get<const HeirEvent>(&feature) ){
    if(debug(5)){
      dout()<<"H: "<<*heirEvent;
    }
    score = m_pHeirDistrib->lookup(*heirEvent);
  } else if(const ValencyEvent* v = boost::get<const ValencyEvent>(&feature)){
    if(debug(5)){
      dout()<<"V: "<<*v;
    }
    score = m_pValDistrib->lookup(*v);
  } else if(const PAVOrderEvent* ordE = boost::get<const PAVOrderEvent>(&feature)){
    if(debug(5)){
      dout()<<"O: "<< *ordE;
    }
    score = m_pPermDistrib->lookup(*ordE);
  } else {
    // should never get to here
    THROW("Unknown feature type!" << endl);
  }

  if(debug(5)){
    dout() << " " << score << endl;
  }

  return score;
}


//! Increment features from a grammar 
void HeadWMTGB::increment(const Grammar* pG ) {
  const HeadWMTGB* pHWMTG = dynamic_cast<const HeadWMTGB*>(pG);
  if (not pHWMTG)
    THROW("Grammar type mismatch in HeadWMTGB::increment");
  _super::increment(pHWMTG);
  m_pValDistrib->increment(*(pHWMTG->m_pValDistrib));
  m_pDepDistrib->increment(*(pHWMTG->m_pDepDistrib));

  m_pHeirRoleVectorDistrib->increment(*(pHWMTG->m_pHeirRoleVectorDistrib));
  m_pHeirDistrib->increment(*(pHWMTG->m_pHeirDistrib));
  m_pPermDistrib->increment(*(pHWMTG->m_pPermDistrib));
  
  static double numDHDistribs = pow(2,(double)arity())-2;
  for(arity_t d = 0; d < numDHDistribs; ++d)
    m_depHeirDistribs[d]->increment(*(pHWMTG->m_depHeirDistribs[d]));
}

// return the sum of the scores of all features
SCORE HeadWMTGB::mass() {
  SCORE ret = _super::mass();
  ret += m_pValDistrib->sum();
  ret += m_pHeirRoleVectorDistrib->sum();
  ret += m_pHeirDistrib->sum();
  ret += m_pDepDistrib->sum();
  for(arity_t d = 0; d < pow(2,(double)arity())-2; d++)
    ret += m_depHeirDistribs[d]->sum();
  ret += m_pPermDistrib->sum();

  return ret;
};

// increment features. Note that score is not log.
void HeadWMTGB:: increment(const FeatureTypes& feature, SCORE score) {
  if(const Dependency* dep = boost::get<const Dependency>(&feature) ){
    int i;
    if(debug(3))cerr << "HeadWMTGB" << m_uniqueID << "::incr by " << score << " Dep " << *dep << endl;
    getRelevantScoreSet(*dep,i )->increment(*dep, score);
  } else if(const HeirEvent* heirEvent = boost::get<const HeirEvent>(&feature) ){
    if(debug(3)) cerr << "HeadWMTGB" << m_uniqueID << "::incr by " << score << " HeirEvent " << *heirEvent << endl;
    m_pHeirDistrib->increment(*heirEvent, score);
  } else if(const HeirRoleVectorEvent* heirVectorEvent =
	    boost::get<const HeirRoleVectorEvent>(&feature) ){
    if(debug(3)) cerr << "HeadWMTGB" << m_uniqueID << "::incr by " << score << " HeirVE " << *heirVectorEvent << endl;
    m_pHeirRoleVectorDistrib->increment(*heirVectorEvent, score);
  } else if(const ValencyEvent* v = boost::get<const ValencyEvent>(&feature)){
    if(debug(3)) cerr << "HeadWMTGB" << m_uniqueID << "::incr by " << score << " Valency " << *v << endl;
    m_pValDistrib->increment(*v, score);
  } else
    if(const PAVOrderEvent* ordE = boost::get<const PAVOrderEvent>(&feature)){
      if(debug(3)) cerr << "HeadWMTGB" << m_uniqueID << "::incr by " << score << " PAVOrderEvent " << *ordE << endl;
      m_pPermDistrib->increment(*ordE, score);
    }

    else {
      THROW("Cannot increment features." << endl);
    }
}

// Read from istream
istream& HeadWMTGB::get(istream& in) {
  arity_t d;
  vector<SCORE> minprbs;

  _super::get(in);

  /*
   * read valency event distribution.
   */
  cerr<<"Reading valency event distribution ..."<<endl;
  ValencyEvent valE(arity());

  // if(m_estType == MLE){
  m_pValDistrib->read(in, &valE);
  // } else {
  // cerr<<"Unknown estimator type!"<<endl;
  // exit(1);
  // }

  NTLink lnl1, lnl2;
  NTLink ntlnk;

  /*
   * read heir vector event distribution.
   */
  cerr<<"Reading heir vector event distribution ..."<<endl;
  HeirRoleVectorEvent heirVE(arity());
  HeirRoleVector hv = heirVE.heirVector();
  ImDom<NTLink, short> pe = heirVE.getParValPair();

  // if(m_estType == MLE){
  m_pHeirRoleVectorDistrib->read(in, &heirVE);
  // } else {
  // cerr<<"Unknown estimator type!"<<endl;
  // exit(1);
  // }

  /*
   * read heir event distribution.
   */
  cerr<<"Reading heir event distribution ..."<<endl;
  HeirEvent heirE(arity());
  m_pHeirDistrib->read(in, &heirE);

  /*
   * read dependency event distribution.
   */
  cerr<<"Reading dependency distribution ..."<<endl;
  Dependency depE(arity());
  m_pDepDistrib->read(in, &depE);

  cerr<<"Reading dependencyHeir distributions ..."<<endl;
  for(d = 0; d < pow(2,(double)arity())-2; d++){
    unsigned int rhsSize = Dependency::getLHSsizefromNumber(d, arity());
    unsigned int lhsSize = arity()*4 - rhsSize;

    Vector<unsigned short> rhv(rhsSize,0);
    Vector<unsigned short> lhv(lhsSize,0);

    m_depHeirDistribs[d]->read(in, &depE);
  }

  cerr<<"Reading permutation distributions ..."<<endl;
  PAVOrderEvent ordE(arity());
  m_pPermDistrib->read(in, &ordE);

  return in;
}

//! Write
ostream& HeadWMTGB:: put(ostream& out) const {
  arity_t d;

  _super::put(out);

  cerr<<"Writing valency event distribution ..."<<endl;
  out<<*m_pValDistrib;

  cerr<<"Writing heir vector event distribution ..."<<endl;
  out<<*m_pHeirRoleVectorDistrib;

  cerr<<"Writing heir event distribution ..."<<endl;
  out<<*m_pHeirDistrib;

  cerr<<"Writing dependency distribution ..."<<endl;
  out<<*m_pDepDistrib;
  cerr<<"Writing Dependency/Heir distribution ..."<<endl;
  for(d = 0; d < pow(2,(double)arity())-2; d++)
    out<<*m_depHeirDistribs[d];

  cerr<<"Writing permutation distributions ..."<<endl;
  out<<*m_pPermDistrib;

  return out;
}

/**
 * Decomposes an MTG production into features
 */
SPVector<HeadWMTGB::FeatureTypes> HeadWMTGB::
getFeatureBag(Production* pProd) const {

  assert(pProd->rank() <= 2);

  SPVector<FeatureTypes> features;

  // valency event
  features.push_back(BSP<FeatureTypes>
		     (new FeatureTypes(ValencyEvent(pProd->lhs(), pProd->rank()))));

  HeadProduction* pHProd = dynamic_cast<HeadProduction*>(pProd);

  // if it's not headed, then there are no other features of interest
  if (not pHProd)
    return features;

  // heir role vector and heir event
  if(pHProd->rank() >= 1){
    Vector<HeirRoleVectorEvent> hvs(1, HeirRoleVectorEvent(ImDom<NTLink, short>(pHProd->lhs(), pHProd->rank()), pHProd->heirRole()));
    Vector<HeirRoleVectorEvent>::iterator it;
    for(it = hvs.begin(); it != hvs.end(); it++)
      features.push_back(BSP<FeatureTypes>(new FeatureTypes(HeirRoleVectorEvent(*it))));
    NTLink* nt = pHProd->heirRole().getHeirEvent(pHProd->rhsLinks());
    features.push_back(BSP<FeatureTypes>(new FeatureTypes(HeirEvent(pHProd->lhs(), *nt))));
    delete nt;
  }

  // dependency events
  for(size_t i = 0; i < pHProd->rank(); i++)
    if(boost::get<NTLink>(&pHProd->rhsLinks()[i])){
      NTLink* nt = pHProd->heirRole().getDependency(boost::get<NTLink>(pHProd->rhsLinks()[i]),i);
      if(nt != NULL){
	features.push_back(BSP<FeatureTypes>(new FeatureTypes(Dependency(pHProd->lhs(), *nt))));
      }
      delete nt;
    }

  // permutation event, only comes from n-ary productions with n > 1.
  if(pHProd->rank() > 1){
    PAVOrderEvent::FirstType links;
    links.parent() = boost::get<NTLink>(pHProd->rhsLinks()[0]);
    links.child() = boost::get<NTLink>(pHProd->rhsLinks()[1]);
    features.push_back(BSP<FeatureTypes>
		       (new FeatureTypes(PAVOrderEvent(links, pHProd->rhsPerm()))));
  }

  return features;

}

BSP<BPO::options_description> HeadWMTGB::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  return optionsDescription;
}
