#include "Hedge.hpp"
#include "Exceptions.hpp"

//! Constructor.
Hedge::Hedge(const arity_t& dim)
  : _super(DSV(dim)), m_label(NTLink(dim)) 
{}

//! copy constructor.
Hedge::Hedge(const Hedge& other) 
  :_super(other), m_label(other.m_label)
{}

//! Constructor.
Hedge::Hedge(const NTLink& label, const DSV& span)
  : _super(span), m_label(label) 
{}


//! constructor from tree iterator and DSV
Hedge::Hedge(const HeadMTree::iterator& it, const DSV& span)
  : _super(span), m_label(BGET_NT_LABEL(*it))
{}

Hedge::~Hedge() {
  if (Dbug::isDebugAll(6))
    cerr << "Destroying Hedge " << m_uniqueID << endl;
}

int Hedge::cmp(const Item* item, bool sameType) const {
  int ret = _super::cmp(item, sameType);
  if (ret != 0)
    return ret;

  // constituent label
  const Hedge* hedge = static_cast<const Hedge*>(item);
  NTLink label = hedge->getNTLink();
  if((ret = m_label.cmp(label)) != 0) return ret;

  return 0;
}

bool Hedge::operator == (const Hedge& other) const {
  if (not (m_label == other.m_label))
    return false;
  if (not (m_span == other.m_span))
    return false;
  if (not (m_fanoutV == other.m_fanoutV))
    return false;
  // it is necessary to compare sizes, cuz otherwise different items
  // can appear the same if they differ only in their size on the
  // output dimension during translation
  if (not (m_sizes == other.m_sizes))
    return false;

  return true;
}


//! checks whether two items are equal in content and type
bool Hedge::equal_to(const Item& item) const {
  const Hedge* other = dynamic_cast<const Hedge*>(&item);
  if(other){
    // same type.
    return operator==(*other);
  } else {
    // different type.
    return false;
  }
}

//! checks whether two items are equal in content and type
bool Hedge::equal_to(const Item* item) const {
  const Hedge* other = dynamic_cast<const Hedge*>(item);
  if(other){
    // same type.
    return operator==(*other);
  } else {
    // different type.
    return false;
  }
}

/// checks whether two items are equal in content and type, with one
/// exception: If the other item has DSV components that are inactive
/// in this hedge, then that difference is ignored.
bool Hedge::equalUpToInactiveDSpans(const Item& item) const {
  const Hedge* other = dynamic_cast<const Hedge*>(&item);
  if(not other)
    // different type
    return false;
  if (not (m_label == other->m_label))
    return false;
  if (not (m_fanoutV == other->m_fanoutV))
    return false;
  // minus one cuz sizes are 1-indexed
  arity_t lastActiveSpan = m_span.size() - 1;
  for(arity_t dim = 0; dim < lastActiveSpan; ++dim)
    if (not (m_span[dim] == other->m_span[dim]))
      return false;
  
  return true;
}

//! Returns the number of non-null dims.
size_t Hedge::numActiveDims() const {
  size_t ret = 0;

  for(unsigned int i = 0; i < m_fanoutV.size(); i++){
    if(m_fanoutV[i] > 0) {
      ++ret;
    }
  }
  return ret;
}

size_t Hedge::volume() const
{ return ((DSV*)(this->getDSV()))->getSize(); }

size_t Hedge::hash() const {
  static HashNS::hash< DSV > hDSV;
  static HashNS::hash< NTLink > hNTL;
  size_t ret = hNTL(m_label) * 1063
    + hDSV(m_span);
  return ret;
}

Hedge& Hedge::operator=(const Hedge& hedge) {
  // Base class
  _super::operator=(hedge);

  // span
  m_span.DSV::operator =( *((DSV*)hedge.getDSV()));

  // label
  NTLink label = hedge.getNTLink();

  m_label = label;

  return *this;
}

//! Sets the label.
void Hedge::setLabel(const NTLink& l) {
   m_label = l;
}

//! Sets the constituent label of the item on just one dimension
void Hedge::setLabel(const arity_t d, const Nonterminal* l) {
  m_label.setNT(d,l->clone());
}

const Nonterminal* Hedge::getNTLink(arity_t d) const
{ return m_label.getNT(d); }

void Hedge::setNTLink(arity_t d, Nonterminal* n)
{ m_label.setNT(d,n); }

NTLink& Hedge::getNTLink()
{ return m_label; }

const NTLink& Hedge::getNTLink()const
{ return m_label; }

// return the preterminal label of the headItem on the given dimension
SynCat Hedge::pretermLabel(arity_t dim) const {
  const Item* pI = headItem(dim);
  const Hedge* pH = dynamic_cast<const Hedge*>(pI);
  return pH->getNTLink().getNT(dim)->getSynCat();
};

//! Dump for debugging.
void Hedge::dump(ostream& o) const {
  o<<"Hedge:= " << m_uniqueID
   << " DSV=" << m_span;
  if(m_pHiddenDSVs) {
    o<< " hDSV=";
    Vector<DSV>::const_iterator vci;
    for(vci = m_pHiddenDSVs->begin(); vci !=  m_pHiddenDSVs->end(); ++vci)
      o << *vci << " ";
  }
  o<< "sizes=";
  print_vec(o, m_sizes);
  o <<" Label=("<<m_label<<"); "
    <<"g= "<<getInsideScore()<<"; "
    <<"h= "<<getOutsideScore()<<"; ";
  if(m_spInf) {
    o <<"inferred from ";
    m_spInf->dumpAnts(o);
  };
}

Item* Hedge::get1DItem(arity_t d)const{
  NTLink ln = NTLink(1);
  ln.setNT(0,m_label.getNT(d)->clone());

  DSV dspan = DSV(1);
  dspan[0] = m_span[d];

  Hedge* ret = new Hedge(ln,dspan);
  return ret;

};
