#ifndef _Hedge_H_
#define _Hedge_H_

#include <vector>
#include <assert.h>
#include "DSV.hpp"
#include "NTLink.hpp"
#include "Item.hpp"
#include "Exceptions.hpp"
#include "HeirRoleVector.hpp"
#include "HeadMTree.hpp"

class Inference;

using namespace std;

//! This class represents a Hyper Edge item.
class Hedge : public Item {
    typedef Item _super;
public:

  //! constructor based on arity only
  Hedge(const arity_t&);

  //! constructor from constituent label and DSV
  Hedge(const NTLink& label, const DSV& span);

  //! constructor from tree iterator and DSV
  Hedge(const HeadMTree::iterator& it, const DSV& span);

  //! copy constructor
  Hedge(const Hedge&);

  //! Destructor
  virtual ~Hedge();

  // returns a copy of this
  virtual Item* clone() const {
    return new Hedge(*this);
  }

  ///////////////////////////////////////////////////////////////
  // Comparison Interfaces //
  ///////////////////////////////////////////////////////////////

  //! Gives an order to two Hedges on agenda.
  //! \return : - 0 equal;
  //! - 1 greater;
  //! - -1 smaller;
  virtual int cmp(const Item* hedge, bool sameType = true) const;

  //! Equality test by comparing signature: (label, span, fanoutv).
  virtual bool operator == (const Hedge& hedge) const;

  //! Compares whether two items are equal in content and type.
  //! Compared with operator==, this method is virtual.
  virtual bool equal_to(const Item& item) const;

//! Compares whether two items are equal in content and type.
  //! Compared with operator==, this method is virtual.
  virtual bool equal_to(const Item* item) const;

  /// checks whether two items are equal in content and type, with one
  /// exception: If the other item has DSV components that are inactive
  /// in this hedge, then that difference is ignored.
  bool equalUpToInactiveDSpans(const Item& item) const;
  
  /////////////////////////////////////////////////////////////
  // Information Access //
  ////////////////////////////////////////////////////////////

  //! Returns the language dimensionality.
  virtual arity_t arity() const { return m_label.arity();}

  /// marginalization
  virtual Item* get1DItem(arity_t d) const;

  //! Sets the label.
  virtual void setLabel(const NTLink&);

  //! Sets the constituent label of the item on just one dimension
  virtual void setLabel(const arity_t, const Nonterminal*);

  //! Returns the label as a NTLink
  NTLink& getNTLink();

  //! Returns copy of the label as a NTLink
  const NTLink& getNTLink() const;

  //! Returns the label of dim d
  const Nonterminal* getNTLink(arity_t d) const;

  void setNTLink(arity_t d, Nonterminal* n);

  //! for hashing
  virtual size_t hash() const;

  //! for duplication
  virtual Hedge& operator=(const Hedge& hedge);

  size_t volume() const;

  //! returns the number of non-null dims
  size_t numActiveDims() const;

  // return the preterminal label of the headItem on the given dimension
  SynCat pretermLabel(arity_t dim) const;

  // dump for debugging
  virtual void dump(ostream& o) const;

  //!unencodes the heir role vector stored in the lexical heads of the cons
  inline static HeirRoleVector unencodeHeads(const Hedge* trigger, const Hedge* matchingItem, Hedge* cons){

    //Now we have the HeirRoleVector encoded in the Antecedents lexical heads.
    //Unencode it.
    HeirRoleVector hrv(trigger->arity());
    for(unsigned int i = 0; i < trigger->arity(); i++){
      int s = (cons)->getNTLink().getTLink()[i];
      hrv.setRole(i,s);
      if(s == 1)
	(cons)->getNTLink().setTerminal((trigger)->getNTLink().getTLink()[i],i);
      if(s == 2)
      	(cons)->getNTLink().setTerminal((matchingItem)->getNTLink().getTLink()[i],i);
    }
    return hrv;

  }
  
protected:

  //! The label of the item.
  NTLink m_label;
};

namespace HashNS {
  //! Template specialization of STL's hash object.
  template<> struct hash<Hedge> {
    size_t operator() (const Hedge& hedge) const {return hedge.hash();}
  };

  template<> struct hash< Hedge* >
  { size_t operator() (const Hedge* pitem) const {return pitem->hash();} };

  template<> struct hash< BSP<Hedge> >
  { size_t operator() (const BSP<Hedge> pitem) const {return pitem->hash();} };

  template<> struct hash< BSP<const Hedge> >
  { size_t operator() (const BSP<const Hedge> pitem) const {return pitem->hash();} };
}

#endif
