#ifndef _HEIRVECTOR_H_
#define _HEIRVECTOR_H_

#include "Vector.hpp"
#include "GPDefs.hpp"
#include "NTLink.hpp"
#include "ProdRHSElem.hpp"
#include "RawHeirRoleVector.hpp"
#include <boost/serialization/serialization.hpp>
#include "serialize-variant.hpp"
#include "Vocabulary.hpp"

//! "HeirRoleVector" class. Tells which child is the heir in each
//! dimension.
/// This class is implemented using a combination of two design
/// patterns (see Gamma et al. (1994) _Design Patterns_). It is a
/// flyweight and simultaneosly a proxy. The flyweight factory *is*
/// the proxy. Like all flyweights, it uses a vocabulary to avoid
/// creating multiple copies of the same object. Each instance needs
/// to record only an index into the vocabulary. In the present case,
/// the vocabulary lives as a static object inside the flyweight
/// class. Thus, other classes don't know that this class is a
/// flyweight, which better encapsulates the implementation.
class HeirRoleVector{
  typedef RawHeirRoleVector _super;
public:

  //! default constructor
  HeirRoleVector();

 //! constructor from a RawHeirRoleVector
  HeirRoleVector(const RawHeirRoleVector& raw);

  //! Constructor.
  HeirRoleVector(const arity_t& dim);

  //! Constructor.
  HeirRoleVector(const arity_t& dim, short s);
  
  /// constructor from super
  HeirRoleVector(const Vector<short>& other);

  virtual ~HeirRoleVector() {};
  
  //! Return a normalized version of the HeirRoleVector
  //! (make sure first position is 1)
  HeirRoleVector normal() const;

  //!Normalize this HeirRoleVector
  //! (make sure first position is 1)
  //! N.B. this should ONLY be used when being made into an event.
  //! Once in a distribution, we only care about normalized HVs, but
  //! when an HV is in a production, we can't normalize it unless we
  //! switch around all the other parts of the production to match.
  void normalize();

  HeirRoleVector get1DItem(arity_t d)const;

  //! detects whether the HeirRoleVector is normalized.
  bool isNormal()const;

  //! returns true if this heir vector is head split.
  bool headSplit() const;

  //! dimension of heirVector.
  arity_t arity() const {
    return sm_voc.getWord(m_index)->arity();}

  //! Clear the content, but retain the dimension.
  //  void clear();

  //! invert the HeirRoleVector (1->2 2->1 0->0)
  HeirRoleVector invert() const;

  //! Judges whether empty.
  bool empty() const;

  //! For hashing.
  size_t hash() const;

  //! Number of dimensions of heirVector.
  arity_t arity(){
   return sm_voc.getWord(m_index)->arity();
  }

  //! set the role in dimension a to s
  void setRole(arity_t a, short s);
 
  //! get the role in dimension a
  short getRole(arity_t a)const;
 

  //! Equality test
  bool operator == (const HeirRoleVector& other) const {
    return (m_index == other.m_index);
  }

  //! Less than test.
  bool operator < (const HeirRoleVector& other) const{
  return (m_index < other.m_index);
  }

  //! Less than test.
  bool operator > (const HeirRoleVector& other) const{
  return (m_index > other.m_index);
  }

  int cmp(const HeirRoleVector& other) const;

  //!output.
  ostream& serialize(ostream&out) const;

  //! Judges whether the indices are not increasing order in the
  //! first dimension. <1,2> -- false; <2,1> -- true.
  bool improperOrder() const;

  //! Returns the number of roles on dimension d.
  short rank() const;

  //! Remove one role, and normalize the PAV.
  void deleteRoleAndNormalize(const short i);

  //! Remove one role, and DOT NOT normalize the PAV.
  void deleteRoleAndNormalizeOnly(const short i);

  short maxRole() const;

  set<short> roleSet() const;

  //! Inputs HV.
  friend istream& operator>>(istream& in, HeirRoleVector& hv);

  //! Outputs HV.
  friend ostream& operator<<(ostream& out, const HeirRoleVector& hv);

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    ar & boost::serialization::base_object<Vector<short> >(*this);
  };

  ostream& put(ostream& out) const;

  //! Returns true is there is any information about heirs.
  bool containsHeir() const;

  //! returns true if the HeirRoleVector contains "role".
  bool contains(short role)const;

  //We are passed in a Vector which is the RHS,and
  // we export an Heir, given the this HeirRoleVector.
  NTLink* getHeirEvent (const Vector<ProdRHSElem>& d) const;

  NTLink* getHeirEvent (const NTLink& l1, const NTLink& l2) const;

  NTLink* getDependency(const NTLink&, short val) const;

  VocabularyIndex  getIndex()const{
    return m_index;
  }

  static Vocabulary<RawHeirRoleVector>  getVocab(){
    return sm_voc;
  }

protected: // data

  // currently an unsigned int,
  // but can probably be changed to unsigned short
  VocabularyIndex m_index;

  // currently hashed in both directions,
  // but one dir should be a vector to be faster
  static Vocabulary< RawHeirRoleVector > sm_voc;



};

namespace HashNS{
  template<>
  struct hash<HeirRoleVector > {
    size_t operator()(const HeirRoleVector& hv) const { return hv.hash();}
  };
}

#endif
