#ifndef _HeirTree_C_
#define _HeirTree_C_
#include "HeirTree.hpp"
#include "HeadMTreeNode.hpp"
#include "HeirRoleVector.hpp"
#include "Exceptions.hpp"
#include "boost/tuple/tuple.hpp"

// Many of the methods of the tree base class should have been static
// but aren't.  They take an iterator and a tree, but don't actually
// use the tree.  Instead of passing around an extra tree param, we
// use a static dummy tree where necessary.
template<class _DecoratedTree >
HeirTree<_DecoratedTree> HeirTree<_DecoratedTree>::dummyTree = HeirTree<_DecoratedTree>();

// Backup my own head and let the head pointer point to the
// head of the decorated object.
template<class _DecoratedTree >
HeirTree<_DecoratedTree>:: HeirTree(const _super& base) {
  // __head_bk = this->getHead();
  //setHead(base.getHead());
}

// restore the my head pointer.
template<class _DecoratedTree >
HeirTree<_DecoratedTree>::~HeirTree() {} //setHead(__head_bk);}

template<class _DecoratedTree >
void HeirTree<_DecoratedTree>::initHeirVectors(const iterator& it) {
  if( (BGET_HEIRVEC(*it)).size() == 0){
    (BGET_HEIRVEC(*it)) = Vector<unsigned int>(BGET_NT_LABEL(*it).arity());
    for(unsigned int i = 0; i < (BGET_HEIRVEC(*it)).size(); i++){
      (BGET_HEIRVEC(*it))[i] = 0;
    }
  }
}

template<class _DecoratedTree >
void HeirTree<_DecoratedTree>::setHeir(const iterator& it) {
  if(dummyTree.is_valid(it)) {
    iterator it_parent = dummyTree.parent(it);
    if(dummyTree.is_valid(it_parent)) {
      if( (BGET_HEIRVEC(*it_parent)).size() == 0){
	initHeirVectors(it_parent);
      }
      iterator children;
      unsigned int pos = 1;
      bool found = false;
      for(children = it_parent.begin(); children != it_parent.end(); children++){
	if(children == it){
	  found = true;
	  break;
	}
	else pos++;
      }
      assert(found);
      for(unsigned int i = 0; i < BGET_HEIRVEC(*it_parent).size(); i++)
	BGET_HEIRVEC(*it_parent)[i] = pos;
    }
  } else {
    THROW("Error: try to set an invalid tree iterator to be an heir!"<<endl);
  }
}

template<class _DecoratedTree >
bool HeirTree<_DecoratedTree>::isHeir(const iterator& it)  {
  if(dummyTree.is_valid(it)) {
    const iterator it_parent = dummyTree.parent(it);
    if(dummyTree.is_valid(it_parent)) {
      unsigned int pos = 1;
      bool found = false;
      for(iterator children = it_parent.begin(); children != it_parent.end(); children++){
	if(children == it){
	  found = true;
	  break;
	}
	else pos++;
      }
      assert(found);
	
      for(unsigned int i = 0; i < BGET_HEIRVEC(*it_parent).size(); i++){
	if(BGET_HEIR( BGET_HEIRVEC(*it_parent)[i]) == pos) {
	  return true;
	}
      }
      return false;
    }
    else return true;
  }
  else {
    THROW("Error: judge heir for an undefined iterator.\n" << endl);
  }
  //Never gets here but compiler was warning.
  return true;
}

/// overload for callers that happen to supply a parent, which we will ignore
template<class _DecoratedTree >
bool HeirTree<_DecoratedTree>::isHeir(const iterator& child, const iterator& parent)  {
  return isHeir(child);
}

template<class _DecoratedTree >
bool HeirTree<_DecoratedTree>::isHeir(const iterator& it, const arity_t& d)  {
  if(dummyTree.is_valid(it)) {
    const iterator it_parent = dummyTree.parent(it);

    if(dummyTree.is_valid(it_parent)) {
      unsigned int pos = 1;
      bool found = false;
      for(iterator children = it_parent.begin(); children != it_parent.end(); children++){
	if(children == it){
	  found = true;
	  break;
	}
	else pos++;
      }
      assert(found);
      
      if(BGET_HEIR( BGET_HEIRVEC(*it_parent)[d]) == pos)
	return true;
      else
	return false;
    }
    else return false;
  }
  else {
    THROW("Error: judge heir for an undefined iterator." << endl);
  }
  //Never gets here but compiler was warning.
  return true;
}

/// overload for callers that happen to supply a parent, which we will ignore
template<class _DecoratedTree >
bool HeirTree<_DecoratedTree>::isHeir(const iterator& child, const iterator& parent, const arity_t& d) {
  return isHeir(child, d);
}

/// returns an iterator to the heir node on the given dimension
template<class _DecoratedTree >
typename HeirTree<_DecoratedTree>::iterator 
HeirTree<_DecoratedTree>::
findHeir(const iterator & it, arity_t dim) {
  typename HeirTree<_DecoratedTree>::sibling_iterator sit = dummyTree.begin(it);
  while (sit != dummyTree.end(it) && ! (isHeir(sit, dim)))
    ++sit;
  return sit;
};

template<class _DecoratedTree >
void HeirTree<_DecoratedTree>::setHeirQueue(const iterator& it, queue<short>& d){

  if(dummyTree.is_valid(it)){
    while(!d.empty()){
      short s = d.front();
      d.pop();
      iterator it_parent = dummyTree.parent(it);
      if(dummyTree.is_valid(it_parent)) {
	  //if our HeirVector has not been initialized.
	  if( (BGET_HEIRVEC(*it_parent)).size() == 0){
	    initHeirVectors(it_parent);
	  }
	  
	  unsigned int pos = 1;
	  bool found = false;
	  for(iterator children = it_parent.begin(); children != it_parent.end(); children++){
	    if(children == it){
	      found = true;
	      break;
	    }
	    else pos++;
	  }
	  assert(found);
	  
    	  BGET_HEIRVEC(*it_parent)[s] = pos;
	}
    }
  }
  else {
    THROW("Error: try to set an invalid tree iterator to be an heir!"<<endl);
  }
}

//! Returns a queue which has all the dimensions for which
//! this item is an heir in the parent.
template<class _DecoratedTree >
queue<short> HeirTree<_DecoratedTree >::getHeirQueue(const iterator& it) {
  queue<short> ret;
  if(dummyTree.is_valid(it)){
    const iterator it_parent = dummyTree.parent(it);
    if(dummyTree.is_valid(it_parent)){
      for(unsigned int i = 0; i < BGET_HEIRVEC(*it_parent).size(); i++){
	unsigned int pos = 1;
	bool found = false;
	for(iterator children = it_parent.begin(); children != it_parent.end(); children++){
	    if(children == it){
	      found = true;
	      break;
	    }
	    else pos++;
	}
	assert(found);
	
	if(BGET_HEIR(BGET_HEIRVEC(*it_parent)[i]) == pos){
	  ret.push(i);
	}
      }
    }
  }
  return ret;
}

template<class _DecoratedTree >
HeirRoleVector HeirTree<_DecoratedTree >::getHRV(const iterator& it) {
  HeirRoleVector hrv((BGET_NT_LABEL(*it).size()));
  short role = 0;
  typename HeirTree<_DecoratedTree >::sibling_iterator it_sib;
  queue<iterator> heirIts;
  for(it_sib = it.begin(); it_sib != it.end(); ++it_sib) {
    ++role;
    if(isHeir(it_sib)) {
      queue<short> q = getHeirQueue(it_sib);
      while(!q.empty()) {
	hrv.setRole(q.front(),role);
	q.pop();
      }
    }
  }

  return hrv;
};

// returns the highest node in the tree that has the input node as its head
// the base case is that a node is its own head
template<class _DecoratedTree >
typename HeirTree<_DecoratedTree >::iterator 
HeirTree<_DecoratedTree >::maximalProjection(const iterator& it) {
  // if the current node is the heir of its parent ...
  if (isHeir(it)) {
    const iterator it_parent = dummyTree.parent(it);
    // .. and the parent exists
    if (dummyTree.is_valid(it_parent))
      // then recurse up
      return maximalProjection(it_parent);
    else
      return it;
  } else
    return it;
};

#endif
