/*
 * A Tree structure, with some heir involving operations, designed
 * to incrementally decorate another type of tree structures.
 */
#ifndef _HeirTree_H_
#define _HeirTree_H_

#include "NTLink.hpp"
#include "Vector.hpp"
#include "HeirRoleVector.hpp"
#include <iterator>
#include <queue>
#include "boost/tuple/tuple.hpp"
#define BGET_CONST_HEIR(it) static_cast<const unsigned int>((it))
//! An HeirTree is a static decorator pattern. It is used to add the heir
//! related functionalities to another tree.
template<class _DecoratedTree>
class HeirTree : public _DecoratedTree {
protected:
  typedef typename _DecoratedTree::tree_node tree_node;
public:
  typedef typename _DecoratedTree::iterator iterator;
  typedef _DecoratedTree _super;
  typedef typename _super::value_type value_type;

  //! Empty constructor.
  HeirTree() { }//__head_bk = this->getHead();}

  //! Constructor. Takes the reference of the decorated object as input.
  HeirTree(const _super& base);

  //! Destructor. Doesn't actually release the memory, but
  //! only restores the "head" pointer in tree to be the __head_bk.
  ~HeirTree();

  //! Sets a node (pointed to by the input iterator) to be the heir
  //! by setting the parents heir pointers to it.
  static void setHeir(const iterator & );
  /// overload for callers that happen to supply a parent, which we will ignore
  static void setHeir(const iterator &, const iterator & );

  //! Initialize the vector of Heirs pointed to by the iterator.
  static void initHeirVectors(const iterator & );

  //! Judges whether a node pointed to by the input iterator
  //! is an heir or not. Top node is an heir.
  static bool isHeir(const iterator & it);
  /// overload for callers that happen to supply a parent, which we will ignore
  static bool isHeir(const iterator& child, const iterator& parent);

  //! Judges whether a node pointed to by the input iterator
  //! is an heir or not. Top node is an heir.
  static bool isHeir(const iterator & it, const arity_t & d);
  /// overload for callers that happen to supply a parent, which we will ignore
  static bool isHeir(const iterator& child, const iterator& parent, const arity_t & d);

  //! Sets the heir of the parent to point to this position in
  //! the tree, in every dimension in queue.
  static void setHeirQueue(const iterator & , queue<short> & );

  //! Returns the HRV for the node
  static HeirRoleVector getHRV(const iterator &);

  //! Returns a queue which has all the dimensions for which
  //! this item is an heir in the parent.
  static queue<short> getHeirQueue(const iterator &);

  /// returns the highest node in the tree that has the input node as its head
  /// the base case is that a node is its own head
  static iterator maximalProjection(const iterator &);

  /// returns an iterator to the heir node on the given dimension
  static iterator findHeir(const iterator & it, arity_t dim);

  // Many of the methods of the tree base class should have been
  // static but aren't.  They take an iterator and a tree, but don't
  // actually use the tree.  Instead of passing around an extra tree
  // param, we use a static dummy tree where necessary.
  static HeirTree<_DecoratedTree> dummyTree;

  //! The backup of my own tree node, so that the head pointer
  //! can point to the head of the decorated object.
  // tree_node* __head_bk;
};

#ifndef _HeirTree_C_
#include "HeirTree.cpp"
#endif

#endif
