#include "HiddenDSVs.hpp"
#include "Dbug.hpp"

// If a grammar is being used as a supervising grammar for learning,
// then we want to update its consequents' hiddenDSVs structure.  If
// only one of the antecedents has hiddenDSVs, then those hiddenDSVs
// are copied to the consequent.  If both antecedents have hiddenDSVs,
// then some elements of their cross-product must be compatible with
// the input PAV.  Every such element is summed and copied to the
// hiddenDSV of the consequent.  If no element is compatible with the
// PAV, then the antecedents cannot compose, so we abort.  @return
// true unless the input hiddenDSVs are incompatible with the PAV
// @param outHDSVs is the output hiddenDSVs, passed by reference so
// that the pointer can be changed
bool hDSV::crossProdWithPA (const Vector<DSV>* pHDSVs1, const Vector<DSV>* pHDSVs2, 
			    const arity_t constrainedDim, const PrecArray& inPA, 
			    Vector<DSV>*& outHDSVs)  {
  if ( pHDSVs1 != NULL && pHDSVs2 == NULL)
    outHDSVs = new Vector<DSV>(*pHDSVs1); // copy every element
  else if ( pHDSVs1 == NULL && pHDSVs2 != NULL)
    outHDSVs = new Vector<DSV>(*pHDSVs2); // copy every element
  else if ( pHDSVs1 != NULL && pHDSVs2 != NULL) {
    outHDSVs = new Vector<DSV>;
    // try every element of the cross-product
    Vector<DSV>::const_iterator svci1, svci2;
    for(svci1 = pHDSVs1->begin(); svci1 != pHDSVs1->end(); ++svci1)
      for(svci2 = pHDSVs2->begin(); svci2 != pHDSVs2->end(); ++svci2) {
	if (Dbug::isDebugAll(5))
	  cerr << "hDSV::crossProd of " << *svci1 << " and " << *svci2 << endl;
	// punt if they overlap
	if ((*svci1).overlapped((*svci2)))
	  continue;
	// DSV relativization
	RawPAV rPAV(*svci1, *svci2);
	PrecArray candPA = rPAV.getPA(constrainedDim);
	if (candPA == inPA) {
	  // add em up and save em
	  DSV combinedDSV(*svci1 + *svci2);
	  outHDSVs->push_back(combinedDSV);
	}
      }
    if (outHDSVs->empty()) {
      // The hidden DSVs of the antecedents are incompatible with the
      // input PAV, so there are no possible consequents.
      delete outHDSVs;

      return false; /// antecedent pHDSVs exist but are incompatible
	}
  }

  return true;
}

/// If a grammar is being used as a supervising grammar for
/// learning, then we want to update its consequents' hiddenDSVs
/// structure.  If only one of the antecedents has hiddenDSVs, then
/// those hiddenDSVs are copied to the consequent.  If both
/// antecedents have hiddenDSVs, then we take their cross-product.
/// Every element of the cross-product is summed and copied to the
/// hiddenDSV of the consequent.  @return true unless the result is
/// the empty set.  @param outHDSVs is the output hiddenDSVs, passed
/// by reference so that the pointer can be changed
bool hDSV::crossProd (const Vector<DSV>* pHDSVs1, const Vector<DSV>* pHDSVs2, 
		      Vector<DSV>*& outHDSVs) {
  if ( pHDSVs1 != NULL && pHDSVs2 == NULL)
    outHDSVs = new Vector<DSV>(*pHDSVs1); // copy every element
  else if ( pHDSVs1 == NULL && pHDSVs2 != NULL)
    outHDSVs = new Vector<DSV>(*pHDSVs2); // copy every element
  else if ( pHDSVs1 != NULL && pHDSVs2 != NULL) {
    outHDSVs = new Vector<DSV>;
    // try every element of the cross-product
    Vector<DSV>::const_iterator svci1, svci2;
    for(svci1 = pHDSVs1->begin(); svci1 != pHDSVs1->end(); ++svci1)
      for(svci2 = pHDSVs2->begin(); svci2 != pHDSVs2->end(); ++svci2) {
	// add em up and save em
	if (Dbug::isDebugAll(5))
	  cerr << "hDSV::crossProd of " << *svci1 << " and " << *svci2 << endl;
	DSV combinedDSV(*svci1 + *svci2);
	// punt if they overlap
	if (not combinedDSV.empty()) {
	  outHDSVs->push_back(combinedDSV);
	}
      }
    if (outHDSVs->empty()) {
      // The hidden DSVs of the antecedents are incompatible with the
      // input PAV, so there are no possible consequents.
      delete outHDSVs;
	
      return false; // antecedent pHDSVs exist but are incompatible
    }
  }
    
  return true;
}
