#include "File.hpp"
#include "IndexedWMTG.hpp"
#include "TLink.hpp"
#include "COMPILER-DEPENDENT_HEADERS.hpp"
#include "HeadMTree.hpp"
#include "Hedge.hpp"
#include "OptionsPrinter.hpp"

//! Destructor. To clear the memory allocated.
IndexedWMTG::~IndexedWMTG() {
  // delete the grammar index.
  if(m_pGrmIndex) { delete m_pGrmIndex;}
}

//! constructor
IndexedWMTG::
IndexedWMTG(arity_t dm,
	    const string& configFile,
	    char** overrideArgV,
	    size_t overrideArgC,
	    BSP<BPO::options_description> optionsDescription)
  : _grammar(dm),
    _super(dm, configFile, overrideArgV, overrideArgC, 
	   addOptionDescriptionsTo(optionsDescription)),
    m_pGrmIndex(new GrammarIndex(m_arity)), m_negateCounts(false) {
  
  /*
   * Compile the config file.
   */
  BPO::variables_map vm;
  if(configFile != ""){
    iFile ifs(configFile.c_str());
    store(parse_config_file(ifs, *optionsDescription), vm);
  }
  notify(vm);

  if(vm.count("NegateCounts")) m_negateCounts = vm["NegateCounts"].as<bool>();

  if(debug(1)) {
    cerr << "IndexedWMTG options:" << endl;
    cerr << OptionsPrinter(*optionsDescription, vm) << endl;
  }
}

//! input
istream& IndexedWMTG::get(istream& is){
  _super::get(is);
  buildGrammarIndex();
  return is;
}

// assignment
IndexedWMTG& IndexedWMTG::operator = (const IndexedWMTG& other) {
  _super::operator=(other);

  // copy it
  m_pGrmIndex = new GrammarIndex(*other.m_pGrmIndex);

  return *this;
}

//! update the score of a production, NOT propagating to subfeatures
void IndexedWMTG::updateFeatureScore(Production* pProd , SCORE sc ) {
  if (m_pavVcb.contains(pProd->rhsPerm())) {
    m_pGrmIndex->insert(pProd);
    _super::updateFeatureScore(pProd, sc);
  }
}

//! Returns all Z's given X and Y such that Z->XY (or Z->YX).
//! Note that if there is a grammar index, the returned
//! result will be retrieved from the index. Otherwise,
//! the result will be obtained based on the nt sets.
void IndexedWMTG::getPossibleCons(const Item* item1,
				  const Item* item2,
				  const Pair<bool,PAV>& arrangement,
				  vector<Item* >&cons) {
  const PAV& pav = arrangement.second;
  role_t roleOfTrigger = arrangement.first ? 1 : 2;

  //keeps track if we have swapped items.
  bool swapped = false;

  //If PAV not in normal form, we switch the antecedents and
  //Heirvector to make it in normal form.
  if(!pav.isNormalFormForRole(roleOfTrigger)){
    swap(item1,item2);
    swapped = true;
  }

  if(debug(5)){
    cerr << "IndexedWMTG::getPossibleCons " << *((Hedge*)(item1)) << endl
	 << " and " <<endl << *((Hedge*)(item2)) << endl;
    cerr << "pav: " << arrangement.second << (swapped ? "(swapped)" : "(not swapped)") << endl;
  }

  // Gets information from item1 and item2
  const Hedge* thisHedge = static_cast<const Hedge*>(item1);
  const Hedge* otherHedge = static_cast<const Hedge*>(item2);

  const DSV* thisSpan = thisHedge->getDSV();
  const DSV* otherSpan = otherHedge->getDSV();

  const NTLink& link1 = thisHedge->getNTLink();
  const NTLink& link2 = otherHedge->getNTLink();

  DEBUG(5, "calling getParents on " << link1 << " and " << link2 << endl);
  Production* pProd;
  BSP<Production> spProd;
  set<NTLink> parents;
  m_pGrmIndex->getParents(link1, link2, pav, parents);
  
  set<NTLink>::iterator i;
  for(i = parents.begin(); i != parents.end(); i++){
    
    DSV combinedSpan = *thisSpan + *otherSpan;
    const NTLink& newLabel = *i;
    Item* h = new Hedge(newLabel, combinedSpan);
    DEBUG(5, "IndexedWMTG: final getParents is " << *((Hedge*)h) << endl);
    
    //! Create an MTG Production and set costs
    Vector<ProdRHSElem> rhsVec;
    rhsVec.push_back(link1);
    rhsVec.push_back(link2);
    pProd = new Production(newLabel, rhsVec, pav);
    DEBUG(5, "MtgProd " << pProd << endl);
    SCORE prodScore = this->getProdScore(pProd);
    //    spProd = BSP<Production>(pProd);
    //    h->setProd(spProd);
    h->setModelScore(prodScore);
    cons.push_back(h);
  }
}



//! Returns all Z's given X and Y such that Z->XY (or Z->YX).
//! Note that if there is a grammar index, the returned
//! result will be retrieved from the index. Otherwise,
//! the result will be obtained based on the nt sets.
void IndexedWMTG::getPossibleConsHeaded(const Item* item1,
					const Item* item2,
					const Pair<bool,PAV>& arrangement,
					vector<HRVIpair>&cons) {

  const PAV& pav = arrangement.second;
  role_t roleOfTrigger = arrangement.first ? 1 : 2;

  //keeps track if we have swapped items.
  bool swapped = false;

  //If PAV not in normal form, we switch the antecedents and
  //Heirvector to make it in normal form.
  if(!pav.isNormalFormForRole(roleOfTrigger)){
    swap(item1,item2);
    swapped = true;
  }

  if(debug(5)){
    cerr << "IndexedWMTG::getPossibleConsHeaded for " << *((Hedge*)(item1)) << endl
	 << " and " <<endl << *((Hedge*)(item2)) << endl;
    cerr << "pav: " << arrangement.second << (swapped ? "(swapped)" : "(not swapped)") << endl;
  }

  // Gets information from item1 and item2
  const Hedge* thisHedge = static_cast<const Hedge*>(item1);
  const Hedge* otherHedge = static_cast<const Hedge*>(item2);

  const DSV* thisSpan = thisHedge->getDSV();
  const DSV* otherSpan = otherHedge->getDSV();

  const NTLink& link1 = thisHedge->getNTLink();
  const NTLink& link2 = otherHedge->getNTLink();

  DEBUG(5, "calling getParents on " << link1 << " and " << link2 << endl);
  Production* pProd;
  BSP<Production> spProd;
  //headed case -- use HeirVectors
  if(m_isHeaded){
    set<Pair<HeirRoleVector,NTLink> > parents;
    m_pGrmIndex->getParents(link1, link2, pav, parents);
  
    set<Pair<HeirRoleVector,NTLink> >::iterator i;
    for(i = parents.begin(); i != parents.end(); i++){
      DEBUG(5, "IndexedWMTG: before hedge: hv is " << i->first << endl);
      DSV combinedSpan = *thisSpan + *otherSpan;
      const NTLink& newLabel = i->second;
      Item* h = new Hedge(newLabel, combinedSpan);
      DEBUG(5, "IndexedWMTG: final getParents is " << *((Hedge*)h));

      //! Create an MTG Production and set costs
      Vector<ProdRHSElem> rhsVec;
      rhsVec.push_back(link1);
      rhsVec.push_back(link2);
      pProd = new HeadProduction(newLabel, rhsVec, pav, i->first);

      HeirRoleVector hv = swapped ? i->first.invert() : i->first;

      DEBUG(5, "using prod " << *pProd << endl);
      SCORE prodScore = this->getProdScore(pProd);
      //      spProd = BSP<Production>(pProd);
      //      h->setProd(spProd);
      h->setModelScore(prodScore);
      delete pProd;
      cons.push_back(pair<HeirRoleVector,Item*>(hv,h));
    }
  }
}

//! returns possible consequents of a TLink antecedent
void IndexedWMTG::getPossibleConsForTLink(const TLink& tl, const DSV& dsv, vector<Item*>& ret) const {
  //arity of grammar
  arity_t dim= tl.arity();
  //non emtpy dim
  arity_t nonEmptyDim;
  //active dim found
  bool found = false;

  //search for active dimension
  for (unsigned int i = 0; i < dim; i++){
    if(tl[i] != NULLT){
      nonEmptyDim = i;
      found = true;
      break;
    }
  }

  //if no dimension is active.
  if(!found)
    return;

  PAV pav(dim);
  PrecArray pa1;
  PrecArray pa0;
  pa1.push_back(1);
  Vector<PrecArray> v(2);
  HeirRoleVector hv(2,0);

  if(nonEmptyDim == 0){
    v[0] = pa1;
    v[1] = pa0;
  } else if(nonEmptyDim == 1){
    v[0] = pa0;
    v[1] = pa1;
  } else
    THROW("All components are empty!");

  set<NTLink> possibleTagsForTLink;
  getPossibleTagsForTLink(tl, possibleTagsForTLink);
  
  if(possibleTagsForTLink.empty()){
    if(m_smoothScans){
      NTLink syncats;
      // keep the lexical heads, but assign dummy syncats
      Nonterminal* pUnk = new Nonterminal(UNKNT);
      Nonterminal* pNull = new Nonterminal(NULLNT);
      if (nonEmptyDim == 0)
	syncats = NTLink(pUnk, pNull);
      else
	syncats = NTLink(pNull, pUnk);
      NTLink ntl(syncats, tl);
      Item* h = new Hedge(ntl,dsv);
      h->setModelScore(m_smoothingCost);
      ret.push_back(h);
    }
  } else {
    set<NTLink>::const_iterator it;
    for(it = possibleTagsForTLink.begin(); it != possibleTagsForTLink.end(); it++){
      hv.setRole(nonEmptyDim,1);
      Vector<ProdRHSElem> rhsVec;
      rhsVec.push_back(tl);
      Production* p = new  HeadProduction(NTLink(*it,tl), rhsVec, PAV(v), hv);
      if (debug(5))
	cerr << "IWMTG::getPossConsForTL computing prodscore for " << *p;
      SCORE prodScore = this->getProdScore(p);
      delete p;
    
      Item* h = new Hedge(*it,dsv);
      h->setModelScore(prodScore);
      ret.push_back(h);
    }
  }

  return;
}
//! Returns all the possible pre-terminals lexicalized by
//! input terminal link.
void IndexedWMTG::getPossibleTagsForTLink(const TLink& tlink,
					  set<NTLink>& returnSet) const {
  assert(arity());
  m_pGrmIndex->getParents(tlink, returnSet);
}

//! Returns all the possible arrangements satisfying the requirement
//! by the fanout vector.
//! If the labeled DSV has no licensed arrangement, then
//! returns an undefined arrangement.
//! \param label : the lexicalized constituent for which we are
//! querying its licensed Arrangements (an arrangement
//! is a pair of PAV and a role; the role is the one
//! that we are concerned with).
//! \param span : the span covered the label.
//! \param fanoutV : the fanout vector of the item where we
//! get the label and the span. Note that the fanout
//! vector of the span argument is only a part of this
//! fanoutV in translation --- the spans on the dimensions
//! to be translated are not formed yet.
//! \param arrangements : the set of arrangements satisfying the fanout
//! vector constraint. The return
//! PAVs must have a role with the same fanoutV as the
//! specified fanoutV.
void IndexedWMTG::getArrangementFor(BSP<set<Pair<bool,PAV> > >& arrangements, const Item* trigger) const {
#ifdef USE_ARMT_INDEX
  m_pGrmIndex->getArrangementFor(arrangements, trigger);
#else
  _super::getArrangementFor(arrangements, trigger);
#endif
}

//! Build the index of the grammar.
void IndexedWMTG::buildGrammarIndex() {
  m_pGrmIndex->clear();
  /*
   * Iterate over the productions in the treebank
   * grammar, and increment them into the index.
   */
  const_iterator i;
  for(i = m_prods.begin(); i != m_prods.end(); ++i){
    Production* pProd = i->first;
    // index it only if its PAV is in the vocab of permissible PAVs
    const PAV& aPAV = pProd->rhsPerm();
    if (m_pavVcb.contains(aPAV)) {
      m_pGrmIndex->insert(i->first);
      SCORE newScore = i->second;
        // next, update the w2wTM
      W2WLink lk;
      if (pProd->getTLink(lk)) {
	ScoreSet<W2WLink>::iterator w2wi = m_w2wTM.find(lk);
	if (w2wi == m_w2wTM.end()) {
	  m_w2wTM.insert(lk, newScore);
	} else {
	  SCORE oldScore = w2wi->second;
	  if (m_costAggregator == SUM) {
	    w2wi->second = oldScore + newScore;
	  } else { // MAX or MAXSUM should use min [sic] (not max)
	    w2wi->second = min(oldScore, newScore);
	  }
	}
      }
    }
  }
}

void IndexedWMTG::estimate() {
    // the prob of a production's RHS is its joint prob divided by the prob of the LHS
  _super::estimate();
  if(m_negateCounts)
    m_prods *= -1;
};



void IndexedWMTG::clear() {
  _super::clear();
  m_pGrmIndex->clear();
}

BSP<BPO::options_description> 
IndexedWMTG::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  optionsDescription->add_options()
    ("NegateCounts", BPO::value<bool>()->default_value(false), "Set to true to negate counts read in.")
    ;
  return optionsDescription;
}
