#ifndef _IndexedWMTG_H_
#define _IndexedWMTG_H_

#include "MultiMap.hpp"
#include "StringVocabulary.hpp"
#include "Set.hpp"
#include "Histogram.hpp"
#include "GrammarIndex.hpp"
#include "NaiveWMTG.hpp"
#include <boost/program_options.hpp>
#include <boost/serialization/serialization.hpp>

class IndexedWMTG : public NaiveWMTG {
  typedef Grammar _grammar;
  typedef NaiveWMTG _super;
  BSP<BPO::options_description> 
  addOptionDescriptionsTo(BSP<BPO::options_description> priorOptions);

protected:
  typedef _super::iterator iterator;
  typedef _super::const_iterator const_iterator;

  // added for serialization, but kept protected from outside classes
  IndexedWMTG() {};

public:

  /**
   * @param configFile The config file of the grammar.
   */
  IndexedWMTG(arity_t dm, const string& configFile,
	      char** overrideArgV = NULL, size_t overrideArgC = 0,
	      BSP<BPO::options_description> priorOptions =
	      BSP<BPO::options_description>(new BPO::options_description())
	      );

  //! Destructor.
  virtual ~IndexedWMTG();

  //! Assignment.
  virtual IndexedWMTG& operator = (const IndexedWMTG&);

  //! input
  virtual istream& get(istream& is);

  virtual void getPossibleCons(const Item* item1, const Item* item2, const Pair<bool,PAV>& pav,
			       vector<Item*>& cons);
  
  virtual void getPossibleConsHeaded(const Item* item1, const Item* item2, const Pair<bool,PAV>& pav,
				     vector<HRVIpair >& cons);

  //!given a tlink and DSV, return a vector of Items that have
  // the item as a consequent.  These represent "terminal 
  // productions.
  virtual void getPossibleConsForTLink(const TLink& tl, const DSV& dsv, vector<Item*>& vItem) const;
  //! Returns all the possible arrangements satisfying the requirement
  //! by the fanout vector.
  //! If the labeled DSV has no licensed arrangement, then
  //! returns an undefined arrangement.
  //! \param label : the lexicalized constituent for which we are
  //! querying its licensed Arrangements (an arrangement
  //! is a pair of PAV and a role; the role is the one
  //! that we are concerned with).
  //! \param span : the span covered the label.
  //! \param fanoutV : the fanout vector of the item where we
  //! get the label and the span. Note that the fanout
  //! vector of the span argument is only a part of this
  //! fanoutV in translation --- the spans on the dimensions
  //! to be translated are not formed yet.
  //! \param arrgs : the set of arrangements satisfying the fanout
  //! vector constraint. The return
  //! PAVs must have a role with the same fanoutV as the
  //! specified fanoutV.
  virtual void getArrangementFor(BSP<set<Pair<bool,PAV> > >& arrgs, const Item* trigger) const;

  //! update the score of a production, NOT propagating to subfeatures
  virtual void updateFeatureScore(Production* , SCORE );

  // getter for w2w dict
  virtual BSP<W2WDictionary> getW2WDict() const {
    return m_pGrmIndex->getW2WDict();
  }

  /** Clears the content (only) of the containers. */
  virtual void clear();

  //! Build the index of the grammar. This method is called in the
  //! constructor, so it should not be virtual.
  void buildGrammarIndex();

  virtual void estimate();

protected: // methods

  //! Returns all the possible pre-terminals lexicalized by the input
  //! TLink. The input TLink is complete (D-dimensional). Note that,
  //! if a Grammar Index is in use then, this method will return the
  //! POS tags for the input TLink based on the content of the
  //! index. Otherwise, the returned result is the pre-terminal set.
  virtual void getPossibleTagsForTLink(const TLink&, set<NTLink>& returnSet) const;

  friend class boost::serialization::access;

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    ar & boost::serialization::base_object<_super>(*this);
    ar & m_pGrmIndex;
  };

private: // data

  //! The index to the grammar.
  GrammarIndex* m_pGrmIndex;
  
  //! If true, negate counts after estimation
  bool  m_negateCounts;

};

#endif
