#ifndef _Inference_H_
#define _Inference_H_

#include <vector>
#include "Item.hpp"
#include "GPDefs.hpp"
#include "UniqueID.hpp"
#include "PAV.hpp"
#include "StringVocabulary.hpp"
#include "DebugMacros.hpp"
#include "Exceptions.hpp"

class Item;
class Grammar;

using namespace std;

//! Base class of the Inference class family.
class Inference : public UniqueID {

public:

  Inference(const PAV&);
  Inference(const PAV&, const bool);
  
  //! Destructor.
  virtual ~Inference();

  //! output adaptors
  friend ostream& operator<<(ostream& o, const Inference& h);
  friend ostream& operator<<(ostream& o, Inference* pI);

  // returns a copy of this
  virtual Inference* clone() const = 0;

  //! how many antecedents are there?
  virtual unsigned int rank() const = 0;

  //! Gets the consequent
  virtual Item* getConseq() const = 0;

  //! Gets the antecedents. The first element in the returned
  //! vector is the heir.
  virtual vector<const Item*> getAnt() const = 0;

  virtual vector<const Item*>::iterator begin() = 0;

  virtual vector<const Item*>::iterator end()  = 0;

  virtual bool equalTo(const Inference& inf) const;

  //! Assigns a rank number to a type of inference.
  virtual int typeRank() const = 0;

  //! Dumps inference content.
  virtual void dump(ostream& out) const;
  //! Print out the inference, where terminals are strings instead of ids.
  virtual void dump(ostream& out, StringVocabulary* terminalSet) const;
  //! Print out the inference, where terminals and nonterminals are
  //! strings, instead of ids.
  virtual void dump(ostream& out, StringVocabulary* terminalSet, 
		    StringVocabulary* nonTermSet) const;
  virtual void dumpAnts(ostream& out) const = 0;

  //! Returns the PAV 
  const PAV getPAV() const;
  PAV getPAV();

  //! Returns the PAV and heir role 
  const Pair<bool,PAV> getArnmt() const;
  Pair<bool,PAV> getArnmt();

  //! Returns true if this inference involves head splitting.
  virtual bool headSplit() const { return false;}

  // find the smallest descendant containing this item's head on the given dimension
  // properly implemented only in head-bearing subclass(es)
  virtual Item* headItem(arity_t dim) const {
    return NULL;
  };

  //! compare two inferences
  virtual int cmp(const Inference* inf) const;

  //! Set the consequent item of this inference.
  virtual void setConseq(const Item* item){
    THROW("Not implemented: Inference::setConseq(Item*)");
  }

protected: // data

  // the PAV of the antecedents
  PAV m_PAV;

  // if the inferences is headed, then this variable indicates whether
  // role 1 of m_PAV is the head or not
  bool m_headIsFirst;
};

#endif

