#include "Item.hpp"
#include <iostream>
#include "Exceptions.hpp"
#include "MoreComparators.hpp"

size_t Item::sm_batchID = 0;

//! Constructor.
Item::Item() : m_batchID(sm_batchID),
	       m_pHiddenDSVs(NULL),
	       m_modelScore(0.0),
	       m_insideScore(0.0),
	       m_outsideScore(0.0),
	       m_auxScore(0.0)
{}

//! constructor from span and inference
Item::Item(DSV span, vector<size_t> sizeV) 
  : m_batchID(sm_batchID),
    m_span(span),
    m_pHiddenDSVs(NULL),
    m_sizes(sizeV),
    m_fanoutV(span.getFanoutV()),
    m_modelScore(0.0),
    m_insideScore(0.0),
    m_outsideScore(0.0),
    m_auxScore(0.0) {

  // if we didn't get a size vector, then compute it
  if (m_sizes.empty()) {
    size_t sz = m_span.size();
    m_sizes.resize(sz);
    for(arity_t d = 0; d < sz; ++d)
      m_sizes[d] = span.length(d);
  }
}

//copy constructor
Item::Item(const Item& other) 
  :  m_batchID(other.m_batchID),
     m_span(other.m_span),
     m_sizes(other.m_sizes),
     m_fanoutV(other.m_fanoutV),
     //     m_spProd(other.m_spProd),
     m_modelScore(other.m_modelScore),
     m_insideScore(other.m_insideScore),
     m_outsideScore(other.m_outsideScore),
     m_auxScore(other.m_auxScore) {
  if (other.m_spInf) // if it exists
     // the inference is cloned so that it can be modified via the new
     // item without affecting the inference of the other item
    m_spInf.reset(other.m_spInf->clone());
  if (other.m_pHiddenDSVs) {
    size_t vSize = other.m_pHiddenDSVs->size();
    m_pHiddenDSVs = new Vector<DSV>(vSize);
    for (unsigned short it = 0; it < vSize; ++it)
      m_pHiddenDSVs->operator[](it) = other.m_pHiddenDSVs->operator[](it);
  } else
    m_pHiddenDSVs = NULL;
}

Item::~Item() {
  delete m_pHiddenDSVs;
}

//! assignment
Item& Item::operator= (const Item& other) {
  m_modelScore = other.m_modelScore;
  m_insideScore = other.m_insideScore;
  m_outsideScore = other.m_outsideScore;
  m_auxScore = other.m_auxScore;
  m_spInf = other.m_spInf;
  //  setProd(other.getProd());
  m_fanoutV = other.m_fanoutV;

  if (other.m_pHiddenDSVs) {
    size_t vSize = other.m_pHiddenDSVs->size();
    m_pHiddenDSVs = new Vector<DSV>(vSize);
    for (unsigned short it = 0; it < vSize; ++it)
      m_pHiddenDSVs->operator[](it) = other.m_pHiddenDSVs->operator[](it);
  };

  return *this;
}

/// this non-virtualizable operator is defined in terms of a virtual
/// method. Note, however, that the result is non-symmetric.  E.g.,
/// it'll return true if the other item is an instance of a subclass
/// of this item, and equal only on the shared members.
bool Item:: operator == (const Item& other) const {
  return this->equal_to(other);
}

int Item :: cmp(const Item* other, bool sameType) const {
  int ret;
  // fanout vector.
  if((ret = MC::cmp(getFanoutV(), other->getFanoutV())) != 0) return ret;

  // span
  if((ret = m_span.cmp(*(other->getDSV()))) != 0) return ret;

  return 0;
}

//! Returns the inference that derives this item.
BSP<Inference> Item:: getInference() const { return m_spInf; }

//! Sets the inference that derives this item.
void Item:: setInference(Inference* pInf) {
  m_spInf.reset(pInf);
}
void Item:: setInference(BSP<Inference> spInf) {
  m_spInf = spInf; 
}

//! Gets the inside score (or Viterbi inside score).
SCORE Item:: getInsideScore() const {return m_insideScore;}

//! Sets the inside score (or Viterbi inside score).
void Item:: setInsideScore(const SCORE& score) {
  if (Dbug::isDebugAll(5))
    cerr << "setting inside score " << score << " for item " << m_uniqueID << endl;
  m_insideScore = score;
}

//! Gets the model score
SCORE Item:: getModelScore() const {return m_modelScore;}

//! Sets the model score
void Item:: setModelScore(const SCORE& score) {m_modelScore = score;}

//! Gets the outside score (or Viterbi outside score).
SCORE Item:: getOutsideScore() const {return m_outsideScore;}

//! Sets the outside score (or Viterbi outside score).
void Item:: setOutsideScore(const SCORE& score) {m_outsideScore = score;}

//! Gets the auxiliary score (or Viterbi outside score).
SCORE Item:: getAuxScore() const {return m_auxScore;}

//! Sets the auxiliary score (or Viterbi outside score).
void Item:: setAuxScore(const SCORE& score) {
  if (Dbug::isDebugAll(5))
    cerr << "setting aux score " << score << " for " << *this;
  m_auxScore = score;
}

//! Get the total score (inside + outside)
SCORE Item:: getScore() const{return m_insideScore + m_outsideScore;};

//! move the dspan on the given dim from the DSV to a hiddenDSV
void Item::hideDSV(arity_t firstOutDim) {

  DSV newDSV(m_span.size());
  for(arity_t ar = firstOutDim; ar < m_span.size(); ++ar) {
    newDSV[ar] = m_span[ar];
    m_span[ar].clear();
  };

  if (not m_pHiddenDSVs)
    m_pHiddenDSVs = new Vector<DSV>(1, newDSV);
};

//! generate new items by unhiding the hidden DSVs
void Item::restoreHDSVs(arity_t firstOutDim, vector<Item*>& pIVec) const {

  // Scan consequents have no hidden DSVs
  if ( ! m_pHiddenDSVs)
    return;

  Vector<DSV>::const_iterator vdci;
  for(vdci = m_pHiddenDSVs->begin(); vdci != m_pHiddenDSVs->end(); ++vdci) {
    Item* newI = clone();
    newI->m_span += *vdci;
    delete newI->m_pHiddenDSVs;
    newI->m_pHiddenDSVs = NULL;
    pIVec.push_back(newI);
  }
};

FanoutVect Item::getFanoutV() const {
  return m_fanoutV;
}

/*! Sets the Fanout Vector. The input fanout vector
 * should be consistent with that is induced from
 * the d-span.
 */
void Item::setFanoutV(const FanoutVect& fov) {
  m_fanoutV = fov;
}

// sets the item's inference and vice-versa
void Item::fuseInference(BSP<Inference> inference){
  //this is added for charts that do not create a copy of the item
  //on insert.
  //TODO - fuseInference should be added to charts that do create a copy
  if(m_spInf==inference) return;

  m_spInf = inference;
  inference->setConseq(this);
}

// find the smallest descendant containing this item's head
Item* Item::headItem(arity_t dim) const {
  return m_spInf->headItem(dim);
};

// return the position of the headWord of this item on the given dimension of its sentence
unsigned short Item::headPosn(arity_t dim) const {
  const Item* hItem = headItem(dim);
  const DSpan& theDSpan = hItem->getDSpan(dim);
  if (theDSpan.size() < 1)
    THROW("head item has no span on dim " << dim);
  // right boundary of first span on given dim
  return theDSpan[0].second;
};

// return the number of antecedents for this item
unsigned short Item::numAnts() const {
  return m_spInf->getAnt().size();
};

//! increment the static batchID.
void Item::incrementBatchID() {
  ++sm_batchID;
}

//! Output Adapter.
ostream& operator<<(ostream& o, Item* pI) {
  if (pI)
    pI->dump(o);
  else
    o << "<NULL ITEM>";
  
  return o;
};
ostream& operator<<(ostream& o, const Item& h) {
  h.dump(o);
  return o;
}

