#ifndef _Item_H_
#define _Item_H_

#include "GPDefs.hpp"
#include "DSV.hpp"
#include "StringVocabulary.hpp"
#include "UniqueID.hpp"
#include "Dbug.hpp"
#include "Exceptions.hpp"
#include "FanoutVect.hpp"
#include "Inference.hpp"

// forward declaration is necessary to break circular inclusion
class Inference;

//! Item
class Item : public UniqueID {

protected: // no public default
  //! Empty constructor
  Item();

public:
  //! constructor from DSV and sizeV
  Item(DSV aDSV, vector<size_t> sizeV = vector<size_t>());

  //! copy constructor
  Item(const Item& item);

  //! Destructor.
  virtual ~Item();

  //! Equality test.
  bool operator == (const Item&) const;

  //! Checks whether two items are the same in content and type.
  //! Compared with the operator== method, this method is virtual.
  virtual bool equal_to(const Item&) const = 0;

  virtual bool equal_to(const Item*) const = 0;

  // returns a copy of this
  virtual Item* clone() const = 0;

  //! Returns the inference that derives this item.
  virtual BSP<Inference> getInference() const;

  //! Sets the inference that derives this item.
  virtual void setInference(Inference* pInf);
  virtual void setInference(BSP<Inference> spInf);

  //! Set the inference in this item, and set the items conseq to this
  //! inference
  virtual void fuseInference(BSP<Inference> inference);

  //! Returns the language dimensionality.
  virtual arity_t arity() const = 0;

  //! Gets the inside score (or Viterbi inside score).
  virtual SCORE getInsideScore() const;

  //! Sets the inside score (or Viterbi inside score).
  virtual void setInsideScore(const SCORE& score);

  //! Gets the model score
  virtual SCORE getModelScore() const;

  //! Sets the model score
  virtual void setModelScore(const SCORE& score);

  //! Gets the outside score (or Viterbi outside score).
  virtual SCORE getOutsideScore() const;

  //! Sets the outside score (or Viterbi outside score).
  virtual void setOutsideScore(const SCORE& score);

  //! Gets the auxiliary score (or Viterbi outside score).
  virtual SCORE getAuxScore() const;

  //! Sets the auxiliary score (or Viterbi outside score).
  virtual void setAuxScore(const SCORE& score);

  //! Get the total score (inside + outside)
  virtual SCORE getScore() const;

  //! Returns the number of non-null dims.
  virtual size_t numActiveDims() const = 0;

  //! Returns the size of the span of the derived item.
  virtual size_t volume() const = 0;

  //! returns a 1d item made up from dimension d.
  virtual Item* get1DItem(arity_t d)const{THROW("Item::get1DItem not implemented");};

  //! For hashing.
  virtual size_t hash() const = 0;

  //! For the comparison within the same type.
  virtual int cmp(const Item* other, bool sameType= true) const;

  //! Assignment.
  Item& operator= (const Item& item);

  //! output adaptors
  friend ostream& operator<<(ostream& o, const Item& h);
  friend ostream& operator<<(ostream& o, Item* pI);

  //! Dumps the content, used for debugging.
  virtual void dump(ostream& o) const = 0;
  virtual void dump(ostream& o, StringVocabulary* terminalSet) const
  {THROW("Item::dump(out, terminalSet) not implemented");}
  virtual void dump(ostream& o, StringVocabulary* terminalSet,
		    StringVocabulary* nonTermSet) const
  {THROW("Item::dump(out, terminalSet, nonTermSet) not implemented");}

  //! Returns the span covered by the item.
  const DSV* getDSV() const {
    return &m_span;
  }
  DSV* getDSV() {
    return &m_span;
  }

  //! set DSV
  void setDSV(DSV& aDSV) {m_span = aDSV;}

  //! set hidden DSVs
  void setHiddenDSVs(Vector<DSV>* pDSVV) {m_pHiddenDSVs = pDSVV;}

  //! get hidden DSVs
  const Vector<DSV>* getHiddenDSVs() const {return m_pHiddenDSVs;}
  Vector<DSV>* getHiddenDSVs() {return m_pHiddenDSVs;}

  //! Returns the span on just one dimension
  const DSpan& getDSpan(arity_t d) const {
    return m_span[d];
  }
  DSpan& getDSpan(arity_t d) {
    return m_span[d];
  }

  /// copy the DSV to a hiddenDSV, and erase the dspans on the output
  /// components (from the firstOutDim upwards)
  void hideDSV(arity_t firstOutDim);

  //! generate new items by unhiding each of the hidden DSVs
  void restoreHDSVs(arity_t firstOutDim, vector<Item*>& pIVec) const;

  //! Returns the fanout vector of the item
  virtual FanoutVect getFanoutV() const;

  //! Sets the fanout of the item
  virtual void setFanoutV(const FanoutVect& fov);

  // test for emptiness on given dim
  bool empty(arity_t dim) const {
    return (m_sizes[dim] == 0);
  }
  // get sizes
  size_t getSize(arity_t dim) const {
    return m_sizes[dim];
  }
  vector<size_t> getSizes()const{
    return m_sizes;
  }
  // set sizes
  void setSizes(vector<size_t> v){
    m_sizes = v;
  }
  // set size of the given component
  void setSizes(arity_t dim, size_t sz){
    m_sizes[dim] = sz;
  }

  /// dummy method for when the Item is being treated as a node in a HedgeTree facade.
  /// (other types of trees have nodes that can lack labels)
  bool hasLabel(arity_t dim) const {
    return true;
  };

  // find the smallest descendant containing this item's head on the given dimension
  Item* headItem(arity_t dim) const;

  // return the position of the headWord of this item on the given dimension of its sentence
  unsigned short headPosn(arity_t dim) const;

  // return the number of antecedents for this item
  unsigned short numAnts() const;

  //! returns the ID of the batch this was created in.
  size_t getBatchID() const {return m_batchID; }

  //! returns the BatchID of the batch this was created in.
  static void incrementBatchID();

protected:
  //! unique identifier for each batch
  unsigned int m_batchID;

  //! span
  DSV m_span;

  /// During learning, the target string is unknown to the learning
  /// logic, but the supervising logic needs an efficient way to keep
  /// track of the target spans that an item might cover (e.g., based
  /// on its label, or tree constraints).  So we memoize these spans
  /// in this special data structure, where it is invisible to the
  /// learning logic.  The elements of this Vector are meant to be
  /// disjunctive, but they are all stored in one item, rather than in
  /// separate items, because otherwise a learning logic might get
  /// confused about why there are multiple items that it thinks are
  /// identical (or at least that's the idea -- it currently doesn't
  /// work when position tree iterators are involved).  The whole
  /// structure is here as just a pointer to minimize overhead for
  /// when it is not needed.
  Vector<DSV>* m_pHiddenDSVs;

  //! number of terminals in yield of each component.
  vector<size_t> m_sizes;

  /// D-dimensional fanout vector. Includes the part induced from
  /// the d-dimensional d-span (D <= d), and the translated
  /// part. 
  FanoutVect m_fanoutV;

  //! The inference that derives the item.
  BSP<Inference> m_spInf;

  //! The id for tracking; this one is static for the class, so it
  //keeps track of the largest uid assigned to any Item so far.
  static size_t sm_batchID;

  //! the inside score contributed by the model, i.e. not including the costs of antecedents
  SCORE m_modelScore;

  //! the inside score
  SCORE m_insideScore;

  //! the outside score
  SCORE m_outsideScore;

  //! the outside score
  SCORE m_auxScore;

};

namespace HashNS {
  //! Template specialization of STL's hash object.
  template<>
  struct hash< Item >
  { size_t operator() (const Item& item) const {return item.hash();} };

  template<>
  struct hash< Item* >
  { size_t operator() (const Item* pitem) const {return pitem->hash();} };

  template<>
  struct hash< BSP<Item> >
  { size_t operator() (const BSP<Item> pitem) const {return pitem->hash();} };
}

#endif
