#include "FeatureGenerator.hpp"

#include <string>
#include <stdio.h>
#include <stdlib.h>

#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

#include "NTLink.hpp"
#include "Hedge.hpp"
#include "HedgeTree.hpp"
#include "HeadProduction.hpp"
#include "Inference.hpp"

template<> inline
bool FeatureGenerator<HedgeTree>::prepareFor(Vector<const HedgeTree*>& treeVec) {
  return true;
}

template<> inline
bool FeatureGenerator<HedgeTree>::prepareFor(const HedgeTree* pPT, const arity_t focusDim) {
  return true;
};

/// get the lexical heads and SynCats of the item itself
template<> inline
void FeatureGenerator<HedgeTree>::
getFocusFeats(ScoreSet<FeatureType>& featCount, 
	      const HedgeTree::iterator& treeIt, const arity_t focusDim) const {
  string dimNumberString = string("dim") + char('0' + focusDim);

  // label
  // feature name will be like dimX.focusLabel==Z, where X is the dimension number, 
  // and Z is the label
  const NTLink& theNTL = treeIt->getNTLink();
  const Nonterminal* theNT = theNTL.getNT(focusDim);
  const SynCat theSC = theNT->getSynCat();
  const string* pLabel = GlobalVocab::s_synCatSet->getWord(theSC);
  string labelFeatName = dimNumberString + string(" focusLabel") + " = " + *pLabel;
  addOneFeature(featCount, labelFeatName);

  // terminal
  // feature name will be like dimX.focusWord==Z, where X is the dimension number, 
  // and Z is the word itself
  const LexNT* theLNT = dynamic_cast<const LexNT*>(theNT);
  if (! theLNT)
    THROW("NT is not a LexNT");
  const Terminal theTerm = theLNT->lex();
  const string* pWord = GlobalVocab::s_terminalSet->getWord(theTerm);
  string wordFeatName = dimNumberString + string(" focusWord") + " = " + *pWord;
  addOneFeature(featCount, wordFeatName);
};

// collect all of the dependents on the path from inIter down to stopIt on the focusDim
template<> inline
void FeatureGenerator<HedgeTree>::
getDependentFeats(ScoreSet<FeatureType>& featCount, 
		  const HedgeTree::iterator& inIter, 
		  const arity_t focusDim, const DSpan& theDSpan,
		  const HedgeTree::iterator& stopIt) const {

  const string* pWord;
  const string* pLabel;
  string featName;
  string dimNumberString = string("dim") + char('0' + focusDim);
  HedgeTree::sibling_iterator si = HedgeTree::dummyTree.begin(inIter);
  const HedgeTree::sibling_iterator& sentinel = HedgeTree::dummyTree.end(inIter); 
  for (; si != sentinel; ++si) {
    // proceed only if the node is active on focusDim
    const Item* pI = *si;
    if (pI->empty(focusDim)) // change in this specialization
      continue;

    if (! pI->hasLabel(focusDim)) // change in this specialization
      continue;

    const Hedge* pH = dynamic_cast<const Hedge*>(*si);
    const Nonterminal* pNT = pH->getNTLink().getNT(0); // change in this specialization
    SynCat depLabel = pNT->getSynCat();
    pLabel = GlobalVocab::s_synCatSet->getWord(depLabel);

    if (HedgeTree::dummyTree.isHeir(si, inIter, focusDim)) {
      // we get only the label cuz the headWord would be the same as
      // that of the parent
      featName = dimNumberString + string(" headChildLabel ") + *pLabel;
      addOneFeature(featCount, featName);
      // cast up to more generic iterator type
      HedgeTree::iterator lowerIter(pH);
      //dont recurse below the stop iterator
      if(*si != stopIt)
	getDependentFeats(featCount, lowerIter, focusDim, theDSpan); // recurse
    } else {
      const LexNT* pLNT = dynamic_cast<const LexNT*>(pNT);
      if (! pLNT)
	THROW("dunno how to get lexical heads from unlexicalized Hedges yet");
      Terminal dependentTerm = pLNT->lex();
      pWord = GlobalVocab::s_terminalSet->getWord(dependentTerm);
      featName = dimNumberString + string(" dependentWord ") + *pWord;
      addOneFeature(featCount, featName);
      // now the label
      featName = dimNumberString + string(" dependentLabel ") + *pLabel;
      addOneFeature(featCount, featName);
    }
  }
}

// structural features include:
// * label of the maximal projection
// * dependents at every X-bar level
// * label/headWord of the parent of the maximal projection (= dependency head)
// * label of the head sibling of the maximal projection
// * whether the head sibling is to the left or to the right
// * label/headWord of all other siblings of the maximal projection
template<> inline
void FeatureGenerator<HedgeTree>::
getStructuralFeatsAbove(ScoreSet<FeatureType>& featCount, const HedgeTree::iterator& treeIt, 
			const arity_t focusDim, const DSpan& theDSpan) const {
  const string* pWord;
  string featName;
  string dimNumberString = string("dim") + char('0' + focusDim);
  // first, find the maximal projection
  HedgeTree::iterator maxProj = HedgeTree::maximalProjection(treeIt);
  const Nonterminal* pMPNT = BGET_NT_LABEL(*maxProj).getNT(0);
  SynCat maxProjLabel = pMPNT->getSynCat();
  const string* pLabel = GlobalVocab::s_synCatSet->getWord(maxProjLabel);
  featName = dimNumberString + string(" maxProjLabel ") + *pLabel;
  addOneFeature(featCount, featName);

  // next, collect the dependents at every X-bar level
  getDependentFeats(featCount, maxProj, focusDim, theDSpan, treeIt);

  // next, the parent of the maximal projection, if it exists
  HedgeTree::iterator maxProjParent = HedgeTree::dummyTree.parent(maxProj);
  if (HedgeTree::dummyTree.is_valid(maxProjParent)) {
    const Nonterminal* pMPParNT  = BGET_NT_LABEL(*maxProjParent).getNT(0);
    SynCat maxProjParentLabel = pMPParNT->getSynCat();
    pLabel = GlobalVocab::s_synCatSet->getWord(maxProjParentLabel);
    featName = dimNumberString + string(" maxProjParentLabel ") + *pLabel;
    addOneFeature(featCount, featName);
    const LexNT* pMParLNT = dynamic_cast<const LexNT*>(pMPParNT);
    if (pMParLNT) {
      Terminal maxProjParentHeadPosn = pMParLNT->lex();
      // - 1 cuz the positions are 1-indexed but the sentences are 0-indexed
      Terminal maxProjParentHead = (*m_pSentTuple)[focusDim][ maxProjParentHeadPosn - 1];
      pWord = GlobalVocab::s_terminalSet->getWord(maxProjParentHead);
      featName = dimNumberString + string(" maxProjParentHead ") + *pWord;
      addOneFeature(featCount, featName);

      // in which direction is the head?
      const Nonterminal* pFocusNT  = BGET_NT_LABEL(*treeIt).getNT(0);
      const LexNT* pFocusLNT = dynamic_cast<const LexNT*>(pFocusNT);
      if (pFocusLNT) {
	Terminal focusHeadPosn = pFocusLNT->lex();
	if ( maxProjParentHeadPosn > focusHeadPosn)
	  featName = dimNumberString + string(" headFollows");
	else 
	  featName = dimNumberString + string(" headPrecedes");
	addOneFeature(featCount, featName);
      }
    }

    // next, all the siblings of the maximal projection
    // iterate over all the children of the parent of the max proj
    HedgeTree::sibling_iterator si;
    for (si = HedgeTree::dummyTree.begin(maxProjParent);
	 si != HedgeTree::dummyTree.end(maxProjParent); ++si) {
      if (*si == maxProj) // change in this specialization
	continue; // skip the max proj itself
      const Hedge* pH = dynamic_cast<const Hedge*>(*si);
      pMPNT = pH->getNTLink().getNT(0); // change in this specialization
      SynCat siblingLabel = pMPNT->getSynCat();
      pLabel = GlobalVocab::s_synCatSet->getWord(siblingLabel);
      if (HedgeTree::dummyTree.isHeir(si, maxProjParent, focusDim)) {
	// we get only the label cuz the headWord would be the same as
	// that of the parent of the maxProj
	featName = dimNumberString + string(" maxProjHeadSiblingLabel ") + *pLabel;
	addOneFeature(featCount, featName);
      } else {
	featName = dimNumberString + string(" maxProjSiblingLabel ") + *pLabel;
	addOneFeature(featCount, featName);
	const LexNT* pMPLNT = dynamic_cast<const LexNT*>(pMPNT);
	if (pMPLNT) {
	  Terminal maxProjSiblingHeadPosn = pMPLNT->lex();
	  // - 1 cuz the positions are 1-indexed but the sentences are 0-indexed
	  Terminal maxProjSiblingHead = (*m_pSentTuple)[focusDim][ maxProjSiblingHeadPosn - 1];
	  pWord = GlobalVocab::s_terminalSet->getWord(maxProjSiblingHead);
	  featName = dimNumberString + string(" maxProjSiblingHead ") + *pWord;
	  addOneFeature(featCount, featName);
	}
      }
    }
  }
};
