#ifndef _LearnableGrammar_H_
#define _LearnableGrammar_H_

#include <set>
#include "Dbug.hpp"
#include "GPDefs.hpp"
#include "MoreExceptions.hpp"

#include "Grammar.hpp"
#include "HeadMTree.hpp"

#include <boost/serialization/serialization.hpp>
#include <boost/serialization/is_abstract.hpp>

using namespace std;

class UpdateRule;

//! Abstract class to be implemented by all Learnable Grammars
class LearnableGrammar : virtual public Grammar {
  typedef Grammar _super;

protected:
  // for serialization
  LearnableGrammar() : m_pSentTuple(NULL) {}

public:

  //! constructor to build the grammar
  LearnableGrammar(arity_t dim) : _super(dim), m_pSentTuple(NULL) {}

  //! destructor
  virtual ~LearnableGrammar() {}

  //! clear the contents
  void clear();

  virtual bool prepareFor(SentenceTuple& tuple);
  virtual bool prepareFor(HeadMTree* pTree);
  virtual bool prepareFor(const Vector<HeadMTree*>& treeVec, vector<unsigned int>& startingPosns);

  //! Decompose a multitree in the way specified by the grammar and learn from it.
  virtual void increment(HeadMTree&, SCORE = UNITYSCORE) = 0;

  //! Decompose a multitree into inferences in the way specified by the grammar.
  virtual void tree2infs(Vector<Inference*>&, HeadMTree&);

  //! create an inference from the given tree node and add it to the set
  virtual void treeNode2infs(Vector<Inference*>&, const HeadMTree::iterator&, 
			     const DSV&, const DSV&, const DSV&) = 0;

  //! Increment features from another grammar
  virtual void increment(const Grammar* ) = 0;

  //! return the sum of the scores of all features
  virtual SCORE mass() = 0;

  //! perform any "wrap-up" operations needed besides normalization
  //! e.g., some grammars use this to rebuild indexes
  virtual void estimate() {} // do nothing by default

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    cerr << "SER LEARNABLE" << endl;
    ar & boost::serialization::base_object<_super>(*this);
  }

  /// call-back method to simulate virtual template parameter for UpdateRule::trees2gradient()
  /// computes the margin and return its total cost
  virtual SCORE trees2gradient(const vector<HeadMTree::iterator>& supervisorMargin, 
			       const vector<HeadMTree::iterator>& learnerMargin,
			       const vector<HeadMTree::iterator>& intersectionMargin,
			       Grammar* pCurrModel,
			       const UpdateRule* uRule) = 0;

protected: // data

  // cache sent tuple currently being processed
  SentenceTuple* m_pSentTuple;
  // cache tree currently being processed
  HeadMTree* m_pTree;

};

BOOST_IS_ABSTRACT(LearnableGrammar);

#endif
