#include "Weighted.hpp"
#include "LearnableWMTG.hpp"
#include "NaiveWMTG.hpp"
#include "HeadMTree.hpp"
#include "HeadProduction.hpp"
#include "HeadCompose.hpp"
#include "Scan.hpp"
#include "UpdateRule.hpp"
#include "GlobalVocab.hpp"
#include <boost/shared_ptr.hpp>

//! constructor
LearnableWMTG::
LearnableWMTG(arity_t dm, const string& configFile,
	      BSP<BPO::options_description> optionsDescription)
  : _grammar(dm),
    _learnable(dm),
    _super(dm, configFile, addOptionDescriptionsTo(optionsDescription)) {}

void LearnableWMTG::increment(HeadMTree& tr, SCORE count) {
  vector<BSP<Production> > prodVec;
  tree2features(tr, prodVec);
  vector<BSP<Production> >::iterator pvi;
  for(pvi = prodVec.begin(); pvi != prodVec.end(); ++pvi){
    Production* p = (*pvi).get()->clone();
    increment(p, count);
  }
}

//! clear the content
void LearnableWMTG::clear() {
  _super::clear();
  _learnable::clear();
};

//! Decompose a multitree into inferences in the way specified by the grammar.
void LearnableWMTG::treeNode2infs(Vector<Inference*>& ret, const HeadMTree::iterator& it, 
				  const DSV& conseqDSV, const DSV& ant1DSV, const DSV& ant2DSV) {
  Inference* pInf;
  short fanout = it.number_of_children();
  if (fanout == 0) {
    // do nothing
  } else if (fanout == 1) {
    Hedge* pConseq = new Hedge(it, conseqDSV);
    HeadMTree::sibling_iterator sit = HeadMTree::dummyTree.begin(it);
    const TLink& antTL = BGET_TERM_SPAN_LABEL(*sit).first;
    pInf = new Scan(pConseq, antTL);
    ret.push_back(pInf);
  } else if (it.number_of_children() == 2) {
    Hedge* pConseq = new Hedge(it, conseqDSV);
    PAV thePAV = BGET_PAV(*it);
    const Pair<bool, PAV> arment(true,thePAV);
    HeadMTree::sibling_iterator sit = HeadMTree::dummyTree.begin(it);
    Hedge* ant1 = new Hedge(sit, ant1DSV);
    ++sit;
    Hedge* ant2 = new Hedge(sit, ant2DSV);
    if (m_isHeaded) {
      HeirRoleVector hrv = HeadMTree::getHRV(it);
      pInf = new HeadCompose(ant1, ant2, pConseq, arment, hrv);
    } else
      pInf = new Compose(ant1, ant2, pConseq, arment);

    ret.push_back(pInf);
  } else 
    THROW("Dunno what to do with fanout of " << fanout);
};

/// call-back method to simulate virtual template parameter for UpdateRule::trees2gradient()
/// computes the margin and return its total cost
SCORE LearnableWMTG::trees2gradient(const vector<HeadMTree::iterator>& supervisorMargin, 
				    const vector<HeadMTree::iterator>& learnerMargin,
				    const vector<HeadMTree::iterator>& intersectionMargin, 
				    Grammar* pCurrModel,
				    const UpdateRule* uRule) {

  NaiveWMTG* pCurrGrammar = dynamic_cast<NaiveWMTG*>(pCurrModel);

  SPVector<Production> supervisorFeats, knownSupervisorFeats, unknownSupervisorFeats, learnerFeats, intersectionFeats;
  size_t supervisorFeatVecSize = 0;
  size_t learnerFeatVecSize = 0;
  size_t intersectionFeatVecSize = 0;
  vector<HeadMTree::iterator>::const_iterator tici;
  SPVector<Production>::const_iterator vpci;

  if (debug(3))
    cerr << "learner margin (LearnableWMTG):" << endl;
  // convert each node in the learner's margin to features
  for(tici = learnerMargin.begin(); tici != learnerMargin.end(); ++tici) {
    learnerFeatVecSize += treeNode2features(*tici, learnerFeats);
  };

  // convert each node in the supervisor's margin to features
  if (debug(3))
    cerr << "supervisor margin (LearnableWMTG):" << endl;
  for(tici = supervisorMargin.begin(); tici != supervisorMargin.end(); ++tici) {
    supervisorFeatVecSize += treeNode2features(*tici, supervisorFeats);
  };

  // convert each node in the intersection's margin to features
  if (debug(3))
    cerr << "intersection margin (LearnableWMTG):" << endl;
  for(tici = intersectionMargin.begin(); tici != intersectionMargin.end(); ++tici) {
    intersectionFeatVecSize += treeNode2features(*tici, intersectionFeats);
  };

  // find out how many features are previously unseen
  size_t knownFeats = 0;
  for(vpci = supervisorFeats.begin(); vpci != supervisorFeats.end(); ++vpci) {
    if (pCurrGrammar->knows(vpci->get())) {
      ++knownFeats;
      knownSupervisorFeats.push_back(*vpci);
    } else
      unknownSupervisorFeats.push_back(*vpci);
  }

  // get feature update sizes
  SCORE supervisorFeatGradient, learnerFeatGradient, unseenFeatGradient,intersectionFeatGradient;
  uRule->getFeatureGradients(knownFeats, learnerFeatVecSize,intersectionFeatVecSize,
			     supervisorFeatGradient, learnerFeatGradient, intersectionFeatGradient, 
			     unseenFeatGradient);

  SCORE posMass = 0.0;
  SCORE negMass = 0.0;
  // update previously unseen supervisor features
  for(vpci = unknownSupervisorFeats.begin(); vpci != unknownSupervisorFeats.end(); ++vpci) {
    updateFeatureScore(vpci->get()->clone(), unseenFeatGradient);
    negMass += unseenFeatGradient;
  }

  // update each known supervisor feature
  for(vpci = knownSupervisorFeats.begin(); vpci != knownSupervisorFeats.end(); ++vpci) {
    updateFeatureScore(vpci->get()->clone(), supervisorFeatGradient);
    negMass += supervisorFeatGradient;
  }

  // update each known intersection feature
  for(vpci = intersectionFeats.begin(); vpci != intersectionFeats.end(); ++vpci) {
    updateFeatureScore(vpci->get()->clone(), intersectionFeatGradient);
    negMass += intersectionFeatGradient;
  }

  // update each learner feature
  for(vpci = learnerFeats.begin(); vpci != learnerFeats.end(); ++vpci) {
    updateFeatureScore(vpci->get()->clone(), learnerFeatGradient);
    posMass += learnerFeatGradient;
  }

  // each unseen feature counts a whole 1, since we definitely want to
  // keep training if we've just seen stuff that we've never seen
  // before.
  SCORE ret = posMass - negMass + unknownSupervisorFeats.size();
  if (debug(2)) {
    cerr << unknownSupervisorFeats.size() << " unknownSuperFeats; ";
    cerr << knownSupervisorFeats.size() << " knownSuperFeats; ";
    cerr << learnerFeats.size() << " learnerFeats" << endl;
    //     cerr << "LearnableWMTG update mass = " << posMass - negMass
    // 	 << " ConvergenceValue = " << ret << endl;
  }

  return ret; 
};

BSP<BPO::options_description> 
LearnableWMTG::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  //optionsDescription->add_options()
  // ("IndexedWMTGTestOption", BPO::value<string>()->default_value("default value"), "test option")
  // ;
  return optionsDescription;
}
