#ifndef _LearnableWMTG_H_
#define _LearnableWMTG_H_

#include <iterator>
#include "MultiMap.hpp"
#include "Histogram.hpp"
#include "Set.hpp"
#include "WMTG.hpp"
#include "HeadMTree.hpp"
#include "StringVocabulary.hpp"
#include "SPVector.hpp"

#include "LearnableGrammar.hpp"
#include "Item.hpp"
#include "UpdateRule.hpp"

#include <boost/serialization/serialization.hpp>
#include <boost/serialization/is_abstract.hpp>
#include <boost/shared_ptr.hpp>

class HeadMTree;
class NaiveWMTG;

// abstract class to be implemented by all Learnable WMTGs, adds MTG
// specific increment methods.
class LearnableWMTG : public LearnableGrammar, public WMTG {
  typedef WMTG _super;
  typedef Grammar _grammar;
  typedef LearnableGrammar _learnable;
  BSP<BPO::options_description> 
  addOptionDescriptionsTo(BSP<BPO::options_description> priorOptions);

protected:
  // for serialization
  LearnableWMTG() {};

public:

  //! constructors -- even though this is an abstract class,
  //! constructors are necessary for initializing the data members
  LearnableWMTG(arity_t dim, bool inGCNF)
    : _grammar(dim), _learnable(dim), _super(dim, inGCNF)
  {}

  //! constructor
  LearnableWMTG(arity_t dm, const string& configFile,
		BSP<BPO::options_description> optionsDescription);

  //! Destructor.
  virtual ~LearnableWMTG() {}

  //! clear the contents
  void clear();

  // test whether a feature exists in the model
  virtual bool knows(Production* pProd) const = 0;

  //! Decompose a multitree into productions and increment them in.
  virtual void increment(HeadMTree&, SCORE = UNITYSCORE);

  //! Increment a production, propagating to subfeatures
  virtual void increment(Production* , SCORE ) = 0;

  //! create an inference from the given tree node and add it to the set
  virtual void treeNode2infs(Vector<Inference*>&, const HeadMTree::iterator&, const DSV& conseqDSV, 
			     const DSV& antDSV1 = DSV(), const DSV& antDSV2 = DSV());

  //! update the score of a production, NOT propagating to subfeatures
  virtual void updateFeatureScore(Production* , SCORE ) = 0;

  /// call-back method to simulate virtual template parameter for UpdateRule::trees2gradient()
  /// computes the margin and return its total cost
  virtual SCORE trees2gradient(const vector<HeadMTree::iterator>& supervisorMargin, 
			       const vector<HeadMTree::iterator>& learnerMargin,
			       const vector<HeadMTree::iterator>& intersectionMargin,
			       Grammar* pCurrModel,
			       const UpdateRule* uRule);

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    ar & boost::serialization::base_object<_learnable>(*this);
    ar & boost::serialization::base_object<_super>(*this);
  }
};

BOOST_IS_ABSTRACT(LearnableWMTG);

#endif
