#include "LexLinkBasedEstimator.hpp"
#include "Item.hpp"
#include "Hedge.hpp"
#include "HeadProduction.hpp"
#include "Exceptions.hpp"

//! Pre-computes the data needed in the computation of this heuristic.
void LexLinkBasedEstimator::precompute(NaiveWMTG *grammar) {
  m_minWordCosts.clear();
  m_minWordCosts.resize(grammar->arity());

  for(NaiveWMTG::iterator itr = grammar->begin(); itr!=grammar->end(); itr++) {
    HeadProduction *prd = dynamic_cast<HeadProduction*>(itr->first);
	if(prd == NULL) {
	  THROW("LexLinkBasedEstimator only works with HeadProductions");
	}
	// We are only interested in nonterminal productions as almost 
	// every terminal production will have a cost of 0
	// since we're using GCNF terminals have one element on 
	// the rhs, while nonterms have two
	Vector<ProdRHSElem> rhs = prd->rhsLinks();
	if(rhs.size() == 2) {
	  SCORE cost = itr->second;

	  // find the minimum production cost per rhs link
	  Vector<ProdRHSElem>::iterator itr = rhs.begin(); 
	  for(; itr!=rhs.end(); itr++) {
	    TLink *tmp = boost::get<TLink>(&(*itr));
		TLink link;
		if(!tmp) {
		  link = boost::get<NTLink>((*itr)).getTLink();
		} else {
		  link = *tmp;
		}

	    LinkCostMap::iterator res = m_minNontermCost.find(link);
		if(res == m_minNontermCost.end()) {
		  m_minNontermCost.insert(LinkCostMap::value_type(link, cost));
		  insertMinWord(link, cost);
		} else {
		  SCORE oldMin = res->second;
		  if(cost < oldMin) {
		    m_minNontermCost.insert(LinkCostMap::value_type(link, cost));
		    insertMinWord(link, cost);
		  }
		}
	  }
	}
  }

  cout << "precompute done" << endl;
}

void LexLinkBasedEstimator::insertMinWord(const TLink &link, SCORE cost) {
  for(arity_t d=0; d<link.arity(); d++) {
    Terminal t = link.at(d);
	WordCostMap minCost = m_minWordCosts.at(d);
	WordCostMap::iterator i = minCost.find(t);
	if(i == minCost.end()) {
	  minCost.insert(WordCostMap::value_type(t, cost));
	} else {
	  SCORE tmp = i->second;
	  if(cost < tmp)
	    minCost.insert(WordCostMap::value_type(i->first, cost));
	}
  }
}

void LexLinkBasedEstimator::initialize(SentenceTuple &sent) {
  sentence = sent;
  cout << "initialize done" << endl;;
}

//! Returns h(e,d).
//! \param e : the item for which we are computing its heuristic.
//! \param d : language dimension.
SCORE LexLinkBasedEstimator::h(Item* i, const arity_t& d) {
  Hedge* e;
  if(!(e = dynamic_cast<Hedge*>(i))){
    THROW("Error: unknown Item type in LexLinkBasedEstimator" << endl);
  }

  // the sentence length on the dth dimension.
  int len = sentence.lengths()[d];

  DSpan span1d = (*(e->getDSV()))[d];

  SCORE ret = 0.0;

  // only use the words that aren't covered by the item
  for(short i = 0; i < len; ++i){
    Location loc(1, i);
    DSV tempSpan(loc);
    if(!tempSpan.overlapped(span1d)){
	  Terminal t = sentence[d][i];
	  WordCostMap::iterator res;
	  try {
	    res = m_minWordCosts.at(d).find(t);
	  } catch(...) {
	    THROW("can't get min word costs for dim " << d << ".  Perhaps initialize wasn't called?" << endl);
	  }
	  if(res != m_minWordCosts[d].end()) {
        ret += res->second;
	  }
    }
  }

  cout << "ret " << ret << endl;
  return ret;
}

//! Returns the heuristic merit for an item.
//! \param sntT : the input sentence tuple.
//! \param item : the item for which we are computing its merit.
SCORE LexLinkBasedEstimator::operator()(SentenceTuple &sent, const Item* i) {
  const Hedge* item;
  if(!(item = dynamic_cast<const Hedge*>(i))) {
    THROW("unknown Item type in LexLinkBasedEstimator" << endl);
  }

  NTLink lab = item->getNTLink();
  arity_t d;
  // return 0 if no lex head in one or more dimensions
  for(d = 0; d < lab.arity(); ++d) {
    if((static_cast<const LexNT*>((lab).getNT(d)))->lex() == UNDEF_T) {
      return 0.0;
    }
  }

  // for monolingual case we don't use this heuristic.
  if(sent.arity() <= 1 && sent.arity() == item->arity()) {
    cerr<<"In monolingual parsing, the Lexical Link based heuristic"
        <<" is disabled."<<endl;
    return 0.0;
  }

  SCORE ret = POSINF;
  for(arity_t d = 0; d < sent.arity(); ++d){
    SCORE temp = h(const_cast<Hedge*>(item), d);
    ret = (temp < ret) ? temp : ret;
  }
  return ret;
}

//! Resets the class. After this operation, the temporary
//! tables are cleared, but the tables of maximum production
//! merits are kept unchanged.
void LexLinkBasedEstimator::reset() {
}

//! Get.
istream& LexLinkBasedEstimator::get(istream&in) {
  string line;

  getline(in, line, '\n');
  istringstream ist(line);
  ist >> m_default;

  int isD = 0;
  while(getline(in, line, '\n')){
    cerr<<"++++++++"<<line<<"*******"<<endl;
    if(line == "gN"){
      isD = 1;
    } else {
      cerr<<"********"<<line<<"*******"<<endl;
      vector<string> strs = split(line, "\t");
      assert(strs.size() == 2);

      istringstream ist1(strs[0]);
      TLink tlink;
      ist1 >> tlink;

      istringstream ist2(strs[1]);
      SCORE merit;
      ist2 >> merit;

      if(isD == 1){
        m_minNontermCost[tlink] = merit;
      }
    }
  }

  return in;
}

//! Put.
//! The default value m_default is also output because
//! this default value in this class is not obtained from
//! the grammar, instead, it is obtained from
//! the smoothing strategy in other HeadWMTGs.
ostream& LexLinkBasedEstimator::put(ostream&out) const {
  // the default value:
  out<<m_default<<endl;

  /*
  LinkCostMap::const_iterator it_hash;

  // output table \hat{g}_{D}(l).
  out<<"gN"<<endl;

  for(it_hash = m_minNontermCost.begin(); 
      it_hash != m_minNontermCost.end(); ++it_hash){
    out<<it_hash->first<<"\t"<<it_hash->second<<endl;
  }
  */

  return out;
}

