/*
 * Lexical link based OCE function.
 */
#ifndef _LexLinkBasedEstimator_H_
#define _LexLinkBasedEstimator_H_

#include <vector>
#include "OutsideCostEstimator.hpp"
#include "HeadWMTG.hpp"

using namespace std;

//! The lexical link based OCE function. 
/*! It returns the heuristic merit of an item based on the 
 * lexical links belonging to the yield of the outer trees 
 * of the item.
 */
class LexLinkBasedEstimator : public OutsideCostEstimator {
  typedef OutsideCostEstimator _super;
public:
  //! Constructor. By default, the lexical head is a contributor
  //! to the heuristic of an item.
  LexLinkBasedEstimator(const string& configFile) 
    : _super(configFile) {
    m_default = 0;
  }

  //! Destructor.
  ~LexLinkBasedEstimator() {}

  //! Pre-computes the data needed in the computation of this
  //! heuristic. Note that, in alignment, this method
  //! should be called AFTER the default small value is computed
  //! in ApproxHeadWMTG. By default, the lexical head is a contributor
  //! to the heuristic of an item.
  //! \param grammar : this model decides how a production is
  //!                  decomposed, and how the decomposed events
  //!                  are scored. Note that this model should
  //!                  be estimated from the following treebank
  //!                  grammar BEFORE this method is called.
  void precompute(NaiveWMTG *grammar);

  //! Initializes the estimator based on the sentence tuple. This sentence
  //! tuple is to be parsed by the main multiparser.
  void initialize(SentenceTuple &sent);

  //! Returns the heuristic merit for an item.
  //! \param sntT : the input sentence tuple.
  //! \param item : the item for which we are computing its merit.
  SCORE operator()(SentenceTuple &sent, const Item* item);

  //! Resets the class. After this operation, the temporary
  //! tables are cleared, but the tables of maximum production
  //! merits are kept unchanged.
  void reset();

  //! If the input param is true, then the lexical head of an item
  //! will be considered as a contributor to heuristic value;
  //! otherwise, the lexical head is considered as a contributor
  //! to inside value.
  //void setLexicalHeadAsContributorToHeuristic(bool yes = true);

  //! Get.
  istream& get(istream&);

  //! Put.
  ostream& put(ostream&) const;

private:
  //! Returns h(e,d).
  //! \param e : the item for which we are computing its heuristic.
  //! \param d : language dimension.
  SCORE h(Item* e, const arity_t& d);

  // Permanent data
  /////////////////////////////////////////////////////////////////

  typedef hash_map<TLink, SCORE> LinkCostMap;
  //typedef pair<arity_t, Terminal> DimWord;
  //typedef hash_map<DimWord, SCORE> MinWordCostMap;
  typedef hash_map<Terminal, SCORE> WordCostMap;

  void insertMinWord(const TLink &link, SCORE cost);

  //! the minimum of the cost of any nonterminal production
  //! whose RHS contains a dependent lexicalized by a terminal link.
  LinkCostMap m_minNontermCost;

  //! the minimum cost of a word in a given dimension
  vector<WordCostMap> m_minWordCosts;
  //MinWordCostMap m_minWordCost;

  //! Default returned value for an query doesn't appear in both
  //! m_glN and m_glT.
  SCORE m_default;

  // Temporary data
  /////////////////////////////////////////////////////////////////

  SentenceTuple sentence;
};

#endif
