#ifndef LexNT_H
#define LexNT_H

#include "IOUtil.hpp"
#include "GPDefs.hpp"
#include "Nonterminal.hpp"
#include "Exceptions.hpp"
#include <cassert>
#include <boost/serialization/serialization.hpp>

using namespace std;

//! A LexNT is a lexicalized nonterminal, which has a terminal and SynCat.
class LexNT : public Nonterminal {

public:

  //! Needed for hashing
  static const size_t sm_bigValForHash = 1109;

  //! Default constructor.
  LexNT();

  //! Constructor from the NT and its head.
  LexNT(const SynCat& a, const Terminal& b);

  //! Destructor.
  virtual ~LexNT() {}

  //! returns reference to terminal
  Terminal& lex();

  //! returns copy of terminal
  Terminal lex() const;

  //! return true if this is an inactive LexNT
  bool isInactive() const;

  //! Convert a LexNT to a VocabularyIndex via bit manipulation.
  VocabularyIndex toID() const;

  //! To compare with others like me, use the implementer class.
  inline bool operator==(const LexNT& other) const{
    return nt() == other.nt() && lex() == other.lex();
  }

  //! compare to another LexNT (as ref)
  inline bool operator==(const Nonterminal& other) const{
    const LexNT& castOther = static_cast<const LexNT&>(other);
    return (nt() == other.nt() && lex() == castOther.lex());
  }

  //! compare to another LexNT (as pointer)
  inline bool operator==(const Nonterminal* other) const{
    return operator==(*other);
  }

  //! To compare with others like me, use the implementer class.
  bool operator!=(const LexNT& other) const;

  virtual size_t hash() const;

  ////////////////////////////Operators from Pair//////////////////

  LexNT& operator= (const LexNT& other);
  bool operator<(const LexNT& other) const;

  inline bool operator<(const Nonterminal* other) const {
    const LexNT* otherLexNT = dynamic_cast<const LexNT*>(other);
    assert(otherLexNT != 0);
    return make_pair(nt(), lex()) < make_pair(otherLexNT->nt(), otherLexNT->lex());
  }

  bool operator > (const LexNT& other) const;

  //! Output.
  ostream& put(ostream& out) const;

  //! output with vocab deciphering where input boolean says if it's a
  // position tree or not.
  ostream& put(ostream& out, StringVocabulary* ntset,
	       StringVocabulary* terminalset, bool isPositionTree) const;

  //! output with vocab deciphering
  ostream& put(ostream& out, StringVocabulary* ntset,
	       StringVocabulary* terminalset) const{
    return put(out,ntset,terminalset,false);
  }

  //! serialize.
  ostream& serialize(ostream& out) const;

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    ar & boost::serialization::base_object<Nonterminal>(*this);
    ar & m_terminal;
  };

  //! Input.
  istream& get(istream& in);

  istream& get(istream& in, StringVocabulary& ntVocab, StringVocabulary& termVocab,bool isPositionTree=false);

  //! Output.
  friend ostream& operator << (ostream& out, const LexNT& obj) {
    return obj.put(out);
  }

  //! Input.
  friend istream& operator >> (istream& in, LexNT& obj) {
    return obj.get(in);
  }

  //! Returns a pointer to a copy of this object.
  inline virtual LexNT* clone() const{
    return new LexNT(*this);
  };

  //! Exports the object content using a vector.
  //! The order is SynCat, and then Terminal.
  vector<unsigned int> sequentialize() const;

  //! compare to another LexNT.
  virtual int cmp(const Nonterminal& other) const;

  //! compare to another LexNT via pointer
  virtual int cmp(const Nonterminal* other) const;

  //! return a delexicalized version of this NTLink
  virtual Nonterminal* delex() const;

  //! return a vector of unsigned ints representing this
  //! nonterminal.
  virtual vector<unsigned short> vectorDecompose()const;

  virtual void setHeirMark(){
    m_terminal = HeirMark;
  }

  virtual void setNull();

  virtual bool isNull() const;

  virtual bool isDummy() const;

protected:
  //! member terminal
  Terminal m_terminal;

};

namespace HashNS {
  template<> struct hash<LexNT> {
    size_t operator()(const LexNT& obj) const {
      return obj.hash();
    }
  };

  template<> struct hash<LexNT*> {
    size_t operator()(const LexNT* obj) const { return obj->hash();
    cerr << "HASH LEXNT:" << endl;
    cerr << *obj << endl;
    cerr << obj->hash() << endl;
    }
  };
}

#endif
