#include "File.hpp"
#include "Logic.hpp"
#include "Chart.hpp"
#include "ChartBuilder.hpp"
#include "MaxCoverComparator.hpp"
#include "Grammar.hpp"
#include "Inference.hpp"
#include "TLink.hpp"
#include "SentenceTuple.hpp"
#include "PrunerBuilder.hpp"
#include "Exceptions.hpp"
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"
#include "Path.hpp"

size_t Logic::sm_logicCount = 0;

// Constructor(S).
Logic::Logic(const string& configFile,
	     BSP<BPO::options_description> optionsDescription)
  : m_uniqueID(++sm_logicCount), m_numMatches(0), m_numUniqueMatches(0),
    m_arity(0), m_inputArity(0), 
    m_pGrammar(NULL), m_pPruner(NULL), m_pGoal(NULL) {

  arity_t dim = 0;
  arity_t inputArity = 0;
  string optPrunerBuilderConfigFile;
  string optChartConfigFile;
  unsigned int debug = 0;

  addOptionDescriptionsTo(optionsDescription);
 
  BPO::variables_map vm;
   iFile ifs(configFile.c_str());
   store(parse_config_file(ifs, *optionsDescription), vm);
   notify(vm);
 
  if(vm.count("D")) dim = vm["D"].as<unsigned int>();
   if (vm.count("ChartConfigFile")){
    optChartConfigFile = vm["ChartConfigFile"].as<string>();
  } else {
    THROW ("Error:  A configuration file must be specified for the chart.");
  }     
   if(vm.count("InputArity")) inputArity = vm["InputArity"].as<unsigned int>();
  if(vm.count("PrunerBuilderConfigFile")) 
    optPrunerBuilderConfigFile = vm["PrunerBuilderConfigFile"].as<string>();
  if(vm.count("debug")) debug = vm["debug"].as<unsigned int>();

  if(dim == 0)
    THROW("Error: The Logic dimensionality is zero!"<<endl);

  if (debug > 0)
    debugme(debug);

  if(this->debug(1)) {
    cerr << "Option values of \"Logic\"" << endl << endl;
    cerr << OptionsPrinter(*optionsDescription, vm) << endl;
  }

  m_arity = dim;
  // if input arity not specified in config file, then default to logic's arity
  if (inputArity == 0)
    m_inputArity = m_arity;
  else
    m_inputArity = inputArity;
   
  // build the chart
  string chartConfigFile = Path::maybe_compose(configFile, optChartConfigFile);
  m_spMainChart.reset(ChartBuilder::Instance()->build(chartConfigFile));
 
  // builds the pruner
  string prunerBuilderConfigFile = Path::maybe_compose(configFile, optPrunerBuilderConfigFile);
  m_pPruner = PrunerBuilder::Instance()->build(prunerBuilderConfigFile);

}

//! Destructor.
Logic::~Logic() {
  //  if(m_spMainChart) {delete m_spMainChart; }
  if(m_pPruner){ delete m_pPruner; }
  if(m_pGoal) { delete m_pGoal;}

  //don't delete grammar, since Logic did not create it.

  //for(unsigned int i = 0; i <= m_savedCharts.size(); i++)
  // delete m_savedCharts[i];
}

//! Initialize the Logic using a sentence tuple.
void Logic::initialize(SentenceTuple& snt) {
  m_numMatches = 0;
  m_numUniqueMatches = 0;
  clear();
  
  if (! m_pGrammar) {
    THROW("Logic::initialize: null grammar!" << endl);
  };

//   arity_t inArity = snt.size();
//   if (inArity < m_inputArity) {
//     THROW("Input arity less than expected by Logic.");
//   } else if (inArity == m_inputArity)
     m_sent = snt;
//   else { // inArity > m_inputArity
//     m_sent.clear();
//     for(arity_t dim = 0; dim < m_inputArity; ++dim)
//       m_sent.push_back(snt[dim]);
//   }      

  // useful if the grammar does caching, or if the grammar is based on moving streams
  m_pGrammar->prepareFor(snt);
}

//! get the grammar
Grammar* Logic::getGrammar() {
  return m_pGrammar;
}

//! set the grammar
void Logic::setGrammar(Grammar* gr) {
  m_pGrammar = gr;
}

//! set the chart
void Logic::setChart(BSP<Chart> ch) {
  m_spMainChart = ch;
}

TerminationTest* Logic::getGoal() {
  return m_pGoal;
}

void Logic::setGoal(TerminationTest& goal) {
  m_pGoal = &goal;
}

//! Get the score of an Item.
inline pair<SCORE,SCORE> Logic::getIOScores(const Item* item) const {
  return m_spMainChart->getIOScores(item);
}

//! Check whether a given Item is already on the chart.
inline bool Logic::derived(const Item* item) const {
  return m_spMainChart->derived(item);
}

//! Export the set of items that maximally cover the
//! entire span, and the spans of items are not overlapped.
void Logic::exportMaxCover(Vector<const Item*>& items) {
  /*
   * Extract the set of spans from the chart, and sort them in ascending
   * order of the span size.
   */
  Vector<const Item*> itemVec;

  m_spMainChart->exportAll(itemVec);

  // sort by span size.
  MaxCoverComparator mcc(m_pGrammar, m_inputArity);
  sort(itemVec.begin(), itemVec.end(), mcc);

  /*
   * Greedily find the maximum constituents, and extract the subtree
   * rooted at the constituents.
   */
  Vector<const Item*>::iterator it;
  set<DSV> extractedSpans;
  set<DSV>::iterator it1;
  DSV tempDSV;
  for(it = itemVec.begin(); it != itemVec.end(); ++it){
    // skip items that are not composed with any part of the input
    if ((*it)->getDSV()->empty())
      continue;

    /*
     * test whether the current span is overlapped with other spans
     * that have been extracted
     */
    bool overlapping = false;
    for(it1 = extractedSpans.begin(); it1 != extractedSpans.end(); ++it1){
      // try to add them; if they overlap it'll come back empty
      tempDSV = *((*it)->getDSV()) + (*it1);
      if (tempDSV.empty()) {
	overlapping = true;
	break;
      }
    }
    if(not overlapping){
      if(debug(4)) { cout<<"Extracting "<<*((Item*)(*it))<<endl;}
      extractedSpans.insert(*((*it)->getDSV()));
      items.push_back((Item*)(*it));
    }
  }

  if(debug(4))
    printStats(cout);
}


//! Prints some parsing statistics
inline void Logic::printStats(ostream& out) const {
  m_spMainChart->printStats(out);
}

//! Returns the total number of items contained.
inline size_t Logic::size() const {
  return m_spMainChart->size();
}

/**
 * @brief Expands an item, generating new items from inferences.
 * @param triggerItem The triggering item for inference firing.
 * @param itemList The newly generated items.
 * The vector is cleared at the beginning of the method.
 * @return The pointer to the item where the
 * trigger item is stored. Null is returned if the trigger
 * item is NULL or the inference deriving the trigger item
 * is pruned.
 */
// This is the template method design pattern, which is implemented by
// calling protected methods memoize() and genNewInfs().
// Returns true if control of triggerItem was given to a BSP
/// Item is passed in by reference to pointer, so that if the
/// underlying chart swaps it, the caller can still use it.
MemoResult Logic::expand(Item*& triggerItem, vector<Item*> & retItemVec) {
  retItemVec.clear();

  MemoResult ret = IGNORE;
  vector<Item*> localItemVec;
  if(triggerItem){
    //! If the input item has infinite score, which means that
    //! this item is underivable, then ignore it.
    if(triggerItem->getScore() >= POSINF) {//==
      if(debug(2))
	dout()<<"post-agenda pruning (inf cost) " << triggerItem; 
      // ret = Logic::IGNORE;
      return ret;
    }

    // Check pruning strategy, which can be viewed as a side condition
    // of inference rules
    if(m_pPruner && m_pPruner->prunable(triggerItem)){
      if(debug(2))
	dout()<<"post-agenda pruning (by pruner) " << triggerItem; 
      return ret;
    }

    // this Inference is needed in case we do an insertion
    BSP<Inference> inference = triggerItem->getInference();
    // memoization can happen in the chart, the pruner's index, and/or elsewhere
    ret = memoize(triggerItem);
    // generate new inferences
    switch (ret) {
    case INSERT : {
      // Insertion does a copy of the item, which does
      // not copy the inference. This function updates
      // the inference and makes sure it has two way pointers.
      //      triggerItem->fuseInference(inference);

      // Generates new items.
      genNewInfs(triggerItem, localItemVec);
      break;
    }
    default: break;
    }

    // neither of these checks are necessary for Scans and Loads
    for(vector<Item*>::iterator it = localItemVec.begin(); it != localItemVec.end(); it++){
      // Here is a hack specific to Viterbi semirings and best-first
      // search.  Under these conditions, an item will have lowest cost
      // the first time that it is derived.  So there's no point
      // considering it again.
      if (derived(*it)) {
	if(debug(3)){ 
	  dout()<<"pre-agenda pruning (already derived) " <<endl; 
	  (*it)->dump(cerr);
	  cerr << endl;
	}
	delete (*it);
      } // if pruning, delete it
      else if(m_pPruner && m_pPruner->prunable(*it)){
	if(debug(3)){ 
	  dout()<<"pre-agenda pruning (based on pruner) " <<endl; 
	  (*it)->dump(cerr);
	  cerr << endl;
	}
	delete (*it);
      }
      else//else, keep it, we are not pruning
	retItemVec.push_back(*it);
    }  
  }
  // If the trigger is NULL, then parsing has just begun, and we need
  // to generate the first batch of inferences.
  else {
    genNewInfs(triggerItem, retItemVec);
  }

  return ret;
}

/**
 * Clears the logic. (The contents of the working chart and pruning
 * strategy are cleared. The working charts are deleted. The sentence
 * is cleared with the dimensionality preserved. Temporary working
 * data are also cleared.
 */
void Logic::clear() {
  if(m_spMainChart) { m_spMainChart->clear(); }

//   for(vector<BSP<Chart> >::iterator i = m_savedCharts.begin();
//       i != m_savedCharts.end(); ++i){
//     delete *i;
//   }

  m_savedCharts.clear();

  if(m_pPruner) {m_pPruner->reset();}

  m_sent.clear();
}

//! Backup the working chart and then clear its content.
void Logic::saveCharts() {
  // must be implemented in the subclass
  THROW("Logic::saveCharts() not implemented");

  /*
    m_savedCharts.push_back(m_spMainChart);
    m_spMainChart = m_spMainChart->memorylessClone();
  */
}

/** Adds this class' option description to the option description list of the hierarchy */
BSP<BPO::options_description> Logic::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  optionsDescription->add_options()
    ("D", BPO::value<unsigned int>(), "Logic dimensionality")
    ("InputArity", BPO::value<unsigned int>(), "How many components to consider in each input sentence tuple?")
    ("PrunerBuilderConfigFile", BPO::value<string>(), "The config file for the pruner builder.")
    ("debug", BPO::value<unsigned int>()->default_value(0), "The debug level.")
    ("ChartConfigFile", BPO::value<string>(), "")
    ;
  return optionsDescription;
}
