#ifndef _Logic_H_
#define _Logic_H_

#include <vector>
#include <set>
#include "SentenceTuple.hpp"
#include "GPDefs.hpp"
#include "PruningStrategy.hpp"
#include "HeadMTree.hpp"
#include "TerminationTest.hpp"
#include <boost/program_options.hpp>
#include "Dbug.hpp"

class Grammar;
class Chart;
class Inference;
class TLink;

using namespace std;

//! Factory of inferences.
class Logic : public Dbug {

private:

  //! adds the additional options that the configFile contains
  BSP<BPO::options_description> 
  addOptionDescriptionsTo(BSP<BPO::options_description> priorOptions);

protected:

  //! Constructor accessible from subclasses only
  Logic(const string& configFile,
	BSP<BPO::options_description> priorOptions =
	BSP<BPO::options_description>(new BPO::options_description())
	);

public:

  //! Destructor.
  virtual ~Logic();

  virtual TerminationTest* getGoal();
  virtual void setGoal(TerminationTest&);

  /// Expands an item, generating new inferences.  Returns true if
  /// input item was memoized.
  /// Item is passed in by reference to pointer, so that if the
  /// underlying chart swaps it, the caller can still use it.
  virtual MemoResult expand(Item*&, vector<Item*>&);

  /** Save the current charts and then clear its content. */
  virtual void saveCharts();

  //! Get the score of an Item.
  virtual pair<SCORE,SCORE> getIOScores(const Item* item) const;

  //! Check whether a given Item has already been derived
  virtual bool derived(const Item* item) const;

  /*! Exports a parse forest of subtrees, each subtree being associated
   * with its inside score.
   *
   * The union of the yields of the subtrees will cover the
   * entire sentence tuple, and there is no crossing brackets
   * among the yields of the subtrees.
   *
   * This method is usually called if there is no complete
   * parse obtained.
   */
  virtual WTreeVec exportParses() = 0;

  //! Export a set of non-overlapping items that maximally cover the
  //! entire span.
  virtual void exportMaxCover(Vector<const Item*>& ret);

  //! Prints some parsing statistics
  virtual void printStats(ostream& out) const;

  //! Returns the total number of items contained.
  virtual size_t size() const;

  /************************************************************************
                     Logic Component Accessing Methods
   ************************************************************************/

  //! get the grammar
  virtual Grammar* getGrammar();

  /** Gets the chart. */
  virtual const BSP<Chart> getChart() const { return m_spMainChart;}
  virtual BSP<Chart> getChart() { return m_spMainChart;}

  /** Gets the pruner. */
   virtual const PruningStrategy* pruner() const { return m_pPruner;}
   virtual PruningStrategy* pruner() { return m_pPruner;}

   /** Returns the reference to the sentence tuple. */
   SentenceTuple& snt() { return m_sent;}
   const SentenceTuple& snt() const { return m_sent;}

   //! dimensionality of inferences
   arity_t arity() const { return m_arity;}

   //! number of input components
   arity_t inputArity() const { return m_inputArity;}

  /************************************************************************
                     Logic State Mutating Methods
   ************************************************************************/

  /** Initialize the Logic components using a sentence tuple.*/
  virtual void initialize(SentenceTuple&);

  //! set the grammar
  virtual void setGrammar(Grammar* gr);

  //! set the chart -- useful for sharing charts among different concurrent logics
  virtual void setChart(BSP<Chart> ch);

  /**
   * Clears the logic. (The contents of the working chart and pruning
   * strategy are cleared. The working charts are deleted. The
   * sentence is cleared with the dimensionality preserved. Temporary
   * working data are also cleared.
   */
  virtual void clear();

  size_t getNumMatches(){return m_numMatches;};
  size_t getNumUniqueMatches(){return m_numUniqueMatches;};

protected: // methods

  //! store the item in the chart
  virtual MemoResult memoize(Item* & triggerItem) = 0;

  //! Generate new inferences triggered by the input item.
  virtual void genNewInfs(const Item* triggeringItem,
			  vector<Item*>& infVec) = 0;

protected: // data

  /// for diagnostics
  static size_t sm_logicCount; // how many have we constructed so far?
  size_t m_uniqueID;  // for each instance
  
  //number of proposed matches
  size_t m_numMatches;

  //number of unique proposed matches
  size_t m_numUniqueMatches;

  //! inference dimensionality.
  arity_t m_arity;

  //! How many components to consider from each input sentence tuple?
  arity_t m_inputArity;

  //! The sentence being parsed.
  SentenceTuple m_sent;

  //! The grammar
  Grammar* m_pGrammar;

  /** The working chart in use in the current parsing. */
  BSP<Chart> m_spMainChart;

  /** The vector of charts from previous passes of multi-pass parsing.
   */
  vector<BSP<Chart> > m_savedCharts;

  /**
   * The pruning strategy. We put it in Logic because pruning strategies
   * can be viewed as side conditions to fire inferences.
   */
  PruningStrategy* m_pPruner;

  //! The goal of the inference engine, if any
  TerminationTest* m_pGoal;
};

#endif
