/// \brief Multitree

#ifndef _MTree_H_
#define _MTree_H_

#include "StringVocabulary.hpp"
#include "Terminal.hpp"
#include "SynCat.hpp"
#include "SentenceTuple.hpp"
#include "tree.hpp"
#include "HeadMTreeNode.hpp"
#include "PAVTree.hpp"

class MTree;
// weighted tree vector
typedef vector<pair<MTree*, SCORE> > WTreeVec;

class MTree : public PAVTree<tree<HeadMTreeNode> > {
public:

  virtual ~MTree(){};

  //MTree output
  friend ostream& operator<<(ostream& out, const MTree& mTree);
  friend ostream& operator<<(ostream& out, MTree& mTree);

  //! Prints a subtree
  virtual void print(ostream& out, const MTree* tr, iterator it) const {
    THROW("Not implemented");
  }

  //   //! The order of Kids.
  //   friend bool operator<(const iterator it1, const iterator it2);
  
  //! Gets the yield of the MTree and returns it in a SentenceTuple.
  static SentenceTuple yield(iterator i);
  SentenceTuple yield();
  static void yield_helper(Vector<Terminal>& os, iterator it, 
			   unsigned int segment, arity_t d);

  //! The Charniak dummy root in the multitree
  //! is the first kid of the multitree root.
  void removeCharniakDummyRoot( );

  //! Copy the nonterminal label from the first dim
  //! to the remaining dims in the same NTLink.
  void copySynCats();

  //! Remove the dummy root and EOS node in a parse tree of our
  //! own format.
  void removeDummyNodes( );

  //! Naturalizes a parse tree by removing
  //! dummy nodes (e.g., EOSH or ROOT), and by organizing
  //! siblings in linear order.
  void naturalize(StringVocabulary& synCatVcb);

  //! Replace the lexical heads of all nonterminals with the dummy terminal.
  void delexicalize(Terminal& dummy, SynCat& dummyNT, SynCat& inactiveNT);

  //! delexicalize the tree (hack of setting to 0).
  void delexTree();
  void delexTree_helper(iterator it);

  /** Returns true if the node yields a base phrase */
  bool isBasePhrase(iterator& it) const;

  void getLeftIters(iterator& tr, arity_t& d, vector<iterator>& ret);

  /**
   * Reintegerize the NTs in the multitree based on new
   * NT Vocabs. NTs will be added into the new NT
   * Vocabs if they don't occur in the new ones.
   *
   * @param tr The multitree.
   * @param newNtVcb The new nonterminal vocabs, each per language.
   * @param oldNtVcb The old nonterminal vocabs, each per language.
   * @param newTVcb The new terminal vocabs, each per language.
   * @param oldTVcb The old terminal vocabs, each per language.
   */
  void reintNTs( vector<StringVocabulary*> newNtVcb,
		 vector<StringVocabulary*> oldNtVcb,
		 vector<StringVocabulary*> newTVcb,
		 vector<StringVocabulary*> oldTVcb);

  void appendSpans(vector<unsigned int> &);
  void appendSpansHelper(iterator& it, arity_t d, unsigned int& position, int segment);

  void findFirstDiscontinuity(iterator& outit) ;
  void findFirstDiscontinuityHelper(iterator it, iterator& outit);

  /// The purpose of this method is to abstract away from design
  /// decisions about whether leaves should contain only terminals
  /// or also the preterminal labels.  As of Mar'06, HeadMTrees store
  /// only terminal info in leaves, but PositionTrees store both the
  /// terminal and its POS tag.
  /// if iter is not a leaf, then throw exception
  static const SynCat pretermLabel(const iterator& iter, arity_t dim);

  /**
   * Reintegerize the NTs in the multitree based on new
   * NT Vocabs.
   */
  void reintNTs(MTree& tr,
		vector<StringVocabulary*> newNtVcb,
		vector<StringVocabulary*> oldNtVcb);

  // static wrapper that helps to reduce methods called directly on the iterator
  static unsigned int number_of_children(const iterator& it) {
    return it.number_of_children();
  }

  static arity_t s_arity;
};


#endif
