#include "MaxCoverComparator.hpp"
#include "Vector.hpp"
#include "MoreComparators.hpp"
#include "DSV.hpp"
#include "Hedge.hpp"
#include "HeadPMTGB.hpp"
#include "Dbug.hpp"

MaxCoverComparator::MaxCoverComparator(Grammar* g, arity_t ar)
  : m_pWMTG(dynamic_cast<WMTG*>(g)), m_pW2WTM(NULL), 
    m_inputArity(ar) {
  
  W2WTM* pw2wtm = dynamic_cast<W2WTM*>(m_pWMTG);
  if (pw2wtm) 
    m_pW2WTM = addressof(pw2wtm->getW2WTM());
};

bool MaxCoverComparator::operator()(const Item* p1, const Item* p2) const {
  vector<size_t> sizes1 = p1->getSizes();
  vector<size_t> sizes2 = p2->getSizes();
  const DSV* dsv1 = p1->getDSV();
  const DSV* dsv2 = p2->getDSV();

  //First compare by dimensionality
  if(p1->getFanoutV().dimensionality() > p2->getFanoutV().dimensionality()){
    return true;
  }
  else if(p1->getFanoutV().dimensionality() < p2->getFanoutV().dimensionality()){
    return false;
  }

  //Next compare by product of existing dimensions, but only on the input side
  unsigned int prod1 = 1;
  unsigned int prod2 = 1;
  // both items should have same dimensionality at this point
  arity_t relevantDims = std::min((unsigned int)dsv1->size(), (unsigned int)m_inputArity);
  for(unsigned int i = 0; i < relevantDims; ++i)
    if(dsv1->length(i) > 0)
      prod1 *= dsv1->length(i);
  for(unsigned int i = 0; i < relevantDims; ++i)
    if(dsv2->length(i) > 0)
      prod2 *= dsv2->length(i);
  if(prod1 < prod2){
    return false;
  } else if (prod2 < prod1){
    return true;
  }

  prod1 = 1;
  prod2 = 1;
  for(unsigned int i = 0; i < sizes1.size(); i++)
    if(sizes1[i] > 0)
      prod1 *= sizes1[i];
  for(unsigned int i = 0; i < sizes2.size(); i++)
    if(sizes2[i] > 0)
      prod2 *= sizes2[i];
  if(prod1 < prod2){
    return true;
  } else if (prod2 < prod1){
    return false;
  }
  
  // special case for 1Dx1D (consequents of preterminal productions)
  if (m_pW2WTM) {
    const PAV& pav1 = p1->getInference()->getPAV();
    const PAV& pav2 = p2->getInference()->getPAV();
    if (pav1.getPA(0).size() == 1 
      && pav1.getPA(1).size() == 1
      && pav2.getPA(0).size() == 1 
      && pav2.getPA(1).size() == 1) {
      const Hedge* pH1 = dynamic_cast<const Hedge*>(p1);
      Terminal t10 = pH1->getNTLink().getTLink()[0];
      Terminal t11 = pH1->getNTLink().getTLink()[1];
      W2WLink lk1(t10, t11);
      
      const Hedge* pH2 = dynamic_cast<const Hedge*>(p2);
      Terminal t20 = pH2->getNTLink().getTLink()[0];
      Terminal t21 = pH2->getNTLink().getTLink()[1];
      W2WLink lk2(t20, t21);
      try {
	SCORE w2w1 = m_pW2WTM->lookup(lk1);
	SCORE w2w2 = m_pW2WTM->lookup(lk2);
	if (Dbug::isDebugAll(5))
	  cerr << "MCC: comparing " << *pH1 << " and " << *pH2
	       << " Terminals " << t10 << " " << t11 << " " << t20 << " " << t21
	       << " Scores: " << w2w1 << " " << w2w2 << endl;      
	if(w2w1 < w2w2){
	  return true;
	} else if (w2w2 < w2w1){
	  return false;
	}
      } catch (KeyNotFound<W2WLink>) {}
    }
  }

  // Next compare by cost, lower cost goes first.
  SCORE rootCost1 = m_pWMTG->rootCost(p1);
  SCORE rootCost2 = m_pWMTG->rootCost(p2);
  SCORE s1 = p1->getScore() + rootCost1;
  SCORE s2 = p2->getScore() + rootCost2;
  cerr << "";
  if(s1 < s2) {
    return true;
  } else if(s1 > s2) {
    return false;
  }

  // Next compare by auxiliary cost (will be zeros if unused)
  SCORE as1 = p1->getAuxScore() + rootCost1;
  SCORE as2 = p2->getAuxScore() + rootCost2;
  if(as1 < as2) {
    return true;
  } else if(as1 > as2) {
    return false;
  }

  //Next compare component-size by size (smallest is best)
  for(unsigned int i = 0; i < sizes1.size() && i < sizes2.size(); i++){
    if(sizes1[i] < sizes2[i]){
      return true;
    } else if(sizes1[i] > sizes2[i]){
      return false;
    }
  }

  //Next compare by NTLink
  const Hedge* h1 = dynamic_cast<const Hedge*>(p1);
  const Hedge* h2 = dynamic_cast<const Hedge*>(p2);
  if (! h1 or ! h2)
    THROW("Non-hedges!");
  int cmpr = h1->getNTLink().cmp(h2->getNTLink());
  if(cmpr > 0){
    return true;
  } else if(cmpr < 0){
    return false;
  }

  //Next compare by span.
  cmpr = dsv1->cmp(*dsv2);
  if(cmpr > 0){
    return true;
  } else if (cmpr < 0){
    return false;
  }

  //Next compare by hidden span.
  const Vector<DSV>* hdsv1 = p1->getHiddenDSVs();
  const Vector<DSV>* hdsv2 = p2->getHiddenDSVs();
  if (hdsv1 or hdsv2) {
    if (not hdsv1)
      return true;
    else if (not hdsv2)
      return false;

    cmpr = MC::cmp(*hdsv1, *hdsv2);
    if(cmpr > 0){
      return true;
    } else if (cmpr < 0){
      return false;
    }
  }

  // if we get to this point, then the two pointers better
  // point to the same item
  if(p1 == p2){
    return true;
  } else {
    THROW("Identical items being compared:" << endl 
	  << *h1 << "and" << endl << *h2);
  }

  return false;
};
