
#ifndef _NTLink_H_
#define _NTLink_H_

#include "Set.hpp"
#include "GPDefs.hpp"
#include "MoreAlgorithms.hpp"
#include "StringVocabulary.hpp"
#include "Nonterminal.hpp"
#include "LexNT.hpp"
#include "RawNTLink.hpp"
#include "TLink.hpp"
//#include "GPcmp.hpp"
#include <boost/serialization/serialization.hpp>
#include "boost/multi_array.hpp"
using namespace std;

//! This class holds a vector of Nonterminal pointers. Each of these
//! "nonterminals" has at least a SynCat. One should use caution when
//! using this class, since it hold pointers.

class NTLink {
 typedef RawNTLink _super;
public:

  //! Empty constructor.
  NTLink();

  //! Constructs an _d dimensional NTLink, and each
  //! dimension is initialized with nonterminal nt;
  NTLink(const arity_t& _d, const Nonterminal& nt);

  //! Constructs an _d dimensional NTLink, and each
  //! dimension is initialized with LexNT nt;
  NTLink(const arity_t& _d, const LexNT& nt);

  //! Constructs an 2 dimensional NTLink, and each
  //! dimension has one of the inputs;
  NTLink(Nonterminal* nt1, Nonterminal* nt2);

  //! Constructs an NTlink with "LexNT's" as its members.
  NTLink(const NTLink&,const TLink&);

  //! Constructor (from a Vector).
  //NTLink(const Nonterminal* ve[]);
  NTLink(const vector<Nonterminal*> &vec);
  
  NTLink(const RawNTLink& raw) ;

  //! Constructor (from a nonterminal, only for D == 1).
  NTLink(const arity_t d);

  //! delete elements and clear
  void clear();

  //! comparator
  int cmp(const NTLink& other) const;

  //! Destructor. Destroys the Nonterminal pointers.
  inline ~NTLink() {  }

  //! D-dimensional cross-product of the input NTs.
  static void hyperCrossProduct(vector<Set<Nonterminal*,less<Nonterminal*> > >& sets, set<NTLink>& result);

  //! D-dimensional cross-product of the input NTs.
  static void hyperCrossProduct2(vector<set<Nonterminal*, less<Nonterminal*> > >& sets, set<NTLink>& result);

  //! return a delexicalized version of this NTLink
  void delex();

  //returns reference to element
  const Nonterminal* getNT(unsigned int i)const{
    return sm_voc.getWord(m_index)->getNT(i);
  }

  //returns reference to element
  void setNT(unsigned int i, Nonterminal* nt);

  //< operator
  inline bool operator<(const NTLink& other)const{
    return (m_index < other.m_index);
  }

  // > operator
  inline bool operator>(const NTLink& other)const{
    return (m_index > other.m_index);
  }

  // size (or dimensionality) of link.
  inline unsigned int size()const{
    return  sm_voc.getWord(m_index)->size();
  }

  // makes a copy of this NTLink
  inline NTLink* clone() const{
    return new NTLink(*this);
  };

  // reverse the elements in link
  void reverse();

  //! return a vector of terminals associated with his NTLink.
  //! Can only be used when elements are LexNTs
  TLink getTLink() const;

  //! sets the terminals in this NTLink.
  //! Can only be used when elements are LexNTs
  void setTLink(const TLink& tl) ;

  void setTerminal(const Terminal& t, arity_t d);

  //! Returns the dimensionality.
  inline arity_t arity() const { return sm_voc.getWord(m_index)->arity();}

  // == operator (based on values)
  inline bool operator==(const NTLink& other) const {
    return (m_index == other.m_index);
  }

  inline size_t hash() const {
    return m_index;
  }

  //! sets all links to null
  void setNull();

  //! Returns a NTLink with each element to be NULLNT.
  static NTLink nullNTLink(const arity_t& d );

  //! Returns a Dummy NTLink.
  static NTLink dummyNTLink(const arity_t& d) ;

  //! Returns a Unk NTLink.
  static NTLink unkNTLink(const arity_t& d) ;

  //! Returns a unknown preterminal NTLink.
  static NTLink prtmUnkNTLink(const arity_t& d) ;

  //! Write.
  ostream& put(ostream& out) const;

  //! write with nonterminal info.
  ostream& put(ostream& out, vector<StringVocabulary*> ntvVec)const;

  //! write with nonterminal and terminal info. (only used with
  //! lexical case)
  ostream& put(ostream& out, StringVocabulary* ntSet,
           StringVocabulary* terminalSet) const;

  //! Overload only not to output the space at the end.
  friend ostream& operator << (ostream& out, const NTLink& obj);

  //! Read.
  istream& get(istream& in) ;

  //! If arity is not specified, reads line by line.
  //! Otherwise, reads in terms of arity number.
  friend istream& operator >> (istream& in, NTLink& obj);

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    // ar & boost::serialization::base_object<m_array>(*this);
  };

  bool isDummy()const;

protected: // data

  // currently an unsigned int,
  // but can probably be changed to unsigned short
  VocabularyIndex m_index;

  // currently hashed in both directions,
  // but one dir should be a vector to be faster
  static Vocabulary<RawNTLink > sm_voc;
};

namespace HashNS {
  template<> struct hash< NTLink > {
    size_t operator()(const NTLink& e) const{
      return e.hash();
    }
  };

}

#endif
