#include "File.hpp"
#include "NaiveWMTG.hpp"
#include "TLink.hpp"
#include "COMPILER-DEPENDENT_HEADERS.hpp"
#include "HeadMTree.hpp"
#include "MultinomialDistribution.hpp"
#include "GlobalVocab.hpp"
#include "Hedge.hpp"
#include "HeadProduction.hpp"
#include "Exceptions.hpp"
#include "ScoreSetPtrs.hpp"
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

//! Destructor. To clear the memory allocated.
NaiveWMTG::~NaiveWMTG() {
  m_prods.clear();
}

NaiveWMTG::
NaiveWMTG(arity_t dm,
	  const string& configFile,
	  char** overrideArgV,
	  size_t overrideArgC,
	  BSP<BPO::options_description> optionsDescription
	  )
  : _grammar(dm), 
    _super(dm, configFile, addOptionDescriptionsTo(optionsDescription)) {
  string optTreebankGrammarFile;

  /*
   * Compile the config file.
   */

  BPO::variables_map vm;
  if(overrideArgV){
    if(this->debug(1)){
      cerr<<"NaiveWMTG: Overriding config file arguments..."<<endl;
    }
    store(BPO::command_line_parser(overrideArgC, overrideArgV).
	  options(*optionsDescription).run(), vm);
  }
  if(configFile != ""){
    iFile ifs(configFile.c_str());
    store(parse_config_file(ifs, *optionsDescription), vm);
  }
  notify(vm);

  if(vm.count("TreebankGrammar")) optTreebankGrammarFile = vm["TreebankGrammar"].as<string>();

  if(this->debug(1)) {
    cerr << "NaiveWMTG options:" << endl;
    cerr << OptionsPrinter(*optionsDescription, vm) << endl;
  }

  if(optTreebankGrammarFile != ""){
    iFile iF(optTreebankGrammarFile);
    Production* pProd;
    if (m_isHeaded)
      pProd = new HeadProduction;
    else 
      pProd = new Production;
    m_prods.read(iF, pProd);
    delete pProd;
    iF.close();
  }
}

size_t NaiveWMTG::size() const {
  return m_prods.size();
};

// return the sum of the scores of all features
SCORE NaiveWMTG::mass() {
  return m_prods.sum();
};

//! begin iterator
NaiveWMTG::iterator NaiveWMTG::begin() {
  return m_prods.begin();
};

//! const begin iterator
NaiveWMTG::const_iterator NaiveWMTG::begin() const {
  return m_prods.begin();
};

//! end iterator
NaiveWMTG::iterator NaiveWMTG::end() {
  return m_prods.end();
};

//! const end.iterator
NaiveWMTG::const_iterator NaiveWMTG::end() const {
  return m_prods.end();
};

//! input
istream& NaiveWMTG::get(istream& is){
  // We read in the PAV vocab, but if we already have one, then we
  // ignore the one that was read in.  That way, we can override the
  // PAV vocab of the grammar being read using a config file.
  Vocabulary<PAV> pavVcb;
  is >> pavVcb;
  if (m_pavVcb.empty()) // use it only if existing one is empty
    insertPAV(pavVcb);
  cerr<<"Reading Productions ..."<<endl;
  Production* pProd;
  if (m_isHeaded)
    pProd = new HeadProduction;
  else 
    pProd = new Production;
  m_prods.read(is, pProd);
  delete pProd;
  return is;
}

//! output
ostream& NaiveWMTG::put(ostream& os) const {
  _super::put(os);
  cerr<<"Writing PAVs ..."<<endl;
  os << m_pavVcb;
  os << m_prods;
  return os;
}

//! Increment productions from a grammar with iterators
void NaiveWMTG::increment(const Grammar* pG ) {
  const NaiveWMTG* pNaive = dynamic_cast<const NaiveWMTG*>(pG);
  if (not pNaive)
    THROW("Grammar type mismatch in  NaiveWMTG::increment");
  for(const_iterator nwci = pNaive->begin(); nwci != pNaive->end(); ++nwci) {
    pair<const Production*, SCORE> scoredProd = *nwci;
    Production* pr = scoredProd.first->clone();
    updateFeatureScore(pr, scoredProd.second);
  }
}

void NaiveWMTG::increment(Production* rule, SCORE count) {
  updateFeatureScore(rule, count);
}

//! update the score of a production, NOT propagating to subfeatures
void NaiveWMTG::updateFeatureScore(Production* pProd , SCORE sc) {
  if (debug(3))
    cerr << "NaiveWMTG " << m_uniqueID << "::incr by " << sc << " " << *pProd;

  W2WLink lk;
  bool found = pProd->getTLink(lk);

  ScoreSetPtrs<Production>::iterator ssci = m_prods.find(pProd);
  // update the prod itself
  SCORE newScore;
  if (ssci == m_prods.end()) {
    m_prods.insert(pProd, sc);
    newScore = sc;
  } else {
    //this production is not going to be inserted.
    // so it must be deleted
    delete pProd;
    pProd = 0;
    ssci->second += sc;
    // next, update the w2wTM
    newScore = ssci->second;
  }

  // update the w2wTM
  if (found) {
    ScoreSet<W2WLink>::iterator w2wi = m_w2wTM.find(lk);
    if (w2wi == m_w2wTM.end()) {
      m_w2wTM.insert(lk, newScore);
    } else {
      SCORE oldScore = w2wi->second;
      if (m_costAggregator == SUM) {
	w2wi->second = oldScore + newScore;
      } else { // MAX or MAXSUM should use min [sic] (not max)
	w2wi->second = min(oldScore, newScore);
      }
    }
  }
}

bool NaiveWMTG::knows(Production* pProd) const {
  return (m_prods.find(pProd) != m_prods.end());
}

//! determine whether a certain inference should be allowed
//! for a NaiveWMTG it is allowed iff its prod is in the prod set
bool NaiveWMTG::allowInf(const Inference* pInf) {
  Production prod(pInf);
  ScoreSetPtrs<Production>::const_iterator pi = 
    m_prods.find(addressof(prod));

  return (pi != m_prods.end());
}

//! Returns the score (such as count or log prob) of a MTG production.
SCORE NaiveWMTG::getProdScore(Production* prod) const {
  ScoreSetPtrs<Production>::const_iterator hpci = m_prods.find(prod);
  if (hpci == m_prods.end()){
    if(debug(3))
      cerr << *prod << " not found! " <<endl;
    return POSINF;
  } else {
    return hpci->second;
  }
}

//! returns possible consequents of a TLink antecedent
void NaiveWMTG::getPossibleConsForTLink(const TLink& tl, const DSV& dsv, vector<Item*>& ret) const {
  //arity of grammar
  arity_t dim= tl.arity();
  //non emtpy dim
  arity_t nonEmptyDim;
  //active dim found
  bool found = false;

  //search for active dimension
  for (unsigned int i = 0; i < dim; i++){
    if(tl[i] != NULLT){
      nonEmptyDim = i;
      found = true;
      break;
    }
  }

  //if no dimension is active.
  if(!found)
    return;

  set<NTLink> possibleTagsForTLink;
  getPossibleTagsForTLink(tl, possibleTagsForTLink);
  set<NTLink>::const_iterator it;
  
  if(possibleTagsForTLink.empty()){
    if(m_smoothScans){
      NTLink syncats;
      // keep the lexical heads, but assign dummy syncats
      Nonterminal* pUnk = new Nonterminal(UNKNT);
      Nonterminal* pNull = new Nonterminal(NULLNT);
      if (nonEmptyDim == 0)
	syncats = NTLink(pUnk, pNull);
      else
	syncats = NTLink(pNull, pUnk);
      NTLink ntl(syncats, tl);
      Item* h = new Hedge(ntl,dsv);
      ret.push_back(h);
    }
  } else {
    for(it = possibleTagsForTLink.begin(); it != possibleTagsForTLink.end(); it++){
      Item* h = new Hedge(*it,dsv);
      ret.push_back(h);
    }
  }

  return;
}

//! Returns all the possible pre-terminals lexicalized by
//! input terminal link.
void NaiveWMTG::getPossibleTagsForTLink(const TLink& link, set<NTLink>& resultSet) const{
  assert(arity());

  set<NTLink> prtmset;
  genNTLinks(GlobalVocab::s_synCatSet, prtmset);

  for(set<NTLink>::const_iterator it = prtmset.begin(); it != prtmset.end();
      ++it){
    NTLink nl = *it;

    /*
     * We always set the tag of the null to be UNK_PRTM.
     */
    for(arity_t d = 0; d < link.arity(); ++d){
      if(link[d] == NULLT){ nl.setNT(d,new Nonterminal(UNK_PRTM));}
    }
    resultSet.insert(NTLink(nl, link));
  }
}

//! Generates the set of ntLinks, each one formed by taking nonterminals
//! from the NT Vocabularies on corresponding dimensions.
//! Note that null NTLink and dummy NTLink are not generated.
void NaiveWMTG::
genNTLinks(StringVocabulary* ntSet, set<NTLink>& ntLinks) const {
  if(inGCNF())
    gen1DNTLinks(ntSet, ntLinks);
  else
    genDDNTLinks(ntSet, ntLinks);
}

//! Generates the set of ntLinks, each one formed by taking nonterminals
//! from the NT Vocabularies on corresponding dimensions.
//! Notice that null NTLink and dummy NTLink are not generated.
void NaiveWMTG::
genDDNTLinks(StringVocabulary* ntSet, set<NTLink>& ntLinks) const {
  size_t i;
  arity_t d;
  size_t total = 1;
  arity_t dimTotal = ntSet->size();
  vector<StringVocabulary::const_iterator> iters(dimTotal);

  // total number of ntLinks to be generated.
  for(d = 0; d < dimTotal; ++d){
    total *= ntSet->size();
  }

  // initializes iterators.
  for(d = 0; d < dimTotal; ++d){
    iters[d] = ntSet->begin();
  }

  // begin to generate ...
  for(i = 0; i < total; ++i){
    // form the ntlink based on the nonterminals pointed to
    // by the iterators.
    NTLink ntlink(dimTotal);
    for(d = 0; d < dimTotal; ++d){
      ntlink.setNT(d,new Nonterminal(SynCat(iters[d]->second)));
    }

    // memoize the formed ntlink if it is not a dummy ntlink
    // and not a null ntlink.
    if((!(ntlink == NTLink::dummyNTLink(dimTotal))) &&
       (!(ntlink == NTLink::nullNTLink(dimTotal)))){
      ntLinks.insert(ntlink);
    }

    // move the iterators to the next ...
    for(d = 0; d < dimTotal; ++d){
      ++iters[d];

      // if not point to the end, then move next.
      if(iters[d] != ntSet->end()){
	break;
      } else {
	iters[d] = ntSet->begin();
      }
    }
  }
}

//! Generates the set of ntLinks, each one formed by taking nonterminals
//! from the NT Vocabularies on corresponding dimensions.
//! Notice that null NTLink and dummy NTLink are not generated.
void NaiveWMTG::
gen1DNTLinks(StringVocabulary* ntSet, set<NTLink>& ntLinks) const {
  arity_t d,d2;
  arity_t dimTotal = ntSet->size();
  vector<StringVocabulary::const_iterator> iters(dimTotal);

  // initializes iterators.
  for(d = 0; d < dimTotal; ++d){
    iters[d] = ntSet->begin();
  }

  for(d = 0; d < dimTotal; ++d){
    NTLink ntlink(dimTotal);
    for(d2 = 0; d < dimTotal; ++d){
      if(d2 == d)
	ntlink.setNT(d,new Nonterminal(SynCat(iters[d]->second)));
      else
	ntlink.setNT(d,new Nonterminal(DummyNT));
    }
    ntLinks.insert(ntlink);
  }
}

/// Compile w2w model from the productions, assuming GCNF.  Currently,
/// this ignores links involving a null in the 0th component.
BSP<W2WDictionary> NaiveWMTG::getW2WDict() const {
  BSP<W2WDictionary> ret( new W2WDictionary);
  const_iterator ci;
  W2WLink lk;
  for(ci = begin(); ci != end(); ++ci) {
    if (ci->first->getTLink(lk)) {
      ret->insert(lk);
    }
  }
  return ret;
}

/** Clears the content (only) of the containers. */
void NaiveWMTG::clear() {
  _super::clear();

  // erase production container
  m_prods.clear();
}

/** Adds this class' option description to the option description list of the hierarchy */
BSP<BPO::options_description> 
NaiveWMTG::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  optionsDescription->add_options()
    ("TreebankGrammar", BPO::value<string>(), "Treebank grammar")
    ;
  return optionsDescription;
}
