#ifndef _NaiveWMTG_H_
#define _NaiveWMTG_H_

#include "MultiMap.hpp"
#include "StringVocabulary.hpp"
#include "Set.hpp"
#include "Histogram.hpp"
#include "MultinomialDistribution.hpp"
#include "LearnableWMTG.hpp"
#include "HasW2WDict.hpp"
#include "W2WTM.hpp"
#include "ScoreSetPtrs.hpp"
#include <boost/program_options.hpp>
#include <boost/serialization/serialization.hpp>

//! A naive implementation of WMTG, using just a histogram of production rules.
class NaiveWMTG : public LearnableWMTG, public HasW2WDict, public W2WTM {
  typedef LearnableWMTG _super;
  typedef Grammar _grammar;
 public:

  //! Iterator to iterate all the MTG productions.
  typedef ScoreSetPtrs<Production>::iterator iterator;
  typedef ScoreSetPtrs<Production>::const_iterator const_iterator;

  /// constructor from config file
  NaiveWMTG(arity_t dm,
	    const string& configFile,
	    char** overrideArgV = NULL,
	    size_t overrideArgC = 0,
	    BSP<BPO::options_description> priorOptions 
	    = BSP<BPO::options_description>(new BPO::options_description())
	    );

  //! Destructor.
  virtual ~NaiveWMTG();

  //! Assignment.
  // NaiveWMTG& operator = (const NaiveWMTG&);

  virtual size_t size() const;

  //! begin iterator
  iterator begin();

  //! const begin iterator
  const_iterator begin() const;

  //! end iterator
  iterator end();

  //! const end.iterator
  const_iterator end() const;

  bool knows(Production* pProd) const;

  // return the sum of the scores of all features
  virtual SCORE mass();

  //! Increment a production
  virtual void increment(Production* , SCORE = UNITYSCORE);

  //! Increment features from a grammar
  virtual void increment(const Grammar* );

  //! update the score of a production, NOT propagating to subfeatures
  virtual void updateFeatureScore(Production* pProd, SCORE sc);

  //! input
  virtual istream& get(istream& is);

  //! output
  virtual ostream& put(ostream& os) const;

  //! determine whether a certain inference should be allowed
  virtual bool allowInf(const Inference* pInf);

  //! returns possible consequents of a TLink antecedent
  virtual void getPossibleConsForTLink(const TLink& t, const DSV& d, vector<Item*>& vItem) const;

  /// Compile w2w model from the productions, assuming GCNF.
  /// Currently, this ignores links involving a null in the 0th
  /// component.
  virtual BSP<W2WDictionary> getW2WDict() const;

  ////////////////////////////////////////////////////////////////////
  // Input & Output //
  ///////////////////////////////////////////////////////////////////

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    ar & boost::serialization::base_object<_super>(*this);
    ar & m_prods;
  };

  /** Clears the content (only) of the containers. */
  virtual void clear();

protected: // methods

  // added for serialization
  NaiveWMTG() {};

  //! Returns all the possible pre-terminals lexicalized by the input TLink.
  virtual void getPossibleTagsForTLink(const TLink&, set<NTLink>& returnSet) const;

  //! Returns the score (such as count or log prob) of a MTG production.
  virtual SCORE getProdScore(Production* ) const;

private: // methods

  //! Generates the set of ntlinks, each one formed by taking nonterminals
  //! from the NT Vocabularies on corresponding dimensions.
  //! Note that null NTLink and dummy NTLink are not generated.
  void genNTLinks(StringVocabulary*, set<NTLink>& ) const;

  void gen1DNTLinks(StringVocabulary*, set<NTLink>& ) const;

  void genDDNTLinks(StringVocabulary*, set<NTLink>& ) const;

  BSP<BPO::options_description> addOptionDescriptionsTo(BSP<BPO::options_description> priorOptions);

protected: // data

  //! A Histogram of MTG productions.
  ScoreSetPtrs<Production> m_prods;

};

#endif
