#include "NodeRefHedge.hpp"

//! Constructor. The same specification as that in the
//! base class. Note that the tree iterator is unspecified
//! after construction, and needs to be explicitly set using
//! method setNode.
NodeRefHedge::
NodeRefHedge(const NTLink& label, const DSV& span, const vector<short>& heads)
  : _super(label, span),
    m_headPosns(heads),
    m_childCounter(label.size()){
  m_itNode.resize(span.arity());
}

//! Constructor. The same specification as that in the
//! base class. Note that the tree iterator is unspecified
//! after construction, and needs to be explicitly set using
//! method setNode.
NodeRefHedge::
NodeRefHedge(const NTLink& label, const DSV& span,  const vector<short>& heads,
	     vector<PositionTree::iterator> vecIt)
  : _super(label, span), 
    m_headPosns(heads), 
    m_childCounter(label.size(),1),
    m_itNode(vecIt) {
  for(unsigned int i = 0; i < vecIt.size(); i++){
    if(m_itNode[i].node == NULL)
      m_childCounter[i] = 0;
  }
}

//! Constructor.
NodeRefHedge::
NodeRefHedge(const vector<Item*>& vec) : Hedge(vec.size()),  m_childCounter(vec.size()){
  size_t arity = vec.size();
  m_sizes.resize(arity);
  m_headPosns.resize(arity);
  m_itNode.resize(arity);
  //if the input item is 1D, then we only want to pull information
  // out of the first dimension. (it's the only one that exists!)
  if(vec[0]->arity() == 1)
    for(arity_t i = 0; i < arity; i++){
      if(dynamic_cast<NodeRefHedge*>(vec[i])){
	m_sizes[i]  = ((NodeRefHedge*)vec[i])->getSizes()[0];
	m_span[i] = ((NodeRefHedge*)vec[i])->getDSpan(0);
	m_label.setNT(i,(((NodeRefHedge*)vec[i])->getNTLink()).getNT(0)->clone());
	m_itNode[i] = ((NodeRefHedge*)vec[i])->getNode()[0];
	m_headPosns[i] = ((NodeRefHedge*)vec[i])->getHeads()[0];
	m_childCounter[i] = ((NodeRefHedge*)vec[i])->getChildCounter()[0];
      }
      //in case it's another type of Hedge
      else{
	m_sizes[i]  = (static_cast<Hedge*>(vec[i]))->getSizes()[0];
	m_span[i] = static_cast<Hedge*>(vec[i])->getDSpan(0);
	m_label.setNT(i,( static_cast<Hedge*>(vec[i])->getNTLink()).getNT(0)->clone());
	m_itNode[i] = NULL;
	m_headPosns[i] = 0;
	m_childCounter[i] = 0;
      }

    }
  else //else, we pull out of the ith dimension.
    for(arity_t i = 0; i < arity; i++){
      if(dynamic_cast<NodeRefHedge*>(vec[i])){
	m_sizes[i]  = ((NodeRefHedge*)vec[i])->m_sizes[i];
	m_span[i] = ((NodeRefHedge*)vec[i])->getDSpan(i);
	m_label.setNT(i,(((NodeRefHedge*)vec[i])->getNTLink()).getNT(i)->clone());
	m_itNode[i] = ((NodeRefHedge*)vec[i])->getNode()[i];
	m_headPosns[i] = ((NodeRefHedge*)vec[i])->getHeads()[i];
	m_childCounter[i] = ((NodeRefHedge*)vec[i])->getChildCounter()[i];
      }
      //in case it's another type of Hedge
      else{
	m_sizes[i]  = (static_cast<Hedge*>(vec[i]))->getSizes()[i];
	m_span[i] = static_cast<Hedge*>(vec[i])->getDSpan(i);
	m_label.setNT(i,( static_cast<Hedge*>(vec[i])->getNTLink()).getNT(i)->clone());
	m_itNode[i] = NULL;
	m_headPosns[i] = 0;
	m_childCounter[i] = 0;
      }
    }
}

//copy constructor
NodeRefHedge::NodeRefHedge(const NodeRefHedge& other) 
  : _super(other), m_headPosns(other.m_headPosns), m_childCounter(other.m_childCounter),m_itNode(other.m_itNode)
{}

//! Assignment (label, span, fanoutv, node_it).
NodeRefHedge& NodeRefHedge::operator = (const NodeRefHedge& nrhedge) {
  _super::operator=(nrhedge);
  m_itNode = nrhedge.m_itNode;
  m_headPosns = nrhedge.m_headPosns;
  return *this;
}

int NodeRefHedge::cmp(const Item* item, bool sameType) const {
  const NodeRefHedge* pHedge = dynamic_cast<const NodeRefHedge*>(item);
  int ret;
  // fanout vector
  if((ret = MC::cmp(getFanoutV(), pHedge->getFanoutV())) != 0) return ret;
  // constituent label
  NTLink label = pHedge->getNTLink();
  if((ret = m_label.cmp(label)) != 0) return ret;
  // heads
  //  if((ret = MC::cmp(m_headPosns, pHedge->getHeads())) != 0) return ret;
  // span
  if((ret = m_span.cmp((DSV&)(*(pHedge->getDSV())))) != 0) return ret;

  return 0;
}

//! Equality test by comparing signature: (label, span, fanoutv, node_it).
bool NodeRefHedge::operator == (const NodeRefHedge& hedge) const {
  return ((_super::operator==(hedge)) 
	  && (getNode() == hedge.getNode()));
}

//! Compares whether two items are equal in content and type.
//! Compared with operator==, this method is virtual.
bool NodeRefHedge::equal_to(const Item& item) const {
  const NodeRefHedge* p_other = dynamic_cast<const NodeRefHedge*>(&item);
  if(p_other){
    return operator==(*p_other);
  } else {
    return false;
  }
}

/// output
void NodeRefHedge::dump(ostream& o) const{
  _super::dump(o);
  o << " {";
  for(unsigned int i = 0; i < m_itNode.size(); i++){
    PositionTree::iterator ptd = m_itNode[i];
    if (ptd != NULL){
      o << " treeIter= ";
      ptd->printPositionTreeNode(o);
      o << " nodeSize=  ";
      o << m_childCounter[i] ;
    } else{
      o << " <BlankMTNode>";
    }
  };
  o << " }";
  o << " heads @ " << getHeads();
}

Item* NodeRefHedge::get1DItem(arity_t d) const {
  NTLink ln = NTLink(1);
  ln.setNT(0,m_label.getNT(d)->clone());

  DSV dspan = DSV(1);
  dspan[0] = m_span[d];
  vector<PositionTree::iterator> vecIt(1);
  vecIt[0] = m_itNode[d];
  vector<short> headPV(1);
  headPV[0] = m_headPosns[d];
  
  NodeRefHedge* ret = new NodeRefHedge(ln, dspan, headPV, vecIt);
  return ret;
};
