/*
 * Class NodeRefHedge is an Hedge item containing
 * an iterator to a tree node.
 */

#ifndef _NodeRefHedge_H_
#define _NodeRefHedge_H_

#include "Hedge.hpp"
#include "PositionTree.hpp"
#include <vector>

/**
 * Class NodeRefHedge is an Hedge item containing
 * an iterator to a tree node.
 */
class NodeRefHedge : public Hedge {
  typedef Hedge _super;
 public:

  //! Constructor. The same specification as that in the
  //! base class. Note that the tree iterator is unspecified
  //! after construction, and needs to be explicitly set using
  //! setNode().
  NodeRefHedge(const NTLink& labels, const DSV& spans, const vector<short>& heads);

  //! Constructor. The same specification as that in the
  //! based class. Note that here the position tree node is set.
  NodeRefHedge(const NTLink& labels, const DSV& spans, const vector<short>& heads,
	       vector<PositionTree::iterator> vecIt);

  //! constructor from vector of D 1D items
  NodeRefHedge(const vector<Item*>& vec);

  NodeRefHedge(const NodeRefHedge& nhr);
  NodeRefHedge();

  //! Destructor.
  virtual ~NodeRefHedge() {};

  // returns a copy of this
  virtual Item* clone() const {
    return new NodeRefHedge(*this);
  }

  //! Sets/Gets the tree node iterator.
  void setNode(const vector<PositionTree::iterator>& vecIt) { m_itNode = vecIt;}
  const vector<PositionTree::iterator>& getNode() const { return m_itNode;}

  //! set/get the head positions
  void setHeads(const vector<short>& heads) { m_headPosns = heads;}
  const vector<short>& getHeads() const { return m_headPosns;}

  //! set a count of the number of children of the 
  // pointer that this item covers.
  void setChildCounter(const vector<short>& ns) { m_childCounter = ns;}

  //! get a count of the number of children of the 
  // pointer that this item covers.
  const vector<short>& getChildCounter() const { return m_childCounter;}

  //! returns a 1d Hedge, extracted from dimension d.
  Item* get1DItem(arity_t d) const;

  //! Assignment (label, span, fanoutv, node_it).
  NodeRefHedge& operator = (const NodeRefHedge& hedge);

  //! Equality test by comparing signature: (label, span, fanoutv, node_it).
  bool operator == (const NodeRefHedge& hedge) const;

  //! Compares whether two items are equal in content and type.
  //! Compared with operator==, this method is virtual.
  virtual bool equal_to(const Item& item) const;

  //! Output Adaptor.
  //  friend ostream& operator<<(ostream& o, const NodeRefHedge& h);
  virtual void dump(ostream& o) const;

  int cmp(const Item* item, bool sameType) const;

private:
  
  // positions of the head of this item in each component of the input sentence
  vector<short> m_headPosns;

  //! number of children under m_itNode that have been covered by this item
  vector<short> m_childCounter;
  
  vector<PositionTree::iterator> m_itNode;
};

namespace HashNS {
  //! Template specialization of STL's hash object.
  template<> struct hash< NodeRefHedge >
  { size_t operator() (const NodeRefHedge& item) const {return item.hash();} };

  template<> struct hash< NodeRefHedge* >
  { size_t operator() (const NodeRefHedge* pitem) const {return pitem->hash();} };

  template<> struct hash< BSP<NodeRefHedge> >
  { size_t operator() (const BSP<NodeRefHedge> pitem) const {return pitem->hash();} };

  template<> struct hash< BSP<const NodeRefHedge> >
  { size_t operator() (const BSP<const NodeRefHedge> pitem) const {return pitem->hash();} };
}

#endif
