#ifndef NONTERMINAL_H
#define NONTERMINAL_H

#include "GPDefs.hpp"
#include "IOUtil.hpp"
#include "StringVocabulary.hpp"
// #include "MoreExceptions.hpp"
#include <boost/serialization/serialization.hpp>

//! This class represents the base class for what ever is being rewritten
//! in a grammar. At the very least, you must rewrite the SynCat. However
//! if you want to add more information to the nonterminal
//! (i.e. lexicalization) you will need to subclass this class.

class Nonterminal{

public:

  //! Default constructor (mainly used for i/o)
  Nonterminal();

  //! Constructor with category
  Nonterminal(SynCat cat);

  //! destructor
  virtual ~Nonterminal();

  //! Getter for Syntactic Category
  virtual SynCat getSynCat()const;

  virtual SynCat &getSynCat();

  //! Setter for Syntactic Category
  virtual void setSynCat(const SynCat c);

  virtual SynCat& nt(){
    return m_synCat;
  }
  virtual SynCat nt() const{
    return m_synCat;
  }

  // THE FOLLOWING NEED TO BE OVERLOADED WHEN YOU SUBCLASS!

  virtual bool isInactive() const;

  //! output
  friend ostream& operator << (ostream& out, const Nonterminal& obj);

  //! output with vocab deciphering
  virtual ostream& put(ostream& out)const;;

  //! output with vocab deciphering
  virtual ostream& put(ostream& out, StringVocabulary* ntset)const;

  //! output with vocab deciphering
  virtual ostream& put(ostream& out, StringVocabulary* ntset, StringVocabulary* tset)const{
    return this->put(out,ntset);
  };

  //! input
  friend istream& operator >> (istream& in, Nonterminal& obj);

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    ar & m_synCat;
  };

  //! reads a nonterminal from input and returns a pointer to a
  // newly instantiated Nonterminal of the appropriate type.
  static Nonterminal* readNonterminal(istream& in);

  virtual Nonterminal *clone() const;

  virtual istream& get(istream& in);

  //! = operator
  virtual void operator =(const SynCat cat);

  //! < operator
  virtual bool operator <(const Nonterminal& other)const;

  //! < operator
  virtual bool operator <(const Nonterminal* other)const;

  //! == operator
  virtual bool operator ==(const Nonterminal& other)const;
  //! == operator
  virtual bool operator ==(const Nonterminal* other)const;

  virtual int cmp(const Nonterminal& other) const;

  virtual int cmp(const Nonterminal* other) const;

  //! return a delexicalized version of this NTLink
  virtual Nonterminal* delex() const{
    return new Nonterminal(*this);
  };

  virtual size_t hash() const;

  virtual vector<unsigned short> vectorDecompose()const;

  virtual void setHeirMark(){}

  virtual void setNull();

  virtual bool isNull() const;

  virtual bool isDummy() const;

protected:

  //Syncactic Category, (present in all types of Nonterminals)
  SynCat m_synCat;

};

namespace std {
  template<>
  struct less<Nonterminal*> {
    bool operator()(const Nonterminal* p1, const Nonterminal* p2) const {
      return (*p1 < *p2);
    }
  };
};

namespace HashNS {
  template<> struct hash<Nonterminal> {
    size_t operator()(const Nonterminal& obj) const {
      return obj.hash();
    }
  };
}

#endif

