#include "OptimalEstimator.hpp"

#include "OptionsPrinter.hpp"
#include "Path.hpp"
#include "GrammarBuilder.hpp"
#include "Chart.hpp"
#include "HeadProduction.hpp"

OptimalEstimator::OptimalEstimator(const string& configFile) 
    : _super(configFile) {

  string parserFile;
  string grammarFile;
  unsigned int debugLevel;

  // compile the config file.
  BPO::options_description config("Configuration");
  config.add_options()
      ("ParserConfig", BPO::value<string>(&parserFile),
       "The config file for the parser.")
      ("Debug", BPO::value<unsigned int>(&debugLevel)->default_value(debuglevel()),
       "The debug level for this instance")
      ;
  BPO::variables_map vm;
  cerr << "DimSplitEstimator: Compiling config file "
       << configFile << endl;
  iFile ifs(configFile.c_str());
  store(parse_config_file(ifs, config), vm);
  notify(vm);

  if(debugLevel != debuglevel())
    debugme(debugLevel);

  if(debug(1)) {
    cerr << "Option values of \"DimSplitEstimator\"" << endl << endl;
    cerr << OptionsPrinter(config, vm) << endl;
  }

  if(parserFile == "") {
    THROW("parser config file not given");
  }
  m_parserConfigFile = Path::maybe_compose(configFile, parserFile);
}

OptimalEstimator::~OptimalEstimator() {
  // The grammar will be deleted somewhere else 
  // since it's a pointer from the main parser
  delete m_parser;
}

void OptimalEstimator::precompute(NaiveWMTG *grammar) {
  m_grammar = grammar;

  // Create a parser
  Parser *p = new Parser(m_parserConfigFile, grammar);
  if(!p)
    THROW("Parser could not be created");

  m_parser = p;
  m_score = 0.0;
}

void OptimalEstimator::initialize(SentenceTuple &sent) {
  _super::initialize(sent);
  m_score = 0.0;
  m_parser->clear();
  WTreeVec tree = m_parser->getParse(sent);
  for(WTreeVec::iterator itr=tree.begin(); 
      itr!=tree.end(); itr++) {
    cout << "adding: " << itr->second << endl;
    cout << *(itr->first) << endl;
    m_score += itr->second;
  }
  if(tree.size() == 0) {
    cerr << "could not get parse for " << sent << endl;
  }
}

SCORE OptimalEstimator::operator()(SentenceTuple &sent, const Item *item) {
  const Hedge *hedge = dynamic_cast<const Hedge*>(item);
  if(!hedge) {
    THROW("item must be a Hedge");
  }

  // Find the item in the chart whose span matches the given Hedge
  BSP<Chart> chart = m_parser->getLogic()->getChart();
  // The first score is the inside score, the second is the outside score
  SCORE itemScore = (chart->getIOScores(hedge)).first;
  SCORE cost = m_score - itemScore;
  if(cost < 0)
    cost = 0.0;

  return cost;
}

void OptimalEstimator::reset() {
  m_parser->clear();
}

