#include "OutsideCostEstimatorBuilder.hpp"
#include "Exceptions.hpp"

#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"
#include "File.hpp"
#include "Path.hpp"

#include "LengthBasedEstimator.hpp"
#include "LexLinkBasedEstimator.hpp"
#include "DimSplitEstimator.hpp"
#include "OptimalEstimator.hpp"

OutsideCostEstimatorBuilder* OutsideCostEstimatorBuilder::_instance = 0;

OutsideCostEstimatorBuilder* OutsideCostEstimatorBuilder::Instance(){
  if(_instance == 0){
    _instance = new OutsideCostEstimatorBuilder;
  }
  return _instance;
}

// Builds the outside cost estimator. And if the data file
// is specified, load it in.
OutsideCostEstimator* OutsideCostEstimatorBuilder::build(const string& configFile,
							 Grammar* pGrm) {
  string estimateType;
  string oceConfigFile;

  // compile the config file.
  BPO::options_description config("Configuration");
  config.add_options()
      ("EstimatorType", BPO::value<string>(&estimateType),
       "The type of the outside cost estimator.")
      ("ConfigFile", BPO::value<string>(&oceConfigFile),
       "The data file for the outside cost estimator.")
      ;
  BPO::variables_map vm;
  cerr << "OutsideCostEstimatorBuilder: Compiling config file "
       << configFile << endl;
  iFile ifs(configFile.c_str());
  store(parse_config_file(ifs, config), vm);
  notify(vm);

  if(debug(1)) {
      cerr << "Option values of \"OutsideCostEstimatorBuilder\""
	   << endl << endl;
      cerr << OptionsPrinter(config, vm) << endl;
  }

  if(estimateType == "") {
    THROW("Error: Please specify the outside cost estimate!"<<endl);
  }
  if(oceConfigFile.length() == 0) {
    THROW("Please specify a data file!"<<endl);
  }

  estimateType = tolower(estimateType);
  oceConfigFile = Path::maybe_compose(configFile, oceConfigFile);

  OutsideCostEstimator * oce;

  if(estimateType== "length"){
    oce = new LengthBasedEstimator(oceConfigFile);
  } else if(estimateType == "lexlink"){
    oce = new LexLinkBasedEstimator(oceConfigFile);
  } else if(estimateType == "dimsplit") {
    oce = new DimSplitEstimator(oceConfigFile);
  } else if(estimateType == "optimal") {
    oce = new OptimalEstimator(oceConfigFile);
  } else {
    oce = NULL;
    THROW("OCEBuilder: unknown estimate type!"<<endl);
  }

  return oce;
}

