#include <map>
#include <set>
#include "PAV.hpp"
#include "DSV.hpp"
#include "Exceptions.hpp"
#include "MoreComparators.hpp"
#include "GPDefs.hpp"

using namespace std;

Vocabulary<RawPAV> PAV::sm_voc;

PAV::PAV()
  : m_index(sm_voc.addWord(_super()))
{}

PAV::PAV(const arity_t& dim)
  : m_index(sm_voc.addWord(_super(dim)))
{}

//! constructor from one PrecArray
PAV::PAV(const PrecArray& pa)
  : m_index(sm_voc.addWord(_super(pa)))
{}

//! constructor from a RawPAV
PAV::PAV(const RawPAV& raw)
  : m_index(sm_voc.addWord(_super(raw)))
{}

//! constructor from a Vector<PrecArray>
PAV::PAV(const Vector<PrecArray>& pavec)
  : m_index(sm_voc.addWord(_super(pavec)))
{}

//! constructor from an arity, a PrecArray, and the position of the PrecArray
//! the other components will be empty
PAV::PAV(const arity_t& dim, const PrecArray& pa, const arity_t& psn)
  : m_index(sm_voc.addWord(_super(dim, pa, psn)))
{}

//! constructor from a fan-out vector
//! all elements will have the same role
PAV::PAV(const FanoutVect& fov)
  : m_index(sm_voc.addWord(_super(fov)))
{}

//! constructor from a single DSV
//! all elements will have the same role
PAV::PAV(const DSV& dsv)
  : m_index(sm_voc.addWord(_super(dsv)))
{}

PAV::PAV(const DSV& dsv1, const DSV& dsv2)
  : m_index(sm_voc.addWord(_super(dsv1, dsv2)))
{}

PAV::PAV(const set<DSV >& dsvset)
  : m_index(sm_voc.addWord(_super(dsvset)))
{}

//! Equality test.
bool PAV::operator == (const PAV& other) const {
  return (m_index == other.m_index);
}

//! Less than test.
bool PAV::operator < (const PAV& other) const {
  return (m_index < other.m_index);
}

//! Greater than test.
bool PAV::operator > (const PAV& other) const {
  return (m_index > other.m_index);
}

//! Assignment.
PAV& PAV::operator=(const PAV& other) {
  m_index = other.m_index;
  return *this;
}

// //! indexing
// PrecArray& PAV::operator[] (const arity_t index) {
// return static_cast<PrecArray&>(_super::operator[](index));
// }

const PrecArray& PAV::getPA(const arity_t dim) const {
  return sm_voc.getWord(m_index)->getPA(dim);
}
// const PrecArray& PAV::operator[] (const arity_t index) const {
// return static_cast<const PrecArray&>(_super::operator[](index));
// }

arity_t PAV::arity() const {
  return sm_voc.getWord(m_index)->arity();
}

// void PAV::clear()
// {
// for(arity_t d = 0; d < arity(); d++) (*this)[d].clear();
// }

bool PAV::empty() const {
  return sm_voc.getWord(m_index)->empty();
}

bool PAV::hasGap() const {
  return sm_voc.getWord(m_index)->hasGap();
}

bool PAV::hasGap(arity_t dim) const {
  return sm_voc.getWord(m_index)->hasGap(dim);
}

bool PAV::monotonic(arity_t d) const {
  return sm_voc.getWord(m_index)->monotonic(d);
}

size_t PAV::hash() const {
  return m_index;
}

// the fanout of one dimension
unsigned short PAV::getFanout(const arity_t dim) const {
  return sm_voc.getWord(m_index)->getFanout(dim);
};

// Returns a fanout vector for one role.
FanoutVect PAV::getFanoutV(const role_t role) const {
  return sm_voc.getWord(m_index)->getFanoutV(role);
}

// Returns a fanout vector of the entire PAV.
FanoutVect PAV::getFanoutV() const {
  return sm_voc.getWord(m_index)->getFanoutV();
}

istream& operator>>(istream& in, PAV& pav) {
  RawPAV raw;
  in >> raw;
  pav.m_index = PAV::sm_voc.addWord(raw);
  return in;
}

ostream& operator<<(ostream& out, const PAV& pav) {
  out << *(PAV::sm_voc.getWord(pav.m_index));
  return out;
}

ostream& PAV::serialize(ostream&out) const {
  THROW("PAV::serialize not implemented");

  return out;
}

//! Returns the fanout (scalar) for the whole PAV
//! This is very different from the getFanoutV method for one role!
short PAV::fanout() const {
  return sm_voc.getWord(m_index)->fanout();
}

int PAV::cmp(const PAV& other) const {
  return MC::cmp((int)m_index, (int)(other.m_index));
}

PAV PAV::swapElement(const PAV& orig) {
  RawPAV raw = _super::swapElement(*(sm_voc.getWord(orig.m_index)));
  PAV pav = PAV(raw);
  return pav;
}

//! reverse order of components
PAV PAV::reverse(const PAV& orig) {
  RawPAV raw = _super::reverse(*(sm_voc.getWord(orig.m_index)));
  return PAV(raw);
}

PAV PAV::normalize(const PAV& orig) {
  RawPAV raw = _super::normalize(*(sm_voc.getWord(orig.m_index)));
  return PAV(raw);
}

PAV PAV::deleteRoleAndNormalize(const PAV& orig, const role_t delRole) {
  RawPAV raw = _super::deleteRoleAndNormalize(*(sm_voc.getWord(orig.m_index)), delRole);
  return PAV(raw);
}

bool PAV::improperOrder() const {
  return sm_voc.getWord(m_index)->improperOrder();
};

//! Returns the set of roles in this PAV.
set<role_t> PAV::roleSet() const {
  return sm_voc.getWord(m_index)->roleSet();
}

/**
 * Return the maximum role. This PAV do not have to be normalized.
 */
role_t PAV::maxRole() const {
  return sm_voc.getWord(m_index)->maxRole();
}

short PAV::rank() const {
  return sm_voc.getWord(m_index)->rank();
}

//! Returns the number of roles on dimension d.
short PAV::rank(arity_t d) const {
  return sm_voc.getWord(m_index)->rank(d);
}

/** Generate all possible d-dimensional (first argument) PAVs with
    fanout (second argument) less than or equal to a limit. The
    fanout limit is also applied to each role. The generated PAVs
    are accumulated into a Vocabulary (third argument). */
void PAV::generate(const arity_t& dim, short fanoutLmt, Vocabulary<PAV>& ret) {
  Vocabulary<RawPAV> rawVocab;
  RawPAV::generate(dim, fanoutLmt, rawVocab);
  // convert from raw to flyweight
  Vocabulary<RawPAV>::const_iterator vci;
  for(vci = rawVocab.begin(); vci!= rawVocab.end(); ++vci) {
    PAV aPAV(vci->first);
    ret.addWord(aPAV);
  }
}

//Check if the PAV is in normal form, from the perspective of aRole
bool PAV::isNormalFormForRole(role_t aRole) const{
  return sm_voc.getWord(m_index)->isNormalFormForRole(aRole);
}

// returns true if condition is met.
//  @param vecRoles what role to look for in each dimension.
//  @param segment which continuous segment to look in (0 based)
//  @param startOrEnd 0:nonexistent 1:start 2: end 3:ignore
//  must meet ALL conditions in each dimension
bool PAV::endsBegWith(const vector<role_t>& vecRoles, const vector<unsigned int>& segment, const vector<unsigned int>& startOrEnd)const{
  return sm_voc.getWord(m_index)->endsBegWith(vecRoles,segment,startOrEnd);
}



