#ifndef _PAV_H_
#define _PAV_H_

#include <set>
#include <memory>
#include <map>
#include "Vector.hpp"
#include "Vocabulary.hpp"
#include "FanoutVect.hpp"
#include "RawPAV.hpp"
#include <boost/serialization/serialization.hpp>

class DSV;

//! \brief precedence array vector
/// This class is implemented using a combination of two design
/// patterns (see Gamma et al. (1994) _Design Patterns_). It is a
/// flyweight and simultaneosly a proxy. The flyweight factory *is*
/// the proxy. Like all flyweights, it uses a vocabulary to avoid
/// creating multiple copies of the same object. Each instance needs
/// to record only an index into the vocabulary. In the present case,
/// the vocabulary lives as a static object inside the flyweight
/// class. Thus, other classes don't know that this class is a
/// flyweight, which better encapsulates the implementation.
class PAV {
  typedef RawPAV _super;
public:

  //! empty constructor, necessary for serialization
  PAV();

  //! Constructor.
  PAV(const arity_t& dim);

  //! constructor from one PrecArray
  PAV(const PrecArray& pa);

  //! constructor from an arity, a PrecArray, and the position of the PrecArray
  //! the other components will be empty
  PAV(const arity_t& dim, const PrecArray& pa, const arity_t& psn);

  //! constructor from a Vector<PrecArray>
  PAV(const Vector<PrecArray>& pa);

  //! constructor from a RawPAV
  PAV(const RawPAV& raw);

  //! constructor from a fan-out vector
  //! all elements will have the same role
  PAV(const FanoutVect& fov);

  //! constructor from a single DSV
  //! all elements will have the same role
  PAV(const DSV& dsv);

  //! constructor from two DSpans.
  PAV(const DSV& span1, const DSV& span2);

  //! constructor from set of DSpans.
  PAV(const set<DSV>& spans);

  //! Equality test.
  bool operator == (const PAV& other) const;

  //! Less than test.
  bool operator < (const PAV& other) const;

  //! Greater than test.
  bool operator > (const PAV& other) const;

  //! Assignment.
  PAV& operator=(const PAV& other);

  //! indexing
  // const PrecArray& operator[] (const arity_t index) const;
  const PrecArray& getPA(const arity_t index) const;

  arity_t arity() const;

  //! Judges whether empty.
  bool empty() const;

  //! Judges whether there is a gap in at least one dimension.
  bool hasGap() const;

  //! Judges whether there is a gap in the given dimension.
  bool hasGap(arity_t dim) const;

  //! are the precedences monotonic in component d?
  bool monotonic(arity_t d) const;

  //! For hashing.
  size_t hash() const;

  //! fanout
  short fanout() const;

  // the fanout of one dimension
  unsigned short getFanout(const arity_t dim) const;

  //! Returns a fanout vector for one role
  FanoutVect getFanoutV(const role_t theRole) const;

  //! Returns the fanout vector of the entire PAV.
  FanoutVect getFanoutV() const;

  //! For ordering.
  int cmp(const PAV& other) const;

  ///////////////////////////////////////////////////////////////////////////
  //////// The following methods are now const instead of destructive ///////
  //! Remove one role, and normalize the PAV.
  static PAV deleteRoleAndNormalize(const PAV& orig, const role_t i);

  //! reverse order of components
  static PAV reverse(const PAV& orig);

  //! Exchanges the value of elements: 1->2, 2->1, and 0 remained.
  //! Currently only works for binary case.
  static PAV swapElement(const PAV& orig);

  //! Normalizes the PAV.
  static PAV normalize(const PAV& orig);

  ///////////////////////////////////////////////////////////////////////////

  //! Judges whether the indices are not increasing order in the
  //! first dimension. <1,2> -- false; <2,1> -- true.
  bool improperOrder() const;

  //! Returns the set of roles in this PAV.
  set<role_t> roleSet() const;

  //! The number of different roles
  short rank() const;

  //! Returns the number of roles on dimension d.
  short rank(arity_t d) const;

  /**
   * Return the maximum role. This PAV do not have to be normalized.
   */
  role_t maxRole() const;

  //! Decides if PAV is in normal form, for a given role (typically,
  //! you pass the role that is supposed to be the head)
  bool isNormalFormForRole(role_t aRole) const;

  //! returns true if condition is met.
  //!  @param vecRoles what role to look for in each dimension.
  //!  @param segment which continuous segment to look in (0 based)
  //!  @param startOrEnd 0:nonexistent 1:start 2: end 3:ignore
  //!  must meet ALL conditions in each dimension
  bool endsBegWith(const vector<role_t>& vecRoles, const vector<unsigned int>& segment, const vector<unsigned int>& startOrEnd)const;
  
  //! Generate all possible d-dimensional (first argument) PAVs with
  //! fanout (second argument) less than a limit.
  //! The fanout limit is also applied to each role.
  //! The generated PAVs are accumulated into a Vocabulary (third argument).
  static void generate(const arity_t&, short, Vocabulary<PAV>&);

  istream& get(istream& in);

  //! Inputs PAV.
  friend istream& operator>>(istream& in, PAV& pav);

  //! Outputs PAV.
  friend ostream& operator<<(ostream& out, const PAV& pav);

  ostream& serialize(ostream&out) const;

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    ar & m_index;
    ar & sm_voc;
  }

  //! return index for PAV
  VocabularyIndex  getIndex()const{
    return m_index;
  }
  
  //! return Vocab of RAW PAVs
  static Vocabulary<RawPAV>  getVocab(){
    return sm_voc;
  }

private: // data

  // currently an unsigned int,
  // but can probably be changed to unsigned short
  VocabularyIndex m_index;

  // currently hashed in both directions,
  // but one dir should be a vector to be faster
  static Vocabulary< RawPAV > sm_voc;

};

namespace HashNS{
  template<>
  struct hash<PAV > {
    size_t operator()(const PAV& pav) const { return pav.hash();}
  };
}

#endif
