#include "File.hpp"
#include "Parser.hpp"
#include "Path.hpp"

#include "SearchStrategyBuilder.hpp"
#include "LogicBuilder.hpp"
#include "TerminationTest.hpp"

#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

// Constructor with a grammar and a configure file.
Parser::Parser(const string& configFile, 
	       Grammar* pGrammar,
	       BSP<BPO::options_description> optionsDescription)
  : m_pLogic(NULL), m_pSearchStrategy(NULL), m_pGrammar(pGrammar),
    m_pInfStream(NULL), m_inputNumber(0), 
    m_multiPassParse(false), m_printStats(true), m_printParseTime(false) {
  unsigned int dim;
  string optLogicBuilderConfigFile;
  string optSearchConfigFile;
  string optInferenceFile;
  string className = "Generalized Parser";

  /*
   * Compile the config file.
   */
  addOptionDescriptionsTo(optionsDescription);

  BPO::variables_map vm;
  cerr << "Parser: Compiling config file " << configFile << endl;
  iFile ifs(configFile.c_str());
  store(parse_config_file(ifs, *optionsDescription), vm);
  notify(vm);

  if(debug(1)) {
      cerr << "Option values of \"Parser\"" << endl << endl;
      cerr << OptionsPrinter(*optionsDescription, vm) << endl;
  }

  if(vm.count("D"))
    dim = vm["D"].as<unsigned int>();
  if(vm.count("MultiPassParsing"))
    m_multiPassParse = vm["MultiPassParsing"].as<bool>();
  if(vm.count("InferenceFile")) {
    optInferenceFile = vm["InferenceFile"].as<string>();
    m_pInfStream = new oFile(optInferenceFile);
  }
  if(vm.count("LogicBuilderConfigFile"))
    optLogicBuilderConfigFile = vm["LogicBuilderConfigFile"].as<string>();
  if(vm.count("SearchStrategyConfigFile"))
    optSearchConfigFile = vm["SearchStrategyConfigFile"].as<string>();
  if(vm.count("PrintStats"))
    m_printStats = vm["PrintStats"].as<bool>();
  if(vm.count("PrintParseTime"))
    m_printParseTime = vm["PrintParseTime"].as<bool>();

  // build the Logic
  string logicBuilderConfigFile = Path::maybe_compose(configFile, optLogicBuilderConfigFile);
  m_pLogic = LogicBuilder::Instance()->build(logicBuilderConfigFile);
  if(!m_pLogic){
    cerr<<"Warning: No Logic was built."<<endl;
  }

  // build the SearchStrategy
  string searchConfigFile = Path::maybe_compose(configFile, optSearchConfigFile);
  m_pSearchStrategy = SearchStrategyBuilder::Instance()->build(searchConfigFile);
  if(!m_pSearchStrategy){
    cerr<<"Warning: No SearchStrategy is built."<<endl;
  }

  // inform the logic, search strategy, etc., about the grammar
  setGrammar(pGrammar);
}

Parser::~Parser() {
  delete m_pLogic;
  delete m_pSearchStrategy;
  delete m_parseStats.getRecursive();
}

//returns a pointer to the grammar
Grammar* Parser::getGrammar() const {
  return m_pGrammar;
}

//sets the grammar pointer for the parser and the associated logic.
void Parser::setGrammar(Grammar* pGrammar) {
  if (pGrammar) {
    m_pGrammar = pGrammar;
    m_pLogic->setGrammar(pGrammar);
    m_pSearchStrategy->setGrammar(pGrammar);
    if (debug(6))
      cerr << "Parser setting the following grammar:" << endl << *pGrammar;
  }
}

//compute the parse and return the resulting vector of tree,value pairs
WTreeVec Parser:: getParse(SentenceTuple& sntT) {

  initialize(sntT);
  return getParse();
}

//compute the parse and return the resulting vector of tree,value pairs
WTreeVec Parser:: getParse() {

  if(m_multiPassParse){
    multiPassParse();
  } else {
    onePassParse();
  }

  if(m_printStats) {
    if(m_pLogic && debug(4)){
      m_pLogic->printStats(dout());
    }

    if (debug(1)) {
      dout() << m_parseStats;
      dout() << "Unique MATCHES PROPOSED: " << m_pLogic->getNumUniqueMatches() << endl;
    }
  }

  WTreeVec ret = exportParses();
  if (debug(3)) {
    cerr << "parse tree" << endl;
    if (ret.empty())
      cerr << " is empty." << endl;
    else {
      WTreeVec::const_iterator it;
      for(it = ret.begin(); it != ret.end(); ++it){
	if(it == ret.begin()){
	  cerr<<*it->first;
	} else {
	  cerr<<"\t"<<*it->first;
	}
      }
      cerr << endl;
    }
  }

  return ret;
}

WTreeVec Parser:: exportParses() {
  return m_pLogic->exportParses();
}

// Attempt to achieve the parsing goal but do not compute or return the tree
bool Parser:: silentParse(SentenceTuple& sntT) {
  initialize(sntT);

  bool goalAchieved;
  // calls either onePassParse or multiPassParse
  if(m_multiPassParse){
    goalAchieved = multiPassParse();
  } else {
    goalAchieved = onePassParse();
  }

  if(m_pLogic && debug(6)){
    m_pLogic->printStats(dout());
  }

  return goalAchieved;
}

//! Clears the parser's state. Note that this does not include the grammar.
void Parser::clear() {
  m_pLogic->clear();
  m_pSearchStrategy->clear();
}

//! Initializes the parser, e.g., seed the agenda.
void Parser :: initialize(SentenceTuple& sntT) {
  assert(m_pSearchStrategy && m_pLogic );

  ++m_inputNumber;

  // the following is presumed to also call Grammar->prepareFor(sntT)
  m_pLogic->initialize(sntT);

  // Also initialize the search strategy
  m_pSearchStrategy->initialize(sntT);

  m_parseStats.clear();
}

//! push items on the Agenda
size_t Parser::enqueue(SearchStrategy* ss, vector<Item*>& itemList) {
  for(vector<Item*>::const_iterator it = itemList.begin(); it!= itemList.end(); it++){  
    Item* derivedItem = (Item*)(*it);
    ss->addItem(derivedItem);
  }
  return m_pSearchStrategy->size();
}

/* One pass parsing until the goal is reached or the parsing
 fails. Use goal.trace() to see which sub-goal is achieved or whether
 the parsing fails. */
bool Parser:: onePassParse() {

  TerminationTest::GoalID goalID;
  TerminationTest* pParseGoal = m_pLogic->getGoal();
  // at the beginning, the goal is unachieved.
  vector<short> lens = (m_pLogic->snt()).lengths();
  DSV aDSV(lens);
  pParseGoal->reset(aDSV);
  pParseGoal->reset();
  int goalCount = 0;

  vector<Item*> itemList;
  MemoResult memoRes;

  // the item popped off the agenda, possibly empty
  Item* pTrigger = NULL;

  Timer timer;
  timer.start();
  do {
    if( ! m_pSearchStrategy->empty()) {
      // Pop an item from agenda;
      pTrigger = m_pSearchStrategy->nextItem();
      m_parseStats.numPopInf++;

      // if the inference stream is set, then dump the trigger's inference
      if (m_pInfStream) {
	*m_pInfStream << m_inputNumber << " ";
	BSP<Inference> spInf = pTrigger->getInference();
	*m_pInfStream << spInf << endl;
      }
    }

    // check if the trigger achieves the goal
    if (pParseGoal->achieved(pTrigger, m_pLogic, goalID))
      ++goalCount;

    // expand the trigger, generating new items
    memoRes = m_pLogic->expand(pTrigger, itemList);
    if (memoRes == IGNORE) // trigger control was not given to shared pointer
      delete pTrigger;
    
    //increment inference counter
    m_parseStats.numGenInf += itemList.size();

    //push items onto the agenda
    size_t aSize = enqueue(m_pSearchStrategy, itemList);
    m_parseStats.maxAgdSize = (m_parseStats.maxAgdSize > aSize)
      ? m_parseStats.maxAgdSize : aSize;

  } while(!m_pSearchStrategy->empty() && goalCount == 0);

  timer.end();
  if(m_printParseTime)
    m_parseStats.elapsedSec = timer.diff();
  else
    m_parseStats.elapsedSec = 0.0;
  m_parseStats.numMatches = m_pLogic->getNumMatches();
  //  delete pTrigger;
  return (goalCount > 0);
}

bool Parser::multiPassParse() {
  
  THROW("multiPassParsing currently broken, " 
	<< "because we've done away with item reference counting.  "
	<< "If we ever want to resurrect it, "
	<< "then it should be done using shared pointers.");

//   // parse based on the current setting and with the complete
//   // parse as the goal (of course, it might fail).
//   TerminationTest* goalptr = m_pLogic->getGoal();
//   while(1){

//     // at the beginning, the goal is unachieved.
//     // reset the goal such that no goal is reached.
//     goalptr->reset();

//     // backup the counters before each parsing..
//     ParseStats statsBk = m_parseStats;
//     // reset the statistics.
//     m_parseStats.clear();

//     onePassParse();

//     bool parsingFail = false;
//     //if(m_parseStats.numGenInf == 0) { parsingFail = true;}
//     if(m_parseStats.numGenInf == m_parseStats.filtered)
//       { parsingFail = true;}

//     if(debug(3)){
//       cerr<<"STATS:"<<m_parseStats.numGenInf<<"| "
// 	  <<m_parseStats.filtered<<endl;
//     }

//     // revise the new counters.
//     m_parseStats += statsBk;

//     // if (i) the goal is reached, or (ii) there are no items
//     // generated in the previous parsing, or (iii) it is not
//     // multi-pass parsing, we break the loop.
//     if(goalptr->achieved() || parsingFail || !m_multiPassParse)
//       {break;}

//     // clear the search strategy to be ready for the next parsing.
//     m_pSearchStrategy->clear();

//     //we reparse based on the current logic status.
//     //To do so, we first greedily extract the set of items
//     //whose spans don't overlap, and cover the entire sentence tuple.
//     vector<Item*> items;
//     if(debug(3)){ cerr<<"EXTRACTING MAXIMUM COVER"<<endl; }
//     m_pLogic->exportMaxCover(items);
//     // push the seeds onto the agenda.
//     vector<Item*>::const_iterator it;
//     if(debug(3)){
//       cerr<<"The current pruning th: "<<m_pLogic->pruner()->getCurTh()<<endl;
//     }

//     for(it = items.begin(); it != items.end(); ++it){
//       Inference* inf = (*it)->getInference();
//       // increase the number of reference times.
//       (*it)->numRef()++;
//       m_pSearchStrategy->addItem(*it);
//       if(debug(3)){
// 	inf->dump(cerr); cerr<<endl;
//       }
//     }
//     if(debug(3)){ cerr<<"END EXTRACTING MAXIMUM COVER"<<endl; }

//     // backup the working chart and clear it.
//     m_pLogic->saveCharts();

//     // lower the threshold of the parsing.
//     if(m_pLogic->pruner()) { m_pLogic->pruner()->lowerTh(); }
//   }

//   return goalptr->achieved();
}

/** Adds this class' option description to the option description l
ist of the hierarchy */
BSP<BPO::options_description> 
Parser::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription) {
  optionsDescription->add_options()
    ("D", BPO::value<unsigned int>(), "The dimension of the parser.")
    ("PrintStats", BPO::value<bool>(), "print parse stats?")
    ("PrintParseTime", BPO::value<bool>(), "print parse timing info?")
    ("MultiPassParsing", BPO::value<bool>(), "Use multi-pass parsing.")
    ("InferenceFile", BPO::value<string>(), "file in which to dump inferences")
    ("LogicBuilderConfigFile", BPO::value<string>(), "The configure file of the Logic.")
    ("SearchStrategyConfigFile", BPO::value<string>(), 
     "The configure file of the Search Strategy.")
    ;

  return optionsDescription;
}
