//! \brief The generalized parser.
/** Most applications will need to call only the constructor and the
getParse() method. Other methods are there for special purposes.
*/

#ifndef _Parser_H_
#define _Parser_H_

#include "AbstractParser.hpp"
#include "ParseStats.hpp"
#include "Timer.hpp"
#include "Grammar.hpp"
#include "Logic.hpp"
#include "Exceptions.hpp"
#include "SearchStrategy.hpp"

class Logic;

using namespace std;

/// The generalized parser.
class Parser : public AbstractParser {
    typedef AbstractParser _super;

private:

  //! add options from the configFile
  BSP<BPO::options_description> 
  addOptionDescriptionsTo(BSP<BPO::options_description> priorOptions);

public:

  // constructor from a config file
  Parser(const string& configFile, 
	 Grammar* pGrammar = NULL,
	 BSP<BPO::options_description> priorOptions =
	 BSP<BPO::options_description>(new BPO::options_description()));

  // the grammar is not deleted. You need to delete the
  // grammar outside Parser.
  virtual ~Parser();

  //! compute the parse and return the resulting vector of tree/value pairs
  virtual WTreeVec getParse(SentenceTuple&);

  //! This overloaded version reparses the currently initialized SentenceTuple.
  //! tThis could be useful if, e.g., the grammar is updated.
  virtual WTreeVec getParse();

  /// send back the resulting parses, e.g. after a silentParse()
  virtual WTreeVec exportParses();

  //! Attempt to achieve the parsing goal but do not compute or return the tree.
  // This is useful in cases where some parent process needs info
  // about some subparses, but not necessarily the whole tree.
  // \return bool whether the goal has been reached
  virtual bool silentParse(SentenceTuple&);

  //! Clears the parser by clearing the contents of its components.
  //! The components are still there, not being deleted.
  //! The grammar is not cleared.
  virtual void clear();

  ////////////////////////////////////////////////////////////////////
  // configure the parser.
  ///////////////////////////////////////////////////////////////////

  //! Gets a reference to the logic
  virtual const Logic* getLogic() {return m_pLogic;}

  //! Set the model for the parser
  virtual void setGrammar(Grammar* pGrammar);

  //! get the model for the parser
  virtual Grammar* getGrammar() const;

  //! Set whether to print the parse stats
  void setPrintStats(bool flag) {m_printStats = flag;}

 protected: // methods

  // no default constructor from outside
  Parser() {}

  /// Initializes the parser for new input.
  virtual void initialize(SentenceTuple&);

  /// One pass parsing until the goal is reached or the parsing
  /// fails. Use goal.trace() to see which sub-goal is achieved
  /// or whether the parsing fails.
  /// Whenever the agenda is empty, the parser restarts!!!
  virtual bool onePassParse();

  /** Parse, using multiple passes if necessary until the parsing goal
  // is found. Note that if the multiPassParse in the constructor is
  // false, then it is just ordinary parsing. Use the goal.trace() to
  // see which sub-goal is achieved or whether the parsing fails.
  // Whenever the agenda is empty, the parser restarts!!!
  */
  /* multiPassParse is a backoff technique that is used when parser fails
     to find a parse tree with sufficiently high probability within a
     sufficiently short time period. It involves making use of
     partial parses for the sentence tuple and backing off to a
     simpler smoothing model.
  */
  virtual bool multiPassParse();

  //! add items to the open list
  size_t enqueue(SearchStrategy* ss, vector<Item*>& itemList);

 private:

  /// This method kept private to avoid use/redefinition.
  void parse() {
    THROW("Please use either getParse() or silentParse() to avoid confusion." << endl);
  }

 protected: // components

  //! The reference to the Logic.
  Logic* m_pLogic;

  //! The reference to the SearchStrategy.
  SearchStrategy* m_pSearchStrategy;

  //! The reference to the grammar
  Grammar* m_pGrammar;

  /// output file handle for dumping inferences, if desired
  oFile* m_pInfStream;

  /// the number, starting from 1, of the input we're currently working on
  unsigned int m_inputNumber;

  // diagnostic info
  ParseStats m_parseStats;

 protected: // configuration variables

  //! if true, then we use multi-pass parsing
  bool m_multiPassParse;

  //! if true, then print out the parse statistics
  bool m_printStats;

  //! If true parsing time for each tuple is printed
  bool m_printParseTime;
};

#endif
