#include "PartialWMTG.hpp"

#include <algorithm>

#include "GPDefs.hpp"
#include "Pair.hpp"
#include "Exceptions.hpp"
#include "MoreExceptions.hpp"
#include "NTLink.hpp"
#include "LexNT.hpp"
#include "PAV.hpp"
#include "HeadMTree.hpp"
#include "HiddenDSVs.hpp"

#include "Item.hpp"
#include "NodeRefHedge.hpp"
#include "GlobalVocab.hpp"
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

PartialWMTG::
PartialWMTG(arity_t dm, const string& configFile,
	 BSP<BPO::options_description> optionsDescription)
  : _grammar(dm), 
    _super(dm, configFile, addOptionDescriptionsTo(optionsDescription)),
   m_rightAttach(false), m_pCurrSent(NULL) {

  /*
   * Compile the config file.
   */
  BPO::variables_map vm;
  cerr << "PartialWMTG: Compiling config file " << configFile << endl;
  iFile ifs(configFile.c_str());
  store(parse_config_file(ifs, *optionsDescription), vm);
  notify(vm);

  if(debug(1)) {
    cerr << "Option values of \"PartialWMTG\"" << endl << endl;
    cerr << OptionsPrinter(*optionsDescription, vm) << endl;
  }

  if(vm.count("TreeDim"))
    m_constrainedDim = vm["TreeDim"].as<unsigned int>();
  if(vm.count("RightAttach"))
    m_rightAttach = vm["RightAttach"].as<bool>();
}

PartialWMTG:: ~PartialWMTG(){}

// Updates the content of the grammar.
bool PartialWMTG::prepareFor(SentenceTuple& tuple) {
  // remember the current sentence
  m_pCurrSent = &tuple[m_constrainedDim];

  return true;
}

void PartialWMTG::getPossibleCons(const Item* item1, const Item* item2,
			       const Pair<bool,PAV>& arrg,
			       vector<Item*>& possConseqs) {
  //NTlink for consequent to be returned
  NTLink outLabel(1);
  NTLink outLabel2(1);
  //HeirVector for consequent to be returned
  HeirRoleVector finalHRV(1);
  //DSV for consequent to be returned
  DSV outDSV(1);

  // get information from item1
  const NodeRefHedge * nrh1 = static_cast<const NodeRefHedge*>(item1);
  const DSV * pDSV1 = nrh1->getDSV();
  const DSpan& inSpan1 = (*pDSV1)[m_constrainedDim];
  //  PositionTree::iterator iPT1 = NULL;
  NTLink label1 = nrh1->getNTLink();

  // get information from item2
  const NodeRefHedge * nrh2 = static_cast<const NodeRefHedge*>(item2);
  const DSV * pDSV2 = nrh2->getDSV();
  const DSpan& inSpan2 = (*pDSV2)[m_constrainedDim];
  //  PositionTree::iterator iPT2 = NULL;
  NTLink label2 = nrh2->getNTLink();

  if(debug(3)){
    static int invocationNumber = 0;
    ++invocationNumber;
    cerr << "PartialWMTG: getPossibleCons " << invocationNumber << " for "
	 << label1 << " and " << label2 << endl;
    cerr << "ANT1: " << *nrh1 << "ANT2: " << *nrh2 << endl;
  }

  //Lowest subsuming node not used by PartialWMTG
  //PositionTree::iterator lsnIt;

  //span is sum of two spans
  DSV combinedDSV(*pDSV1 + *pDSV2);
  if (combinedDSV.size() > m_constrainedDim)
    outDSV[0] = combinedDSV[m_constrainedDim];

  if(outDSV[0].length() >= 1)
    //since we have no way of knowing the heir, we simply mark it as unknown
    // and let the composite grammar decide
    finalHRV.setRole(0,-1);
  else
    finalHRV.setRole(0,-1);

  outLabel.setNT(0,label1.getNT(m_constrainedDim)->clone());

  /////////////////
  //
  // This section has a first cut approach to implementing
  // Binarization techniques 
  //
  ////////////////

  //this is a filter for empty components. 
  // you can only attach an 1D item to the "left" 
  if(m_rightAttach){
    if(pDSV1->empty(1-m_constrainedDim) and 
       !pDSV2->empty(m_constrainedDim) ){
      bool found = false;
      for(unsigned int i = 0; i < inSpan1.size(); i++){
	if(
	   inSpan1[i].second
	   == 
	   (inSpan2[0].first)){
	  found = true;
	  break;
	}
	// if the 1D item is not being attach directly to the left
	// filter
	if(!found){
	  if(debug(3)){
	    cerr << "right attach filter: ";
	    item1->dump(cerr); item2->dump(cerr);
	    cerr << endl;
	  }
	  return;
	}
      }
    }
    else if(pDSV2->empty(1-m_constrainedDim) and
	    !pDSV1->empty(m_constrainedDim) ){
      bool found = false;
      for(unsigned int i = 0; i < inSpan2.size(); i++){
	if((inSpan2[i]).second == (inSpan1[0].first))
	  found = true;
	break;
      }
      if(!found){
	if(debug(3)){
	  cerr << "right attach filter2: ";
	  item1->dump(cerr); item2->dump(cerr);
	  cerr << endl;
	}
	return;
      }
    }
  }//end 1D filter/////////////////////////////

  //only delex if its not a WHOLEDIM production
  // this check could be done more efficiently than adding...
  bool wholeDim = false;
  if(combinedDSV.size() > 1
     && combinedDSV[0].size() != 0 
     && combinedDSV[1].size() != 0){
    outLabel.delex();
  } else if(combinedDSV[m_constrainedDim].size() > 0){
    return;
  }else{
    wholeDim = true;
  }

  // If the input DSpans are empty on the contrained component, and
  // this grammar is being used as a supervising grammar for learning,
  // then we want to update the item's hiddenDSVs structure.  If there
  // are two antecedent hiddenDSVs and their sum is incompatible with
  // the input PAV on the constrained component, then the antecedents
  // cannot compose, so we abort.
  Vector<DSV>* pOutHDSVs = NULL;
  if (inSpan1.empty() and inSpan2.empty()) {
    PrecArray inPA = arrg.second.getPA(m_constrainedDim);
    bool compatible = hDSV::crossProdWithPA(nrh1->getHiddenDSVs(), 
					    nrh2->getHiddenDSVs(), 
					    m_constrainedDim, inPA, pOutHDSVs);
    if (not compatible)
      return; // without returning any possConseqs
  }

  //vector with one (dummy) position tree pointer in it
  vector<PositionTree::iterator> vecIt(1);

  vector<short> outHeadPosns(1);
  vector<short> outHeadPosns2(1);
  // pick the rightmost;
  // this also ensures that we don't pick 0 when there's a real span somewhere
  outHeadPosns[0] = nrh1->getHeads()[m_constrainedDim];

  //construct hedge item
  Item* outItem = new NodeRefHedge(outLabel, outDSV, outHeadPosns, vecIt);
  if (pOutHDSVs)
      outItem->setHiddenDSVs(pOutHDSVs);
  

  bool itemMade = false;
  //insert item and return
  if(nrh1->getHeads()[m_constrainedDim] != 0 && nrh1->getHeads()[1-m_constrainedDim] != 0){
  if(debug(5)){
    bool tmp = GlobalVocab::s_noVocabsOut;
    GlobalVocab::s_noVocabsOut = false;
    cerr << "PartialWMTG adding NodeRefHedge(a): " << *((NodeRefHedge*)outItem) << endl;
    GlobalVocab::s_noVocabsOut = tmp;
  }
    itemMade = true;
    possConseqs.push_back(outItem);
  }

  outLabel2.setNT(0,label2.getNT(m_constrainedDim)->clone());
  
  if(!wholeDim)
    outLabel2.delex();

  outHeadPosns2[0] = nrh2->getHeads()[m_constrainedDim];
  Item* outItem2 = new NodeRefHedge(outLabel2, outDSV, outHeadPosns2, vecIt);
  if(nrh2->getHeads()[m_constrainedDim] != 0 && nrh2->getHeads()[1-m_constrainedDim] != 0){
    itemMade = true;
    if(debug(5)){
      bool tmp = GlobalVocab::s_noVocabsOut;
      GlobalVocab::s_noVocabsOut = false;
      cerr << "PartialWMTG adding NodeRefHedge(b): " << *((NodeRefHedge*)outItem2) << endl;
      GlobalVocab::s_noVocabsOut = tmp;
    }
    possConseqs.push_back(outItem2);
  }
  
  if(!itemMade){
    if(wholeDim){
      if(debug(5)){
	bool tmp = GlobalVocab::s_noVocabsOut;
	GlobalVocab::s_noVocabsOut = false;
	cerr << "PartialWMTG adding NodeRefHedge(c): " << *((NodeRefHedge*)outItem) << endl;
	 cerr << "PartialWMTG adding NodeRefHedge(c): " << *((NodeRefHedge*)outItem2) << endl;
	GlobalVocab::s_noVocabsOut = tmp;
      }
      possConseqs.push_back(outItem);
      if(nrh1->getHeads()[m_constrainedDim] != 0)
	possConseqs.push_back(outItem2);
    }
    else{

      if(nrh1->getHeads()[m_constrainedDim] != 0){
	if(debug(5)){
	  bool tmp = GlobalVocab::s_noVocabsOut;
	  GlobalVocab::s_noVocabsOut = false;
	  cerr << "PartialWMTG adding NodeRefHedge(d): " << *((NodeRefHedge*)outItem) << endl;
	  GlobalVocab::s_noVocabsOut = tmp;
	}
	possConseqs.push_back(outItem);
      }
      else{
	if(debug(5)){
	  bool tmp = GlobalVocab::s_noVocabsOut;
	  GlobalVocab::s_noVocabsOut = false;
	  cerr << "PartialWMTG adding NodeRefHedge(e): " << *((NodeRefHedge*)outItem2) << endl;
	  GlobalVocab::s_noVocabsOut = tmp;
	}
	
	possConseqs.push_back(outItem2);
      }
    }
  }
}

void PartialWMTG::getPossibleConsHeaded(const Item* item1, const Item* item2,
			       const Pair<bool,PAV>& arrg,
			       vector<pair<HeirRoleVector,Item*> >& possConseqs) {
  //do we need to return multiple heads?
  bool needBoth = false;
  

  //if we are not ignoring heads, then the grammar will return an unknown head 
  // as a 0 is an heirRole, and this we do not need to generate both options
  //we only need Both if there is a 1D child, since there will be a w2w score.  
  // (otherwise, the head will not play a role)
  if(m_ignoreHeads && item1->getSizes()[m_constrainedDim] && item2->getSizes()[m_constrainedDim] && (!item1->getSizes()[1-m_constrainedDim] || !item2->getSizes()[1-m_constrainedDim])){
    needBoth = true;
  }

  //NTlink for consequent to be returned
  NTLink outLabel(1), outLabel2(1);
  //HeirVector for consequent to be returned
  HeirRoleVector finalHRV(1), finalHRV2(1);
  //DSV for consequent to be returned
  DSV outDSV(1);

  // get information from item1
  const NodeRefHedge * nrh1 = static_cast<const NodeRefHedge*>(item1);
  const DSV * pDSV1 = nrh1->getDSV();
  const DSpan& inSpan1 = (*pDSV1)[m_constrainedDim];
  //  PositionTree::iterator iPT1 = NULL;
  NTLink label1 = nrh1->getNTLink();

  // get information from item2
  const NodeRefHedge * nrh2 = static_cast<const NodeRefHedge*>(item2);
  const DSV * pDSV2 = nrh2->getDSV();
  const DSpan& inSpan2 = (*pDSV2)[m_constrainedDim];
  //  PositionTree::iterator iPT2 = NULL;
  NTLink label2 = nrh2->getNTLink();

  if(debug(3)){
    static int invocationNumber = 0;
    ++invocationNumber;
    cerr << "PartialWMTG: getPossibleCons " << invocationNumber << " for "
	 << label1 << " and " << label2 << endl;
    cerr << "ANT1: " << *nrh1 << "ANT2: " << *nrh2 << endl;
  }

  //Lowest subsuming node not used by PartialWMTG
  //PositionTree::iterator lsnIt;

  //span is sum of two spans
  DSV combinedDSV(*pDSV1 + *pDSV2);
  if (combinedDSV.size() > m_constrainedDim)
    outDSV[0] = combinedDSV[m_constrainedDim];

  
    //since we have no way of knowing the heir, we simply mark it as unknown
    // and let the composite grammar decide
  vector<short> outHeadPosns(1), outHeadPosns2(1);


  if(!m_ignoreHeads){
    if(outDSV[0].length() >= 1)
      finalHRV.setRole(0,-1);
    else
      finalHRV.setRole(0,0);
    outLabel.setNT(0,label1.getNT(m_constrainedDim)->clone());
    // pick the rightmost;
    // this also ensures that we don't pick 0 when there's a real span somewhere
    outHeadPosns[0] = max(nrh1->getHeads()[m_constrainedDim],nrh2->getHeads()[m_constrainedDim]);
  }else if(needBoth){
    finalHRV.setRole(0,1);
    finalHRV2.setRole(0,2);
    outLabel.setNT(0,label1.getNT(m_constrainedDim)->clone());
    outLabel2.setNT(0,label2.getNT(m_constrainedDim)->clone());
    outHeadPosns[0] = nrh1->getHeads()[m_constrainedDim];
    outHeadPosns2[0] = nrh2->getHeads()[m_constrainedDim];
  }else if(item1->getSizes()[m_constrainedDim]) {
    finalHRV.setRole(0,1);
    outLabel.setNT(0,label1.getNT(m_constrainedDim)->clone());
    outHeadPosns[0] = nrh1->getHeads()[m_constrainedDim];
  }else if(item2->getSizes()[m_constrainedDim]) {
    finalHRV.setRole(0,2);
    outLabel.setNT(0,label2.getNT(m_constrainedDim)->clone());
    outHeadPosns[0] = nrh2->getHeads()[m_constrainedDim];
  }else{
    finalHRV.setRole(0,0);
    outLabel.setNT(0,label1.getNT(m_constrainedDim)->clone());
    outHeadPosns[0] = nrh1->getHeads()[m_constrainedDim];
  }
  
  /////////////////
  //
  // This section has a first cut approach to implementing
  // Binarization techniques 
  //
  ////////////////

  //this is a filter for empty components. 
  // you can only attach an 1D item to the "left" 
  if(m_rightAttach){
    if(pDSV1->empty(1-m_constrainedDim) and 
       !pDSV2->empty(m_constrainedDim) ){
      bool found = false;
      for(unsigned int i = 0; i < inSpan1.size(); i++){
	if(inSpan1[i].second == (inSpan2[0].first)) {
	  found = true;
	  break;
	}
	// if the 1D item is not being attach directly to the left
	// filter
	if(!found){
	  if(debug(3)){
	    cerr << "right attach filter: ";
	    item1->dump(cerr); item2->dump(cerr);
	    cerr << endl;
	  }
	  return;
	}
      }
    }
    else if(pDSV2->empty(1-m_constrainedDim) and
	    !pDSV1->empty(m_constrainedDim) ){
      bool found = false;
      for(unsigned int i = 0; i < inSpan2.size(); i++){
	if((inSpan2[i]).second == (inSpan1[0].first))
	  found = true;
	break;
      }
      if(!found){
	if(debug(3)){
	  cerr << "right attach filter2: ";
	  item1->dump(cerr); item2->dump(cerr);
	  cerr << endl;
	}
	return;
      }
    }
  }//end 1D filter
  
  // If the input DSpans are empty on the contrained component, and
  // this grammar is being used as a supervising grammar for learning,
  // then we want to update the item's hiddenDSVs structure.  If there
  // are two antecedent hiddenDSVs and their sum is incompatible with
  // the input PAV on the constrained component, then the antecedents
  // cannot compose, so we abort.
  Vector<DSV>* pOutHDSVs = NULL;
  if (inSpan1.empty() and inSpan2.empty()) {
    PrecArray inPA = arrg.second.getPA(m_constrainedDim);
    bool compatible = hDSV::crossProdWithPA(nrh1->getHiddenDSVs(), 
					    nrh2->getHiddenDSVs(), 
					    m_constrainedDim, inPA, pOutHDSVs);
    if (not compatible)
      return; // without returning any possConseqs
  }

  //vector with one (dummy) position tree pointer in it
  vector<PositionTree::iterator> vecIt(1);

  
  //construct hedge item
  Item* outItem = new NodeRefHedge(outLabel, outDSV, outHeadPosns, vecIt);
  if (pOutHDSVs)
      outItem->setHiddenDSVs(pOutHDSVs);
  
  if(debug(5)){
    bool tmp = GlobalVocab::s_noVocabsOut;
    GlobalVocab::s_noVocabsOut = false;
    cerr << "PartialWMTG adding NodeRefHedge1: " << *((NodeRefHedge*)outItem) << endl;
    GlobalVocab::s_noVocabsOut = tmp;
  }
  //insert item and return
  possConseqs.push_back(pair<HeirRoleVector,Item*>(finalHRV,outItem));
  
  if(needBoth){
    //construct hedge item
    outItem = new NodeRefHedge(outLabel2, outDSV, outHeadPosns2, vecIt);
    if (pOutHDSVs)
      outItem->setHiddenDSVs(pOutHDSVs);
    
    if(debug(5)){
      bool tmp = GlobalVocab::s_noVocabsOut;
      GlobalVocab::s_noVocabsOut = false;
      cerr << "PartialWMTG adding NodeRefHedge2: " << *((NodeRefHedge*)outItem) << endl;
      GlobalVocab::s_noVocabsOut = tmp;
    }
    //insert item and return
    possConseqs.push_back(pair<HeirRoleVector,Item*>(finalHRV2,outItem));
  }
}

/// Returns all the possible lexicalized pre-terminals for a terminal
/// link covering the input span. The tree iterators in the returned
/// set point to the leaves in the restricting tree. Item costs remain
/// at the initial zero.  Note that the returned items are
/// *one-dimensional*; they are combined with items from other
/// dimensions by the grammar that contains this one.
void
PartialWMTG::getPossibleConsForTLink(const TLink& inTL, const DSV& inDSV, vector<Item*>& ret) const {
  if(debug(5)){
    bool tmp = GlobalVocab::s_noVocabsOut;
    GlobalVocab::s_noVocabsOut = false;
    GlobalVocab::s_noVocabsOut = tmp;
  }

  PositionTree::iterator i = 0;
  vector<PositionTree::iterator> viPT(1);
  viPT[0] = i;
  Terminal theTerminal = inTL[m_constrainedDim];
  TLink outTL(1, theTerminal);
  NTLink outNTL(1, SynCat(0));
  NTLink outLNTL(outNTL, outTL);
  DSpan inSpan;
  if (inDSV.size() > m_constrainedDim)
    inSpan = inDSV[m_constrainedDim];
  DSV outDSV(1);
  outDSV[0] = inSpan;
  short inPosn;
  if (not inSpan.empty())
    // the word position is the right boundary of the 0th contiguous span
    inPosn = inSpan[0].second;
  else
    inPosn = 0;
  vector<short> outHeadPosns(1, inPosn);
  
  Item* nrh = new NodeRefHedge(outLNTL, outDSV, outHeadPosns, viPT);

  // If the inSpan is empty in the active component, but the TLink is
  // not null, then this is a Load.  This situation can arise if this
  // grammar is being used as a supervising grammar for learning, in
  // which case we still want to populate the item's hiddenDSVs
  // structure for efficiency.  So we find all possible spans for it
  // in the string, which the grammar keeps track of in m_pCurrSent.


  // I AM COMMENTING THE FOLLOWING CODE OUT BECAUSE IT EXPECTS A SNT TUPLE IN THE 
  // TRANSLATION SIDE OF THINGS.  IF ITS A LOAD, THE SNT TUPLE HAS NO SIZE.


//   if (theTerminal != NULLT and inSpan.empty()) {

//     Vector<DSV>* pDSVV = new Vector<DSV>(1);
//     unsigned short len = m_pCurrSent->size();
//     for(unsigned short osi = 0; osi < len; ++osi)
//       if ( (*m_pCurrSent)[osi] == theTerminal) {
// 	DSpan aSpan = DSpan(DSpan::Span(osi, osi+1));
// 	DSV hiddenDSV(m_arity);
// 	hiddenDSV[m_constrainedDim] = aSpan;
// 	pDSVV->push_back(hiddenDSV);
//       }
//     if (not pDSVV->empty())
//       // transfer ownership to the item
//       nrh->setHiddenDSVs(pDSVV);
//     else
//       delete pDSVV;
//   }

  if(debug(5)){
    bool tmp = GlobalVocab::s_noVocabsOut;
    GlobalVocab::s_noVocabsOut = false;
    cerr << "PartialWMTG adding NodeRefHedge: " << *((NodeRefHedge*)nrh) << endl;
    GlobalVocab::s_noVocabsOut = tmp;
  }

  ret.push_back(nrh);
  return;
}

/** Adds this class' option description to the option description list of the hierarchy */
BSP<BPO::options_description> 
PartialWMTG::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  optionsDescription->add_options()
    ("TreeDim",  BPO::value<unsigned int>(), "The constrained dimension")
    ("RightAttach",  BPO::value<bool>(), 
     "True if we want to only attach epsilons in one direction")
    ;
  return optionsDescription;
}
