#include "PositionTree.hpp"
#include "GlobalVocab.hpp"
#include "Exceptions.hpp"

/*
 * Output of PositionTreeNode
 */
void PositionTree::printPositionTreeNode(ostream& out, const PositionTreeNode& data, bool useGlobalVocab) {
  //heir vector of size 0 means yield.
  if(BGET_HEIRVEC(data).size() == 0)
    out<<BGET_DSV(data);

  if(useGlobalVocab)
    (static_cast<const LexNT*>((BGET_POSITION_LABEL(data).getNT(0))))
      ->put(out, GlobalVocab::s_synCatSet, GlobalVocab::s_terminalSet, true);
  else
    (static_cast<const LexNT*>((BGET_POSITION_LABEL(data).getNT(0))))->put(out);
}

/*
 * Output of PositionTree
 */
void PositionTree::printPositionTree(ostream& out, PositionTree::iterator it, bool useGlobalVocab) {
  PositionTree::sibling_iterator sib_it;

  out<<"(";
  if(is_valid(it) && it != end()){
    if(isHeir(it) && it != begin() ) {
      //if it is an heir, find out in what dimensions.
      queue<short> d = getHeirQueue(it);
      //output the appropriate dimensions for the heir.
      while(!d.empty()){
	out<<"*";
	out << ((short)d.front());
	d.pop();
      }
      out << " ";
    }
    it->printPositionTreeNode(out, useGlobalVocab);
    for(sib_it = it.begin(); sib_it != it.end(); ++sib_it)
      printPositionTree(out, sib_it, useGlobalVocab);
  }
  out<<")";
}

void PositionTree::printPositionTree(ostream& out, bool useGlobalVocab) {
  PositionTree::printPositionTree(out, begin(), useGlobalVocab);
}

/*
 * Input of {\em 1-D} PositionTreeNode.
 * Used to read Charniak parse.
 */
void PositionTree::readPositionTreeNode(istream& in, PositionTreeNode& data, arity_t d) {
  // PositionTree::InitPositionTreeNode(data);
  data.InitPositionTreeNode(1);
  BGET_SPAN(data).clear();

  // read NTLink.
  LexNT* lnt = new LexNT();
  //string str;

  if(!GlobalVocab::s_noVocabsIn){
    in >> *lnt;
    in >> ws;
  }
  else{
    lnt->get(in,*GlobalVocab::s_synCatSet, *GlobalVocab::s_terminalSet, true);
    in >> ws;
  }

  BGET_POSITION_LABEL(data).setNT(0,lnt);

  // set the arity of span
  // if(BGET_SPAN(data).arity() == 0)
  // BGET_SPAN(data).setDimity(BGET_POSITION_LABEL(data).arity());
}

/*
 * Input of {\em 1-D} PositionTree.
 */
void PositionTree::readPositionTree(istream& in, PositionTree::iterator it, arity_t d) {
  char ch;

  PositionTreeNode data(1);

  if(in.eof()) return;
  in>>ch; if(in.eof()) return ;

  if(ch != '('){
    THROW("In PostionTree input, expected a '(', got "<<ch<<endl);
  }

  /*
   * Read data for the node pointed to by the input iterator.
   */
  // readPositionTreeNode(in,data,d);
  data.readPositionTreeNode(in, d);
  if(it == begin()) {
    it = set_head(data);
  } else {
    *it = data;
  }

  /*
   * Read children
   */
  while(in.peek() != ')'){
    PositionTree::iterator it1;
    it1 = append_child(it,data);
    readPositionTree(in, it1,d);
  }

  in >> ch;
  in >> ws;

  /*
   * Add children span and set heir.
   */
  if(it.number_of_children() == 0){
    short position = (static_cast<const LexNT*>((BGET_POSITION_LABEL(*it).getNT(0))))->lex();
    BGET_SPAN(*it)[0].push_back(DSpan::Span(position - 1, position ));
  }
  else{
    PositionTree::sibling_iterator it_sibling;
    // set heir
    for(it_sibling = it.begin(); it_sibling != it.end(); ++it_sibling){
      if(index(it_sibling) == ((static_cast<const LexNT*>(BGET_POSITION_LABEL(*it).getNT(0))))->lex()){
	setHeir(it_sibling);
	LexNT* nt = (static_cast<LexNT*>(BGET_POSITION_LABEL(*it).getNT(0)->clone()));
	nt->lex()= (static_cast<const LexNT*>(BGET_POSITION_LABEL(*it_sibling).getNT(0)))->lex();
	BGET_POSITION_LABEL(*it).setNT(0,nt);
	break;
      }
    }
    // add span.
    for(it_sibling = it.begin(); it_sibling != it.end(); ++it_sibling){
      addSpan(it, BGET_SPAN(*it_sibling));
    }
  }
}

void PositionTree::readPositionTree(istream& in, arity_t d) {
  PositionTree::readPositionTree(in, begin(), d);

  if(in.peek() == '\n') {
    in >> ws;
  }
}

//! Adds root node and EOSH to a position tree.
//! NB that the positions are not accurate since
//! these nodes dont have positions in the tree.
void PositionTree::addRoot(const PositionTree& posTree, PositionTree & retTree){

  arity_t dim;
  if(posTree.begin() == posTree.end()){
    cerr<<"Empty Tree, no dummy root added."<<endl;
    return;
  }
  dim = BGET_NT_LABEL(*posTree.begin()).arity();
  DSV currentRootSpan = BGET_SPAN(*posTree.begin());

  PositionTreeNode dataRoot(dim);
  dataRoot.InitPositionTreeNode(dim);

  PositionTreeNode dataEosh(dim);
  dataEosh.InitPositionTreeNode(dim);

  BGET_POSITION_LABEL(dataEosh)= NTLink(1);
  BGET_POSITION_LABEL(dataRoot)= NTLink(1);

  for(arity_t d = 0; d < dim; ++d){
    // either empty span or an entire span.
    assert(currentRootSpan[d].size() <= 1);
    short position = currentRootSpan[d].size() ? currentRootSpan[d].back().second : 0;
    BGET_SPAN(dataEosh)[d].push_back(DSV::Span(position, position + 1));
    BGET_POSITION_LABEL(dataEosh).setNT(d,new LexNT(RootTag,EOSH));
    BGET_SPAN(dataRoot)[d].push_back(DSV::Span(0 , position + 1));
    BGET_POSITION_LABEL(dataRoot).setNT(d,new LexNT(Root,EOSH));
  }

  //set up our new tree with the root node.
  PositionTree::iterator itRoot = retTree.set_head(dataRoot);
  
  PositionTree::iterator pi = retTree.append_child(itRoot, dataEosh);

  retTree.replace(pi, posTree.begin());
  


  //these set up the correct pointers to alter the tree.
  //  retTree.begin().node->first_child = posTree.begin().node;
  //retTree.begin().node->last_child = posTree.begin().node;
  //posTree.begin().node->next_sibling = 0;
  //posTree.begin().node->parent = retTree.begin().node;

  //set the EOSH to be the heir.
  pi = retTree.append_child(itRoot, dataEosh);
  retTree.setHeir(pi);
}

bool operator<(const PositionTree::iterator it1, const PositionTree::iterator it2) {
  if(BGET_SPAN(*it1) < BGET_SPAN(*it2)){ return true;}
  else if(BGET_SPAN(*it2) < BGET_SPAN(*it1)){ return false;}
  else {
    return it1.node < it2.node;
  }
}

// return the position of the headword of a node
// zero means unknown or nonexistent
short PositionTree::headPosn(const iterator& it, arity_t dim) {
  if (it.number_of_children() == 0) {
    const DSV& theDSV = BGET_DSV(*it);
    if (theDSV.empty(dim))
      return 0;
    else 
      return theDSV[dim][0].second; // right boundary of 0th DSpan on relevant dim
  } else if (it.number_of_children() == 1) {
    return headPosn( it.begin(), dim);
  } else {
    // recurse to the node which is the heir on that dim
    const iterator heirIt = findHeir(it, dim);
    return headPosn( heirIt, dim);
  }
};

// The purpose of this method is to abstract away from design
// decisions about whether leaves should contain only terminals
// or also the preterminal labels.  As of Mar'06, HeadMTrees store
// only terminal info in leaves, but PositionTrees store both the
// terminal and its POS tag.
// if iter is not a leaf, then throw exception
const SynCat PositionTree::pretermLabel(const iterator& iter, arity_t dim) {
    if (iter.number_of_children() > 0) {
      THROW("iter does not point to a leaf");
    }

    const SynCat ret = BGET_NT_LABEL(*iter).getNT(dim)->getSynCat();
    return ret;
};

// find the lowest node under the given iterator that covers the given DSV
PositionTree::iterator PositionTree::lowestSubsumingNode(const iterator& it,const DSV& dsv){
  if(BGET_SPAN(*it).contains(dsv)){
    PositionTree::sibling_iterator it_sibling;
    // set heir
    for(it_sibling = it.begin(); it_sibling != it.end(); ++it_sibling){
      if(BGET_SPAN(*it_sibling).contains(dsv)){
	return lowestSubsumingNode(it_sibling,dsv); 
      }
    }
  }
  return it;
}

// find the lowest node under the given iterator that covers the given DSV on the given dim
PositionTree::iterator PositionTree::lowestSubsumingNode(const iterator& it,const DSV& dsv, const arity_t dim){
  if(BGET_SPAN(*it).contains(dim,dsv)){
    PositionTree::sibling_iterator it_sibling;
    // set heir
    for(it_sibling = it.begin(); it_sibling != it.end(); ++it_sibling){
      if(BGET_SPAN(*it_sibling).contains(dim,dsv)){
	return lowestSubsumingNode(it_sibling,dsv,dim); 
      }
    }
  }
  return it;
}
