/// \brief Headed Multitree with word positions instead of words, used
/// by PseudoMTG to mesh with w2w models that represent w2w links as
/// pairs of word positions

#ifndef PositionTree_H_
#define PositionTree_H_

#include <set>
#include "tree.hpp"
#include "HeirTree.hpp"
#include "SpanTree.hpp"
#include "Vector.hpp"
#include "PAVTree.hpp"
#include "PositionTreeNode.hpp"
#include "NTLink.hpp"
#include "SentenceTuple.hpp"

class PositionTree : public SpanTree<HeirTree<tree<PositionTreeNode> > >{

public:

  //! Output of PositionTree.
  void printPositionTree(ostream& out, bool useGlobalVocab = false);
  //! Output of PositionTree.
  void printPositionTree(ostream& out, PositionTree::iterator it, bool useGlobalVocab = false);

  //!Input of {\em 1-D} PositionTree.
  void readPositionTree(istream& in, arity_t d);
  //!Input of {\em 1-D} PositionTree.
  void readPositionTree(istream& in, PositionTree::iterator it, arity_t d);

  //! PositionTreeNode output.
  void printPositionTreeNode(ostream&, const PositionTreeNode&, bool useGlobalVocab = false);

  //! PositionTreeNode input.
  void readPositionTreeNode(istream&, PositionTreeNode&, arity_t d);

  //! Initializes the PositionTreeNode.
  void InitPositionTreeNode(PositionTreeNode& data);

  static iterator lowestSubsumingNode(const iterator& it,const DSV& dsv);
  static iterator lowestSubsumingNode(const iterator& it,const DSV& dsv, const arity_t dim);

  //! Adds root node and EOSH to a position tree.
  //! NB that the positions are not accurate since
  //! these nodes dont have positions in the tree.
  void addRoot(const PositionTree& posTree, PositionTree & retTree);

//   //! Helper for converting a positionTree to a multitree
//   void positionTree2LexTree(HeadMTree& lexTree, const SentenceTuple& sntTp) 
//     THROW("Invalid Tree");
//   void positionTree2LexTree_helper(PositionTree::iterator itPosTree,
// 				   HeadMTree& lexTree, HeadMTree::iterator itLexTree,
// 				   const SentenceTuple& sntTp);

  //! The order of Kids.
  friend bool operator<(const PositionTree::iterator it1, const PositionTree::iterator it2);

  friend bool operator<(const PositionTreeNode& t1, const PositionTreeNode& t2);

  /// return the position of the headword of a node
  /// zero means unknown or nonexistent
  static short headPosn(const iterator& it, arity_t dim);

  /// The purpose of this method is to abstract away from design
  /// decisions about whether leaves should contain only terminals
  /// or also the preterminal labels.  As of Mar'06, HeadMTrees store
  /// only terminal info in leaves, but PositionTrees store both the
  /// terminal and its POS tag.
  /// if iter is not a leaf, then throw exception
  static const SynCat pretermLabel(const iterator& iter, arity_t dim);

  // static wrapper that helps to reduce methods called directly on the iterator
  static unsigned int number_of_children(const iterator& it) {
    return it.number_of_children();
  }

};

bool operator<(const PositionTree::iterator it1, const PositionTree::iterator it2);

bool operator<(const PositionTreeNode& t1, const PositionTreeNode& t2);


namespace std {
  template<>
  struct less<PositionTreeNode> {
    bool operator()(const PositionTreeNode& t1, const PositionTreeNode& t2) const {
      if(BGET_CONST_SPAN(t1) < BGET_CONST_SPAN(t2)) {
	return true;
      }
      else { return false;}
    }
  };
};

#endif
