#include "Production.hpp"
#include "Dbug.hpp"
#include "TLink.hpp"
#include "NTLink.hpp"
#include "Compose.hpp"
#include "Scan.hpp"

Production::
Production(const Production& other)
  : m_rhs(other.m_rhs),
    m_lhs(other.m_lhs),
    m_PAV(other.m_PAV)
{}

//! constructor from inference
Production::Production(const Inference* pInf) 
  : m_lhs(static_cast<const Hedge*>(pInf->getConseq())->getNTLink()),
    m_PAV(pInf->getPAV()) {
  const Scan* pScan = dynamic_cast<const Scan*>(pInf);
  if (pScan) {
    const TLink& theTL = pScan->getTLink();
    m_rhs.push_back(ProdRHSElem(theTL));
  } else {
    const Compose* pComp = dynamic_cast<const Compose*>(pInf);
    if (pComp) {
      vector<const Item*> iVec = pComp->getAnt();
      vector<const Item*>::const_iterator vici = iVec.begin();
      vector<const Item*>::const_iterator sentinel = iVec.end();
      while (vici != sentinel) {
	const Hedge* pH = dynamic_cast<const Hedge*>(*vici);
	if (pH) {
	  const NTLink& theNT = pH->getNTLink();
	  m_rhs.push_back(ProdRHSElem(theNT));
	} else
	  THROW("Antecedent is not a hedge.");
	++vici;
      }
    } else
      THROW("Unknown inference type");
  }
}

Production::
Production(const HeadMTree::iterator& it)
  : m_lhs(BGET_NT_LABEL(*it).size()) {

  HeadMTree::sibling_iterator it_sib;
  for(it_sib = it.begin(); it_sib != it.end(); ++it_sib) {
    //rhs label
    if (Pair<TLink, DSV>* p = boost::get< Pair<TLink, DSV> >( &BGET_EITHER_LABEL(*it_sib) ) )
      rhsLinks().push_back(p->first);
    else
      if (NTLink* p = boost::get< NTLink >( &BGET_EITHER_LABEL(*it_sib) ) )
	rhsLinks().push_back(*p);
      else
	if (TLink* p = boost::get< TLink >( &BGET_EITHER_LABEL(*it_sib) ) )
	  rhsLinks().push_back(*p);
  }
  
  // lhs always NT link.
  lhs() = BGET_NT_LABEL(*it);
  
  // pav
  rhsPerm() = BGET_PAV(*it);
}

/*! Constructor.
 * if links vector is empty, then it is terminal production.
 */
Production::
Production(const NTLink& lhss,
	   const Vector<ProdRHSElem>& rhslinks,
	   const PAV& pav)
  : m_rhs(rhslinks),
    m_lhs(lhss),
    m_PAV(pav)
{}

Production::~Production() {
  if (Dbug::isDebugAll(6))
    cerr << "Destroying " << *this << endl;;
};

// Equality test
bool
Production::operator==(const Production& other) const {
  return (lhs() == other.lhs() 
	  && rhsLinks() == other.rhsLinks()
	  && rhsPerm() == other.rhsPerm());
}

bool Production::operator<(const Production& other) const {
  return (cmp(other) < 0);
}

bool Production::isW2WProduction() const{
  if(rank() != 2)
    return false;
  const NTLink* nt1  = boost::get<const NTLink>(addressof(m_rhs[0]));
  const NTLink* nt2 = boost::get<const NTLink>(addressof(m_rhs[1]));
  if(!(nt1 && nt2))
    return false;
  if((nt1->getNT(0)->nt() == NULLNT || nt2->getNT(1)->nt() == NULLNT))
    return true;
  if((nt1->getNT(1)->nt() == NULLNT || nt2->getNT(0)->nt() == NULLNT))
    return true;
//  if((*nt1)[0]->nt() == NULLNT && (*nt2)[0]->nt() == NULLNT)
//    return true;
//  if((*nt1)[1]->nt() == NULLNT && (*nt2)[1]->nt() == NULLNT)
  //   return true;
  return false;
}

int Production::cmp(const Production& other) const {
  int ret;

  //lhs
  if((ret = m_lhs.cmp(other.m_lhs)) != 0) return ret;

  // pav
  if((ret = m_PAV.cmp(other.m_PAV)) != 0) return ret;

  // rhs: first by length
  size_t thisSize = m_rhs.size();
  size_t otherSize = other.m_rhs.size();
  if((ret = MC::cmp(thisSize, otherSize)) != 0) return ret;
  Vector<ProdRHSElem>::const_iterator it1(m_rhs.begin());
  Vector<ProdRHSElem>::const_iterator it2(other.m_rhs.begin());
  for (; it1 != m_rhs.end(); ++it1, ++it2) {
    if (const NTLink* pNTL1 = boost::get< const NTLink >( &(*it1)) )
      if (boost::get< const TLink >(&(*it2)) )
	return 1; // tlinks always less than ntlinks
      else if (const NTLink* pNTL2 = boost::get< const NTLink >(&(*it2)) )
	return (*pNTL1).cmp(*pNTL2);
      else
	if (const TLink* pTL1 = boost::get< const TLink >(&(*it1)) )
	  if (boost::get< const NTLink >(&(*it2)) )
	    return -1; // tlinks always less than ntlinks
	  else if (const TLink* pTL2 = boost::get< const TLink >(&(*it2)) )
	    return MC::cmp(*pTL1,*pTL2);
  };

  return 0;
}

// Returns the lhs.
const NTLink Production::lhs() const {
  return m_lhs;
}

// Returns the lhs.
NTLink& Production::lhs() {
  return m_lhs;
}

//! Returns the links at RHS.
const Vector<ProdRHSElem>& Production:: rhsLinks() const {
  return m_rhs;
}

//! Returns the links at RHS.
Vector<ProdRHSElem>& Production:: rhsLinks() {
  return m_rhs;
}

//! Returns the permutation at RHS.
const PAV& Production:: rhsPerm() const {
  return m_PAV;
}

//! Returns the permutation at RHS.
PAV& Production::rhsPerm() {
  return m_PAV;
}


/// If possible, extract a W2WLink from a Production.  This can only
/// succeed if the production is binary.  If successful, return true,
/// o.w. return false.  If successful, \param lk will contain the
/// extracted link.  Currently, this ignores links involving a null in
/// the 0th component.
bool Production::getTLink(W2WLink& lk) const {
  if (rank() != 2)
    return false;
  const NTLink* nt1 = boost::get<const NTLink>(addressof(m_rhs[0]));
  const NTLink* nt2 = boost::get<const NTLink>(addressof(m_rhs[1]));
  if( nt1 && nt2) {
    TLink tl1 = nt1->getTLink();
    TLink tl2 = nt2->getTLink();
    // Assuming they are both 2D, even if one of the slots is inactive (null).
    // Also assuming that each is active in at least one slot.
    if (not tl1.active(0) && tl2.active(0)) {
      lk = W2WLink(tl2[0], tl1[1]);
      return true;
    } else if (not tl1.active(1) && tl2.active(1)) {
      lk = W2WLink(tl1[0], tl2[1]);
      return true;
    } else if (not tl2.active(0) && tl1.active(0)) {
      lk = W2WLink(tl1[0], tl2[1]);
      return true;
    } else if (not tl2.active(1) && tl1.active(1)) {
      lk = W2WLink(tl2[0], tl1[1]);
      return true;
    }
  } else 
    THROW("Binary production has something other than NTLinks on the RHS!");

  return false;
}

// Assignment
Production&
Production::operator=(const Production& other) {
  UniqueID::operator=(other);
  lhs() = other.lhs();
  rhsLinks() = other.rhsLinks();
  rhsPerm() = other.rhsPerm();
  return *this;
}

void Production::clear() {
  rhsLinks().clear();
  rhsPerm() = PAV();
}

bool Production::isTerminalProduction() const{
  for(unsigned int i = 0; i < rhsLinks().size(); i++){
    if(!(boost::get<TLink>(&rhsLinks()[i])))
      return false;
  }
  return true;
}

ostream& operator<<(ostream& out,
	            const Production& aProd) {
  aProd.put(out);
  out << endl;
  return out;
}

ostream& Production::put(ostream& out) const {
  if (Dbug::isDebugAll(3))
    cerr << "Prod:= " << m_uniqueID;

  /*output lhs*/
  out << lhs();

  /*output rhs links*/
  for(size_t r = 0; r < rank(); ++r) {
    out<<"^"<<rhsLinks()[r];
  }

  /*output rhs permutation.*/
  // terminal production doesn't output the permutation.
  if(rank()) {
    out<<"^"<<rhsPerm();
  }

  return out;
}

istream& operator>>(istream& in,
	            Production& aProd) {
  aProd.clear();
  in >> ws;
  string buffer;
  getline(in, buffer, '\n');

  vector<string> strs = split(buffer, "^");
  deque<string> strs1;
  copy(strs.begin(), strs.end(), back_inserter<deque<string> >(strs1));

  // every aProd at least has lhs label and heirRole (even if there is no heir).
  if(strs1.size() <= 1) return in;

  // heirRoleVector
  // There is still an heirRole even if there is no
  // heir so we need to parse it out of the way
  strs = split(strs1.back().c_str(), " ");
  HeirRoleVector hv(strs.size());
  for(unsigned int i = 0; i < strs.size(); i++) {
    hv.setRole(i,atoi(strs[i].c_str()));
  }

  strs1.pop_back();

  // lhs
#ifdef USE_STRSTREAM
  istrstream ist(strs1[0].c_str(), strs1.size());
#else
  istringstream ist(strs1[0]);
#endif
  NTLink el(hv.arity());
  ist >> ws;
  ist >> el;
  aProd.lhs() = el;
  strs1.pop_front();

  // pav. If, now, the size of strs1 is still nonzero, then
  // the rank will not be zero.
  if(strs1.size()) {
#ifdef USE_STRSTREAM
    istrstream ist1(strs1.back().c_str(), strs1.back().size());
#else
    istringstream ist1(strs1.back());
#endif
    ist >> ws;
    ist1 >> aProd.rhsPerm();
    strs1.pop_back();
  }

  // rhs label.
  while(strs1.size()) {
#ifdef USE_STRSTREAM
    istrstream ist1(strs1.front().c_str(), strs1.front().size());
#else
    istringstream ist1(strs1.front());
#endif
    in >> ws;
    ist1 >> ws;
    //if there is a {, then it's an NTLink
    if(strs1.front().c_str()[0] == '{' ||
       ( strs1.front().c_str()[1] == '{' && strs1.front().c_str()[0] == ' ')) {
      NTLink elem(hv.arity());
      ist1 >> elem;
      aProd.rhsLinks().push_back(elem);
    } else {
      //else it's a TLink
      TLink elem;
      ist1 >> elem;
      aProd.rhsLinks().push_back(elem);
    }
    strs1.pop_front();
  }
  return in;
}

istream& Production::get(istream& in){
  in >> *this;
  return in;
}

size_t Production::hash() const {
  size_t ret;
  // lhs
  ret = lhs().hash();

  // rhs label.
  for(size_t i = 0; i < rank(); ++i) {

    if(const NTLink* nt = boost::get<NTLink>(&rhsLinks()[i]))
      ret += nt->hash() * (i+1);
    else if(const TLink* t = boost::get<TLink>(&rhsLinks()[i]))
      ret += t->hash() * (i+1);
    // else if(const Pair<TLink, DSV>* p = boost::get<Pair<TLink, DSV> >(&rhsLinks()[i]))
    //	ret += p->first.hash() * (i+1);
    else
      THROW("Error: cannot hash." << endl);
  }

  // pav.
  if(rank()) {
    ret *= 10;
    ret += rhsPerm().hash();
  }

  ret *= 10;

  return ret;
}

// Returns the fanout of this production.
short Production::fanout() const {
  return rhsPerm().fanout();
}

// A => \epsilon, where A might be NULLNT[NULLT].
bool Production::isEpsilonProduction() const {
  if(rank() == 0 /*valency zero*/ &&
     rhsPerm().rank() == 0 /*empty span*/ &&
     lhs().getTLink() == TLink(arity(), NULLT) ) {
    return true;
  } else {
    return false;
  }
}

// A=>B, no matter what the rank of PAV is.
bool Production::isUnary() const {
  if(rank() == 1 /**/) {
    return true;
  } else {
    return false;
  }
}

// Remove a role, including the link and the role in the PAV.
void Production::eraseRole(short r) {
  rhsLinks().erase(rhsLinks().begin() + (r -1));
  rhsPerm() = PAV::deleteRoleAndNormalize(rhsPerm(), r);
}

//! Remove the \epsilon RHS elements, modifying roles in PAV and
//! heir index accordingly.
void Production::removeRHSEpsilon() {
  // backward!!
  for(short r = rank(); r > (short)0; --r) {
    // if \epsilon
    if(boost::get<NTLink>(rhsLinks()[r-1])==NTLink::nullNTLink(arity()) && /*null label*/
       rhsPerm().getFanoutV(r) == FanoutVect(rhsPerm().arity(), 0)) /*empty span*/ {
      eraseRole(r);
    }
  }
}

namespace MC {
  template<> int cmp<Production>(const Production& p1, const Production& p2) {
    return p1.cmp(p2);
  }
}
