#ifndef _Production_H_
#define _Production_H_

#include "HeadMTree.hpp"
#include "NTLink.hpp"
#include "PAV.hpp"
#include "Bigram.hpp"
#include "Vocabulary.hpp"
#include "ProdRHSElem.hpp"
#include "Exceptions.hpp"
#include "MoreComparators.hpp"
#include <boost/serialization/serialization.hpp>

class Inference;

//! Basic MTG production.
/*!
 * These productions don't contain heir information.
 */
class Production : public UniqueID {
 public:

  //! Empty constructor.
  Production() {}

  //! constructor
  Production(const NTLink& lhs,
	     const Vector<ProdRHSElem>& rhsLinks,
	     const PAV& pav);

  //! constructor from inference
  Production(const Inference* pInf);

  //! constructor from a HeadMTree iterator
  Production(const HeadMTree::iterator&);

  //! copy constructor
  Production(const Production&);

  //! virtual destructor
  virtual ~Production();

  //! Assignment.
  virtual Production& operator=(const Production&);

  //! Equality test.
  virtual bool operator==(const Production&) const;

  // comparator
  int cmp(const Production& other) const;

  // less function
  bool operator<(const Production& other) const;

  ///////////////////////////////////////////////////////////
  // Operations on LHS //
  ///////////////////////////////////////////////////////////

  //! Returns the lhs.
  const NTLink lhs() const;

  //! Returns the lhs.
  NTLink& lhs();

  ///////////////////////////////////////////////////////////
  // Operations on RHS //
  ///////////////////////////////////////////////////////////

  //! Returns the links at RHS.
  const Vector<ProdRHSElem >& rhsLinks() const;

  //! Returns the links at RHS.
  Vector<ProdRHSElem >& rhsLinks();

  //! Returns the permutation.
  const PAV& rhsPerm() const;

  //! Returns the permutation.
  PAV& rhsPerm() ;
  
  
  /// Extract a W2WLink from a binary Production, if possible.  If
  /// successful, return true, o.w. return false.  If successful,
  /// \param lk will contain the extracted link.  Currently, this
  /// ignores links involving a null in the 0th component.
  virtual bool getTLink(W2WLink& lk) const;

  /// Equals the valency value of the valency event
  /// Note: This will overestimate the rank when the rule has both
  /// terminals and nonterminals on the RHS.
  size_t rank() const{ return rhsLinks().size();}

  //! returns true if this is a terminal production
  bool isTerminalProduction() const;

  //! returns true is this production attaches two 1D productions
  //  in opposite components into a 2D production.
  virtual bool isW2WProduction() const;

//   //! Returns the bigram on the inside of the RHS of the last component
//   virtual Bigram extractInnerBigrams()const{throw NotImplemented();};
  
  //! what do you think this does?
  virtual void clear();

  //! For hashing.
  virtual size_t hash() const;

  //! Returns the fanout of this production.
  //! Equals to the fanout of its PAV.
  short fanout() const;

  //! A=>epsilon.
  bool isEpsilonProduction() const;

  //! A=>B, no matter what the rank of PAV is.
  bool isUnary() const;

  //! Remove a RHS element by removing both the link and the role in the
  //! PAV.
  //! \param r : the 1-started link index.
  virtual void eraseRole(short r);

  //! Remove the epsilon RHS elements, modifying roles in PAV and
  //! heir index accordingly.
  void removeRHSEpsilon();
  
   //! Returns the language dimensionality.
  arity_t arity() const {return lhs().size();}

  // makes a copy of this Production
  virtual Production* clone() const{
    return new Production(*this);
  };

  ///////////////////////////////////////////////////////////
  // Input / Output //
  ///////////////////////////////////////////////////////////

  //!Output a Production
  friend ostream& operator << (ostream&, const Production&);
  //!Output a production. (needed for polymorphism)
  virtual ostream& put(ostream&) const;

  //!Input a production
  friend istream& operator >> (istream&, Production&);
  //!Input a production. (needed for polymorphism)
  virtual istream& get(istream&);

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    ar & m_lhs;
    ar & m_rhs;
    ar & m_PAV;
  };
  
  //! Return the data simply as a vector of ints.
  virtual Vector<int> makeVec() const{return Vector<int>();};

 protected:

  Vector<ProdRHSElem > m_rhs;

  NTLink m_lhs;

  PAV m_PAV;
};

namespace std {
  template<> struct less<Production> {
    bool operator()(Production& p1, Production& p2) const {
      return (p1 < p2);
    };
    bool operator()(const Production& p1, const Production& p2) const {
      return (p1 < p2);
    };
    bool operator()(const BSP<Production> p1, const BSP<Production> p2) const {
      return (*p1 < *p2);
    };
  };
};

namespace MC {
  template<> int cmp<Production>(const Production& p1, const Production& p2);
};

namespace HashNS {
  //! Template specialization of STL's hash object.
  template<> class hash<Production> {
  public:
    size_t operator()(const Production& obj) const
    { return obj.hash(); }
  };

  template<> class hash<Production*> {
  public:
    size_t operator()(const Production* obj) const
    { return obj->hash(); }
  };

  template<> class hash<BSP<Production> > {
  public:
    size_t operator()(const BSP<Production> obj) const
    { return obj->hash(); }
  };
};

#endif
