#include "PrunerBuilder.hpp"

#include <fstream>
#include "Path.hpp"
#include "Exceptions.hpp"
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

#include "RelativePruningStrategy.hpp"
#include "AbsoluteCostPruningStrategy.hpp"
#include "CompositePruningStrategy.hpp"
#include "AbsoluteCountPruningStrategy.hpp"

PrunerBuilder* PrunerBuilder::_instance = 0;

PrunerBuilder* PrunerBuilder::Instance(){
  if(_instance == 0){
    _instance = new PrunerBuilder;
  }
  return _instance;
}

//! build pruner
PruningStrategy* PrunerBuilder::build(const string& configFile) {
  string optPruningStrategy;
  string optPruningConfigFile;

  // compile the config file.
  BPO::options_description config("Configuration");
  config.add_options()
    ("PruningStrategy", BPO::value<string>(&optPruningStrategy), "The type of the pruning strategy.")
    ("PrunerConfigFile", BPO::value<string>(&optPruningConfigFile), "The pruning config file.");
  BPO::variables_map vm;
  cerr << "PrunerBuilder: Compiling config file " << configFile << endl;
  iFile ifs(configFile.c_str());
  store(parse_config_file(ifs, config), vm);
  notify(vm);

  if(debug(1)) {
    cerr << "Option values of \"PrunerBuilder\"" << endl << endl;
    cerr << OptionsPrinter(config, vm) << endl;
  }
  if(optPruningStrategy == ""){
    THROW("Error: Pruning strategy unspecified!"<<endl);
  }

  string strPruningStrategy = optPruningStrategy;
  strPruningStrategy = tolower(strPruningStrategy);

  PruningStrategy* pruner;

  if (strPruningStrategy == "composite") {
    pruner = new CompositePruningStrategy();
    vector<string> strvec = split(optPruningConfigFile, "|");
    for (unsigned int i = 0; i < strvec.size(); i++) {
      vector<string> prunerDefine = split(strvec[i], ",");
      if (prunerDefine.size() < 2) {
	THROW("The config file is incorrectly formatted. It should be "
	      << "of the form: "
	      << "\"relative,relativeConfigFileName|absoluteCost,absoluteCostConfigFileName,absoluteCount, absoluteCountConfigFileName\""
	      << endl);
      }
      ((CompositePruningStrategy*)pruner)->add(buildOne(prunerDefine[0], prunerDefine[1]));
    }
    return pruner;
  }

  if(strPruningStrategy == "relative" || strPruningStrategy == "absoluteCost" || strPruningStrategy == "absoluteCount") {
    string pruningConfigFile = Path::maybe_compose(configFile, optPruningConfigFile);
    pruner = buildOne(strPruningStrategy, pruningConfigFile);
  } else {
    THROW("Error: unknown pruning strategy!"<<endl);
  }
  return pruner;
}

PruningStrategy* PrunerBuilder::buildOne(const string& prunerType, const string& configFile) {
  PruningStrategy* pruner;
  if(prunerType == "relative"){
    pruner = new RelativePruningStrategy(configFile);
  } else if(prunerType == "absoluteCost"){
    pruner = new AbsoluteCostPruningStrategy(configFile);
  } else if(prunerType == "absoluteCount"){
    pruner = new AbsoluteCountPruningStrategy(configFile);
  } else {
    THROW("Error: unknown pruning strategy!"<<endl);
  }
  return pruner;
}
