#include "PseudoMTG.hpp"
#include "CPIterator.hpp"
#include "NodeRefHedge.hpp"
#include "Scan.hpp"
#include "HeadCompose.hpp"
#include "HeadProduction.hpp"
#include "HiddenDSVs.hpp"
#include "LabelingStrategy.hpp"

#include "Path.hpp"
#include "PAV.hpp"
#include "LexNT.hpp"
#include "GlobalVocab.hpp"
#include "Exceptions.hpp"
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"
#include "MeasureMemory.hpp"

//! Constructor.
PseudoMTG::PseudoMTG(arity_t dm, const string& configFile, 
		     BSP<BPO::options_description> optionsDescription)
  : _grammar(dm), 
    _super(dm, configFile, addOptionDescriptionsTo(optionsDescription)),
    m_w2wTM(NULL),m_rightBin(false){
  string optW2WModelFile;

  /*
   * Compile the config file.
   */

  BPO::variables_map vm;
  cerr << "PseudoMTG: Compiling config file " << configFile << endl;
  iFile ifs(configFile.c_str());
  store(parse_config_file(ifs, *optionsDescription), vm);
  notify(vm);

  if(this->debug(1)) {
    cerr << "Option values of \"PseudoMTG\"" << endl << endl;
    cerr << OptionsPrinter(*optionsDescription, vm) << endl;
  }

  if(vm.count("w2wModel"))
    optW2WModelFile = vm["w2wModel"].as<string>();
  if(vm.count("RightBin"))
    m_rightBin = vm["RightBin"].as<bool>();

  if(optW2WModelFile == ""){
    m_pW2WModelStream = NULL;
  }else{
    // set up the W2W model steam
    m_pW2WModelStream = new iFile(optW2WModelFile);
  }
}

PseudoMTG::~PseudoMTG() {
  for(arity_t d = 0; d < m_monoGrammars.size(); ++d){
    delete m_monoGrammars[d];
  }
  delete m_w2wTM;
  delete m_pW2WModelStream;
}

ostream& PseudoMTG::put(ostream& o) const {
  o << "PseudoMTG" << endl;
  return o;
}

void PseudoMTG::add(Grammar* grammar){
  m_monoGrammars.push_back(dynamic_cast<PartialWMTG*>(grammar));
}

//! Returns the language dimensionality.
arity_t PseudoMTG::arity() const {
  return (arity_t)m_monoGrammars.size();
}

void PseudoMTG:: getArrangementFor(BSP<set<Pair<bool,PAV> > >& arrgs, const Item* trigger) const {
  // start with the super's set
  _super::getArrangementFor(arrgs, trigger);
  // let subgrammars filter this set
  for(arity_t dim = 0; dim < m_arity; ++dim)
    m_monoGrammars[dim]->filterArrangementFor(arrgs, trigger);
};

//! Returns all the possible pre-terminals lexicalized by
//! input terminal link.
//! Item costs remain at the initial zero.
void PseudoMTG::getPossibleConsForTLink(const TLink& tl, const DSV& dspan, vector<Item*>& ret) const {
  if(debug(2)){
    cerr <<"entering PsMTG::getPossConsForTlink " << tl << " with DSV " << dspan << endl;
  }

  unsigned int d;
  vector<Item*> tagset[arity()];

  for(d = 0; d < arity(); ++d)
    m_monoGrammars[d]->getUnaryCons(tl, dspan, tagset[d]);

  CPIterator<vector<Item*>, vector<Item*>, Item*>
    cpi(tagset[0].begin(), tagset[0].end(), tagset[1].begin(), tagset[1].end());
  vector<Item*> vec(arity());
  while (cpi.valid()) {
    vec[0] = cpi.firstDeref();
    vec[1] = cpi.secondDeref();
    NodeRefHedge* nrh = new NodeRefHedge(vec);
    Vector<DSV>* pOutHDSVs = NULL;
    bool compatible = hDSV::crossProd(vec[0]->getHiddenDSVs(), vec[1]->getHiddenDSVs(), 
		    pOutHDSVs);
    if (compatible)
      nrh->setHiddenDSVs(pOutHDSVs);
    // the item's scores remain at the initial zero
    if(debug(5)){
      bool tmp = GlobalVocab::s_noVocabsOut;
      GlobalVocab::s_noVocabsOut = false;
      cerr << "PsMTG made a " << *nrh << endl;
      GlobalVocab::s_noVocabsOut = tmp;
    }
    ret.push_back(nrh);
    cpi.increment();
  }

  // clean up
  vector<Item*>::iterator i;
  for(d = 0; d < arity(); ++d){
    for(i = tagset[d].begin(); i != tagset[d].end(); i++){
      delete *i;
    }
  }
  return;
}

bool PseudoMTG::prepareFor(SentenceTuple& tuple) {

  /*
   * prepare each submodel with the SentenceTuple
   */
  for(arity_t d = 0; d < arity(); ++d)
    m_monoGrammars[d]->prepareFor(tuple);

  
  if(m_pW2WModelStream){
    // load w2w links 
    W2WTMType hist;
    
    *m_pW2WModelStream >> hist;
    
    // create the W2W Model based on the histogram of terminal links.
    W2WTMType::iterator itHist;
    W2WTMType hist1;
    for(itHist = hist.begin(); itHist != hist.end(); ++itHist){
      //cap score at 1.
      if(1 < itHist->second)
	hist1.increment(itHist->first, POSINF);
      else
      hist1.increment(itHist->first, ((-1) * log(itHist->second)));
    }
    delete m_w2wTM;
    m_w2wTM = new W2WTMType(hist1);
   
    // add link for end-of-sentence head
    // this is a temporary hack that adds 10 to align with previous scheme....
    PosLinkType eosLink;
    eosLink.first = tuple[0].size();
    eosLink.second = tuple[1].size();
    m_w2wTM->increment(eosLink, ZEROSCORE);
  }
  return true;
}

/// Determine whether a certain inference should be allowed.  A
/// PseudoMTG allows all Scans.  It allows Loads that match some word
/// in the output component.  For a Compose, it presumes that the
/// antecedents are OK, but needs to check w2w constraints, tree
/// constraints (including labels), and consequent fan-out.
bool PseudoMTG::allowInf(BSP<Inference> spInf) {
  
  Hedge* pConseq = static_cast<Hedge*>(spInf->getConseq());
  arity_t firstOutDim = 1;

  if (debug(3))
    cerr << "PsMTG::allowInf for conseq " << pConseq;

  // For efficiency, we sneakily substitute a NodeRefHedge for every
  // Hedge, without telling the caller :O. 
  if (BSP<Scan> spScan = dynamic_pointer_cast<Scan>(spInf)) {
    // substitute NodeRefHedge
    TLink tl = spScan->getTLink();
    DSV* pDSV = pConseq->getDSV();
    vector<Item*> allowedConseqVec;
    getUnaryCons(tl, *pDSV, allowedConseqVec);
    while (! allowedConseqVec.empty()) {
      Item* pPar = allowedConseqVec.back();
      allowedConseqVec.pop_back();
      // hidden DSVs would be already set
      //      pPar->hideDSV(firstOutDim);
      if (pConseq->equal_to(*pPar)) {
	if (debug(3))
	  cerr<< "PsMTG::allowInf substituting " << *pConseq << " with " << *pPar;
	// substitute NodeRefHedge
	spInf->setConseq(pPar); // also deletes previous consequent
	pPar->setInference(spInf);
	// clean up the rest of the possible allowedConseqVec
	while (! allowedConseqVec.empty()) {
	  Item* pPar = allowedConseqVec.back();
	  allowedConseqVec.pop_back();
	  delete pPar;
	};
	return true;
      }
      delete pPar;
    };

    // allow all Scans and Loads for now
    return true;

  } else if (BSP<HeadCompose> spComp = dynamic_pointer_cast<HeadCompose>(spInf)) {
    // First, check that both antecedents are NodeRefHedges
    const NodeRefHedge* pAnt1;
    const NodeRefHedge* pAnt2;
    vector<const Item*> antVec = spComp->getAnt();
    // assuming exactly two antecedents
    if ((pAnt1 = dynamic_cast<const NodeRefHedge*>(antVec[0]))
	&& 
	(pAnt2 = dynamic_cast<const NodeRefHedge*>(antVec[1]))) {
      HeirRoleVector testHRV = spComp->heirIndices();
      // VERY INEFFICIENT DIRTY HACK :) !!! For now, just
      // getPossibleCons of the antecedents and check if the consequent
      // is one of them.
      vector<HRVIpair> allowedConseqVec;
      getHeadedBinaryCons(pAnt1, pAnt2, spComp->getArnmt(), allowedConseqVec);
      vector<Item*> pIVec1, pIVec2;
      vector<Item*>::const_iterator vici1, vici2;
      // restore all possible hiddenDSVs for both ants, and
      // get cons for each pairing
      pAnt1->restoreHDSVs(firstOutDim, pIVec1);
      pAnt2->restoreHDSVs(firstOutDim, pIVec2);
      for(vici1 = pIVec1.begin(); vici1 != pIVec1.end(); ++vici1)
	for(vici2 = pIVec2.begin(); vici2 != pIVec2.end(); ++vici2)
	  getHeadedBinaryCons(*vici1, *vici2, spComp->getArnmt(), allowedConseqVec);
      // clean up
      for(vici1 = pIVec1.begin(); vici1 != pIVec1.end(); ++vici1)
	delete *vici1;
      for(vici2 = pIVec2.begin(); vici2 != pIVec2.end(); ++vici2)
	delete *vici2;	  
      while (! allowedConseqVec.empty()) {
	HRVIpair thePair = allowedConseqVec.back();
	allowedConseqVec.pop_back();

	HeirRoleVector& currHRV = thePair.first;
	Item* pPar = thePair.second;
	pPar->hideDSV(firstOutDim);
	if (testHRV == currHRV 
	    && pConseq->equal_to(*pPar)) {
	  if (debug(3))
	    cerr<< "PsMTG::allowInf substituting " << *pConseq << " with " << *pPar;
	  // substitute NodeRefHedge
	  spInf->setConseq(pPar); // also deletes previous consequent
	  pPar->setInference(spInf);
	  // clean up the rest of the possible allowedConseqVec
	  while (! allowedConseqVec.empty()) {
	    HRVIpair thePair = allowedConseqVec.back();
	    allowedConseqVec.pop_back();
	    delete thePair.second;
	  };
	  return true;
	}
	delete pPar;
      };

      return false;
    } else
      // One of the ants is not a noderefhedge, which can happen if a
      // Scan/Load got accepted but not connected to a constraining
      // tree, because it can't be part of the final parse.
      return false;
  } else if (BSP<Compose> spComp = dynamic_pointer_cast<Compose>(spInf)) {
    // First, check that both antecedents are NodeRefHedges
    const NodeRefHedge* pAnt1;
    const NodeRefHedge* pAnt2;
    vector<const Item*> antVec = spComp->getAnt();
    // assuming exactly two antecedents
    if ((pAnt1 = dynamic_cast<const NodeRefHedge*>(antVec[0]))
	&& 
	(pAnt2 = dynamic_cast<const NodeRefHedge*>(antVec[1]))) {
      // VERY INEFFICIENT DIRTY HACK :) !!! For now, just
      // getPossibleCons of the antecedents and check if the consequent
      // is one of them.
      vector<Item*> allowedConseqVec;
      getBinaryCons(pAnt1, pAnt2, spComp->getArnmt(), allowedConseqVec);
      vector<Item*> pIVec1, pIVec2;
      vector<Item*>::const_iterator vici1, vici2;
      // restore all possible hiddenDSVs for both ants, and
      // get cons for each pairing
      pAnt1->restoreHDSVs(firstOutDim, pIVec1);
      pAnt2->restoreHDSVs(firstOutDim, pIVec2);
      for(vici1 = pIVec1.begin(); vici1 != pIVec1.end(); ++vici1)
	for(vici2 = pIVec2.begin(); vici2 != pIVec2.end(); ++vici2)
	  getBinaryCons(*vici1, *vici2, spComp->getArnmt(), allowedConseqVec);
      // clean up
      for(vici1 = pIVec1.begin(); vici1 != pIVec1.end(); ++vici1)
	delete *vici1;
      for(vici2 = pIVec2.begin(); vici2 != pIVec2.end(); ++vici2)
	delete *vici2;	  
      while (! allowedConseqVec.empty()) {
	Item* pPar = allowedConseqVec.back();
	allowedConseqVec.pop_back();
	pPar->hideDSV(firstOutDim);
	if (pConseq->equal_to(*pPar)) {
	  if (debug(3))
	    cerr<< "PsMTG::allowInf substituting " << *pConseq << " with " << *pPar;
	  // substitute NodeRefHedge
	  spInf->setConseq(pPar); // also deletes previous consequent
	  pPar->setInference(spInf);
	  // clean up the rest of the possible allowedConseqVec
	  while (! allowedConseqVec.empty()) {
	    Item* pPar = allowedConseqVec.back();
	    allowedConseqVec.pop_back();
	    delete pPar;
	  };
	  return true;
	}
	delete pPar;
      };

      return false;

    } else
      // One of the ants is not a noderefhedge, which can happen if a
      // Scan/Load got accepted but not connected to a constraining
      // tree, because it can't be part of the final parse.
      return false;
  } else
    THROW("Unknown inference type; cannot proceed!");
}

bool PseudoMTG::filter(const Item* item1in,
		       const Item* item2in,
		       const Pair<bool,PAV>& arrg){
  //  return false;
  /////////////////
    //
    // This section has a first cut approach to implementing
    // Binarization techniques 
    //
    ////////////////
    PAV pav;
    const Item* item1;
    const Item* item2;
    if(arrg.first){
      item1 = item1in;
      item2 = item2in;
    }
    else{
      item1 = item2in;
      item2 = item1in;
    }
    

    pav = arrg.second;
    vector<vector<unsigned int> > vec1(pav.arity());
    vector<vector<unsigned int> > vec2(pav.arity());
    //1 means item1 LEFT or ABOVE item2.
    //2 means item2 ABOVE or LEFT of item1

    if(debug(4)){
      cerr << "CheckFilter of " << pav; item1->dump(cerr); item2->dump(cerr);cerr << endl;
    }

    vector<role_t> vect(2,2);
    vector<unsigned int> ends(2,0);
    vector<unsigned int> vecSeg(2,0);  
    
    PAV pavEp;
    PAV pavt;

    bool found = false;
  
    if(!dynamic_cast<const LexNT*>(((Hedge*)item1)->getNTLink().getNT(0))){
      if(dynamic_cast<const LexNT*>(((Hedge*)item2)->getNTLink().getNT(0))){
	pavEp = item2->getInference()->getPAV();
	pavt = item1->getInference()->getPAV();
	found = true;
      }
    }
    else if(!dynamic_cast<const LexNT*>(((Hedge*)item2)->getNTLink().getNT(0))){
      if(dynamic_cast<const LexNT*>(((Hedge*)item1)->getNTLink().getNT(0))){
	pavEp = item1->getInference()->getPAV();
	pavt = item2->getInference()->getPAV();
	found = true;
      }
    }

    if(found){
      vect[0] = 1;
      ends[0] = 0;
      ends[1] = 3;
      //if epsilon in first dimension
      // which means no 1 in first dimension on epsilon side
      if(pavEp.endsBegWith(vect,vecSeg,ends)){
	vect[0] = 2;
	ends[0] = 0;
	ends[1] = 3;
	//if no 2 in first dimension of non ep
	if(pavt.endsBegWith(vect,vecSeg,ends)){
	  found = false;
	}
      }
    
      //ataching left
      else{
	vect[0] = 2;
	ends[0] = 0;
	ends[1] = 3;
	//if no 2 in first dimension
	if(pavt.endsBegWith(vect,vecSeg,ends)){
	  found = false;
	}else{
	  vect[1] = 2;
	  ends[1] = 0;
	  ends[0] = 3;
	  //if no 2 in 2nd dimension
	  if(pavt.endsBegWith(vect,vecSeg,ends)){
	    found = false;
	  }else{
	    //no 1 in 2nd dimension
	    vect[1] = 1;
	    if(pavt.endsBegWith(vect,vecSeg,ends)){
	      found = false;
	    }      
	  }
	}
      }
      if(found){
	if(debug(3)){         
	  cerr << "pavt " << pavt << endl;
	  cerr << "pavEp " << pavEp << endl;
	  cerr << "ep ! attaching to other Ep." << endl;
	}
	return true;
      }
    }
  
    bool adFound = false;
	
    int last = 0;
    int num2s = 0;
    int num1s = 0;
    int num0s = 0;
    bool nonZero1 =false;
    for(unsigned int j = 0; j < pav.getPA(0).size();j++){
      if(pav.getPA(0)[j] == 1){
	if(last == 2){
	  adFound = true;
	  nonZero1 = true;
	  vec1[num2s-1].push_back(2);
	}
	num1s++;
      }
      else if(pav.getPA(0)[j] == 2){
	if(last == 1){
	  adFound = true;
	  nonZero1 = true;
	  vec1[num1s-1].push_back(1);
	}
	num2s++;
      }
      else{
	num0s++;
	vec1[num0s-1].push_back(0);
      }
      last = pav.getPA(0)[j];
    }
  
    last = 0;
    num2s = 0;
    num1s = 0;
    num0s = 0;
    bool nonZero2 =false;
    for(unsigned int j = 0; j < pav.getPA(1).size();j++){
      if(pav.getPA(1)[j] == 1){
	if(last == 2){
	  adFound = true;
	  nonZero2 = true;
	  vec2[num2s-1].push_back(2);
	}
	num1s++;
      }
      else if(pav.getPA(1)[j] == 2){
	if(last == 1){
	  adFound = true;
	  nonZero2 = true;
	  vec2[num1s-1].push_back(1);
	}
	num2s++;
      }
      else{
	num0s++;
	vec2[num0s-1].push_back(0);
      }
    
      last = pav.getPA(1)[j];
    }
  
    vector<size_t> item1Sizes = item1->getSizes();
    vector<size_t> item2Sizes = item2->getSizes();

    if(!adFound){
      if((item1Sizes[0] >= 1 && item1Sizes[1] >= 1) &&(item2Sizes[0] >= 1 && item2Sizes[1] >= 1)){
	return true;
      }
    }

    //now we have a vector showing the orientations.  
    // A 1 means there was a 12 in the PAV, and a 2 means there was a 21. A 0 means there was neither.
  
    //the next step is to take cross product and do appropriate checks for each circumstance.
    for(unsigned int i = 0; i < vec1.size(); i++){
      for(unsigned int ii = 0;  ii < vec1[i].size();ii++){
	for(unsigned int j = 0; j < vec2.size(); j++){
	  for(unsigned int jj = 0; jj <  vec2[j].size();jj++){
	    vector<unsigned int> vecSeg;
	    vecSeg.push_back(i);
	    vecSeg.push_back(j);
	  
	    if(debug(5)){
	      cerr << "FILTER INFO" << endl;
	      cerr << "filt pos 1: " << vec1[i][ii] << endl;
	      cerr << "filt pos 2: " << vec2[j][jj] << endl;
	      cerr << "item1: "; item1->dump(cerr); cerr <<endl;
	      cerr << "item2: "; item2->dump(cerr); cerr <<endl;
	    }
	  
	    if(vec1[i][ii] == 1 && vec2[j][jj] == 1){
	      if(item1Sizes[0] <= 1 && item1Sizes[1] <= 1)
		break;
	      vector<role_t> vect(2,2);
	      //ends in 2 
	      //ends in 2
	      vector<unsigned int> ends(2,2);

	      PAV pavt = item1->getInference()->getPAV();
	      if(debug(4)){
		cerr << "inside PAV:" << pavt;
	      }
	     
	      //if ends in 2,2
	      if(pavt.endsBegWith(vect,vecSeg,ends)){
		vect[0] = 1;
		ends[0] = 0;
		ends[1] = 3;
		if(!pavt.endsBegWith(vect,vecSeg,ends)){
		  vect[1] = 1;
		  ends[1] = 0;
		  ends[0] = 3;
		  if(!pavt.endsBegWith(vect,vecSeg,ends)){
		    return true;
		  }
		}
	      }

	      //COMMENT OUT BELOW


	      // 	    vect[0] = 2;
	      // 	    ends[0] = 0;
	      // 	    ends[1] = 3;
	    
	      // 	    //if no 2 in first dimension
	      // 	    if(pavt.endsBegWith(vect,vecSeg,ends)){
	      // 	      return true;
	      // 	    }

	      // 	    vect[0] = 1;
	      // 	    ends[0] = 0;
	      // 	    ends[1] = 3;
	      // 	    //if no 1 in first dimension
	      // 	    if(pavt.endsBegWith(vect,vecSeg,ends)){
	      // 	      return true;
	      // 	    }

	      // 	    vect[0] = 0;
	      // 	    vect[1] = 2;
	      // 	    ends[0] = 3;
	      // 	    ends[1] = 0;
	      // 	    //if no 2 in second dimension
	      // 	    if(pavt.endsBegWith(vect,vecSeg,ends)){
	      // 	      return true;
	      // 	    }

	      // 	    vect[0] = 0;
	      // 	    vect[1] = 1;
	      // 	    ends[0] = 3;
	      // 	    ends[1] = 0;
	      // 	    //if no 1 in second dimension
	      // 	    if(pavt.endsBegWith(vect,vecSeg,ends)){
	      // 	      return true;
	      // 	    }

	    }
	    else if (vec1[i][ii] == 1 && vec2[j][jj] == 2){
	      if(item1Sizes[0] <= 1 && item1Sizes[1] <= 1)
		break;
	      vector<role_t> vect(2,2);
	      vector<unsigned int> ends(2,2);
	      ends[1] = 1;
	    
	      PAV pavt = item1->getInference()->getPAV();
	      if(debug(4)){
		cerr << "1 2 section: "  << endl;
		cerr << "insidePAV " << pavt << endl;
	      }
	      if(pavt.endsBegWith(vect,vecSeg,ends)){
		vect[0] = 1;
		ends[0] = 0;
		ends[1] = 3;
		if(!pavt.endsBegWith(vect,vecSeg,ends)){
		  vect[1] = 1;
		  ends[1] = 0;
		  ends[0] = 3;
		  if(!pavt.endsBegWith(vect,vecSeg,ends)){
		    return true;
		  }	      
		}
	      }
	      //	    COMMENT
	      // 	    vect[0] = 0;
	      // 	    vect[1] = 2;
	      // 	    ends[0] = 3;
	      // 	    ends[1] = 0;

	      // 	    if(pavt.endsBegWith(vect,vecSeg,ends)){
	      // 	      return true;
	      // 	    }

	      // 	    vect[0] = 0;
	      // 	    vect[1] = 1;
	      // 	    ends[0] = 3;
	      // 	    ends[1] = 0;

	      // 	    if(pavt.endsBegWith(vect,vecSeg,ends)){
	      // 	      return true;
	      // 	    }

	    }
	    else if (vec1[i][ii] == 2 && vec2[j][jj] == 1){
	      if(item2Sizes[0] <= 1 && item2Sizes[1] <= 1)
		break;
	      vector<role_t> vect(2,2);
	      vector<unsigned int> ends(2,2);
	      ends[1] = 1;

	      PAV pavt = item2->getInference()->getPAV();
	      if(debug(4)){
		cerr << "insidePAV:" << pavt;
	      }
	      if(pavt.endsBegWith(vect,vecSeg,ends)){
		vect[0] = 1;
		ends[0] = 0;
		ends[1] = 3;
		if(!pavt.endsBegWith(vect,vecSeg,ends)){
		  vect[1] = 1;
		  ends[1] = 0;
		  ends[0] = 3;
		  if(!pavt.endsBegWith(vect,vecSeg,ends)){
		    return true;
		  }
		}
	      }
	    }
	    else if (vec1[i][ii] == 2 && vec2[j][jj] == 2){
	      if(item2Sizes[0] <= 1 && item2Sizes[1] <= 1)
		break;
	      vector<role_t> vect(2,2);
	      vector<unsigned int> ends(2,2);

	      PAV pavt = item2->getInference()->getPAV();
	      if(debug(4)){
		cerr << "insidePAV:" << pavt;
	      }
	      if(pavt.endsBegWith(vect,vecSeg,ends)){
		vect[0] = 1;
		ends[0] = 0;
		ends[1] = 3;
		if(!pavt.endsBegWith(vect,vecSeg,ends)){
		  vect[1] = 1;
		  ends[1] = 0;
		  ends[0] = 3;
		  if(!pavt.endsBegWith(vect,vecSeg,ends)){
		    return true;
		  }
		}
	      }
	    }
	    else if (vec1[i][ii] == 0 && vec2[j][jj] == 2){
	      if(nonZero1 && nonZero2)
		break;
	      if(item1Sizes[0] <= 1 && item1Sizes[1] <= 1)
		break;
	      vector<role_t> vect(2,1);
	      vector<unsigned int> ends(2,1);

	      PAV pavt = item1->getInference()->getPAV();
	      if(debug(4)){
		cerr << "insidePAV:" << pavt;
	      }
	      if(pavt.endsBegWith(vect,vecSeg,ends)){
		return true;
	      }

	      vect[0] = 0;
	      vect[1] = 2;
	      ends[0] = 3;
	      ends[1] = 0;

	      if(pavt.endsBegWith(vect,vecSeg,ends)){
		return true;
	      }

	      vect[0] = 0;
	      vect[1] = 1;
	      ends[0] = 3;
	      ends[1] = 0;

	      if(pavt.endsBegWith(vect,vecSeg,ends)){
		return true;
	      }

	    }
	    else if (vec1[i][ii] == 0 && vec2[j][jj] == 1){
	      if(nonZero1 && nonZero2)
		break;	    
	      if(item1Sizes[0] <= 1 && item1Sizes[1] <= 1)
		break;
	      vector<role_t> vect(2,2);
	      vector<unsigned int> ends(2,2);
	      //if ends in 2 2
	      PAV pavt = item1->getInference()->getPAV();
	      if(debug(4)){
		cerr << "insidePAV:" << pavt;
	      }
	      if(pavt.endsBegWith(vect,vecSeg,ends)){
		//must be 1 in both dimensions
		vect[0] = 1;
		ends[0] = 0;
		ends[1] = 3;
		if(!pavt.endsBegWith(vect,vecSeg,ends)){
		  vect[1] = 1;
		  ends[1] = 0;
		  ends[0] = 3;
		  if(!pavt.endsBegWith(vect,vecSeg,ends)){
		    return true;
		  }
		  return true;
		}
	      }
	      vect[0] = 1;
	      ends[0] = 1;
	      vect[1] = 1;
	      ends[1] = 2;
	    
	      if(pavt.endsBegWith(vect,vecSeg,ends)){
		//must be 2 in both dimensions
		vect[0] = 2;
		ends[0] = 0;
		ends[1] = 3;
		if(!pavt.endsBegWith(vect,vecSeg,ends)){
		  vect[1] = 2;
		  ends[1] = 0;
		  ends[0] = 3;
		  if(!pavt.endsBegWith(vect,vecSeg,ends)){
		    return true;
		  }
		}
	      }
	    }	  
	  }
	}
      }
    }
    return false;
}

//! Returns all Z's given X and Y such that Z->XY (or Z->YX).
//! \param parents : the parent constituents that will be returned.
void PseudoMTG::getPossibleCons(const Item* item1,
				const Item* item2,
				const Pair<bool,PAV>& arrg,
				vector<Item*> & parents){
  const NodeRefHedge * ant1 = static_cast<const NodeRefHedge*>(item1);
  const NodeRefHedge * ant2 = static_cast<const NodeRefHedge*>(item2);

  if(debug(3)){
    NTLink label1 = ant1->getNTLink();
    NTLink label2 = ant2->getNTLink();
    cerr << "PseudoMTG::getPossibleCons for " << endl;
    cerr << "ANT1: " << *ant1 << "ANT2: " << *ant2 
	 << "heir: " << (int)arrg.first << " PAV: " << arrg.second << endl;
  }

  if(m_rightBin)
    if(filter(item1, item2,arrg)){
      if(debug(3)){cerr << "FILTERED = YES" << endl;}
      return;
    }
   
  //vector of sets which will hold 1D items, one generated
  //from each dimensions contraining grammar.
  vector<Item*> possiblePar[arity()];

  /*
   *First call getPossibleCons on each subgrammar
   */
  for(unsigned short d = 0; d < arity(); ++d)
    // call each model's getPossible parent.
    m_monoGrammars[d]->getBinaryCons(ant1, ant2, arrg, possiblePar[d]);

  // check to see if we should bother
  for(unsigned int i = 0; i < arity(); i++){
    //if a component is empty (i.e. no parents found in that component.)
    if(possiblePar[i].empty()){
      if(debug(5))
	cerr << "PseudoMTG::getPossibleCons: no possible cons on dim " << i << endl;
      vector<Item*>::iterator it;
      for(unsigned int j = 0; j < arity(); j++)
	for(it = possiblePar[j].begin(); it != possiblePar[j].end(); it++)
	  //Delete the hedges we are not going to use.
	  delete *it;
	      
      return;
    }
  }

  vector<vector<Item*> > vset(arity());
  for(unsigned int i = 0; i < arity(); i++){
    //Now we set the Label back to its original lexical self, 
    //and pull out the HeirRole in each dimension.
    vector<Item*>::iterator it;
    for(it = possiblePar[i].begin(); it != possiblePar[i].end(); it++){
      vset[i].push_back(*it);
    }
  }

  /*
   * Take the cross product of items in each component
   */
  vector<Item*> splitItem(2);
  CPIterator< vector<Item* >, vector<Item* > >
    cp(vset[0].begin(), vset[0].end(), vset[1].begin(), vset[1].end());
  while (cp.valid()) {
    splitItem[0] = cp.firstDeref();
    splitItem[1] = cp.secondDeref();

    if(debug(5)){
      bool tmp = GlobalVocab::s_noVocabsOut;
      GlobalVocab::s_noVocabsOut = false;
      cerr << "STATS ON cross-product entry" << endl;
      cerr << "Item1: " << *((NodeRefHedge*)splitItem[0]);
      cerr << "Item2: " << *((NodeRefHedge*)splitItem[1]);
      GlobalVocab::s_noVocabsOut = tmp;
    }
    
    //create the consequent
    NodeRefHedge* conseq = new NodeRefHedge(splitItem);

    HeirRoleVector hv(arity(),0);
    if(dynamic_cast<const LexNT*>(conseq->getNTLink().getNT(0))) {
      if(conseq->getNTLink().getTLink()[0] != 0){
	if(conseq->getNTLink().getTLink()[0] == ant1->getNTLink().getTLink()[0]){
	  hv.setRole(0,1);
	} else if (conseq->getNTLink().getTLink()[0] == ant2->getNTLink().getTLink()[0]){
	  hv.setRole(0,2);
	}
      }
      if(conseq->getNTLink().getTLink()[1] != 0){
	if(conseq->getNTLink().getTLink()[1] == ant1->getNTLink().getTLink()[1]){
	  hv.setRole(1,1);
	} else if (conseq->getNTLink().getTLink()[1] == ant2->getNTLink().getTLink()[1]){
	  hv.setRole(1,2);
	}
      }
    }
    
  
    if(debug(5))
      cerr << "PsMTG::getPossCons will return "<< *conseq << endl;
    
    //store the item we just generated
    parents.push_back(conseq);
    
    cp.increment();
  }
  vector<Item*>::iterator it2;
  for(unsigned int j = 0; j < arity(); j++){
    for(it2 = vset[j].begin(); it2 != vset[j].end(); it2++){
      //Delete the hedges we are not going to use.
      delete *it2;
    }
  }
}

  /// Returns the cost (such as negative log prob) of an inference
  /// by default, use the corresponding production rule
SCORE PseudoMTG::getInfScore(Inference* inf) const{
  
  HeirRoleVector hv;
  if(HeadCompose* h = dynamic_cast<HeadCompose*>(inf)){
    hv = h->heirIndices();
  }else{
    //    THROW("getInfScore on non headed Compose in PseudoMTG");
  }

  const FanoutVect fov = inf->getConseq()->getFanoutV();
  char totalFanOut = accumulate(fov.begin(), fov.end(), 0);
  SCORE fanOutCost = totalFanOut - inf->getConseq()->numActiveDims();
  assert(fanOutCost >= 0.0);
  
  SCORE w = getW2WCost(hv, static_cast<const NodeRefHedge*>(inf->getAnt()[0]),  static_cast<const NodeRefHedge*>(inf->getAnt()[1]));
  
  
  return (w + fanOutCost);
}


//! Returns all Z's given X and Y such that Z->XY (or Z->YX).
//! \param parents : the parent constituents that will be returned.
void PseudoMTG::getPossibleConsHeaded(const Item* item1,
				      const Item* item2,
				      const Pair<bool,PAV>& arrg,
				      vector<pair<HeirRoleVector,Item*> > & parents){
  const NodeRefHedge * ant1 = static_cast<const NodeRefHedge*>(item1);
  const NodeRefHedge * ant2 = static_cast<const NodeRefHedge*>(item2);
  const bool& theRole = arrg.first;
  const PAV& thePAV = arrg.second;

  if(debug(3)){
    NTLink label1 = ant1->getNTLink();
    NTLink label2 = ant2->getNTLink();
    cerr << "PseudoMTG::getPossibleConsHeaded for " << endl;
    cerr << "ANT1: " << *ant1 << "ANT2: " << *ant2
	 << "heir: " << (int)arrg.first << " PAV: " << thePAV << endl;
  }

  if(m_rightBin)
    if(filter(item1, item2, arrg)){
      if(debug(3)){cerr << "FILTERED = YES" << endl;}
      return;
    }
   
  //vector of sets which will hold 1D items, one generated
  //from each dimension's constraining grammar
  vector<HRVIpair > possiblePar[arity()];

  /*
   *First call getPossibleCons on each subgrammar
   */
  for(unsigned short d = 0; d < arity(); ++d)
    // call each model's getPossible parent.
    m_monoGrammars[d]->getHeadedBinaryCons(item1, item2, arrg, possiblePar[d]);

  // check to see if we should bother
  for(unsigned int i = 0; i < arity(); i++){
    //if a component is empty (i.e. no parents found in that component.)
    if(possiblePar[i].empty()){
      if(debug(5))
	cerr << "PseudoMTG::getPossibleConsHeaded: no possible cons on dim " << i << endl;
      vector<pair<HeirRoleVector,Item*> >::iterator it;
      for(unsigned int j = 0; j < arity(); j++)
	for(it = possiblePar[j].begin(); it != possiblePar[j].end(); it++)
	  //Delete the hedges we are not going to use.
	  delete it->second;
	      
      return;
    }
  }

  // the short in the pair is the heir info
  vector<vector<pair<short,Item*> > > vset(arity());
  for(unsigned int i = 0; i < arity(); i++){
    //Now we pull out the HeirRole in each dimension.
    vector<pair<HeirRoleVector,Item*> >::iterator it;
    for(it = possiblePar[i].begin(); it != possiblePar[i].end(); it++){
      short s = it->first.getRole(0);
      vset[i].push_back(pair<short,Item*>(s,it->second));
    }
  }

  /*
   * Take the cross product of items in each component
   */
  HeirRoleVector hv(2,0);
  vector<Item*> splitItem(2);
  CPIterator< vector<pair<short,Item*> >, vector<pair<short,Item*> >, pair<short,Item*> >
    cp(vset[0].begin(), vset[0].end(), vset[1].begin(), vset[1].end());
  while (cp.valid()) {
    hv.setRole(0,cp.firstDeref().first);
    hv.setRole(1,cp.secondDeref().first);
    
    splitItem[0] = cp.firstDeref().second;
    splitItem[1] = cp.secondDeref().second;

    if(debug(5)){
      bool tmp = GlobalVocab::s_noVocabsOut;
      GlobalVocab::s_noVocabsOut = false;
      cerr << "STATS ON cross-product entry" << endl;
      cerr << "Item1: " << *((NodeRefHedge*)splitItem[0]);
      cerr << "Item2: " << *((NodeRefHedge*)splitItem[1]);
      GlobalVocab::s_noVocabsOut = tmp;
    }
    
    /*
     * Examine to see if we found heads in at least one of the expanding
     * dimensions.  (Note that an heirRole of -1 means "possible head"
     * while 0 means no head.
     */
    bool foundHead = false;
    bool atLeastOneExpandingDim = false;
    bool negFound = false;
    //for all components
    // the termination condition of this loop works even if item1
    // arity is more than item2 arity 
    for(arity_t i = 0; i < item1->arity(); i++){
      //if both input spans are not empty
      if(!item1->getDSV()->empty(i) &&!item2->getDSV()->empty(i)){
	atLeastOneExpandingDim = true;
	//if its > 0, we have a confirmed head.
	//If it's < 0, we have a possible head
	if(hv.getRole(i) != 0){
	  foundHead = true;
	  //a possible head means we have to do more checking later.
	  //the 1D grammar was not able to determine if it's a head
	  //conclusively.
	  if(hv.getRole(i) == -1)
	    negFound = true;
	  else
	    break;
	}
      }
    }
    //if no expandingDimensions have heads, we cannot combine the items
    if(!foundHead && atLeastOneExpandingDim && !m_ignoreHeads){
       if(debug(5)){
      	cerr << endl << "no expanding dimensions found.  ABORT COMBINE"  << endl;
       }
        cp.increment();
       continue;
    }
    //if a "possible heir" was found, there must be at least one confirmed heir
    // in at least one dimension.
    if(negFound){
      bool posFound = false;
      // the termination condition of this loop 
      // works even if item1 arity is more than item2 arity
      for(arity_t i = 0; i < item1->arity(); i++){
	if(hv.getRole(i) > 0){
	  posFound = true;
	  break;
	}
      }
      if(!posFound && !m_ignoreHeads){
	//no confirmed heirs any where.
	// so two items cannot combine
	if(debug(5)){
	  cerr << "no confirmed heirs found.  ABORT COMBINE"  << endl;
	}

	hv.setRole(0,1);
	hv.setRole(1,1);
	cp.increment();
	continue;
      }
    }

    //create the consequent
    NodeRefHedge* conseq = new NodeRefHedge(splitItem);
    Vector<DSV>* pOutHDSVs = NULL;
    bool compatible = hDSV::crossProd(splitItem[0]->getHiddenDSVs(), 
				      splitItem[1]->getHiddenDSVs(), pOutHDSVs);
    if (compatible)
      conseq->setHiddenDSVs(pOutHDSVs);

    //items array holds both input items
    const Item* items[2]; items[0] = item1; items[1] = item2;
    // Identify heads and propogate correct information in each dimension
    for(arity_t i = 0; i < item1->arity(); i++){
      //if the heir role is empty or unknown
      if(hv.getRole(i) <= 0){
	//if both antecedents have empty spans
	if(item1->getSizes()[i] == 0 && item2->getSizes()[i] == 0)
	  assert(hv.getRole(i) == 0);
	//else if item1 has an empty span, the heir should be item 2
	else if(item1->getSizes()[i] == 0 && not item2->getSizes()[i] == 0){
	  //set the correct lexical head.
	  conseq->getNTLink().setNT(i,((Hedge*)item2)->getNTLink().getNT(i)->clone());
	  hv.setRole(i,2);
	}
	//else if item2 has an empty span, the heir should be item1
	else if( not item1->getSizes()[i] == 0 && item2->getSizes()[i] == 0){
	  //set the correct lexical head.
	  conseq->getNTLink().setNT(i,((Hedge*)item1)->getNTLink().getNT(i)->clone());
	  hv.setRole(i,1);
	}

	//else neither have empty span
	//just copy the heir from the other component into this one if it is present.
	// if its not, assume 1.
	else{
	  if(!m_ignoreHeads){
	    hv.setRole(i,hv.getRole(1-i));
	    if(hv.getRole(i) == 1){
	      conseq->getNTLink().setTerminal(ant1->getNTLink().getTLink()[i],i);
	    }
	    else if (hv.getRole(i) == 2){
	      conseq->getNTLink().setTerminal(ant2->getNTLink().getTLink()[i],i);
	    }
	  }
	  else{
	    if(hv.getRole(1-i)!= -1 and hv.getRole(1-i)!= 0){
	      hv.setRole(i,hv.getRole(1-i));
	    }
	    else{ //assume 1 
	      hv.setRole(i,1);
	    }
	    if(hv.getRole(i) == 1){
	      conseq->getNTLink().setTerminal(ant1->getNTLink().getTLink()[i],i);
	      vector<short> v = conseq->getHeads();
	      v[i] = ant1->getHeads()[i];
	      conseq->setHeads(v);
	    }
	    else if (hv.getRole(i) == 2){
	      conseq->getNTLink().setTerminal(ant2->getNTLink().getTLink()[i],i);
	      vector<short> v = conseq->getHeads();
	      v[i] = ant2->getHeads()[i];
	      conseq->setHeads(v);
	    }
	  }
	  SynCat sc =  conseq->getNTLink().getNT(i)->getSynCat();
	  Nonterminal* n = ((Hedge*)items[hv.getRole(i)-1])->getNTLink().getNT(i)->clone();
	  n->setSynCat(sc);
	  conseq->getNTLink().setNT(i,n);
	}
      }
    }
      
    // change labels as needed
    computeLabels( *ant1, *ant2, theRole, thePAV, hv, *conseq);
 
    SCORE s = getW2WCost(hv, ant1, ant2);
    // compute fan-out cost
    const FanoutVect fov = conseq->getFanoutV();
    char totalFanOut = accumulate(fov.begin(), fov.end(), 0);
    // additional cost of 1.0 for every gap
    SCORE fanOutCost = totalFanOut - conseq->numActiveDims();
    assert(fanOutCost >= 0.0);
    conseq->setModelScore(s + fanOutCost);
    if(debug(5))
      cerr << "PsMTG::getPossConsH will return HV [" << hv << "] with " << *conseq;
    
    //store the item we just generated
    parents.push_back(HRVIpair(hv,conseq));
    
    cp.increment();
  }
  vector<pair<short,Item*> >::iterator it2;
  for(unsigned int j = 0; j < arity(); j++){
    for(it2 = vset[j].begin(); it2 != vset[j].end(); it2++){
      //Delete the hedges we are not going to use.
      delete it2->second;
    }
  }
}

SCORE PseudoMTG::getW2WCost(const HeirRoleVector& hv,
			    const NodeRefHedge* ant1,const NodeRefHedge* ant2 ) const{
  PosLinkType tl;      // THIS IS THE LINK WE ARE TESTING  

  bool bothLexicalized = false;
  if(!m_isHeaded){
    if(dynamic_cast<const LexNT*>(ant1->getNTLink().getNT(0)) 
       && dynamic_cast<const LexNT*>(ant2->getNTLink().getNT(0))){
      bothLexicalized = true;
    }
    else if(dynamic_cast<const LexNT*>(ant1->getNTLink().getNT(0)) 
	    && !dynamic_cast<const LexNT*>(ant2->getNTLink().getNT(0))){
      tl.first = ant1->getHeads()[0];
      tl.second = ant1->getHeads()[1];
    }
    else if(!dynamic_cast<const LexNT*>(ant1->getNTLink().getNT(0)) 
	    && dynamic_cast<const LexNT*>(ant2->getNTLink().getNT(0))){
      tl.first = ant2->getHeads()[0];
      tl.second = ant2->getHeads()[1];
    }
    else {
      return 0;
    }
  }

  if(m_isHeaded || bothLexicalized){
    vector<short> ant1Heads = ant1->getHeads();
    vector<short> ant2Heads = ant2->getHeads();
    
    int deltaPos;
       
    switch(getHeirLayout(ant1Heads,ant2Heads,deltaPos)){
    case NODELTAS:
      //no deltas in this production.
      //A B/ A B  (no deltas)
      // no prob score needed
      // normally, we would add antecedents
      return 0;
      
      //CASE  A DELTA / DELTA A (deltaPos = 1)
      //or DELTA A / A DELTA (deltaPos = 0)
    case CROSSWAYS:
      if(deltaPos == 0){
	tl.first = ant2Heads[0];
	tl.second = ant1Heads[1];
      }
      else{
	tl.first = ant1Heads[0];
	tl.second = ant2Heads[1];
      }
      break;
      // A DELTA / B DELTA
      // or DELTA A / DELTA B
    case WHOLEDIM:
      //      if(!m_isHeaded){rerurn POSINF};
      if(hv.getRole(0) == 1 || hv.getRole(1) == 1 ){
	tl.first = ant2Heads[0];
	tl.second = ant2Heads[1];
      }
      else if(hv.getRole(0) == 2 || hv.getRole(1) == 2){
	tl.first = ant1Heads[0];
	tl.second = ant1Heads[1];
      }else if(bothLexicalized){
	return POSINF;
      }
      //cerr << "HeirVector: " << hv << endl;
      //THROW("Odd HeirVector in W2WCost: ");
      
      break;
      
      //DELTA A / B B
      //A DELTA / B B
      //A A / B DELTA
      //A A / DELTA B
    case ONEDELTA:
      if(deltaPos == 0 || deltaPos == 1){
	tl.first = ant1Heads[0];
	tl.second = ant1Heads[1];
      }
      else if (deltaPos == 2 || deltaPos == 3){
	tl.first = ant2Heads[0];
	tl.second = ant2Heads[1];
      }
    }
  }
  
  SCORE w2wScore;
  
  if(m_w2wTM){
    try {
      // assumes the w2wTM is already in negative log space
      w2wScore = m_w2wTM->lookup(tl);
    }catch(KeyNotFound<PosLinkType> ex1){
      // if w2wlink cannot be found in the w2wTM, then cost is inf.
      w2wScore = POSINF;
    }
  }
  else{
    w2wScore = 0;
  }

  // we are returning a neg logprob, so this should be > 0
  if (w2wScore < 0) {
    THROW("TCMTG: w2wScore for tl " << tl << " is "
	  << w2wScore << " but should be a negative log-prob!" << endl
	  << *m_w2wTM << endl);
  };

  // check for NaN
  if (w2wScore != w2wScore)
    throw NumericalError();

  if (debug(5)) {
    // Dereferencing the vocabs won't work cuz the noderefhedge here
    // is always 1D, even if it pertains to the 2nd dimension.  In
    // that case, NTLink::put has no way of knowing *which* vocab to
    // deref.
    static int invocationNumber = 0;
    ++invocationNumber;
    bool tmp = GlobalVocab::s_noVocabsOut;
    GlobalVocab::s_noVocabsOut = false;
    dout() << "W2W cost for PosLink " << tl << " = " << w2wScore
	   << " at invocation " << invocationNumber << endl;
    dout() << "heir role was " << hv << endl;
    GlobalVocab::s_noVocabsOut = tmp;
  }

  return w2wScore;
}

/** This function takes in two tlinks and returns the heir
    layout. This heir layout is a fingerprint to decide how to apply
    the w2w models. posDelta tells more specifically where the first
    delta occurs. */
PseudoMTG::HeirLayout PseudoMTG::getHeirLayout(const vector<short>& ant1Heads, 
					       const vector<short>& ant2Heads,
					       int& posDelta) const {
  if(arity() != 2)
    THROW( "PseudoMTG::getHeirLayout for dim != 2 not implemented");

  //CASE 0: A B/ A B (no deltas)
  if(ant1Heads[0] != UNDEF_POS &&
     ant1Heads[1] != UNDEF_POS &&
     ant2Heads[0] != UNDEF_POS &&
     ant2Heads[1] != UNDEF_POS ){
    // there is no w2w to apply here it does not effect the probability.
    return NODELTAS;
  }

  //CASE 1: A DELTA / DELTA A
  if(ant1Heads[0] != UNDEF_POS &&
     ant1Heads[1] == UNDEF_POS &&
     ant2Heads[0] == UNDEF_POS &&
     ant2Heads[1] != UNDEF_POS ){
    posDelta = 1;
    return CROSSWAYS;
  }

  //CASE 1:DELTA A / A DELTA
  if(ant1Heads[0] == UNDEF_POS &&
     ant1Heads[1] != UNDEF_POS &&
     ant2Heads[0] != UNDEF_POS &&
     ant2Heads[1] == UNDEF_POS ){
    posDelta = 0;
    return CROSSWAYS;
  }

  //CASE 2: A DELTA / B DELTA
  if(ant1Heads[0] != UNDEF_POS &&
     ant1Heads[1] == UNDEF_POS &&
     ant2Heads[0] != UNDEF_POS &&
     ant2Heads[1] == UNDEF_POS ){
    return WHOLEDIM;
  }

  //CASE 2: DELTA A / DELTA B
  if(ant1Heads[0] == UNDEF_POS &&
     ant1Heads[1] != UNDEF_POS &&
     ant2Heads[0] == UNDEF_POS &&
     ant2Heads[1] != UNDEF_POS ){
    return WHOLEDIM;
  }

  //CASE 3:DELTA A / B B
  if(ant1Heads[0] == UNDEF_POS){
    posDelta = 0;
    return ONEDELTA;
  }
  //CASE 3:A DELTA / B B
  else if(ant1Heads[1] == UNDEF_POS){
    posDelta = 1;
    return ONEDELTA;
  }
  //CASE 3:A A / DELTA B
  else if(ant2Heads[0] == UNDEF_POS){
    posDelta = 2;
    return ONEDELTA;
  }
  //CASE 3: A A/ B DELTA
  else if(ant2Heads[1] == UNDEF_POS){
    posDelta = 3;
    return ONEDELTA;
  } else
    THROW("PseudoMTG::getHeirLayout : invalid" << endl);

  //will never reach here:
  return ONEDELTA;
}

void PseudoMTG::clear() {
  _super::clear();

  if(m_pW2WModelStream){
    rewind(*m_pW2WModelStream);
  }

  for(arity_t d = 0; d < arity(); ++d){
    m_monoGrammars[d]->clear();
  }
}

/// Compute the nonterminal label for a consequent item, given its
/// d-span, its PAV, its lexical heads, and its two antecedents. In
/// other words, all the info is expected to be pre-computed except
/// the nonterminal labels.
void PseudoMTG::computeLabels(const NodeRefHedge& ant1, const NodeRefHedge& ant2,
			      const bool theRole, const PAV& thePAV, 
			      const HeirRoleVector& hVec, NodeRefHedge& conseq) const {

  arity_t arity = conseq.arity();
  arity_t d, emptyHeirD = 0;
  bool emptyHeirFound = false;

  if (debug(3)) {
    cerr << "PsMTG::cL Ant1: " << ant1;
    cerr << "PsMTG::cL Ant2: " << ant2;
    cerr << "PsMTG::cL Conseq: " << conseq;
    cerr << "PsMTG::cL HeirRoleVector: " << hVec << endl;
  } else if (debug(2)) {
    NTLink lnt = conseq.getNTLink();
    cerr << "PsMTG::cL Incoming Label: " << lnt << endl;
  };

  // main loop
  for(d = 0; d < arity; ++d) {
    // If the heir is empty on this dimension, then we'll copy the label
    // from the first active dimension, keeping the epsilon head.
    if (hVec.getRole(d) == 1 && ant1.getNTLink(d)->isInactive()
	|| hVec.getRole(d) == 2 && ant2.getNTLink(d)->isInactive()) {
      if((ant1.getNTLink(d)->isInactive()) && (ant2.getNTLink(d)->isInactive()))
	continue;
      // This assumes that the heir cannot be empty on all dimensions.
      assert(! emptyHeirFound);
      emptyHeirD = d;
      emptyHeirFound = true;
      continue;
    } else
      // otherwise, compute it normally
      computeLabelsHelper(ant1, ant2, theRole, thePAV, conseq, d, hVec.getRole(d));
  };

  //  bool exists;
  if (emptyHeirFound) {
    // Find the first dimension on which both antecedents are active.
    for(d = 0; d < arity; ++d) {
      if (! ant1.getSizes()[d] == 0 && ! ant2.getSizes()[d] == 0){
	// firstActiveD = d;
	// Copy the NT from this dimension to the empty heir.
	VocabularyIndex nt = conseq.getNTLink(d)->nt();
	// in case it hasn't been seen on the other dimension, add it
	// to the relevant vocab
	// (no longer necessary since vocabs were merged -IDM)
	//	const string* label = GlobalVocab::s_synCatSet->getWord(nt, exists);
	//	assert(exists);
	//      VocabularyIndex newNT = GlobalVocab::s_synCatSet->addWord(*label);

	LexNT* newLNT = new LexNT(nt, UNDEF_T);
	conseq.setLabel(emptyHeirD, newLNT);
	delete newLNT;
	emptyHeirFound = false;
	break;
      };
    };
    // if emptyHeirFound is still set at this point,
    // then there is no fully active dimension, so bail
    if(m_isHeaded)   
      assert(! emptyHeirFound);
  };

  if (debug(2)) {
    cerr << "PsMTG::cL Outgoing Label= " <<conseq.getNTLink() << endl;
  };
}

// this helper assumes that the relevant dimension is active in at least one component
void PseudoMTG::computeLabelsHelper(const NodeRefHedge& ant1, const NodeRefHedge& ant2,
				    const bool theRole, const PAV& thePAV, 
				    NodeRefHedge& conseq, 
				    const arity_t d, const role_t heirAntecedent) const {
  VocabularyIndex nt1, nt2;

  PositionTree::iterator mti = conseq.getNode()[d];
  if(mti == 0)
    return;

  if (debug(3)){
    cerr << "PsMTG::cLH ConstrainingNode = ";
    mti->printPositionTreeNode(cerr);
  }

  DSpan constrainingDSpan = BGET_CONST_SPAN(*mti)[0];

  DSV* newDSV = conseq.getDSV();
  if (debug(3)) {
    cerr << "; d = " << (int)d;
    if (debug(4)) {
      cerr << "; NewSpan= " << (*newDSV) << "; MargSpan= " << constrainingDSpan;
    }
    cerr << endl;
  }
  DSpan newSpan = (*newDSV)[d];
  if (newSpan == constrainingDSpan) {
    // if the spans match, then just copy the NT
    // it's [0] because the constraining tree is expected to be just 1D

    NTLink lntl = conseq.getNTLink();

    LexNT* newLNT( (static_cast<LexNT*>(BGET_POSITION_LABEL(*mti).getNT(0)->clone())));
    newLNT->lex() = lntl.getTLink()[(int)d];

    conseq.setLabel(d, newLNT);
    delete newLNT;

  } else 
    // if it's a subspan, then create compound NT label
    // note that the so-called compound might not be compound,
    // because NULLs are ignored
    if (newSpan.containedIn(constrainingDSpan)){
      nt1 = ant1.getNTLink(d)->nt();
      if (debug(3))
	cerr << "PseudoMTG: Vocab Index= " << (int)nt1 << endl;
      nt2 = ant2.getNTLink(d)->nt();
      if (debug(3))
	cerr << "PseudoMTG: Vocab Index= " << (int)nt1 << endl;

      VocabularyIndex newNT = LabelingStrategy::BottomUpConcat1DHeaded(d, nt1, nt2, theRole, 
								       thePAV, heirAntecedent);

      const Terminal term = (static_cast<const LexNT*>(conseq.getNTLink(d)))->lex();
      const LexNT lnt(newNT, term);
      conseq.setLabel(d, addressof(lnt));

    } else {
      // otherwise, there's a problem
      THROW("d: " << (int)d << endl << "; NewSpan= " << newSpan
	    << "; MargSpan= " << constrainingDSpan << endl);
    }

  if (debug(3)) {
    cerr << " ; New label = ";
    conseq.getNTLink(d)->put(cerr);
    cerr << endl;
  };
}

/** Adds this class' option description to the option description list of the hierarchy */
BSP<BPO::options_description> 
PseudoMTG::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  optionsDescription->add_options()
    ("w2wModel", BPO::value<string>(), "the W2W translation model")
    ("RightBin", BPO::value<bool>(), "True if the grammar is rightBin")
    ;
  return optionsDescription;
}
