#ifndef _PseudoMTG_H_
#define _PseudoMTG_H_

#include "File.hpp"
#include "MoreAlgorithms.hpp"
#include "Link.hpp"

#include "WMTG.hpp"
#include "TreeConstrainedMTG.hpp"

#include "PAVOrderEvent.hpp"
#include "Pair.hpp"
#include "HeirRoleVector.hpp"

using namespace std;

class NodeRefHedge;

/// A grammar that behaves like an MTG despite containing no explicit
/// production rules, typically used for hierarchical alignment under
/// lower-dimensional constraints.
class PseudoMTG : public WMTG {
  typedef WMTG _super;
  typedef Grammar _grammar;

  typedef Link<short> PosLinkType;
  typedef ScoreSet<PosLinkType > W2WTMType;

  BSP<BPO::options_description> 
  addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription);

 public:
  PseudoMTG(arity_t arity, const string& configFile,
	   BSP<BPO::options_description> priorOptions = 
	   BSP<BPO::options_description>(new BPO::options_description()));

  virtual ~PseudoMTG();

  //! add a constraint grammar
  void add(Grammar* grammar);
  /**
   * Updates the content of the grammar.
   * @param tuple The sentence tuple.
   * @return If there is an empty parse read, then return false.
   */
  virtual bool prepareFor(SentenceTuple& tuple);

  //! determine whether a certain inference should be allowed
  virtual bool allowInf(BSP<Inference> spInf);

  //! returns true if the items should be filtered out due to 
  // binarization scheme  (WORK IN PROGRSS) 
  bool filter(const Item* item1, const Item* item2, const Pair<bool,PAV>& arrg);
  
  //! Returns all Z's given X and Y such that Z->XY (or Z->YX).
  //! \param heir : the heir constituent.
  //! \param depend : the dependent constituent.
  //! \param parents : the parent constituents that will be returned.
  virtual void getPossibleCons(const Item* heir,
		       const Item* depend,
		       const Pair<bool,PAV>& pav,
		       vector<Item*>& parents);
  virtual void getPossibleConsHeaded(const Item* heir,
		       const Item* depend,
		       const Pair<bool,PAV>& pav,
		       vector<pair<HeirRoleVector,Item*> >& parents);
  virtual void getPossibleConsForTLink(const TLink& tl, const DSV& dspan, vector<Item*>& vItem) const;

  istream& get(istream&) {
    THROW("PseudoMTG::get not implemented");
  }

  ostream& put(ostream& o) const;

  //! Clear the content.
  void clear();

  //! Returns the language dimensionality.
  arity_t arity() const;

  void getArrangementFor(BSP<set<Pair<bool,PAV> > >& arrgs, const Item* trigger) const;

protected: // methods

  enum HeirLayout {NODELTAS, CROSSWAYS, WHOLEDIM, ONEDELTA};
  HeirLayout getHeirLayout(const vector<short>& , const vector<short>& , int&) const;

  //! getter for w2w models
  vector<W2WTMType*>* getW2WModels() const ;

  /**
     Compute the nonterminal label for a consequent item, given its
     d-span, its PAV, its lexical heads, and its two antecedents. In
     other words, all the info is expected to be pre-computed except
     the nonterminal labels.
  */
  void computeLabels(const NodeRefHedge& ant1, const NodeRefHedge& ant2,
		     const bool theRole, const PAV& thePAV, const HeirRoleVector& hVec, 
		     NodeRefHedge& conseq) const;

  void computeLabelsHelper(const NodeRefHedge& ant1, const NodeRefHedge& ant2,
			   const bool theRole, const PAV& thePAV, 
			   NodeRefHedge& conseq, arity_t d, role_t heirAntecedent) const;

  /// Returns the cost (such as negative log prob) of an inference
  /// by default, use the corresponding production rule
  virtual SCORE getInfScore(Inference*) const;

  SCORE getW2WCost(const HeirRoleVector& hv, 
		   const NodeRefHedge* ant1, const NodeRefHedge* ant2 ) const;

 protected: // data
  //! 1D submodels
  vector<PartialWMTG*> m_monoGrammars;

  //! input stream of W2W model associated with this grammar
  istream* m_pW2WModelStream;

  //! Reference to a W2WTM
  W2WTMType* m_w2wTM;
  
  //! if true, force right binarization
  bool m_rightBin;

};
#endif
