#include <map>
#include <set>
#include "RawHeirRoleVector.hpp"
#include "GPDefs.hpp"

using namespace std;

static const int SERIALIZE_DIM = 100000;
static const short HASH_VALUE = 17;
static const char* DIM_SEP = "|";
static const short OTHER = 0;

void RawHeirRoleVector::clear() {
  for(arity_t d = 0; d < arity(); d++)
    (*this)[d] = OTHER;
}

bool RawHeirRoleVector::empty() const {
  for(arity_t d = 0; d < arity(); d++)
    if((*this)[d] != OTHER) return false;

  return true;
}

size_t RawHeirRoleVector::hash() const {
  size_t ret = 0;
  for(arity_t d = 0; d < arity(); d++){
    ret *= (short)HASH_VALUE;
    ret += (*this)[d];
  }
  return ret;
}

RawHeirRoleVector RawHeirRoleVector::invert() const{
  RawHeirRoleVector ret (*this);
  for(arity_t i = 0; i < arity(); i++){
    switch((*this)[i]){
    case 1:
      ret[i] = 2;
      break;
    case 2:
      ret[i] = 1;
      break;
    case 0:
      break;
    }
  }

  return ret;
}

RawHeirRoleVector RawHeirRoleVector::get1DItem(arity_t d)const{
  RawHeirRoleVector ret(1);
  ret[0] = (*this)[d];
  return ret;
}

// Binarized the PAV by taking r as one role, and all the others as one role.
void RawHeirRoleVector::binarize(const short r) {
  for(arity_t d = 0; d < arity(); d++){
    for(unsigned int i = 0; i < (*this).size(); i++){
      // change the roles.
      if((*this)[i] == r){
	(*this)[i] = (r == 1) ? 1 : 2;
      } else if((*this)[i] != 0){
	(*this)[i] = (r == 1) ? 2 : 1;
      }
    }
  }
}

ostream& operator<<(ostream& out, const RawHeirRoleVector& hv) {
  for(arity_t d = 0; d < hv.arity(); d++) {
    if(d != 0) out<<" ";
    out << hv[d];
  }
  return out;
}

ostream& RawHeirRoleVector::put(ostream& out) const{
  for(arity_t d = 0; d < arity(); d++) {
    if(d != 0) out<<DIM_SEP;
    out << (*this)[d];
  }
  return out;
}

int RawHeirRoleVector::cmp(const RawHeirRoleVector& other) const {
  assert(arity() == other.arity());
  for(arity_t i = 0; i < arity(); i++){
    if(other[i] < (*this)[i])
      return 1;
    else if(other[i] > (*this)[i])
      return -1;
  }
  return 0;
}

void RawHeirRoleVector::swapElement() {
  for(arity_t d = 0; d < arity(); d++)
    if((*this)[d] == 1)
      (*this)[d] = 2;
    else if((*this)[d] == 2)
      (*this)[d] = 1 ;
}

//! Returns the set of roles in this PAV.
set<short> RawHeirRoleVector::roleSet() const {

  set<short> roleset;

  for(arity_t d = 0; d < arity(); ++d){
    if((*this)[d]){
      roleset.insert((*this)[d]);
    }
  }
  return roleset;
}

/**
 * Return the maximum role. This RawHeirRoleVector do not have to be normalized.
 */
short RawHeirRoleVector::maxRole() const {
  short role = 0;
  for(arity_t d = 0; d < arity(); ++d){
    if (role < (*this)[d])
      role = (*this)[d];
  }
  assert(role);
  return role;
}

ostream& RawHeirRoleVector::serialize(ostream& out) const {
  for(arity_t d = 0; d < arity(); d++) {
    if(d != 0) out<<" "<<SERIALIZE_DIM<<" ";
    out<<(*this)[d];
  }
  return out;
}

//! (make sure first position is 1)
RawHeirRoleVector RawHeirRoleVector::normal() const{
  RawHeirRoleVector ret = (*this);
  ret.normalize();
  return ret;
}

//! returns true if this heir vector is head split.
bool RawHeirRoleVector::headSplit() const{
  RawHeirRoleVector::const_iterator i;
  set<short> inds;
  for(i = (*this).begin(); i != (*this).end(); ++i){
    inds.insert(*i);
  }

  if(inds.size() > 1){
    return true;
  } else {
    return false;
  }

};

//!Normalized this RawHeirRoleVector
//! (make sure first position is 1)
//! N.B. this should ONLY be used when being made into an event.
//! Once in a distribution, we only care about normalized HVs, but
//! when an HV is in a production, we can't normalize it unless we
//! switch around all the other parts of the production to match.
void RawHeirRoleVector::normalize(){
  if(!isNormal())
    *this = this->invert();
}

//! detects whether the RawHeirRoleVector is normalized.
bool RawHeirRoleVector::isNormal()const{
  if((*this)[0] == 1)
    return true;
  if((*this)[0] == 0)
    return true;
  if((*this)[0] == 2 && (*this)[1] == 0)
    return true;
  else
    return false;
}

short RawHeirRoleVector::rank() const {
  if(arity() == 0) return 0;
  set<short> roleset = roleSet();
  return (short)roleset.size();
}

void RawHeirRoleVector::deleteRoleAndNormalize(const short delRole) {
  throw NotImplemented();
}

//! Returns true is there is any information about heirs.
bool RawHeirRoleVector::containsHeir() const{
  for(unsigned int i = 0; i < (*this).size(); i++) {
      if((*this)[i] != OTHER){
	return true;
      }
    }
  return false;
}

//! Return true if the role is in the heirVector
bool RawHeirRoleVector::contains(short val)const{
  for(unsigned int i = 0; i < (*this).size(); i++) {
      if((*this)[i] == val){
	return true;
      }
    }
  return false;
}

//We are passed in a Vector which is the RHS,and
// we export an Heir, given the this RawHeirRoleVector.
NTLink* RawHeirRoleVector::getHeirEvent(const Vector<ProdRHSElem>& d) const{

  NTLink* nl = new NTLink(this->size());

  for(unsigned int i = 0; i < this->size(); i++) {
    if((*this)[i] != 0){
      if(const NTLink* ntl = boost::get<const NTLink>((&(d[((*this)[i]-1)]))))
	(*nl).setNT(i,new Nonterminal((*ntl).getNT(i)->nt()));
      else if (const TLink* tl = boost::get<const TLink>((&(d[((*this)[i]-1)])))) {
	tl = tl; // duck compiler warning
	(*nl).setNT(i, new Nonterminal(UNDEF_NT));
      }
    }
    else
      (*nl).setNT(i,new Nonterminal(UNDEF_NT));
  }

  return nl;
}

//We are passed in a NTLink1 and 2.
// we export an HeirEvent, given the this is the RawHeirRoleVector.
NTLink* RawHeirRoleVector::getHeirEvent(const NTLink& link1, const NTLink& link2) const{
  NTLink* nl = new NTLink( this->size());

  for(unsigned int i = 0; i < this->size(); i++){
    if((*this)[i] == 1)
      (*nl).setNT(i,new Nonterminal(link1.getNT(i)->nt()));
    else
      (*nl).setNT(i,new Nonterminal(link2.getNT(i)->nt()));
  }
  return nl;
}

//! We are given and NTLink and we create a dependency from it.
NTLink* RawHeirRoleVector::getDependency(const NTLink& ntl, short role) const{

  bool found = false;
  NTLink* lntl = new NTLink( this->size());
  for(unsigned int i = 0; i < this->size(); i++)
    if((*this)[i] != role+1){
      found = true;
      (*lntl).setNT(i,ntl.getNT(i)->clone());
    }
    else{
      (*lntl).setNT(i,ntl.getNT(i)->clone());
      Nonterminal* nt = (*lntl).getNT(i)->clone();
      nt->setHeirMark();
      (*lntl).setNT(i,nt);
    }

  if(!found){
    delete lntl;
    return NULL;
  } else
    return lntl;
}
