#ifndef _RAWHEIRVECTOR_H_
#define _RAWHEIRVECTOR_H_

#include "Vector.hpp"
#include "GPDefs.hpp"
#include "NTLink.hpp"
#include "ProdRHSElem.hpp"
#include <boost/serialization/serialization.hpp>
#include "serialize-variant.hpp"

//! "RawHeirRoleVector" class. Tells which child is the heir in each
//! dimension. SHOULD PROBABLY BE REPLACED WITH BITS.
class RawHeirRoleVector: public Vector<short> {
  typedef Vector<short> _super;
public:

  //! default constructor
  RawHeirRoleVector() : _super(){}

  //! copy constructor
  RawHeirRoleVector(const RawHeirRoleVector& other) : _super(other){}

  /// constructor from super
  RawHeirRoleVector(const Vector<short>& other) : _super(other){}

  //! Constructor.
  RawHeirRoleVector(const arity_t& dim) : Vector<short >(dim) {
    for(size_t i = 0; i < dim; i++)
      (*this)[i] = 0;
  }

  //! Constructor.
  RawHeirRoleVector(const arity_t& dim, short s) : Vector<short >(dim, s) {}

  virtual ~RawHeirRoleVector() {};

  //! Return a normalized version of the RawHeirRoleVector
  //! (make sure first position is 1)
  RawHeirRoleVector normal() const;

  //!Normalized this RawHeirRoleVector
  //! (make sure first position is 1)
  //! N.B. this should ONLY be used when being made into an event.
  //! Once in a distribution, we only care about normalized HVs, but
  //! when an HV is in a production, we can't normalize it unless we
  //! switch around all the other parts of the production to match.
  void normalize();

  RawHeirRoleVector get1DItem(arity_t d)const;

  //! detects whether the RawHeirRoleVector is normalized.
  bool isNormal()const;

  //! returns true if this heir vector is head split.
  bool headSplit() const;

  //! dimension of heirVector.
  arity_t arity() const {
    return (arity_t)size();}

  //! Clear the content, but retain the dimension.
  void clear();

  //! invert the RawHeirRoleVector (1->2 2->1 0->0)
  RawHeirRoleVector invert() const;

  //! Judges whether empty.
  bool empty() const;

  //! resize the RawHeirRoleVector (number of dimensions)
  void setDimity(arity_t d){
    resize(d);
  };

  //! For hashing.
  size_t hash() const;

  //! Number of dimensions of heirVector.
  arity_t arity(){
    return (*this).size();
  }

  //! Assignment.
  RawHeirRoleVector& operator = (const RawHeirRoleVector& other)
  { Vector<short>::operator=(other); return *this; }

  //! Equality test
  bool operator == (const RawHeirRoleVector& other) const {
    return (Vector<short >::operator==(other)) ;
  }

  //! Equality test with short. (if all are equal to short)
  bool operator == ( short other) const {
    for(unsigned int i = 0; i < (*this).size(); i++){
      if((*this)[i] != other)
	return false;
    }
    return true;
  }

  //! binarize the heir vector.
  void binarize(const short r);

  //! Less than test.
  bool operator < (const RawHeirRoleVector& other) const
  { return (cmp(other) == -1); }

  //! decrement
  RawHeirRoleVector operator-(const short val) const{
    //cerr << "this: \'" << (*this) << "\'" << endl;
    // cerr << val << endl;
    // cerr << (*this).size();
    RawHeirRoleVector ret((*this).size());
    //cerr << (ret.size()) << endl;
    for(unsigned int i = 0; i < (*this).size(); i++){
      ret[i] = (*this)[i] - val;
    }
    return ret;
  }

  //! Assignment
  RawHeirRoleVector& operator=(short other){
    (*this).setAll(other);
    return *this;
  }

  //!Sets all elements to val
  void setAll(short val){
    for(unsigned int i = 0; i < (*this).size(); i++)
      (*this)[i] = val;
  }

  int cmp(const RawHeirRoleVector& other) const;

  //!output.
  ostream& serialize(ostream&out) const;

  //! Exchanges the value of elements: 1->2, 2->1, and 0 remained.
  //! Currently only works for binary case.
  void swapElement();

  //! Judges whether the indices are not increasing order in the
  //! first dimension. <1,2> -- false; <2,1> -- true.
  bool improperOrder() const;

  //! Returns the number of roles on dimension d.
  short rank() const;

  //! Remove one role, and normalize the PAV.
  void deleteRoleAndNormalize(const short i);

  //! Remove one role, and DOT NOT normalize the PAV.
  void deleteRoleAndNormalizeOnly(const short i);

  short maxRole() const;

  set<short> roleSet() const;

  //! Inputs HV.
  //friend istream& operator>>(istream& in, RawHeirRoleVector& hv);

  //! Outputs HV.
  friend ostream& operator<<(ostream& out, const RawHeirRoleVector& hv);

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    ar & boost::serialization::base_object<Vector<short> >(*this);
  };

  ostream& put(ostream& out) const;

  //! Returns true is there is any information about heirs.
  bool containsHeir() const;

  //! returns true if the RawHeirRoleVector contains "role".
  bool contains(short role)const;

  //We are passed in a Vector which is the RHS,and
  // we export an Heir, given the this RawHeirRoleVector.
  NTLink* getHeirEvent (const Vector<ProdRHSElem>& d) const;

  NTLink* getHeirEvent (const NTLink& l1, const NTLink& l2) const;

  NTLink* getDependency(const NTLink&, short val) const;
};

namespace HashNS{
  template<>
  struct hash<RawHeirRoleVector > {
    size_t operator()(const RawHeirRoleVector& hv) const { return hv.hash();}
  };
}

#endif
