#include "NTLink.hpp"
#include "LexNT.hpp"
#include "GlobalVocab.hpp"
#include "Exceptions.hpp"
#include "Set.hpp"

RawNTLink::
RawNTLink(){
  m_array = new Nonterminal*[GlobalVocab::s_arity];
  for(arity_t i = 0; i < GlobalVocab::s_arity; i++)
    m_array[i] = NULL;
  m_size = GlobalVocab::s_arity;
}

// Constructor which sets dimension. (Does not init pointers)
RawNTLink::
RawNTLink(const arity_t d) {
 m_array = new Nonterminal*[d];
 m_size = d;
 for(arity_t i = 0; i < m_size; i++)
   m_array[i] = NULL;
}

// Constructor which does init pointers to Nonterminals
RawNTLink::
RawNTLink(const arity_t& _d, const Nonterminal& nt){
 m_array = new Nonterminal*[_d];
  for(arity_t d = 0; d < _d; d++){
    (m_array[d]) = nt.clone();
  }
  m_size = _d;
}

// Constructor which does init pointers to LexNTs
RawNTLink::
RawNTLink(const arity_t& _d, const LexNT& nt) {
 m_array = new Nonterminal*[_d];
  for(arity_t d = 0; d < _d; d++){
    (m_array[d]) = nt.clone();
  }
  m_size = _d;
}

// Constructor from two nonterms, one in each dimension.
RawNTLink::
RawNTLink(Nonterminal* nt1, Nonterminal* nt2){
 m_array = new Nonterminal*[2];
  m_array[0] = nt1;
  m_array[1] = nt2;
  m_size = 2;
}


// Constructor
RawNTLink::RawNTLink(const vector<Nonterminal*> &vec) {
  m_array = new Nonterminal*[vec.size()];
  for(arity_t d = 0; d < vec.size(); d++) {
    m_array[d] = vec[d]->clone();
  }
  m_size = vec.size();
}

//Constructor
RawNTLink::RawNTLink(const RawNTLink& ntLink, const TLink& tlink){
 m_array = new Nonterminal*[ntLink.size()];
  for(arity_t d = 0; d < ntLink.arity(); d++){
    m_array[d] = new LexNT(ntLink.m_array[d]->getSynCat(), tlink[d]);
  }
  m_size = ntLink.size();
}

//Clear
void RawNTLink::clear(){
   for(arity_t d = 0; d < this->size(); d++){
     if(m_array[d] != NULL)
       delete (m_array[d]);
   }
}

//Returns vector of terminals. (only works with LexNT elements.)
TLink RawNTLink::getTLink() const {
  if(!(dynamic_cast<LexNT*>(m_array[0])))
    THROW("getTLink() with nonLexNT not implemented");

  TLink tLink(size());
  for(arity_t i = 0; i < size(); i++)
    tLink[i] = (static_cast<LexNT*>(m_array[i]))->lex();
  return tLink;
}

//sets vector of terminals. (only works with LexNT elements.)
void RawNTLink::setTLink(const TLink& tl) {
  if(!(dynamic_cast<LexNT*>(m_array[0])))
    THROW("getTLink() with nonLexNT not implemented");

  for(arity_t i = 0; i < size(); i++)
    (static_cast<LexNT*>(m_array[i]))->lex() = tl[i];
}
//sets vector of terminals. (only works with LexNT elements.)
void RawNTLink::setTerminal(const Terminal& t, arity_t d) {
  if(!(dynamic_cast<LexNT*>(m_array[d])))
    THROW("getTLink() with nonLexNT not implemented");

  (static_cast<LexNT*>(m_array[d]))->lex() = t;
}

void RawNTLink::delex(){
  for(arity_t i = 0; i < size(); i++){
    if(dynamic_cast<LexNT*>(m_array[i])){
      SynCat sc = m_array[i]->getSynCat();
      delete m_array[i];
      m_array[i] = new Nonterminal(sc);
    }
  }
 }

// reverses all elements
void RawNTLink::reverse(){
  unsigned int j = this->size() - 1;
  for(unsigned int i = 0; i < this->size(); i++){
    if(i != j){
      Nonterminal* tmp = m_array[i];
      m_array[i] = m_array[j];
      m_array[j] = tmp;
    }
  }
}

//! D-dimensional cross-product
//! takes a vector of sets of Xs and returns a set of vectors of Xs.
//! \todo recode this recursively over dim pairs and make it generic
void RawNTLink::
hyperCrossProduct(vector<Set<Nonterminal*, less<Nonterminal*> > >& sets, set<RawNTLink>& result) {
  if(sets.size() == 0) return;

  size_t d;
  set<Nonterminal*>::iterator it1;
  vector<vector<Nonterminal*> > prod;
  vector<Nonterminal*> toDelete;

  // create 1d vector of singleton vectors of the 1st set
  for(it1 = sets[0].begin(); it1 != sets[0].end(); ++it1){
    Nonterminal* ntp = (*it1)->clone();
    toDelete.push_back(ntp);

    prod.push_back(vector<Nonterminal*>(1, ntp));
  }

  for(d = 1; d < sets.size(); d++){
    vector<pair<vector<Nonterminal*>, Nonterminal*> > prod1;
    prod1 = cross_product(prod.begin(), prod.end(),
			  sets[d].begin(), sets[d].end());

    prod.clear();
    // add each nt to each vector of nts
    vector<pair<vector<Nonterminal*>, Nonterminal*> > :: iterator it2;
    for(it2 = prod1.begin(); it2 != prod1.end(); ++it2){
      it2->first.push_back(it2->second);
      // push the updated vector of nts back onto the vector of vectors stack
      prod.push_back(it2->first);
    }
  }
  // iterate thru the results and clone/rearrange them for output
  vector<vector<Nonterminal*> > :: iterator it3;
  for(it3 = prod.begin(); it3 != prod.end(); ++it3){
    RawNTLink lexnts(sets.size());
    for(d = 0; d < sets.size(); ++d){
      lexnts.m_array[d] = (*it3)[d]->clone();
    }
    result.insert(lexnts);
  }

  vector<Nonterminal*>:: iterator it2;
  for(it2 = toDelete.begin(); it2 != toDelete.end(); ++it2){
    delete *it2;
  }
}

//! D-dimensional cross-product
//! takes a vector of sets of Xs and returns a set of vectors of Xs.
//! \todo recode this recursively over dim pairs and make it generic
void RawNTLink::
hyperCrossProduct2(vector<set<Nonterminal*, less<Nonterminal*> > >& sets, set<RawNTLink>& result) {
  if(sets.size() == 0) return;

  size_t d;
  set<Nonterminal*>::iterator it1;
  vector<vector<Nonterminal*> > prod;

  // create 1d vector of singleton vectors of the 1st set
  for(it1 = sets[0].begin(); it1 != sets[0].end(); ++it1){
    prod.push_back(vector<Nonterminal*>(1, *it1));
  }

  for(d = 1; d < sets.size(); d++){
    vector< pair< vector<Nonterminal*>, Nonterminal*> > prod1;
    vector< pair< vector<Nonterminal*>, Nonterminal*> > :: iterator it2;
    prod1 = cross_product(prod.begin(), prod.end(),
			  sets[d].begin(), sets[d].end());

    prod.clear();
    // add each nt to each vector of nts
    for(it2 = prod1.begin(); it2 != prod1.end(); ++it2){
      it2->first.push_back(it2->second);
      // push the updated vector of nts back onto the vector of vectors stack
      prod.push_back(it2->first);
    }
  }
  // iterate thru the results and clone/rearrange them for output
  vector<vector<Nonterminal*> > :: iterator it3;
  for(it3 = prod.begin(); it3 != prod.end(); ++it3){
    RawNTLink lexnts(sets.size());
    for(d = 0; d < sets.size(); ++d){
      lexnts.m_array[d] = ((*it3)[d]->clone());
    }
    result.insert(lexnts);
  }
}

//! Returns a RawNTLink with each element to be NULLNT.
RawNTLink RawNTLink::nullNTLink(const arity_t& dim) {
  RawNTLink ntlink(dim);
  for(arity_t i = 0; i < dim; i++) 
    *(ntlink.m_array[i]) = NULLNT;
  return ntlink;
}

//! Returns a Unk RawNTLink.
RawNTLink RawNTLink::unkNTLink(const arity_t& d) {
  RawNTLink ntlink(d);
  for(arity_t i = 0; i < d; i++) 
    ntlink.m_array[i] = new Nonterminal(UNKNT);
  return ntlink;
}

//! Returns a unknown preterminal RawNTLink.
RawNTLink RawNTLink::prtmUnkNTLink(const arity_t& d) {
  RawNTLink ntlink(d, Nonterminal(UNK_PRTM));
  return ntlink;
}

//! Returns a Dummy RawNTLink.
RawNTLink RawNTLink:: dummyNTLink(const arity_t& dim) {
  RawNTLink ntlink(dim);
  for(arity_t i = 0; i < dim; i++){
    ntlink.m_array[i] = new Nonterminal(DummyNT);
  }
  return ntlink;
}

//! Write.
ostream& RawNTLink::put(ostream& out) const{
  //if there are no vocabs being output
  if(GlobalVocab::s_noVocabsOut){
    for(arity_t d = 0; d < arity(); d++) {
      if(d !=0) out<<" | ";
      if(m_array[d] != NULL)
	m_array[d]->put(out, GlobalVocab::s_synCatSet, GlobalVocab::s_terminalSet);
      else
	out << "NULL";
    }
  }
  else{ // integerized output
    for(arity_t d = 0; d < arity(); d++) {
      if(d !=0) out<<" | ";
      if(m_array[d] != NULL)
	m_array[d]->put(out);
      else
        out << 0;
    }
  }
  return out;
}

// For ordering. Arbitrary.
int RawNTLink::cmp(const RawNTLink& other) const {
  int ret = 0;
  assert(this->size() == other.size());
  unsigned int si = this->size();

  for(unsigned int i = 0 ; i != si; i++){
    const Nonterminal* pNT1 = m_array[i];
    const Nonterminal* pNT2 = other.m_array[i];
    if (pNT1 == NULL && pNT2 == NULL)
      continue;
    if (pNT1 != NULL && pNT2 == NULL) {
      ret = -1;
      break;
    };
    if (pNT1 == NULL && pNT2 != NULL) {
      ret = 1;
      break;
    };
    ret = pNT1->cmp(pNT2);
    if(ret != 0)
      break;
  }

  return ret;
}

//output
ostream& RawNTLink::put(ostream& out, vector<StringVocabulary*> ntvVec)const{
  for(arity_t d = 0; d < arity(); ++d){
    if(d){ out << " | "; }

    if(ntvVec.size() > d && ntvVec[d]){
      bool exists;
      const string* nt = ntvVec[d]->getWord(m_array[d]->getSynCat(), exists);
      assert(exists);
      out << *nt;
    } else { out<<(*m_array[d]); }
  }
  return out;
}

//output
ostream& RawNTLink::put(ostream& out, StringVocabulary* ntSet, StringVocabulary* terminalSet) const{
  for(arity_t d = 0; d < size(); ++d){
    if(d != 0)
      out << " | ";
    m_array[d]->put(out,ntSet,terminalSet);
  }
  return out;
}

//! Overload only not to output the space at the end.
ostream& operator << (ostream& out, const RawNTLink& obj){
  obj.put(out);
  return out;
}

//! Read.
istream& RawNTLink::get(istream& in) {
  char ch;
  string str;
  in >> ws;
  //arity_t dim = this->size();
  // for(unsigned int i = 0; i < this->size(); i ++){
  // if(m_array[i] != NULL)
  // delete m_array[i];
  // }

  unsigned int d = 0;
  if(size() == 0){
    while(true){
      GlobalVocab::s_ioDIM = d;
      m_array[d] = (Nonterminal::readNonterminal(in));
      in >> ws;
      if (in.peek() != '|'){
	break;
      }
      in >> ch;
      d++;
    }
  }
  else {
    for(unsigned int d = 0; d < this->size(); d++){
       GlobalVocab::s_ioDIM = d;
       char ch;
      if(m_array[d] == NULL){
	m_array[d] = (Nonterminal::readNonterminal(in));
      }
      else{
	m_array[d]->get(in);
      }
      if(d < this->size()-1)
	in >> ch;
    }
  }
  return in;
}

//! If arity is not specified, reads line by line.
//! Otherwise, reads in terms of arity number.
istream& operator >> (istream& in, RawNTLink& obj){
  obj.get(in);
  return in;
}

void RawNTLink::setNull(){
  for(arity_t i = 0; i < m_size; i++)
    m_array[i]->setNull();
}

bool RawNTLink::isDummy() const{
  for(arity_t i = 0; i < m_size; i++)
    if(!m_array[i]->isDummy())
      return false;
  return true;
}

//returns reference to element
const Nonterminal* RawNTLink::getNT(unsigned int i)const{
  return m_array[i];
}

//returns reference to element
void RawNTLink::setNT(unsigned int i, Nonterminal* nt){
  if(m_array[i])
    delete m_array[i];
  m_array[i] = nt;
}
