
#ifndef _RawNTLink_H_
#define _RawNTLink_H_

#include "Set.hpp"
#include "GPDefs.hpp"
#include "MoreAlgorithms.hpp"
#include "StringVocabulary.hpp"
#include "Nonterminal.hpp"
#include "LexNT.hpp"
#include "TLink.hpp"
//#include "GPcmp.hpp"
#include <boost/serialization/serialization.hpp>
#include "boost/multi_array.hpp"
using namespace std;

//! This class holds a vector of Nonterminal pointers. Each of these
//! "nonterminals" has at least a SynCat. One should use caution when
//! using this class, since it hold pointers.

class RawNTLink {
public:

  //! Needed for hashing. Set to max number of nonterminals in the second
  //! language.
  static const size_t sm_bigValForHash = 1109;

  //! Empty constructor.
  RawNTLink();

  //! Constructs an _d dimensional RawNTLink, and each
  //! dimension is initialized with nonterminal nt;
  RawNTLink(const arity_t& _d, const Nonterminal& nt);

  //! Constructs an _d dimensional RawNTLink, and each
  //! dimension is initialized with LexNT nt;
  RawNTLink(const arity_t& _d, const LexNT& nt);

  //! Constructs an 2 dimensional RawNTLink, and each
  //! dimension has one of the inputs;
  RawNTLink(Nonterminal* nt1, Nonterminal* nt2);

  //! Constructs an NTlink with "LexNT's" as its members.
  RawNTLink(const RawNTLink&,const TLink&);

  //! Constructor (from a Vector).
  //RawNTLink(const Nonterminal* ve[]);
  RawNTLink(const vector<Nonterminal*> &vec);

  // copy constructor
  inline RawNTLink(const RawNTLink& ntl){
    m_array = new Nonterminal*[ntl.size()];
    for(unsigned int i = 0; i < ntl.size(); i++){
      if( ntl.getNT(i) != NULL) {
	Nonterminal* nt = ntl.getNT(i)->clone();
	m_array[i] = nt;
      } else
	m_array[i] = NULL;
    }
    m_size = ntl.size();
  }

  //! Constructor (from a nonterminal, only for D == 1).
  RawNTLink(const arity_t d);

  //! delete elements and clear
  void clear();

  //! comparator
  int cmp(const RawNTLink& other) const;

  //! Destructor. Destroys the Nonterminal pointers.
  ~RawNTLink() {
    for(arity_t d = 0; d < this->size(); d++){
      if (m_array[d])
        delete (m_array[d]);
    }
    delete[] m_array;
  }

  //! D-dimensional cross-product of the input NTs.
  static void hyperCrossProduct(vector<Set<Nonterminal*,less<Nonterminal*> > >& sets, set<RawNTLink>& result);

  //! D-dimensional cross-product of the input NTs.
  static void hyperCrossProduct2(vector<set<Nonterminal*, less<Nonterminal*> > >& sets, set<RawNTLink>& result);

  //! return a delexicalized version of this RawNTLink
  void delex();

  //< operator
  inline bool operator<(const RawNTLink& other)const{
    for(unsigned int i = 0; i < other.size(); i++){
      if(m_array[i]->operator<(other.getNT(i))){
    return true;
      };
    }
    return false;
  }
  //= operator (copies values using clone(), not pointers)
  inline RawNTLink& operator=(const RawNTLink& ntl){

    // this->resize(ntl.size());
    //assert(ntl.size() == this->size());

    // if the sizes differ then we should allocate a new array
    if(ntl.size() != this->size()) {
      for(size_t i=0; i<this->size(); i++) {
        if(m_array[i])
          delete m_array[i];
      }
      delete[] m_array;
      m_array = new Nonterminal*[ntl.size()];
      m_size = ntl.size();
      for(size_t i=0; i<size(); i++)
        m_array[i] = NULL;
    }

    for(unsigned int i = 0; i < size(); i++){
      if(m_array[i])
        delete m_array[i];
      if(ntl.getNT(i))
        m_array[i] = ntl.getNT(i)->clone();
      else
        m_array[i] = NULL;
    }
    return *this;
  }

  // > operator
  inline bool operator>(const RawNTLink& other)const{
    return (!(this->operator<(other)) &&
         !(this->operator==(other)));
  }

  // size (or dimensionality) of link.
  inline unsigned int size()const{
    return m_size;
  }

  // makes a copy of this RawNTLink
  inline RawNTLink* clone() const{
    return new RawNTLink(*this);
  };

  // reverse the elements in link
  void reverse();

  //! return a vector of terminals associated with his RawNTLink.
  //! Can only be used when elements are LexNTs
  TLink getTLink() const;

  //! sets the terminals in this RawNTLink.
  //! Can only be used when elements are LexNTs
  void setTLink(const TLink& tl) ;

  void setTerminal(const Terminal& t, arity_t d);

  //! Returns the dimensionality.
  inline arity_t arity() const { return this->size();}

  // == operator (based on values)
  inline bool operator==(const RawNTLink& other) const {
     if(this->size() != other.size())
      return false;
    
    for(unsigned int j = 0; j < other.size(); j++){
      Nonterminal* pNT1 = m_array[j];
      Nonterminal* pNT2 = other.m_array[j];
      if (pNT1 == NULL && pNT2 == NULL)
	continue;
      else if (pNT1 != NULL && pNT2 == NULL) {
	return false;
      }
      else if (pNT1 == NULL && pNT2 != NULL) {
	return false;
      }
      else if(typeid(*pNT1) != typeid(*pNT2))
	return false;
      if(pNT1->cmp(pNT2) != 0){
	return false;	
      }
    }
    return true;
  }

  inline size_t hash() const {
    size_t ret = 0;
    arity_t a = arity();
    for(arity_t i = 0; i < a; i++){ 
      Nonterminal* nt = m_array[i];
      if(nt){
	ret *= sm_bigValForHash;
	ret += nt->hash(); 
      }
    }
    return ret+(arity()*sm_bigValForHash);
  }

  //! sets all links to null
  void setNull();

  //! Returns a RawNTLink with each element to be NULLNT.
  static RawNTLink nullNTLink(const arity_t& d );

  //! Returns a Dummy RawNTLink.
  static RawNTLink dummyNTLink(const arity_t& d) ;

  //! Returns a Unk RawNTLink.
  static RawNTLink unkNTLink(const arity_t& d) ;

  //! Returns a unknown preterminal RawNTLink.
  static RawNTLink prtmUnkNTLink(const arity_t& d) ;

  //! Write.
  ostream& put(ostream& out) const;

  //! write with nonterminal info.
  ostream& put(ostream& out, vector<StringVocabulary*> ntvVec)const;

  //! write with nonterminal and terminal info. (only used with
  //! lexical case)
  ostream& put(ostream& out, StringVocabulary* ntSet, StringVocabulary* terminalSet) const;

  //! Overload only not to output the space at the end.
  friend ostream& operator << (ostream& out, const RawNTLink& obj);

  //! Read.
  istream& get(istream& in) ;

  //! If arity is not specified, reads line by line.
  //! Otherwise, reads in terms of arity number.
  friend istream& operator >> (istream& in, RawNTLink& obj);

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    // ar & boost::serialization::base_object<m_array>(*this);
  };

  bool isDummy()const;


  //returns Nonterminal* from dimension i.
  const Nonterminal* getNT(unsigned int i)const;

  //sets noterminal pointer in dimension i
  void setNT(unsigned int i, Nonterminal* nt);


protected:
  arity_t m_size;
  Nonterminal** m_array;

};

namespace HashNS {
  template<> struct hash< RawNTLink > {
    size_t operator()(const RawNTLink& e) const{
      return e.hash();
    }
  };

}

namespace HashNS {
  template<> struct hash< Nonterminal* > {
    size_t operator()(const Nonterminal* e) const{
      return e->hash();
    }
  };

}

typedef RawNTLink PositionRawNTLink;

#endif
