#include <map>
#include <set>
#include "RawPAV.hpp"
#include "DSV.hpp"
#include "Exceptions.hpp"
#include "MoreComparators.hpp"
#include "GPDefs.hpp"

using namespace std;

static const char GAPVALUE = 0;
static const int SERIALIZE_DIM = 100000;
static const short HASH_VALUE = 17;
static const char* DIM_SEP = "|";

RawPAV::RawPAV(const arity_t& dim) : Vector< PrecArray >(dim) { }

//! constructor from one PrecArray
RawPAV::RawPAV(const PrecArray& pa) : Vector< PrecArray >(1, pa) { }

//! constructor from a Vector<PrecArray>
RawPAV::RawPAV(const Vector<PrecArray>& pa) : Vector< PrecArray >(pa) { }

//! constructor from an arity, a PrecArray, and the position of the PrecArray
//! the other components will be empty
RawPAV::RawPAV(const arity_t& dim, const PrecArray& pa, const arity_t& psn)
  // initialize with empties
  : Vector< PrecArray >(dim, PrecArray()) {
  this->operator[](dim) = pa;
}

//! constructor from a fan-out vector
//! all elements will have the same role
RawPAV::RawPAV(const FanoutVect& fov)
  // initialize with empties
  : Vector< PrecArray >(fov.size(), PrecArray()) {
  for(unsigned int i = 0; i < fov.size(); ++i) {
    if (fov[i] > 0)
      operator[](i).push_back(1);
    // for every element beyond one, first insert a gap
    if (fov[i] > 1)
      for(int j = fov[i]; j > 1; --j) {
        operator[](i).push_back(0);
        operator[](i).push_back(1);
      };
  };
}

//! constructor from a single DSV
//! all elements will have the same role
RawPAV::RawPAV(const DSV& dsv)
  // initialize with empties
  : Vector< PrecArray >(dsv.size(), PrecArray()) {
  for(unsigned int i = 0; i < dsv.size(); ++i) {
    size_t currfanout = dsv[i].size();
    if (currfanout > 0)
      operator[](i).push_back(1);
    // for every element beyond one, first insert a gap
    if (currfanout > 1)
      for(int j = currfanout; j > 1; --j) {
        operator[](i).push_back(0);
        operator[](i).push_back(1);
      };
  };
}

RawPAV::RawPAV(const DSV& theDSV1, const DSV& theDSV2)
  : Vector< PrecArray>(max(theDSV1.arity(), theDSV2.arity())) {
  unsigned int i,j;
  arity_t d;
  //cerr << "RawPAV constructor bin:" << theDSV1 << "; " << theDSV2 << endl;

  for(d = 0; d < arity(); d++){
    size_t size1 = 0;
    size_t size2 = 0;
    if (theDSV1.size() > d)
      size1 =  theDSV1[d].size();
    if (theDSV2.size() > d)
      size2 =  theDSV2[d].size();
    // i and j index the contiguous elements of the d'th d-spans.
    i = 0; j = 0;
    while(i < size1 && j < size2){
      // if the end of the first span precedes the start of the second
      if(theDSV1[d][i].second <= theDSV2[d][j].first) {
	(*this)[d].push_back(1);
	if(theDSV1[d][i].second != theDSV2[d][j].first)
	  (*this)[d].push_back(0);//gap
	i++;
      }
      // if the start of the first span precedes the start of the first
      else if(theDSV2[d][j].second <= theDSV1[d][i].first) {
	(*this)[d].push_back(2);
	if(theDSV2[d][j].second != theDSV1[d][i].first)
	  (*this)[d].push_back(0); //gap
	j++;
      }
      else {
	cerr << "spans " << i << " and " << j << " on dim " << (int)d << " in DSVs " 
	     << theDSV1 << " and " << theDSV2 << " are overlapped." << endl;
	THROW("");
      }
    }

    // more elements in the first dspan
    if(i != size1){
      for(unsigned int ii = i; ii < size1; ii++){
	(*this)[d].push_back(1);
	if(ii != size1 - 1)
	  (*this)[d].push_back(0);
      }
    }

    // more elements in the 2nd dspan
    if(j != size2){
      for(unsigned int jj = j; jj < size2; jj++){
	(*this)[d].push_back(2);
	if(jj != size2 - 1)
	  (*this)[d].push_back(0);
      }
    }
  }
}

RawPAV::RawPAV(const set<DSV >& theDSVs) {
  // set the arity.
  //assert(theDSVs.size() != 0);
  if(theDSVs.size() == 0) {
    _super::resize(0);
    return;
  }

  _super::resize(theDSVs.begin()->arity());

  set<DSV >::const_iterator iDSV;
  arity_t d;

  /*cerr << "RawPAV constructor notbin:";
    for(iDSV = theDSVs.begin(); iDSV!=theDSVs.end(); iDSV++) {
    cerr << *iDSV << ";";
    }
    cerr << endl;
  */
  unsigned int curval[theDSVs.size()];

  for(d = 0; d < arity(); d++){

    unsigned int i;
    for (i = 0; i<theDSVs.size(); i++){
      curval[i] = 0;
    }

    //while all curvals have not reached the end
    bool doneSearchTheDSV = false;

    //end from the previous chosen span
    int lastlowestend = 0;

    while(!doneSearchTheDSV) {
      int choice = -1;
      unsigned int i = 0;
      //start and end of chosen span section
      int loweststart = 100000;
      int lowestend = 100000;

      //find the lowest start subspan and set choice to the position of this value
      //i = 0; iDSV points to the first item
      for(iDSV = theDSVs.begin(); iDSV!=theDSVs.end(); iDSV++, i++) {
	//check if this span is complete
	if(curval[i]<(*iDSV)[d].size()) {
	  if( (*iDSV)[d][curval[i]].first <= loweststart) {
	    loweststart = (*iDSV)[d][curval[i]].first;
	    lowestend = (*iDSV)[d][curval[i]].second;
	    choice = i;
	  }
	}
      }

      //if we were able to choose a value, add it to the list, else we are done
      if(choice >=0) {
	//push gap if the item's start is not equal to the last end
	if(lastlowestend && loweststart!=lastlowestend)
	  (*this)[d].push_back(0);

	(*this)[d].push_back(choice+1);

	lastlowestend = lowestend;
	curval[choice]++;
      } else {
	doneSearchTheDSV = true;
      }
    }
  }
}

//! Equality test.
bool RawPAV::operator == (const RawPAV& other) const {
  return (Vector<PrecArray >::operator==(other)) ;
}

//! Less than test.
bool RawPAV::operator < (const RawPAV& other) const {
  return (cmp(other) == -1);
}

//! Assignment.
RawPAV& RawPAV::operator=(const RawPAV& other) {
  Vector<PrecArray >::operator=(other);
  return *this;
}

//! indexing
PrecArray& RawPAV::operator[] (const arity_t index) {
  return static_cast<PrecArray&>(_super::operator[](index));
}
const PrecArray& RawPAV::getPA(const arity_t index) const {
  return static_cast<const PrecArray&>(_super::operator[](index));
}
// const PrecArray& RawPAV::operator[] (const arity_t index) const {
// return static_cast<const PrecArray&>(_super::operator[](index));
// }

arity_t RawPAV::arity() const {
  return (arity_t)size();
}

void RawPAV::clear() {
  for(arity_t d = 0; d < arity(); d++) (*this)[d].clear();
}

bool RawPAV::empty() const {
  for(arity_t d = 0; d < arity(); d++)
    if(getPA(d).size()) return false;

  return true;
}

bool RawPAV::monotonic(arity_t d) const {
  PrecArray pa = getPA(d);
  size_t pasz = pa.size();
  for(unsigned char d = 1; d < pasz; ++d)
    if(pa[d] < pa[d-1]) return false;

  return true;
}

bool RawPAV::hasGap() const{
  for(unsigned int i = 0; i < size(); i++){
    const PrecArray& thePA = getPA(i);
    for(unsigned int j = 0; j < thePA.size(); j++)
      if(thePA[j] == 0)
	return true;
  }
  return false;
}

bool RawPAV::hasGap(arity_t dim) const {
  const PrecArray& thePA = getPA(dim);
  for(unsigned int j = 0; j < thePA.size(); j++)
    if(thePA[j] == 0) 
      return true;
  return false;
}

size_t RawPAV::hash() const {
  size_t ret = 0;
  for(arity_t d = 0; d < arity(); d++){
    for(unsigned int i = 0; i < getPA(d).size(); i++){
      ret *= (short)HASH_VALUE;
      ret += getPA(d)[i];
    }
  }
  return ret;
}

// the fanout of one dimension
unsigned short RawPAV::getFanout(const arity_t dim) const {
  const PrecArray& thePA = getPA(dim);
  unsigned short ret = 0;
  for(unsigned int i = 0; i < thePA.size(); i++) {
    // add 1 only if thePA is non-empty
    if (i==0)
      ++ret;
    // if gap
    if(thePA[i] == 0)
      ++ret;
  }
  return ret;
}

// Returns a fanout vector for one role.
FanoutVect RawPAV::getFanoutV(const role_t role) const{
  FanoutVect ret(arity());
  for(arity_t d = 0; d < arity(); d++) {
    ret[d] = 0;
    const PrecArray& thePA = getPA(d);
    for(unsigned int i = 0; i < thePA.size(); i++)
      if(thePA[i] == role)
        ret[d]++;
  }
  return ret;
}

// Returns a fanout vector of the entire RawPAV.
FanoutVect RawPAV::getFanoutV() const {
  FanoutVect ret(arity());
  for(arity_t d = 0; d < arity(); d++) {
    ret[d] = 0;
    const PrecArray& thePA = getPA(d);
    for(unsigned int i = 0; i < thePA.size(); i++) {
      if(i == 0) {
	ret[d]++;
      }
      // if gap
      if(thePA[i] == 0){
        ret[d]++;
      }
    }
  }
  return ret;
}

istream& operator>>(istream& in, RawPAV& pav) {
  char ch;

  in >> ws;
  in >> ch;

  if(in.eof())
    return in;

  if(ch != '[') {
    THROW(" '[' expected in pav, got "<<ch<<endl);
  }

  // enlarges the dimensionality.
  pav.clear();
  pav.resize(1);
  unsigned int d = 0;
  while(1){
    in >> ws;
    while(in.peek() != '|' && in.peek() != ']'){
      unsigned int i;
      in >> i;
      pav[d].push_back(i);
      in >> ws;
    }
    if(in.peek() == '|'){
      in >> ch;
      pav.resize(pav.size() + 1);
      d++;
      in >> ws;
    }
    else if(in.peek() == ']') {
      in >> ch;
      in >> ws;
      break;
    }
  }
  return in;

}

istream& RawPAV::get(istream& in) {
  string str;
  if(in.peek() == '\n')
    getline(in, str, '\n');
  clear();
  getline(in, str, '\n');
  vector<string> strv;
  char buf[100];
  sprintf(buf, "%d", SERIALIZE_DIM);
  if(str.find(buf) != string::npos){
    strv= split(str, DIM_SEP);
  } else {
    strv = split(str, DIM_SEP);
  }
  //assert(strv.size() == pav.arity());
  resize(strv.size());
  for(unsigned int i = 0; i < strv.size(); i++){
#ifdef USE_STRSTREAM
    istrstream ist(strv[i].c_str(), strv[i].size());
#else
    istringstream ist(strv[i]);
#endif
    ist>>(*this)[i];
  }
  return in;
}

ostream& operator<<(ostream& out, const RawPAV& pav) {
  out << "[ ";
  for(arity_t d = 0; d < pav.arity(); d++) {
    if(d != 0)
      out<<DIM_SEP << " ";

    const PrecArray& pa = pav.getPA(d);
    if(pa.size() == 0)
      out<<" ";
    else
      out<<pa;

    /* if(d!=D-1)
       out << " | " ;*/
  }

  out << " ]";
  return out;
}

ostream& RawPAV::serialize(ostream&out) const {
  for(arity_t d = 0; d < arity(); d++) {
    if(d != 0)
      out<<" "<<SERIALIZE_DIM<<" ";

    if(getPA(d).size() == 0)
      out<<" ";
    else
      out<<getPA(d);
  }

  return out;
}

//! Returns the fanout (scalar) for the whole RawPAV
//! This is very different from the getFanoutV method for one role!
short RawPAV::fanout() const {
  // fanout of a RawPAV happens to be the # of non-empty dimensions plus the # of gaps
  short ret = 0;
  for(arity_t i = 0; i < arity(); i++) {
    if (getPA(i).size())
      ++ret;
    for(unsigned int j = 0; j < getPA(i).size(); j++)
      if(getPA(i)[j] == 0)
	++ret;
  };

  return ret;
}

int RawPAV::cmp(const RawPAV& other) const {
  assert(arity() == other.arity());
  int ret;
  for(arity_t i = 0; i < arity(); i++)
    if((ret = MC::cmp(getPA(i) , other.getPA(i))) != 0)
      return ret;

  return 0;
}

RawPAV RawPAV::swapElement(const RawPAV& orig) {
  RawPAV ret(orig);

  for(arity_t d = 0; d < orig.arity(); d++)
    for(unsigned int i = 0; i < ret[d].size(); i++)
      if(ret[d][i] == 1)
        ret[d][i] = 2;
      else if(ret[d][i] == 2)
        ret[d][i]=1;

  return ret;
}

//! reverse order of components
RawPAV RawPAV::reverse(const RawPAV& orig) {
  RawPAV ret(orig);
  ret._super::reverse();

  return ret;
}

// void RawPAV::swapElement()
// {
// for(arity_t d = 0; d < arity(); d++)
// for(unsigned int i = 0; i < operator[](d).size(); i++)
// if(getPA(d)[i] == 1)
// 	(*this)[d][i] = 2;
// else if(getPA(d)[i] == 2)
// 	(*this)[d][i]=1;
// }

// //! reverse order of components
// void RawPAV::reverse() {
// _super::reverse();
// }

bool RawPAV::improperOrder() const {
  unsigned int i1 = MAX_UINT, i2 = MAX_UINT;

  if(size() == 0) return false;

  for(unsigned int i = 0; i < getPA(0).size(); i++)
    if(getPA(0)[i] == 1 && i1 == MAX_UINT)
      i1 = i;
    else if(getPA(0)[i] == 2 && i2 == MAX_UINT)
      i2 = i;

  if(i1 != MAX_UINT) {
    if(i2 != MAX_UINT) {
      if(i2 < i1)
	return true;
      else
	return false;
    }
    else
      return false;
  }
  else if(i2 != MAX_UINT)
    return true;
  else
    return false;
};

//! Normalizes the RawPAV with role preserved. That is, turn 1102201 into 10201.
//! This method doesn't roles, for example, it turns 220101 into 20101.
void RawPAV::rolePreserveNormalize() {
  for(arity_t d = 0; d < arity(); ++d){
    PrecArray::iterator it1D = (*this)[d].begin();
    while(it1D != (*this)[d].end()){
      if(it1D != (*this)[d].begin() && *it1D == *(it1D -1)){
	it1D = (*this)[d].erase(it1D);
      } else {
	++it1D;
      }
    }

    // remove the gap in the front or at the end.
    if(!(*this)[d].empty() && (*this)[d].front() == 0){
      (*this)[d].erase((*this)[d].begin());
    }
    if(!(*this)[d].empty() && (*this)[d].back() == 0){
      (*this)[d].pop_back();
    }
  }
}

//! Returns the set of roles in this RawPAV.
set<role_t> RawPAV::roleSet() const {
  set<role_t> roleset;

  for(arity_t d = 0; d < arity(); ++d){
    for(size_t i = 0; i < getPA(d).size(); ++i){
      if(getPA(d)[i]){
	roleset.insert(getPA(d)[i]);
      }
    }
  }
  return roleset;
}

RawPAV RawPAV::normalize(const RawPAV& orig) {
  arity_t d;

  // re-index the roles (roles are indexes).
  map<role_t, role_t> reindex;
  for( d = 0; d < orig.arity(); ++d){
    for(size_t i = 0; i < orig.getPA(d).size(); ++i){
      if(role_t r = orig.getPA(d)[i]){
	if(reindex.find(r) == reindex.end()){
	  reindex[r] = reindex.size() ;
	}
      }
    }
  }

  RawPAV ret(orig);
  ret.rolePreserveNormalize();

  // apply the reindex
  for(d = 0; d < ret.arity(); ++d){
    for(size_t i = 0; i < ret[d].size(); ++i){
      if(ret[d][i]){
	ret[d][i] = reindex[orig.getPA(d)[i]];
      }
    }
  }

  return ret;
}

// void RawPAV::normalize()
// {

// arity_t d;

// rolePreserveNormalize();

// // re-index the roles (roles are indexes).
// map<role_t, role_t> reindex;
// for( d = 0; d < arity(); ++d){
// for(size_t i = 0; i < (*this)[d].size(); ++i){
// if(role_t r= (*this)[d][i]){
// 	if(reindex.find(r) == reindex.end()){
// 	  reindex[r] = reindex.size() ;
// 	}
// }
// }
// }

// RawPAV tmp = *this;

// // apply the reindex to this RawPAV.
// for(d = 0; d < arity(); ++d){
// for(size_t i = 0; i < (*this)[d].size(); ++i){
// if((*this)[d][i]){
// 	(*this)[d][i] = reindex[tmp[d][i]];
// }
// }
// }
// }

/**
 * Return the maximum role. This RawPAV do not have to be normalized.
 */
role_t RawPAV::maxRole() const {
  role_t role = 0;
  for(arity_t d = 0; d < arity(); ++d){
    for(size_t i = 0; i < getPA(d).size(); ++i){
      role = (role < getPA(d)[i]) ? getPA(d)[i] : role;
    }
  }

  return role;
}

short RawPAV::rank() const {
  if(arity() == 0)
    return 0;

  set<role_t> roleset = roleSet();
  return (short)roleset.size();
}

//! Returns the number of roles on dimension d.
short RawPAV::rank(arity_t d) const {
  //short ret = 0;
  role_t role;
  size_t s;
  s = getPA(d).size();
  set<role_t> roles;
  for(size_t i = 0; i < s; i++){
    role = getPA(d)[i];
    if(role){
      roles.insert(role);
    }
  }

  return roles.size();
}

RawPAV RawPAV::deleteRoleAndNormalize(const RawPAV& orig, const role_t delRole) {

  // deletion of gaps should not rely on this representation of gaps,
  // nor use this method
  assert(delRole != GAPVALUE);

  // uncomment to debug
  // cerr << "Deleting role " << delRole << " from " << orig << endl;

  RawPAV ret(orig);

  for(arity_t i = 0; i < orig.arity(); i++) {

    // PrecArray* a1dPtr = *(operator[](i));
    PrecArray::const_iterator aio;
    PrecArray::iterator air;

    // replace all the deleted role's positions with gaps, and shift higher roles down
    for(aio = orig.getPA(i).begin(), air = ret[i].begin();
        aio != orig.getPA(i).end();
        ++aio, ++air) {
      if (*aio == delRole)
        *air = GAPVALUE;
      else if (*aio > delRole)
        --(*air);
    }

    // then, merge adjacent gaps
    air = ret[i].begin();
    while(air != ret[i].end()) {
      if (*air == GAPVALUE && (air + 1) != ret[i].end() && *(air + 1) == GAPVALUE)
        // erase() should automatically advance the iterator to the next element
        ret[i].erase(air);
      else
        ++air;
    }

    // don't forget to remove leading and trailing gaps!
    if (ret[i].size() && ret[i][0] == GAPVALUE)
      ret[i].erase(ret[i].begin());
    size_t remaining = ret[i].size();
    if(remaining && ret[i][remaining - 1] == GAPVALUE)
      ret[i].erase(ret[i].end() - 1);
  }
  // uncomment to debug
  // cerr << "Result: " << orig << " of rank " << rank() << endl;

  return ret;
}

// void RawPAV::deleteRoleAndNormalize(const role_t delRole) {

// // deletion of gaps should not rely on this representation of gaps,
// // nor use this method
// assert(delRole != GAPVALUE);

// // uncomment to debug
// // cerr << "Deleting role " << delRole << " from " << *this << endl;

// for(arity_t i = 0; i < arity(); i++) {

// // PrecArray* a1dPtr = *(operator[](i));
// PrecArray::iterator ai;

// // replace all the deleted role's positions with gaps, and shift higher roles down
// for(ai = (*this)[i].begin(); ai != (*this)[i].end(); ++ai) {
// if (*ai == delRole)
// 	*ai = GAPVALUE;
// else if (*ai > delRole)
// 	--(*ai);
// }

// // then, merge adjacent gaps
// ai = (*this)[i].begin();
// while(ai != (*this)[i].end()) {
// if (*ai == GAPVALUE && (ai + 1) != (*this)[i].end() && *(ai + 1) == GAPVALUE)
// 	// erase() should automatically advance the iterator to the next element
// 	(*this)[i].erase(ai);
// else
// 	++ai;
// }

// // don't forget to remove leading and trailing gaps!
// if ((*this)[i].size() && (*this)[i][0] == GAPVALUE)
// (*this)[i].erase((*this)[i].begin());
// size_t remaining = (*this)[i].size();
// if(remaining && (*this)[i][remaining - 1] == GAPVALUE)
// (*this)[i].erase((*this)[i].end() - 1);
// }
// // uncomment to debug
// // cerr << "Result: " << *this << " of rank " << rank() << endl;
// }

/** Generate all possible d-dimensional (first argument) RawPAVs with
    fanout (second argument) less than or equal to a limit. The
    fanout limit is also applied to each role. The generated RawPAVs
    are accumulated into a Vocabulary (third argument). */
void RawPAV::generate(const arity_t& dim, short fanoutLmt, Vocabulary<RawPAV>& pavVocabulary) {
  unsigned int i;
  vector<role_t> pavRole1(fanoutLmt * 2 - 1);
  vector<role_t> pavJoin(2*pavRole1.size() + 1);

  // fill the roles for pav 1.
  for(i = 0; i < pavRole1.size(); ++i){
    pavRole1[i] = (i % 2) ? 0 : 1;
  }
  // fill the roles for pavJoin.
  vector<role_t>::iterator itPavJoin = copy(pavRole1.begin(), pavRole1.end(), pavJoin.begin());
  *itPavJoin = 0; // fill a gap.
  while(++itPavJoin != pavJoin.end()){
    (*itPavJoin) = ((itPavJoin - pavJoin.begin()) % 2) ? 0 : 2;
  }

  do{
    generate_helper(pavJoin, dim, fanoutLmt, pavVocabulary);
  } while(next_permutation(pavJoin.begin(), pavJoin.end()));

  do{
    generate_helper(pavRole1, dim, fanoutLmt, pavVocabulary);
  } while(next_permutation(pavRole1.begin(), pavRole1.end()));
}

/** A helper function used in generate(...). It generates all legal
    (constrained by fanout limit) RawPAVs based on an unrolled (by
    concatenating on dimensions) unnormalized (e.g., might having 11102,
    or 2011) RawPAV (vector of roles). Currently only support 1-D and
    2-D. */
void RawPAV::generate_helper(vector<role_t> 			&unNormalizedRawPAV,
			  const arity_t 		&dim,
			  short 			fanoutLmt,
			  Vocabulary<RawPAV> &pavVocabulary) {
  if(dim != 1 && dim != 2){
    THROW("Unsupported dimensionality: " << dim);
  }

  vector<role_t>::iterator itRawPAV;
  /*
   * Judge whether it is in mad order, that is, the
   * first role is not 1 if there are roles in.
   */
  itRawPAV = find_if(unNormalizedRawPAV.begin(), unNormalizedRawPAV.end(), bind2nd(not_equal_to<role_t>(), 0));
  // if there is no role or the first role is not 1, return
  if(itRawPAV == unNormalizedRawPAV.end() || *itRawPAV != (role_t)1){
    return;
  }

  //Create an RawPAV (maybe unnormalized) from the unrolled RawPAV (vector of roles).

  if(dim == 1){
    RawPAV anRawPAV(dim);
    anRawPAV[0].vector<role_t>::operator=(unNormalizedRawPAV);
    anRawPAV.rolePreserveNormalize();
    if(anRawPAV.fanout() <= fanoutLmt){
      pavVocabulary.addWord(anRawPAV);
    }
  } else if(dim == 2){
    for(itRawPAV = unNormalizedRawPAV.begin(); itRawPAV != unNormalizedRawPAV.end(); ++itRawPAV){

      RawPAV anRawPAV(dim);
      copy(unNormalizedRawPAV.begin(), itRawPAV, back_inserter(anRawPAV[0]));
      copy(itRawPAV, unNormalizedRawPAV.end(), back_inserter(anRawPAV[1]));
      anRawPAV.rolePreserveNormalize();

      if(anRawPAV.fanout() <= fanoutLmt){
	FanoutVect fov = anRawPAV.getFanoutV();
	if(fov[0] == 0 && fov[1] == 2 || fov[0] == 2 && fov[1] == 0)
	  continue;
	pavVocabulary.addWord(anRawPAV);
      }
    }
    // end
    RawPAV anRawPAV(dim);
    copy(unNormalizedRawPAV.begin(), unNormalizedRawPAV.end(), back_inserter(anRawPAV[0]));
    anRawPAV.rolePreserveNormalize();
    if(anRawPAV.fanout() <= fanoutLmt){
      FanoutVect fov = anRawPAV.getFanoutV();
      if(fov[0] == 0 && fov[1] == 2 || fov[0] == 2 && fov[1] == 0)
	return;
      pavVocabulary.addWord(anRawPAV);
    }
  }
}

//Check if the RawPAV is in normal form, from the perspective of aRole
bool RawPAV::isNormalFormForRole(role_t aRole) const {
  bool normalForm = true;
  for (unsigned int d = 0; d < arity(); d++){
    if(getPA(d).size() > 0){
      if( getPA(d)[0] != aRole ) {
	normalForm = false;
      }
      break;
    }
  }
  return normalForm;
}

bool RawPAV::endsWith(arity_t dim, role_t role, unsigned int segmentNum) const {
  unsigned int segCount = 0;
  role_t last = 0;
  for(unsigned int j = 0; j <= getPA(dim).size(); j++){
    if(j == getPA(dim).size()){
      if(last != role){
	return false;
      }
      else
	return true;
    }
    if(getPA(dim)[j] == 0){
      segCount++;
      if(segCount > segmentNum){
	if(last != role)
	  return false;
	else
	  return true;
      }
    }
    
    last = getPA(dim)[j];
  }
  return false;
}

bool RawPAV::beginsWith(arity_t dim, role_t role, unsigned int segmentNum)const{
  unsigned int segCount = 0;
  role_t last = 0;
  for(unsigned int j = 0; j < getPA(dim).size(); j++){
    if(getPA(dim)[j] == 0){
      segCount++;
    }
    else{
      if(last == 0){
	if(segCount == segmentNum){
	  if(role != getPA(dim)[j])
	    return false;
	  else
	    return true;
	}
      }
    }
    
    last = getPA(dim)[j];
  }
  return false;
}

bool RawPAV::missing(arity_t dim, role_t role, unsigned int segmentNum)const{
  unsigned int segCount = 0;
  for(unsigned int j = 0; j < getPA(dim).size(); j++){
    if(getPA(dim)[j] == 0){
      segCount++;
    }
    else if(getPA(dim)[j] == role)
      if(segCount == segmentNum)
	return false;
    else if(segCount > segmentNum)
      return true;
  }
  return true;
}

//! Returns true if PAV ends/beg/does not have  in each dimension
bool RawPAV::endsBegWith(const vector<role_t>& vecRoles, const vector<unsigned int>& segment, const vector<unsigned int>& startOrEnd)const{
  for(unsigned int i = 0; i < this->size(); i++){
    //1 means begin with
    if(startOrEnd[i] == 1){
      if(!beginsWith(i,vecRoles[i],segment[i]))
	return false;
    }
    //2 means ends with
    else if(startOrEnd[i] == 2){
      if(!endsWith(i,vecRoles[i],segment[i]))
	return false;
    }
    //0 means does not contain
    else if(startOrEnd[i] == 0){
      if(!missing(i,vecRoles[i],segment[i]))
	return false;
    }
    //else, ignore.
  }
   return true;
}
  
