#ifndef _RawPAV_H_
#define _RawPAV_H_

#include <set>
#include <memory>
#include <map>
#include "Vector.hpp"
#include "Vocabulary.hpp"
#include "FanoutVect.hpp"
#include "PrecArray.hpp"
#include <boost/serialization/serialization.hpp>

class DSV;

//! precedence array vector
class RawPAV: public Vector< PrecArray > {
  typedef Vector<PrecArray> _super;
public:

  //! Empty constructor.
  RawPAV() { resize(0);}

  //! Constructor.
  RawPAV(const arity_t& dim);

  //! constructor from one PrecArray
  RawPAV(const PrecArray& pa);

  //! constructor from an arity, a PrecArray, and the position of the PrecArray
  //! the other components will be empty
  RawPAV(const arity_t& dim, const PrecArray& pa, const arity_t& psn);

  //! constructor from a Vector<PrecArray>
  RawPAV(const Vector<PrecArray>& pa);

  //! constructor from a fan-out vector
  //! all elements will have the same role
  RawPAV(const FanoutVect& fov);

  //! constructor from a single DSV
  //! all elements will have the same role
  RawPAV(const DSV& dsv);

  //! constructor from two DSpans.
  RawPAV(const DSV& span1, const DSV& span2);

  //! constructor from set of DSpans.
  RawPAV(const set<DSV>& spans);

  //! Equality test.
  bool operator == (const RawPAV& other) const;

  //! Less than test.
  bool operator < (const RawPAV& other) const;

  //! Assignment.
  RawPAV& operator=(const RawPAV& other);

  //! indexing
  // const PrecArray& operator[] (const arity_t index) const;
  const PrecArray& getPA(const arity_t index) const;

  arity_t arity() const;

  //! Judges whether empty.
  bool empty() const;

  //! Judges whether there is a gap in at least one dimension.
  bool hasGap() const;

  //! Judges whether there is a gap in the given dimension.
  bool hasGap(arity_t dim) const;

  //! are the precedences monotonic in component d?
  bool monotonic(arity_t d) const;

  //! For hashing.
  size_t hash() const;

  //! fanout
  short fanout() const;

  // the fanout of one dimension
  unsigned short getFanout(const arity_t dim) const;

  //! Returns a fanout vector for one role
  FanoutVect getFanoutV(const role_t side) const;

  //! Returns the fanout vector of the entire RawPAV.
  FanoutVect getFanoutV() const;

  //! For ordering.
  int cmp(const RawPAV& other) const;

  ///////////////////////////////////////////////////////////////////////////
  //////// The following methods are now const instead of destructive ///////
  //! Remove one role, and normalize the RawPAV.
  static RawPAV deleteRoleAndNormalize(const RawPAV& orig, const role_t i);

  //! reverse order of components
  static RawPAV reverse(const RawPAV& orig);

  //! Exchanges the value of elements: 1->2, 2->1, and 0 remained.
  //! Currently only works for binary case.
  static RawPAV swapElement(const RawPAV& orig);

  //! Normalizes the RawPAV.
  static RawPAV normalize(const RawPAV& orig);

  ///////////////////////////////////////////////////////////////////////////

  //! Judges whether the indices are not increasing order in the
  //! first dimension. <1,2> -- false; <2,1> -- true.
  bool improperOrder() const;

  //! Returns the set of roles in this RawPAV.
  set<role_t> roleSet() const;

  //! The number of different roles
  short rank() const;

  //! Returns the number of roles on dimension d.
  short rank(arity_t d) const;

  /**
   * Return the maximum role. This RawPAV do not have to be normalized.
   */
  role_t maxRole() const;

  //! Decides if RawPAV is in normal form, for a given role (typically,
  //! you pass the role that is supposed to be the head)
  bool isNormalFormForRole(role_t aRole) const;



  //! returns true if condition is met.
  //!  @param vecRoles what role to look for in each dimension.
  //!  @param segment which continuous segment to look in (0 based)
  //!  @param startOrEnd 0:nonexistent 1:start 2: end 3:ignore
  //!  must meet ALL conditions in each dimension
  bool endsBegWith(const vector<role_t>& vecRoles, const vector<unsigned int>& segment, const vector<unsigned int>& startOrEnd)const;

  //! returns true if 
  //! in dimension "dim" 
  //! begins with "role" 
  //! in continuous segment "segmentNumber"
  bool beginsWith(arity_t dim, role_t role, unsigned int segmentNum)const;

  //! returns true if in dimension "dim" ends with the "role" in continuous segment "segmentNumber"
  //! returns true if 
  //! in dimension "dim" 
  //! endswith "role" 
  //! in continuous segment "segmentNumber"
  bool endsWith(arity_t dim, role_t role, unsigned int segmentNum)const;

  //! returns true if 
  //! in dimension "dim" 
  //! does not contain "role" 
  //! in continuous segment "segmentNumber"
  bool missing(arity_t dim, role_t role, unsigned int segmentNum)const;
  
  //! Generate all possible d-dimensional (first argument) RawPAVs with
  //! fanout (second argument) less than a limit.
  //! The fanout limit is also applied to each role.
  //! The generated RawPAVs are accumulated into a Vocabulary (third argument).
  static void generate(const arity_t&, short, Vocabulary<RawPAV>&);

  istream& get(istream& in);

  //! Inputs RawPAV.
  friend istream& operator>>(istream& in, RawPAV& pav);

  //! Outputs RawPAV.
  friend ostream& operator<<(ostream& out, const RawPAV& pav);

  ostream& serialize(ostream&out) const;

  //! Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    ar & boost::serialization::base_object<_super>(*this);
  }

private:

  //! the const version is public, but the intrusive version if private
  PrecArray& operator[] (const arity_t index);

  //! Clear the contents, but retain the dimension.
  void clear();

  //! Normalizes the RawPAV with role preserved. That is, turn 1102201 into 10201.
  //! This method descent roles, for example, it turns 220101 into 20101.
  void rolePreserveNormalize();

  //! A helper function used in generate(...). It generates all legal
  //! (constrained by fanout limit) RawPAVs based on an unrolled
  //! (by concatenating on dimensions) unnormalized (e.g., might having
  //! 11102, or 2011) RawPAV (vector of roles).
  //! Currently only support 1-D and 2-D.
  static void generate_helper(vector<role_t> &unNormalizedRawPAV,
                            const arity_t &dim,
  	   	            short fanoutLmt,
		            Vocabulary<RawPAV> &pavVocabulary);
};

namespace HashNS{
  template<>
  struct hash<RawPAV > {
    size_t operator()(const RawPAV& pav) const { return pav.hash();}
  };
}

#endif
