#include <string>
#include <fstream>
#include "GPDefs.hpp"
#include "Exceptions.hpp"

#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

#include "RelativePruningStrategy.hpp"
#include "FactoredItemIndex.hpp"
#include "NaiveItemIndex.hpp"
#include "Hedge.hpp"

/*
 * Constructor.
 * @param configFile The config file.
 */
RelativePruningStrategy::RelativePruningStrategy (const string& configFile) 
  : m_pruneDummiesAggressively(false) {

  unsigned int arity ;
  string itemIndexType;

  // compile the config file.
  BPO::options_description config("Configuration");
  config.add_options()
    ("ItemIndexType", BPO::value<string>(&itemIndexType), "The type of the item index.")
    ("Dimensionality", BPO::value<unsigned int>(&arity), "How many components for factored index?")
    ("RelativeThresholdAdditive", BPO::value<double>(), "The additive difference threshold")
    ("RelativeThresholdFactor", BPO::value<double>(), "The multiplicative difference threshold")
    ("PruneDummiesAggressively", BPO::value<bool>(), "Prune dummies that are not best in their DSV?")
    ;
  BPO::variables_map vm;
  cerr << "RelativePruningStrategy: Compiling config file " << configFile << endl;
  iFile ifs(configFile.c_str());
  store(parse_config_file(ifs, config), vm);
  notify(vm);

  m_t = 0;
  if(vm.count("RelativeThresholdAdditive")) m_t = vm["RelativeThresholdAdditive"].as<double>();
  if(vm.count("RelativeThresholdFactor")) m_t = -log(vm["RelativeThresholdFactor"].as<double>());
  if(vm.count("PruneDummiesAggressively")) m_pruneDummiesAggressively = vm["PruneDummiesAggressively"].as<bool>();

  if(debug(1)) {
    cerr << "Option values of \"RelativePruningStrategy\"" << endl << endl;
    cerr << OptionsPrinter(config, vm) << endl;
  }
  
  if(itemIndexType == ""){
    THROW("Error: The type of item index is not specified for "
	<<"the relative pruning strategy!"<<endl);
  }

  string strItemIndexType = tolower(itemIndexType);

  if(strItemIndexType == "factored"){
    // default arity = 2
    m_pItemIndex = new FactoredItemIndex(arity);

  } else if(strItemIndexType == "naive"){
    m_pItemIndex = new NaiveItemIndex();

  } else if(strItemIndexType == "none") {
    m_pItemIndex = NULL;
    if (m_t >= 0) {
      THROW("Error: your threshold is below one, but your item "
	    << "index type is \'none\'." << endl);
    }
  } else {
    THROW("Error: unknown item index type!"<<endl);
  }

}

bool RelativePruningStrategy::prunable(const Item* item) const {
  arity_t d;
  const DSV* span = item->getDSV();

  // the number of times that the factored item index says "prune"
  size_t numBad = 0;
  // the number of factored indexes (should be 1 for terminal/preterminal scan/load items).
  size_t numFactors = min(span->size(), m_pItemIndex->numFactors());

  // retrieves items from the chart according to the span
  for(d = 0; d < numFactors; ++d){
    vector<Item*> edges;
    // retrieve method also returns the best item f-value
    // in the same index entry.
    SCORE best = m_pItemIndex->retrieve(*span, d, edges);
    SCORE itemScore = item->getScore();
    if (debug(3))
      cerr << ": item score = " << itemScore << " ; competing score = " << best << endl;

    if (m_pruneDummiesAggressively) {
      // if it's a dummy, then it loses to any item with same span and
      // at least as low a cost
      const Hedge* pH;
      if ((pH = dynamic_cast<const Hedge*>(item)))
	if (pH->getNTLink().isDummy())
	  if (best <= itemScore)
	    return true;
    }

    // If any of the indexes say that this item
    // should be pruned, then we prune it.
    if(prunable(itemScore, best)){
      ++numBad;
      if (debug(3))
	cerr << "Item pruned!" << endl;
      return true;
    }
  }

  return false;
}

bool RelativePruningStrategy::prunable(SCORE score, SCORE best) const {
  // If the relative pruning threshold is less than 0, then
  // there is no pruning at all.
  if(m_t < 0){ return false; }

  SCORE diff = score - best;
  if(diff > m_t){
    return true;
  } else {
    return false;
  }
}

/** Inserts a chart index into the item index table.*/
void RelativePruningStrategy::memoize(const Item* item) {
  if(m_pItemIndex)
    m_pItemIndex->insert(item);
}

/* Resets the thresholds to the initial values, clears the item index.*/
void RelativePruningStrategy::reset() {
  if(m_pItemIndex){
    m_pItemIndex->clear();
  }
}
