#include "HeadMTree.hpp"
#include "ScoreSet.hpp"
#include "GPDefs.hpp"
#include "TranslationEvaluator.hpp"
#include "Production.hpp"
#include "HeadProduction.hpp"
#include "Exceptions.hpp"

//!  EMUpdate rule encapsulates comparson between
//!  proposed and supervisory standard parses, as well as
//!  updates to the learner's model.
 
TranslationEvaluator::TranslationEvaluator(){ }

/**
 * This method will compare a supervisory parse with a 
 * candidate parse using a unigram comparison of the 
 * lexical items, and will return the resulting score
 * which is calculated as the number of matched words
 * in the two sentences, divided by the total number
 * of words in the two sentences.
 */
SCORE TranslationEvaluator::evalByWords(
               WTreeVec &supervisorParse, 
               WTreeVec &candidateParse) const{
  SCORE transScore = 0.0;
  Histogram<Terminal> supervisorScores; //histogram of terminals in supervisor
  Histogram<Terminal> candidateScores; //histogram of terminals in candidate
  int totalCount = 0;

  // increment the supervisor parse into the scoreset
  cerr << "supervising observations: " ;
  for (unsigned int i = 0; i < supervisorParse.size(); ++i){
    MTree tr = *supervisorParse[i].first;
    cerr << "supervising tree" << tr << endl;
    SentenceTuple result = tr.yield();
    cerr << result;
    // walk the observation sequence building a Histogram of Terminals
    Vector<Terminal> terminals = result[1];
    cerr << "size " << terminals.size();
    for (unsigned int j = 0; j < terminals.size(); ++j){
      totalCount++;
      cerr << terminals[j];
      supervisorScores.increment(terminals[j]);
    }
  }
  cerr << endl;

  
  // increment the candidate parse into the scoreset
  cerr << "candidate observations: ";
  for (unsigned int i = 0; i < candidateParse.size(); ++i){
    MTree tr = *candidateParse[i].first;
    SentenceTuple result = tr.yield();

    // walk the observation sequence building a Histogram of Terminals
    Vector<Terminal> terminals = result[1];
    if (terminals.size() > 0){
      cerr << "candidate observations: " << terminals.size();
      for (unsigned int j = 0; j < terminals.size(); ++j){
        totalCount++;
        cerr << terminals[j];
        candidateScores.increment(terminals[j]);
      }
    }
  }
  cerr << endl;

  // we now have a histogram for each terminal sequence.
  // their combined size - 2*their intersection will
  // give the number of unmatched terminals.
  Histogram<Terminal> intersect = candidateScores.intersect(supervisorScores);
  transScore = (2*candidateScores.sum())/totalCount;

  return transScore;
} 

/**
 * This method will compare a supervisory parse with a
 * candidate parse by comparing the tree structure of 
 * the two parses.  This will work by dividing the 
 * number of matching productions in a parse by
 * the total number of productions in the parse.
 * It will then return the resulting score.
 */
SCORE TranslationEvaluator::evalByParse(
                  WTreeVec &supervisorParse,
                  WTreeVec &candidateParse) const{
  SCORE transScore = 0.0;
  ScoreSet<Production> prodScores; //create a new ScoreSet object
  HeadMTree::iterator supervisorIt;
  HeadMTree::iterator candidateIt;
  //HeadMTree::iterator vectorIt;
  int totalCount = 0;
  int differenceCount = 0;

  // increment the supervisor parse into the scoreset
  for (unsigned int i = 0; i < supervisorParse.size(); ++i){
    MTree tr = *supervisorParse[i].first;
    for (supervisorIt = tr.begin(); 
         supervisorIt != tr.end(); ++supervisorIt){
      if (boost::get<NTLink>(&BGET_EITHER_LABEL(*supervisorIt))){
        totalCount++;
        Production* aProd = new HeadProduction(supervisorIt);
        prodScores.increment(*aProd);
      }
    }
  }

  // find the difference between the two by decrementing the prodScores
  // and counting the items remaining in prodScores plus the 
  // unmatched items.
  for (unsigned int i = 0; i < candidateParse.size(); ++i){
    MTree tr = *candidateParse[i].first;
    for (candidateIt = tr.begin();
         candidateIt != tr.end(); ++candidateIt){
      try {
        if (boost::get<NTLink>(&BGET_EITHER_LABEL(*candidateIt))){
          totalCount++;
          Production* aProd = new HeadProduction(candidateIt);
          prodScores.decrement(*aProd);
        }
    
      } catch(KeyNotFound<Production>){
        differenceCount++;
      } 
    }
  }
  // add the mass remaining in the score set to the differences found
  // so far
  transScore = (prodScores.sum() + differenceCount)/totalCount;
  cerr << "score by parses: " << transScore << endl;
  return transScore;
}

