#include "TreeConstrainedMTG.hpp"

#include <algorithm>
#include <queue>
#include <set>

#include "FanoutVect.hpp"
#include "GPDefs.hpp"
#include "Pair.hpp"
#include "Exceptions.hpp"
#include "MoreExceptions.hpp"
#include "NTLink.hpp"
#include "DSV.hpp"
#include "HeadMTree.hpp"
#include "HiddenDSVs.hpp"

#include "Item.hpp"
#include "NodeRefHedge.hpp"
#include "GlobalVocab.hpp"
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

TreeConstrainedMTG::
TreeConstrainedMTG(arity_t dm,
		   const string& configFile,
		   BSP<BPO::options_description> configOptions)
  : _grammar(dm),
    _super(dm, configFile, addOptionDescriptionsTo(configOptions)),
    m_pRestrictingTree(NULL) {

  cerr << "TreeConstrainedMTG: Compiling config file " << configFile << endl;
  BPO::variables_map vm;
  //variables_map vm;
  iFile ifs(configFile.c_str());
  store(parse_config_file(ifs, *configOptions), vm);

  //was causing segFault in OPT mode
  //notify(vm);

  string optMonoTB;
  if(vm.count("TreeStream")){
    optMonoTB = vm["TreeStream"].as<string>();
  }

  if(debug(1)) {
    cerr << "Option values of \"TreeConstrainedMTG\"" << endl << endl;
    cerr << OptionsPrinter(*configOptions, vm) << endl;
  }
  m_pMonoTBStream = new iFile(optMonoTB.c_str());
}

TreeConstrainedMTG:: ~TreeConstrainedMTG(){
  delete m_pRestrictingTree;
  delete m_pMonoTBStream;
}

// prepare for another pass through the constraining treebank
void TreeConstrainedMTG::clear() {
  rewind(*m_pMonoTBStream);
}

// Updates the content according to a parse and a sentence tuple.
bool TreeConstrainedMTG::prepareFor(SentenceTuple& sntT) {

  // remember the current sentence
  m_pCurrSent = &sntT[m_constrainedDim];

  // get new constraining tree
  PositionTree parseTree;
  parseTree.readPositionTree(*(m_pMonoTBStream),m_constrainedDim);

  // delete previous parseTree
  delete m_pRestrictingTree;

  m_pRestrictingTree = new PositionTree;
  m_pRestrictingTree->addRoot(parseTree, *m_pRestrictingTree);

  m_span2leaf.clear();
  PositionTree::iterator i;
  for(i = m_pRestrictingTree->begin(); i != m_pRestrictingTree->end(); ++i) {
    /* if it is a leaf */
    if(!i.number_of_children()){
      DSpan aSpan =(BGET_SPAN(*i))[0];
      m_span2leaf[aSpan].insert(i);
    }
  }

  if((BGET_SPAN(*m_pRestrictingTree->begin()))[0][0].second != sntT[m_constrainedDim].size()) {
    cerr << "1D TREE SPAN SIZE: " << (BGET_SPAN((*m_pRestrictingTree->begin())))[0][0].second << endl;
    cerr << "SNT TUPLE SIZE: " << sntT[m_constrainedDim].size() << endl;
    cerr << "ON DIM: " << (int)(m_constrainedDim) << endl;
    THROW("Mismatch in span size of 1D tree and sentence tuples. ");
  }
  

  return true;
}

void TreeConstrainedMTG::
filterArrangementFor(BSP<set<Pair<bool,PAV> > >& arrgs, const Item* trigger) const {
  const NodeRefHedge* pNRH = dynamic_cast<const NodeRefHedge*>(trigger);
  if (! pNRH)
    THROW("dunno what to do with items other than NodeRefHedges here");

  const DSpan& theDSpan = (*(trigger->getDSV()))[m_constrainedDim];
  const PositionTree::iterator& iPT = pNRH->getNode()[m_constrainedDim];
  // DSpan of lowest subsuming PositionTreeNode
  const DSpan& lsnDSpan = iPT->getDSV()[m_constrainedDim];
  unsigned short triggerLeftEdge = theDSpan[0].first;
  unsigned short triggerRightEdge = theDSpan[theDSpan.size() - 1].second;
  unsigned short lsnLeftEdge = lsnDSpan[0].first;
  unsigned short lsnRightEdge = lsnDSpan[lsnDSpan.size() - 1].second;
  // if one edge matches but the other doesn't, then we can filter
  if (triggerLeftEdge == lsnLeftEdge && triggerRightEdge < lsnRightEdge) {
    // stub
  } else if (triggerRightEdge == lsnRightEdge && triggerLeftEdge > lsnLeftEdge) {
    // stub
  }
};

void TreeConstrainedMTG::getPossibleCons(const Item* item1, const Item* item2, const Pair<bool,PAV>& arrg,
					 vector<Item*>& possConseqs) {
  //Will contain NTlink for consequent to be returned.
  NTLink outLabel(1);
  //Will contain NTlink for alternative consequent to be returned.
  // (in the case of no heads)
  NTLink outLabel2(1);
  //Will contain HeirVector for consequent to be returned.
  HeirRoleVector outHRV(1);
  //Will contain DSV for consequent to be returned.
  DSV outDSV(1);
  //True if the span is empty on one component in both labels.
  bool oneDimEmpty = false;
  bool noEmptyIts = false;

  if(debug(5)){
    bool tmp = GlobalVocab::s_noVocabsOut;
    GlobalVocab::s_noVocabsOut = false;
    cerr << "TreeConstrainedMTG::getPossibleCons QUERY: " << endl
	 << *((NodeRefHedge*)item1) << endl
	 << " AND " << *((NodeRefHedge*)item2)<< endl;
    GlobalVocab::s_noVocabsOut = tmp;
  }

  // Get information from item1
  const NodeRefHedge * nrhedge1 = static_cast<const NodeRefHedge*>(item1);
  const DSV * inDSV1 = item1->getDSV();
  const DSpan& inSpan1 = (*inDSV1)[m_constrainedDim];
  PositionTree::iterator iPT1 = nrhedge1->getNode()[m_constrainedDim];
  NTLink label1 = nrhedge1->getNTLink();
  vector<short> inHeadPosn1 = nrhedge1->getHeads();

  // Get information from item2
  const NodeRefHedge * nrhedge2 = static_cast<const NodeRefHedge*>(item2);
  const DSV * inDSV2 = nrhedge2->getDSV();
  const DSpan& inSpan2 = (*inDSV2)[m_constrainedDim];
  PositionTree::iterator iPT2 = nrhedge2->getNode()[m_constrainedDim];
  NTLink label2 = nrhedge2->getNTLink();
  vector<short> inHeadPosn2 = nrhedge2->getHeads();

  DSV combinedDSV(*inDSV1 + *inDSV2);
  if(combinedDSV.size() < 2
     || combinedDSV[0].size() == 0 
     || combinedDSV[1].size() == 0){
    oneDimEmpty = true;
  }
  if (combinedDSV.size() > m_constrainedDim)
    outDSV[0] = combinedDSV[m_constrainedDim];

  //Lowest subsuming node
  PositionTree::iterator lsnIt;

  //If both position tree pointers are null, we assume they can combine to a NULL.
  //(independent rewriting)
  if(iPT1 == 0 && iPT2 == 0){
    //vector with one position tree pointer in it
    vector<PositionTree::iterator> vecIt(1, iPT1);
    outLabel.setNT(0,label1.getNT(m_constrainedDim)->clone());
    //if one is null in one dimension, then don't climb the position tree
  } else if(iPT1 == 0 && iPT2 != 0) {
    lsnIt = iPT2;
    outHRV.setRole(0,2);
    outLabel.setNT(0,BGET_POSITION_LABEL(*iPT2).getNT(0)->clone());
     
    if(dynamic_cast<const LexNT*>(label2.getNT(m_constrainedDim))){
      TLink tl(item1->arity());
      tl[0] = label2.getTLink()[m_constrainedDim];
      outLabel.setTLink(tl);
    }
    //if iPT2 is the inactive pointer (null in other dimension)
  } else if (iPT1 != 0 && iPT2 == 0) {
    lsnIt = iPT1;
    outHRV.setRole(0,1);
    outLabel.setNT(0,BGET_POSITION_LABEL(*iPT1).getNT(0)->clone());
    if(dynamic_cast<const LexNT*>(label1.getNT(m_constrainedDim))){
      TLink tl(1);
      tl[0] = label1.getTLink()[m_constrainedDim];
      outLabel.setTLink(tl);
    }
    //lastly, if neither position tree pointer is 0
  } else {
    noEmptyIts = true;
  
    //this is true if it1 points to a node in the position tree that 
    //is completely covered by the item. If so, we can progress up the tree.
    //if not, we have to wait until the other children are covered
    bool complete1 = true;
    if(nrhedge1->getChildCounter()[m_constrainedDim] > 1){
      complete1 = false;
    }
      
    bool complete2 = true;
    if(nrhedge2->getChildCounter()[m_constrainedDim] > 1){
      complete2 = false;
    }
    
    //Find the lowest subsuming node of the pair of the input iterators
    lsnIt = computeTreeIterator(iPT1, iPT2, complete1,complete2);
    //if the two don't share a subsuming node, exit
    if(lsnIt == NULL){
      if(debug(5))cerr << "LSN NOT FOUND " << endl;
      return;
    }
    
    ///// BEGIN EPSILON FILTER
    //filter out 1D attachments in one direction
    if(m_rightAttach){
      //we can only filter this if the word to the right of the
      //epsilon is in the same lowest subsuming node.
      if((BGET_SPAN(*lsnIt)[0][0].first > inSpan1[0].second) && 
	 (BGET_SPAN(*lsnIt)[0][0].first > inSpan2[0].second)){
	if(inDSV1->empty(1-m_constrainedDim) and 
	   !inDSV2->empty(m_constrainedDim) ){
	  bool found = false;
	  for(unsigned int i = 0; i < inSpan1.size(); i++){
	    if(
	       inSpan1[i].second
	       == 
	       (inSpan2[0].first)){
	      found = true;
	      break;
	    }
	    if(!found){
	      if(debug(3)){
		cerr << "right attach filter: ";
		item1->dump(cerr); item2->dump(cerr);
		cerr << endl;
	      }
	      return;
	    }
	  }
	}
	else if(inDSV2->empty(1-m_constrainedDim) and
		!inDSV1->empty(m_constrainedDim) ){
	  bool found = false;
	  for(unsigned int i = 0; i < inSpan2.size(); i++){
	    if((inSpan2[i]).second == (inSpan1[0].first))
	      found = true;
	    break;
	  }
	  if(!found){
	    if(debug(3)){
	      cerr << "right attach filter2: ";
	      item1->dump(cerr); item2->dump(cerr);
	      cerr << endl;
	    }
	    return;
	  }
	}
      }
    }
    ///////  END EPSILON FILTER ///////////

    //calculate the heir vector for this situation
    outHRV = findHeirRoleVector(iPT1,iPT2);
    if(oneDimEmpty){
      if(outHRV.getRole(0) == 0)
	outHRV.setRole(0,1);
    }
    //calculate the label
    outLabel = findHeirNTLink(lsnIt, outHRV, label1, label2);
    
    if(oneDimEmpty){
      HeirRoleVector hrv = outHRV.invert();
      outLabel2 = findHeirNTLink(lsnIt, hrv, label1, label2);
    }
  }

  // If the input DSpans are empty on the contrained component, and
  // this grammar is being used as a supervising grammar for learning,
  // then we want to update the item's hiddenDSVs structure.  If there
  // are two antecedent hiddenDSVs and their sum is incompatible with
  // the input PAV on the constrained component, then the antecedents
  // cannot compose, so we abort.
  Vector<DSV>* pOutHDSVs = NULL;
  if (inSpan1.empty() and inSpan2.empty()) {
    PrecArray inPA = arrg.second.getPA(m_constrainedDim);
    bool compatible = hDSV::crossProdWithPA(nrhedge1->getHiddenDSVs(), 
					    nrhedge2->getHiddenDSVs(), 
					    m_constrainedDim, inPA, pOutHDSVs);
    if (not compatible)
      return; // without returning any possConseqs
  }

  //vector with one position tree pointer in it
  vector<PositionTree::iterator> vecIt(1, lsnIt);
  
  //if the grammar is Headed, we set the terminal link based 
  //on the heirRoleVecor, and send it back
  if(!oneDimEmpty)
    outLabel.delex();
  
  
  vector<short> outHeadPosn(1);
  if (outHRV.getRole(0) == 1)
    outHeadPosn[0] = inHeadPosn1[m_constrainedDim];
  else
    outHeadPosn[0] = inHeadPosn2[m_constrainedDim];

  //construct hedge item
  Item* outItem= new NodeRefHedge(outLabel, outDSV, outHeadPosn, vecIt);

  //set the child counters for the new node
  if(vecIt[0].node != NULL){
    vector<short> counters(add_uneven_vec(nrhedge1->getChildCounter(), nrhedge2->getChildCounter()));    
    //if we have not completed this node yet, i.e. if we are still waiting for
    // other children, sum the counters.
    if((unsigned int)counters[m_constrainedDim] != vecIt[0].number_of_children()){
      ((NodeRefHedge*)outItem)->setChildCounter(vector<short>(1,counters[m_constrainedDim]));
    }else{
      // else, we have finished the node.  Reset the counters to 1.
      vector<short> counters(1,1);
      ((NodeRefHedge*)outItem)->setChildCounter(counters);
    }
  }else{
    //node is NULL, so counter set to 0.
    vector<short> counters(1,0);
    ((NodeRefHedge*)outItem)->setChildCounter(counters);
  }

  if (pOutHDSVs)
    outItem->setHiddenDSVs(pOutHDSVs);
  // compute fan-out cost
  FanoutVect fov = outItem->getFanoutV();
  char totalFanOut = accumulate(fov.begin(), fov.end(), 0);
  // additional cost of 1.0 for every gap
  SCORE fanOutCost = totalFanOut - outItem->numActiveDims();
  assert(fanOutCost >= 0.0);
  //! score it.
  outItem->setModelScore(fanOutCost);
  if(debug(4)){
    bool tmp = GlobalVocab::s_noVocabsOut;
    GlobalVocab::s_noVocabsOut = false;
    cerr << "TCMTG:getPossibleCons returning: outHRV: " << outHRV
	 << " and item: " << *((NodeRefHedge*)outItem);
    cerr << endl;
    GlobalVocab::s_noVocabsOut = tmp;
  }

  //insert item and return
  possConseqs.push_back(outItem);
  
  //now repeat for other option (when not headed)
  if(noEmptyIts && oneDimEmpty){
    if (outHRV.invert().getRole(0) == 1)
      outHeadPosn[0] = inHeadPosn1[m_constrainedDim];
    else
      outHeadPosn[0] = inHeadPosn2[m_constrainedDim];
    
    outItem = new NodeRefHedge(outLabel2, outDSV, outHeadPosn, vecIt);

    //set the child counters for the new node
    if(vecIt[0].node != NULL){
      vector<short> counters(add_uneven_vec(nrhedge1->getChildCounter(), nrhedge2->getChildCounter()));    
      //if we have not completed this node yet, i.e. if we are still waiting for
      // other children, sum the counters.
      if((unsigned int)counters[m_constrainedDim] != vecIt[0].number_of_children()){
	((NodeRefHedge*)outItem)->setChildCounter(vector<short>(1,counters[m_constrainedDim]));
      }else{
	// else, we have finished the node.  Reset the counters to 1.
	vector<short> counters(1,1);
	((NodeRefHedge*)outItem)->setChildCounter(counters);
      }
    }else{
      //node is NULL, so counter set to 0.
      vector<short> counters(1,0);
      ((NodeRefHedge*)outItem)->setChildCounter(counters);
    }
    if (pOutHDSVs) {
      // copy it, otherwise ownership is ambiguous
      Vector<DSV>* pOutHDSVs2 = new Vector<DSV>(*pOutHDSVs);
      outItem->setHiddenDSVs(pOutHDSVs2);
    }
    FanoutVect fov = outItem->getFanoutV();
    char totalFanOut = accumulate(fov.begin(), fov.end(), 0);
    SCORE fanOutCost = totalFanOut - outItem->numActiveDims();
    assert(fanOutCost >= 0.0);
    outItem->setModelScore(fanOutCost);
    if(debug(4)){
      bool tmp = GlobalVocab::s_noVocabsOut;
      GlobalVocab::s_noVocabsOut = false;
      cerr << "TCMTG:getPossibleCons2 returning: outHRV: " << outHRV
	   << " and item: " << *((NodeRefHedge*)outItem);
      cerr << endl;
      GlobalVocab::s_noVocabsOut = tmp;
    }
    
    possConseqs.push_back(outItem);
  }
}

void TreeConstrainedMTG::
getPossibleConsHeaded(const Item* item1, const Item* item2, const Pair<bool,PAV>& arrg,
		      vector<HRVIpair>& possConseqs) {

  //Will contain NTlink for consequent to be returned.
  NTLink outLabel(1), outLabel2(1);
  //Will contain HeirVector for consequent to be returned.
  HeirRoleVector outHRV(1), outHRV2(1);
  //Will contain DSV for consequent to be returned.
  DSV outDSV(1);
  vector<short> outHeadPosn(1), outHeadPosn2(1);

  //True if the span is empty on one component in both labels.
  bool oneDimEmpty = false;
  bool needBoth = true;

  if(debug(5)){
    bool tmp = GlobalVocab::s_noVocabsOut;
    GlobalVocab::s_noVocabsOut = false;
    cerr << "TreeConstrainedMTG::getPossibleConsHeaded QUERY: " << endl
	 << *((NodeRefHedge*)item1) << endl
	 << " AND " << *((NodeRefHedge*)item2)<< endl;
    GlobalVocab::s_noVocabsOut = tmp;
  }

  // Get information from item1
  const NodeRefHedge * nrhedge1 = static_cast<const NodeRefHedge*>(item1);
  const DSV * inDSV1 = item1->getDSV();
  const DSpan& inSpan1 = (*inDSV1)[m_constrainedDim];
  PositionTree::iterator iPT1 = nrhedge1->getNode()[m_constrainedDim];
  NTLink label1 = nrhedge1->getNTLink();
  vector<short> inHeadPosn1 = nrhedge1->getHeads();


  // Get information from item2
  const NodeRefHedge * nrhedge2 = static_cast<const NodeRefHedge*>(item2);
  const DSV * inDSV2 = nrhedge2->getDSV();
  const DSpan& inSpan2 = (*inDSV2)[m_constrainedDim];
  PositionTree::iterator iPT2 = nrhedge2->getNode()[m_constrainedDim];
  NTLink label2 = nrhedge2->getNTLink();
  vector<short> inHeadPosn2 = nrhedge2->getHeads();

  // If the input DSpans are empty on the contrained component, and
  // this grammar is being used as a supervising grammar for learning,
  // then we want to update the item's hiddenDSVs structure.  If there
  // are two antecedent hiddenDSVs and their sum is incompatible with
  // the input PAV on the constrained component, then the antecedents
  // cannot compose, so we abort.
  Vector<DSV>* pOutHDSVs = NULL;
  if (inSpan1.empty() and inSpan2.empty()) {
    PrecArray inPA = arrg.second.getPA(m_constrainedDim);
    bool compatible = hDSV::crossProdWithPA(nrhedge1->getHiddenDSVs(), 
					    nrhedge2->getHiddenDSVs(), 
					    m_constrainedDim, inPA, pOutHDSVs);
    if (not compatible)
      return; // without returning any possConseqs
  }

  //Get DSV of consequent.
  DSV combinedDSV(*inDSV1 + *inDSV2);
  if(combinedDSV.size() < 2
     || combinedDSV[0].size() == 0 
     || combinedDSV[1].size() == 0){
    oneDimEmpty = true;
  }

  //if we are not ignoring heads, then there may be a -1 returned in the heir role, so we
  // dont needBoth head options.
  //also, if one span is empty, there is only one option for the head. 
  if (!m_ignoreHeads or inSpan1.empty() or inSpan2.empty())
    needBoth = false;

    
  if (combinedDSV.size() > m_constrainedDim)
    outDSV[0] = combinedDSV[m_constrainedDim];

  //get lowest subsuming node
  PositionTree::iterator lsnIt;

  //If both position tree pointers are null, we assume they can combine to a NULL.
  //(independent rewriting)
  if(iPT1 == 0 && iPT2 == 0){
    //vector with one position tree pointer in it
    vector<PositionTree::iterator> vecIt(1, iPT1);
    outLabel.setNT(0,label1.getNT(m_constrainedDim)->clone());
    
    //if one is null in one dimension, then don't climb the position tree
  } else if(iPT1 == 0 && iPT2 != 0) {
    lsnIt = iPT2;
    outHRV.setRole(0,2);
    outLabel.setNT(0,BGET_POSITION_LABEL(*iPT2).getNT(0)->clone());
    if(dynamic_cast<const LexNT*>(label2.getNT(m_constrainedDim))){
      TLink tl(item1->arity());
      tl[0] = label2.getTLink()[m_constrainedDim];
      outLabel.setTLink(tl);
    }
    //if iPT2 is the inactive pointer
  } else if (iPT1 != 0 && iPT2 == 0) {
    lsnIt = iPT1;
    outHRV.setRole(0,1);
    outLabel.setNT(0,BGET_POSITION_LABEL(*iPT1).getNT(0)->clone());
    if(dynamic_cast<const LexNT*>(label1.getNT(m_constrainedDim))){
      TLink tl(1);
      tl[0] = label1.getTLink()[m_constrainedDim];
      outLabel.setTLink(tl);
    }
    //lastly, if neither position tree pointer is 0
  } else {
    //this is true if it1 points to a node in the position tree that 
    //is completely covered by the item. If so, we can progress up the tree.
    //if not, we have to wait until the other children are covered
    bool complete1 = true;
    if(nrhedge1->getChildCounter()[m_constrainedDim] > 1){
      complete1 = false;
    }
      
    bool complete2 = true;
    if(nrhedge2->getChildCounter()[m_constrainedDim] > 1){
      complete2 = false;
    }

    //Find the lowest subsuming node of the pair of the input iterators
    lsnIt = computeTreeIterator(iPT1, iPT2, complete1, complete2);

    //if the two don't share a subsuming node, exit
    if(lsnIt == NULL){
      if(debug(5))cerr << "LSN NOT FOUND " << endl;
      return;
    }
    
    ///// BEGIN EPSILON FILTER
    //filter out 1D attachments in one direction
    if(m_rightAttach){
      //we can only filter this if the word to the right of the
      //epsilon is in the same lowest subsuming node.
      if((BGET_SPAN(*lsnIt)[0][0].first > inSpan1[0].second) && 
	 (BGET_SPAN(*lsnIt)[0][0].first > inSpan2[0].second)){
	if(inDSV1->empty(1-m_constrainedDim) and 
	   !inDSV2->empty(m_constrainedDim) ){
	  bool found = false;
	  for(unsigned int i = 0; i < inSpan1.size(); i++){
	    if(inSpan1[i].second == (inSpan2[0].first)) {
	      found = true;
	      break;
	    }
	    if(!found){
	      if(debug(3)){
		cerr << "right attach filter: ";
		item1->dump(cerr); item2->dump(cerr);
		cerr << endl;
	      }
	      return;
	    }
	  }
	}
	else if(inDSV2->empty(1-m_constrainedDim) and
		!inDSV1->empty(m_constrainedDim) ){
	  bool found = false;
	  for(unsigned int i = 0; i < inSpan2.size(); i++){
	    if((inSpan2[i]).second == (inSpan1[0].first))
	      found = true;
	    break;
	  }
	  if(!found){
	    if(debug(3)){
	      cerr << "right attach filter2: ";
	      item1->dump(cerr); item2->dump(cerr);
	      cerr << endl;
	    }
	    return;
	  }
	}
      }
    }
    ///////  END EPSILON FILTER ///////////


    //calculate the heir vector for this situation
    outHRV = findHeirRoleVector(iPT1,iPT2);
    if(oneDimEmpty){
      if(outHRV.getRole(0) == 0)
	outHRV.setRole(0,1);
    }

    //calculate the label
    outLabel = findHeirNTLink(lsnIt, outHRV, label1, label2);
    
    //set up item 2

    outHRV2 = outHRV.invert();
    if(oneDimEmpty){
      if(outHRV2.getRole(0) == 0)
	outHRV2.setRole(0,1);
    }
    outLabel2 = findHeirNTLink(lsnIt, outHRV2, label1, label2);
  }
  
  if (outHRV.getRole(0) == 1)
    outHeadPosn[0] = inHeadPosn1[m_constrainedDim];
  else
    outHeadPosn[0] = inHeadPosn2[m_constrainedDim];
  if (outHRV2.getRole(0) == 1)
    outHeadPosn2[0] = inHeadPosn1[m_constrainedDim];
  else
    outHeadPosn2[0] = inHeadPosn2[m_constrainedDim];
  

  //vector with one position tree pointer in it
  vector<PositionTree::iterator> vecIt(1, lsnIt);

  //
  // set up first item
  //


  //construct hedge item
  Item* outItem= new NodeRefHedge(outLabel, outDSV, outHeadPosn, vecIt);

  //set the child counters for the new node
  if(vecIt[0].node != NULL){
    vector<short> counters(add_uneven_vec(nrhedge1->getChildCounter(), nrhedge2->getChildCounter()));    
    //if we have not completed this node yet, i.e. if we are still waiting for
    // other children, sum the counters.
    if((unsigned int)counters[m_constrainedDim] != vecIt[0].number_of_children()){
      ((NodeRefHedge*)outItem)->setChildCounter(vector<short>(1,counters[m_constrainedDim]));
    }else{
      // else, we have finished the node.  Reset the counters to 1.
      vector<short> counters(1,1);
      ((NodeRefHedge*)outItem)->setChildCounter(counters);
    }
  }else{
    //node is NULL, so counter set to 0.
    vector<short> counters(1,0);
    ((NodeRefHedge*)outItem)->setChildCounter(counters);
  }
  
  if (pOutHDSVs)
    outItem->setHiddenDSVs(pOutHDSVs);
  // compute fan-out cost
  FanoutVect fov = outItem->getFanoutV();
  char totalFanOut = accumulate(fov.begin(), fov.end(), 0);
  // additional cost of 1.0 for every gap
  SCORE fanOutCost = totalFanOut - outItem->numActiveDims();
  assert(fanOutCost >= 0.0);
  //! score it.
  outItem->setModelScore(fanOutCost);
  if(debug(4)){
    bool tmp = GlobalVocab::s_noVocabsOut;
    GlobalVocab::s_noVocabsOut = false;
    cerr << "TCMTG:getPossibleCons returning: outHRV: " << outHRV
	 << " and item: " << *((NodeRefHedge*)outItem);
    cerr << endl;
    GlobalVocab::s_noVocabsOut = tmp;
  }

  //insert item and return
  possConseqs.push_back(pair<HeirRoleVector,Item*>(outHRV,outItem));


  //
  // set up second item
  //
  if(needBoth){
    
    //construct hedge item
    outItem= new NodeRefHedge(outLabel2, outDSV, outHeadPosn2, vecIt);
    
    //set the child counters for the new node
    if(vecIt[0].node != NULL){
      vector<short> counters(add_uneven_vec(nrhedge1->getChildCounter(), nrhedge2->getChildCounter()));    
      //if we have not completed this node yet, i.e. if we are still waiting for
      // other children, sum the counters.
      if((unsigned int)counters[m_constrainedDim] != vecIt[0].number_of_children()){
	((NodeRefHedge*)outItem)->setChildCounter(vector<short>(1,counters[m_constrainedDim]));
      }else{
	// else, we have finished the node.  Reset the counters to 1.
	vector<short> counters(1,1);
	((NodeRefHedge*)outItem)->setChildCounter(counters);
      }
    }else{
      //node is NULL, so counter set to 0.
      vector<short> counters(1,0);
      ((NodeRefHedge*)outItem)->setChildCounter(counters);
    }
    
    if (pOutHDSVs)
      outItem->setHiddenDSVs(pOutHDSVs);
    // compute fan-out cost
    fov = outItem->getFanoutV();
    totalFanOut = accumulate(fov.begin(), fov.end(), 0);
    // additional cost of 1.0 for every gap
    fanOutCost = totalFanOut - outItem->numActiveDims();
    assert(fanOutCost >= 0.0);
    //! score it.
    outItem->setModelScore(fanOutCost);
    if(debug(4)){
      bool tmp = GlobalVocab::s_noVocabsOut;
      GlobalVocab::s_noVocabsOut = false;
      cerr << "TCMTG:getPossibleCons2 returning: outHRV: " << outHRV2
	   << " and item: " << *((NodeRefHedge*)outItem);
      cerr << endl;
      GlobalVocab::s_noVocabsOut = tmp;
    }
    
    //insert item and return
    possConseqs.push_back(pair<HeirRoleVector,Item*>(outHRV2,outItem));
  }
  
}

//! Returns a pointer to the lowest subsuming node of iPT1 and iPT2 in the position tree.
//! if none is found, the bool found is set to false.
PositionTree::iterator
TreeConstrainedMTG::computeTreeIterator(PositionTree::iterator& iPT1, PositionTree::iterator& iPT2,
					bool complete1, bool complete2){
  //pointers to mark parents
  PositionTree::iterator iPT1Par = iPT1;
  PositionTree::iterator iPT2Par = iPT2;
  
  //pointers to be returned
  PositionTree::iterator iPT1ret= iPT1;
  PositionTree::iterator iPT2ret= iPT2;

  PositionTree atree;

  if(iPT1 != 0 && iPT2 != 0){
    //if we are ready to progress up the tree (all children are covered)
    if(complete1){
      //while we have a unary production, move up again to find LSN
      iPT1Par = atree.parent(iPT1Par);
      while(iPT1Par.node != NULL && iPT1Par.number_of_children() == 1){
	if(atree.parent(iPT1Par).node != NULL){
	  iPT1ret = iPT1Par;
	  iPT1Par = atree.parent(iPT1Par);
	}
	else break;
      }
    }
    //if we are ready to progress up the tree (all children are covered)
    if(complete2){
      iPT2Par = atree.parent(iPT2Par);
      while(iPT2Par.node != NULL && iPT2Par.number_of_children() == 1){
	if(atree.parent(iPT2Par).node != NULL){
	  iPT2ret = iPT2Par;
	  iPT2Par = atree.parent(iPT2Par);
	}
	else break;
      }
    }
    //now we have the parent and child pointers.
    if(debug(5)){
      bool tmp = GlobalVocab::s_noVocabsOut;
      GlobalVocab::s_noVocabsOut = false;
      if(iPT1 != 0)
	cerr << "iPT1: " << BGET_POSITION_LABEL(*iPT1) << endl;
      if(iPT2 != 0)
	cerr << "iPT2: " << BGET_POSITION_LABEL(*iPT2) << endl;
      GlobalVocab::s_noVocabsOut = tmp;
    }
    //if they share a pointer, this is a typical sibling case.
    if(iPT1Par == iPT2Par){
      iPT1 = iPT1ret;
      iPT2 = iPT2ret;
      return iPT1Par;
    }else if (iPT1Par == iPT2){ //else one covers the other
      //we dont want iPT2ret because the iPT2 pointer stays put.
      iPT1 = iPT1ret; 
      return iPT1Par;
    }else if( iPT2Par == iPT1){
      //we dont want iPT1ret because the iPT1 pointer stays put.
      iPT2 = iPT2ret;
      return iPT2Par;
    } 
  } else if(iPT1 == 0 && iPT2 != 0) { //else one is NULL.  (Then it just stays put)
    return iPT2;
  } else if (iPT1 != 0 && iPT2 == 0) {
    return iPT1;
  }
  //if you reach here, the two nodes do not share a common parent.
  return NULL;
}

//! given iPT1 and iPT2, gets heir vector. If none is found, backs off
//! to iPT1other and iPT2other.
HeirRoleVector TreeConstrainedMTG::findHeirRoleVector(PositionTree::iterator& iPT1, 
						      PositionTree::iterator& iPT2){
  PositionTree atree;
  HeirRoleVector ret(1, 0);

  if((iPT1 == 0 && iPT2 != 0) || (iPT1 != 0 && iPT2 == 0)){
    if (iPT1 == 0)
      ret.setRole(0,2);
    else
      ret.setRole(0,1);
    return ret;
  }

  // if one item is the parent of the other in the position tree, then
  // the "parent" is going to be the heir for sure. This is sort of
  // tricky, but true in our algorithm.
  if( iPT2 == atree.parent(iPT1)){
    ret.setRole(0,2);
    return ret;
  }

  if( iPT1 == atree.parent(iPT2)){
    ret.setRole(0,1);
    return ret;
  }

  queue<short> q1 = atree.getHeirQueue(iPT1);
  queue<short> q2 = atree.getHeirQueue(iPT2);

  short s;
  while(!q1.empty()){
    s = ((short)q1.front());
    if(s == 0){
      //this item is heir!
      ret.setRole(0,1);
    }
    q1.pop();
  }

  while(!q2.empty()){
    s = ((short)q2.front());
    if(s == 0){
      ret.setRole(0,2);
    }
    q2.pop();
  }

  return ret;
}

//!gets NTLink for this situation.
NTLink TreeConstrainedMTG::findHeirNTLink(PositionTree::iterator& parent, HeirRoleVector &hv, 
					  NTLink& label1, NTLink& label2){
  NTLink ntl(1);
  TLink tl(1);

  //set Nonterminal
  ntl.setNT(0,BGET_POSITION_LABEL(*parent).getNT(0)->clone());

  //only do this if grammar is headed
  //  if(m_isHeaded){
  //t terminal (if known)
  if(dynamic_cast<const LexNT*>(label1.getNT(0)) && dynamic_cast<const LexNT*>(label2.getNT(0))){
    if(hv.getRole(0) == 1){
      tl[0] = label1.getTLink()[m_constrainedDim];
    }
    else if (hv.getRole(0) == 2){
      tl[0] =  label2.getTLink()[m_constrainedDim];
    }
    return NTLink(ntl,tl);
  }else{
    ntl.delex();
    return ntl;
  }
}

/// Returns all the possible lexicalized pre-terminals for a terminal
/// link covering the input span.  The tree iterators in the returned
/// set point to the leaves in the restricting tree. Note that the
/// returned items are *one-dimensional*; they are combined with items
/// from other dimensions by the grammar that contains this one.
void
TreeConstrainedMTG::getPossibleConsForTLink(const TLink& inTL, const DSV& inDSV, vector<Item*>& vItem) const {
  if(debug(5)){
    bool tmp = GlobalVocab::s_noVocabsOut;
    GlobalVocab::s_noVocabsOut = false;
    cerr <<"entering TreeConstrainedMTG::getPossibleConsForTlink" << endl;
    cerr <<"TLink= " << inTL << "; DSV= " << inDSV << endl;
    GlobalVocab::s_noVocabsOut = tmp;
  }

  NTLink outNTL(1);
  vector<PositionTree::iterator> viPT(1);
  
  // figure out the head position
  arity_t inSize = inDSV.size();
  DSpan inSpan;
  if (inSize > m_constrainedDim)
    inSpan = inDSV[m_constrainedDim];
  
  Terminal theTerminal = inTL[m_constrainedDim];
  if (theTerminal == NULLT) {
    PositionTree::iterator iPT = 0;
    viPT[0] = iPT;
    outNTL.setNT(0,new Nonterminal(0));

    TLink t(1, theTerminal);
    NTLink outLNTL(outNTL,t);
    DSV theDSV(1);

    vector<short> outHeadPosn(1);
    if (inSize > m_constrainedDim and not inSpan.empty())
      outHeadPosn[0] = inSpan[0].second;
    else
      outHeadPosn[0] = 0;
    
    Item* nrh = new NodeRefHedge(outLNTL, theDSV, outHeadPosn, viPT);
   
    if(debug(5)){
      bool tmp = GlobalVocab::s_noVocabsOut;
      GlobalVocab::s_noVocabsOut = false;
      cerr << "TCMTG adding NodeRefHedge: " << *((NodeRefHedge*)nrh) << endl;
      GlobalVocab::s_noVocabsOut = tmp;
    }
 
    vItem.push_back(nrh);
    return;
  }

  // if we're here, then the inTL is not null
  vector<DSpan> sameDimSpans; // possible spans in the *same* component

  // If the inDSV is empty in the active component, then this is a
  // Load.  This situation can arise if the TCMTG is being used as a
  // supervising grammar for learning, in which case we still want to
  // link the item to the constraining tree for efficiency.  We also
  // want to populate the item's hiddenDSVs structure.  So we find all
  // possible spans for it in the m_pCurrSent and create an item for
  // each possible span.
  if (inSize > m_constrainedDim and not inSpan.empty()) {
    DSpan inSpan = inDSV[m_constrainedDim];
    sameDimSpans.push_back(inSpan);
  } else
    for(unsigned short osi = 0; osi < m_pCurrSent->size(); ++osi){
      if ( (*m_pCurrSent)[osi] == theTerminal) {
	DSpan aSpan = DSpan(DSpan::Span(osi, osi+1));
	sameDimSpans.push_back(aSpan);
      }
    }

  while (! sameDimSpans.empty()) {
    DSpan relevantSpan = sameDimSpans.back();
    sameDimSpans.pop_back();

    hash_map<DSpan, set<PositionTree::iterator> > :: const_iterator ii
      = m_span2leaf.find(relevantSpan);
    if(ii != m_span2leaf.end()){
      set<PositionTree::iterator> :: const_iterator k;
      // this loop works, but the current code will never use more
      // than one iteration
      for(k = ii->second.begin(); k != ii->second.end(); ++k){
	PositionTree::iterator i = *k;
	outNTL.setNT(0,(BGET_POSITION_LABEL(*i)).getNT(0)->clone());
	viPT[0] = i;
      }
      //span not found
    } else {
      PositionTree::iterator iPT = 0;
      viPT[0] = iPT;
      outNTL.setNT(0,new Nonterminal(0));
    }
    TLink outTL( 1, theTerminal);
    NTLink outLNTL(outNTL, outTL);
    DSV outDSV(1);
    if (inSize > m_constrainedDim and not inSpan.empty())
      outDSV[0] = relevantSpan;
    vector<short> outHeadPosn(1);
    outHeadPosn[0] = relevantSpan[0].second;
    Item* nrh = new NodeRefHedge(outLNTL, outDSV, outHeadPosn, viPT);

    if (inSize <= m_constrainedDim || inSpan.empty()) {
      // set the hiddenDSV
      DSV hiddenDSV(m_arity);
      hiddenDSV[m_constrainedDim] = relevantSpan;
      Vector<DSV>* pDSVV = new Vector<DSV>(1, hiddenDSV);
      // transfer ownership to the item
      nrh->setHiddenDSVs(pDSVV);
    }
    if(debug(5)){
      // Dereferencing the vocabs won't work cuz the noderefhedge here
      // is always 1D, even if it pertains to the 2nd dimension.  In
      // that case, NTLink::put has no way of knowing *which* vocab to
      // deref.

      bool tmp = GlobalVocab::s_noVocabsOut;
      GlobalVocab::s_noVocabsOut = false;
      cerr << "TCMTG::getPossConsForT adding NodeRefHedge: " << *((NodeRefHedge*)nrh) << endl;
      GlobalVocab::s_noVocabsOut = tmp;
    }
    vItem.push_back(nrh);
  }
  return;
}


/** Adds this class' option description to the option description list of the hierarchy */
BSP<BPO::options_description> 
TreeConstrainedMTG::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  //optionsDescription->add_options()
  // ("SomeBottomUpLogicOption", BPO::value<string>()->default_value("my default"), "test option")
  // ;
  optionsDescription->add_options()
    ("TreeStream", BPO::value<string>(), "stream of constraining trees")
    ;

  return optionsDescription;
}
