#ifndef _TreeConstrainedMTG_H_
#define _TreeConstrainedMTG_H_

#include "File.hpp"
#include "DSV.hpp"
#include "StringVocabulary.hpp"
#include "MultiMap.hpp"
#include "MoreAlgorithms.hpp"
#include "PartialWMTG.hpp"
#include "PositionTree.hpp"
#include "HeirEvent.hpp"

/// a WMTG that accepts/generates inferences according to a constraining tree
class TreeConstrainedMTG: public PartialWMTG {
  typedef PartialWMTG _super;
  typedef Grammar _grammar;

  BSP<BPO::options_description> 
  addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription);

public:

  /**
   * Constructor.
   * @param configFile The config file of the grammar.
   */
  TreeConstrainedMTG(arity_t d, const string& configFile,
		     BSP<BPO::options_description> priorOptions =
		     BSP<BPO::options_description>(new BPO::options_description())
		     );

  virtual ~TreeConstrainedMTG();

  // prepare for another pass through the constraining treebank
  void clear();

  /**
   * Updates the content of the grammar.
   * @param tuple The sentence tuple.
   * @return If there is an empty parse read, then return false.
   */
  virtual bool prepareFor(SentenceTuple& tuple);

  virtual void filterArrangementFor(BSP<set<Pair<bool,PAV> > >& arrgs, const Item* trigger) const;

  //! Returns all Z's given X and Y such that Z->XY (or Z->YX).
  //! \param heir : the heir constituent.
  //! \param depend : the dependent constituent.
  //! \param conss : the parent constituents that will be returned.
  void getPossibleCons(const Item* heir,
		       const Item* depend,
		       const Pair<bool,PAV>& pav,
		       vector< Item*>& conss);

  //! Returns all Z's given X and Y such that Z->XY (or Z->YX).
  //! \param heir : the heir constituent.
  //! \param depend : the dependent constituent.
  //! \param conss : the parent constituents and HeirRoleVector that will be returned.
  void getPossibleConsHeaded(const Item* heir,
			     const Item* depend,
			     const Pair<bool,PAV>& pav,
			     vector<HRVIpair>& conss);
  /*
   * Input/Output.
   */
  istream& get(istream&) {
    THROW("TreeConstrainedMTG::get not implemented");
  }

  ostream& put(ostream&) const{
    THROW("TreeConstrainedMTG::put not implemented");
  }

  virtual void getPossibleConsForTLink(const TLink& t, const DSV& d, vector<Item*>& vItem) const;

  //! Returns the unsmoothed log prob of event.
  SCORE getUnSmoothedProb(const TLink&) const;

private: // methods

  //! return an HeirRoleVector that matches the inforation in it1 and
  //! it2. If none is found, back off to it1other and it2other.
  HeirRoleVector findHeirRoleVector(PositionTree::iterator& it1, 
				    PositionTree::iterator& it2);

  //! finds the heirNTLink. Used to get NTLink for a parent item.
  NTLink findHeirNTLink( PositionTree::iterator& parent, 
			 HeirRoleVector &hv, NTLink&, NTLink&);

  //! used to find the LSN of the two iterators. If none is found,
  //! Null iterator is returned.  complete1 and complete2 tell if 
  //! the it1 and it2 pointers are ready to traverse up the tree.
  PositionTree::iterator computeTreeIterator(PositionTree::iterator& it1, 
						 PositionTree::iterator& it2,
						 bool complete1,bool complete2);

protected: // data

  /// Index to the spans of tree leaves. Constructed in the
  /// prepareFor(). The value type of the hash_map is a set for obsolete
  /// reasons.  However, it might be useful in future if there is
  /// more than one constraining tree for a given component.
  hash_map<DSpan, set<PositionTree::iterator> > m_span2leaf;

  //!1d tree
  PositionTree* m_pRestrictingTree;

  //! stream of 1d trees
  iFile* m_pMonoTBStream;

};

#endif
