#ifndef _UpdateRule_H_
#define _UpdateRule_H_

#include "Weighted.hpp"
#include "SPVector.hpp"
#include "Production.hpp"
#include "HeadMTree.hpp"
#include "Parser.hpp"

class LearnableGrammar;

//! Defines an abstract kernel for comparing two parses.
class UpdateRule : public Dbug {
public:

  UpdateRule(bool keepW2W = true) : m_keepW2W(keepW2W) {};

  virtual ~UpdateRule() { };

  /// update the model and return the cost or margin
  SCORE getGradient(AbstractParser* supervisor, Parser* learner, 
		    SentenceTuple& sT, LearnableGrammar* gradient,
		    ostream& treeStream) const;

  /// compute the gradient for each feature -- here is the default for MLE
  virtual void getFeatureGradients( size_t supervisorMarginSize, size_t learnerMarginSize, 
				    size_t intersectionSize, 
				    SCORE& supervisorGradient, SCORE& learnerGradient, 
				    SCORE& intersectionGradient, 
				    SCORE& unseenFeatGradient) const {
    supervisorGradient = 0.0;
    learnerGradient = 1.0;
    intersectionGradient = 0.0;
    unseenFeatGradient = 1.0;
  }

protected: // methods
  virtual SCORE getGradientHelper(const WTreeVec& learnerParses, 
				  const WTreeVec& supervisoryParses,
				  vector<HeadMTree::iterator>& supervisorMargin, 
				  vector<HeadMTree::iterator>& learnerMargin, 
				  vector<HeadMTree::iterator>& vecIntersection) const;


  /// \param supervisorMargin will contain all the supervisorFeats
  /// that are not learnerFeats \param learnerMargin will contain the
  /// converse \param supervisorFeats and \param learnerFeats cannot
  /// be const because they will be sorted
  template<class FType>
  void computeFeatureMargins(SPVector<FType>& supervisorFeats, 
			     SPVector<FType>& learnerFeats,
			     SPVector<FType>& supervisorMargin, 
			     SPVector<FType>& learnerMargin) const {

    // short-circuit for EM
    if (supervisorFeats.empty())
      learnerMargin = learnerFeats;
    // short-circuit for fresh learner
    else if (learnerFeats.empty())
      supervisorMargin = supervisorFeats;
    else {

      // compute the differences in both directions
      MA::unsorted_set_2sided_difference(supervisorFeats.begin(), supervisorFeats.end(),
					 learnerFeats.begin(), learnerFeats.end(),
					 back_inserter(supervisorMargin), 
					 back_inserter(learnerMargin),
					 MA::less_sptr<FType>());
    }

    size_t supervisorFeatVecSize = supervisorFeats.size();
    size_t learnerFeatVecSize = learnerFeats.size();
    size_t supervisorMarginSize = supervisorMargin.size();
    size_t learnerMarginSize = learnerMargin.size();
    size_t intersectionSize = (supervisorFeatVecSize + learnerFeatVecSize 
			       - supervisorMarginSize - learnerMarginSize ) / 2;
    if (debug(3)) {
      cerr << "Supervisor feats: " << endl;
      supervisorFeats.put(cerr, "\n");
      cerr << "Learner feats: " << endl;
      learnerFeats.put(cerr, "\n");
    }

    if (debug(1))
      cerr << "Sizes: intersection " << intersectionSize 
	   << " ; learner margin " << learnerMarginSize 
	   << " ; supervisor margin " <<  supervisorMarginSize  << endl;
  }

protected: // data
  //! if set to false, we do not return w2wNodes that are roots as mistakes
  bool m_keepW2W;

};

#endif
