#include "WMTG.hpp"
#include "File.hpp"
#include "HeadCompose.hpp"
#include "HeadProduction.hpp"
#include "Scan.hpp"
#include "LabelingStrategy.hpp"

//! constructors
WMTG::WMTG(const arity_t arity,
	   bool inGCNF,
	   bool isHeaded)
  : _super(arity), m_fanoutLimit(arity),
    m_rootLab(defaultRootLabel()), m_inGCNF(inGCNF), 
    m_isHeaded(isHeaded), m_smoothingCost(0.0), m_smoothScans(false),
    m_ignoreHeads(false), m_costAggregator(SUM)
{}

//! constructor
WMTG::
WMTG(arity_t dm,
     const string& configFile,
     BSP<BPO::options_description> optionsDescription)
  : _super(dm), m_fanoutLimit(dm), m_rootLab(defaultRootLabel()), 
    m_smoothScans(false), m_ignoreHeads(false),m_costAggregator(SUM) { 

  addOptionDescriptionsTo(optionsDescription);

  string allowedPAVFile, smoothingPAVFile;
  string optCtlNTVcbFilesOut;
  unsigned int debugLevel = 0;
  /*
   * Compile the config file.
   */
  BPO::variables_map vm;
  if(configFile != ""){
    iFile ifs(configFile.c_str());
    store(parse_config_file(ifs, *optionsDescription), vm);
  }
  notify(vm);

  if(vm.count("D")) m_arity = (arity_t) vm["D"].as<unsigned int>();
  if(vm.count("FanOut")) m_fanoutLimit = vm["FanOut"].as<unsigned int>();
  if(vm.count("GCNF")) m_inGCNF = vm["GCNF"].as<bool>();
  if(vm.count("Headed")) m_isHeaded = vm["Headed"].as<bool>();
  if(vm.count("PAV")) allowedPAVFile = vm["PAV"].as<string>();
  if(vm.count("smoothPAV")) smoothingPAVFile = vm["smoothPAV"].as<string>();
  if(vm.count("smoothCost")) m_smoothingCost = vm["smoothCost"].as<float>();
  if(vm.count("oN")) optCtlNTVcbFilesOut = vm["oN"].as<string>();
  if(vm.count("debug")) debugLevel = vm["debug"].as<unsigned int>();
  if(vm.count("smoothScans")) m_smoothScans = vm["smoothScans"].as<bool>();
  if(vm.count("ignoreHeads")) m_ignoreHeads = vm["ignoreHeads"].as<bool>();
  if(vm.count("costAggregator")) {
    const string& costAggregator = vm["costAggregator"].as<string>();
    if (costAggregator == "SUM")
      m_costAggregator = SUM;
    else if (costAggregator == "MAX")
      m_costAggregator = MAX;
    else if (costAggregator == "MAXSUM")
      m_costAggregator = MAXSUM;
    else
      THROW("Unknown cost aggregator!");
  }

  if (debugLevel > 0)
    debugme(debugLevel);

  if(debug(1)) {
    cerr << "WMTG options:" << endl;
    cerr << OptionsPrinter(*optionsDescription, vm) << endl;
  }
  if (optCtlNTVcbFilesOut != "")
    // save it in a more stable place
    m_ctlNTVcbFilesOut = string(optCtlNTVcbFilesOut);

  if(dm != (size_t)m_arity)
    THROW(dm - m_arity <<" more NT set(s) need(s) to be specified." << endl);

  /*
   * Load allowed PAVs
   */
  if(allowedPAVFile != "") {
    Vocabulary<PAV> pavVcb;
    if(debug(2)){ cerr<<"Loading allowed PAVs ..."; }
    iFile pavF(allowedPAVFile);
    pavF >> pavVcb;
    pavF.close();
    insertPAV(pavVcb);
  }

  /*
   * Load smoothing PAVs
   */
  if(smoothingPAVFile != "") {
    if(debug(2)){ cerr<<"Loading smoothing PAVs ..."; }
    iFile pavF(smoothingPAVFile);
    pavF >> m_smoothingPAVs;
    pavF.close();
  }
}

//! destructor
WMTG::~WMTG() {}

//! assignment
WMTG& WMTG::operator=(const WMTG& other) {
  _super::operator=(other);
  m_rootLab = other.m_rootLab;
  m_fanoutLimit = other.m_fanoutLimit;
  m_inGCNF = other.m_inGCNF;
  m_isHeaded = other.m_isHeaded;

  return (*this);
}

void WMTG::insertPAV(const PAV& aPAV) {
  /*
   * Filter illegal PAVs.
   */
  if(aPAV.fanout() > fanoutLimit())
    return;
  FanoutVect fov1(aPAV.getFanoutV(1)); // 1 is role.
  FanoutVect fov2(aPAV.getFanoutV(2)); // 2 is role.
  if(fov1.overallFanout() > fanoutLimit() || fov2.overallFanout() > fanoutLimit()){
    return;
  }

  // add it into the pav vocab.
  m_pavVcb.addWord(aPAV);

  // Insert an Arrangement for fov of role 1.
  fov2arnmt_t::iterator f2ai;
  f2ai = m_fov2arnmt.find(fov1);
  if (f2ai == m_fov2arnmt.end()) {
    m_fov2arnmt.insert(make_pair(fov1, arnmtSet_t()));
    f2ai = m_fov2arnmt.find(fov1);
  }
  f2ai->second.insert(Pair<bool,PAV>(true, aPAV));

  // Insert an Arrangement for fov of role 2.
  f2ai = m_fov2arnmt.find(fov2);
  if (f2ai == m_fov2arnmt.end()) {
    m_fov2arnmt.insert(make_pair(fov2, arnmtSet_t()));
    f2ai = m_fov2arnmt.find(fov2);
  }
  f2ai->second.insert(Pair<bool,PAV>(false, aPAV));
}

//! records a vocab of PAVs
void WMTG::insertPAV(Vocabulary<PAV>& pavVcb) {
  Vocabulary<PAV>::const_iterator it_pav_vcb;
  for(it_pav_vcb = pavVcb.begin(); it_pav_vcb != pavVcb.end(); ++it_pav_vcb){
    insertPAV(it_pav_vcb->first);
  }
}

//! Returns the possible arrangements for an item
void WMTG:: getArrangementFor(BSP<set<Pair<bool,PAV> > >& arnmts, const Item* trigger) const {
  fov2arnmt_t::const_iterator f2ai = m_fov2arnmt.find(trigger->getFanoutV());
  if (f2ai == m_fov2arnmt.end()) {
    arnmts.reset(new set< Pair<bool,PAV> >);
  } else 
    arnmts.reset(new set<Pair<bool,PAV> >(f2ai->second));
}

//! grammar's default start symbol
NTLink
WMTG::defaultRootLabel() const {
  return NTLink(arity(),Nonterminal(Root));
}

//! grammar's start symbol
NTLink
WMTG::rootLabel() const {
  return m_rootLab;
}

void WMTG::setRootLabel(const NTLink& lab) {
  m_rootLab = lab;
}

//! Returns the fanout limit
short WMTG::fanoutLimit() const {
  return m_fanoutLimit;
}

bool WMTG::inGCNF() const {
  return m_inGCNF;
}

bool WMTG::isHeaded() const {
  return m_isHeaded;
}

/** Clears the content (only) of the containers. */
void WMTG::clear() {
  // read the list of names of output synCatVcb files
  arity_t d;
  vector<oFile*> outsynCatVcbF;
  if(m_ctlNTVcbFilesOut != ""){
    vector<string> fileList = split(m_ctlNTVcbFilesOut,"|");
    vector<string>::iterator itFileList;
    for(itFileList = fileList.begin(); itFileList != fileList.end();
	++itFileList) {
      outsynCatVcbF.push_back(new oFile(itFileList->c_str()));
    }
  }
  // output the new synCatVcbs.
  const StringVocabulary* synCatVcb = ntSet();
  for(d = 0; d < outsynCatVcbF.size(); ++d){
    // output the synCatVcb
    *outsynCatVcbF[d]<<*synCatVcb;
    outsynCatVcbF[d]->close();
    delete outsynCatVcbF[d];
  }

  // clear terminalSet and ntSet *after* dumping them
  _super::clear();
}

Hedge* WMTG::makeDummyConseq(const Hedge* pI1, const Hedge* pI2) {
  
  NTLink label(NTLink::dummyNTLink(m_arity));
  if (m_isHeaded) {
    // add the 1st antecedent's lexical heads
    label = NTLink(label, (pI1->getNTLink().getTLink()));
  }
  const DSV* pDSV1 = pI1->getDSV();
  const DSV* pDSV2 = pI2->getDSV();
  const DSV& theDSV(*pDSV1 + *pDSV2);
  Hedge * pConseq = new Hedge(label, theDSV);
  pConseq->setModelScore(m_smoothingCost);
  return pConseq;
}

HeirRoleVector WMTG::makeDummyHRV(arity_t ar) {
  // a vector of all 1's
  return HeirRoleVector(ar, 1);
}

/// This is totally hardwired for 2D
Hedge* WMTG::makePretermInference(const Hedge* pH1, const Hedge* pH2) {
  // the PAV has just one role per component
  bool theRole(true);
  PrecArray thePA1(1, 1); // one copy of a 1
  PrecArray thePA2(1, 2); // one copy of a 2
  Vector<PrecArray> prePAV;
  
  prePAV.push_back(thePA1);
  prePAV.push_back(thePA2);

  PAV thePAV(prePAV);
  Pair<bool, PAV> arment(theRole, thePAV);
  // the conseq label is a concat of the antecedent labels
  NTLink theLabel = LabelingStrategy::BottomUpConcat(pH1->getNTLink(), pH2->getNTLink(), 
						     theRole, thePAV);
  // the conseq DSV will be a func of the ant DSVs as usual
  DSV theDSV(*(pH1->getDSV()) + *(pH2->getDSV()));
  Hedge* pConseq = new Hedge(theLabel, theDSV);
  
  Compose* theInf;
  if (m_isHeaded) {
    // the HRV has a 1 in comp 1 and a 2 in comp 2
    Vector<short> hrv(2);
    hrv[0] = 1;
    hrv[1] = 2;
    // the label should be lexicalized
    TLink theTLink(2);
    theTLink[0] = pH1->getNTLink().getTLink()[0];
    theTLink[1] = pH2->getNTLink().getTLink()[1];
    theLabel = NTLink(theLabel, theTLink);
    pConseq->setLabel(theLabel);
    // the Compose will be headed if the WMTG is
    theInf = new HeadCompose(pH1, pH2, pConseq, arment, hrv);
  } else {
    theInf = new Compose(pH1, pH2, pConseq, arment);
  }
  vector<size_t> sizes(add_uneven_vec(pH1->getSizes(), pH2->getSizes()));
  pConseq->setSizes(sizes);
  FanoutVect fov( thePAV.getFanoutV());
  pConseq->setFanoutV(fov);
  pConseq->setModelScore(m_smoothingCost);
  static SCORE inCost, auxCost;
  switch (m_costAggregator) {
  case MAXSUM:
    inCost = max(pH1->getInsideScore(), pH2->getInsideScore(),
		 pConseq->getModelScore());
    pConseq->setInsideScore(inCost);
    auxCost = pH1->getAuxScore() + pH2->getAuxScore() 
      + pConseq->getModelScore();
    pConseq->setAuxScore(auxCost);
    break;
  case MAX:
    inCost = max(pH1->getInsideScore(), pH2->getInsideScore(),
		 pConseq->getModelScore());
    pConseq->setInsideScore(inCost);
    break;
  default: // SUM
    inCost = pH1->getInsideScore() + pH2->getInsideScore() 
      + pConseq->getModelScore();
    pConseq->setInsideScore(inCost);
    break;
  }

  return pConseq;
}

/// non-virtual wrapper for getPossibleCons
void WMTG::getBinaryCons(const Item* heir, const Item* depend,
			 const Pair<bool,PAV>& arment, vector<Item*> & ret) {
  getPossibleCons(heir, depend, arment, ret);

  const PAV& thePAV = arment.second;
  // if this is one of the smoothing PAVs and no possible consequents
  // were found, then we add a dummy consequent
  if (m_smoothingPAVs.contains(thePAV) && ret.empty()) {
    const Hedge * heirHedge = static_cast<const Hedge*>(heir);
    const Hedge * depHedge = static_cast<const Hedge*>(depend);
    // To avoid spurious ambiguity, check that the ant on the left in
    // 0th component is NOT a dummy.
    NTLink testNTL;
    const DSpan& span1 = heir->getDSV()->operator[](0);
    const DSpan& span2 = depend->getDSV()->operator[](0);
    if (span1.to(span2) == DSpan::LEFT)
      testNTL = heirHedge->getNTLink();
    else
      testNTL = depHedge->getNTLink();
    if (not testNTL.isDummy())
      ret.push_back(makeDummyConseq(heirHedge, depHedge));
  }


};

/// non-virtual wrapper for getPossibleConsHeaded
void WMTG::getHeadedBinaryCons(const Item* heir, const Item* depend,
			       const Pair<bool,PAV>& arment,
			       vector<HRVIpair > & ret) {
  getPossibleConsHeaded(heir, depend, arment, ret);

  const PAV& thePAV = arment.second;
  // if this is one of the smoothing PAVs and no possible consequents
  // were found, then we add a dummy consequent
  if (m_smoothingPAVs.contains(thePAV) && ret.empty()) {
    const Hedge * heirHedge = static_cast<const Hedge*>(heir);
    const Hedge * depHedge = static_cast<const Hedge*>(depend);
    // To avoid spurious ambiguity, check that the ant on the left in
    // 0th component is NOT a dummy.
    NTLink testNTL;
    const DSpan& span1 = heir->getDSV()->operator[](0);
    const DSpan& span2 = depend->getDSV()->operator[](0);
    if (span1.to(span2) == DSpan::LEFT)
      testNTL = heirHedge->getNTLink();
    else
      testNTL = depHedge->getNTLink();
    if (not testNTL.isDummy()) {
      const HeirRoleVector& theHRV(makeDummyHRV(m_arity));
      Hedge* pConseq(makeDummyConseq(heirHedge, depHedge));
      ret.push_back( HRVIpair(theHRV, pConseq));
    }
  }

 };

/// non-virtual wrapper for getPossibleConsForTLink
void WMTG::getUnaryCons(const TLink& tl, const DSV& theDSV, vector<Item*>& ret) const {
  getPossibleConsForTLink(tl, theDSV, ret);

  if (ret.empty())
    return;

  //search for active dimension and compute sizes and fanout
  arity_t dim = tl.size();
  arity_t activeDim;
  vector<size_t> sizes(dim);
  FanoutVect fov(dim);
  for (unsigned int i = 0; i < dim; i++){
    if(tl[i] == NULLT){
      fov[i] = 0;
      sizes[i] = 0;
    } else {
      activeDim = i;
      fov[i] = 1;
      sizes[i] = 1;
    }
  }

  vector<Item*>::iterator vii;
  for(vii= ret.begin(); vii != ret.end(); ++vii) {
    // set sizes, fanout, and cost
    Item* pItem = *vii;
    pItem->setSizes(sizes);
    pItem->setFanoutV(fov);
    // no antecedents, so just the cost of the inference itself
    pItem->setInsideScore(pItem->getModelScore());
    if (m_costAggregator == MAXSUM)
      pItem->setAuxScore(pItem->getModelScore());
  }
  return;
};

size_t WMTG::tree2features( const HeadMTree& tr, vector<BSP<Production> >& ret) const {
  // do NOT clear ret, in case the caller wants to loop over a bunch of trees
  //  ret.clear();
  size_t sz = 0;
  HeadMTree::iterator mtrit;
  for(mtrit = tr.begin(); mtrit != tr.end(); ++mtrit)
    sz += treeNode2features( mtrit, ret);
  return sz;
};

size_t WMTG::treeNode2features( const HeadMTree::iterator& inNode, 
				vector<BSP<Production> >& ret) const {
  // do NOT clear ret, in case the caller wants to loop over a bunch of trees
  // ret.clear();
  size_t sz = 0;
  if(boost::get<NTLink>(&BGET_EITHER_LABEL(*inNode))){
    ++sz;
    if (m_isHeaded) {
      BSP<Production> spProd(new HeadProduction(inNode));
      if (debug(3))
	cerr << *spProd << endl;
      ret.push_back(spProd);	
    } else {
      BSP<Production> spProd(new Production(inNode));
      if (debug(3))
	cerr << *spProd << endl;
      ret.push_back(spProd);
    }
  }
  return sz;
};

BSP<Production> WMTG::getFeatureBag(const Item* pI) const {
  const Hedge* pH = dynamic_cast<const Hedge*>(pI);
  if (not pH) 
    THROW("Not a hedge; dunno what to do with it.");
  
  BSP<Production> ret;
  const NTLink& lhs = pH->getNTLink();
  BSP<Inference> pInf = pI->getInference();
  if (BSP<Scan> pScan = dynamic_pointer_cast<Scan>(pInf)) {
    const FanoutVect& fov(pI->getFanoutV());
    const PAV thePAV(fov);
    const TLink& tl(pScan->getTLink());
    const ProdRHSElem& elem = tl;
    const Vector<ProdRHSElem> rhs(1, elem);
    if (m_isHeaded) {
      HeirRoleVector hrv(pI->arity(),0);
      const arity_t activeDim = tl.firstActiveComponent();
      hrv.setRole(activeDim,1);
      ret.reset(new HeadProduction(lhs, rhs, thePAV, hrv));
    } else {
      ret.reset(new Production(lhs, rhs, thePAV));
    }
  } else if (BSP<Compose> pComp = dynamic_pointer_cast<Compose>(pInf)) {
    const PAV& thePAV(pComp->getPAV());
    const vector<const Item*>& ants(pComp->getAnt());
    Vector<ProdRHSElem> rhs(2);
    rhs[0] = static_cast<const Hedge*>(ants[0])->getNTLink();
    rhs[1] = static_cast<const Hedge*>(ants[1])->getNTLink();
    if (BSP<HeadCompose> pHComp = dynamic_pointer_cast<HeadCompose>(pComp)) {
      const HeirRoleVector& hrv(pHComp->heirIndices());
      ret.reset(new HeadProduction(lhs, rhs, thePAV, hrv));
    } else {
      ret.reset(new Production(lhs, rhs, thePAV));
    }
  } else
    THROW("Unknown inference type.");

  return ret;
}

// Returns the cost (such as negative log prob) of an inference
// by default, use the corresponding production rule
SCORE WMTG::getInfScore(Inference* pInf) const {
  Production* pProd;
  SCORE ret;
  if (m_isHeaded){
    
    pProd = new HeadProduction(pInf);
    HeirRoleVector hr = (dynamic_cast<HeadProduction*>(pProd))->heirRole();
    
    vector<HeirRoleVector> vec, newVec;
    vec.push_back(hr);

    //make copies letting all ambiguous 0s be a 1 or 2.
    for(unsigned int i = 0; i < hr.arity(); i++){
      //if it's a 0, but it has a size
      if(hr.getRole(i) == 0 && pInf->getConseq()->getSizes()[i] != 0){
	vector<HeirRoleVector>  newVec;
	for(unsigned int j = 0; j < vec.size(); j++){
	  //create two versions of each vector, replacing the 0s.
	  vec[j].setRole(i,1);
	  newVec.push_back(vec[j]);
	  vec[j].setRole(i,2);
	  newVec.push_back(vec[j]);
	}
	vec = newVec;
      }
    }
    
    //check to see which is the best of all replacements
    SCORE best = POSINF;
    HeirRoleVector bestHRV;
    for(unsigned int i = 0; i < vec.size(); i++){
      static_cast<HeadProduction*>(pProd)->setHeirRole(vec[i]);
      SCORE s =  getProdScore(pProd);
      //if we found a better one
      if(s <= best){
	best = s;
	bestHRV = vec[i];
      }
    }
    
    static_cast<HeadCompose*>(pInf)->setHeirIndices(bestHRV);
    delete pProd;
    ret = best;
  }
  else{
    pProd = new Production(pInf);
    ret =  getProdScore(pProd);
    delete pProd;
  }
  
  return ret;
};

void WMTG::getInferences(const Item* trigger, 
			 const Item* matcher,
			 const Pair<bool,PAV>& arrangement,
			 vector<Inference*>& vecInf){
  PAV thePAV = arrangement.second;
  if(m_isHeaded){
    vector<HRVIpair > possConsequents;
    // Get possible consequents (parents) for trigger and matcher
    getHeadedBinaryCons((Item*)trigger, matcher, arrangement, possConsequents);
    
    if (not  possConsequents.empty()) {
      vector<size_t> sizes(add_uneven_vec(trigger->getSizes(), matcher->getSizes()));
      FanoutVect fov( thePAV.getFanoutV());
      
      static SCORE inCost, auxCost;
      
      // For each possible consequent
      vector<pair<HeirRoleVector,Item*> >::iterator iterPa;
      for(iterPa = possConsequents.begin(); iterPa != possConsequents.end(); iterPa++){
	
	Hedge* conseq = static_cast<Hedge*>(iterPa->second);
	
	// compose a new inference
	Inference* pComp = new HeadCompose(trigger, matcher, conseq, 
					     arrangement, iterPa->first); 
	//set its attributes including model score
	conseq->setSizes(sizes);
	conseq->setFanoutV(fov);
	conseq->setModelScore(getInfScore(pComp));
	 
	//set its inside score
	switch (m_costAggregator) {
	case MAXSUM:
	  inCost = max(trigger->getInsideScore(), matcher->getInsideScore(),
		       conseq->getModelScore());
	  conseq->setInsideScore(inCost);
	  auxCost = trigger->getAuxScore() + matcher->getAuxScore() 
	    + conseq->getModelScore();
	  conseq->setAuxScore(auxCost);
	  break;
	case MAX:
	  inCost = max(trigger->getInsideScore(), matcher->getInsideScore(),
		       conseq->getModelScore());
	  conseq->setInsideScore(inCost);
	  break;
	default: // SUM
	  inCost = trigger->getInsideScore() + matcher->getInsideScore() 
	    + conseq->getModelScore();
	  conseq->setInsideScore(inCost);
	  break;
	}
	//push if on the back of the vector to be returned
	if (debug(5))
	  cerr << "WMTG::getInfs made new HeadCompose: " << *pComp << endl;
	vecInf.push_back(pComp);
      }
    }
  } else {
    vector<Item*> possConsequents;
    // Get possible consequents (parents) for trigger and matcher
    getBinaryCons(trigger, matcher, arrangement, possConsequents);
    
    // For each possible consequent
    vector<Item*>::iterator iterPa;
    for(iterPa = possConsequents.begin(); iterPa != possConsequents.end(); iterPa++){
      Hedge* conseq = static_cast<Hedge*>(*iterPa);
         
      // compose a new inference
      Inference* pComp = new Compose(trigger, matcher, conseq, arrangement); 
      conseq->setModelScore(getInfScore(pComp));
      vecInf.push_back(pComp);
    }
  }
}

BSP<BPO::options_description> 
WMTG::addOptionDescriptionsTo(BSP<BPO::options_description> optionsDescription){
  optionsDescription->add_options()
    ("D", BPO::value<unsigned int>(), "The dimension of the grammar")
    ("FanOut", BPO::value<unsigned int>(), "Grammar fanout")
    ("GCNF", BPO::value<bool>()->default_value(true), "Grammar is in GCNF format")
    ("Headed", BPO::value<bool>()->default_value(false), "Grammar is headed")
    ("PAV", BPO::value<string>(), "PAV Vocabulary file")
    ("smoothPAV", BPO::value<string>(), "file of smoothing PAVs")
    ("smoothCost", BPO::value<float>(), "cost of items generated by smoothing")
    ("oN", BPO::value<string>(), "Control file for output nonterminal vocabs")
    ("debug", BPO::value<unsigned int>()->default_value(0), "The debug level.")
    ("smoothScans", BPO::value<bool>(),"True is scans are smoothed")
    ("ignoreHeads", BPO::value<bool>(),"True if we are ignoring head restrictions.")
    ("costAggregator", BPO::value<string>()->default_value("SUM"), 
     "operator for deriving inference costs: SUM, MAX or MAXSUM")
    ;

  return optionsDescription;
}
