#ifndef _WMTG_H_
#define _WMTG_H_

#include <boost/serialization/serialization.hpp>
#include <boost/serialization/is_abstract.hpp>
#include <iterator>
#include <ext/hash_map>
#include <set>

#include "OptionsPrinter.hpp"

//#include "MultiMap.hpp"
#include "Histogram.hpp"
#include "Set.hpp"
#include "WVector.hpp"
#include "StringVocabulary.hpp"

#include "GlobalVocab.hpp"
#include "Production.hpp"
#include "Hedge.hpp"
#include "Grammar.hpp"
#include "Item.hpp"
#include "PAV.hpp"

class HeadMTree;

class WMTG : virtual public Grammar {
  typedef Grammar _super;

  BSP<BPO::options_description> 
  addOptionDescriptionsTo(BSP<BPO::options_description> priorOptions);

protected:
  // for serialization
  WMTG():m_isHeaded(true) {};

public:

  typedef pair<HeirRoleVector,Item*> HRVIpair;

  /// constructors -- even though this is an abstract class,
  /// constructors are necessary for initializing the data members
  WMTG(arity_t dim, 
       bool inGCNF = false, 
       bool isHeaded = true);

  //! constructor
  WMTG(arity_t dm, const string& configFile,
       BSP<BPO::options_description> optionsDescription);

  /// Destructor.
  virtual ~WMTG();

  /// assignment
  WMTG& operator=(const WMTG& other);

  /// fanout limit
  virtual short fanoutLimit() const;
  virtual short fanout() const {return fanoutLimit();}

  /// returns true if the grammar is in GCNF
  bool inGCNF() const;
  /// returns true if the grammar is headed
  bool isHeaded() const;

  /// Binarizes all the productions.
  virtual void binarize(vector<StringVocabulary*>&, char* binarizationMethod = "all") {
    THROW("Not Implemented");
  }

  void insertPAV(const PAV& pav);
  void insertPAV(Vocabulary<PAV>& pavVcb);

  /// Serialization
  template<class Archive>
  void serialize(Archive & ar, unsigned int version) {
    ar & boost::serialization::base_object<_super>(*this);
    ar & m_rootLab;
    ar & m_fanoutLimit;
    ar & m_inGCNF;
    ar & m_ctlNTVcbFilesOut;
  };

  //! The cost associated with having this item as a root.
  virtual SCORE rootCost(const Item*)const{return 0;}

  /// break the tree down into a vector of prod rules
  /// return the number of prods in the vector
  size_t tree2features( const HeadMTree& inTree, vector<BSP<Production> >& ret) const;

  /// break a tree node down into a vector of features
  /// return the number of features so generated
  size_t treeNode2features(const HeadMTree::iterator& inNode, 
			   vector<BSP<Production> >& ret) const;

  /// returns the feature bag pertaining to this item.  In a WMTG, the
  /// feature bag contains just one thing -- the production rule,
  /// which can be headed or not.
  BSP<Production> getFeatureBag(const Item* pI) const;

  /// Returns the cost (such as negative log prob) of an inference
  /// by default, use the corresponding production rule
  virtual SCORE getInfScore(Inference*) const;

  /// non-virtual wrapper for getPossibleCons
  void getBinaryCons(const Item* heir, const Item* depend,
		     const Pair<bool,PAV>& arment, vector<Item*> & parents);

  /// non-virtual wrapper for getPossibleConsHeaded
  void getHeadedBinaryCons(const Item* heir, const Item* depend,
			   const Pair<bool,PAV>& arment,
			   vector<HRVIpair>& parents);

  /// non-virtual wrapper for getPossibleConsForTLink
  void getUnaryCons(const TLink& theTL, const DSV& theDSV, vector<Item*>& vItem) const;

  /// Returns the possible arrangements
  /// When the role in the PAV is true, the trigger is represented by a 1 in the PAV.
  /// When the role is false, the trigger is represented by a 2 in the PAV. 
  virtual void getArrangementFor(BSP<set<Pair<bool,PAV> > >& arrgs, const Item* trigger) const;

  /// Returns the start symbol of the grammar.
  virtual NTLink defaultRootLabel() const;

  virtual NTLink rootLabel() const;
  /** sets the root label*/
  virtual void setRootLabel(const NTLink& lab);

  /** Clears the content (only) of the containers. */
  virtual void clear();

  /** given two items, gets the conqequent and returns an inference */
  virtual void getInferences(const Item*, const Item*, const Pair<bool, PAV>&, vector<Inference*>&);

  /// get features of a consequent
  /// by default there aren't any, but subclasses might override
  virtual void getFeats(Hedge* pConseq, WVector<GlobalVocab::FeatureType>& featVec) {};

  /// make an item whose span is the sum of the antecedent spans, and
  /// whose label is the Dummy NT
  Hedge* makeDummyConseq(const Hedge* pI1, const Hedge* pI2);
  /// make an HRV of all 1's
  HeirRoleVector makeDummyHRV(arity_t ar);
  /// assuming that pH1 is 1D in component 1 and pH2 is 1D in component
  /// 2, compose them and return the parent
  Hedge* makePretermInference(const Hedge* pH1, const Hedge* pH2);

protected: //methods

  /// Returns the cost (such as negative log prob) of a MTG production.
  virtual SCORE getProdScore(Production*) const {
    THROW("WMTG::getProdScore not implemented");
  };

  /// Returns all Z's given X and Y such that Z->XY (or Z->YX).
  /// Note that if there is a grammar index, the returned
  /// result will be retrieved from the index. Otherwise,
  /// the result will be the obtained based on the nt sets.
  /// \param heir : the heir constituent.
  /// \param depend : the dependent constituent.
  /// \param parents : the parent constituents that will be returned.
  virtual void getPossibleCons(const Item* heir, const Item* depend,
			       const Pair<bool,PAV>& arment,
			       vector<Item*> & parents){
    THROW("WMTG::getPossibleCons not implemented");
  }
  
  /// Returns all Z's given X and Y such that Z->XY (or Z->YX).
  /// Note that if there is a grammar index, the returned
  /// result will be retrieved from the index. Otherwise,
  /// the result will be the obtained based on the nt sets.
  /// \param heir : the heir constituent.
  /// \param depend : the dependent constituent.
  /// \param parents : the parent constituents that will be returned and heir info
  virtual void getPossibleConsHeaded(const Item* heir, const Item* depend,
				     const Pair<bool,PAV>& arment,
				     vector<HRVIpair > & parents){
    THROW("WMTG::getPossibleConsHeaded not implemented");
  } 
  
  virtual void getPossibleConsForTLink(const TLink& t, const DSV& d, vector<Item*>&) const = 0;

protected: // data

  /// Fan-out limit
  short m_fanoutLimit;

  /// root label
  NTLink m_rootLab;

  /// tells if grammar is in GCNF
  bool m_inGCNF;

  /// Most WMTG subclasses have some way of modifying their own
  //NTsets. It is usually useful to keep track of these modifications
  //after the grammar object is gone. For this purpose, we use a file
  //of filenames where the grammar's NTsets can be dumped.
  string m_ctlNTVcbFilesOut;

  //true if grammar is headed
  bool m_isHeaded;

  //! The Vocabulary of possible legal (under the fanout limit) PAVs 
  Vocabulary<PAV> m_pavVcb;

  //! The Vocabulary of smoothing PAVs -- i.e. PAVs that are always allowed
  Vocabulary<PAV> m_smoothingPAVs;
  /// the cost of items generated by smoothing
  float m_smoothingCost;
  
  //! true if we want to allow unseen scans to have a dummy POS
  bool m_smoothScans;

  //! true if heads are being ignored and passed up randomly. ( They will
  // not play a roll in an alignment.)
  bool m_ignoreHeads;

  /// A map from a fanout vector to all PAVs with a role having the same fanout vector
  typedef set<Pair<bool,PAV> > arnmtSet_t; 
  typedef hash_map<FanoutVect, arnmtSet_t > fov2arnmt_t;
  fov2arnmt_t m_fov2arnmt;

  /// How are inference costs derived?
  enum costAggregationOperator {SUM, MAX, MAXSUM};
  costAggregationOperator m_costAggregator;

};

BOOST_IS_ABSTRACT(WMTG);

#endif
