#include "File.hpp"
#include "Path.hpp"
#include "SentenceTuple.hpp"
#include "GlobalVocab.hpp"
#include "MaxIterationsTester.hpp"
#include <stdexcept>
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

#include "DiscrimLearner.hpp"
#include "GrammarBuilder.hpp"

using namespace std;

/*
 * Command line
 */
string sntFileName;
int debugLevel = 0;
// grammarBuilder config file
string optLGBuilderConfigFile;
// the config file for the Learner class.
string optDLConfigFile;
string mainConfigFile;
string tbPrefix, modelPrefix;
bool debugWithLinenumbers = false;
unsigned int maxIter = 0;

int main(int argc, char** argv) {
  try{

    BPO::options_description generic("Generic options");
    generic.add_options()
      ("config-file,c", BPO::value<string>(), "config file")
      ("help,h", "help message")
      ;

    // Declare a group of options that will be allowed both on command
    // line and in config file
    BPO::options_description config("Configuration");
    config.add_options()
      ("sentences,s", BPO::value<string>(&sntFileName), 
       "Sentence tuple file name")
      ("TreeBankFilePrefix", BPO::value<string>(&tbPrefix), "")
      ("ModelFilePrefix", BPO::value<string>(&modelPrefix), "")
      ("dl-config,e", BPO::value<string>(&optDLConfigFile),
       "The config file for the discrimLearner class")
      ("LearningGrammarBuilderConfigFile,g", BPO::value<string>(&optLGBuilderConfigFile),
       "The learning grammarBuilder config file")
      ("MaxIterations", BPO::value<unsigned int>(&maxIter), "")
      ("debug,d", BPO::value<int>(&debugLevel)->default_value(0), 
       "Debug level")
      ("debug-with-linenumbers", BPO::value<bool>(&debugWithLinenumbers)->default_value(false), 
       "Append file names and line numbers to debug output")
      ("noVocabsIn,v", BPO::value<bool>(&GlobalVocab::s_noVocabsIn)->default_value(false), 
       "If true, we are not using integerized input")
      ("noVocabsOut,o", BPO::value<bool>(&GlobalVocab::s_noVocabsOut)->default_value(false), 
       "It true, we are not using integerized output")
      ;

    BPO::options_description cmdline_options;
    cmdline_options.add(generic).add(config); // .add(hidden);

    BPO::options_description config_file_options;
    config_file_options.add(config); // .add(hidden);

    BPO::options_description visible("Allowed options");
    visible.add(generic).add(config);

    BPO::variables_map vm;

    store(BPO::command_line_parser(argc, argv).
          options(cmdline_options).run(), vm);

    if(vm.count("config-file")){
      mainConfigFile = vm["config-file"].as<string>();
      cerr << "discrimTrain: Compiling config file " << mainConfigFile << endl;
      iFile ifs(mainConfigFile.c_str());
      store(parse_config_file(ifs, config_file_options), vm);
    }

    notify(vm);

    if (vm.count("help")) {
      cerr << "discrimTrain: the training program" << endl;
      cerr << "Usage: discrimTrain [options]" << endl << endl;;
      cerr << visible << "\n";
      return EXIT_SUCCESS;
    }

    if(debugLevel >= 1){
      cerr << "Option values of \"discrimTrain\"" << endl;
      cerr << OptionsPrinter(config, vm) << endl;
    }

    // validation
    if(argc == 1){
      cerr << "Wrong number of arguments. Try --help" << endl;
      exit(EXIT_FAILURE);
    }

    if(sntFileName == ""){
      cerr << "discrimtrain: Option --sentences is mandatory. Try --help" << endl;
      exit(EXIT_FAILURE);
    }

    if(optLGBuilderConfigFile == ""){
      cerr << "Option --LearningGrammarBuilderConfig needed. Try --help" << endl;
      exit(EXIT_FAILURE);
    }
    // end: validation

    Dbug debugObj;
    Dbug::setDebugWithLineInfo(debugWithLinenumbers);
    debugObj.debugall(debugLevel);

    iFile in_snt(sntFileName.c_str());

    // create the initial grammar
    const string learningGrammarBuilderConfigFile = 
      Path::maybe_compose(mainConfigFile, optLGBuilderConfigFile);
    Grammar* learningGrammar = 
      GrammarBuilder::Instance()->build(learningGrammarBuilderConfigFile);

    // create the learner object
    string dlConfigFile = Path::maybe_compose(mainConfigFile, optDLConfigFile);
    DiscrimLearner dl(dlConfigFile, learningGrammar);

    ConvergenceTester<SCORE>* pConvtester = new MaxIterationsTester(maxIter);

    // I/O streams
    iFile inData(sntFileName);
    oFile* pTreeStream = NULL;

    SentenceTuple sntT;
    WTreeVec parses;
    WTreeVec::iterator iSubtrees;
    unsigned int iterCounter = 0;

    // Behold The Main Loop
    do {
      if(debugLevel >= 0) {
	cerr<<"discrimTrain starting pass # " << ++iterCounter
	    << " thru the data." << endl;
      }
     
      // prepare to dump trees
      if(tbPrefix != "") {
	char outTreebankFile[100];
	sprintf(outTreebankFile, "%s_%d",tbPrefix.c_str(), iterCounter);
	pTreeStream = new oFile(outTreebankFile);
      }

      // rewind, etc.
      dl.clear();

      unsigned int sntCounter = 0;
      rewind(inData);
      while(inData >> sntT) {
	// the input sentence tuple.
	if(debugLevel >= 0) {
	  cerr<<"Learning from sentence tuple No. " << ++sntCounter 
	      << " of lengths " << sntT.lengths() << endl;
	  cerr<<sntT;
	}

	dl.initialize(sntT);
      
	if (debugLevel >= 4)
	  GlobalVocab::put(cerr);

	dl.learn();

	if(pTreeStream) {
	  parses = dl.exportParses();
	  for(iSubtrees = parses.begin(); iSubtrees != parses.end(); ++iSubtrees){
	    *pTreeStream << *iSubtrees->first << "\t";
	  }
	  *pTreeStream << endl;
	  // clean up
	  for(iSubtrees = parses.begin(); iSubtrees != parses.end(); ++iSubtrees){
	    delete iSubtrees->first;
	  }
	}
      }
      pTreeStream->close();

      // dump the learned model
      if(modelPrefix != ""){
	cerr << "Dumping model." << endl;
	char outModelFile[100];
	sprintf(outModelFile, "%s_%d", modelPrefix.c_str(), iterCounter);
	ofstream modelStream(outModelFile);
	modelStream << *learningGrammar;
      } else
	// just dump to standard out
	cout << *learningGrammar;
      
    } while(! pConvtester->satisfied(iterCounter) );

    inData.close();

    // clean up
//     cerr << "Cleaning up." << endl;
//     delete learningGrammar;

  } catch(exception& e){
    cerr << "discrimTrain exception: " << endl << e.what() << endl;
    exit(EXIT_FAILURE);
  }
}
