#include <stdexcept>
#include <iostream>
#include "File.hpp"

#include "MoreStrings.hpp"

#include "Parser.hpp"
#include "GrammarBuilder.hpp"
#include "GlobalVocab.hpp"

#include "StringVocabulary.hpp"
#include "SentenceTuple.hpp"
#include "HeadMTree.hpp"
#include "Exceptions.hpp"
#include "Dbug.hpp"
#include "MeasureMemory.hpp"
//#include "MazMemPool.hpp"
#include "Path.hpp"
#include <stdio.h>

#ifdef BOOST_MEM_POOL
#include "boost/pool/pool.hpp"
#include "boost/pool/object_pool.hpp"
#include "boost/pool/pool_alloc.hpp"
#include "BoostMemPool.hpp"
#endif

#include "boost/filesystem/path.hpp"
#include "boost/filesystem/operations.hpp"

#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

using namespace std;

/*
 * Command line: "opt" means command line options.
 */

string mainConfigFile;

/* THE INPUTS. */
// the sentence tuples.
string optTupleFile;

// the file to dump the parsed treebank
string optTreebank;

// the file to dump the feature vocabulary
string optFeatVocab;

// the config file of the genpar.
string optParserConfigFile;

// the config file of the grammar.
string optGrammarBuilderConfigFile;

string optRunFile;

// the debugging level.
int debug;

bool debugWithLinenumbers = false;

bool useRhsCompletionIndex = false;

bool resumeParsing = false;

bool runMany = false;
int runOne = 0;

int main(int argc, char** argv) {
    // uncomment to set output precision
    // cerr.setf(ios::fixed, ios::floatfield);
    // cerr.precision(40);
    BPO::options_description generic("Generic options");
    generic.add_options()
      ("config-file,c", BPO::value<string>(), "config file")
      ("help,h", "help message")
      ;

    // Declare a group of options that will be
    // allowed both on command line and in
    // config file
    BPO::options_description config("Configuration");
    config.add_options()
      ("sentences,s", BPO::value<string>(&optTupleFile), "The sentence tuple file")
      ("treebank,b", BPO::value<string>(&optTreebank), "The file to hold the output treebank")
      ("featVocab", BPO::value<string>(&optFeatVocab), "The file to hold the feature vocabulary")
      ("genpar-config", BPO::value<string>(&optParserConfigFile), "The config file for the Generalized Parser.")
      ("grammarbuilder-config", BPO::value<string>(&optGrammarBuilderConfigFile), "The config file for the GrammarBuilder")
      ("debug,d", BPO::value<int>(&debug)->default_value(0), "debug")
      ("debug-with-linenumbers", BPO::value<bool>(&debugWithLinenumbers)->default_value(false), "Append file names and line numbers to debug output") ("noVocabsIn,v", BPO::value<bool>(&GlobalVocab::s_noVocabsIn)->default_value(false), "If true, we are not using integerized input")
      ("noVocabsOut,o", BPO::value<bool>(&GlobalVocab::s_noVocabsOut)->default_value(false), "If true, we are not using integerized output")
      ("resume,r", BPO::value<bool>(&resumeParsing)->default_value(false), "if true, resume from where it last left off.")
      ("runOne,o", BPO::value<int>(&runOne)->default_value(0), "if has value,only run this sentence. ")
      ("runFile,o", BPO::value<string>(&optRunFile), "a file with a list of lines to run. ")
      ;

    BPO::options_description cmdline_options;
    cmdline_options.add(generic).add(config); // .add(hidden);

    BPO::options_description config_file_options;
    config_file_options.add(config); // .add(hidden);

    BPO::options_description visible("Allowed options");
    visible.add(generic).add(config);

    BPO::variables_map vm;

    // store(BPO::command_line_parser(argc, argv).
    // options(cmdline_options).positional(p).run(), vm);
    store(BPO::command_line_parser(argc, argv).
	  options(cmdline_options).run(), vm);

    if(vm.count("config-file")){
      mainConfigFile = vm["config-file"].as<string>();
      cerr << "gp: Compiling config file " << mainConfigFile << endl;
      iFile ifs(mainConfigFile.c_str());
      store(parse_config_file(ifs, config_file_options), vm);
    }

    notify(vm);

    if (vm.count("help")) {
      cerr << "Usage: gp [options] [< <sentence_tuples>]" << endl << endl;;
      cerr << visible << "\n";
      return 0;
    }

    // VALIDATION
    if(argc == 1){
      // cmdPar.dieUsage(argv[0], cmds, CmdNumber(cmds));
      THROW("Wrong number of arguments. Try --help" << endl);
    }

    if(optGrammarBuilderConfigFile == ""){
      THROW("Option --grammarbuilder-config needed. Try --help" << endl);
    }
    // END VALIDATION

    // set Dbug's static debugAll level
    Dbug::setDebugWithLineInfo(debugWithLinenumbers);
    Dbug debugger;
    debugger.debugall(debug);
    if(debug >= 1){
      cerr << "Option values of \"gp\"" << endl << endl;
      cerr << OptionsPrinter(config, vm) << endl;
    }

    // Construct and load the grammar
    string grammarBuilderConfigFile = Path::maybe_compose(mainConfigFile,
							  optGrammarBuilderConfigFile);
    Grammar* grammar = GrammarBuilder::Instance()->build(grammarBuilderConfigFile);

    // Construct the parser
    string genParConfigFile = Path::maybe_compose(mainConfigFile, optParserConfigFile);
    Parser* pParser = new Parser(genParConfigFile, grammar);

    // for tree dumping
    ostream* tbout;

    //number of trees parsed in prior run
    //(to resume from where we left off)
    int numTreesParsed = 0;

    if(resumeParsing){
      bool tmp = GlobalVocab::s_noVocabsIn;
      GlobalVocab::s_noVocabsIn = GlobalVocab::s_noVocabsOut;

      if(optTreebank != "") {
	istream* tbin = new iFile(optTreebank.c_str());
	HeadMTree hmt;
	string line;
	while(getline(*tbin,line, '\n')) {
	  vector<string> strvec = split(line, "\t");
	  for(unsigned int i = 0; i < strvec.size(); ++i){
	    istringstream ist(strvec[i]);
	    HeadMTree parse;
	    ist >> parse;
	  }
	  numTreesParsed++;
	}
	cerr << "Number of trees already parsed is: " << numTreesParsed << endl;
      }
      GlobalVocab::s_noVocabsIn = tmp;
    }

    queue<unsigned int> linesToRun;
    if(optRunFile != ""){
      if(runOne){
	cerr << "cant use runOne and runFile" << endl;
	exit(1);
      }
      runMany = true;
      istream* iRun = new iFile(optRunFile.c_str());
      int x;	
      while( *iRun >> x){
	linesToRun.push(x);
      }
    }
    
    if(runOne){
      runMany = true;
      linesToRun.push(runOne);
      numTreesParsed = runOne-1;
    }

    if(resumeParsing){
      if(optTreebank != "") {
	tbout = new oFile(optTreebank.c_str(),ios_base::app);
      } else {
	cerr << "Must have an output file name if resume is set to true." << endl;
	exit(1);
      }
    }
    else{
      if(optTreebank != "") {
	tbout = new oFile(optTreebank.c_str());
      } else {
	tbout = &cout;
      }
    }

    // the input sentence tuples
    SentenceTuple sntT;
    istream* ist;
    if(optTupleFile != ""){
      ist = new iFile(optTupleFile);
    } else {
      ist = &cin;
    }

    //cycle through the input tuples that we want to discard since
    //they've already been parsed
    size_t inCounter = 0;
    
    if(linesToRun.size() != 0){
      for(; inCounter < linesToRun.front()-1; inCounter++){
	*ist >>sntT;
	grammar->prepareFor(sntT);
      }
      linesToRun.pop();
    }


    Timer tm;
    WTreeVec parses;
    WTreeVec :: iterator itP;

    // Behold The Main Loop.  This loop takes reads in sentences and 
    // passes them to the parser.
    while(*ist >>sntT) { //read in sentenceTuple.

      if(debug >= 0) {
	cerr<<"Parsing sentence No. " << ++inCounter 
	    << " of lengths " << sntT.lengths() << endl;
      }
      //start the timer (for stats)
      tm.start();

      // THE CENTER OF IT ALL: This does the parsing.
      parses = pParser->getParse(sntT);
      
      // output the trees we got back.
      for(itP = parses.begin(); itP != parses.end(); ++itP){
	  *tbout<<*itP->first<<"\t";
      }
      *tbout<<endl;

      // print scores
      cerr<<"SCORE:\t";
      for(itP = parses.begin(); itP != parses.end(); ++itP)
	cerr<<itP->second<<"\t";
      cerr<<endl;

      // print parse trees
      cerr<<"PARSE TREE:\t";
      for(itP = parses.begin(); itP != parses.end(); ++itP){
	cerr<<*itP->first<<"\t";
      }
      cerr<<endl;

     // clean up, get ready for the next round
      pParser->clear();

      // the following should really be done inside Parser::clear()
      // delete the trees we got back.
      for(itP = parses.begin(); itP != parses.end(); ++itP)
	delete itP->first;

      // clear mem pool
      //      MazMemPool::clear_memory(); // David's pool.
#ifdef BOOST_MEM_POOL
      clear_boost_pools(); // boost pool.
#endif

      // stop timer
      tm.end();
      if (debug >=2)
	cerr<<"PARSING TIME:\t"<<tm.diff()<<endl;
      if(runOne)
	break;
      if(runMany && linesToRun.size() == 0)
	break;
      if(runMany){
	for(; inCounter < linesToRun.front()-1; inCounter++){
	  *ist >>sntT;
	  grammar->prepareFor(sntT);
	}
	linesToRun.pop();
      }
   
    }
    // END OF MAIN LOOP
    cerr<<"PARSING ENDED!\n";

    if (optFeatVocab != "") {
      oFile fvFile(optFeatVocab);
      fvFile << *(GlobalVocab::s_featSet);
    }

    if (debug >= 2) {
      dumpMemoryUsage();
    }

    // inform the grammar that it needs to save state
    grammar->clear();

    // clean up time
    if(optTupleFile != "")
      delete ist;

    if (optTreebank != "")
      delete tbout;

    delete pParser;
    delete grammar;

    return 0;

}

