#include <stdexcept>
#include "Path.hpp"
#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"
#include "MeasureMemory.hpp"
#include "Exceptions.hpp"
#include "File.hpp"
#include "GlobalVocab.hpp"

#include "GrammarBuilder.hpp"
#include "NaiveWMTG.hpp"
#include "LearnableGrammar.hpp"
#include "NegLogProbTester.hpp"
#include "MaxIterationsTester.hpp"
#include "IOUtil.hpp"
#include "Parser.hpp"
#include "DummyParser.hpp"
#include "UpdateRuleBuilder.hpp"
#include "UpdateRule.hpp"
#include "MStep.hpp"
#include "GradientMStep.hpp"
#include "Timer.hpp"
#include "ConvergenceTester.hpp"
#include "ParserBuilder.hpp"
#include "ActiveLearningFilter.hpp"
#include "ActiveFilterBuilder.hpp"

using namespace std;

int main(int argc, char** argv) {

  try{
    /******************* BEGIN CONFIG ***************************************/
    int debugLevel = 0;

    string sntFileName;
    string optInitLGBuilderConfigFile, optLGBuilderConfigFile, optSGBuilderConfigFile;
    string mainConfigFile;
    bool debugWithLinenumbers = false;
    string optInitLPBuilderConfigFile, optLPBuilderConfigFile, optSPBuilderConfigFile;
    string testerType;
    string tb_prefix, model_prefix, rule_prefix;
    string logFile;
    double convTH; 
    string updateRule;
    unsigned int updateFrequency;
    string mStepType;
    string activeLearningFilterType;
    unsigned int grammarDumpInterval;
    
    BPO::options_description generic("Generic options");
    generic.add_options()
      ("config-file,c", BPO::value<string>(), "config file")
      ("help,h", "help message")
      ;

    // Declare a group of options that will be allowed both on command
    // line and in config file
    BPO::options_description config("Configuration");
    config.add_options()
      ("sentences,s", BPO::value<string>(&sntFileName), 
       "Sentence tuple file name")
      ("InitLearningGrammarBuilderConfigFile,i", BPO::value<string>(&optInitLGBuilderConfigFile), 
       "The grammarBuilder config file for the initial learning grammar")
      ("LearningGrammarBuilderConfigFile,g", BPO::value<string>(&optLGBuilderConfigFile), 
       "The grammarBuilder config file for learned grammars")
      ("debug,d", BPO::value<int>(&debugLevel)->default_value(0), 
       "Debug level")
      ("debug-with-linenumbers", BPO::value<bool>(&debugWithLinenumbers)->default_value(false), 
       "Append file names and line numbers to debug output")
      ("noVocabsIn,v", BPO::value<bool>(&GlobalVocab::s_noVocabsIn)->default_value(false), 
       "If true, we are not using integerized input")
      ("noVocabsOut,o", BPO::value<bool>(&GlobalVocab::s_noVocabsOut)->default_value(false), 
       "It true, we are not using integerized output")
      ("LearningParserBuilderConfigFile", BPO::value<string>(&optLPBuilderConfigFile), 
       "The config file for the non-initial parser")
      ("InitParserConfigFile", BPO::value<string>(&optInitLPBuilderConfigFile), 
       "The config file for the initial parser")
      ("TreeBankFilePrefix", BPO::value<string>(&tb_prefix), "")
      ("ModelFilePrefix", BPO::value<string>(&model_prefix), "")
      ("RuleFilePrefix", BPO::value<string>(&rule_prefix), "")
      ("log", BPO::value<string>(&logFile), "")
      ("ConvergenceTesterType", BPO::value<string>(&testerType), "")
      ("ConvergeTH", BPO::value<double>(&convTH), "")
      ("SupervisingParserConfigFile", BPO::value<string>(&optSPBuilderConfigFile), 
       "Configuration file for the parser which will generate the supervising parses")
      ("UpdateFrequency", BPO::value<unsigned int>(&updateFrequency), 
       "The number of examples between updates to our model.")
      ("SupervisingGrammarBuilderConfigFile", BPO::value<string>(&optSGBuilderConfigFile),
       "The configuration file for the supervising grammar.")
      ("UpdateRule", BPO::value<string>(&updateRule), 
       "The type of update to be made to our model.")
      ("MStepType", BPO::value<string>(&mStepType), "")
      ("ActiveLearningFilterType", BPO::value<string>(&activeLearningFilterType), "")
      ("GrammarDumpInterval", BPO::value<unsigned int>(&grammarDumpInterval), "")
      ;

    BPO::options_description cmdline_options;
    cmdline_options.add(generic).add(config); // .add(hidden);

    BPO::options_description config_file_options;
    config_file_options.add(config); // .add(hidden);

    BPO::options_description visible("Allowed options");
    visible.add(generic).add(config);

    BPO::variables_map vm;

    // store(BPO::command_line_parser(argc, argv).
    // options(cmdline_options).positional(p).run(), vm);
    store(BPO::command_line_parser(argc, argv).
          options(cmdline_options).run(), vm);

    if(vm.count("config-file")){
      mainConfigFile = vm["config-file"].as<string>();
      cerr << "treeLearn: Compiling config file " << mainConfigFile << endl;
      iFile ifs(mainConfigFile.c_str());
      store(parse_config_file(ifs, config_file_options), vm);
    }

    notify(vm);

    if (vm.count("help")) {
      cerr << "treeLearn: the training program" << endl;
      cerr << "Usage: treeLearn [options]" << endl << endl;;
      cerr << visible << "\n";
      return EXIT_SUCCESS;
    }
 
    // if no interval has been established, then 
    // do not dump the grammar
    
    if (!vm.count("GrammarDumpInterval")){
      grammarDumpInterval = 0;
    }  

    if(debugLevel >= 1){
      cerr << "Option values of \"treeLearn\"" << endl;
      cerr << OptionsPrinter(config, vm) << endl;
    }

    /******************* END CONFIG ***************************************/

    /******************* BEGIN VALIDATION ************************************/
    if(argc == 1){
      cerr << "Wrong number of arguments. Try --help" << endl;
      exit(EXIT_FAILURE);
    }

    if(sntFileName == ""){
      cerr << "treeLearn: Option --sentences is mandatory. Try --help" << endl;
      exit(EXIT_FAILURE);
    }

    if(optLGBuilderConfigFile == ""){
      cerr << "Option --grammarbuilder-config needed. Try --help" << endl;
      exit(EXIT_FAILURE);
    }

    if (optLPBuilderConfigFile == ""){
      cerr << "no parser config file specified!" << endl;
      exit(EXIT_FAILURE);
    }
    /******************* END VALIDATION ************************************/

    /************************* BEGIN BUILDING ********************************/
    Dbug debugObj;
    Dbug::setDebugWithLineInfo(debugWithLinenumbers);
    debugObj.debugall(debugLevel);

    string initLGBuilderConfigFile, lGBuilderConfigFile, sGBuilderConfigFile;
    string initLPBuilderConfigFile, lPBuilderConfigFile, sPBuilderConfigFile;
    ConvergenceTester<SCORE>* terminationTester(NULL);
    Parser* pLearningParser(NULL);
    AbstractParser* pSupervisoryParser(NULL);
    UpdateRule* pUpdateRule(NULL);
    MaxIterationsTester* pUpdateTester(NULL);
    MStep* pMStep(NULL);
    ActiveLearningFilter* pActiveFilter(NULL);
    char model_file_name[100];
    char treebankFile[100];
    char grammarDumpFile[100];

    // prepare builder config file for the learning grammar
    // initial builder config file defaults to the non-initial file
    lGBuilderConfigFile = Path::maybe_compose(mainConfigFile, optLGBuilderConfigFile);
    if (optInitLGBuilderConfigFile != "")
      initLGBuilderConfigFile = Path::maybe_compose(mainConfigFile, optInitLGBuilderConfigFile);
    else
      initLGBuilderConfigFile = lGBuilderConfigFile;
    
    // create the initial learning grammar
    // Grammar* initialGrammar = GrammarBuilder::Instance()->build(initLGBuilderConfigFile);
    // LearnableGrammar* model = dynamic_cast<LearnableGrammar*>(initialGrammar);
    // if (not model)
    //  THROW("treeLearn requires LearnableGrammar");

    // prepare builder config file for the learning parser
    // initial builder config file defaults to the non-initial file
    lPBuilderConfigFile = Path::maybe_compose(mainConfigFile, optLPBuilderConfigFile);
    if (optInitLPBuilderConfigFile != "") {
      initLPBuilderConfigFile = Path::maybe_compose(mainConfigFile, optInitLPBuilderConfigFile);
    } else {
      initLPBuilderConfigFile = lPBuilderConfigFile;
    }
    
    // create the initial learning parser
    // pLearningParser = new Parser(initLPBuilderConfigFile, model);
    pLearningParser = dynamic_cast<Parser*>(ParserBuilder::Instance()->build(initLPBuilderConfigFile));
    LearnableGrammar* model = dynamic_cast<LearnableGrammar*>(pLearningParser->getGrammar());
    if (optSPBuilderConfigFile != "") {
      string sParserBuilderConfigFile = Path::maybe_compose(mainConfigFile, optSPBuilderConfigFile);
      cerr << "supervising parser config: " << optSPBuilderConfigFile << endl;
      //pSupervisoryParser = new DummyParser(sParserBuilderConfigFile,NULL);
      pSupervisoryParser = ParserBuilder::Instance()->build(sParserBuilderConfigFile);
    }
    
    // current supervisory parser hardwired to be DummyParser, which doesn't need a grammar
//     if (optSGBuilderConfigFile != "") {
//       sGBuilderConfigFile = Path::maybe_compose(mainConfigFile, optSGBuilderConfigFile);
//       Grammar* supervisoryGrammar = GrammarBuilder::Instance()->build(sGBuilderConfigFile);
//       cerr << "grammar builder config: " << optSGBuilderConfigFile << endl;
//     }
    
    // create the update rule, this is dependent on the learning algorithm
    pUpdateRule = UpdateRuleBuilder::Instance()->build(updateRule);
    pUpdateTester = new MaxIterationsTester(updateFrequency);
    
    // create the MStep object
    if (mStepType == "gradient"){
      pMStep = new GradientMStep();
    } else { // default
      pMStep = new MStep();
    }

    LearnableGrammar* gradient = dynamic_cast<LearnableGrammar*>
      (GrammarBuilder::Instance()->build(lGBuilderConfigFile, 
					 model->ntSet(), model->terminalSet()));

    // create the convergenceTester
    if (testerType == "MaxIterations"){
      terminationTester = dynamic_cast<ConvergenceTester<SCORE>*>
	(new MaxIterationsTester(static_cast<unsigned int>(convTH)));
    } else { // default
      terminationTester = new NegLogProbTester(convTH);
    }

    //  default active learning filter is the null filter 
    //if (activeLearningFilterType == ""){
    //  activeLearningFilterType = "NULL";
    //}

    pActiveFilter = ActiveFilterBuilder::Instance()->build(activeLearningFilterType);
  
      // set the treebank file name prefix.
    if(tb_prefix == "") {
      tb_prefix = "tb";
    }

    // set the model file name prefix.
    if(model_prefix == ""){
      model_prefix = "model";
    }

    // set the rule file name prefix.
    if(rule_prefix == ""){
      rule_prefix = "rule";
    }
    
    // open the log file
    ofstream logStream;
    if(logFile != ""){
      logStream.open(logFile.c_str());
    }

    // training data
    iFile data(sntFileName);
    
    SCORE corpusScore;
    Timer timer;

    /************************* END BUILDING ********************************/

    // main loop
    int corpusIter = 0;
    do {

      ++corpusIter;
      if(logFile != "") logStream<<"ITERATION "<<corpusIter<<endl;
      cerr<<"ITERATION "<<corpusIter<<endl;

      if(model_prefix != "")
        sprintf(model_file_name, "%s_%d", model_prefix.c_str(), corpusIter);
      else 
	model_file_name[0] = '\0';
      ofstream modelStream(model_file_name);

      // rewind the data to the beginning.
      rewind(data);

      // prepare to dump the multitrees
      if(tb_prefix != "")
        sprintf(treebankFile, "%s_%d",tb_prefix.c_str(), corpusIter);
      else
        treebankFile[0] = '\0';
      ofstream treeStream(treebankFile);

      // in a probabilistic setting the corpusScore is a neg-log-likelihood
      // otherwise, it can be a cost or margin
      corpusScore = ZEROSCORE;

      SentenceTuple sntT;

      int obsCount = 0;

      if(logStream)
	logStream<<"==============================================="<<endl;

      pUpdateTester->reset();

      // start the timer
      timer.start();

      while( data >> sntT ){
        if (pActiveFilter->filter(sntT)){
	  cerr<<"sentence # " <<++obsCount<<endl;
	  corpusScore += 
	    pUpdateRule->getGradient(pSupervisoryParser, pLearningParser, sntT, gradient, treeStream);
	  if (debugLevel >= 2) {
	    cerr << "old model mass = " << model->mass();
	    cerr << " gradient mass = " << gradient->mass();
	  }
	  // check to see if we need to update the model on this example
	  if (pUpdateTester->satisfied()){
	    pMStep->minimizeCost(model, gradient);
            // if the dumpInterval has passed then print out the grammar
	    if (grammarDumpInterval > 0){
              if (obsCount%grammarDumpInterval == 0){
                sprintf(grammarDumpFile, "%s_%d","GrammarAtExample",obsCount);
		ofstream grammarStream(grammarDumpFile);
                grammarStream << *model;
                grammarStream << endl;
                grammarStream.close();
               
              }
            } 
	  }
	  if (debugLevel >= 2) {
	    cerr << " new model mass = " << model->mass() << endl;
	  }
        }
      }

      // if we are in batch mode then update the model after the corpus loop
      if (pUpdateTester->batch()){
	pMStep->minimizeCost(model, gradient);
      }
      timer.end();

      if(logStream)
	logStream << obsCount << " sentence tuple(s) processed.\n";

      if(treeStream)
	treeStream.close();
    
      if(logFile != "") {
        // formating of floating-point output.
        logStream.setf(ios_base::fixed, ios_base::floatfield);
        logStream<<"log probability at iteration "
	    << corpusIter << " = " << corpusScore << endl;
        logStream<<"time cost for E-Step is "<<timer.diff()<<" sec(s)."<<endl;
      }
      // formatting of floating-point output
      cerr.setf(ios_base::fixed, ios_base::floatfield);
      cerr << "log probability at iteration " << corpusIter << " = " << corpusScore << endl;
      //      cerr << "time cost for E-Step is "<<timer.diff()<<" sec(s)."<<endl;

      /*
       * dump the model
       */
      if(modelStream) {
        cerr<<"Dumping model..."<<endl;
        modelStream<<*model;
        modelStream.close();
      }

      // change the parser, if necessary
      if (corpusIter == 1 && lPBuilderConfigFile != initLPBuilderConfigFile){ 
	// pLearningParser = new Parser(lPBuilderConfigFile, model);
	if(pLearningParser)
	  delete pLearningParser;
        pLearningParser = dynamic_cast<Parser*>(ParserBuilder::Instance()->build(lPBuilderConfigFile));
        
      } // else 
	// update the parser's grammar
	pLearningParser->setGrammar(model);

      // the convergence tester tests whether the iterations
      // will be finished
    } while(! terminationTester->satisfied(corpusScore) );

    cerr.setf(ios_base::fixed, ios_base::floatfield);
    cerr << "log probability at end = " << corpusScore << endl;
    if(logFile != "") {
      logStream.setf(ios_base::fixed, ios_base::floatfield);
      logStream << "log probability at end = " << corpusScore << endl;
    }

    if (debugLevel >= 2)
      dumpMemoryUsage();

    // tell it to save state if it wants to
    model->clear();
    delete terminationTester;
    delete pUpdateRule;
    delete pMStep;
    delete pUpdateTester;
    delete pLearningParser;
    delete pSupervisoryParser;
    // delete initialGrammar;
    delete[] argv;
    
    data.close();

  } catch(exception& e){
    cerr << "treeLearn exception: " << endl << e.what() << endl;
    exit(EXIT_FAILURE);
  }
}
