//! initializes a grammar from a treebank
#ifdef USE_SERIALIZATION
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/xml_oarchive.hpp>
#include "ArchiveMaker.hpp"
#endif

#include "COMPILER-DEPENDENT_HEADERS.hpp"
#include <vector>
#include <string>
#include "File.hpp"
#include "Path.hpp"
#include "Item.hpp"
#include "HeadMTree.hpp"
#include "GlobalVocab.hpp"
#include "GrammarBuilder.hpp"
#include "LearnableGrammar.hpp"
#include "Exceptions.hpp"
#include "MeasureMemory.hpp"
#include "MultinomialDistribution.hpp"

#include <boost/program_options.hpp>
#include "OptionsPrinter.hpp"

#ifdef USE_SERIALIZATION
#include <boost/serialization/serialization.hpp>
#endif

using namespace std;

// the sentence tuples.
string optTupleFile;
// the input multitreebank
string optTreebank;
// the config file of the grammar
string optGrammarBuilderConfigFile;
// output file for the grammar
string optGrammarOutFileName;
string outFileType;
// output file for the inferences
string optInfsOutFileName;
// the file to dump the feature vocabulary
string optFeatVocab;
// the debug level
int debugLevel;

int main(int argc, char** argv) {

  MTreeNode::includingSpans = true;
  static string mainConfigFile;

  BPO::options_description generic("Generic options");
  generic.add_options()
    ("config-file,c", BPO::value<string>(), "config file")
    ("help,h", "help message")
    ;

  // Declare a group of options that will be
  // allowed both on command line and in
  // config file
  BPO::options_description config("Configuration");
  config.add_options()
    ("sentences,s", BPO::value<string>(&optTupleFile), "The sentence tuple file")
    ("treebank,b", BPO::value<string>(&optTreebank), "Input of a multi-treebank")
    ("grammarOutput", BPO::value<string>(&optGrammarOutFileName), "The file to hold the output grammar")
    ("infsOutput", BPO::value<string>(&optInfsOutFileName), "where to dump inferences (if desired)")
    ("featVocab", BPO::value<string>(&optFeatVocab), "The file to hold the feature vocabulary")
    ("fileType", BPO::value<string>(&outFileType), "The output grammar file format")
    ("grammarbuilder-config", BPO::value<string>(&optGrammarBuilderConfigFile), 
     "The config file for the GrammarBuilder")
    ("debug,d", BPO::value<int>(&debugLevel)->default_value(0), "debug")
    ("noVocabsIn,v", BPO::value<bool>(&GlobalVocab::s_noVocabsIn)->default_value(false), 
     "If true, we are not using integerized input")
    ("noVocabsOut,o", BPO::value<bool>(&GlobalVocab::s_noVocabsOut)->default_value(false), 
     "If true, we are not using integerized output")
    ;

  BPO::options_description cmdline_options;
  cmdline_options.add(generic).add(config); // .add(hidden);
  BPO::options_description config_file_options;
  config_file_options.add(config); // .add(hidden);
  BPO::options_description visible("Allowed options");
  visible.add(generic).add(config);
  BPO::variables_map vm;
  store(BPO::command_line_parser(argc, argv).options(cmdline_options).run(), vm);

  if(vm.count("config-file")){
    mainConfigFile = vm["config-file"].as<string>();
    cerr << "trees2grammar: Compiling config file " << mainConfigFile << endl;
    iFile ifs(mainConfigFile.c_str());
    store(parse_config_file(ifs, config_file_options), vm);
  }
  notify(vm);

  if (vm.count("help")) {
    cerr << "initgrammar: Grammar Initializer" << endl << endl;
    cerr << "Usage: initgrammar [options] > <grammar_file>]" << endl << endl;;
    cerr << visible << "\n";
    return 0;
  }

  if(debugLevel >= 1){
    cerr << "Option values of \"initgrammar\"" << endl;
    cerr << OptionsPrinter(config, vm) << endl;
  }

  if(argc == 1){
    // cmdPar.dieUsage(argv[0], cmds, CmdNumber(cmds));
    THROW("Wrong number of arguments. Try --help" << endl);
  }

  Dbug debugger;
  debugger.debugall(debugLevel);

  // the multitreebank
  if(optTreebank == ""){
    THROW("Input of a treebank is required."<<endl);
  }
  iFile mTreebankF(optTreebank);

  istream* sntStream=NULL;
  if(optTupleFile != ""){
    sntStream = new iFile(optTupleFile);
  }else if (optInfsOutFileName != ""){
    THROW("If outputing features, then a sentence tuple file is required. Please set \'s\' in the config file to point to the corrent sentence tuple file"<< endl);
  }
    
  // initialize the grammar to build
  // note: can be any learnable grammar, depending on config file
  string grammarBuilderConfigFile =
    Path::maybe_compose(mainConfigFile, optGrammarBuilderConfigFile);
  LearnableGrammar* grm = dynamic_cast<LearnableGrammar*>
    (GrammarBuilder::Instance()->build(grammarBuilderConfigFile));

  // grammar output
  ostream* grOutStream = NULL;
  if (optGrammarOutFileName != "")
    grOutStream = new oFile(optGrammarOutFileName);
  else if (outFileType == "tar" || outFileType == "bar")
    THROW("Cannot create an archive if no output file is specified");

  // inferences output
  ostream* infOutStream = NULL;
  if (optInfsOutFileName != "")
    infOutStream = new oFile(optInfsOutFileName);

  string line;
  size_t counter = 0;
  Vector<Inference*> infVec;
  Vector<Inference*>::const_iterator vici;
  Vector<HeadMTree*> treeVec;
  Vector<HeadMTree*>::const_iterator tvci;
  SentenceTuple snt;
  while(getline(mTreebankF, line)) {
    ++counter;
    vector<unsigned int> startingPosns(2,0);
    vector<string> strvec = split(line, "\t");
    // if(counter % 1000 == 0){
    cerr <<"Reading multitree no. "<< counter << "\n";
    // }

    //sentence is only needed if we want to output features.
    if(sntStream){
      *sntStream >> snt;
      grm->prepareFor(snt);
    }
    treeVec.clear();
    for(size_t i = 0; i < strvec.size(); ++i){
      istringstream ist(strvec[i]);

      HeadMTree* pTree = new HeadMTree;
      ist>> *pTree;

      if(pTree->begin() == pTree->end())
	continue;

      treeVec.push_back(pTree);
    }
    // prepare to get infs from trees
    if (optInfsOutFileName != "")
      grm->prepareFor(treeVec, startingPosns);

    const Vector<HeadMTree*>::const_iterator sentinel = treeVec.end();
    for(tvci = treeVec.begin(); tvci != sentinel; ++tvci) {
      // if we need the grammar
      if (optGrammarOutFileName != "")
	grm->increment(**tvci);
      // if we need the inferences
      if (optInfsOutFileName != "")
	grm->tree2infs(infVec, **tvci);
    }
    if (optInfsOutFileName != "") {
      // dump all the infs for this set of subtrees
      for(vici = infVec.begin(); vici != infVec.end(); ++vici) {
	*infOutStream << counter << " ";
	*infOutStream << *vici << endl;
	// clean up to avoid memory leaks
	vector<const Item*> ants = (*vici)->getAnt();
	vector<const Item*>::iterator itemIt;
	for(itemIt = ants.begin(); itemIt != ants.end(); ++itemIt)
	  delete *itemIt;	
	delete (*vici)->getConseq(); // also deletes the inference
      }
      infVec.clear();
    }
  }

  // estimate back-off distributions, do indexing, etc.
  grm->estimate();

  // dump feature vocab
  if (optFeatVocab != "") {
    oFile fvFile(optFeatVocab);
    fvFile << *(GlobalVocab::s_featSet);
  }

  // dump grammar
  if (outFileType == "txt") {
    cout << *grm;
    outFileType = "tar"; // write both text output and archive to validate
  }
#ifdef USE_SERIALIZATION
  if (outFileType == "bar") {
    boost::archive::binary_oarchive oa(*grOutStream);
    ArchiveMaker am;
    am.writeBinaryArchive(oa, grm); }
  else if (outFileType == "tar") {
    boost::archive::text_oarchive oa(*grOutStream);
    ArchiveMaker am;
    am.writeTextArchive(oa, grm); }
#endif
  else
    THROW("Unknown output file type for the grammar in trees2grammar");

  // clear() also dumps the revised vocabs
  grm->clear();
  delete grm;

  delete infOutStream;
  delete grOutStream;
  delete sntStream;

}


